!function(t){function e(n){if(r[n])return r[n].exports;var i=r[n]={i:n,l:!1,exports:{}};return t[n].call(i.exports,i,i.exports,e),i.l=!0,i.exports}var n=window.webpackJsonp;window.webpackJsonp=function(r,o,a){for(var s,u,l,c=0,f=[];c<r.length;c++)u=r[c],i[u]&&f.push(i[u][0]),i[u]=0;for(s in o)Object.prototype.hasOwnProperty.call(o,s)&&(t[s]=o[s]);for(n&&n(r,o,a);f.length;)f.shift()();if(a)for(c=0;c<a.length;c++)l=e(e.s=a[c]);return l};var r={},i={1:0};return e.e=function(t){function n(){o.onerror=o.onload=null,clearTimeout(a);var e=i[t];0!==e&&(e&&e[1](new Error("Loading chunk "+t+" failed.")),i[t]=void 0)}if(0===i[t])return Promise.resolve();if(i[t])return i[t][2];var r=document.getElementsByTagName("head")[0],o=document.createElement("script");o.type="text/javascript",o.charset="utf-8",o.async=!0,o.timeout=12e4,e.nc&&o.setAttribute("nonce",e.nc),o.src=e.p+""+t+"-27896e7b8aabcfde5b72.js";var a=setTimeout(n,12e4);o.onerror=o.onload=n;var s=new Promise(function(e,n){i[t]=[e,n]});return i[t][2]=s,r.appendChild(o),s},e.m=t,e.c=r,e.i=function(t){return t},e.d=function(t,n,r){e.o(t,n)||Object.defineProperty(t,n,{configurable:!1,enumerable:!0,get:r})},e.n=function(t){var n=t&&t.__esModule?function(){return t.default}:function(){return t};return e.d(n,"a",n),n},e.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},e.p="",e.oe=function(t){throw console.error(t),t},e(e.s=140)}({0:function(t,e,n){n(36),t.exports=angular},1:function(t,e,n){var r,i;/*!
 * jQuery JavaScript Library v3.1.1
 * https://jquery.com/
 *
 * Includes Sizzle.js
 * https://sizzlejs.com/
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license
 * https://jquery.org/license
 *
 * Date: 2016-09-22T22:30Z
 */
!function(e,n){"use strict";"object"==typeof t&&"object"==typeof t.exports?t.exports=e.document?n(e,!0):function(t){if(!t.document)throw new Error("jQuery requires a window with a document");return n(t)}:n(e)}("undefined"!=typeof window?window:this,function(n,o){"use strict";function a(t,e){e=e||ot;var n=e.createElement("script");n.text=t,e.head.appendChild(n).parentNode.removeChild(n)}function s(t){var e=!!t&&"length"in t&&t.length,n=yt.type(t);return"function"!==n&&!yt.isWindow(t)&&("array"===n||0===e||"number"==typeof e&&e>0&&e-1 in t)}function u(t,e,n){return yt.isFunction(e)?yt.grep(t,function(t,r){return!!e.call(t,r,t)!==n}):e.nodeType?yt.grep(t,function(t){return t===e!==n}):"string"!=typeof e?yt.grep(t,function(t){return ct.call(e,t)>-1!==n}):Et.test(e)?yt.filter(e,t,n):(e=yt.filter(e,t),yt.grep(t,function(t){return ct.call(e,t)>-1!==n&&1===t.nodeType}))}function l(t,e){for(;(t=t[e])&&1!==t.nodeType;);return t}function c(t){var e={};return yt.each(t.match(Pt)||[],function(t,n){e[n]=!0}),e}function f(t){return t}function p(t){throw t}function h(t,e,n){var r;try{t&&yt.isFunction(r=t.promise)?r.call(t).done(e).fail(n):t&&yt.isFunction(r=t.then)?r.call(t,e,n):e.call(void 0,t)}catch(t){n.call(void 0,t)}}function d(){ot.removeEventListener("DOMContentLoaded",d),n.removeEventListener("load",d),yt.ready()}function v(){this.expando=yt.expando+v.uid++}function g(t){return"true"===t||"false"!==t&&("null"===t?null:t===+t+""?+t:Ut.test(t)?JSON.parse(t):t)}function m(t,e,n){var r;if(void 0===n&&1===t.nodeType)if(r="data-"+e.replace(Lt,"-$&").toLowerCase(),n=t.getAttribute(r),"string"==typeof n){try{n=g(n)}catch(t){}Vt.set(t,e,n)}else n=void 0;return n}function y(t,e,n,r){var i,o=1,a=20,s=r?function(){return r.cur()}:function(){return yt.css(t,e,"")},u=s(),l=n&&n[3]||(yt.cssNumber[e]?"":"px"),c=(yt.cssNumber[e]||"px"!==l&&+u)&&Bt.exec(yt.css(t,e));if(c&&c[3]!==l){l=l||c[3],n=n||[],c=+u||1;do o=o||".5",c/=o,yt.style(t,e,c+l);while(o!==(o=s()/u)&&1!==o&&--a)}return n&&(c=+c||+u||0,i=n[1]?c+(n[1]+1)*n[2]:+n[2],r&&(r.unit=l,r.start=c,r.end=i)),i}function b(t){var e,n=t.ownerDocument,r=t.nodeName,i=Gt[r];return i?i:(e=n.body.appendChild(n.createElement(r)),i=yt.css(e,"display"),e.parentNode.removeChild(e),"none"===i&&(i="block"),Gt[r]=i,i)}function $(t,e){for(var n,r,i=[],o=0,a=t.length;o<a;o++)r=t[o],r.style&&(n=r.style.display,e?("none"===n&&(i[o]=qt.get(r,"display")||null,i[o]||(r.style.display="")),""===r.style.display&&Wt(r)&&(i[o]=b(r))):"none"!==n&&(i[o]="none",qt.set(r,"display",n)));for(o=0;o<a;o++)null!=i[o]&&(t[o].style.display=i[o]);return t}function w(t,e){var n;return n="undefined"!=typeof t.getElementsByTagName?t.getElementsByTagName(e||"*"):"undefined"!=typeof t.querySelectorAll?t.querySelectorAll(e||"*"):[],void 0===e||e&&yt.nodeName(t,e)?yt.merge([t],n):n}function x(t,e){for(var n=0,r=t.length;n<r;n++)qt.set(t[n],"globalEval",!e||qt.get(e[n],"globalEval"))}function _(t,e,n,r,i){for(var o,a,s,u,l,c,f=e.createDocumentFragment(),p=[],h=0,d=t.length;h<d;h++)if(o=t[h],o||0===o)if("object"===yt.type(o))yt.merge(p,o.nodeType?[o]:o);else if(Qt.test(o)){for(a=a||f.appendChild(e.createElement("div")),s=(Kt.exec(o)||["",""])[1].toLowerCase(),u=Zt[s]||Zt._default,a.innerHTML=u[1]+yt.htmlPrefilter(o)+u[2],c=u[0];c--;)a=a.lastChild;yt.merge(p,a.childNodes),a=f.firstChild,a.textContent=""}else p.push(e.createTextNode(o));for(f.textContent="",h=0;o=p[h++];)if(r&&yt.inArray(o,r)>-1)i&&i.push(o);else if(l=yt.contains(o.ownerDocument,o),a=w(f.appendChild(o),"script"),l&&x(a),n)for(c=0;o=a[c++];)Jt.test(o.type||"")&&n.push(o);return f}function C(){return!0}function k(){return!1}function S(){try{return ot.activeElement}catch(t){}}function T(t,e,n,r,i,o){var a,s;if("object"==typeof e){"string"!=typeof n&&(r=r||n,n=void 0);for(s in e)T(t,s,n,r,e[s],o);return t}if(null==r&&null==i?(i=n,r=n=void 0):null==i&&("string"==typeof n?(i=r,r=void 0):(i=r,r=n,n=void 0)),i===!1)i=k;else if(!i)return t;return 1===o&&(a=i,i=function(t){return yt().off(t),a.apply(this,arguments)},i.guid=a.guid||(a.guid=yt.guid++)),t.each(function(){yt.event.add(this,e,i,r,n)})}function E(t,e){return yt.nodeName(t,"table")&&yt.nodeName(11!==e.nodeType?e:e.firstChild,"tr")?t.getElementsByTagName("tbody")[0]||t:t}function M(t){return t.type=(null!==t.getAttribute("type"))+"/"+t.type,t}function D(t){var e=se.exec(t.type);return e?t.type=e[1]:t.removeAttribute("type"),t}function A(t,e){var n,r,i,o,a,s,u,l;if(1===e.nodeType){if(qt.hasData(t)&&(o=qt.access(t),a=qt.set(e,o),l=o.events)){delete a.handle,a.events={};for(i in l)for(n=0,r=l[i].length;n<r;n++)yt.event.add(e,i,l[i][n])}Vt.hasData(t)&&(s=Vt.access(t),u=yt.extend({},s),Vt.set(e,u))}}function O(t,e){var n=e.nodeName.toLowerCase();"input"===n&&Xt.test(t.type)?e.checked=t.checked:"input"!==n&&"textarea"!==n||(e.defaultValue=t.defaultValue)}function N(t,e,n,r){e=ut.apply([],e);var i,o,s,u,l,c,f=0,p=t.length,h=p-1,d=e[0],v=yt.isFunction(d);if(v||p>1&&"string"==typeof d&&!gt.checkClone&&ae.test(d))return t.each(function(i){var o=t.eq(i);v&&(e[0]=d.call(this,i,o.html())),N(o,e,n,r)});if(p&&(i=_(e,t[0].ownerDocument,!1,t,r),o=i.firstChild,1===i.childNodes.length&&(i=o),o||r)){for(s=yt.map(w(i,"script"),M),u=s.length;f<p;f++)l=i,f!==h&&(l=yt.clone(l,!0,!0),u&&yt.merge(s,w(l,"script"))),n.call(t[f],l,f);if(u)for(c=s[s.length-1].ownerDocument,yt.map(s,D),f=0;f<u;f++)l=s[f],Jt.test(l.type||"")&&!qt.access(l,"globalEval")&&yt.contains(c,l)&&(l.src?yt._evalUrl&&yt._evalUrl(l.src):a(l.textContent.replace(ue,""),c))}return t}function P(t,e,n){for(var r,i=e?yt.filter(e,t):t,o=0;null!=(r=i[o]);o++)n||1!==r.nodeType||yt.cleanData(w(r)),r.parentNode&&(n&&yt.contains(r.ownerDocument,r)&&x(w(r,"script")),r.parentNode.removeChild(r));return t}function R(t,e,n){var r,i,o,a,s=t.style;return n=n||fe(t),n&&(a=n.getPropertyValue(e)||n[e],""!==a||yt.contains(t.ownerDocument,t)||(a=yt.style(t,e)),!gt.pixelMarginRight()&&ce.test(a)&&le.test(e)&&(r=s.width,i=s.minWidth,o=s.maxWidth,s.minWidth=s.maxWidth=s.width=a,a=n.width,s.width=r,s.minWidth=i,s.maxWidth=o)),void 0!==a?a+"":a}function I(t,e){return{get:function(){return t()?void delete this.get:(this.get=e).apply(this,arguments)}}}function j(t){if(t in ge)return t;for(var e=t[0].toUpperCase()+t.slice(1),n=ve.length;n--;)if(t=ve[n]+e,t in ge)return t}function F(t,e,n){var r=Bt.exec(e);return r?Math.max(0,r[2]-(n||0))+(r[3]||"px"):e}function q(t,e,n,r,i){var o,a=0;for(o=n===(r?"border":"content")?4:"width"===e?1:0;o<4;o+=2)"margin"===n&&(a+=yt.css(t,n+zt[o],!0,i)),r?("content"===n&&(a-=yt.css(t,"padding"+zt[o],!0,i)),"margin"!==n&&(a-=yt.css(t,"border"+zt[o]+"Width",!0,i))):(a+=yt.css(t,"padding"+zt[o],!0,i),"padding"!==n&&(a+=yt.css(t,"border"+zt[o]+"Width",!0,i)));return a}function V(t,e,n){var r,i=!0,o=fe(t),a="border-box"===yt.css(t,"boxSizing",!1,o);if(t.getClientRects().length&&(r=t.getBoundingClientRect()[e]),r<=0||null==r){if(r=R(t,e,o),(r<0||null==r)&&(r=t.style[e]),ce.test(r))return r;i=a&&(gt.boxSizingReliable()||r===t.style[e]),r=parseFloat(r)||0}return r+q(t,e,n||(a?"border":"content"),i,o)+"px"}function U(t,e,n,r,i){return new U.prototype.init(t,e,n,r,i)}function L(){ye&&(n.requestAnimationFrame(L),yt.fx.tick())}function H(){return n.setTimeout(function(){me=void 0}),me=yt.now()}function B(t,e){var n,r=0,i={height:t};for(e=e?1:0;r<4;r+=2-e)n=zt[r],i["margin"+n]=i["padding"+n]=t;return e&&(i.opacity=i.width=t),i}function z(t,e,n){for(var r,i=(G.tweeners[e]||[]).concat(G.tweeners["*"]),o=0,a=i.length;o<a;o++)if(r=i[o].call(n,e,t))return r}function W(t,e,n){var r,i,o,a,s,u,l,c,f="width"in e||"height"in e,p=this,h={},d=t.style,v=t.nodeType&&Wt(t),g=qt.get(t,"fxshow");n.queue||(a=yt._queueHooks(t,"fx"),null==a.unqueued&&(a.unqueued=0,s=a.empty.fire,a.empty.fire=function(){a.unqueued||s()}),a.unqueued++,p.always(function(){p.always(function(){a.unqueued--,yt.queue(t,"fx").length||a.empty.fire()})}));for(r in e)if(i=e[r],be.test(i)){if(delete e[r],o=o||"toggle"===i,i===(v?"hide":"show")){if("show"!==i||!g||void 0===g[r])continue;v=!0}h[r]=g&&g[r]||yt.style(t,r)}if(u=!yt.isEmptyObject(e),u||!yt.isEmptyObject(h)){f&&1===t.nodeType&&(n.overflow=[d.overflow,d.overflowX,d.overflowY],l=g&&g.display,null==l&&(l=qt.get(t,"display")),c=yt.css(t,"display"),"none"===c&&(l?c=l:($([t],!0),l=t.style.display||l,c=yt.css(t,"display"),$([t]))),("inline"===c||"inline-block"===c&&null!=l)&&"none"===yt.css(t,"float")&&(u||(p.done(function(){d.display=l}),null==l&&(c=d.display,l="none"===c?"":c)),d.display="inline-block")),n.overflow&&(d.overflow="hidden",p.always(function(){d.overflow=n.overflow[0],d.overflowX=n.overflow[1],d.overflowY=n.overflow[2]})),u=!1;for(r in h)u||(g?"hidden"in g&&(v=g.hidden):g=qt.access(t,"fxshow",{display:l}),o&&(g.hidden=!v),v&&$([t],!0),p.done(function(){v||$([t]),qt.remove(t,"fxshow");for(r in h)yt.style(t,r,h[r])})),u=z(v?g[r]:0,r,p),r in g||(g[r]=u.start,v&&(u.end=u.start,u.start=0))}}function Y(t,e){var n,r,i,o,a;for(n in t)if(r=yt.camelCase(n),i=e[r],o=t[n],yt.isArray(o)&&(i=o[1],o=t[n]=o[0]),n!==r&&(t[r]=o,delete t[n]),a=yt.cssHooks[r],a&&"expand"in a){o=a.expand(o),delete t[r];for(n in o)n in t||(t[n]=o[n],e[n]=i)}else e[r]=i}function G(t,e,n){var r,i,o=0,a=G.prefilters.length,s=yt.Deferred().always(function(){delete u.elem}),u=function(){if(i)return!1;for(var e=me||H(),n=Math.max(0,l.startTime+l.duration-e),r=n/l.duration||0,o=1-r,a=0,u=l.tweens.length;a<u;a++)l.tweens[a].run(o);return s.notifyWith(t,[l,o,n]),o<1&&u?n:(s.resolveWith(t,[l]),!1)},l=s.promise({elem:t,props:yt.extend({},e),opts:yt.extend(!0,{specialEasing:{},easing:yt.easing._default},n),originalProperties:e,originalOptions:n,startTime:me||H(),duration:n.duration,tweens:[],createTween:function(e,n){var r=yt.Tween(t,l.opts,e,n,l.opts.specialEasing[e]||l.opts.easing);return l.tweens.push(r),r},stop:function(e){var n=0,r=e?l.tweens.length:0;if(i)return this;for(i=!0;n<r;n++)l.tweens[n].run(1);return e?(s.notifyWith(t,[l,1,0]),s.resolveWith(t,[l,e])):s.rejectWith(t,[l,e]),this}}),c=l.props;for(Y(c,l.opts.specialEasing);o<a;o++)if(r=G.prefilters[o].call(l,t,c,l.opts))return yt.isFunction(r.stop)&&(yt._queueHooks(l.elem,l.opts.queue).stop=yt.proxy(r.stop,r)),r;return yt.map(c,z,l),yt.isFunction(l.opts.start)&&l.opts.start.call(t,l),yt.fx.timer(yt.extend(u,{elem:t,anim:l,queue:l.opts.queue})),l.progress(l.opts.progress).done(l.opts.done,l.opts.complete).fail(l.opts.fail).always(l.opts.always)}function X(t){var e=t.match(Pt)||[];return e.join(" ")}function K(t){return t.getAttribute&&t.getAttribute("class")||""}function J(t,e,n,r){var i;if(yt.isArray(e))yt.each(e,function(e,i){n||De.test(t)?r(t,i):J(t+"["+("object"==typeof i&&null!=i?e:"")+"]",i,n,r)});else if(n||"object"!==yt.type(e))r(t,e);else for(i in e)J(t+"["+i+"]",e[i],n,r)}function Z(t){return function(e,n){"string"!=typeof e&&(n=e,e="*");var r,i=0,o=e.toLowerCase().match(Pt)||[];if(yt.isFunction(n))for(;r=o[i++];)"+"===r[0]?(r=r.slice(1)||"*",(t[r]=t[r]||[]).unshift(n)):(t[r]=t[r]||[]).push(n)}}function Q(t,e,n,r){function i(s){var u;return o[s]=!0,yt.each(t[s]||[],function(t,s){var l=s(e,n,r);return"string"!=typeof l||a||o[l]?a?!(u=l):void 0:(e.dataTypes.unshift(l),i(l),!1)}),u}var o={},a=t===Le;return i(e.dataTypes[0])||!o["*"]&&i("*")}function tt(t,e){var n,r,i=yt.ajaxSettings.flatOptions||{};for(n in e)void 0!==e[n]&&((i[n]?t:r||(r={}))[n]=e[n]);return r&&yt.extend(!0,t,r),t}function et(t,e,n){for(var r,i,o,a,s=t.contents,u=t.dataTypes;"*"===u[0];)u.shift(),void 0===r&&(r=t.mimeType||e.getResponseHeader("Content-Type"));if(r)for(i in s)if(s[i]&&s[i].test(r)){u.unshift(i);break}if(u[0]in n)o=u[0];else{for(i in n){if(!u[0]||t.converters[i+" "+u[0]]){o=i;break}a||(a=i)}o=o||a}if(o)return o!==u[0]&&u.unshift(o),n[o]}function nt(t,e,n,r){var i,o,a,s,u,l={},c=t.dataTypes.slice();if(c[1])for(a in t.converters)l[a.toLowerCase()]=t.converters[a];for(o=c.shift();o;)if(t.responseFields[o]&&(n[t.responseFields[o]]=e),!u&&r&&t.dataFilter&&(e=t.dataFilter(e,t.dataType)),u=o,o=c.shift())if("*"===o)o=u;else if("*"!==u&&u!==o){if(a=l[u+" "+o]||l["* "+o],!a)for(i in l)if(s=i.split(" "),s[1]===o&&(a=l[u+" "+s[0]]||l["* "+s[0]])){a===!0?a=l[i]:l[i]!==!0&&(o=s[0],c.unshift(s[1]));break}if(a!==!0)if(a&&t.throws)e=a(e);else try{e=a(e)}catch(t){return{state:"parsererror",error:a?t:"No conversion from "+u+" to "+o}}}return{state:"success",data:e}}function rt(t){return yt.isWindow(t)?t:9===t.nodeType&&t.defaultView}var it=[],ot=n.document,at=Object.getPrototypeOf,st=it.slice,ut=it.concat,lt=it.push,ct=it.indexOf,ft={},pt=ft.toString,ht=ft.hasOwnProperty,dt=ht.toString,vt=dt.call(Object),gt={},mt="3.1.1",yt=function(t,e){return new yt.fn.init(t,e)},bt=/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,$t=/^-ms-/,wt=/-([a-z])/g,xt=function(t,e){return e.toUpperCase()};yt.fn=yt.prototype={jquery:mt,constructor:yt,length:0,toArray:function(){return st.call(this)},get:function(t){return null==t?st.call(this):t<0?this[t+this.length]:this[t]},pushStack:function(t){var e=yt.merge(this.constructor(),t);return e.prevObject=this,e},each:function(t){return yt.each(this,t)},map:function(t){return this.pushStack(yt.map(this,function(e,n){return t.call(e,n,e)}))},slice:function(){return this.pushStack(st.apply(this,arguments))},first:function(){return this.eq(0)},last:function(){return this.eq(-1)},eq:function(t){var e=this.length,n=+t+(t<0?e:0);return this.pushStack(n>=0&&n<e?[this[n]]:[])},end:function(){return this.prevObject||this.constructor()},push:lt,sort:it.sort,splice:it.splice},yt.extend=yt.fn.extend=function(){var t,e,n,r,i,o,a=arguments[0]||{},s=1,u=arguments.length,l=!1;for("boolean"==typeof a&&(l=a,a=arguments[s]||{},s++),"object"==typeof a||yt.isFunction(a)||(a={}),s===u&&(a=this,s--);s<u;s++)if(null!=(t=arguments[s]))for(e in t)n=a[e],r=t[e],a!==r&&(l&&r&&(yt.isPlainObject(r)||(i=yt.isArray(r)))?(i?(i=!1,o=n&&yt.isArray(n)?n:[]):o=n&&yt.isPlainObject(n)?n:{},a[e]=yt.extend(l,o,r)):void 0!==r&&(a[e]=r));return a},yt.extend({expando:"jQuery"+(mt+Math.random()).replace(/\D/g,""),isReady:!0,error:function(t){throw new Error(t)},noop:function(){},isFunction:function(t){return"function"===yt.type(t)},isArray:Array.isArray,isWindow:function(t){return null!=t&&t===t.window},isNumeric:function(t){var e=yt.type(t);return("number"===e||"string"===e)&&!isNaN(t-parseFloat(t))},isPlainObject:function(t){var e,n;return!(!t||"[object Object]"!==pt.call(t))&&(!(e=at(t))||(n=ht.call(e,"constructor")&&e.constructor,"function"==typeof n&&dt.call(n)===vt))},isEmptyObject:function(t){var e;for(e in t)return!1;return!0},type:function(t){return null==t?t+"":"object"==typeof t||"function"==typeof t?ft[pt.call(t)]||"object":typeof t},globalEval:function(t){a(t)},camelCase:function(t){return t.replace($t,"ms-").replace(wt,xt)},nodeName:function(t,e){return t.nodeName&&t.nodeName.toLowerCase()===e.toLowerCase()},each:function(t,e){var n,r=0;if(s(t))for(n=t.length;r<n&&e.call(t[r],r,t[r])!==!1;r++);else for(r in t)if(e.call(t[r],r,t[r])===!1)break;return t},trim:function(t){return null==t?"":(t+"").replace(bt,"")},makeArray:function(t,e){var n=e||[];return null!=t&&(s(Object(t))?yt.merge(n,"string"==typeof t?[t]:t):lt.call(n,t)),n},inArray:function(t,e,n){return null==e?-1:ct.call(e,t,n)},merge:function(t,e){for(var n=+e.length,r=0,i=t.length;r<n;r++)t[i++]=e[r];return t.length=i,t},grep:function(t,e,n){for(var r,i=[],o=0,a=t.length,s=!n;o<a;o++)r=!e(t[o],o),r!==s&&i.push(t[o]);return i},map:function(t,e,n){var r,i,o=0,a=[];if(s(t))for(r=t.length;o<r;o++)i=e(t[o],o,n),null!=i&&a.push(i);else for(o in t)i=e(t[o],o,n),null!=i&&a.push(i);return ut.apply([],a)},guid:1,proxy:function(t,e){var n,r,i;if("string"==typeof e&&(n=t[e],e=t,t=n),yt.isFunction(t))return r=st.call(arguments,2),i=function(){return t.apply(e||this,r.concat(st.call(arguments)))},i.guid=t.guid=t.guid||yt.guid++,i},now:Date.now,support:gt}),"function"==typeof Symbol&&(yt.fn[Symbol.iterator]=it[Symbol.iterator]),yt.each("Boolean Number String Function Array Date RegExp Object Error Symbol".split(" "),function(t,e){ft["[object "+e+"]"]=e.toLowerCase()});var _t=/*!
 * Sizzle CSS Selector Engine v2.3.3
 * https://sizzlejs.com/
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license
 * http://jquery.org/license
 *
 * Date: 2016-08-08
 */
function(t){function e(t,e,n,r){var i,o,a,s,u,l,c,p=e&&e.ownerDocument,d=e?e.nodeType:9;if(n=n||[],"string"!=typeof t||!t||1!==d&&9!==d&&11!==d)return n;if(!r&&((e?e.ownerDocument||e:U)!==N&&O(e),e=e||N,R)){if(11!==d&&(u=mt.exec(t)))if(i=u[1]){if(9===d){if(!(a=e.getElementById(i)))return n;if(a.id===i)return n.push(a),n}else if(p&&(a=p.getElementById(i))&&q(e,a)&&a.id===i)return n.push(a),n}else{if(u[2])return Z.apply(n,e.getElementsByTagName(t)),n;if((i=u[3])&&x.getElementsByClassName&&e.getElementsByClassName)return Z.apply(n,e.getElementsByClassName(i)),n}if(x.qsa&&!W[t+" "]&&(!I||!I.test(t))){if(1!==d)p=e,c=t;else if("object"!==e.nodeName.toLowerCase()){for((s=e.getAttribute("id"))?s=s.replace(wt,xt):e.setAttribute("id",s=V),l=S(t),o=l.length;o--;)l[o]="#"+s+" "+h(l[o]);c=l.join(","),p=yt.test(t)&&f(e.parentNode)||e}if(c)try{return Z.apply(n,p.querySelectorAll(c)),n}catch(t){}finally{s===V&&e.removeAttribute("id")}}}return E(t.replace(st,"$1"),e,n,r)}function n(){function t(n,r){return e.push(n+" ")>_.cacheLength&&delete t[e.shift()],t[n+" "]=r}var e=[];return t}function r(t){return t[V]=!0,t}function i(t){var e=N.createElement("fieldset");try{return!!t(e)}catch(t){return!1}finally{e.parentNode&&e.parentNode.removeChild(e),e=null}}function o(t,e){for(var n=t.split("|"),r=n.length;r--;)_.attrHandle[n[r]]=e}function a(t,e){var n=e&&t,r=n&&1===t.nodeType&&1===e.nodeType&&t.sourceIndex-e.sourceIndex;if(r)return r;if(n)for(;n=n.nextSibling;)if(n===e)return-1;return t?1:-1}function s(t){return function(e){var n=e.nodeName.toLowerCase();return"input"===n&&e.type===t}}function u(t){return function(e){var n=e.nodeName.toLowerCase();return("input"===n||"button"===n)&&e.type===t}}function l(t){return function(e){return"form"in e?e.parentNode&&e.disabled===!1?"label"in e?"label"in e.parentNode?e.parentNode.disabled===t:e.disabled===t:e.isDisabled===t||e.isDisabled!==!t&&Ct(e)===t:e.disabled===t:"label"in e&&e.disabled===t}}function c(t){return r(function(e){return e=+e,r(function(n,r){for(var i,o=t([],n.length,e),a=o.length;a--;)n[i=o[a]]&&(n[i]=!(r[i]=n[i]))})})}function f(t){return t&&"undefined"!=typeof t.getElementsByTagName&&t}function p(){}function h(t){for(var e=0,n=t.length,r="";e<n;e++)r+=t[e].value;return r}function d(t,e,n){var r=e.dir,i=e.next,o=i||r,a=n&&"parentNode"===o,s=H++;return e.first?function(e,n,i){for(;e=e[r];)if(1===e.nodeType||a)return t(e,n,i);return!1}:function(e,n,u){var l,c,f,p=[L,s];if(u){for(;e=e[r];)if((1===e.nodeType||a)&&t(e,n,u))return!0}else for(;e=e[r];)if(1===e.nodeType||a)if(f=e[V]||(e[V]={}),c=f[e.uniqueID]||(f[e.uniqueID]={}),i&&i===e.nodeName.toLowerCase())e=e[r]||e;else{if((l=c[o])&&l[0]===L&&l[1]===s)return p[2]=l[2];if(c[o]=p,p[2]=t(e,n,u))return!0}return!1}}function v(t){return t.length>1?function(e,n,r){for(var i=t.length;i--;)if(!t[i](e,n,r))return!1;return!0}:t[0]}function g(t,n,r){for(var i=0,o=n.length;i<o;i++)e(t,n[i],r);return r}function m(t,e,n,r,i){for(var o,a=[],s=0,u=t.length,l=null!=e;s<u;s++)(o=t[s])&&(n&&!n(o,r,i)||(a.push(o),l&&e.push(s)));return a}function y(t,e,n,i,o,a){return i&&!i[V]&&(i=y(i)),o&&!o[V]&&(o=y(o,a)),r(function(r,a,s,u){var l,c,f,p=[],h=[],d=a.length,v=r||g(e||"*",s.nodeType?[s]:s,[]),y=!t||!r&&e?v:m(v,p,t,s,u),b=n?o||(r?t:d||i)?[]:a:y;if(n&&n(y,b,s,u),i)for(l=m(b,h),i(l,[],s,u),c=l.length;c--;)(f=l[c])&&(b[h[c]]=!(y[h[c]]=f));if(r){if(o||t){if(o){for(l=[],c=b.length;c--;)(f=b[c])&&l.push(y[c]=f);o(null,b=[],l,u)}for(c=b.length;c--;)(f=b[c])&&(l=o?tt(r,f):p[c])>-1&&(r[l]=!(a[l]=f))}}else b=m(b===a?b.splice(d,b.length):b),o?o(null,a,b,u):Z.apply(a,b)})}function b(t){for(var e,n,r,i=t.length,o=_.relative[t[0].type],a=o||_.relative[" "],s=o?1:0,u=d(function(t){return t===e},a,!0),l=d(function(t){return tt(e,t)>-1},a,!0),c=[function(t,n,r){var i=!o&&(r||n!==M)||((e=n).nodeType?u(t,n,r):l(t,n,r));return e=null,i}];s<i;s++)if(n=_.relative[t[s].type])c=[d(v(c),n)];else{if(n=_.filter[t[s].type].apply(null,t[s].matches),n[V]){for(r=++s;r<i&&!_.relative[t[r].type];r++);return y(s>1&&v(c),s>1&&h(t.slice(0,s-1).concat({value:" "===t[s-2].type?"*":""})).replace(st,"$1"),n,s<r&&b(t.slice(s,r)),r<i&&b(t=t.slice(r)),r<i&&h(t))}c.push(n)}return v(c)}function $(t,n){var i=n.length>0,o=t.length>0,a=function(r,a,s,u,l){var c,f,p,h=0,d="0",v=r&&[],g=[],y=M,b=r||o&&_.find.TAG("*",l),$=L+=null==y?1:Math.random()||.1,w=b.length;for(l&&(M=a===N||a||l);d!==w&&null!=(c=b[d]);d++){if(o&&c){for(f=0,a||c.ownerDocument===N||(O(c),s=!R);p=t[f++];)if(p(c,a||N,s)){u.push(c);break}l&&(L=$)}i&&((c=!p&&c)&&h--,r&&v.push(c))}if(h+=d,i&&d!==h){for(f=0;p=n[f++];)p(v,g,a,s);if(r){if(h>0)for(;d--;)v[d]||g[d]||(g[d]=K.call(u));g=m(g)}Z.apply(u,g),l&&!r&&g.length>0&&h+n.length>1&&e.uniqueSort(u)}return l&&(L=$,M=y),v};return i?r(a):a}var w,x,_,C,k,S,T,E,M,D,A,O,N,P,R,I,j,F,q,V="sizzle"+1*new Date,U=t.document,L=0,H=0,B=n(),z=n(),W=n(),Y=function(t,e){return t===e&&(A=!0),0},G={}.hasOwnProperty,X=[],K=X.pop,J=X.push,Z=X.push,Q=X.slice,tt=function(t,e){for(var n=0,r=t.length;n<r;n++)if(t[n]===e)return n;return-1},et="checked|selected|async|autofocus|autoplay|controls|defer|disabled|hidden|ismap|loop|multiple|open|readonly|required|scoped",nt="[\\x20\\t\\r\\n\\f]",rt="(?:\\\\.|[\\w-]|[^\0-\\xa0])+",it="\\["+nt+"*("+rt+")(?:"+nt+"*([*^$|!~]?=)"+nt+"*(?:'((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\"|("+rt+"))|)"+nt+"*\\]",ot=":("+rt+")(?:\\((('((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\")|((?:\\\\.|[^\\\\()[\\]]|"+it+")*)|.*)\\)|)",at=new RegExp(nt+"+","g"),st=new RegExp("^"+nt+"+|((?:^|[^\\\\])(?:\\\\.)*)"+nt+"+$","g"),ut=new RegExp("^"+nt+"*,"+nt+"*"),lt=new RegExp("^"+nt+"*([>+~]|"+nt+")"+nt+"*"),ct=new RegExp("="+nt+"*([^\\]'\"]*?)"+nt+"*\\]","g"),ft=new RegExp(ot),pt=new RegExp("^"+rt+"$"),ht={ID:new RegExp("^#("+rt+")"),CLASS:new RegExp("^\\.("+rt+")"),TAG:new RegExp("^("+rt+"|[*])"),ATTR:new RegExp("^"+it),PSEUDO:new RegExp("^"+ot),CHILD:new RegExp("^:(only|first|last|nth|nth-last)-(child|of-type)(?:\\("+nt+"*(even|odd|(([+-]|)(\\d*)n|)"+nt+"*(?:([+-]|)"+nt+"*(\\d+)|))"+nt+"*\\)|)","i"),bool:new RegExp("^(?:"+et+")$","i"),needsContext:new RegExp("^"+nt+"*[>+~]|:(even|odd|eq|gt|lt|nth|first|last)(?:\\("+nt+"*((?:-\\d)?\\d*)"+nt+"*\\)|)(?=[^-]|$)","i")},dt=/^(?:input|select|textarea|button)$/i,vt=/^h\d$/i,gt=/^[^{]+\{\s*\[native \w/,mt=/^(?:#([\w-]+)|(\w+)|\.([\w-]+))$/,yt=/[+~]/,bt=new RegExp("\\\\([\\da-f]{1,6}"+nt+"?|("+nt+")|.)","ig"),$t=function(t,e,n){var r="0x"+e-65536;return r!==r||n?e:r<0?String.fromCharCode(r+65536):String.fromCharCode(r>>10|55296,1023&r|56320)},wt=/([\0-\x1f\x7f]|^-?\d)|^-$|[^\0-\x1f\x7f-\uFFFF\w-]/g,xt=function(t,e){return e?"\0"===t?"�":t.slice(0,-1)+"\\"+t.charCodeAt(t.length-1).toString(16)+" ":"\\"+t},_t=function(){O()},Ct=d(function(t){return t.disabled===!0&&("form"in t||"label"in t)},{dir:"parentNode",next:"legend"});try{Z.apply(X=Q.call(U.childNodes),U.childNodes),X[U.childNodes.length].nodeType}catch(t){Z={apply:X.length?function(t,e){J.apply(t,Q.call(e))}:function(t,e){for(var n=t.length,r=0;t[n++]=e[r++];);t.length=n-1}}}x=e.support={},k=e.isXML=function(t){var e=t&&(t.ownerDocument||t).documentElement;return!!e&&"HTML"!==e.nodeName},O=e.setDocument=function(t){var e,n,r=t?t.ownerDocument||t:U;return r!==N&&9===r.nodeType&&r.documentElement?(N=r,P=N.documentElement,R=!k(N),U!==N&&(n=N.defaultView)&&n.top!==n&&(n.addEventListener?n.addEventListener("unload",_t,!1):n.attachEvent&&n.attachEvent("onunload",_t)),x.attributes=i(function(t){return t.className="i",!t.getAttribute("className")}),x.getElementsByTagName=i(function(t){return t.appendChild(N.createComment("")),!t.getElementsByTagName("*").length}),x.getElementsByClassName=gt.test(N.getElementsByClassName),x.getById=i(function(t){return P.appendChild(t).id=V,!N.getElementsByName||!N.getElementsByName(V).length}),x.getById?(_.filter.ID=function(t){var e=t.replace(bt,$t);return function(t){return t.getAttribute("id")===e}},_.find.ID=function(t,e){if("undefined"!=typeof e.getElementById&&R){var n=e.getElementById(t);return n?[n]:[]}}):(_.filter.ID=function(t){var e=t.replace(bt,$t);return function(t){var n="undefined"!=typeof t.getAttributeNode&&t.getAttributeNode("id");return n&&n.value===e}},_.find.ID=function(t,e){if("undefined"!=typeof e.getElementById&&R){var n,r,i,o=e.getElementById(t);if(o){if(n=o.getAttributeNode("id"),n&&n.value===t)return[o];for(i=e.getElementsByName(t),r=0;o=i[r++];)if(n=o.getAttributeNode("id"),n&&n.value===t)return[o]}return[]}}),_.find.TAG=x.getElementsByTagName?function(t,e){return"undefined"!=typeof e.getElementsByTagName?e.getElementsByTagName(t):x.qsa?e.querySelectorAll(t):void 0}:function(t,e){var n,r=[],i=0,o=e.getElementsByTagName(t);if("*"===t){for(;n=o[i++];)1===n.nodeType&&r.push(n);return r}return o},_.find.CLASS=x.getElementsByClassName&&function(t,e){if("undefined"!=typeof e.getElementsByClassName&&R)return e.getElementsByClassName(t)},j=[],I=[],(x.qsa=gt.test(N.querySelectorAll))&&(i(function(t){P.appendChild(t).innerHTML="<a id='"+V+"'></a><select id='"+V+"-\r\\' msallowcapture=''><option selected=''></option></select>",t.querySelectorAll("[msallowcapture^='']").length&&I.push("[*^$]="+nt+"*(?:''|\"\")"),t.querySelectorAll("[selected]").length||I.push("\\["+nt+"*(?:value|"+et+")"),t.querySelectorAll("[id~="+V+"-]").length||I.push("~="),t.querySelectorAll(":checked").length||I.push(":checked"),t.querySelectorAll("a#"+V+"+*").length||I.push(".#.+[+~]")}),i(function(t){t.innerHTML="<a href='' disabled='disabled'></a><select disabled='disabled'><option/></select>";var e=N.createElement("input");e.setAttribute("type","hidden"),t.appendChild(e).setAttribute("name","D"),t.querySelectorAll("[name=d]").length&&I.push("name"+nt+"*[*^$|!~]?="),2!==t.querySelectorAll(":enabled").length&&I.push(":enabled",":disabled"),P.appendChild(t).disabled=!0,2!==t.querySelectorAll(":disabled").length&&I.push(":enabled",":disabled"),t.querySelectorAll("*,:x"),I.push(",.*:")})),(x.matchesSelector=gt.test(F=P.matches||P.webkitMatchesSelector||P.mozMatchesSelector||P.oMatchesSelector||P.msMatchesSelector))&&i(function(t){x.disconnectedMatch=F.call(t,"*"),F.call(t,"[s!='']:x"),j.push("!=",ot)}),I=I.length&&new RegExp(I.join("|")),j=j.length&&new RegExp(j.join("|")),e=gt.test(P.compareDocumentPosition),q=e||gt.test(P.contains)?function(t,e){var n=9===t.nodeType?t.documentElement:t,r=e&&e.parentNode;return t===r||!(!r||1!==r.nodeType||!(n.contains?n.contains(r):t.compareDocumentPosition&&16&t.compareDocumentPosition(r)))}:function(t,e){if(e)for(;e=e.parentNode;)if(e===t)return!0;return!1},Y=e?function(t,e){if(t===e)return A=!0,0;var n=!t.compareDocumentPosition-!e.compareDocumentPosition;return n?n:(n=(t.ownerDocument||t)===(e.ownerDocument||e)?t.compareDocumentPosition(e):1,1&n||!x.sortDetached&&e.compareDocumentPosition(t)===n?t===N||t.ownerDocument===U&&q(U,t)?-1:e===N||e.ownerDocument===U&&q(U,e)?1:D?tt(D,t)-tt(D,e):0:4&n?-1:1)}:function(t,e){if(t===e)return A=!0,0;var n,r=0,i=t.parentNode,o=e.parentNode,s=[t],u=[e];if(!i||!o)return t===N?-1:e===N?1:i?-1:o?1:D?tt(D,t)-tt(D,e):0;if(i===o)return a(t,e);for(n=t;n=n.parentNode;)s.unshift(n);for(n=e;n=n.parentNode;)u.unshift(n);for(;s[r]===u[r];)r++;return r?a(s[r],u[r]):s[r]===U?-1:u[r]===U?1:0},N):N},e.matches=function(t,n){return e(t,null,null,n)},e.matchesSelector=function(t,n){if((t.ownerDocument||t)!==N&&O(t),n=n.replace(ct,"='$1']"),x.matchesSelector&&R&&!W[n+" "]&&(!j||!j.test(n))&&(!I||!I.test(n)))try{var r=F.call(t,n);if(r||x.disconnectedMatch||t.document&&11!==t.document.nodeType)return r}catch(t){}return e(n,N,null,[t]).length>0},e.contains=function(t,e){return(t.ownerDocument||t)!==N&&O(t),q(t,e)},e.attr=function(t,e){(t.ownerDocument||t)!==N&&O(t);var n=_.attrHandle[e.toLowerCase()],r=n&&G.call(_.attrHandle,e.toLowerCase())?n(t,e,!R):void 0;return void 0!==r?r:x.attributes||!R?t.getAttribute(e):(r=t.getAttributeNode(e))&&r.specified?r.value:null},e.escape=function(t){return(t+"").replace(wt,xt)},e.error=function(t){throw new Error("Syntax error, unrecognized expression: "+t)},e.uniqueSort=function(t){var e,n=[],r=0,i=0;if(A=!x.detectDuplicates,D=!x.sortStable&&t.slice(0),t.sort(Y),A){for(;e=t[i++];)e===t[i]&&(r=n.push(i));for(;r--;)t.splice(n[r],1)}return D=null,t},C=e.getText=function(t){var e,n="",r=0,i=t.nodeType;if(i){if(1===i||9===i||11===i){if("string"==typeof t.textContent)return t.textContent;for(t=t.firstChild;t;t=t.nextSibling)n+=C(t)}else if(3===i||4===i)return t.nodeValue}else for(;e=t[r++];)n+=C(e);return n},_=e.selectors={cacheLength:50,createPseudo:r,match:ht,attrHandle:{},find:{},relative:{">":{dir:"parentNode",first:!0}," ":{dir:"parentNode"},"+":{dir:"previousSibling",first:!0},"~":{dir:"previousSibling"}},preFilter:{ATTR:function(t){return t[1]=t[1].replace(bt,$t),t[3]=(t[3]||t[4]||t[5]||"").replace(bt,$t),"~="===t[2]&&(t[3]=" "+t[3]+" "),t.slice(0,4)},CHILD:function(t){return t[1]=t[1].toLowerCase(),"nth"===t[1].slice(0,3)?(t[3]||e.error(t[0]),t[4]=+(t[4]?t[5]+(t[6]||1):2*("even"===t[3]||"odd"===t[3])),t[5]=+(t[7]+t[8]||"odd"===t[3])):t[3]&&e.error(t[0]),t},PSEUDO:function(t){var e,n=!t[6]&&t[2];return ht.CHILD.test(t[0])?null:(t[3]?t[2]=t[4]||t[5]||"":n&&ft.test(n)&&(e=S(n,!0))&&(e=n.indexOf(")",n.length-e)-n.length)&&(t[0]=t[0].slice(0,e),t[2]=n.slice(0,e)),t.slice(0,3))}},filter:{TAG:function(t){var e=t.replace(bt,$t).toLowerCase();return"*"===t?function(){return!0}:function(t){return t.nodeName&&t.nodeName.toLowerCase()===e}},CLASS:function(t){var e=B[t+" "];return e||(e=new RegExp("(^|"+nt+")"+t+"("+nt+"|$)"))&&B(t,function(t){return e.test("string"==typeof t.className&&t.className||"undefined"!=typeof t.getAttribute&&t.getAttribute("class")||"")})},ATTR:function(t,n,r){return function(i){var o=e.attr(i,t);return null==o?"!="===n:!n||(o+="","="===n?o===r:"!="===n?o!==r:"^="===n?r&&0===o.indexOf(r):"*="===n?r&&o.indexOf(r)>-1:"$="===n?r&&o.slice(-r.length)===r:"~="===n?(" "+o.replace(at," ")+" ").indexOf(r)>-1:"|="===n&&(o===r||o.slice(0,r.length+1)===r+"-"))}},CHILD:function(t,e,n,r,i){var o="nth"!==t.slice(0,3),a="last"!==t.slice(-4),s="of-type"===e;return 1===r&&0===i?function(t){return!!t.parentNode}:function(e,n,u){var l,c,f,p,h,d,v=o!==a?"nextSibling":"previousSibling",g=e.parentNode,m=s&&e.nodeName.toLowerCase(),y=!u&&!s,b=!1;if(g){if(o){for(;v;){for(p=e;p=p[v];)if(s?p.nodeName.toLowerCase()===m:1===p.nodeType)return!1;d=v="only"===t&&!d&&"nextSibling"}return!0}if(d=[a?g.firstChild:g.lastChild],a&&y){for(p=g,f=p[V]||(p[V]={}),c=f[p.uniqueID]||(f[p.uniqueID]={}),l=c[t]||[],h=l[0]===L&&l[1],b=h&&l[2],p=h&&g.childNodes[h];p=++h&&p&&p[v]||(b=h=0)||d.pop();)if(1===p.nodeType&&++b&&p===e){c[t]=[L,h,b];break}}else if(y&&(p=e,f=p[V]||(p[V]={}),c=f[p.uniqueID]||(f[p.uniqueID]={}),l=c[t]||[],h=l[0]===L&&l[1],b=h),b===!1)for(;(p=++h&&p&&p[v]||(b=h=0)||d.pop())&&((s?p.nodeName.toLowerCase()!==m:1!==p.nodeType)||!++b||(y&&(f=p[V]||(p[V]={}),c=f[p.uniqueID]||(f[p.uniqueID]={}),c[t]=[L,b]),p!==e)););return b-=i,b===r||b%r===0&&b/r>=0}}},PSEUDO:function(t,n){var i,o=_.pseudos[t]||_.setFilters[t.toLowerCase()]||e.error("unsupported pseudo: "+t);return o[V]?o(n):o.length>1?(i=[t,t,"",n],_.setFilters.hasOwnProperty(t.toLowerCase())?r(function(t,e){for(var r,i=o(t,n),a=i.length;a--;)r=tt(t,i[a]),t[r]=!(e[r]=i[a])}):function(t){return o(t,0,i)}):o}},pseudos:{not:r(function(t){var e=[],n=[],i=T(t.replace(st,"$1"));return i[V]?r(function(t,e,n,r){for(var o,a=i(t,null,r,[]),s=t.length;s--;)(o=a[s])&&(t[s]=!(e[s]=o))}):function(t,r,o){return e[0]=t,i(e,null,o,n),e[0]=null,!n.pop()}}),has:r(function(t){return function(n){return e(t,n).length>0}}),contains:r(function(t){return t=t.replace(bt,$t),function(e){return(e.textContent||e.innerText||C(e)).indexOf(t)>-1}}),lang:r(function(t){return pt.test(t||"")||e.error("unsupported lang: "+t),t=t.replace(bt,$t).toLowerCase(),function(e){var n;do if(n=R?e.lang:e.getAttribute("xml:lang")||e.getAttribute("lang"))return n=n.toLowerCase(),n===t||0===n.indexOf(t+"-");while((e=e.parentNode)&&1===e.nodeType);return!1}}),target:function(e){var n=t.location&&t.location.hash;return n&&n.slice(1)===e.id},root:function(t){return t===P},focus:function(t){return t===N.activeElement&&(!N.hasFocus||N.hasFocus())&&!!(t.type||t.href||~t.tabIndex)},enabled:l(!1),disabled:l(!0),checked:function(t){var e=t.nodeName.toLowerCase();return"input"===e&&!!t.checked||"option"===e&&!!t.selected},selected:function(t){return t.parentNode&&t.parentNode.selectedIndex,t.selected===!0},empty:function(t){for(t=t.firstChild;t;t=t.nextSibling)if(t.nodeType<6)return!1;return!0},parent:function(t){return!_.pseudos.empty(t)},header:function(t){return vt.test(t.nodeName)},input:function(t){return dt.test(t.nodeName)},button:function(t){var e=t.nodeName.toLowerCase();return"input"===e&&"button"===t.type||"button"===e},text:function(t){var e;return"input"===t.nodeName.toLowerCase()&&"text"===t.type&&(null==(e=t.getAttribute("type"))||"text"===e.toLowerCase())},first:c(function(){return[0]}),last:c(function(t,e){return[e-1]}),eq:c(function(t,e,n){return[n<0?n+e:n]}),even:c(function(t,e){for(var n=0;n<e;n+=2)t.push(n);return t}),odd:c(function(t,e){for(var n=1;n<e;n+=2)t.push(n);return t}),lt:c(function(t,e,n){for(var r=n<0?n+e:n;--r>=0;)t.push(r);return t}),gt:c(function(t,e,n){for(var r=n<0?n+e:n;++r<e;)t.push(r);return t})}},_.pseudos.nth=_.pseudos.eq;for(w in{radio:!0,checkbox:!0,file:!0,password:!0,image:!0})_.pseudos[w]=s(w);for(w in{submit:!0,reset:!0})_.pseudos[w]=u(w);return p.prototype=_.filters=_.pseudos,_.setFilters=new p,S=e.tokenize=function(t,n){var r,i,o,a,s,u,l,c=z[t+" "];if(c)return n?0:c.slice(0);for(s=t,u=[],l=_.preFilter;s;){r&&!(i=ut.exec(s))||(i&&(s=s.slice(i[0].length)||s),u.push(o=[])),r=!1,(i=lt.exec(s))&&(r=i.shift(),o.push({value:r,type:i[0].replace(st," ")}),s=s.slice(r.length));for(a in _.filter)!(i=ht[a].exec(s))||l[a]&&!(i=l[a](i))||(r=i.shift(),o.push({value:r,type:a,matches:i}),s=s.slice(r.length));if(!r)break}return n?s.length:s?e.error(t):z(t,u).slice(0)},T=e.compile=function(t,e){var n,r=[],i=[],o=W[t+" "];if(!o){for(e||(e=S(t)),n=e.length;n--;)o=b(e[n]),o[V]?r.push(o):i.push(o);o=W(t,$(i,r)),o.selector=t}return o},E=e.select=function(t,e,n,r){var i,o,a,s,u,l="function"==typeof t&&t,c=!r&&S(t=l.selector||t);if(n=n||[],1===c.length){if(o=c[0]=c[0].slice(0),o.length>2&&"ID"===(a=o[0]).type&&9===e.nodeType&&R&&_.relative[o[1].type]){if(e=(_.find.ID(a.matches[0].replace(bt,$t),e)||[])[0],!e)return n;l&&(e=e.parentNode),t=t.slice(o.shift().value.length)}for(i=ht.needsContext.test(t)?0:o.length;i--&&(a=o[i],!_.relative[s=a.type]);)if((u=_.find[s])&&(r=u(a.matches[0].replace(bt,$t),yt.test(o[0].type)&&f(e.parentNode)||e))){if(o.splice(i,1),t=r.length&&h(o),!t)return Z.apply(n,r),n;break}}return(l||T(t,c))(r,e,!R,n,!e||yt.test(t)&&f(e.parentNode)||e),n},x.sortStable=V.split("").sort(Y).join("")===V,x.detectDuplicates=!!A,O(),x.sortDetached=i(function(t){return 1&t.compareDocumentPosition(N.createElement("fieldset"))}),i(function(t){return t.innerHTML="<a href='#'></a>","#"===t.firstChild.getAttribute("href")})||o("type|href|height|width",function(t,e,n){if(!n)return t.getAttribute(e,"type"===e.toLowerCase()?1:2)}),x.attributes&&i(function(t){return t.innerHTML="<input/>",t.firstChild.setAttribute("value",""),""===t.firstChild.getAttribute("value")})||o("value",function(t,e,n){if(!n&&"input"===t.nodeName.toLowerCase())return t.defaultValue}),i(function(t){return null==t.getAttribute("disabled")})||o(et,function(t,e,n){var r;if(!n)return t[e]===!0?e.toLowerCase():(r=t.getAttributeNode(e))&&r.specified?r.value:null}),e}(n);yt.find=_t,yt.expr=_t.selectors,yt.expr[":"]=yt.expr.pseudos,yt.uniqueSort=yt.unique=_t.uniqueSort,yt.text=_t.getText,yt.isXMLDoc=_t.isXML,yt.contains=_t.contains,yt.escapeSelector=_t.escape;var Ct=function(t,e,n){for(var r=[],i=void 0!==n;(t=t[e])&&9!==t.nodeType;)if(1===t.nodeType){if(i&&yt(t).is(n))break;r.push(t)}return r},kt=function(t,e){for(var n=[];t;t=t.nextSibling)1===t.nodeType&&t!==e&&n.push(t);return n},St=yt.expr.match.needsContext,Tt=/^<([a-z][^\/\0>:\x20\t\r\n\f]*)[\x20\t\r\n\f]*\/?>(?:<\/\1>|)$/i,Et=/^.[^:#\[\.,]*$/;yt.filter=function(t,e,n){var r=e[0];return n&&(t=":not("+t+")"),1===e.length&&1===r.nodeType?yt.find.matchesSelector(r,t)?[r]:[]:yt.find.matches(t,yt.grep(e,function(t){return 1===t.nodeType}))},yt.fn.extend({find:function(t){var e,n,r=this.length,i=this;if("string"!=typeof t)return this.pushStack(yt(t).filter(function(){for(e=0;e<r;e++)if(yt.contains(i[e],this))return!0}));for(n=this.pushStack([]),e=0;e<r;e++)yt.find(t,i[e],n);return r>1?yt.uniqueSort(n):n},filter:function(t){return this.pushStack(u(this,t||[],!1))},not:function(t){return this.pushStack(u(this,t||[],!0))},is:function(t){return!!u(this,"string"==typeof t&&St.test(t)?yt(t):t||[],!1).length}});var Mt,Dt=/^(?:\s*(<[\w\W]+>)[^>]*|#([\w-]+))$/,At=yt.fn.init=function(t,e,n){var r,i;if(!t)return this;if(n=n||Mt,"string"==typeof t){if(r="<"===t[0]&&">"===t[t.length-1]&&t.length>=3?[null,t,null]:Dt.exec(t),!r||!r[1]&&e)return!e||e.jquery?(e||n).find(t):this.constructor(e).find(t);if(r[1]){if(e=e instanceof yt?e[0]:e,yt.merge(this,yt.parseHTML(r[1],e&&e.nodeType?e.ownerDocument||e:ot,!0)),Tt.test(r[1])&&yt.isPlainObject(e))for(r in e)yt.isFunction(this[r])?this[r](e[r]):this.attr(r,e[r]);return this}return i=ot.getElementById(r[2]),i&&(this[0]=i,this.length=1),this}return t.nodeType?(this[0]=t,this.length=1,this):yt.isFunction(t)?void 0!==n.ready?n.ready(t):t(yt):yt.makeArray(t,this)};At.prototype=yt.fn,Mt=yt(ot);var Ot=/^(?:parents|prev(?:Until|All))/,Nt={children:!0,contents:!0,next:!0,prev:!0};yt.fn.extend({has:function(t){var e=yt(t,this),n=e.length;return this.filter(function(){for(var t=0;t<n;t++)if(yt.contains(this,e[t]))return!0})},closest:function(t,e){var n,r=0,i=this.length,o=[],a="string"!=typeof t&&yt(t);if(!St.test(t))for(;r<i;r++)for(n=this[r];n&&n!==e;n=n.parentNode)if(n.nodeType<11&&(a?a.index(n)>-1:1===n.nodeType&&yt.find.matchesSelector(n,t))){o.push(n);break}return this.pushStack(o.length>1?yt.uniqueSort(o):o)},index:function(t){return t?"string"==typeof t?ct.call(yt(t),this[0]):ct.call(this,t.jquery?t[0]:t):this[0]&&this[0].parentNode?this.first().prevAll().length:-1},add:function(t,e){return this.pushStack(yt.uniqueSort(yt.merge(this.get(),yt(t,e))))},addBack:function(t){return this.add(null==t?this.prevObject:this.prevObject.filter(t))}}),yt.each({parent:function(t){var e=t.parentNode;return e&&11!==e.nodeType?e:null},parents:function(t){return Ct(t,"parentNode")},parentsUntil:function(t,e,n){return Ct(t,"parentNode",n)},next:function(t){return l(t,"nextSibling")},prev:function(t){return l(t,"previousSibling")},nextAll:function(t){return Ct(t,"nextSibling")},prevAll:function(t){return Ct(t,"previousSibling")},nextUntil:function(t,e,n){return Ct(t,"nextSibling",n)},prevUntil:function(t,e,n){return Ct(t,"previousSibling",n)},siblings:function(t){return kt((t.parentNode||{}).firstChild,t)},children:function(t){return kt(t.firstChild)},contents:function(t){return t.contentDocument||yt.merge([],t.childNodes)}},function(t,e){yt.fn[t]=function(n,r){var i=yt.map(this,e,n);return"Until"!==t.slice(-5)&&(r=n),r&&"string"==typeof r&&(i=yt.filter(r,i)),this.length>1&&(Nt[t]||yt.uniqueSort(i),Ot.test(t)&&i.reverse()),this.pushStack(i)}});var Pt=/[^\x20\t\r\n\f]+/g;yt.Callbacks=function(t){t="string"==typeof t?c(t):yt.extend({},t);var e,n,r,i,o=[],a=[],s=-1,u=function(){for(i=t.once,r=e=!0;a.length;s=-1)for(n=a.shift();++s<o.length;)o[s].apply(n[0],n[1])===!1&&t.stopOnFalse&&(s=o.length,n=!1);t.memory||(n=!1),e=!1,i&&(o=n?[]:"")},l={add:function(){return o&&(n&&!e&&(s=o.length-1,a.push(n)),function e(n){yt.each(n,function(n,r){yt.isFunction(r)?t.unique&&l.has(r)||o.push(r):r&&r.length&&"string"!==yt.type(r)&&e(r)})}(arguments),n&&!e&&u()),this},remove:function(){return yt.each(arguments,function(t,e){for(var n;(n=yt.inArray(e,o,n))>-1;)o.splice(n,1),n<=s&&s--}),this},has:function(t){return t?yt.inArray(t,o)>-1:o.length>0},empty:function(){return o&&(o=[]),this},disable:function(){return i=a=[],o=n="",this},disabled:function(){return!o},lock:function(){return i=a=[],n||e||(o=n=""),this},locked:function(){return!!i},fireWith:function(t,n){return i||(n=n||[],n=[t,n.slice?n.slice():n],a.push(n),e||u()),this},fire:function(){return l.fireWith(this,arguments),this},fired:function(){return!!r}};return l},yt.extend({Deferred:function(t){var e=[["notify","progress",yt.Callbacks("memory"),yt.Callbacks("memory"),2],["resolve","done",yt.Callbacks("once memory"),yt.Callbacks("once memory"),0,"resolved"],["reject","fail",yt.Callbacks("once memory"),yt.Callbacks("once memory"),1,"rejected"]],r="pending",i={state:function(){return r},always:function(){return o.done(arguments).fail(arguments),this},catch:function(t){return i.then(null,t)},pipe:function(){var t=arguments;return yt.Deferred(function(n){yt.each(e,function(e,r){var i=yt.isFunction(t[r[4]])&&t[r[4]];o[r[1]](function(){var t=i&&i.apply(this,arguments);t&&yt.isFunction(t.promise)?t.promise().progress(n.notify).done(n.resolve).fail(n.reject):n[r[0]+"With"](this,i?[t]:arguments)})}),t=null}).promise()},then:function(t,r,i){function o(t,e,r,i){return function(){var s=this,u=arguments,l=function(){var n,l;if(!(t<a)){if(n=r.apply(s,u),n===e.promise())throw new TypeError("Thenable self-resolution");l=n&&("object"==typeof n||"function"==typeof n)&&n.then,yt.isFunction(l)?i?l.call(n,o(a,e,f,i),o(a,e,p,i)):(a++,l.call(n,o(a,e,f,i),o(a,e,p,i),o(a,e,f,e.notifyWith))):(r!==f&&(s=void 0,u=[n]),(i||e.resolveWith)(s,u))}},c=i?l:function(){try{l()}catch(n){yt.Deferred.exceptionHook&&yt.Deferred.exceptionHook(n,c.stackTrace),t+1>=a&&(r!==p&&(s=void 0,u=[n]),e.rejectWith(s,u))}};t?c():(yt.Deferred.getStackHook&&(c.stackTrace=yt.Deferred.getStackHook()),n.setTimeout(c))}}var a=0;return yt.Deferred(function(n){e[0][3].add(o(0,n,yt.isFunction(i)?i:f,n.notifyWith)),e[1][3].add(o(0,n,yt.isFunction(t)?t:f)),e[2][3].add(o(0,n,yt.isFunction(r)?r:p))}).promise()},promise:function(t){return null!=t?yt.extend(t,i):i}},o={};return yt.each(e,function(t,n){var a=n[2],s=n[5];i[n[1]]=a.add,s&&a.add(function(){r=s},e[3-t][2].disable,e[0][2].lock),a.add(n[3].fire),o[n[0]]=function(){return o[n[0]+"With"](this===o?void 0:this,arguments),this},o[n[0]+"With"]=a.fireWith}),i.promise(o),t&&t.call(o,o),o},when:function(t){var e=arguments.length,n=e,r=Array(n),i=st.call(arguments),o=yt.Deferred(),a=function(t){return function(n){r[t]=this,i[t]=arguments.length>1?st.call(arguments):n,--e||o.resolveWith(r,i)}};if(e<=1&&(h(t,o.done(a(n)).resolve,o.reject),"pending"===o.state()||yt.isFunction(i[n]&&i[n].then)))return o.then();for(;n--;)h(i[n],a(n),o.reject);return o.promise()}});var Rt=/^(Eval|Internal|Range|Reference|Syntax|Type|URI)Error$/;yt.Deferred.exceptionHook=function(t,e){n.console&&n.console.warn&&t&&Rt.test(t.name)&&n.console.warn("jQuery.Deferred exception: "+t.message,t.stack,e)},yt.readyException=function(t){n.setTimeout(function(){throw t})};var It=yt.Deferred();yt.fn.ready=function(t){return It.then(t).catch(function(t){yt.readyException(t)}),this},yt.extend({isReady:!1,readyWait:1,holdReady:function(t){t?yt.readyWait++:yt.ready(!0)},ready:function(t){(t===!0?--yt.readyWait:yt.isReady)||(yt.isReady=!0,t!==!0&&--yt.readyWait>0||It.resolveWith(ot,[yt]))}}),yt.ready.then=It.then,"complete"===ot.readyState||"loading"!==ot.readyState&&!ot.documentElement.doScroll?n.setTimeout(yt.ready):(ot.addEventListener("DOMContentLoaded",d),n.addEventListener("load",d));var jt=function(t,e,n,r,i,o,a){var s=0,u=t.length,l=null==n;if("object"===yt.type(n)){i=!0;for(s in n)jt(t,e,s,n[s],!0,o,a)}else if(void 0!==r&&(i=!0,yt.isFunction(r)||(a=!0),l&&(a?(e.call(t,r),e=null):(l=e,e=function(t,e,n){return l.call(yt(t),n)})),e))for(;s<u;s++)e(t[s],n,a?r:r.call(t[s],s,e(t[s],n)));return i?t:l?e.call(t):u?e(t[0],n):o},Ft=function(t){return 1===t.nodeType||9===t.nodeType||!+t.nodeType};v.uid=1,v.prototype={cache:function(t){var e=t[this.expando];return e||(e={},Ft(t)&&(t.nodeType?t[this.expando]=e:Object.defineProperty(t,this.expando,{value:e,configurable:!0}))),e},set:function(t,e,n){var r,i=this.cache(t);if("string"==typeof e)i[yt.camelCase(e)]=n;else for(r in e)i[yt.camelCase(r)]=e[r];return i},get:function(t,e){return void 0===e?this.cache(t):t[this.expando]&&t[this.expando][yt.camelCase(e)]},access:function(t,e,n){return void 0===e||e&&"string"==typeof e&&void 0===n?this.get(t,e):(this.set(t,e,n),void 0!==n?n:e)},remove:function(t,e){var n,r=t[this.expando];if(void 0!==r){if(void 0!==e){yt.isArray(e)?e=e.map(yt.camelCase):(e=yt.camelCase(e),e=e in r?[e]:e.match(Pt)||[]),n=e.length;for(;n--;)delete r[e[n]]}(void 0===e||yt.isEmptyObject(r))&&(t.nodeType?t[this.expando]=void 0:delete t[this.expando])}},hasData:function(t){var e=t[this.expando];return void 0!==e&&!yt.isEmptyObject(e)}};var qt=new v,Vt=new v,Ut=/^(?:\{[\w\W]*\}|\[[\w\W]*\])$/,Lt=/[A-Z]/g;yt.extend({hasData:function(t){return Vt.hasData(t)||qt.hasData(t)},data:function(t,e,n){return Vt.access(t,e,n)},removeData:function(t,e){Vt.remove(t,e)},_data:function(t,e,n){return qt.access(t,e,n)},_removeData:function(t,e){qt.remove(t,e)}}),yt.fn.extend({data:function(t,e){var n,r,i,o=this[0],a=o&&o.attributes;if(void 0===t){if(this.length&&(i=Vt.get(o),1===o.nodeType&&!qt.get(o,"hasDataAttrs"))){for(n=a.length;n--;)a[n]&&(r=a[n].name,0===r.indexOf("data-")&&(r=yt.camelCase(r.slice(5)),m(o,r,i[r])));qt.set(o,"hasDataAttrs",!0)}return i}return"object"==typeof t?this.each(function(){Vt.set(this,t)}):jt(this,function(e){var n;if(o&&void 0===e){if(n=Vt.get(o,t),void 0!==n)return n;if(n=m(o,t),void 0!==n)return n}else this.each(function(){Vt.set(this,t,e)})},null,e,arguments.length>1,null,!0)},removeData:function(t){return this.each(function(){Vt.remove(this,t)})}}),yt.extend({queue:function(t,e,n){var r;if(t)return e=(e||"fx")+"queue",r=qt.get(t,e),n&&(!r||yt.isArray(n)?r=qt.access(t,e,yt.makeArray(n)):r.push(n)),r||[]},dequeue:function(t,e){e=e||"fx";var n=yt.queue(t,e),r=n.length,i=n.shift(),o=yt._queueHooks(t,e),a=function(){yt.dequeue(t,e)};"inprogress"===i&&(i=n.shift(),r--),i&&("fx"===e&&n.unshift("inprogress"),delete o.stop,i.call(t,a,o)),!r&&o&&o.empty.fire()},_queueHooks:function(t,e){var n=e+"queueHooks";return qt.get(t,n)||qt.access(t,n,{empty:yt.Callbacks("once memory").add(function(){qt.remove(t,[e+"queue",n])})})}}),yt.fn.extend({queue:function(t,e){var n=2;return"string"!=typeof t&&(e=t,t="fx",n--),arguments.length<n?yt.queue(this[0],t):void 0===e?this:this.each(function(){var n=yt.queue(this,t,e);yt._queueHooks(this,t),"fx"===t&&"inprogress"!==n[0]&&yt.dequeue(this,t)})},dequeue:function(t){return this.each(function(){yt.dequeue(this,t)})},clearQueue:function(t){return this.queue(t||"fx",[])},promise:function(t,e){var n,r=1,i=yt.Deferred(),o=this,a=this.length,s=function(){--r||i.resolveWith(o,[o])};for("string"!=typeof t&&(e=t,t=void 0),t=t||"fx";a--;)n=qt.get(o[a],t+"queueHooks"),n&&n.empty&&(r++,n.empty.add(s));return s(),i.promise(e)}});var Ht=/[+-]?(?:\d*\.|)\d+(?:[eE][+-]?\d+|)/.source,Bt=new RegExp("^(?:([+-])=|)("+Ht+")([a-z%]*)$","i"),zt=["Top","Right","Bottom","Left"],Wt=function(t,e){return t=e||t,"none"===t.style.display||""===t.style.display&&yt.contains(t.ownerDocument,t)&&"none"===yt.css(t,"display")},Yt=function(t,e,n,r){var i,o,a={};for(o in e)a[o]=t.style[o],t.style[o]=e[o];i=n.apply(t,r||[]);for(o in e)t.style[o]=a[o];return i},Gt={};yt.fn.extend({show:function(){return $(this,!0)},hide:function(){return $(this)},toggle:function(t){return"boolean"==typeof t?t?this.show():this.hide():this.each(function(){Wt(this)?yt(this).show():yt(this).hide()})}});var Xt=/^(?:checkbox|radio)$/i,Kt=/<([a-z][^\/\0>\x20\t\r\n\f]+)/i,Jt=/^$|\/(?:java|ecma)script/i,Zt={option:[1,"<select multiple='multiple'>","</select>"],thead:[1,"<table>","</table>"],col:[2,"<table><colgroup>","</colgroup></table>"],tr:[2,"<table><tbody>","</tbody></table>"],td:[3,"<table><tbody><tr>","</tr></tbody></table>"],_default:[0,"",""]};Zt.optgroup=Zt.option,Zt.tbody=Zt.tfoot=Zt.colgroup=Zt.caption=Zt.thead,Zt.th=Zt.td;
var Qt=/<|&#?\w+;/;!function(){var t=ot.createDocumentFragment(),e=t.appendChild(ot.createElement("div")),n=ot.createElement("input");n.setAttribute("type","radio"),n.setAttribute("checked","checked"),n.setAttribute("name","t"),e.appendChild(n),gt.checkClone=e.cloneNode(!0).cloneNode(!0).lastChild.checked,e.innerHTML="<textarea>x</textarea>",gt.noCloneChecked=!!e.cloneNode(!0).lastChild.defaultValue}();var te=ot.documentElement,ee=/^key/,ne=/^(?:mouse|pointer|contextmenu|drag|drop)|click/,re=/^([^.]*)(?:\.(.+)|)/;yt.event={global:{},add:function(t,e,n,r,i){var o,a,s,u,l,c,f,p,h,d,v,g=qt.get(t);if(g)for(n.handler&&(o=n,n=o.handler,i=o.selector),i&&yt.find.matchesSelector(te,i),n.guid||(n.guid=yt.guid++),(u=g.events)||(u=g.events={}),(a=g.handle)||(a=g.handle=function(e){return"undefined"!=typeof yt&&yt.event.triggered!==e.type?yt.event.dispatch.apply(t,arguments):void 0}),e=(e||"").match(Pt)||[""],l=e.length;l--;)s=re.exec(e[l])||[],h=v=s[1],d=(s[2]||"").split(".").sort(),h&&(f=yt.event.special[h]||{},h=(i?f.delegateType:f.bindType)||h,f=yt.event.special[h]||{},c=yt.extend({type:h,origType:v,data:r,handler:n,guid:n.guid,selector:i,needsContext:i&&yt.expr.match.needsContext.test(i),namespace:d.join(".")},o),(p=u[h])||(p=u[h]=[],p.delegateCount=0,f.setup&&f.setup.call(t,r,d,a)!==!1||t.addEventListener&&t.addEventListener(h,a)),f.add&&(f.add.call(t,c),c.handler.guid||(c.handler.guid=n.guid)),i?p.splice(p.delegateCount++,0,c):p.push(c),yt.event.global[h]=!0)},remove:function(t,e,n,r,i){var o,a,s,u,l,c,f,p,h,d,v,g=qt.hasData(t)&&qt.get(t);if(g&&(u=g.events)){for(e=(e||"").match(Pt)||[""],l=e.length;l--;)if(s=re.exec(e[l])||[],h=v=s[1],d=(s[2]||"").split(".").sort(),h){for(f=yt.event.special[h]||{},h=(r?f.delegateType:f.bindType)||h,p=u[h]||[],s=s[2]&&new RegExp("(^|\\.)"+d.join("\\.(?:.*\\.|)")+"(\\.|$)"),a=o=p.length;o--;)c=p[o],!i&&v!==c.origType||n&&n.guid!==c.guid||s&&!s.test(c.namespace)||r&&r!==c.selector&&("**"!==r||!c.selector)||(p.splice(o,1),c.selector&&p.delegateCount--,f.remove&&f.remove.call(t,c));a&&!p.length&&(f.teardown&&f.teardown.call(t,d,g.handle)!==!1||yt.removeEvent(t,h,g.handle),delete u[h])}else for(h in u)yt.event.remove(t,h+e[l],n,r,!0);yt.isEmptyObject(u)&&qt.remove(t,"handle events")}},dispatch:function(t){var e,n,r,i,o,a,s=yt.event.fix(t),u=new Array(arguments.length),l=(qt.get(this,"events")||{})[s.type]||[],c=yt.event.special[s.type]||{};for(u[0]=s,e=1;e<arguments.length;e++)u[e]=arguments[e];if(s.delegateTarget=this,!c.preDispatch||c.preDispatch.call(this,s)!==!1){for(a=yt.event.handlers.call(this,s,l),e=0;(i=a[e++])&&!s.isPropagationStopped();)for(s.currentTarget=i.elem,n=0;(o=i.handlers[n++])&&!s.isImmediatePropagationStopped();)s.rnamespace&&!s.rnamespace.test(o.namespace)||(s.handleObj=o,s.data=o.data,r=((yt.event.special[o.origType]||{}).handle||o.handler).apply(i.elem,u),void 0!==r&&(s.result=r)===!1&&(s.preventDefault(),s.stopPropagation()));return c.postDispatch&&c.postDispatch.call(this,s),s.result}},handlers:function(t,e){var n,r,i,o,a,s=[],u=e.delegateCount,l=t.target;if(u&&l.nodeType&&!("click"===t.type&&t.button>=1))for(;l!==this;l=l.parentNode||this)if(1===l.nodeType&&("click"!==t.type||l.disabled!==!0)){for(o=[],a={},n=0;n<u;n++)r=e[n],i=r.selector+" ",void 0===a[i]&&(a[i]=r.needsContext?yt(i,this).index(l)>-1:yt.find(i,this,null,[l]).length),a[i]&&o.push(r);o.length&&s.push({elem:l,handlers:o})}return l=this,u<e.length&&s.push({elem:l,handlers:e.slice(u)}),s},addProp:function(t,e){Object.defineProperty(yt.Event.prototype,t,{enumerable:!0,configurable:!0,get:yt.isFunction(e)?function(){if(this.originalEvent)return e(this.originalEvent)}:function(){if(this.originalEvent)return this.originalEvent[t]},set:function(e){Object.defineProperty(this,t,{enumerable:!0,configurable:!0,writable:!0,value:e})}})},fix:function(t){return t[yt.expando]?t:new yt.Event(t)},special:{load:{noBubble:!0},focus:{trigger:function(){if(this!==S()&&this.focus)return this.focus(),!1},delegateType:"focusin"},blur:{trigger:function(){if(this===S()&&this.blur)return this.blur(),!1},delegateType:"focusout"},click:{trigger:function(){if("checkbox"===this.type&&this.click&&yt.nodeName(this,"input"))return this.click(),!1},_default:function(t){return yt.nodeName(t.target,"a")}},beforeunload:{postDispatch:function(t){void 0!==t.result&&t.originalEvent&&(t.originalEvent.returnValue=t.result)}}}},yt.removeEvent=function(t,e,n){t.removeEventListener&&t.removeEventListener(e,n)},yt.Event=function(t,e){return this instanceof yt.Event?(t&&t.type?(this.originalEvent=t,this.type=t.type,this.isDefaultPrevented=t.defaultPrevented||void 0===t.defaultPrevented&&t.returnValue===!1?C:k,this.target=t.target&&3===t.target.nodeType?t.target.parentNode:t.target,this.currentTarget=t.currentTarget,this.relatedTarget=t.relatedTarget):this.type=t,e&&yt.extend(this,e),this.timeStamp=t&&t.timeStamp||yt.now(),void(this[yt.expando]=!0)):new yt.Event(t,e)},yt.Event.prototype={constructor:yt.Event,isDefaultPrevented:k,isPropagationStopped:k,isImmediatePropagationStopped:k,isSimulated:!1,preventDefault:function(){var t=this.originalEvent;this.isDefaultPrevented=C,t&&!this.isSimulated&&t.preventDefault()},stopPropagation:function(){var t=this.originalEvent;this.isPropagationStopped=C,t&&!this.isSimulated&&t.stopPropagation()},stopImmediatePropagation:function(){var t=this.originalEvent;this.isImmediatePropagationStopped=C,t&&!this.isSimulated&&t.stopImmediatePropagation(),this.stopPropagation()}},yt.each({altKey:!0,bubbles:!0,cancelable:!0,changedTouches:!0,ctrlKey:!0,detail:!0,eventPhase:!0,metaKey:!0,pageX:!0,pageY:!0,shiftKey:!0,view:!0,char:!0,charCode:!0,key:!0,keyCode:!0,button:!0,buttons:!0,clientX:!0,clientY:!0,offsetX:!0,offsetY:!0,pointerId:!0,pointerType:!0,screenX:!0,screenY:!0,targetTouches:!0,toElement:!0,touches:!0,which:function(t){var e=t.button;return null==t.which&&ee.test(t.type)?null!=t.charCode?t.charCode:t.keyCode:!t.which&&void 0!==e&&ne.test(t.type)?1&e?1:2&e?3:4&e?2:0:t.which}},yt.event.addProp),yt.each({mouseenter:"mouseover",mouseleave:"mouseout",pointerenter:"pointerover",pointerleave:"pointerout"},function(t,e){yt.event.special[t]={delegateType:e,bindType:e,handle:function(t){var n,r=this,i=t.relatedTarget,o=t.handleObj;return i&&(i===r||yt.contains(r,i))||(t.type=o.origType,n=o.handler.apply(this,arguments),t.type=e),n}}}),yt.fn.extend({on:function(t,e,n,r){return T(this,t,e,n,r)},one:function(t,e,n,r){return T(this,t,e,n,r,1)},off:function(t,e,n){var r,i;if(t&&t.preventDefault&&t.handleObj)return r=t.handleObj,yt(t.delegateTarget).off(r.namespace?r.origType+"."+r.namespace:r.origType,r.selector,r.handler),this;if("object"==typeof t){for(i in t)this.off(i,e,t[i]);return this}return e!==!1&&"function"!=typeof e||(n=e,e=void 0),n===!1&&(n=k),this.each(function(){yt.event.remove(this,t,n,e)})}});var ie=/<(?!area|br|col|embed|hr|img|input|link|meta|param)(([a-z][^\/\0>\x20\t\r\n\f]*)[^>]*)\/>/gi,oe=/<script|<style|<link/i,ae=/checked\s*(?:[^=]|=\s*.checked.)/i,se=/^true\/(.*)/,ue=/^\s*<!(?:\[CDATA\[|--)|(?:\]\]|--)>\s*$/g;yt.extend({htmlPrefilter:function(t){return t.replace(ie,"<$1></$2>")},clone:function(t,e,n){var r,i,o,a,s=t.cloneNode(!0),u=yt.contains(t.ownerDocument,t);if(!(gt.noCloneChecked||1!==t.nodeType&&11!==t.nodeType||yt.isXMLDoc(t)))for(a=w(s),o=w(t),r=0,i=o.length;r<i;r++)O(o[r],a[r]);if(e)if(n)for(o=o||w(t),a=a||w(s),r=0,i=o.length;r<i;r++)A(o[r],a[r]);else A(t,s);return a=w(s,"script"),a.length>0&&x(a,!u&&w(t,"script")),s},cleanData:function(t){for(var e,n,r,i=yt.event.special,o=0;void 0!==(n=t[o]);o++)if(Ft(n)){if(e=n[qt.expando]){if(e.events)for(r in e.events)i[r]?yt.event.remove(n,r):yt.removeEvent(n,r,e.handle);n[qt.expando]=void 0}n[Vt.expando]&&(n[Vt.expando]=void 0)}}}),yt.fn.extend({detach:function(t){return P(this,t,!0)},remove:function(t){return P(this,t)},text:function(t){return jt(this,function(t){return void 0===t?yt.text(this):this.empty().each(function(){1!==this.nodeType&&11!==this.nodeType&&9!==this.nodeType||(this.textContent=t)})},null,t,arguments.length)},append:function(){return N(this,arguments,function(t){if(1===this.nodeType||11===this.nodeType||9===this.nodeType){var e=E(this,t);e.appendChild(t)}})},prepend:function(){return N(this,arguments,function(t){if(1===this.nodeType||11===this.nodeType||9===this.nodeType){var e=E(this,t);e.insertBefore(t,e.firstChild)}})},before:function(){return N(this,arguments,function(t){this.parentNode&&this.parentNode.insertBefore(t,this)})},after:function(){return N(this,arguments,function(t){this.parentNode&&this.parentNode.insertBefore(t,this.nextSibling)})},empty:function(){for(var t,e=0;null!=(t=this[e]);e++)1===t.nodeType&&(yt.cleanData(w(t,!1)),t.textContent="");return this},clone:function(t,e){return t=null!=t&&t,e=null==e?t:e,this.map(function(){return yt.clone(this,t,e)})},html:function(t){return jt(this,function(t){var e=this[0]||{},n=0,r=this.length;if(void 0===t&&1===e.nodeType)return e.innerHTML;if("string"==typeof t&&!oe.test(t)&&!Zt[(Kt.exec(t)||["",""])[1].toLowerCase()]){t=yt.htmlPrefilter(t);try{for(;n<r;n++)e=this[n]||{},1===e.nodeType&&(yt.cleanData(w(e,!1)),e.innerHTML=t);e=0}catch(t){}}e&&this.empty().append(t)},null,t,arguments.length)},replaceWith:function(){var t=[];return N(this,arguments,function(e){var n=this.parentNode;yt.inArray(this,t)<0&&(yt.cleanData(w(this)),n&&n.replaceChild(e,this))},t)}}),yt.each({appendTo:"append",prependTo:"prepend",insertBefore:"before",insertAfter:"after",replaceAll:"replaceWith"},function(t,e){yt.fn[t]=function(t){for(var n,r=[],i=yt(t),o=i.length-1,a=0;a<=o;a++)n=a===o?this:this.clone(!0),yt(i[a])[e](n),lt.apply(r,n.get());return this.pushStack(r)}});var le=/^margin/,ce=new RegExp("^("+Ht+")(?!px)[a-z%]+$","i"),fe=function(t){var e=t.ownerDocument.defaultView;return e&&e.opener||(e=n),e.getComputedStyle(t)};!function(){function t(){if(s){s.style.cssText="box-sizing:border-box;position:relative;display:block;margin:auto;border:1px;padding:1px;top:1%;width:50%",s.innerHTML="",te.appendChild(a);var t=n.getComputedStyle(s);e="1%"!==t.top,o="2px"===t.marginLeft,r="4px"===t.width,s.style.marginRight="50%",i="4px"===t.marginRight,te.removeChild(a),s=null}}var e,r,i,o,a=ot.createElement("div"),s=ot.createElement("div");s.style&&(s.style.backgroundClip="content-box",s.cloneNode(!0).style.backgroundClip="",gt.clearCloneStyle="content-box"===s.style.backgroundClip,a.style.cssText="border:0;width:8px;height:0;top:0;left:-9999px;padding:0;margin-top:1px;position:absolute",a.appendChild(s),yt.extend(gt,{pixelPosition:function(){return t(),e},boxSizingReliable:function(){return t(),r},pixelMarginRight:function(){return t(),i},reliableMarginLeft:function(){return t(),o}}))}();var pe=/^(none|table(?!-c[ea]).+)/,he={position:"absolute",visibility:"hidden",display:"block"},de={letterSpacing:"0",fontWeight:"400"},ve=["Webkit","Moz","ms"],ge=ot.createElement("div").style;yt.extend({cssHooks:{opacity:{get:function(t,e){if(e){var n=R(t,"opacity");return""===n?"1":n}}}},cssNumber:{animationIterationCount:!0,columnCount:!0,fillOpacity:!0,flexGrow:!0,flexShrink:!0,fontWeight:!0,lineHeight:!0,opacity:!0,order:!0,orphans:!0,widows:!0,zIndex:!0,zoom:!0},cssProps:{float:"cssFloat"},style:function(t,e,n,r){if(t&&3!==t.nodeType&&8!==t.nodeType&&t.style){var i,o,a,s=yt.camelCase(e),u=t.style;return e=yt.cssProps[s]||(yt.cssProps[s]=j(s)||s),a=yt.cssHooks[e]||yt.cssHooks[s],void 0===n?a&&"get"in a&&void 0!==(i=a.get(t,!1,r))?i:u[e]:(o=typeof n,"string"===o&&(i=Bt.exec(n))&&i[1]&&(n=y(t,e,i),o="number"),null!=n&&n===n&&("number"===o&&(n+=i&&i[3]||(yt.cssNumber[s]?"":"px")),gt.clearCloneStyle||""!==n||0!==e.indexOf("background")||(u[e]="inherit"),a&&"set"in a&&void 0===(n=a.set(t,n,r))||(u[e]=n)),void 0)}},css:function(t,e,n,r){var i,o,a,s=yt.camelCase(e);return e=yt.cssProps[s]||(yt.cssProps[s]=j(s)||s),a=yt.cssHooks[e]||yt.cssHooks[s],a&&"get"in a&&(i=a.get(t,!0,n)),void 0===i&&(i=R(t,e,r)),"normal"===i&&e in de&&(i=de[e]),""===n||n?(o=parseFloat(i),n===!0||isFinite(o)?o||0:i):i}}),yt.each(["height","width"],function(t,e){yt.cssHooks[e]={get:function(t,n,r){if(n)return!pe.test(yt.css(t,"display"))||t.getClientRects().length&&t.getBoundingClientRect().width?V(t,e,r):Yt(t,he,function(){return V(t,e,r)})},set:function(t,n,r){var i,o=r&&fe(t),a=r&&q(t,e,r,"border-box"===yt.css(t,"boxSizing",!1,o),o);return a&&(i=Bt.exec(n))&&"px"!==(i[3]||"px")&&(t.style[e]=n,n=yt.css(t,e)),F(t,n,a)}}}),yt.cssHooks.marginLeft=I(gt.reliableMarginLeft,function(t,e){if(e)return(parseFloat(R(t,"marginLeft"))||t.getBoundingClientRect().left-Yt(t,{marginLeft:0},function(){return t.getBoundingClientRect().left}))+"px"}),yt.each({margin:"",padding:"",border:"Width"},function(t,e){yt.cssHooks[t+e]={expand:function(n){for(var r=0,i={},o="string"==typeof n?n.split(" "):[n];r<4;r++)i[t+zt[r]+e]=o[r]||o[r-2]||o[0];return i}},le.test(t)||(yt.cssHooks[t+e].set=F)}),yt.fn.extend({css:function(t,e){return jt(this,function(t,e,n){var r,i,o={},a=0;if(yt.isArray(e)){for(r=fe(t),i=e.length;a<i;a++)o[e[a]]=yt.css(t,e[a],!1,r);return o}return void 0!==n?yt.style(t,e,n):yt.css(t,e)},t,e,arguments.length>1)}}),yt.Tween=U,U.prototype={constructor:U,init:function(t,e,n,r,i,o){this.elem=t,this.prop=n,this.easing=i||yt.easing._default,this.options=e,this.start=this.now=this.cur(),this.end=r,this.unit=o||(yt.cssNumber[n]?"":"px")},cur:function(){var t=U.propHooks[this.prop];return t&&t.get?t.get(this):U.propHooks._default.get(this)},run:function(t){var e,n=U.propHooks[this.prop];return this.options.duration?this.pos=e=yt.easing[this.easing](t,this.options.duration*t,0,1,this.options.duration):this.pos=e=t,this.now=(this.end-this.start)*e+this.start,this.options.step&&this.options.step.call(this.elem,this.now,this),n&&n.set?n.set(this):U.propHooks._default.set(this),this}},U.prototype.init.prototype=U.prototype,U.propHooks={_default:{get:function(t){var e;return 1!==t.elem.nodeType||null!=t.elem[t.prop]&&null==t.elem.style[t.prop]?t.elem[t.prop]:(e=yt.css(t.elem,t.prop,""),e&&"auto"!==e?e:0)},set:function(t){yt.fx.step[t.prop]?yt.fx.step[t.prop](t):1!==t.elem.nodeType||null==t.elem.style[yt.cssProps[t.prop]]&&!yt.cssHooks[t.prop]?t.elem[t.prop]=t.now:yt.style(t.elem,t.prop,t.now+t.unit)}}},U.propHooks.scrollTop=U.propHooks.scrollLeft={set:function(t){t.elem.nodeType&&t.elem.parentNode&&(t.elem[t.prop]=t.now)}},yt.easing={linear:function(t){return t},swing:function(t){return.5-Math.cos(t*Math.PI)/2},_default:"swing"},yt.fx=U.prototype.init,yt.fx.step={};var me,ye,be=/^(?:toggle|show|hide)$/,$e=/queueHooks$/;yt.Animation=yt.extend(G,{tweeners:{"*":[function(t,e){var n=this.createTween(t,e);return y(n.elem,t,Bt.exec(e),n),n}]},tweener:function(t,e){yt.isFunction(t)?(e=t,t=["*"]):t=t.match(Pt);for(var n,r=0,i=t.length;r<i;r++)n=t[r],G.tweeners[n]=G.tweeners[n]||[],G.tweeners[n].unshift(e)},prefilters:[W],prefilter:function(t,e){e?G.prefilters.unshift(t):G.prefilters.push(t)}}),yt.speed=function(t,e,n){var r=t&&"object"==typeof t?yt.extend({},t):{complete:n||!n&&e||yt.isFunction(t)&&t,duration:t,easing:n&&e||e&&!yt.isFunction(e)&&e};return yt.fx.off||ot.hidden?r.duration=0:"number"!=typeof r.duration&&(r.duration in yt.fx.speeds?r.duration=yt.fx.speeds[r.duration]:r.duration=yt.fx.speeds._default),null!=r.queue&&r.queue!==!0||(r.queue="fx"),r.old=r.complete,r.complete=function(){yt.isFunction(r.old)&&r.old.call(this),r.queue&&yt.dequeue(this,r.queue)},r},yt.fn.extend({fadeTo:function(t,e,n,r){return this.filter(Wt).css("opacity",0).show().end().animate({opacity:e},t,n,r)},animate:function(t,e,n,r){var i=yt.isEmptyObject(t),o=yt.speed(e,n,r),a=function(){var e=G(this,yt.extend({},t),o);(i||qt.get(this,"finish"))&&e.stop(!0)};return a.finish=a,i||o.queue===!1?this.each(a):this.queue(o.queue,a)},stop:function(t,e,n){var r=function(t){var e=t.stop;delete t.stop,e(n)};return"string"!=typeof t&&(n=e,e=t,t=void 0),e&&t!==!1&&this.queue(t||"fx",[]),this.each(function(){var e=!0,i=null!=t&&t+"queueHooks",o=yt.timers,a=qt.get(this);if(i)a[i]&&a[i].stop&&r(a[i]);else for(i in a)a[i]&&a[i].stop&&$e.test(i)&&r(a[i]);for(i=o.length;i--;)o[i].elem!==this||null!=t&&o[i].queue!==t||(o[i].anim.stop(n),e=!1,o.splice(i,1));!e&&n||yt.dequeue(this,t)})},finish:function(t){return t!==!1&&(t=t||"fx"),this.each(function(){var e,n=qt.get(this),r=n[t+"queue"],i=n[t+"queueHooks"],o=yt.timers,a=r?r.length:0;for(n.finish=!0,yt.queue(this,t,[]),i&&i.stop&&i.stop.call(this,!0),e=o.length;e--;)o[e].elem===this&&o[e].queue===t&&(o[e].anim.stop(!0),o.splice(e,1));for(e=0;e<a;e++)r[e]&&r[e].finish&&r[e].finish.call(this);delete n.finish})}}),yt.each(["toggle","show","hide"],function(t,e){var n=yt.fn[e];yt.fn[e]=function(t,r,i){return null==t||"boolean"==typeof t?n.apply(this,arguments):this.animate(B(e,!0),t,r,i)}}),yt.each({slideDown:B("show"),slideUp:B("hide"),slideToggle:B("toggle"),fadeIn:{opacity:"show"},fadeOut:{opacity:"hide"},fadeToggle:{opacity:"toggle"}},function(t,e){yt.fn[t]=function(t,n,r){return this.animate(e,t,n,r)}}),yt.timers=[],yt.fx.tick=function(){var t,e=0,n=yt.timers;for(me=yt.now();e<n.length;e++)t=n[e],t()||n[e]!==t||n.splice(e--,1);n.length||yt.fx.stop(),me=void 0},yt.fx.timer=function(t){yt.timers.push(t),t()?yt.fx.start():yt.timers.pop()},yt.fx.interval=13,yt.fx.start=function(){ye||(ye=n.requestAnimationFrame?n.requestAnimationFrame(L):n.setInterval(yt.fx.tick,yt.fx.interval))},yt.fx.stop=function(){n.cancelAnimationFrame?n.cancelAnimationFrame(ye):n.clearInterval(ye),ye=null},yt.fx.speeds={slow:600,fast:200,_default:400},yt.fn.delay=function(t,e){return t=yt.fx?yt.fx.speeds[t]||t:t,e=e||"fx",this.queue(e,function(e,r){var i=n.setTimeout(e,t);r.stop=function(){n.clearTimeout(i)}})},function(){var t=ot.createElement("input"),e=ot.createElement("select"),n=e.appendChild(ot.createElement("option"));t.type="checkbox",gt.checkOn=""!==t.value,gt.optSelected=n.selected,t=ot.createElement("input"),t.value="t",t.type="radio",gt.radioValue="t"===t.value}();var we,xe=yt.expr.attrHandle;yt.fn.extend({attr:function(t,e){return jt(this,yt.attr,t,e,arguments.length>1)},removeAttr:function(t){return this.each(function(){yt.removeAttr(this,t)})}}),yt.extend({attr:function(t,e,n){var r,i,o=t.nodeType;if(3!==o&&8!==o&&2!==o)return"undefined"==typeof t.getAttribute?yt.prop(t,e,n):(1===o&&yt.isXMLDoc(t)||(i=yt.attrHooks[e.toLowerCase()]||(yt.expr.match.bool.test(e)?we:void 0)),void 0!==n?null===n?void yt.removeAttr(t,e):i&&"set"in i&&void 0!==(r=i.set(t,n,e))?r:(t.setAttribute(e,n+""),n):i&&"get"in i&&null!==(r=i.get(t,e))?r:(r=yt.find.attr(t,e),null==r?void 0:r))},attrHooks:{type:{set:function(t,e){if(!gt.radioValue&&"radio"===e&&yt.nodeName(t,"input")){var n=t.value;return t.setAttribute("type",e),n&&(t.value=n),e}}}},removeAttr:function(t,e){var n,r=0,i=e&&e.match(Pt);if(i&&1===t.nodeType)for(;n=i[r++];)t.removeAttribute(n)}}),we={set:function(t,e,n){return e===!1?yt.removeAttr(t,n):t.setAttribute(n,n),n}},yt.each(yt.expr.match.bool.source.match(/\w+/g),function(t,e){var n=xe[e]||yt.find.attr;xe[e]=function(t,e,r){var i,o,a=e.toLowerCase();return r||(o=xe[a],xe[a]=i,i=null!=n(t,e,r)?a:null,xe[a]=o),i}});var _e=/^(?:input|select|textarea|button)$/i,Ce=/^(?:a|area)$/i;yt.fn.extend({prop:function(t,e){return jt(this,yt.prop,t,e,arguments.length>1)},removeProp:function(t){return this.each(function(){delete this[yt.propFix[t]||t]})}}),yt.extend({prop:function(t,e,n){var r,i,o=t.nodeType;if(3!==o&&8!==o&&2!==o)return 1===o&&yt.isXMLDoc(t)||(e=yt.propFix[e]||e,i=yt.propHooks[e]),void 0!==n?i&&"set"in i&&void 0!==(r=i.set(t,n,e))?r:t[e]=n:i&&"get"in i&&null!==(r=i.get(t,e))?r:t[e]},propHooks:{tabIndex:{get:function(t){var e=yt.find.attr(t,"tabindex");return e?parseInt(e,10):_e.test(t.nodeName)||Ce.test(t.nodeName)&&t.href?0:-1}}},propFix:{for:"htmlFor",class:"className"}}),gt.optSelected||(yt.propHooks.selected={get:function(t){var e=t.parentNode;return e&&e.parentNode&&e.parentNode.selectedIndex,null},set:function(t){var e=t.parentNode;e&&(e.selectedIndex,e.parentNode&&e.parentNode.selectedIndex)}}),yt.each(["tabIndex","readOnly","maxLength","cellSpacing","cellPadding","rowSpan","colSpan","useMap","frameBorder","contentEditable"],function(){yt.propFix[this.toLowerCase()]=this}),yt.fn.extend({addClass:function(t){var e,n,r,i,o,a,s,u=0;if(yt.isFunction(t))return this.each(function(e){yt(this).addClass(t.call(this,e,K(this)))});if("string"==typeof t&&t)for(e=t.match(Pt)||[];n=this[u++];)if(i=K(n),r=1===n.nodeType&&" "+X(i)+" "){for(a=0;o=e[a++];)r.indexOf(" "+o+" ")<0&&(r+=o+" ");s=X(r),i!==s&&n.setAttribute("class",s)}return this},removeClass:function(t){var e,n,r,i,o,a,s,u=0;if(yt.isFunction(t))return this.each(function(e){yt(this).removeClass(t.call(this,e,K(this)))});if(!arguments.length)return this.attr("class","");if("string"==typeof t&&t)for(e=t.match(Pt)||[];n=this[u++];)if(i=K(n),r=1===n.nodeType&&" "+X(i)+" "){for(a=0;o=e[a++];)for(;r.indexOf(" "+o+" ")>-1;)r=r.replace(" "+o+" "," ");s=X(r),i!==s&&n.setAttribute("class",s)}return this},toggleClass:function(t,e){var n=typeof t;return"boolean"==typeof e&&"string"===n?e?this.addClass(t):this.removeClass(t):yt.isFunction(t)?this.each(function(n){yt(this).toggleClass(t.call(this,n,K(this),e),e)}):this.each(function(){var e,r,i,o;if("string"===n)for(r=0,i=yt(this),o=t.match(Pt)||[];e=o[r++];)i.hasClass(e)?i.removeClass(e):i.addClass(e);else void 0!==t&&"boolean"!==n||(e=K(this),e&&qt.set(this,"__className__",e),this.setAttribute&&this.setAttribute("class",e||t===!1?"":qt.get(this,"__className__")||""))})},hasClass:function(t){var e,n,r=0;for(e=" "+t+" ";n=this[r++];)if(1===n.nodeType&&(" "+X(K(n))+" ").indexOf(e)>-1)return!0;return!1}});var ke=/\r/g;yt.fn.extend({val:function(t){var e,n,r,i=this[0];{if(arguments.length)return r=yt.isFunction(t),this.each(function(n){var i;1===this.nodeType&&(i=r?t.call(this,n,yt(this).val()):t,null==i?i="":"number"==typeof i?i+="":yt.isArray(i)&&(i=yt.map(i,function(t){return null==t?"":t+""})),e=yt.valHooks[this.type]||yt.valHooks[this.nodeName.toLowerCase()],e&&"set"in e&&void 0!==e.set(this,i,"value")||(this.value=i))});if(i)return e=yt.valHooks[i.type]||yt.valHooks[i.nodeName.toLowerCase()],e&&"get"in e&&void 0!==(n=e.get(i,"value"))?n:(n=i.value,"string"==typeof n?n.replace(ke,""):null==n?"":n)}}}),yt.extend({valHooks:{option:{get:function(t){var e=yt.find.attr(t,"value");return null!=e?e:X(yt.text(t))}},select:{get:function(t){var e,n,r,i=t.options,o=t.selectedIndex,a="select-one"===t.type,s=a?null:[],u=a?o+1:i.length;for(r=o<0?u:a?o:0;r<u;r++)if(n=i[r],(n.selected||r===o)&&!n.disabled&&(!n.parentNode.disabled||!yt.nodeName(n.parentNode,"optgroup"))){if(e=yt(n).val(),a)return e;s.push(e)}return s},set:function(t,e){for(var n,r,i=t.options,o=yt.makeArray(e),a=i.length;a--;)r=i[a],(r.selected=yt.inArray(yt.valHooks.option.get(r),o)>-1)&&(n=!0);return n||(t.selectedIndex=-1),o}}}}),yt.each(["radio","checkbox"],function(){yt.valHooks[this]={set:function(t,e){if(yt.isArray(e))return t.checked=yt.inArray(yt(t).val(),e)>-1}},gt.checkOn||(yt.valHooks[this].get=function(t){return null===t.getAttribute("value")?"on":t.value})});var Se=/^(?:focusinfocus|focusoutblur)$/;yt.extend(yt.event,{trigger:function(t,e,r,i){var o,a,s,u,l,c,f,p=[r||ot],h=ht.call(t,"type")?t.type:t,d=ht.call(t,"namespace")?t.namespace.split("."):[];if(a=s=r=r||ot,3!==r.nodeType&&8!==r.nodeType&&!Se.test(h+yt.event.triggered)&&(h.indexOf(".")>-1&&(d=h.split("."),h=d.shift(),d.sort()),l=h.indexOf(":")<0&&"on"+h,t=t[yt.expando]?t:new yt.Event(h,"object"==typeof t&&t),t.isTrigger=i?2:3,t.namespace=d.join("."),t.rnamespace=t.namespace?new RegExp("(^|\\.)"+d.join("\\.(?:.*\\.|)")+"(\\.|$)"):null,t.result=void 0,t.target||(t.target=r),e=null==e?[t]:yt.makeArray(e,[t]),f=yt.event.special[h]||{},i||!f.trigger||f.trigger.apply(r,e)!==!1)){if(!i&&!f.noBubble&&!yt.isWindow(r)){for(u=f.delegateType||h,Se.test(u+h)||(a=a.parentNode);a;a=a.parentNode)p.push(a),s=a;s===(r.ownerDocument||ot)&&p.push(s.defaultView||s.parentWindow||n)}for(o=0;(a=p[o++])&&!t.isPropagationStopped();)t.type=o>1?u:f.bindType||h,c=(qt.get(a,"events")||{})[t.type]&&qt.get(a,"handle"),c&&c.apply(a,e),c=l&&a[l],c&&c.apply&&Ft(a)&&(t.result=c.apply(a,e),t.result===!1&&t.preventDefault());return t.type=h,i||t.isDefaultPrevented()||f._default&&f._default.apply(p.pop(),e)!==!1||!Ft(r)||l&&yt.isFunction(r[h])&&!yt.isWindow(r)&&(s=r[l],s&&(r[l]=null),yt.event.triggered=h,r[h](),yt.event.triggered=void 0,s&&(r[l]=s)),t.result}},simulate:function(t,e,n){var r=yt.extend(new yt.Event,n,{type:t,isSimulated:!0});yt.event.trigger(r,null,e)}}),yt.fn.extend({trigger:function(t,e){return this.each(function(){yt.event.trigger(t,e,this)})},triggerHandler:function(t,e){var n=this[0];if(n)return yt.event.trigger(t,e,n,!0)}}),yt.each("blur focus focusin focusout resize scroll click dblclick mousedown mouseup mousemove mouseover mouseout mouseenter mouseleave change select submit keydown keypress keyup contextmenu".split(" "),function(t,e){yt.fn[e]=function(t,n){return arguments.length>0?this.on(e,null,t,n):this.trigger(e)}}),yt.fn.extend({hover:function(t,e){return this.mouseenter(t).mouseleave(e||t)}}),gt.focusin="onfocusin"in n,gt.focusin||yt.each({focus:"focusin",blur:"focusout"},function(t,e){var n=function(t){yt.event.simulate(e,t.target,yt.event.fix(t))};yt.event.special[e]={setup:function(){var r=this.ownerDocument||this,i=qt.access(r,e);i||r.addEventListener(t,n,!0),qt.access(r,e,(i||0)+1)},teardown:function(){var r=this.ownerDocument||this,i=qt.access(r,e)-1;i?qt.access(r,e,i):(r.removeEventListener(t,n,!0),qt.remove(r,e))}}});var Te=n.location,Ee=yt.now(),Me=/\?/;yt.parseXML=function(t){var e;if(!t||"string"!=typeof t)return null;try{e=(new n.DOMParser).parseFromString(t,"text/xml")}catch(t){e=void 0}return e&&!e.getElementsByTagName("parsererror").length||yt.error("Invalid XML: "+t),e};var De=/\[\]$/,Ae=/\r?\n/g,Oe=/^(?:submit|button|image|reset|file)$/i,Ne=/^(?:input|select|textarea|keygen)/i;yt.param=function(t,e){var n,r=[],i=function(t,e){var n=yt.isFunction(e)?e():e;r[r.length]=encodeURIComponent(t)+"="+encodeURIComponent(null==n?"":n)};if(yt.isArray(t)||t.jquery&&!yt.isPlainObject(t))yt.each(t,function(){i(this.name,this.value)});else for(n in t)J(n,t[n],e,i);return r.join("&")},yt.fn.extend({serialize:function(){return yt.param(this.serializeArray())},serializeArray:function(){return this.map(function(){var t=yt.prop(this,"elements");return t?yt.makeArray(t):this}).filter(function(){var t=this.type;return this.name&&!yt(this).is(":disabled")&&Ne.test(this.nodeName)&&!Oe.test(t)&&(this.checked||!Xt.test(t))}).map(function(t,e){var n=yt(this).val();return null==n?null:yt.isArray(n)?yt.map(n,function(t){return{name:e.name,value:t.replace(Ae,"\r\n")}}):{name:e.name,value:n.replace(Ae,"\r\n")}}).get()}});var Pe=/%20/g,Re=/#.*$/,Ie=/([?&])_=[^&]*/,je=/^(.*?):[ \t]*([^\r\n]*)$/gm,Fe=/^(?:about|app|app-storage|.+-extension|file|res|widget):$/,qe=/^(?:GET|HEAD)$/,Ve=/^\/\//,Ue={},Le={},He="*/".concat("*"),Be=ot.createElement("a");Be.href=Te.href,yt.extend({active:0,lastModified:{},etag:{},ajaxSettings:{url:Te.href,type:"GET",isLocal:Fe.test(Te.protocol),global:!0,processData:!0,async:!0,contentType:"application/x-www-form-urlencoded; charset=UTF-8",accepts:{"*":He,text:"text/plain",html:"text/html",xml:"application/xml, text/xml",json:"application/json, text/javascript"},contents:{xml:/\bxml\b/,html:/\bhtml/,json:/\bjson\b/},responseFields:{xml:"responseXML",text:"responseText",json:"responseJSON"},converters:{"* text":String,"text html":!0,"text json":JSON.parse,"text xml":yt.parseXML},flatOptions:{url:!0,context:!0}},ajaxSetup:function(t,e){return e?tt(tt(t,yt.ajaxSettings),e):tt(yt.ajaxSettings,t)},ajaxPrefilter:Z(Ue),ajaxTransport:Z(Le),ajax:function(t,e){function r(t,e,r,s){var l,p,h,$,w,x=e;c||(c=!0,u&&n.clearTimeout(u),i=void 0,a=s||"",_.readyState=t>0?4:0,l=t>=200&&t<300||304===t,r&&($=et(d,_,r)),$=nt(d,$,_,l),l?(d.ifModified&&(w=_.getResponseHeader("Last-Modified"),w&&(yt.lastModified[o]=w),w=_.getResponseHeader("etag"),w&&(yt.etag[o]=w)),204===t||"HEAD"===d.type?x="nocontent":304===t?x="notmodified":(x=$.state,p=$.data,h=$.error,l=!h)):(h=x,!t&&x||(x="error",t<0&&(t=0))),_.status=t,_.statusText=(e||x)+"",l?m.resolveWith(v,[p,x,_]):m.rejectWith(v,[_,x,h]),_.statusCode(b),b=void 0,f&&g.trigger(l?"ajaxSuccess":"ajaxError",[_,d,l?p:h]),y.fireWith(v,[_,x]),f&&(g.trigger("ajaxComplete",[_,d]),--yt.active||yt.event.trigger("ajaxStop")))}"object"==typeof t&&(e=t,t=void 0),e=e||{};var i,o,a,s,u,l,c,f,p,h,d=yt.ajaxSetup({},e),v=d.context||d,g=d.context&&(v.nodeType||v.jquery)?yt(v):yt.event,m=yt.Deferred(),y=yt.Callbacks("once memory"),b=d.statusCode||{},$={},w={},x="canceled",_={readyState:0,getResponseHeader:function(t){var e;if(c){if(!s)for(s={};e=je.exec(a);)s[e[1].toLowerCase()]=e[2];e=s[t.toLowerCase()]}return null==e?null:e},getAllResponseHeaders:function(){return c?a:null},setRequestHeader:function(t,e){return null==c&&(t=w[t.toLowerCase()]=w[t.toLowerCase()]||t,$[t]=e),this},overrideMimeType:function(t){return null==c&&(d.mimeType=t),this},statusCode:function(t){var e;if(t)if(c)_.always(t[_.status]);else for(e in t)b[e]=[b[e],t[e]];return this},abort:function(t){var e=t||x;return i&&i.abort(e),r(0,e),this}};if(m.promise(_),d.url=((t||d.url||Te.href)+"").replace(Ve,Te.protocol+"//"),d.type=e.method||e.type||d.method||d.type,d.dataTypes=(d.dataType||"*").toLowerCase().match(Pt)||[""],null==d.crossDomain){l=ot.createElement("a");try{l.href=d.url,l.href=l.href,d.crossDomain=Be.protocol+"//"+Be.host!=l.protocol+"//"+l.host}catch(t){d.crossDomain=!0}}if(d.data&&d.processData&&"string"!=typeof d.data&&(d.data=yt.param(d.data,d.traditional)),Q(Ue,d,e,_),c)return _;f=yt.event&&d.global,f&&0===yt.active++&&yt.event.trigger("ajaxStart"),d.type=d.type.toUpperCase(),d.hasContent=!qe.test(d.type),o=d.url.replace(Re,""),d.hasContent?d.data&&d.processData&&0===(d.contentType||"").indexOf("application/x-www-form-urlencoded")&&(d.data=d.data.replace(Pe,"+")):(h=d.url.slice(o.length),d.data&&(o+=(Me.test(o)?"&":"?")+d.data,delete d.data),d.cache===!1&&(o=o.replace(Ie,"$1"),h=(Me.test(o)?"&":"?")+"_="+Ee++ +h),d.url=o+h),d.ifModified&&(yt.lastModified[o]&&_.setRequestHeader("If-Modified-Since",yt.lastModified[o]),yt.etag[o]&&_.setRequestHeader("If-None-Match",yt.etag[o])),(d.data&&d.hasContent&&d.contentType!==!1||e.contentType)&&_.setRequestHeader("Content-Type",d.contentType),_.setRequestHeader("Accept",d.dataTypes[0]&&d.accepts[d.dataTypes[0]]?d.accepts[d.dataTypes[0]]+("*"!==d.dataTypes[0]?", "+He+"; q=0.01":""):d.accepts["*"]);for(p in d.headers)_.setRequestHeader(p,d.headers[p]);if(d.beforeSend&&(d.beforeSend.call(v,_,d)===!1||c))return _.abort();if(x="abort",y.add(d.complete),_.done(d.success),_.fail(d.error),i=Q(Le,d,e,_)){if(_.readyState=1,f&&g.trigger("ajaxSend",[_,d]),c)return _;d.async&&d.timeout>0&&(u=n.setTimeout(function(){_.abort("timeout")},d.timeout));try{c=!1,i.send($,r)}catch(t){if(c)throw t;r(-1,t)}}else r(-1,"No Transport");return _},getJSON:function(t,e,n){return yt.get(t,e,n,"json")},getScript:function(t,e){return yt.get(t,void 0,e,"script")}}),yt.each(["get","post"],function(t,e){yt[e]=function(t,n,r,i){return yt.isFunction(n)&&(i=i||r,r=n,n=void 0),yt.ajax(yt.extend({url:t,type:e,dataType:i,data:n,success:r},yt.isPlainObject(t)&&t))}}),yt._evalUrl=function(t){return yt.ajax({url:t,type:"GET",dataType:"script",cache:!0,async:!1,global:!1,throws:!0})},yt.fn.extend({wrapAll:function(t){var e;return this[0]&&(yt.isFunction(t)&&(t=t.call(this[0])),e=yt(t,this[0].ownerDocument).eq(0).clone(!0),this[0].parentNode&&e.insertBefore(this[0]),e.map(function(){for(var t=this;t.firstElementChild;)t=t.firstElementChild;return t}).append(this)),this},wrapInner:function(t){return yt.isFunction(t)?this.each(function(e){yt(this).wrapInner(t.call(this,e))}):this.each(function(){var e=yt(this),n=e.contents();n.length?n.wrapAll(t):e.append(t)})},wrap:function(t){var e=yt.isFunction(t);return this.each(function(n){yt(this).wrapAll(e?t.call(this,n):t)})},unwrap:function(t){return this.parent(t).not("body").each(function(){
yt(this).replaceWith(this.childNodes)}),this}}),yt.expr.pseudos.hidden=function(t){return!yt.expr.pseudos.visible(t)},yt.expr.pseudos.visible=function(t){return!!(t.offsetWidth||t.offsetHeight||t.getClientRects().length)},yt.ajaxSettings.xhr=function(){try{return new n.XMLHttpRequest}catch(t){}};var ze={0:200,1223:204},We=yt.ajaxSettings.xhr();gt.cors=!!We&&"withCredentials"in We,gt.ajax=We=!!We,yt.ajaxTransport(function(t){var e,r;if(gt.cors||We&&!t.crossDomain)return{send:function(i,o){var a,s=t.xhr();if(s.open(t.type,t.url,t.async,t.username,t.password),t.xhrFields)for(a in t.xhrFields)s[a]=t.xhrFields[a];t.mimeType&&s.overrideMimeType&&s.overrideMimeType(t.mimeType),t.crossDomain||i["X-Requested-With"]||(i["X-Requested-With"]="XMLHttpRequest");for(a in i)s.setRequestHeader(a,i[a]);e=function(t){return function(){e&&(e=r=s.onload=s.onerror=s.onabort=s.onreadystatechange=null,"abort"===t?s.abort():"error"===t?"number"!=typeof s.status?o(0,"error"):o(s.status,s.statusText):o(ze[s.status]||s.status,s.statusText,"text"!==(s.responseType||"text")||"string"!=typeof s.responseText?{binary:s.response}:{text:s.responseText},s.getAllResponseHeaders()))}},s.onload=e(),r=s.onerror=e("error"),void 0!==s.onabort?s.onabort=r:s.onreadystatechange=function(){4===s.readyState&&n.setTimeout(function(){e&&r()})},e=e("abort");try{s.send(t.hasContent&&t.data||null)}catch(t){if(e)throw t}},abort:function(){e&&e()}}}),yt.ajaxPrefilter(function(t){t.crossDomain&&(t.contents.script=!1)}),yt.ajaxSetup({accepts:{script:"text/javascript, application/javascript, application/ecmascript, application/x-ecmascript"},contents:{script:/\b(?:java|ecma)script\b/},converters:{"text script":function(t){return yt.globalEval(t),t}}}),yt.ajaxPrefilter("script",function(t){void 0===t.cache&&(t.cache=!1),t.crossDomain&&(t.type="GET")}),yt.ajaxTransport("script",function(t){if(t.crossDomain){var e,n;return{send:function(r,i){e=yt("<script>").prop({charset:t.scriptCharset,src:t.url}).on("load error",n=function(t){e.remove(),n=null,t&&i("error"===t.type?404:200,t.type)}),ot.head.appendChild(e[0])},abort:function(){n&&n()}}}});var Ye=[],Ge=/(=)\?(?=&|$)|\?\?/;yt.ajaxSetup({jsonp:"callback",jsonpCallback:function(){var t=Ye.pop()||yt.expando+"_"+Ee++;return this[t]=!0,t}}),yt.ajaxPrefilter("json jsonp",function(t,e,r){var i,o,a,s=t.jsonp!==!1&&(Ge.test(t.url)?"url":"string"==typeof t.data&&0===(t.contentType||"").indexOf("application/x-www-form-urlencoded")&&Ge.test(t.data)&&"data");if(s||"jsonp"===t.dataTypes[0])return i=t.jsonpCallback=yt.isFunction(t.jsonpCallback)?t.jsonpCallback():t.jsonpCallback,s?t[s]=t[s].replace(Ge,"$1"+i):t.jsonp!==!1&&(t.url+=(Me.test(t.url)?"&":"?")+t.jsonp+"="+i),t.converters["script json"]=function(){return a||yt.error(i+" was not called"),a[0]},t.dataTypes[0]="json",o=n[i],n[i]=function(){a=arguments},r.always(function(){void 0===o?yt(n).removeProp(i):n[i]=o,t[i]&&(t.jsonpCallback=e.jsonpCallback,Ye.push(i)),a&&yt.isFunction(o)&&o(a[0]),a=o=void 0}),"script"}),gt.createHTMLDocument=function(){var t=ot.implementation.createHTMLDocument("").body;return t.innerHTML="<form></form><form></form>",2===t.childNodes.length}(),yt.parseHTML=function(t,e,n){if("string"!=typeof t)return[];"boolean"==typeof e&&(n=e,e=!1);var r,i,o;return e||(gt.createHTMLDocument?(e=ot.implementation.createHTMLDocument(""),r=e.createElement("base"),r.href=ot.location.href,e.head.appendChild(r)):e=ot),i=Tt.exec(t),o=!n&&[],i?[e.createElement(i[1])]:(i=_([t],e,o),o&&o.length&&yt(o).remove(),yt.merge([],i.childNodes))},yt.fn.load=function(t,e,n){var r,i,o,a=this,s=t.indexOf(" ");return s>-1&&(r=X(t.slice(s)),t=t.slice(0,s)),yt.isFunction(e)?(n=e,e=void 0):e&&"object"==typeof e&&(i="POST"),a.length>0&&yt.ajax({url:t,type:i||"GET",dataType:"html",data:e}).done(function(t){o=arguments,a.html(r?yt("<div>").append(yt.parseHTML(t)).find(r):t)}).always(n&&function(t,e){a.each(function(){n.apply(this,o||[t.responseText,e,t])})}),this},yt.each(["ajaxStart","ajaxStop","ajaxComplete","ajaxError","ajaxSuccess","ajaxSend"],function(t,e){yt.fn[e]=function(t){return this.on(e,t)}}),yt.expr.pseudos.animated=function(t){return yt.grep(yt.timers,function(e){return t===e.elem}).length},yt.offset={setOffset:function(t,e,n){var r,i,o,a,s,u,l,c=yt.css(t,"position"),f=yt(t),p={};"static"===c&&(t.style.position="relative"),s=f.offset(),o=yt.css(t,"top"),u=yt.css(t,"left"),l=("absolute"===c||"fixed"===c)&&(o+u).indexOf("auto")>-1,l?(r=f.position(),a=r.top,i=r.left):(a=parseFloat(o)||0,i=parseFloat(u)||0),yt.isFunction(e)&&(e=e.call(t,n,yt.extend({},s))),null!=e.top&&(p.top=e.top-s.top+a),null!=e.left&&(p.left=e.left-s.left+i),"using"in e?e.using.call(t,p):f.css(p)}},yt.fn.extend({offset:function(t){if(arguments.length)return void 0===t?this:this.each(function(e){yt.offset.setOffset(this,t,e)});var e,n,r,i,o=this[0];if(o)return o.getClientRects().length?(r=o.getBoundingClientRect(),r.width||r.height?(i=o.ownerDocument,n=rt(i),e=i.documentElement,{top:r.top+n.pageYOffset-e.clientTop,left:r.left+n.pageXOffset-e.clientLeft}):r):{top:0,left:0}},position:function(){if(this[0]){var t,e,n=this[0],r={top:0,left:0};return"fixed"===yt.css(n,"position")?e=n.getBoundingClientRect():(t=this.offsetParent(),e=this.offset(),yt.nodeName(t[0],"html")||(r=t.offset()),r={top:r.top+yt.css(t[0],"borderTopWidth",!0),left:r.left+yt.css(t[0],"borderLeftWidth",!0)}),{top:e.top-r.top-yt.css(n,"marginTop",!0),left:e.left-r.left-yt.css(n,"marginLeft",!0)}}},offsetParent:function(){return this.map(function(){for(var t=this.offsetParent;t&&"static"===yt.css(t,"position");)t=t.offsetParent;return t||te})}}),yt.each({scrollLeft:"pageXOffset",scrollTop:"pageYOffset"},function(t,e){var n="pageYOffset"===e;yt.fn[t]=function(r){return jt(this,function(t,r,i){var o=rt(t);return void 0===i?o?o[e]:t[r]:void(o?o.scrollTo(n?o.pageXOffset:i,n?i:o.pageYOffset):t[r]=i)},t,r,arguments.length)}}),yt.each(["top","left"],function(t,e){yt.cssHooks[e]=I(gt.pixelPosition,function(t,n){if(n)return n=R(t,e),ce.test(n)?yt(t).position()[e]+"px":n})}),yt.each({Height:"height",Width:"width"},function(t,e){yt.each({padding:"inner"+t,content:e,"":"outer"+t},function(n,r){yt.fn[r]=function(i,o){var a=arguments.length&&(n||"boolean"!=typeof i),s=n||(i===!0||o===!0?"margin":"border");return jt(this,function(e,n,i){var o;return yt.isWindow(e)?0===r.indexOf("outer")?e["inner"+t]:e.document.documentElement["client"+t]:9===e.nodeType?(o=e.documentElement,Math.max(e.body["scroll"+t],o["scroll"+t],e.body["offset"+t],o["offset"+t],o["client"+t])):void 0===i?yt.css(e,n,s):yt.style(e,n,i,s)},e,a?i:void 0,a)}})}),yt.fn.extend({bind:function(t,e,n){return this.on(t,null,e,n)},unbind:function(t,e){return this.off(t,null,e)},delegate:function(t,e,n,r){return this.on(e,t,n,r)},undelegate:function(t,e,n){return 1===arguments.length?this.off(t,"**"):this.off(e,t||"**",n)}}),yt.parseJSON=JSON.parse,r=[],i=function(){return yt}.apply(e,r),!(void 0!==i&&(t.exports=i));var Xe=n.jQuery,Ke=n.$;return yt.noConflict=function(t){return n.$===yt&&(n.$=Ke),t&&n.jQuery===yt&&(n.jQuery=Xe),yt},o||(n.jQuery=n.$=yt),yt})},11:function(t,e,n){n(22),t.exports="ngAnimate"},12:function(t,e,n){n(48),n(38),n(39),n(40),n(41),n(42),n(43),n(47),n(44),n(45),n(46),n(37)},137:function(t,e){!function(){"use strict";function t(){return{restrict:"A",require:"?ngModel",scope:{config:"=?ngIpConfig"},link:function(t,e,n,r){function i(t){t?(u="boolean"==typeof t.requirePort&&t.requirePort,s=u||"boolean"==typeof t.allowPort&&t.allowPort):(s=!1,u=!1),s||!r.$viewValue?r.$parsers[0](r.$viewValue):r.$setViewValue(r.$viewValue.replace(p,""))}function o(t){(t.which<46&&0!==t.which&&8!=t.which&&13!=t.which||47==t.which||t.which>58||58==t.which&&!s)&&t.preventDefault()}function a(t){if(!t)return r.$setValidity("ipAddress",!0),t;d=e[0].selectionStart;var n=!0,i=null;t=t.replace(c,""),t=t.replace(f,".");var o=t.split("."),a=o.length;a<4?n=!1:a>4&&(o.length=4,a=4);for(var v=0;v<a;v++){var g=o[v];if(s)if(v<3)g=g.replace(p,"");else if(g=g.replace(h,""),g.indexOf(":")!=-1){var m=g.split(":");g=m[0],i=m[1].replace(l,"").substring(0,5),(!i||i>65535)&&(n=!1)}else u&&(n=!1);g.length>1?(g=g.replace(l,"").substring(0,3),g>255&&(n=!1)):g||(n=!1),o[v]=g}return t=o.join("."),null!==i&&(t+=":"+i),r.$setValidity("ipAddress",n),r.$setViewValue(t),r.$render(),e[0].setSelectionRange(d,d),t}if(r){var s=!1,u=!1,l=new RegExp("^0","g"),c=new RegExp("^\\.","g"),f=new RegExp("\\.\\.+","g"),p=new RegExp(":","g"),h=new RegExp("^:","g"),d=0;t.$watchCollection("config",i),e.bind("keypress",o),r.$parsers.push(a)}}}}angular.module("ng-ip-address",[]).directive("ngIpAddress",t)}()},140:function(t,e,n){n(0),n(11),n(6),n(3),n(4),n(2),n(7),n(12),n(8),n(1),t.exports=n(5)},2:function(t,e,n){/*!
 * State-based routing for AngularJS
 * @version v1.0.0-beta.1
 * @link https://ui-router.github.io
 * @license MIT License, http://www.opensource.org/licenses/MIT
 */
!function(e,n){t.exports=n()}(this,function(){return function(t){function e(r){if(n[r])return n[r].exports;var i=n[r]={exports:{},id:r,loaded:!1};return t[r].call(i.exports,i,i.exports,e),i.loaded=!0,i.exports}var n={};return e.m=t,e.c=n,e.p="",e(0)}([function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(1)),r(n(53)),r(n(55)),n(58),n(59),n(60),n(61),Object.defineProperty(e,"__esModule",{value:!0}),e.default="ui.router"},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(2)),r(n(46)),r(n(47)),r(n(48)),r(n(49)),r(n(50)),r(n(51)),r(n(52)),r(n(44));var i=n(26);e.UIRouter=i.UIRouter},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(3)),r(n(6)),r(n(7)),r(n(5)),r(n(4)),r(n(8)),r(n(9)),r(n(12))},function(t,e,n){"use strict";function r(t,e,n,r){return void 0===r&&(r=Object.keys(t)),r.filter(function(e){return"function"==typeof t[e]}).forEach(function(r){return e[r]=t[r].bind(n)})}function i(t){void 0===t&&(t={});for(var n=[],r=1;r<arguments.length;r++)n[r-1]=arguments[r];var i=o.apply(null,[{}].concat(n));return e.extend({},i,l(t||{},Object.keys(i)))}function o(t){for(var n=[],r=1;r<arguments.length;r++)n[r-1]=arguments[r];return e.forEach(n,function(n){e.forEach(n,function(e,n){t.hasOwnProperty(n)||(t[n]=e)})}),t}function a(t,e){var n=[];for(var r in t.path){if(t.path[r]!==e.path[r])break;n.push(t.path[r])}return n}function s(t,e,n){void 0===n&&(n=Object.keys(t));for(var r=0;r<n.length;r++){var i=n[r];if(t[i]!=e[i])return!1}return!0}function u(t,e){var n={},r=A(arguments,2);for(var i in e)t(r,i)&&(n[i]=e[i]);return n}function l(t){return u.apply(null,[e.inArray].concat(A(arguments)))}function c(t){return u.apply(null,[T.not(e.inArray)].concat(A(arguments)))}function f(t,e){return d(t,T.prop(e))}function p(t,n){var r=S.isArray(t),i=r?[]:{},o=r?function(t){return i.push(t)}:function(t,e){return i[e]=t};return e.forEach(t,function(t,e){n(t,e)&&o(t,e)}),i}function h(t,n){var r;return e.forEach(t,function(t,e){r||n(t,e)&&(r=t)}),r}function d(t,n){var r=S.isArray(t)?[]:{};return e.forEach(t,function(t,e){return r[e]=n(t,e)}),r}function v(t,e){return t.push(e),t}function g(t,e){return void 0===e&&(e="assert failure"),function(n){if(!t(n))throw new Error(S.isFunction(e)?e(n):e);return!0}}function m(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];if(0===t.length)return[];var n=t.reduce(function(t,e){return Math.min(e.length,t)},9007199254740991);return Array.apply(null,Array(n)).map(function(e,n){return t.map(function(t){return t[n]})})}function y(t,e){var n,r;if(S.isArray(e)&&(n=e[0],r=e[1]),!S.isString(n))throw new Error("invalid parameters to applyPairs");return t[n]=r,t}function b(t){return t.length&&t[t.length-1]||void 0}function $(t,n){return n&&Object.keys(n).forEach(function(t){return delete n[t]}),n||(n={}),e.extend(n,t)}function w(t,e,n){return S.isArray(t)?t.forEach(e,n):void Object.keys(t).forEach(function(n){return e(t[n],n)})}function x(t,e){return Object.keys(e).forEach(function(n){return t[n]=e[n]}),t}function _(t,n){return A(arguments,1).filter(e.identity).reduce(x,t)}function C(t,e){if(t===e)return!0;if(null===t||null===e)return!1;if(t!==t&&e!==e)return!0;var n=typeof t,r=typeof e;if(n!==r||"object"!==n)return!1;var i=[t,e];if(T.all(S.isArray)(i))return k(t,e);if(T.all(S.isDate)(i))return t.getTime()===e.getTime();if(T.all(S.isRegExp)(i))return t.toString()===e.toString();if(T.all(S.isFunction)(i))return!0;var o=[S.isFunction,S.isArray,S.isDate,S.isRegExp];if(o.map(T.any).reduce(function(t,e){return t||!!e(i)},!1))return!1;var a,s={};for(a in t){if(!C(t[a],e[a]))return!1;s[a]=!0}for(a in e)if(!s[a])return!1;return!0}function k(t,e){return t.length===e.length&&m(t,e).reduce(function(t,e){return t&&C(e[0],e[1])},!0)}var S=n(4),T=n(5),E=n(6),M="undefined"==typeof window?{}:window,D=M.angular||{};e.fromJson=D.fromJson||JSON.parse.bind(JSON),e.toJson=D.toJson||JSON.stringify.bind(JSON),e.copy=D.copy||$,e.forEach=D.forEach||w,e.extend=D.extend||_,e.equals=D.equals||C,e.identity=function(t){return t},e.noop=function(){},e.abstractKey="abstract",e.bindFunctions=r,e.inherit=function(t,n){return e.extend(new(e.extend(function(){},{prototype:t})),n)};var A=function(t,e){return void 0===e&&(e=0),Array.prototype.concat.apply(Array.prototype,Array.prototype.slice.call(t,e))};e.inArray=function(t,e){return t.indexOf(e)!==-1},e.removeFrom=T.curry(function(t,e){var n=t.indexOf(e);return n>=0&&t.splice(n,1),t}),e.defaults=i,e.merge=o,e.mergeR=function(t,n){return e.extend(t,n)},e.ancestors=a,e.equalForKeys=s,e.pick=l,e.omit=c,e.pluck=f,e.filter=p,e.find=h,e.mapObj=d,e.map=d,e.values=function(t){return Object.keys(t).map(function(e){return t[e]})},e.allTrueR=function(t,e){return t&&e},e.anyTrueR=function(t,e){return t||e},e.unnestR=function(t,e){return t.concat(e)},e.flattenR=function(t,n){return S.isArray(n)?t.concat(n.reduce(e.flattenR,[])):v(t,n)},e.pushR=v,e.uniqR=function(t,n){return e.inArray(t,n)?t:v(t,n)},e.unnest=function(t){return t.reduce(e.unnestR,[])},e.flatten=function(t){return t.reduce(e.flattenR,[])},e.assertPredicate=g,e.pairs=function(t){return Object.keys(t).map(function(e){return[e,t[e]]})},e.arrayTuples=m,e.applyPairs=y,e.tail=b,e.silenceUncaughtInPromise=function(t){return t.catch(function(t){return 0})&&t},e.silentRejection=function(t){return e.silenceUncaughtInPromise(E.services.$q.reject(t))}},function(t,e,n){"use strict";function r(t){if(e.isArray(t)&&t.length){var n=t.slice(0,-1),r=t.slice(-1);return!(n.filter(i.not(e.isString)).length||r.filter(i.not(e.isFunction)).length)}return e.isFunction(t)}var i=n(5),o=Object.prototype.toString,a=function(t){return function(e){return typeof e===t}};e.isUndefined=a("undefined"),e.isDefined=i.not(e.isUndefined),e.isNull=function(t){return null===t},e.isFunction=a("function"),e.isNumber=a("number"),e.isString=a("string"),e.isObject=function(t){return null!==t&&"object"==typeof t},e.isArray=Array.isArray,e.isDate=function(t){return"[object Date]"===o.call(t)},e.isRegExp=function(t){return"[object RegExp]"===o.call(t)},e.isInjectable=r,e.isPromise=i.and(e.isObject,i.pipe(i.prop("then"),e.isFunction))},function(t,e){"use strict";function n(t){function e(n){return n.length>=r?t.apply(null,n):function(){return e(n.concat([].slice.apply(arguments)))}}var n=[].slice.apply(arguments,[1]),r=t.length;return e(n)}function r(){var t=arguments,e=t.length-1;return function(){for(var n=e,r=t[e].apply(this,arguments);n--;)r=t[n].call(this,r);return r}}function i(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];return r.apply(null,[].slice.call(arguments).reverse())}function o(t,e){return function(){for(var n=[],r=0;r<arguments.length;r++)n[r-0]=arguments[r];return t.apply(null,n)&&e.apply(null,n)}}function a(t,e){return function(){for(var n=[],r=0;r<arguments.length;r++)n[r-0]=arguments[r];return t.apply(null,n)||e.apply(null,n)}}function s(t,e){return function(n){return n[t].apply(n,e)}}function u(t){return function(e){for(var n=0;n<t.length;n++)if(t[n][0](e))return t[n][1](e)}}e.curry=n,e.compose=r,e.pipe=i,e.prop=function(t){return function(e){return e&&e[t]}},e.propEq=n(function(t,e,n){return n&&n[t]===e}),e.parse=function(t){return i.apply(null,t.split(".").map(e.prop))},e.not=function(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n-0]=arguments[n];return!t.apply(null,e)}},e.and=o,e.or=a,e.all=function(t){return function(e){return e.reduce(function(e,n){return e&&!!t(n)},!0)}},e.any=function(t){return function(e){return e.reduce(function(e,n){return e||!!t(n)},!1)}},e.none=e.not(e.any),e.is=function(t){return function(e){return null!=e&&e.constructor===t||e instanceof t}},e.eq=function(t){return function(e){return t===e}},e.val=function(t){return function(){return t}},e.invoke=s,e.pattern=u},function(t,e){"use strict";var n=function(t){return function(){throw new Error(t+"(): No coreservices implementation for UI-Router is loaded. You should include one of: ['angular1.js']")}},r={$q:void 0,$injector:void 0,location:{},locationConfig:{},template:{}};e.services=r,["replace","url","path","search","hash","onChange"].forEach(function(t){return r.location[t]=n(t)}),["port","protocol","host","baseHref","html5Mode","hashPrefix"].forEach(function(t){return r.locationConfig[t]=n(t)})},function(t,e){"use strict";var n=function(){function t(t){this.text=t,this.glob=t.split(".")}return t.prototype.matches=function(t){for(var e=t.split("."),n=0,r=this.glob.length;n<r;n++)"*"===this.glob[n]&&(e[n]="*");return"**"===this.glob[0]&&(e=e.slice(e.indexOf(this.glob[1])),e.unshift("**")),"**"===this.glob[this.glob.length-1]&&(e.splice(e.indexOf(this.glob[this.glob.length-2])+1,Number.MAX_VALUE),e.push("**")),this.glob.length==e.length&&e.join("")===this.glob.join("")},t.is=function(t){return t.indexOf("*")>-1},t.fromString=function(e){return this.is(e)?new t(e):null},t}();e.Glob=n},function(t,e){"use strict";var n=function(){function t(t,e){void 0===t&&(t=[]),void 0===e&&(e=null),this._items=t,this._limit=e}return t.prototype.enqueue=function(t){var e=this._items;return e.push(t),this._limit&&e.length>this._limit&&e.shift(),t},t.prototype.dequeue=function(){if(this.size())return this._items.splice(0,1)[0]},t.prototype.clear=function(){var t=this._items;return this._items=[],t},t.prototype.size=function(){return this._items.length},t.prototype.remove=function(t){var e=this._items.indexOf(t);return e>-1&&this._items.splice(e,1)[0]},t.prototype.peekTail=function(){return this._items[this._items.length-1]},t.prototype.peekHead=function(){if(this.size())return this._items[0]},t}();e.Queue=n},function(t,e,n){"use strict";function r(t,e){return e.length<=t?e:e.substr(0,t-3)+"..."}function i(t,e){for(;e.length<t;)e+=" ";return e}function o(t){return t.replace(/^([A-Z])/,function(t){return t.toLowerCase()}).replace(/([A-Z])/g,function(t){return"-"+t.toLowerCase()})}function a(t){var e=s(t),n=e.match(/^(function [^ ]+\([^)]*\))/);return n?n[1]:e}function s(t){var e=l.isArray(t)?t.slice(-1)[0]:t;return e&&e.toString()||"undefined"}function u(t){function e(t){if(l.isObject(t)){if(n.indexOf(t)!==-1)return"[circular ref]";n.push(t)}return g(t)}var n=[];return JSON.stringify(t,function(t,n){return e(n)}).replace(/\\"/g,'"')}var l=n(4),c=n(10),f=n(3),p=n(5),h=n(11),d=n(19);e.maxLength=r,e.padString=i,e.kebobString=o,e.functionToString=a,e.fnToString=s;var v=null,g=function(t){var e=c.Rejection.isTransitionRejectionPromise;return(v=v||p.pattern([[p.not(l.isDefined),p.val("undefined")],[l.isNull,p.val("null")],[l.isPromise,p.val("[Promise]")],[e,function(t){return t._transitionRejection.toString()}],[p.is(c.Rejection),p.invoke("toString")],[p.is(h.Transition),p.invoke("toString")],[p.is(d.Resolvable),p.invoke("toString")],[l.isInjectable,a],[p.val(!0),f.identity]]))(t)};e.stringify=u,e.beforeAfterSubstr=function(t){return function(e){if(!e)return["",""];var n=e.indexOf(t);return n===-1?[e,""]:[e.substr(0,n),e.substr(n+1)]}}},function(t,e,n){"use strict";var r=n(3),i=n(9);!function(t){t[t.SUPERSEDED=2]="SUPERSEDED",t[t.ABORTED=3]="ABORTED",t[t.INVALID=4]="INVALID",t[t.IGNORED=5]="IGNORED",t[t.ERROR=6]="ERROR"}(e.RejectType||(e.RejectType={}));var o=e.RejectType,a=function(){function t(t,e,n){this.type=t,this.message=e,this.detail=n}return t.prototype.toString=function(){var t=function(t){return t&&t.toString!==Object.prototype.toString?t.toString():i.stringify(t)},e=this.type,n=this.message,r=t(this.detail);return"TransitionRejection(type: "+e+", message: "+n+", detail: "+r+")"},t.prototype.toPromise=function(){return r.extend(r.silentRejection(this),{_transitionRejection:this})},t.isTransitionRejectionPromise=function(e){return e&&"function"==typeof e.then&&e._transitionRejection instanceof t},t.superseded=function(e,n){var r="The transition has been superseded by a different transition (see detail).",i=new t(o.SUPERSEDED,r,e);return n&&n.redirected&&(i.redirected=!0),i},t.redirected=function(e){return t.superseded(e,{redirected:!0})},t.invalid=function(e){var n="This transition is invalid (see detail)";return new t(o.INVALID,n,e)},t.ignored=function(e){var n="The transition was ignored.";return new t(o.IGNORED,n,e)},t.aborted=function(e){var n="The transition has been aborted.";return new t(o.ABORTED,n,e)},t.errored=function(e){var n="The transition errored.";return new t(o.ERROR,n,e)},t}();e.Rejection=a},function(t,e,n){"use strict";var r=n(12),i=n(6),o=n(3),a=n(4),s=n(5),u=n(13),l=n(15),c=n(16),f=n(21),p=n(20),h=n(14),d=n(22),v=n(19),g=n(10),m=n(17),y=n(26),b=0,$=s.prop("self"),w=function(){function t(e,n,r){var a=this;if(this._deferred=i.services.$q.defer(),this.promise=this._deferred.promise,this.treeChanges=function(){return a._treeChanges},this.isActive=function(){return a===a._options.current()},this.router=r,!n.valid())throw new Error(n.error());l.HookRegistry.mixin(new l.HookRegistry,this),this._options=o.extend({current:s.val(this)},n.options()),this.$id=b++;var u=p.PathFactory.buildToPath(e,n);this._treeChanges=p.PathFactory.treeChanges(e,u,this._options.reloadState);var c=this._treeChanges.entering.map(function(t){return t.state});p.PathFactory.applyViewConfigs(r.transitionService.$view,this._treeChanges.to,c);var f=[new v.Resolvable(y.UIRouter,function(){return r},[],void 0,r),new v.Resolvable(t,function(){return a},[],void 0,this),new v.Resolvable("$transition$",function(){return a},[],void 0,this),new v.Resolvable("$stateParams",function(){return a.params()},[],void 0,this.params())],h=this._treeChanges.to[0],d=new m.ResolveContext(this._treeChanges.to);d.addResolvables(f,h.state)}return t.prototype.onBefore=function(t,e,n){throw""},t.prototype.onStart=function(t,e,n){throw""},t.prototype.onExit=function(t,e,n){throw""},t.prototype.onRetain=function(t,e,n){throw""},t.prototype.onEnter=function(t,e,n){throw""},t.prototype.onFinish=function(t,e,n){throw""},t.prototype.onSuccess=function(t,e,n){throw""},t.prototype.onError=function(t,e,n){throw""},t.prototype.$from=function(){return o.tail(this._treeChanges.from).state},t.prototype.$to=function(){return o.tail(this._treeChanges.to).state},t.prototype.from=function(){return this.$from().self},t.prototype.to=function(){return this.$to().self},t.prototype.is=function(e){return e instanceof t?this.is({to:e.$to().name,from:e.$from().name}):!(e.to&&!l.matchState(this.$to(),e.to)||e.from&&!l.matchState(this.$from(),e.from))},t.prototype.params=function(t){return void 0===t&&(t="to"),this._treeChanges[t].map(s.prop("paramValues")).reduce(o.mergeR,{})},t.prototype.injector=function(t){var e=this.treeChanges().to;return t&&(e=p.PathFactory.subPath(e,function(e){return e.state===t||e.state.name===t})),new m.ResolveContext(e).injector()},t.prototype.getResolveTokens=function(){return new m.ResolveContext(this._treeChanges.to).getTokens()},t.prototype.getResolveValue=function(t){var e=new m.ResolveContext(this._treeChanges.to),n=function(t){var n=e.getResolvable(t);if(void 0===n)throw new Error("Dependency Injection token not found: ${stringify(token)}");return n.data};return a.isArray(t)?t.map(n):n(t)},t.prototype.addResolvable=function(t,e){void 0===e&&(e="");var n="string"==typeof e?e:e.name,r=this._treeChanges.to,i=o.find(r,function(t){return t.state.name===n}),a=new m.ResolveContext(r);a.addResolvables([t],i.state)},t.prototype.previous=function(){return this._options.previous||null},t.prototype.options=function(){return this._options},t.prototype.entering=function(){return o.map(this._treeChanges.entering,s.prop("state")).map($)},t.prototype.exiting=function(){return o.map(this._treeChanges.exiting,s.prop("state")).map($).reverse()},t.prototype.retained=function(){return o.map(this._treeChanges.retained,s.prop("state")).map($)},t.prototype.views=function(t,e){void 0===t&&(t="entering");var n=this._treeChanges[t];return n=e?n.filter(s.propEq("state",e)):n,n.map(s.prop("views")).filter(o.identity).reduce(o.unnestR,[])},t.prototype.redirect=function(t){var e=o.extend({},this.options(),t.options(),{previous:this});t=new h.TargetState(t.identifier(),t.$state(),t.params(),e);var n=this.router.transitionService.create(this._treeChanges.from,t),r=this.treeChanges().entering,i=n.treeChanges().entering,a=function(t){return function(e){return t&&e.state.includes[t.name]}},u=f.PathNode.matching(i,r).filter(s.not(a(t.options().reloadState)));return u.forEach(function(t,e){t.resolvables=r[e].resolvables}),n},t.prototype._changedParams=function(){var t=this._treeChanges,e=t.to,n=t.from;if(!this._options.reload&&o.tail(e).state===o.tail(n).state){var r=e.map(function(t){return t.paramSchema}),i=[e,n].map(function(t){return t.map(function(t){return t.paramValues})}),a=i[0],s=i[1],u=o.arrayTuples(r,a,s);return u.map(function(t){var e=t[0],n=t[1],r=t[2];return d.Param.changed(e,n,r)}).reduce(o.unnestR,[])}},t.prototype.dynamic=function(){var t=this._changedParams();return!!t&&t.map(function(t){return t.dynamic}).reduce(o.anyTrueR,!1)},t.prototype.ignored=function(){var t=this._changedParams();return!!t&&0===t.length},t.prototype.hookBuilder=function(){return new c.HookBuilder(this.router.transitionService,this,{transition:this,current:this._options.current})},t.prototype.run=function(){var t=this,e=u.TransitionHook.runSynchronousHooks,n=this.hookBuilder(),i=this.router.globals;i.transitionHistory.enqueue(this);var o=e(n.getOnBeforeHooks());if(g.Rejection.isTransitionRejectionPromise(o)){o.catch(function(){return 0});var a=o._transitionRejection;return this._deferred.reject(a),this.promise}if(!this.valid()){var s=new Error(this.error());return this._deferred.reject(s),this.promise}if(this.ignored())return r.trace.traceTransitionIgnored(this),this._deferred.reject(g.Rejection.ignored()),this.promise;var l=function(){r.trace.traceSuccess(t.$to(),t),t.success=!0,t._deferred.resolve(t.to()),e(n.getOnSuccessHooks(),!0)},c=function(i){r.trace.traceError(i,t),t.success=!1,t._deferred.reject(i),e(n.getOnErrorHooks(),!0)};r.trace.traceTransitionStart(this);var f=function(t,e){return t.then(function(){return e.invokeHook()})};return n.asyncHooks().reduce(f,o).then(l,c),this.promise},t.prototype.valid=function(){return!this.error()},t.prototype.error=function(){var t=this.$to();return t.self[o.abstractKey]?"Cannot transition to abstract state '"+t.name+"'":d.Param.validates(t.parameters(),this.params())?void 0:"Param values not valid for state '"+t.name+"'"},t.prototype.toString=function(){var t=this.from(),e=this.to(),n=function(t){return null!==t["#"]&&void 0!==t["#"]?t:o.omit(t,"#")},r=this.$id,i=a.isObject(t)?t.name:t,u=o.toJson(n(this._treeChanges.from.map(s.prop("paramValues")).reduce(o.mergeR,{}))),l=this.valid()?"":"(X) ",c=a.isObject(e)?e.name:e,f=o.toJson(n(this.params()));return"Transition#"+r+"( '"+i+"'"+u+" -> "+l+"'"+c+"'"+f+" )"},t.diToken=t,t}();e.Transition=w},function(t,e,n){"use strict";function r(t){return t?"[ui-view#"+t.id+" tag "+("in template from '"+(t.creationContext&&t.creationContext.name||"(root)")+"' state]: ")+("fqn: '"+t.fqn+"', ")+("name: '"+t.name+"@"+t.creationContext+"')"):"ui-view (defunct)"}function i(t){return a.isNumber(t)?l[t]:l[l[t]]}var o=n(5),a=n(4),s=n(9),u=function(t){return"[ViewConfig#"+t.$id+" from '"+(t.viewDecl.$context.name||"(root)")+"' state]: target ui-view: '"+t.viewDecl.$uiViewName+"@"+t.viewDecl.$uiViewContextAnchor+"'"};!function(t){t[t.RESOLVE=0]="RESOLVE",t[t.TRANSITION=1]="TRANSITION",t[t.HOOK=2]="HOOK",t[t.INVOKE=3]="INVOKE",t[t.UIVIEW=4]="UIVIEW",t[t.VIEWCONFIG=5]="VIEWCONFIG"}(e.Category||(e.Category={}));var l=e.Category,c=function(){function t(){this._enabled={},this.approximateDigests=0}return t.prototype._set=function(t,e){var n=this;e.length||(e=Object.keys(l).filter(function(t){return isNaN(parseInt(t,10))}).map(function(t){return l[t]})),e.map(i).forEach(function(e){return n._enabled[e]=t})},t.prototype.enable=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];this._set(!0,t)},t.prototype.disable=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];this._set(!1,t)},t.prototype.enabled=function(t){return!!this._enabled[i(t)]},t.prototype.traceTransitionStart=function(t){if(this.enabled(l.TRANSITION)){var e=t.$id,n=this.approximateDigests,r=s.stringify(t);console.log("Transition #"+e+" Digest #"+n+": Started  -> "+r)}},t.prototype.traceTransitionIgnored=function(t){if(this.enabled(l.TRANSITION)){var e=t&&t.$id,n=this.approximateDigests,r=s.stringify(t);console.log("Transition #"+e+" Digest #"+n+": Ignored  <> "+r)}},t.prototype.traceHookInvocation=function(t,e){if(this.enabled(l.HOOK)){var n=o.parse("transition.$id")(e),r=this.approximateDigests,i=o.parse("traceData.hookType")(e)||"internal",a=o.parse("traceData.context.state.name")(e)||o.parse("traceData.context")(e)||"unknown",u=s.functionToString(t.fn);console.log("Transition #"+n+" Digest #"+r+":   Hook -> "+i+" context: "+a+", "+s.maxLength(200,u))}},t.prototype.traceHookResult=function(t,e,n){if(this.enabled(l.HOOK)){var r=o.parse("transition.$id")(n),i=this.approximateDigests,a=s.stringify(t),u=s.stringify(e);console.log("Transition #"+r+" Digest #"+i+":   <- Hook returned: "+s.maxLength(200,a)+", transition result: "+s.maxLength(200,u))}},t.prototype.traceResolvePath=function(t,e,n){if(this.enabled(l.RESOLVE)){var r=n&&n.$id,i=this.approximateDigests,o=t&&t.toString();console.log("Transition #"+r+" Digest #"+i+":         Resolving "+o+" ("+e+")")}},t.prototype.traceResolvableResolved=function(t,e){if(this.enabled(l.RESOLVE)){var n=e&&e.$id,r=this.approximateDigests,i=t&&t.toString(),o=s.stringify(t.data);console.log("Transition #"+n+" Digest #"+r+":               <- Resolved  "+i+" to: "+s.maxLength(200,o))}},t.prototype.traceError=function(t,e){if(this.enabled(l.TRANSITION)){var n=e&&e.$id,r=this.approximateDigests,i=s.stringify(e);console.log("Transition #"+n+" Digest #"+r+": <- Rejected "+i+", reason: "+t)}},t.prototype.traceSuccess=function(t,e){if(this.enabled(l.TRANSITION)){var n=e&&e.$id,r=this.approximateDigests,i=t.name,o=s.stringify(e);console.log("Transition #"+n+" Digest #"+r+": <- Success  "+o+", final state: "+i)}},t.prototype.traceUIViewEvent=function(t,e,n){void 0===n&&(n=""),this.enabled(l.UIVIEW)&&console.log("ui-view: "+s.padString(30,t)+" "+r(e)+n)},t.prototype.traceUIViewConfigUpdated=function(t,e){this.enabled(l.UIVIEW)&&this.traceUIViewEvent("Updating",t," with ViewConfig from context='"+e+"'")},t.prototype.traceUIViewScopeCreated=function(t,e){this.enabled(l.UIVIEW)&&this.traceUIViewEvent("Created scope for",t,", scope #"+e.$id)},t.prototype.traceUIViewFill=function(t,e){this.enabled(l.UIVIEW)&&this.traceUIViewEvent("Fill",t," with: "+s.maxLength(200,e))},t.prototype.traceViewServiceEvent=function(t,e){this.enabled(l.VIEWCONFIG)&&console.log("VIEWCONFIG: "+t+" "+u(e))},t.prototype.traceViewServiceUIViewEvent=function(t,e){this.enabled(l.VIEWCONFIG)&&console.log("VIEWCONFIG: "+t+" "+r(e))},t}();e.Trace=c;var f=new c;e.trace=f},function(t,e,n){"use strict";var r=n(3),i=n(9),o=n(4),a=n(5),s=n(12),u=n(6),l=n(10),c=n(14),f={async:!0,rejectIfSuperseded:!0,current:r.noop,transition:null,traceData:{},bind:null},p=function(){function t(t,e,n,i,o){var a=this;this.transition=t,this.stateContext=e,this.hookFn=n,this.resolveContext=i,this.options=o,this.isSuperseded=function(){return a.options.current()!==a.options.transition},this.options=r.defaults(o,f)}return t.prototype.invokeHook=function(){var t=this,e=t.options,n=t.hookFn;t.resolveContext;if(s.trace.traceHookInvocation(this,e),e.rejectIfSuperseded&&this.isSuperseded())return l.Rejection.superseded(e.current()).toPromise();var r=n.call(e.bind,this.transition,this.stateContext);return this.handleHookResult(r)},t.prototype.handleHookResult=function(t){var e=this;if(o.isDefined(t)){var n=a.pattern([[this.isSuperseded,function(){return l.Rejection.superseded(e.options.current()).toPromise()}],[a.eq(!1),function(){return l.Rejection.aborted("Hook aborted transition").toPromise()}],[a.is(c.TargetState),function(t){return l.Rejection.redirected(t).toPromise()}],[o.isPromise,function(t){return t.then(e.handleHookResult.bind(e))}]]),r=n(t);return r&&s.trace.traceHookResult(t,r,this.options),r}},t.prototype.toString=function(){var t=this,e=t.options,n=t.hookFn,r=a.parse("traceData.hookType")(e)||"internal",o=a.parse("traceData.context.state.name")(e)||a.parse("traceData.context")(e)||"unknown",s=i.fnToString(n);return r+" context: "+o+", "+i.maxLength(200,s)},t.runSynchronousHooks=function(t,e){void 0===e&&(e=!1);for(var n=[],r=0;r<t.length;r++)try{n.push(t[r].invokeHook())}catch(t){if(!e)return l.Rejection.errored(t).toPromise();console.error("Swallowed exception during synchronous hook handler: "+t)}var i=n.filter(l.Rejection.isTransitionRejectionPromise);return i.length?i[0]:n.filter(o.isPromise).reduce(function(t,e){return t.then(a.val(e))},u.services.$q.when())},t}();e.TransitionHook=p},function(t,e){"use strict";var n=function(){function t(t,e,n,r){void 0===n&&(n={}),void 0===r&&(r={}),this._identifier=t,this._definition=e,this._options=r,this._params=n||{}}return t.prototype.name=function(){return this._definition&&this._definition.name||this._identifier},t.prototype.identifier=function(){return this._identifier},t.prototype.params=function(){return this._params},t.prototype.$state=function(){return this._definition},t.prototype.state=function(){return this._definition&&this._definition.self},t.prototype.options=function(){return this._options},t.prototype.exists=function(){return!(!this._definition||!this._definition.self)},t.prototype.valid=function(){return!this.error()},t.prototype.error=function(){var t=this.options().relative;if(!this._definition&&t){var e=t.name?t.name:t;return"Could not resolve '"+this.name()+"' from state '"+e+"'"}return this._definition?this._definition.self?void 0:"State '"+this.name()+"' has an invalid definition":"No such state '"+this.name()+"'"},t}();e.TargetState=n},function(t,e,n){"use strict";function r(t,e){function n(t){for(var e=r,n=0;n<e.length;n++){var i=s.Glob.fromString(e[n]);if(i&&i.matches(t.name)||!i&&e[n]===t.name)return!0}return!1}var r=a.isString(e)?[e]:e,i=a.isFunction(r)?r:n;return!!i(t)}function i(t,e){return function(n,r,i){void 0===i&&(i={});var a=new u(n,r,i);return t[e].push(a),function(){o.removeFrom(t[e])(a)}}}var o=n(3),a=n(4),s=n(7);e.matchState=r;var u=function(){function t(t,e,n){void 0===n&&(n={}),this.callback=e,this.matchCriteria=o.extend({to:!0,from:!0,exiting:!0,retained:!0,entering:!0},t),this.priority=n.priority||0,this.bind=n.bind||null}return t._matchingNodes=function(t,e){if(e===!0)return t;var n=t.filter(function(t){return r(t.state,e)});return n.length?n:null},t.prototype.matches=function(e){var n=this.matchCriteria,r=t._matchingNodes,i={to:r([o.tail(e.to)],n.to),from:r([o.tail(e.from)],n.from),exiting:r(e.exiting,n.exiting),retained:r(e.retained,n.retained),entering:r(e.entering,n.entering)},a=["to","from","exiting","retained","entering"].map(function(t){return i[t]}).reduce(o.allTrueR,!0);return a?i:null},t}();e.EventHook=u;var l=function(){function t(){var t=this;this._transitionEvents={onBefore:[],onStart:[],onEnter:[],onRetain:[],onExit:[],onFinish:[],onSuccess:[],onError:[]},this.getHooks=function(e){return t._transitionEvents[e]},this.onBefore=i(this._transitionEvents,"onBefore"),this.onStart=i(this._transitionEvents,"onStart"),this.onEnter=i(this._transitionEvents,"onEnter"),this.onRetain=i(this._transitionEvents,"onRetain"),this.onExit=i(this._transitionEvents,"onExit"),this.onFinish=i(this._transitionEvents,"onFinish"),this.onSuccess=i(this._transitionEvents,"onSuccess"),this.onError=i(this._transitionEvents,"onError")}return t.mixin=function(t,e){Object.keys(t._transitionEvents).concat(["getHooks"]).forEach(function(n){return e[n]=t[n]})},t}();e.HookRegistry=l},function(t,e,n){"use strict";function r(t){return void 0===t&&(t=!1),function(e,n){var r=t?-1:1,i=(e.node.state.path.length-n.node.state.path.length)*r;return 0!==i?i:n.hook.priority-e.hook.priority}}var i=n(3),o=n(4),a=n(13),s=n(17),u=function(){function t(t,e,n){var o=this;this.$transitions=t,this.transition=e,this.baseHookOptions=n,this.getOnBeforeHooks=function(){return o._buildNodeHooks("onBefore","to",r(),{async:!1})},this.getOnStartHooks=function(){return o._buildNodeHooks("onStart","to",r())},this.getOnExitHooks=function(){return o._buildNodeHooks("onExit","exiting",r(!0),{stateHook:!0})},this.getOnRetainHooks=function(){return o._buildNodeHooks("onRetain","retained",r(!1),{stateHook:!0})},this.getOnEnterHooks=function(){return o._buildNodeHooks("onEnter","entering",r(!1),{stateHook:!0})},this.getOnFinishHooks=function(){return o._buildNodeHooks("onFinish","to",r())},this.getOnSuccessHooks=function(){return o._buildNodeHooks("onSuccess","to",r(),{async:!1,rejectIfSuperseded:!1})},this.getOnErrorHooks=function(){return o._buildNodeHooks("onError","to",r(),{async:!1,rejectIfSuperseded:!1})},this.treeChanges=e.treeChanges(),this.toState=i.tail(this.treeChanges.to).state,this.fromState=i.tail(this.treeChanges.from).state,this.transitionOptions=e.options()}return t.prototype.asyncHooks=function(){var t=this.getOnStartHooks(),e=this.getOnExitHooks(),n=this.getOnRetainHooks(),r=this.getOnEnterHooks(),o=this.getOnFinishHooks(),a=[t,e,n,r,o];return a.reduce(i.unnestR,[]).filter(i.identity)},t.prototype._buildNodeHooks=function(t,e,n,r){var o=this,u=this._matchingHooks(t,this.treeChanges);if(!u)return[];var l=function(n){var u=n.matches(o.treeChanges),l=u[e],c="exiting"===e?o.treeChanges.from:o.treeChanges.to,f=new s.ResolveContext(c);return l.map(function(e){var s=i.extend({bind:n.bind,traceData:{hookType:t,context:e}},o.baseHookOptions,r),u=s.stateHook?e.state:null,l=f.subContext(e.state),c=new a.TransitionHook(o.transition,u,n.callback,l,s);return{hook:n,node:e,transitionHook:c}})};return u.map(l).reduce(i.unnestR,[]).sort(n).map(function(t){return t.transitionHook})},t.prototype._matchingHooks=function(t,e){return[this.transition,this.$transitions].map(function(e){return e.getHooks(t)}).filter(i.assertPredicate(o.isArray,"broken event named: "+t)).reduce(i.unnestR,[]).filter(function(t){return t.matches(e)})},t}();e.HookBuilder=u},function(t,e,n){"use strict";var r=n(3),i=n(5),o=n(12),a=n(6),s=n(18),u=n(19),l=n(20),c=n(9),f=s.resolvePolicies.when,p=[f.EAGER,f.LAZY],h=[f.EAGER],d=function(){function t(t){this._path=t}return t.prototype.getTokens=function(){return this._path.reduce(function(t,e){return t.concat(e.resolvables.map(function(t){return t.token}))},[]).reduce(r.uniqR,[])},t.prototype.getResolvable=function(t){var e=this._path.map(function(t){return t.resolvables}).reduce(r.unnestR,[]).filter(function(e){return e.token===t});return r.tail(e)},t.prototype.subContext=function(e){return new t(l.PathFactory.subPath(this._path,function(t){return t.state===e}))},t.prototype.addResolvables=function(t,e){var n=r.find(this._path,i.propEq("state",e)),o=t.map(function(t){return t.token});n.resolvables=n.resolvables.filter(function(t){return o.indexOf(t.token)===-1}).concat(t)},t.prototype.resolvePath=function(t,e){var n=this;void 0===t&&(t="LAZY");var i=r.inArray(p,t)?t:"LAZY",u=i===s.resolvePolicies.when.EAGER?h:p;o.trace.traceResolvePath(this._path,t,e);var l=this._path.reduce(function(t,i){var o=function(t){return r.inArray(u,t.getPolicy(i.state).when)},a=i.resolvables.filter(o),s=n.subContext(i.state),l=function(t){return t.get(s,e).then(function(e){return{token:t.token,value:e}})};return t.concat(a.map(l))},[]);return a.services.$q.all(l)},t.prototype.injector=function(){return new v(this)},t.prototype.findNode=function(t){return r.find(this._path,function(e){return r.inArray(e.resolvables,t)})},t.prototype.getDependencies=function(t){var e=this.findNode(t),n=l.PathFactory.subPath(this._path,function(t){return t===e})||this._path,i=n.reduce(function(t,e){return t.concat(e.resolvables)},[]).filter(function(e){return e!==t}),o=function(t){var e=i.filter(function(e){return e.token===t});if(e.length)return r.tail(e);var n=a.services.$injector.get(t);
if(!n)throw new Error("Could not find Dependency Injection token: "+c.stringify(t));return new u.Resolvable(t,function(){return n},[],n)};return t.deps.map(o)},t}();e.ResolveContext=d;var v=function(){function t(t){this.context=t,this.native=a.services.$injector}return t.prototype.get=function(t){var e=this.context.getResolvable(t);if(e){if(!e.resolved)throw new Error("Resolvable async .get() not complete:"+c.stringify(e.token));return e.data}return a.services.$injector.get(t)},t.prototype.getAsync=function(t){var e=this.context.getResolvable(t);return e?e.get(this.context):a.services.$q.when(a.services.$injector.get(t))},t}()},function(t,e){"use strict";e.resolvePolicies={when:{LAZY:"LAZY",EAGER:"EAGER"},async:{WAIT:"WAIT",NOWAIT:"NOWAIT",RXWAIT:"RXWAIT"}}},function(t,e,n){"use strict";var r=n(3),i=n(6),o=n(12),a=n(9),s=n(4);e.defaultResolvePolicy={when:"LAZY",async:"WAIT"};var u=function(){function t(e,n,o,a,u){if(this.resolved=!1,this.promise=void 0,e instanceof t)r.extend(this,e);else if(s.isFunction(n)){if(null==e||void 0==e)throw new Error("new Resolvable(): token argument is required");if(!s.isFunction(n))throw new Error("new Resolvable(): resolveFn argument must be a function");this.token=e,this.policy=a,this.resolveFn=n,this.deps=o||[],this.data=u,this.resolved=void 0!==u,this.promise=this.resolved?i.services.$q.when(this.data):void 0}else if(s.isObject(e)&&e.token&&s.isFunction(e.resolveFn)){var l=e;return new t(l.token,l.resolveFn,l.deps,l.policy,l.data)}}return t.prototype.getPolicy=function(t){var n=this.policy||{},r=t&&t.resolvePolicy||{};return{when:n.when||r.when||e.defaultResolvePolicy.when,async:n.async||r.async||e.defaultResolvePolicy.async}},t.prototype.resolve=function(t,e){var n=this,r=i.services.$q,a=function(){return r.all(t.getDependencies(n).map(function(n){return n.get(t,e)}))},s=function(t){return n.resolveFn.apply(null,t)},u=function(t){var e=t.cache();return e.toPromise().then(function(){return e})},l=t.findNode(this),c=l&&l.state,f="RXWAIT"===this.getPolicy(c).async?u:function(t){return t},p=function(t){return n.data=t,n.resolved=!0,o.trace.traceResolvableResolved(n,e),n.data};return this.promise=r.when().then(a).then(s).then(f).then(p)},t.prototype.get=function(t,e){return this.promise||this.resolve(t,e)},t.prototype.toString=function(){return"Resolvable(token: "+a.stringify(this.token)+", requires: ["+this.deps.map(a.stringify)+"])"},t.prototype.clone=function(){return new t(this)},t}();e.Resolvable=u},function(t,e,n){"use strict";var r=n(3),i=n(5),o=n(14),a=n(21),s=function(){function t(){}return t.makeTargetState=function(t){var e=r.tail(t).state;return new o.TargetState(e,e,t.map(i.prop("paramValues")).reduce(r.mergeR,{}))},t.buildPath=function(t){var e=t.params();return t.$state().path.map(function(t){return new a.PathNode(t).applyRawParams(e)})},t.buildToPath=function(e,n){var r=t.buildPath(n);return n.options().inherit?t.inheritParams(e,r,Object.keys(n.params())):r},t.applyViewConfigs=function(e,n,i){n.filter(function(t){return r.inArray(i,t.state)}).forEach(function(i){var o=r.values(i.state.views||{}),a=t.subPath(n,function(t){return t===i}),s=o.map(function(t){return e.createViewConfig(a,t)});i.views=s.reduce(r.unnestR,[])})},t.inheritParams=function(t,e,n){function o(t,e){var n=r.find(t,i.propEq("state",e));return r.extend({},n&&n.paramValues)}function s(e){var i=r.extend({},e&&e.paramValues),s=r.pick(i,n);i=r.omit(i,n);var u=o(t,e.state)||{},l=r.extend(i,u,s);return new a.PathNode(e.state).applyRawParams(l)}return void 0===n&&(n=[]),e.map(s)},t.treeChanges=function(t,e,n){function r(t,n){var r=a.PathNode.clone(t);return r.paramValues=e[n].paramValues,r}for(var o=0,s=Math.min(t.length,e.length),u=function(t){return t.parameters({inherit:!1}).filter(i.not(i.prop("dynamic"))).map(i.prop("id"))},l=function(t,e){return t.equals(e,u(t.state))};o<s&&t[o].state!==n&&l(t[o],e[o]);)o++;var c,f,p,h,d;c=t,f=c.slice(0,o),p=c.slice(o);var v=f.map(r);return h=e.slice(o),d=v.concat(h),{from:c,to:d,retained:f,exiting:p,entering:h}},t.subPath=function(t,e){var n=r.find(t,e),i=t.indexOf(n);return i===-1?void 0:t.slice(0,i+1)},t.paramValues=function(t){return t.reduce(function(t,e){return r.extend(t,e.paramValues)},{})},t}();e.PathFactory=s},function(t,e,n){"use strict";var r=n(3),i=n(5),o=n(22),a=function(){function t(e){if(e instanceof t){var n=e;this.state=n.state,this.paramSchema=n.paramSchema.slice(),this.paramValues=r.extend({},n.paramValues),this.resolvables=n.resolvables.slice(),this.views=n.views&&n.views.slice()}else this.state=e,this.paramSchema=e.parameters({inherit:!1}),this.paramValues={},this.resolvables=e.resolvables.map(function(t){return t.clone()})}return t.prototype.applyRawParams=function(t){var e=function(e){return[e.id,e.value(t[e.id])]};return this.paramValues=this.paramSchema.reduce(function(t,n){return r.applyPairs(t,e(n))},{}),this},t.prototype.parameter=function(t){return r.find(this.paramSchema,i.propEq("id",t))},t.prototype.equals=function(t,e){var n=this;void 0===e&&(e=this.paramSchema.map(i.prop("id")));var o=function(e){return n.parameter(e).type.equals(n.paramValues[e],t.paramValues[e])};return this.state===t.state&&e.map(o).reduce(r.allTrueR,!0)},t.clone=function(e){return new t(e)},t.matching=function(t,e){for(var n=[],r=0;r<t.length&&r<e.length;r++){var i=t[r],a=e[r];if(i.state!==a.state)break;if(!o.Param.equals(i.paramSchema,i.paramValues,a.paramValues))break;n.push(i)}return n},t}();e.PathNode=a},function(t,e,n){"use strict";function r(t){return t=v(t)&&{value:t}||t,s.extend(t,{$$fn:l.isInjectable(t.value)?t.value:function(){return t.value}})}function i(t,e,n,r){if(t.type&&e&&"string"!==e.name)throw new Error("Param '"+r+"' has two type configurations.");return t.type&&e&&"string"===e.name&&h.paramTypes.type(t.type)?h.paramTypes.type(t.type):e?e:t.type?t.type instanceof p.ParamType?t.type:h.paramTypes.type(t.type):n===g.CONFIG?h.paramTypes.type("any"):h.paramTypes.type("string")}function o(t,e){var n=t.squash;if(!e||n===!1)return!1;if(!l.isDefined(n)||null==n)return f.matcherConfig.defaultSquashPolicy();if(n===!0||l.isString(n))return n;throw new Error("Invalid squash policy: '"+n+"'. Valid policies: false, true, or arbitrary string")}function a(t,e,n,r){var i,o,a=[{from:"",to:n||e?void 0:""},{from:null,to:n||e?void 0:""}];return i=l.isArray(t.replace)?t.replace:[],l.isString(r)&&i.push({from:r,to:void 0}),o=s.map(i,u.prop("from")),s.filter(a,function(t){return o.indexOf(t.from)===-1}).concat(i)}var s=n(3),u=n(5),l=n(4),c=n(6),f=n(23),p=n(24),h=n(25),d=Object.prototype.hasOwnProperty,v=function(t){return 0===["value","type","squash","array","dynamic"].filter(d.bind(t||{})).length};!function(t){t[t.PATH=0]="PATH",t[t.SEARCH=1]="SEARCH",t[t.CONFIG=2]="CONFIG"}(e.DefType||(e.DefType={}));var g=e.DefType,m=function(){function t(t,e,n,u){function c(){var e={array:u===g.SEARCH&&"auto"},r=t.match(/\[\]$/)?{array:!0}:{};return s.extend(e,r,n).array}n=r(n),e=i(n,e,u,t);var f=c();e=f?e.$asArray(f,u===g.SEARCH):e;var p=void 0!==n.value,h=l.isDefined(n.dynamic)?!!n.dynamic:!!e.dynamic,d=o(n,p),v=a(n,f,p,d);s.extend(this,{id:t,type:e,location:u,squash:d,replace:v,isOptional:p,dynamic:h,config:n,array:f})}return t.prototype.isDefaultValue=function(t){return this.isOptional&&this.type.equals(this.value(),t)},t.prototype.value=function(t){var e=this,n=function(){if(!c.services.$injector)throw new Error("Injectable functions cannot be called at configuration time");var t=c.services.$injector.invoke(e.config.$$fn);if(null!==t&&void 0!==t&&!e.type.is(t))throw new Error("Default value ("+t+") for parameter '"+e.id+"' is not an instance of ParamType ("+e.type.name+")");return t},r=function(t){var n=s.map(s.filter(e.replace,u.propEq("from",t)),u.prop("to"));return n.length?n[0]:t};return t=r(t),l.isDefined(t)?this.type.$normalize(t):n()},t.prototype.isSearch=function(){return this.location===g.SEARCH},t.prototype.validates=function(t){if((!l.isDefined(t)||null===t)&&this.isOptional)return!0;var e=this.type.$normalize(t);if(!this.type.is(e))return!1;var n=this.type.encode(e);return!(l.isString(n)&&!this.type.pattern.exec(n))},t.prototype.toString=function(){return"{Param:"+this.id+" "+this.type+" squash: '"+this.squash+"' optional: "+this.isOptional+"}"},t.fromConfig=function(e,n,r){return new t(e,n,r,g.CONFIG)},t.fromPath=function(e,n,r){return new t(e,n,r,g.PATH)},t.fromSearch=function(e,n,r){return new t(e,n,r,g.SEARCH)},t.values=function(t,e){return void 0===e&&(e={}),t.map(function(t){return[t.id,t.value(e[t.id])]}).reduce(s.applyPairs,{})},t.changed=function(t,e,n){return void 0===e&&(e={}),void 0===n&&(n={}),t.filter(function(t){return!t.type.equals(e[t.id],n[t.id])})},t.equals=function(e,n,r){return void 0===n&&(n={}),void 0===r&&(r={}),0===t.changed(e,n,r).length},t.validates=function(t,e){return void 0===e&&(e={}),t.map(function(t){return t.validates(e[t.id])}).reduce(s.allTrueR,!0)},t}();e.Param=m},function(t,e,n){"use strict";var r=n(4),i=function(){function t(){this._isCaseInsensitive=!1,this._isStrictMode=!0,this._defaultSquashPolicy=!1}return t.prototype.caseInsensitive=function(t){return this._isCaseInsensitive=r.isDefined(t)?t:this._isCaseInsensitive},t.prototype.strictMode=function(t){return this._isStrictMode=r.isDefined(t)?t:this._isStrictMode},t.prototype.defaultSquashPolicy=function(t){if(r.isDefined(t)&&t!==!0&&t!==!1&&!r.isString(t))throw new Error("Invalid squash policy: "+t+". Valid policies: false, true, arbitrary-string");return this._defaultSquashPolicy=r.isDefined(t)?t:this._defaultSquashPolicy},t}();e.MatcherConfig=i,e.matcherConfig=new i},function(t,e,n){"use strict";function r(t,e){function n(t){return o.isArray(t)?t:o.isDefined(t)?[t]:[]}function r(t){switch(t.length){case 0:return;case 1:return"auto"===e?t[0]:t;default:return t}}function a(t,e){return function(a){if(o.isArray(a)&&0===a.length)return a;var s=n(a),u=i.map(s,t);return e===!0?0===i.filter(u,function(t){return!t}).length:r(u)}}function s(t){return function(e,r){var i=n(e),o=n(r);if(i.length!==o.length)return!1;for(var a=0;a<i.length;a++)if(!t(i[a],o[a]))return!1;return!0}}var u=this;["encode","decode","equals","$normalize"].map(function(e){u[e]=("equals"===e?s:a)(t[e].bind(t))}),i.extend(this,{dynamic:t.dynamic,name:t.name,pattern:t.pattern,is:a(t.is.bind(t),!0),$arrayMode:e})}var i=n(3),o=n(4),a=function(){function t(t){this.pattern=/.*/,i.extend(this,t)}return t.prototype.is=function(t,e){return!0},t.prototype.encode=function(t,e){return t},t.prototype.decode=function(t,e){return t},t.prototype.equals=function(t,e){return t==e},t.prototype.$subPattern=function(){var t=this.pattern.toString();return t.substr(1,t.length-2)},t.prototype.toString=function(){return"{ParamType:"+this.name+"}"},t.prototype.$normalize=function(t){return this.is(t)?t:this.decode(t)},t.prototype.$asArray=function(t,e){if(!t)return this;if("auto"===t&&!e)throw new Error("'auto' array mode is for query parameters only");return new r(this,t)},t}();e.ParamType=a},function(t,e,n){"use strict";function r(t){return null!=t?t.toString().replace(/~/g,"~~").replace(/\//g,"~2F"):t}function i(t){return null!=t?t.toString().replace(/~2F/g,"/").replace(/~~/g,"~"):t}var o=n(3),a=n(4),s=n(5),u=n(6),l=n(24),c=function(){function t(){this.enqueue=!0,this.typeQueue=[],this.defaultTypes={hash:{encode:r,decode:i,is:s.is(String),pattern:/.*/,equals:function(t,e){return t==e}},string:{encode:r,decode:i,is:s.is(String),pattern:/[^\/]*/},int:{encode:r,decode:function(t){return parseInt(t,10)},is:function(t){return a.isDefined(t)&&this.decode(t.toString())===t},pattern:/-?\d+/},bool:{encode:function(t){return t&&1||0},decode:function(t){return 0!==parseInt(t,10)},is:s.is(Boolean),pattern:/0|1/},date:{encode:function(t){return this.is(t)?[t.getFullYear(),("0"+(t.getMonth()+1)).slice(-2),("0"+t.getDate()).slice(-2)].join("-"):void 0},decode:function(t){if(this.is(t))return t;var e=this.capture.exec(t);return e?new Date(e[1],e[2]-1,e[3]):void 0},is:function(t){return t instanceof Date&&!isNaN(t.valueOf())},equals:function(t,e){return["getFullYear","getMonth","getDate"].reduce(function(n,r){return n&&t[r]()===e[r]()},!0)},pattern:/[0-9]{4}-(?:0[1-9]|1[0-2])-(?:0[1-9]|[1-2][0-9]|3[0-1])/,capture:/([0-9]{4})-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1])/},json:{encode:o.toJson,decode:o.fromJson,is:s.is(Object),equals:o.equals,pattern:/[^\/]*/},any:{encode:o.identity,decode:o.identity,equals:o.equals,pattern:/.*/}};var t=function(t,e){return new l.ParamType(o.extend({name:e},t))};this.types=o.inherit(o.map(this.defaultTypes,t),{})}return t.prototype.type=function(t,e,n){if(!a.isDefined(e))return this.types[t];if(this.types.hasOwnProperty(t))throw new Error("A type named '"+t+"' has already been defined.");return this.types[t]=new l.ParamType(o.extend({name:t},e)),n&&(this.typeQueue.push({name:t,def:n}),this.enqueue||this._flushTypeQueue()),this},t.prototype._flushTypeQueue=function(){for(;this.typeQueue.length;){var t=this.typeQueue.shift();if(t.pattern)throw new Error("You cannot override a type's .pattern at runtime.");o.extend(this.types[t.name],u.services.$injector.invoke(t.def))}},t}();e.ParamTypes=c,e.paramTypes=new c},function(t,e,n){"use strict";var r=n(27),i=n(29),o=n(30),a=n(29),s=n(31),u=n(37),l=n(38),c=n(43),f=n(44),p=function(){function t(){this.viewService=new u.ViewService,this.transitionService=new s.TransitionService(this),this.globals=new f.Globals(this.transitionService),this.urlMatcherFactory=new r.UrlMatcherFactory,this.urlRouterProvider=new i.UrlRouterProvider(this.urlMatcherFactory,this.globals.params),this.urlRouter=new a.UrlRouter(this.urlRouterProvider),this.stateRegistry=new l.StateRegistry(this.urlMatcherFactory,this.urlRouterProvider),this.stateProvider=new o.StateProvider(this.stateRegistry),this.stateService=new c.StateService(this),this.viewService.rootContext(this.stateRegistry.root()),this.globals.$current=this.stateRegistry.root(),this.globals.current=this.globals.$current.self}return t}();e.UIRouter=p},function(t,e,n){"use strict";function r(){return{strict:s.matcherConfig.strictMode(),caseInsensitive:s.matcherConfig.caseInsensitive()}}var i=n(3),o=n(4),a=n(28),s=n(23),u=n(22),l=n(25),c=function(){function t(){i.extend(this,{UrlMatcher:a.UrlMatcher,Param:u.Param})}return t.prototype.caseInsensitive=function(t){return s.matcherConfig.caseInsensitive(t)},t.prototype.strictMode=function(t){return s.matcherConfig.strictMode(t)},t.prototype.defaultSquashPolicy=function(t){return s.matcherConfig.defaultSquashPolicy(t)},t.prototype.compile=function(t,e){return new a.UrlMatcher(t,i.extend(r(),e))},t.prototype.isMatcher=function(t){if(!o.isObject(t))return!1;var e=!0;return i.forEach(a.UrlMatcher.prototype,function(n,r){o.isFunction(n)&&(e=e&&o.isDefined(t[r])&&o.isFunction(t[r]))}),e},t.prototype.type=function(t,e,n){var r=l.paramTypes.type(t,e,n);return o.isDefined(e)?this:r},t.prototype.$get=function(){return l.paramTypes.enqueue=!1,l.paramTypes._flushTypeQueue(),this},t}();e.UrlMatcherFactory=c},function(t,e,n){"use strict";function r(t,e){var n=["",""],r=t.replace(/[\\\[\]\^$*+?.()|{}]/g,"\\$&");if(!e)return r;switch(e.squash){case!1:n=["(",")"+(e.isOptional?"?":"")];break;case!0:r=r.replace(/\/$/,""),n=["(?:/(",")|/)?"];break;default:n=["("+e.squash+"|",")?"]}return r+n[0]+e.type.pattern.source+n[1]}var i=n(3),o=n(5),a=n(4),s=n(22),u=n(25),l=n(4),c=n(22),f=n(3),p=n(3),h=function(t,e,n){return t[e]=t[e]||n()},d=function(){function t(e,n){var a=this;this.config=n,this._cache={path:[],pattern:null},this._children=[],this._params=[],this._segments=[],this._compiled=[],this.pattern=e,this.config=i.defaults(this.config,{params:{},strict:!0,caseInsensitive:!1,paramMap:i.identity});for(var l,c,f,p=/([:*])([\w\[\]]+)|\{([\w\[\]]+)(?:\:\s*((?:[^{}\\]+|\\.|\{(?:[^{}\\]+|\\.)*\})+))?\}/g,h=/([:]?)([\w\[\].-]+)|\{([\w\[\].-]+)(?:\:\s*((?:[^{}\\]+|\\.|\{(?:[^{}\\]+|\\.)*\})+))?\}/g,d=0,v=[],g=function(n){if(!t.nameValidator.test(n))throw new Error("Invalid parameter name '"+n+"' in pattern '"+e+"'");if(i.find(a._params,o.propEq("id",n)))throw new Error("Duplicate parameter name '"+n+"' in pattern '"+e+"'")},m=function(t,n){var r=t[2]||t[3],o=n?t[4]:t[4]||("*"===t[1]?".*":null);return{id:r,regexp:o,cfg:a.config.params[r],segment:e.substring(d,t.index),type:o?u.paramTypes.type(o||"string")||i.inherit(u.paramTypes.type("string"),{pattern:new RegExp(o,a.config.caseInsensitive?"i":void 0)}):null}};(l=p.exec(e))&&(c=m(l,!1),!(c.segment.indexOf("?")>=0));)g(c.id),this._params.push(s.Param.fromPath(c.id,c.type,this.config.paramMap(c.cfg,!1))),this._segments.push(c.segment),v.push([c.segment,i.tail(this._params)]),d=p.lastIndex;f=e.substring(d);var y=f.indexOf("?");if(y>=0){var b=f.substring(y);if(f=f.substring(0,y),b.length>0)for(d=0;l=h.exec(b);)c=m(l,!0),g(c.id),this._params.push(s.Param.fromSearch(c.id,c.type,this.config.paramMap(c.cfg,!0))),d=p.lastIndex}this._segments.push(f),i.extend(this,{_compiled:v.map(function(t){return r.apply(null,t)}).concat(r(f)),prefix:this._segments[0]}),Object.freeze(this)}return t.prototype.append=function(t){return this._children.push(t),i.forEach(t._cache,function(e,n){return t._cache[n]=a.isArray(e)?[]:null}),t._cache.path=this._cache.path.concat(this),t},t.prototype.isRoot=function(){return 0===this._cache.path.length},t.prototype.toString=function(){return this.pattern},t.prototype.exec=function(t,e,n,r){function a(t){var e=function(t){return t.split("").reverse().join("")},n=function(t){return t.replace(/\\-/g,"-")},r=e(t).split(/-(?!\\)/),o=i.map(r,e);return i.map(o,n).reverse()}var s=this;void 0===e&&(e={}),void 0===r&&(r={});var u=h(this._cache,"pattern",function(){return new RegExp(["^",i.unnest(s._cache.path.concat(s).map(o.prop("_compiled"))).join(""),s.config.strict===!1?"/?":"","$"].join(""),s.config.caseInsensitive?"i":void 0)}).exec(t);if(!u)return null;var c=this.parameters(),f=c.filter(function(t){return!t.isSearch()}),p=c.filter(function(t){return t.isSearch()}),d=this._cache.path.concat(this).map(function(t){return t._segments.length-1}).reduce(function(t,e){return t+e}),v={};if(d!==u.length-1)throw new Error("Unbalanced capture group in route '"+this.pattern+"'");for(var g=0;g<d;g++){for(var m=f[g],y=u[g+1],b=0;b<m.replace.length;b++)m.replace[b].from===y&&(y=m.replace[b].to);y&&m.array===!0&&(y=a(y)),l.isDefined(y)&&(y=m.type.decode(y)),v[m.id]=m.value(y)}return i.forEach(p,function(t){for(var n=e[t.id],r=0;r<t.replace.length;r++)t.replace[r].from===n&&(n=t.replace[r].to);l.isDefined(n)&&(n=t.type.decode(n)),v[t.id]=t.value(n)}),n&&(v["#"]=n),v},t.prototype.parameters=function(t){return void 0===t&&(t={}),t.inherit===!1?this._params:i.unnest(this._cache.path.concat(this).map(o.prop("_params")))},t.prototype.parameter=function(t,e){void 0===e&&(e={});var n=i.tail(this._cache.path);return i.find(this._params,o.propEq("id",t))||e.inherit!==!1&&n&&n.parameter(t)||null},t.prototype.validates=function(t){var e=this,n=function(t,e){return!t||t.validates(e)};return i.pairs(t||{}).map(function(t){var r=t[0],i=t[1];return n(e.parameter(r),i)}).reduce(i.allTrueR,!0)},t.prototype.format=function(e){function n(t){var n=t.value(e[t.id]),r=t.isDefaultValue(n),i=!!r&&t.squash,o=t.type.encode(n);return{param:t,value:n,isDefaultValue:r,squash:i,encoded:o}}if(void 0===e&&(e={}),!this.validates(e))return null;var r=this._cache.path.slice().concat(this),o=r.map(t.pathSegmentsAndParams).reduce(f.unnestR,[]),s=r.map(t.queryParams).reduce(f.unnestR,[]),u=o.reduce(function(e,r){if(a.isString(r))return e+r;var o=n(r),s=o.squash,u=o.encoded,l=o.param;return s===!0?e.match(/\/$/)?e.slice(0,-1):e:a.isString(s)?e+s:s!==!1?e:null==u?e:a.isArray(u)?e+i.map(u,t.encodeDashes).join("-"):l.type.raw?e+u:e+encodeURIComponent(u)},""),l=s.map(function(t){var e=n(t),r=e.squash,o=e.encoded,s=e.isDefaultValue;if(!(null==o||s&&r!==!1)&&(a.isArray(o)||(o=[o]),0!==o.length))return t.type.raw||(o=i.map(o,encodeURIComponent)),o.map(function(e){return t.id+"="+e})}).filter(i.identity).reduce(f.unnestR,[]).join("&");return u+(l?"?"+l:"")+(e["#"]?"#"+e["#"]:"")},t.encodeDashes=function(t){return encodeURIComponent(t).replace(/-/g,function(t){return"%5C%"+t.charCodeAt(0).toString(16).toUpperCase()})},t.pathSegmentsAndParams=function(t){var e=t._segments,n=t._params.filter(function(t){return t.location===c.DefType.PATH});return p.arrayTuples(e,n.concat(void 0)).reduce(f.unnestR,[]).filter(function(t){return""!==t&&l.isDefined(t)})},t.queryParams=function(t){return t._params.filter(function(t){return t.location===c.DefType.SEARCH})},t.nameValidator=/^\w+([-.]+\w+)*(?:\[\])?$/,t}();e.UrlMatcher=d},function(t,e,n){"use strict";function r(t){var e=/^\^((?:\\[^a-zA-Z0-9]|[^\\\[\]\^$*+?.()|{}]+)*)/.exec(t.source);return null!=e?e[1].replace(/\\(.)/g,"$1"):""}function i(t,e){return t.replace(/\$(\$|\d{1,2})/,function(t,n){return e["$"===n?0:Number(n)]})}function o(t,e,n,r){if(!r)return!1;var i=t.invoke(n,n,{$match:r,$stateParams:e});return!l.isDefined(i)||i}function a(t,e,n){var r=c.services.locationConfig.baseHref();return"/"===r?t:e?r.slice(0,-1)+t:n?r.slice(1)+t:t}function s(t,e,n){function r(t){var e=t(c.services.$injector,f);return!!e&&(l.isString(e)&&(f.replace(),f.url(e)),!0)}if(!n||!n.defaultPrevented){var i,o=t.length;for(i=0;i<o;i++)if(r(t[i]))return;e&&r(e)}}var u=n(3),l=n(4),c=n(6),f=c.services.location,p=function(){function t(t,e){this.rules=[],this.interceptDeferred=!1,this.$urlMatcherFactory=t,this.$stateParams=e}return t.prototype.rule=function(t){if(!l.isFunction(t))throw new Error("'rule' must be a function");return this.rules.push(t),this},t.prototype.otherwise=function(t){if(!l.isFunction(t)&&!l.isString(t))throw new Error("'rule' must be a string or function");return this.otherwiseFn=l.isString(t)?function(){return t}:t,this},t.prototype.when=function(t,e){var n,a=this,s=a.$urlMatcherFactory,p=a.$stateParams,h=l.isString(e);if(l.isString(t)&&(t=s.compile(t)),!h&&!l.isFunction(e)&&!l.isArray(e))throw new Error("invalid 'handler' in when()");var d={matcher:function(t,e){return h&&(n=s.compile(e),e=["$match",n.format.bind(n)]),u.extend(function(){return o(c.services.$injector,p,e,t.exec(f.path(),f.search(),f.hash()))},{prefix:l.isString(t.prefix)?t.prefix:""})},regex:function(t,e){if(t.global||t.sticky)throw new Error("when() RegExp must not be global or sticky");return h&&(n=e,e=["$match",function(t){return i(n,t)}]),u.extend(function(){return o(c.services.$injector,p,e,t.exec(f.path()))},{prefix:r(t)})}},v={matcher:s.isMatcher(t),regex:t instanceof RegExp};for(var g in v)if(v[g])return this.rule(d[g](t,e));throw new Error("invalid 'what' in when()")},t.prototype.deferIntercept=function(t){void 0===t&&(t=!0),this.interceptDeferred=t},t}();e.UrlRouterProvider=p;var h=function(){function t(e){this.urlRouterProvider=e,u.bindFunctions(t.prototype,this,this)}return t.prototype.sync=function(){s(this.urlRouterProvider.rules,this.urlRouterProvider.otherwiseFn)},t.prototype.listen=function(){var t=this;return this.listener=this.listener||f.onChange(function(e){return s(t.urlRouterProvider.rules,t.urlRouterProvider.otherwiseFn,e)})},t.prototype.update=function(t){return t?void(this.location=f.url()):void(f.url()!==this.location&&(f.url(this.location),f.replace()))},t.prototype.push=function(t,e,n){f.url(t.format(e||{})),n&&n.replace&&f.replace()},t.prototype.href=function(t,e,n){if(!t.validates(e))return null;var r=t.format(e);n=n||{absolute:!1};var i=c.services.locationConfig,o=i.html5Mode();if(o||null===r||(r="#"+i.hashPrefix()+r),r=a(r,o,n.absolute),!n.absolute||!r)return r;var s=!o&&r?"/":"",u=i.port();return u=80===u||443===u?"":":"+u,[i.protocol(),"://",i.host(),u,s,r].join("")},t}();e.UrlRouter=h},function(t,e,n){"use strict";var r=n(4),i=n(3),o=function(){function t(e){this.stateRegistry=e,this.invalidCallbacks=[],i.bindFunctions(t.prototype,this,this)}return t.prototype.decorator=function(t,e){return this.stateRegistry.decorator(t,e)||this},t.prototype.state=function(t,e){return r.isObject(t)?e=t:e.name=t,this.stateRegistry.register(e),this},t.prototype.onInvalid=function(t){this.invalidCallbacks.push(t)},t}();e.StateProvider=o},function(t,e,n){"use strict";var r=n(11),i=n(15),o=n(32),a=n(33),s=n(34),u=n(35),l=n(36),c=n(5);e.defaultTransOpts={location:!0,relative:null,inherit:!1,notify:!0,reload:!1,custom:{},current:function(){return null}};var f=function(){function t(t){this._router=t,this.$view=t.viewService,i.HookRegistry.mixin(new i.HookRegistry,this),this._deregisterHookFns={},this.registerTransitionHooks()}return t.prototype.registerTransitionHooks=function(){var t=this._deregisterHookFns;t.redirectTo=this.onStart({to:function(t){return!!t.redirectTo}},u.redirectToHook),t.onExit=this.onExit({exiting:function(t){return!!t.onExit}},l.onExitHook),t.onRetain=this.onRetain({retained:function(t){return!!t.onRetain}},l.onRetainHook),t.onEnter=this.onEnter({entering:function(t){return!!t.onEnter}},l.onEnterHook),t.eagerResolve=this.onStart({},o.eagerResolvePath,{priority:1e3}),t.lazyResolve=this.onEnter({entering:c.val(!0)},o.lazyResolveState,{priority:1e3}),t.loadViews=this.onStart({},a.loadEnteringViews),t.activateViews=this.onSuccess({},a.activateViews),t.updateUrl=this.onSuccess({},s.updateUrl,{priority:9999})},t.prototype.onBefore=function(t,e,n){throw""},t.prototype.onStart=function(t,e,n){throw""},t.prototype.onExit=function(t,e,n){throw""},t.prototype.onRetain=function(t,e,n){throw""},t.prototype.onEnter=function(t,e,n){throw""},t.prototype.onFinish=function(t,e,n){throw""},t.prototype.onSuccess=function(t,e,n){throw""},t.prototype.onError=function(t,e,n){throw""},t.prototype.create=function(t,e){return new r.Transition(t,e,this._router)},t}();e.TransitionService=f},function(t,e,n){"use strict";var r=n(3),i=n(17);e.eagerResolvePath=function(t){return new i.ResolveContext(t.treeChanges().to).resolvePath("EAGER",t).then(r.noop)},e.lazyResolveState=function(t,e){return new i.ResolveContext(t.treeChanges().to).subContext(e).resolvePath("LAZY",t).then(r.noop)}},function(t,e,n){"use strict";function r(t){var e=t.views("entering");if(e.length)return a.services.$q.all(e.map(function(t){return t.load()})).then(o.noop)}function i(t){var e=t.views("entering"),n=t.views("exiting");if(e.length||n.length){var r=t.router.viewService;n.forEach(function(t){return r.deactivateViewConfig(t)}),e.forEach(function(t){return r.activateViewConfig(t)}),r.sync()}}var o=n(3),a=n(6);e.loadEnteringViews=r,e.activateViews=i},function(t,e){"use strict";function n(t){var e=t.options(),n=t.router.stateService,r=t.router.urlRouter;if(e.location&&n.$current.navigable){var i={replace:"replace"===e.location};r.push(n.$current.navigable.url,n.params,i)}r.update(!0)}e.updateUrl=n},function(t,e,n){"use strict";var r=n(4),i=n(6),o=n(14);e.redirectToHook=function(t){function e(e){var n=t.router.stateService;return e instanceof o.TargetState?e:r.isString(e)?n.target(e,t.params(),t.options()):e.state||e.params?n.target(e.state||t.to(),e.params||t.params(),t.options()):void 0}var n=t.to().redirectTo;if(n)return r.isFunction(n)?i.services.$q.when(n(t)).then(e):e(n)}},function(t,e){"use strict";function n(t){return function(e,n){return n[t](e,n)}}e.onExitHook=n("onExit"),e.onRetainHook=n("onRetain"),e.onEnterHook=n("onEnter")},function(t,e,n){"use strict";var r=n(3),i=n(5),o=n(4),a=n(12),s=function(){function t(){var t=this;this.uiViews=[],this.viewConfigs=[],this._viewConfigFactories={},this.sync=function(){function e(t){return t.fqn.split(".").length}function n(t){for(var e=t.viewDecl.$context,n=0;++n&&e.parent;)e=e.parent;return n}var o=t.uiViews.map(function(t){return[t.fqn,t]}).reduce(r.applyPairs,{}),a=function(t){return function(e){if(t.$type!==e.viewDecl.$type)return!1;var n=e.viewDecl,i=n.$uiViewName.split("."),a=t.fqn.split(".");if(!r.equals(i,a.slice(0-i.length)))return!1;var s=1-i.length||void 0,u=a.slice(0,s).join("."),l=o[u].creationContext;return n.$uiViewContextAnchor===(l&&l.name)}},s=i.curry(function(t,e,n,r){return e*(t(n)-t(r))}),u=function(e){var r=t.viewConfigs.filter(a(e));return r.length>1&&r.sort(s(n,-1)),[e,r[0]]},l=function(e){var n=e[0],r=e[1];t.uiViews.indexOf(n)!==-1&&n.configUpdated(r)};t.uiViews.sort(s(e,1)).map(u).forEach(l)}}return t.prototype.rootContext=function(t){return this._rootContext=t||this._rootContext},t.prototype.viewConfigFactory=function(t,e){this._viewConfigFactories[t]=e},t.prototype.createViewConfig=function(t,e){var n=this._viewConfigFactories[e.$type];if(!n)throw new Error("ViewService: No view config factory registered for type "+e.$type);var r=n(t,e);return o.isArray(r)?r:[r]},t.prototype.deactivateViewConfig=function(t){a.trace.traceViewServiceEvent("<- Removing",t),r.removeFrom(this.viewConfigs,t)},t.prototype.activateViewConfig=function(t){a.trace.traceViewServiceEvent("-> Registering",t),this.viewConfigs.push(t)},t.prototype.registerUIView=function(t){a.trace.traceViewServiceUIViewEvent("-> Registering",t);var e=this.uiViews,n=function(e){return e.fqn===t.fqn};return e.filter(n).length&&a.trace.traceViewServiceUIViewEvent("!!!! duplicate uiView named:",t),e.push(t),this.sync(),function(){var n=e.indexOf(t);return n<=0?void a.trace.traceViewServiceUIViewEvent("Tried removing non-registered uiView",t):(a.trace.traceViewServiceUIViewEvent("<- Deregistering",t),void r.removeFrom(e)(t))}},t.prototype.available=function(){return this.uiViews.map(i.prop("fqn"))},t.prototype.active=function(){return this.uiViews.filter(i.prop("$config")).map(i.prop("name"))},t.normalizeUIViewTarget=function(t,e){void 0===e&&(e="");var n=e.split("@"),r=n[0]||"$default",i=o.isString(n[1])?n[1]:"^",a=/^(\^(?:\.\^)*)\.(.*$)/.exec(r);a&&(i=a[1],r=a[2]),"!"===r.charAt(0)&&(r=r.substr(1),i="");var s=/^(\^(?:\.\^)*)$/;if(s.exec(i)){var u=i.split(".").reduce(function(t,e){return t.parent},t);i=u.name}return{uiViewName:r,uiViewContextAnchor:i}},t}();e.ViewService=s},function(t,e,n){"use strict";var r=n(39),i=n(40),o=n(41),a=function(){function t(t,e){this.states={},this.matcher=new r.StateMatcher(this.states),this.builder=new i.StateBuilder(this.matcher,t),this.stateQueue=new o.StateQueueManager(this.states,this.builder,e);var n={name:"",url:"^",views:null,params:{"#":{value:null,type:"hash",dynamic:!0}},abstract:!0},a=this._root=this.stateQueue.register(n);a.navigable=null}return t.prototype.root=function(){return this._root},t.prototype.register=function(t){return this.stateQueue.register(t)},t.prototype.get=function(t,e){var n=this;if(0===arguments.length)return Object.keys(this.states).map(function(t){return n.states[t].self});var r=this.matcher.find(t,e);return r&&r.self||null},t.prototype.decorator=function(t,e){return this.builder.builder(t,e)},t}();e.StateRegistry=a},function(t,e,n){"use strict";var r=n(4),i=function(){function t(t){this._states=t}return t.prototype.isRelative=function(t){return t=t||"",0===t.indexOf(".")||0===t.indexOf("^")},t.prototype.find=function(t,e){if(t||""===t){var n=r.isString(t),i=n?t:t.name;this.isRelative(i)&&(i=this.resolvePath(i,e));var o=this._states[i];return!o||!n&&(n||o!==t&&o.self!==t)?void 0:o}},t.prototype.resolvePath=function(t,e){if(!e)throw new Error("No reference point given for path '"+t+"'");for(var n=this.find(e),r=t.split("."),i=0,o=r.length,a=n;i<o;i++)if(""!==r[i]||0!==i){if("^"!==r[i])break;if(!a.parent)throw new Error("Path '"+t+"' not valid for state '"+n.name+"'");a=a.parent}else a=n;var s=r.slice(i).join(".");return a.name+(a.name&&s?".":"")+s},t}();e.StateMatcher=i},function(t,e,n){"use strict";function r(t){return t.self.$$state=function(){return t},t.self}function i(t){return t.parent&&t.parent.data&&(t.data=t.self.data=l.inherit(t.parent.data,t.data)),t.data}function o(t){var e=function(t,e){return h.Param.fromConfig(e,null,t)},n=t.url&&t.url.parameters({inherit:!1})||[],r=l.values(l.map(l.omit(t.params||{},n.map(p.prop("id"))),e));return n.concat(r).map(function(t){return[t.id,t]}).reduce(l.applyPairs,{})}function a(t){return t.parent?t.parent.path.concat(t):[t]}function s(t){var e=t.parent?l.extend({},t.parent.includes):{};return e[t.name]=!0,
e}function u(t){var e=function(t){return Object.keys(t||{}).map(function(e){return{token:e,val:t[e],deps:void 0}})},n=function(t){return t.$inject||v.services.$injector.annotate(t,v.services.$injector.strictDi)},r=function(t){return!(!t.token||!t.resolveFn)},i=function(t){return!(!t.provide&&!t.token||!(t.useValue||t.useFactory||t.useExisting||t.useClass))},o=function(t){return!!(t&&t.val&&(c.isString(t.val)||c.isArray(t.val)||c.isFunction(t.val)))},a=function(t){return t.provide||t.token},s=p.pattern([[p.prop("resolveFn"),function(t){return new d.Resolvable(a(t),t.resolveFn,t.deps,t.policy)}],[p.prop("useFactory"),function(t){return new d.Resolvable(a(t),t.useFactory,t.deps||t.dependencies,t.policy)}],[p.prop("useClass"),function(t){return new d.Resolvable(a(t),function(){return new t.useClass},[],t.policy)}],[p.prop("useValue"),function(t){return new d.Resolvable(a(t),function(){return t.useValue},[],t.policy,t.useValue)}],[p.prop("useExisting"),function(t){return new d.Resolvable(a(t),function(t){return t},[t.useExisting],t.policy)}]]),u=p.pattern([[p.pipe(p.prop("val"),c.isString),function(t){return new d.Resolvable(t.token,function(t){return t},[t.val],t.policy)}],[p.pipe(p.prop("val"),c.isArray),function(t){return new d.Resolvable(t.token,l.tail(t.val),t.val.slice(0,-1),t.policy)}],[p.pipe(p.prop("val"),c.isFunction),function(t){return new d.Resolvable(t.token,t.val,n(t.val),t.policy)}]]),h=p.pattern([[p.is(d.Resolvable),function(t){return t}],[r,s],[i,s],[o,u],[p.val(!0),function(t){throw new Error("Invalid resolve value: "+f.stringify(t))}]]),g=t.resolve,m=c.isArray(g)?g:e(g);return m.map(h)}var l=n(3),c=n(4),f=n(9),p=n(5),h=n(22),d=n(19),v=n(6),g=function(t){if(!c.isString(t))return!1;var e="^"===t.charAt(0);return{val:e?t.substring(1):t,root:e}},m=function(t,e){return function(n){var r=n,i=g(r.url),o=n.parent,a=i?t.compile(i.val,{params:n.params||{},paramMap:function(t,e){return r.reloadOnSearch===!1&&e&&(t=l.extend(t||{},{dynamic:!0})),t}}):r.url;if(!a)return null;if(!t.isMatcher(a))throw new Error("Invalid url '"+a+"' in state '"+n+"'");return i&&i.root?a:(o&&o.navigable||e()).url.append(a)}},y=function(t){return function(e){return!t(e)&&e.url?e:e.parent?e.parent.navigable:null}};e.resolvablesBuilder=u;var b=function(){function t(t,e){function n(e){return f(e)?null:t.find(l.parentName(e))||c()}this.matcher=t;var l=this,c=function(){return t.find("")},f=function(t){return""===t.name};this.builders={self:[r],parent:[n],data:[i],url:[m(e,c)],navigable:[y(f)],params:[o],views:[],path:[a],includes:[s],resolvables:[u]}}return t.prototype.builder=function(t,e){var n=this.builders,r=n[t]||[];return c.isString(t)&&!c.isDefined(e)?r.length>1?r:r[0]:c.isString(t)&&c.isFunction(e)?(n[t]=r,n[t].push(e),function(){return n[t].splice(n[t].indexOf(e,1))&&null}):void 0},t.prototype.build=function(t){var e=this,n=e.matcher,r=e.builders,i=this.parentName(t);if(i&&!n.find(i))return null;for(var o in r)if(r.hasOwnProperty(o)){var a=r[o].reduce(function(t,e){return function(n){return e(n,t)}},l.noop);t[o]=a(t)}return t},t.prototype.parentName=function(t){var e=t.name||"";return e.indexOf(".")!==-1?e.substring(0,e.lastIndexOf(".")):t.parent?c.isString(t.parent)?t.parent:t.parent.name:""},t.prototype.name=function(t){var e=t.name;if(e.indexOf(".")!==-1||!t.parent)return e;var n=c.isString(t.parent)?t.parent:t.parent.name;return n?n+"."+e:e},t}();e.StateBuilder=b},function(t,e,n){"use strict";var r=n(3),i=n(4),o=n(42),a=function(){function t(t,e,n){this.states=t,this.builder=e,this.$urlRouterProvider=n,this.queue=[]}return t.prototype.register=function(t){var e=this,n=e.states,a=e.queue,s=e.$state,u=r.inherit(new o.State,r.extend({},t,{self:t,resolve:t.resolve||[],toString:function(){return t.name}}));if(!i.isString(u.name))throw new Error("State must have a valid name");if(n.hasOwnProperty(u.name)||r.pluck(a,"name").indexOf(u.name)!==-1)throw new Error("State '"+u.name+"' is already defined");return a.push(u),this.$state&&this.flush(s),u},t.prototype.flush=function(t){for(var e,n,r,i=this,o=i.queue,a=i.states,s=i.builder,u=[],l={};o.length>0;)if(n=o.shift(),e=s.build(n),r=u.indexOf(n),e){if(a.hasOwnProperty(n.name))throw new Error("State '"+name+"' is already defined");a[n.name]=n,this.attachRoute(t,n),r>=0&&u.splice(r,1)}else{var c=l[n.name];if(l[n.name]=o.length,r>=0&&c===o.length)return a;r<0&&u.push(n),o.push(n)}return a},t.prototype.autoFlush=function(t){this.$state=t,this.flush(t)},t.prototype.attachRoute=function(t,e){var n=this.$urlRouterProvider;!e[r.abstractKey]&&e.url&&n.when(e.url,["$match","$stateParams",function(n,i){t.$current.navigable===e&&r.equalForKeys(n,i)||t.transitionTo(e,n,{inherit:!0,location:!1})}])},t}();e.StateQueueManager=a},function(t,e,n){"use strict";var r=n(3),i=n(5),o=function(){function t(t){r.extend(this,t)}return t.prototype.is=function(t){return this===t||this.self===t||this.fqn()===t},t.prototype.fqn=function(){if(!(this.parent&&this.parent instanceof this.constructor))return this.name;var t=this.parent.fqn();return t?t+"."+this.name:this.name},t.prototype.root=function(){return this.parent&&this.parent.root()||this},t.prototype.parameters=function(t){t=r.defaults(t,{inherit:!0});var e=t.inherit&&this.parent&&this.parent.parameters()||[];return e.concat(r.values(this.params))},t.prototype.parameter=function(t,e){return void 0===e&&(e={}),this.url&&this.url.parameter(t,e)||r.find(r.values(this.params),i.propEq("id",t))||e.inherit&&this.parent&&this.parent.parameter(t)},t.prototype.toString=function(){return this.fqn()},t}();e.State=o},function(t,e,n){"use strict";var r=n(3),i=n(4),o=n(8),a=n(6),s=n(20),u=n(21),l=n(31),c=n(10),f=n(14),p=n(22),h=n(7),d=n(3),v=n(3),g=function(){function t(e){this.router=e,this._defaultErrorHandler=function(t){t instanceof Error&&t.stack?console.error(t.stack):t instanceof c.Rejection?(console.error(t),t.detail&&t.detail.stack&&console.error(t.detail.stack)):console.error(t)};var n=["current","$current","params","transition"],r=Object.keys(t.prototype).filter(function(t){return n.indexOf(t)===-1});v.bindFunctions(t.prototype,this,this,r)}return Object.defineProperty(t.prototype,"transition",{get:function(){return this.router.globals.transition},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"params",{get:function(){return this.router.globals.params},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"current",{get:function(){return this.router.globals.current},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"$current",{get:function(){return this.router.globals.$current},enumerable:!0,configurable:!0}),t.prototype._handleInvalidTargetState=function(t,e){function n(){var t=h.dequeue();return void 0===t?c.Rejection.invalid(e.error()).toPromise():g(t).then(m).then(function(t){return t||n()})}var r=this,i=this.router.globals,u=function(){return i.transitionHistory.peekTail()},l=u(),p=s.PathFactory.makeTargetState(t),h=new o.Queue([].concat(this.router.stateProvider.invalidCallbacks)),d=a.services.$q,v=a.services.$injector,g=function(t){return d.when(v.invoke(t,null,{$to$:e,$from$:p}))},m=function(t){if(t instanceof f.TargetState){var e=t;return e=r.target(e.identifier(),e.params(),e.options()),e.valid()?u()!==l?c.Rejection.superseded().toPromise():r.transitionTo(e.identifier(),e.params(),e.options()):c.Rejection.invalid(e.error()).toPromise()}};return n()},t.prototype.reload=function(t){return this.transitionTo(this.current,this.params,{reload:!i.isDefined(t)||t,inherit:!1,notify:!1})},t.prototype.go=function(t,e,n){var i={relative:this.$current,inherit:!0},o=r.defaults(n,i,l.defaultTransOpts);return this.transitionTo(t,e,o)},t.prototype.target=function(t,e,n){if(void 0===n&&(n={}),i.isObject(n.reload)&&!n.reload.name)throw new Error("Invalid reload state object");var r=this.router.stateRegistry;if(n.reloadState=n.reload===!0?r.root():r.matcher.find(n.reload,n.relative),n.reload&&!n.reloadState)throw new Error("No such reload state '"+(i.isString(n.reload)?n.reload:n.reload.name)+"'");var o=r.matcher.find(t,n.relative);return new f.TargetState(t,o,e,n)},t.prototype.transitionTo=function(t,e,n){var i=this;void 0===e&&(e={}),void 0===n&&(n={});var o=this.router,s=o.globals,p=s.transitionHistory;n=r.defaults(n,l.defaultTransOpts),n=r.extend(n,{current:p.peekTail.bind(p)});var h=this.target(t,e,n),d=s.successfulTransitions.peekTail(),v=function(){return[new u.PathNode(i.router.stateRegistry.root())]},g=d?d.treeChanges().to:v();if(!h.exists())return this._handleInvalidTargetState(g,h);if(!h.valid())return r.silentRejection(h.error());var m=function(t){return function(e){if(e instanceof c.Rejection){if(e.type===c.RejectType.IGNORED)return o.urlRouter.update(),s.current;if(e.type===c.RejectType.SUPERSEDED&&e.redirected&&e.detail instanceof f.TargetState){var n=t.redirect(e.detail);return n.run().catch(m(n))}if(e.type===c.RejectType.ABORTED)return o.urlRouter.update(),a.services.$q.reject(e)}var r=i.defaultErrorHandler();return r(e),a.services.$q.reject(e)}},y=this.router.transitionService.create(g,h),b=y.run().catch(m(y));return r.silenceUncaughtInPromise(b),r.extend(b,{transition:y})},t.prototype.is=function(t,e,n){n=r.defaults(n,{relative:this.$current});var o=this.router.stateRegistry.matcher.find(t,n.relative);if(i.isDefined(o))return this.$current===o&&(!i.isDefined(e)||null===e||p.Param.equals(o.parameters(),this.params,e))},t.prototype.includes=function(t,e,n){n=r.defaults(n,{relative:this.$current});var o=i.isString(t)&&h.Glob.fromString(t);if(o){if(!o.matches(this.$current.name))return!1;t=this.$current.name}var a=this.router.stateRegistry.matcher.find(t,n.relative),s=this.$current.includes;if(i.isDefined(a))return!!i.isDefined(s[a.name])&&(!e||d.equalForKeys(p.Param.values(a.parameters(),e),this.params,Object.keys(e)))},t.prototype.href=function(t,e,n){var o={lossy:!0,inherit:!0,absolute:!1,relative:this.$current};n=r.defaults(n,o);var a=this.router.stateRegistry.matcher.find(t,n.relative);if(!i.isDefined(a))return null;n.inherit&&(e=this.params.$inherit(e||{},this.$current,a));var s=a&&n.lossy?a.navigable:a;return s&&void 0!==s.url&&null!==s.url?this.router.urlRouter.href(s.url,p.Param.values(a.parameters(),e),{absolute:n.absolute}):null},t.prototype.defaultErrorHandler=function(t){return this._defaultErrorHandler=t||this._defaultErrorHandler},t.prototype.get=function(t,e){var n=this.router.stateRegistry;return 0===arguments.length?n.get():n.get(t,e||this.$current)},t}();e.StateService=g},function(t,e,n){"use strict";var r=n(45),i=n(8),o=n(3),a=function(){function t(t){var e=this;this.params=new r.StateParams,this.transitionHistory=new i.Queue([],1),this.successfulTransitions=new i.Queue([],1);var n=function(t){e.transition=t,e.transitionHistory.enqueue(t);var n=function(){e.successfulTransitions.enqueue(t),e.$current=t.$to(),e.current=e.$current.self,o.copy(t.params(),e.params)};t.onSuccess({},n,{priority:1e4});var r=function(){e.transition===t&&(e.transition=null)};t.promise.then(r,r)};t.onBefore({},n)}return t}();e.Globals=a},function(t,e,n){"use strict";var r=n(3),i=function(){function t(t){void 0===t&&(t={}),r.extend(this,t)}return t.prototype.$inherit=function(t,e,n){var i,o=r.ancestors(e,n),a={},s=[];for(var u in o)if(o[u]&&o[u].params&&(i=Object.keys(o[u].params),i.length))for(var l in i)s.indexOf(i[l])>=0||(s.push(i[l]),a[i[l]]=this[i[l]]);return r.extend({},a,t)},t}();e.StateParams=i},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(22)),r(n(25)),r(n(45)),r(n(24))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(21)),r(n(20))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(18)),r(n(19)),r(n(17))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(30)),r(n(40)),r(n(42)),r(n(39)),r(n(41)),r(n(38)),r(n(43)),r(n(14))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(16)),r(n(15)),r(n(10)),r(n(11)),r(n(13)),r(n(31))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(28)),r(n(23)),r(n(27)),r(n(29))},function(t,e,n){"use strict";function r(t){for(var n in t)e.hasOwnProperty(n)||(e[n]=t[n])}r(n(37))},function(t,e,n){"use strict";function r(t){var e=f.services.$injector,n=e.get("$controller"),r=e.instantiate;try{var i;return e.instantiate=function(t){e.instantiate=r,i=e.annotate(t)},n(t,{$scope:{}}),i}finally{e.instantiate=r}}function i(t,e){f.services.$injector=t,f.services.$q=e}function o(t){function e(e,r,i,o,a,s){return o.$on("$locationChangeSuccess",function(t){return n.forEach(function(e){return e(t)})}),f.services.locationConfig.html5Mode=function(){var e=t.html5Mode();return e=d.isObject(e)?e.enabled:e,e&&i.history},f.services.template.get=function(t){return a.get(t,{cache:s,headers:{Accept:"text/html"}}).then(h.prop("data"))},p.bindFunctions(e,f.services.location,e,["replace","url","path","search","hash"]),p.bindFunctions(e,f.services.locationConfig,e,["port","protocol","host"]),p.bindFunctions(r,f.services.locationConfig,r,["baseHref"]),w}w=new c.UIRouter,w.stateRegistry.decorator("views",m.ng1ViewsBuilder),w.stateRegistry.decorator("onExit",b.getStateHookBuilder("onExit")),w.stateRegistry.decorator("onRetain",b.getStateHookBuilder("onRetain")),w.stateRegistry.decorator("onEnter",b.getStateHookBuilder("onEnter")),w.viewService.viewConfigFactory("ng1",m.ng1ViewConfigFactory),p.bindFunctions(t,f.services.locationConfig,t,["hashPrefix"]);var n=[];f.services.location.onChange=function(t){return n.push(t),function(){return p.removeFrom(n)(t)}},this.$get=e,e.$inject=["$location","$browser","$sniffer","$rootScope","$http","$templateCache"]}function a(){return w.urlRouterProvider.$get=function(){return w.urlRouter.update(!0),this.interceptDeferred||w.urlRouter.listen(),w.urlRouter},w.urlRouterProvider}function s(){return w.stateProvider.$get=function(){return w.stateRegistry.stateQueue.autoFlush(w.stateService),w.stateService},w.stateProvider}function u(){return w.transitionService.$get=function(){return w.transitionService},w.transitionService}function l(t){t.$watch(function(){g.trace.approximateDigests++})}var c=n(26),f=n(6),p=n(3),h=n(5),d=n(4),v=n(54),g=n(12),m=n(55),y=n(56),b=n(57),$=angular.module("ui.router.angular1",[]);angular.module("ui.router.util",["ng","ui.router.init"]),angular.module("ui.router.router",["ui.router.util"]),angular.module("ui.router.state",["ui.router.router","ui.router.util","ui.router.angular1"]),angular.module("ui.router",["ui.router.init","ui.router.state","ui.router.angular1"]),angular.module("ui.router.compat",["ui.router"]),e.annotateController=r,i.$inject=["$injector","$q"],$.run(i);var w=null;o.$inject=["$locationProvider"],angular.module("ui.router.init",[]).provider("ng1UIRouter",o),angular.module("ui.router.init").run(["ng1UIRouter",function(t){}]),angular.module("ui.router.util").provider("$urlMatcherFactory",["ng1UIRouterProvider",function(){return w.urlMatcherFactory}]),angular.module("ui.router.util").run(["$urlMatcherFactory",function(t){}]),angular.module("ui.router.router").provider("$urlRouter",["ng1UIRouterProvider",a]),angular.module("ui.router.router").run(["$urlRouter",function(t){}]),angular.module("ui.router.state").provider("$state",["ng1UIRouterProvider",s]),angular.module("ui.router.state").run(["$state",function(t){}]),angular.module("ui.router.state").factory("$stateParams",["ng1UIRouter",function(t){return t.globals.params}]),angular.module("ui.router.state").provider("$transitions",["ng1UIRouterProvider",u]),angular.module("ui.router.util").factory("$templateFactory",["ng1UIRouter",function(){return new y.TemplateFactory}]),angular.module("ui.router").factory("$view",function(){return w.viewService}),angular.module("ui.router").factory("$resolve",v.resolveFactory),angular.module("ui.router").service("$trace",function(){return g.trace}),l.$inject=["$rootScope"],e.watchDigests=l,angular.module("ui.router").run(l),e.getLocals=function(t){var e=t.getTokens().filter(d.isString),n=e.map(function(e){return[e,t.getResolvable(e).data]});return n.reduce(p.applyPairs,{})}},function(t,e,n){"use strict";var r=n(42),i=n(21),o=n(17),a=n(3),s=n(40),u={resolve:function(t,e,n){void 0===e&&(e={});var u=new i.PathNode(new r.State({params:{},resolvables:[]})),l=new i.PathNode(new r.State({params:{},resolvables:[]})),c=new o.ResolveContext([u,l]);c.addResolvables(s.resolvablesBuilder({resolve:t}),l.state);var f=function(t){var n=function(t){return s.resolvablesBuilder({resolve:a.map(t,function(t){return function(){return t}})})};c.addResolvables(n(t),u.state),c.addResolvables(n(e),l.state);var r=function(t,e){return t[e.token]=e.value,t};return c.resolvePath().then(function(t){return t.reduce(r,{})})};return n?n.then(f):f({})}};e.resolveFactory=function(){return u}},function(t,e,n){"use strict";function r(t){var e=["templateProvider","templateUrl","template","notify","async"],n=["controller","controllerProvider","controllerAs","resolveAs"],r=["component","bindings"],l=e.concat(n),c=r.concat(l),f={},p=t.views||{$default:o.pick(t,c)};return o.forEach(p,function(e,n){if(n=n||"$default",u.isString(e)&&(e={component:e}),Object.keys(e).length){if(e.component){if(l.map(function(t){return u.isDefined(e[t])}).reduce(o.anyTrueR,!1))throw new Error("Cannot combine: "+r.join("|")+" with: "+l.join("|")+" in stateview: 'name@"+t.name+"'");e.templateProvider=["$injector",function(t){var n=function(t){return e.bindings&&e.bindings[t]||t},r=angular.version.minor>=3?"::":"",o=function(t){var e=a.kebobString(t.name),i=n(t.name);return"@"===t.type?e+"='{{"+r+"$resolve."+i+"}}'":e+"='"+r+"$resolve."+i+"'"},s=i(t,e.component).map(o).join(" "),u=a.kebobString(e.component);return"<"+u+" "+s+"></"+u+">"}]}e.resolveAs=e.resolveAs||"$resolve",e.$type="ng1",e.$context=t,e.$name=n;var c=s.ViewService.normalizeUIViewTarget(e.$context,e.$name);e.$uiViewName=c.uiViewName,e.$uiViewContextAnchor=c.uiViewContextAnchor,f[n]=e}}),f}function i(t,e){var n=t.get(e+"Directive");if(!n||!n.length)throw new Error("Unable to find component named '"+e+"'");return n.map(v).reduce(o.unnestR,[])}var o=n(3),a=n(9),s=n(37),u=n(4),l=n(6),c=n(12),f=n(56),p=n(17),h=n(19);e.ng1ViewConfigFactory=function(t,e){return new m(t,e)},e.ng1ViewsBuilder=r;var d=function(t){return Object.keys(t||{}).map(function(e){return[e,/^([=<@])[?]?(.*)/.exec(t[e])]}).filter(function(t){return u.isDefined(t)&&u.isDefined(t[1])}).map(function(t){return{name:t[1][2]||t[0],type:t[1][1]}})},v=function(t){return d(u.isObject(t.bindToController)?t.bindToController:t.scope)},g=0,m=function(){function t(t,e){this.path=t,this.viewDecl=e,this.$id=g++,this.loaded=!1}return t.prototype.load=function(){var t=this,e=l.services.$q;if(!this.hasTemplate())throw new Error("No template configuration specified for '"+this.viewDecl.$uiViewName+"@"+this.viewDecl.$uiViewContextAnchor+"'");var n=new p.ResolveContext(this.path),r=this.path.reduce(function(t,e){return o.extend(t,e.paramValues)},{}),i={template:e.when(this.getTemplate(r,new f.TemplateFactory,n)),controller:e.when(this.getController(n))};return e.all(i).then(function(e){c.trace.traceViewServiceEvent("Loaded",t),t.controller=e.controller,t.template=e.template})},t.prototype.hasTemplate=function(){return!!(this.viewDecl.template||this.viewDecl.templateUrl||this.viewDecl.templateProvider)},t.prototype.getTemplate=function(t,e,n){return e.fromConfig(this.viewDecl,t,n)},t.prototype.getController=function(t){var e=this.viewDecl.controllerProvider;if(!u.isInjectable(e))return this.viewDecl.controller;var n=l.services.$injector.annotate(e),r=u.isArray(e)?o.tail(e):e,i=new h.Resolvable("",r,n);return i.get(t)},t}();e.Ng1ViewConfig=m},function(t,e,n){"use strict";var r=n(4),i=n(6),o=n(3),a=n(19),s=function(){function t(){}return t.prototype.fromConfig=function(t,e,n){return r.isDefined(t.template)?this.fromString(t.template,e):r.isDefined(t.templateUrl)?this.fromUrl(t.templateUrl,e):r.isDefined(t.templateProvider)?this.fromProvider(t.templateProvider,e,n):null},t.prototype.fromString=function(t,e){return r.isFunction(t)?t(e):t},t.prototype.fromUrl=function(t,e){return r.isFunction(t)&&(t=t(e)),null==t?null:i.services.template.get(t)},t.prototype.fromProvider=function(t,e,n){var s=i.services.$injector.annotate(t),u=r.isArray(t)?o.tail(t):t,l=new a.Resolvable("",u,s);return l.get(n)},t}();e.TemplateFactory=s},function(t,e,n){"use strict";var r=n(6),i=n(53),o=n(17),a=n(3);e.getStateHookBuilder=function(t){return function(e,n){function s(t,e){var n=new o.ResolveContext(t.treeChanges().to);return r.services.$injector.invoke(u,this,a.extend({$state$:e},i.getLocals(n)))}var u=e[t];return u?s:void 0}}},function(t,e,n){"use strict";function r(t,e){var n,r=t.match(/^\s*({[^}]*})\s*$/);if(r&&(t=e+"("+r[1]+")"),n=t.replace(/\n/g," ").match(/^([^(]+?)\s*(\((.*)\))?$/),!n||4!==n.length)throw new Error("Invalid state ref '"+t+"'");return{state:n[1],paramExpr:n[3]||null}}function i(t){var e=t.parent().inheritedData("$uiView"),n=c.parse("$cfg.path")(e);return n?u.tail(n).state.name:void 0}function o(t){var e="[object SVGAnimatedString]"===Object.prototype.toString.call(t.prop("href")),n="FORM"===t[0].nodeName;return{attr:n?"action":e?"xlink:href":"href",isAnchor:"A"===t.prop("tagName").toUpperCase(),clickable:!n}}function a(t,e,n,r,i){return function(o){var a=o.which||o.button,s=i();if(!(a>1||o.ctrlKey||o.metaKey||o.shiftKey||t.attr("target"))){var u=n(function(){e.go(s.state,s.params,s.options)});o.preventDefault();var l=r.isAnchor&&!s.href?1:0;o.preventDefault=function(){l--<=0&&n.cancel(u)}}}}function s(t,e){return{relative:i(t)||e.$current,inherit:!0}}var u=n(3),l=n(4),c=n(5),f=["$state","$timeout",function(t,e){return{restrict:"A",require:["?^uiSrefActive","?^uiSrefActiveEq"],link:function(n,i,l,c){var f=r(l.uiSref,t.current.name),p={state:f.state,href:null,params:null,options:null},h=o(i),d=c[1]||c[0],v=null;p.options=u.extend(s(i,t),l.uiSrefOpts?n.$eval(l.uiSrefOpts):{});var g=function(e){e&&(p.params=angular.copy(e)),p.href=t.href(f.state,p.params,p.options),v&&v(),d&&(v=d.$$addStateInfo(f.state,p.params)),null!==p.href&&l.$set(h.attr,p.href)};f.paramExpr&&(n.$watch(f.paramExpr,function(t){t!==p.params&&g(t)},!0),p.params=angular.copy(n.$eval(f.paramExpr))),g(),h.clickable&&i.bind("click",a(i,t,e,h,function(){return p}))}}}],p=["$state","$timeout",function(t,e){return{restrict:"A",require:["?^uiSrefActive","?^uiSrefActiveEq"],link:function(n,r,i,s){function u(e){h.state=e[0],h.params=e[1],h.options=e[2],h.href=t.href(h.state,h.params,h.options),d&&d(),c&&(d=c.$$addStateInfo(h.state,h.params)),h.href&&i.$set(l.attr,h.href)}var l=o(r),c=s[1]||s[0],f=[i.uiState,i.uiStateParams||null,i.uiStateOpts||null],p="["+f.map(function(t){return t||"null"}).join(", ")+"]",h={state:null,params:null,options:null,href:null},d=null;n.$watch(p,u,!0),u(n.$eval(p)),l.clickable&&r.bind("click",a(r,t,e,l,function(){return h}))}}}],h=["$state","$stateParams","$interpolate","$transitions",function(t,e,n,o){return{restrict:"A",controller:["$scope","$element","$attrs","$timeout",function(e,a,s,c){function f(e,n,r){var o=t.get(e,i(a)),s=p(e,n),u={state:o||{name:e},params:n,hash:s};return $.push(u),w[s]=r,function(){var t=$.indexOf(u);t!==-1&&$.splice(t,1)}}function p(t,n){if(!l.isString(t))throw new Error("state should be a string");return l.isObject(n)?t+u.toJson(n):(n=e.$eval(n),l.isObject(n)?t+u.toJson(n):t)}function h(){for(var t=0;t<$.length;t++)g($[t].state,$[t].params)?d(a,w[$[t].hash]):v(a,w[$[t].hash]),m($[t].state,$[t].params)?d(a,y):v(a,y)}function d(t,e){c(function(){t.addClass(e)})}function v(t,e){t.removeClass(e)}function g(e,n){return t.includes(e.name,n)}function m(e,n){return t.is(e.name,n)}var y,b,$=[],w={};y=n(s.uiSrefActiveEq||"",!1)(e);try{b=e.$eval(s.uiSrefActive)}catch(t){}b=b||n(s.uiSrefActive||"",!1)(e),l.isObject(b)&&u.forEach(b,function(n,i){if(l.isString(n)){var o=r(n,t.current.name);f(o.state,e.$eval(o.paramExpr),i)}}),this.$$addStateInfo=function(t,e){if(!(l.isObject(b)&&$.length>0)){var n=f(t,e,b);return h(),n}},e.$on("$stateChangeSuccess",h),e.$on("$destroy",o.onStart({},function(t){return t.promise.then(h)&&null})),h()}]}}];angular.module("ui.router.state").directive("uiSref",f).directive("uiSrefActive",h).directive("uiSrefActiveEq",h).directive("uiState",p)},function(t,e){"use strict";function n(t){var e=function(e,n,r){return t.is(e,n,r)};return e.$stateful=!0,e}function r(t){var e=function(e,n,r){return t.includes(e,n,r)};return e.$stateful=!0,e}n.$inject=["$state"],e.$IsStateFilter=n,r.$inject=["$state"],e.$IncludedByStateFilter=r,angular.module("ui.router.state").filter("isState",n).filter("includedByState",r)},function(t,e,n){"use strict";function r(t,e,n,r,u){var h=l.parse("viewDecl.controllerAs"),d=l.parse("viewDecl.resolveAs");return{restrict:"ECA",priority:-400,compile:function(r){var u=r.html();return function(r,l){var v=l.data("$uiView");if(v){var g=v.$cfg||{viewDecl:{}};l.html(g.template||u),s.trace.traceUIViewFill(v.$uiView,l.html());var m=t(l.contents()),y=g.controller,b=h(g),$=d(g),w=g.path&&new c.ResolveContext(g.path),x=w&&p.getLocals(w);if(r[$]=x,y){var _=e(y,o.extend({},x,{$scope:r,$element:l}));b&&(r[b]=_,r[b][$]=x),l.data("$ngControllerController",_),l.children().data("$ngControllerController",_),i(n,_,r,g)}if(a.isString(g.viewDecl.component))var C=g.viewDecl.component,k=f.kebobString(C),S=function(){var t=[].slice.call(l[0].children).filter(function(t){return t&&t.tagName&&t.tagName.toLowerCase()===k});return t&&angular.element(t).data("$"+C+"Controller")},T=r.$watch(S,function(t){t&&(i(n,t,r,g),T())});m(r)}}}}}function i(t,e,n,r){!a.isFunction(e.$onInit)||r.viewDecl.component&&d||e.$onInit();var i=o.tail(r.path).state.self,s={bind:e};if(a.isFunction(e.uiOnParamsChanged)){var u=new c.ResolveContext(r.path),l=u.getResolvable("$transition$").data,f=function(t){if(t!==l&&t.exiting().indexOf(i)===-1){var n=t.params("to"),r=t.params("from"),a=t.treeChanges().to.map(function(t){return t.paramSchema}).reduce(o.unnestR,[]),s=t.treeChanges().from.map(function(t){return t.paramSchema}).reduce(o.unnestR,[]),u=a.filter(function(t){var e=s.indexOf(t);return e===-1||!s[e].type.equals(n[t.id],r[t.id])});if(u.length){var c=u.map(function(t){return t.id});e.uiOnParamsChanged(o.filter(n,function(t,e){return c.indexOf(e)!==-1}),t)}}};n.$on("$destroy",t.onSuccess({},f,s))}if(a.isFunction(e.uiCanExit)){var p={exiting:i.name};n.$on("$destroy",t.onBefore(p,e.uiCanExit,s))}}var o=n(3),a=n(4),s=n(12),u=n(55),l=n(5),c=n(17),f=n(9),p=n(53),h=["$view","$animate","$uiViewScroll","$interpolate","$q",function(t,e,n,r,i){function o(t,n){return{enter:function(t,n,r){angular.version.minor>2?e.enter(t,null,n).then(r):e.enter(t,null,n,r)},leave:function(t,n){angular.version.minor>2?e.leave(t).then(n):e.leave(t,n)}}}function c(t,e){return t===e}var f={$cfg:{viewDecl:{$context:t.rootContext()}},$uiView:{}},p={count:0,restrict:"ECA",terminal:!0,priority:400,transclude:"element",compile:function(e,h,d){return function(e,h,v){function g(t){(!t||t instanceof u.Ng1ViewConfig)&&(c(S,t)||(s.trace.traceUIViewConfigUpdated(M,t&&t.viewDecl&&t.viewDecl.$context),S=t,y(t)))}function m(){if(b&&(s.trace.traceUIViewEvent("Removing (previous) el",b.data("$uiView")),b.remove(),b=null),w&&(s.trace.traceUIViewEvent("Destroying scope",M),w.$destroy(),w=null),$){var t=$.data("$uiView");s.trace.traceUIViewEvent("Animate out",t),k.leave($,function(){t.$$animLeave.resolve(),b=null}),b=$,$=null}}function y(t){var r=e.$new();s.trace.traceUIViewScopeCreated(M,r);var o=i.defer(),u=i.defer(),l={$cfg:t,$uiView:M,$animEnter:o.promise,$animLeave:u.promise,$$animLeave:u},c=d(r,function(t){k.enter(t.data("$uiView",l),h,function(){o.resolve(),w&&w.$emit("$viewContentAnimationEnded"),(a.isDefined(C)&&!C||e.$eval(C))&&n(t)}),m()});$=c,w=r,w.$emit("$viewContentLoaded",t||S),w.$eval(_)}var b,$,w,x,_=v.onload||"",C=v.autoscroll,k=o(v,e),S=void 0,T=h.inheritedData("$uiView")||f,E=r(v.uiView||v.name||"")(e)||"$default",M={$type:"ng1",id:p.count++,name:E,fqn:T.$uiView.fqn?T.$uiView.fqn+"."+E:E,config:null,configUpdated:g,get creationContext(){return l.parse("$cfg.viewDecl.$context")(T)}};s.trace.traceUIViewEvent("Linking",M),h.data("$uiView",{$uiView:M}),y(),x=t.registerUIView(M),e.$on("$destroy",function(){s.trace.traceUIViewEvent("Destroying/Unregistering",M),x()})}}};return p}];r.$inject=["$compile","$controller","$transitions","$view","$timeout"];var d="function"==typeof angular.module("ui.router").component;angular.module("ui.router.state").directive("uiView",h),angular.module("ui.router.state").directive("uiView",r)},function(t,e){"use strict";function n(){var t=!1;this.useAnchorScroll=function(){t=!0},this.$get=["$anchorScroll","$timeout",function(e,n){return t?e:function(t){return n(function(){t[0].scrollIntoView()},0,!1)}}]}angular.module("ui.router.state").provider("$uiViewScroll",n)}])})},22:function(t,e){/**
 * @license AngularJS v1.6.1
 * (c) 2010-2016 Google, Inc. http://angularjs.org
 * License: MIT
 */
!function(t,e){"use strict";function n(t,e,n){if(!t)throw vt("areq","Argument '{0}' is {1}",e||"?",n||"required");return t}function r(t,e){return t||e?t?e?(X(t)&&(t=t.join(" ")),X(e)&&(e=e.join(" ")),t+" "+e):t:e:""}function i(t){var e={};return t&&(t.to||t.from)&&(e.to=t.to,e.from=t.from),e}function o(t,e,n){var r="";return t=X(t)?t:t&&tt(t)&&t.length?t.split(/\s+/):[],G(t,function(t,i){t&&t.length>0&&(r+=i>0?" ":"",r+=n?e+t:t+e)}),r}function a(t,e){var n=t.indexOf(e);e>=0&&t.splice(n,1)}function s(t){if(t instanceof nt)switch(t.length){case 0:return t;case 1:if(t[0].nodeType===j)return t;break;default:return nt(u(t))}if(t.nodeType===j)return nt(t)}function u(t){if(!t[0])return t;for(var e=0;e<t.length;e++){var n=t[e];if(n.nodeType===j)return n}}function l(t,e,n){G(e,function(e){t.addClass(e,n)})}function c(t,e,n){G(e,function(e){t.removeClass(e,n)})}function f(t){return function(e,n){n.addClass&&(l(t,e,n.addClass),n.addClass=null),n.removeClass&&(c(t,e,n.removeClass),n.removeClass=null)}}function p(t){if(t=t||{},!t.$$prepared){var e=t.domOperation||rt;t.domOperation=function(){t.$$domOperationFired=!0,e(),e=rt},t.$$prepared=!0}return t}function h(t,e){d(t,e),v(t,e)}function d(t,e){e.from&&(t.css(e.from),e.from=null)}function v(t,e){e.to&&(t.css(e.to),e.to=null)}function g(t,e,n){var r=e.options||{},i=n.options||{},o=(r.addClass||"")+" "+(i.addClass||""),a=(r.removeClass||"")+" "+(i.removeClass||""),s=m(t.attr("class"),o,a);i.preparationClasses&&(r.preparationClasses=C(i.preparationClasses,r.preparationClasses),delete i.preparationClasses);var u=r.domOperation!==rt?r.domOperation:null;return Y(r,i),u&&(r.domOperation=u),s.addClass?r.addClass=s.addClass:r.addClass=null,s.removeClass?r.removeClass=s.removeClass:r.removeClass=null,e.addClass=r.addClass,e.removeClass=r.removeClass,r}function m(t,e,n){function r(t){tt(t)&&(t=t.split(" "));var e={};return G(t,function(t){t.length&&(e[t]=!0)}),e}var i=1,o=-1,a={};t=r(t),e=r(e),G(e,function(t,e){a[e]=i}),n=r(n),G(n,function(t,e){a[e]=a[e]===i?null:o});var s={addClass:"",removeClass:""};return G(a,function(e,n){var r,a;e===i?(r="addClass",a=!t[n]||t[n+q]):e===o&&(r="removeClass",a=t[n]||t[n+F]),a&&(s[r].length&&(s[r]+=" "),s[r]+=n)}),s}function y(t){return t instanceof nt?t[0]:t}function b(t,e,n){var r="";e&&(r=o(e,V,!0)),n.addClass&&(r=C(r,o(n.addClass,F))),n.removeClass&&(r=C(r,o(n.removeClass,q))),r.length&&(n.preparationClasses=r,t.addClass(r))}function $(t,e){e.preparationClasses&&(t.removeClass(e.preparationClasses),e.preparationClasses=null),e.activeClasses&&(t.removeClass(e.activeClasses),e.activeClasses=null)}function w(t,e){var n=e?"-"+e+"s":"";return _(t,[ht,n]),[ht,n]}function x(t,e){var n=e?"paused":"",r=R+lt;return _(t,[r,n]),[r,n]}function _(t,e){var n=e[0],r=e[1];t.style[n]=r}function C(t,e){return t?e?t+" "+e:t:e}function k(t){return[pt,t+"s"]}function S(t,e){var n=e?ft:ht;return[n,t+"s"]}function T(t,e,n){var r=Object.create(null),i=t.getComputedStyle(e)||{};return G(n,function(t,e){var n=i[t];if(n){var o=n.charAt(0);("-"===o||"+"===o||o>=0)&&(n=E(n)),0===n&&(n=null),r[e]=n}}),r}function E(t){var e=0,n=t.split(/\s*,\s*/);return G(n,function(t){"s"===t.charAt(t.length-1)&&(t=t.substring(0,t.length-1)),t=parseFloat(t)||0,e=e?Math.max(t,e):t}),e}function M(t){return 0===t||null!=t}function D(t,e){var n=N,r=t+"s";return e?n+=it:r+=" linear all",[n,r]}function A(){var t=Object.create(null);return{flush:function(){t=Object.create(null)},count:function(e){var n=t[e];return n?n.total:0},get:function(e){var n=t[e];return n&&n.value},put:function(e,n){t[e]?t[e].total++:t[e]={total:1,value:n}}}}function O(t,e,n){G(n,function(n){t[n]=K(t[n])?t[n]:e.style.getPropertyValue(n)})}var N,P,R,I,j=1,F="-add",q="-remove",V="ng-",U="-active",L="-prepare",H="ng-animate",B="$$ngAnimateChildren",z="";void 0===t.ontransitionend&&void 0!==t.onwebkittransitionend?(z="-webkit-",N="WebkitTransition",P="webkitTransitionEnd transitionend"):(N="transition",P="transitionend"),void 0===t.onanimationend&&void 0!==t.onwebkitanimationend?(z="-webkit-",R="WebkitAnimation",I="webkitAnimationEnd animationend"):(R="animation",I="animationend");var W,Y,G,X,K,J,Z,Q,tt,et,nt,rt,it="Duration",ot="Property",at="Delay",st="TimingFunction",ut="IterationCount",lt="PlayState",ct=9999,ft=R+at,pt=R+it,ht=N+at,dt=N+it,vt=e.$$minErr("ng"),gt=["$$rAF",function(t){function e(t){r=r.concat(t),n()}function n(){if(r.length){for(var e=r.shift(),o=0;o<e.length;o++)e[o]();i||t(function(){i||n()})}}var r,i;return r=e.queue=[],e.waitUntilQuiet=function(e){i&&i(),i=t(function(){i=null,e(),n()})},e}],mt=["$interpolate",function(t){return{link:function(e,n,r){function i(t){t="on"===t||"true"===t,n.data(B,t)}var o=r.ngAnimateChildren;tt(o)&&0===o.length?n.data(B,!0):(i(t(o)(e)),r.$observe("ngAnimateChildren",i))}}}],yt="$$animateCss",bt=1e3,$t=3,wt=1.5,xt={transitionDuration:dt,transitionDelay:ht,transitionProperty:N+ot,animationDuration:pt,animationDelay:ft,animationIterationCount:R+ut},_t={transitionDuration:dt,transitionDelay:ht,animationDuration:pt,animationDelay:ft},Ct=["$animateProvider",function(t){var e=A(),n=A();this.$get=["$window","$$jqLite","$$AnimateRunner","$timeout","$$forceReflow","$sniffer","$$rAFScheduler","$$animateQueue",function(t,r,s,u,l,c,g,m){function b(t,e){var n="$$ngAnimateParentKey",r=t.parentNode,i=r[n]||(r[n]=++L);return i+"-"+t.getAttribute("class")+"-"+e}function $(n,r,i,o){var a=e.get(i);return a||(a=T(t,n,o),"infinite"===a.animationIterationCount&&(a.animationIterationCount=1)),e.put(i,a),a}function C(i,a,s,u){var l;if(e.count(s)>0&&(l=n.get(s),!l)){var c=o(a,"-stagger");r.addClass(i,c),l=T(t,i,u),l.animationDuration=Math.max(l.animationDuration,0),l.transitionDuration=Math.max(l.transitionDuration,0),r.removeClass(i,c),n.put(s,l)}return l||{}}function E(t){H.push(t),g.waitUntilQuiet(function(){e.flush(),n.flush();for(var t=l(),r=0;r<H.length;r++)H[r](t);H.length=0})}function A(t,e,n){var r=$(t,e,n,xt),i=r.animationDelay,o=r.transitionDelay;return r.maxDelay=i&&o?Math.max(i,o):i||o,r.maxDuration=Math.max(r.animationDuration*r.animationIterationCount,r.transitionDuration),r}var j=f(r),L=0,H=[];return function(t,n){function l(){g()}function f(){g(!0)}function g(e){if(!(K||Z&&J)){K=!0,J=!1,B.$$skipPreparationClasses||r.removeClass(t,xt),r.removeClass(t,kt),x(Y,!1),w(Y,!1),G(lt,function(t){Y.style[t[0]]=""}),j(t,B),h(t,B),Object.keys(z).length&&G(z,function(t,e){t?Y.style.setProperty(e,t):Y.style.removeProperty(e)}),B.onDone&&B.onDone(),ht&&ht.length&&t.off(ht.join(" "),L);var n=t.data(yt);n&&(u.cancel(n[0].timer),t.removeData(yt)),Q&&Q.complete(!e)}}function $(t){qt.blockTransition&&w(Y,t),qt.blockKeyframeAnimation&&x(Y,!!t)}function T(){return Q=new s({end:l,cancel:f}),E(rt),g(),{$$willAnimate:!1,start:function(){return Q},end:l}}function L(t){t.stopPropagation();var e=t.originalEvent||t,n=e.$manualTimeStamp||Date.now(),r=parseFloat(e.elapsedTime.toFixed($t));Math.max(n-ut,0)>=nt&&r>=it&&(Z=!0,g())}function H(){function e(){if(!K){if($(!1),G(lt,function(t){var e=t[0],n=t[1];Y.style[e]=n}),j(t,B),r.addClass(t,kt),qt.recalculateTimingStyles){if(Ct=Y.getAttribute("class")+" "+xt,Et=b(Y,Ct),jt=A(Y,Ct,Et),Ft=jt.maxDelay,et=Math.max(Ft,0),it=jt.maxDuration,0===it)return void g();qt.hasTransitions=jt.transitionDuration>0,qt.hasAnimations=jt.animationDuration>0}if(qt.applyAnimationDelay&&(Ft="boolean"!=typeof B.delay&&M(B.delay)?parseFloat(B.delay):Ft,et=Math.max(Ft,0),jt.animationDelay=Ft,Vt=S(Ft,!0),lt.push(Vt),Y.style[Vt[0]]=Vt[1]),nt=et*bt,at=it*bt,B.easing){var e,i=B.easing;qt.hasTransitions&&(e=N+st,lt.push([e,i]),Y.style[e]=i),qt.hasAnimations&&(e=R+st,lt.push([e,i]),Y.style[e]=i)}jt.transitionDuration&&ht.push(P),jt.animationDuration&&ht.push(I),ut=Date.now();var o=nt+wt*at,a=ut+o,s=t.data(yt)||[],l=!0;if(s.length){var c=s[0];l=a>c.expectedEndTime,l?u.cancel(c.timer):s.push(g)}if(l){var f=u(n,o,!1);s[0]={timer:f,expectedEndTime:a},s.push(g),t.data(yt,s)}ht.length&&t.on(ht.join(" "),L),B.to&&(B.cleanupStyles&&O(z,Y,Object.keys(B.to)),v(t,B))}}function n(){var e=t.data(yt);if(e){for(var n=1;n<e.length;n++)e[n]();t.removeData(yt)}}if(!K){if(!Y.parentNode)return void g();var i=function(t){if(Z)J&&t&&(J=!1,g());else if(J=!t,jt.animationDuration){var e=x(Y,J);J?lt.push(e):a(lt,e)}},o=Rt>0&&(jt.transitionDuration&&0===Mt.transitionDuration||jt.animationDuration&&0===Mt.animationDuration)&&Math.max(Mt.animationDelay,Mt.transitionDelay);o?u(e,Math.floor(o*Rt*bt),!1):e(),tt.resume=function(){i(!0)},tt.pause=function(){i(!1)}}}var B=n||{};B.$$prepared||(B=p(W(B)));var z={},Y=y(t);if(!Y||!Y.parentNode||!m.enabled())return T();var K,J,Z,Q,tt,et,nt,it,at,ut,lt=[],ft=t.attr("class"),pt=i(B),ht=[];if(0===B.duration||!c.animations&&!c.transitions)return T();var dt=B.event&&X(B.event)?B.event.join(" "):B.event,vt=dt&&B.structural,gt="",mt="";vt?gt=o(dt,V,!0):dt&&(gt=dt),B.addClass&&(mt+=o(B.addClass,F)),B.removeClass&&(mt.length&&(mt+=" "),mt+=o(B.removeClass,q)),B.applyClassesEarly&&mt.length&&j(t,B);var xt=[gt,mt].join(" ").trim(),Ct=ft+" "+xt,kt=o(xt,U),St=pt.to&&Object.keys(pt.to).length>0,Tt=(B.keyframeStyle||"").length>0;if(!Tt&&!St&&!xt)return T();var Et,Mt;if(B.stagger>0){var Dt=parseFloat(B.stagger);Mt={transitionDelay:Dt,animationDelay:Dt,transitionDuration:0,animationDuration:0}}else Et=b(Y,Ct),Mt=C(Y,xt,Et,_t);B.$$skipPreparationClasses||r.addClass(t,xt);var At;if(B.transitionStyle){var Ot=[N,B.transitionStyle];_(Y,Ot),lt.push(Ot)}if(B.duration>=0){At=Y.style[N].length>0;var Nt=D(B.duration,At);_(Y,Nt),lt.push(Nt)}if(B.keyframeStyle){var Pt=[R,B.keyframeStyle];_(Y,Pt),lt.push(Pt)}var Rt=Mt?B.staggerIndex>=0?B.staggerIndex:e.count(Et):0,It=0===Rt;It&&!B.skipBlocking&&w(Y,ct);var jt=A(Y,Ct,Et),Ft=jt.maxDelay;et=Math.max(Ft,0),it=jt.maxDuration;var qt={};if(qt.hasTransitions=jt.transitionDuration>0,qt.hasAnimations=jt.animationDuration>0,qt.hasTransitionAll=qt.hasTransitions&&"all"===jt.transitionProperty,qt.applyTransitionDuration=St&&(qt.hasTransitions&&!qt.hasTransitionAll||qt.hasAnimations&&!qt.hasTransitions),qt.applyAnimationDuration=B.duration&&qt.hasAnimations,qt.applyTransitionDelay=M(B.delay)&&(qt.applyTransitionDuration||qt.hasTransitions),qt.applyAnimationDelay=M(B.delay)&&qt.hasAnimations,qt.recalculateTimingStyles=mt.length>0,(qt.applyTransitionDuration||qt.applyAnimationDuration)&&(it=B.duration?parseFloat(B.duration):it,qt.applyTransitionDuration&&(qt.hasTransitions=!0,jt.transitionDuration=it,At=Y.style[N+ot].length>0,lt.push(D(it,At))),qt.applyAnimationDuration&&(qt.hasAnimations=!0,jt.animationDuration=it,lt.push(k(it)))),0===it&&!qt.recalculateTimingStyles)return T();if(null!=B.delay){var Vt;"boolean"!=typeof B.delay&&(Vt=parseFloat(B.delay),et=Math.max(Vt,0)),qt.applyTransitionDelay&&lt.push(S(Vt)),qt.applyAnimationDelay&&lt.push(S(Vt,!0))}return null==B.duration&&jt.transitionDuration>0&&(qt.recalculateTimingStyles=qt.recalculateTimingStyles||It),nt=et*bt,at=it*bt,B.skipBlocking||(qt.blockTransition=jt.transitionDuration>0,qt.blockKeyframeAnimation=jt.animationDuration>0&&Mt.animationDelay>0&&0===Mt.animationDuration),B.from&&(B.cleanupStyles&&O(z,Y,Object.keys(B.from)),d(t,B)),qt.blockTransition||qt.blockKeyframeAnimation?$(it):B.skipBlocking||w(Y,!1),{$$willAnimate:!0,end:l,start:function(){if(!K)return tt={end:l,cancel:f,resume:null,pause:null},Q=new s(tt),E(H),Q}}}}]}],kt=["$$animationProvider",function(t){function e(t){return t.parentNode&&11===t.parentNode.nodeType}t.drivers.push("$$animateCssDriver");var n="ng-animate-shim",r="ng-anchor",i="ng-anchor-out",o="ng-anchor-in";this.$get=["$animateCss","$rootScope","$$AnimateRunner","$rootElement","$sniffer","$$jqLite","$document",function(t,a,s,u,l,c,f){function p(t){return t.replace(/\bng-\S+\b/g,"")}function h(t,e){return tt(t)&&(t=t.split(" ")),tt(e)&&(e=e.split(" ")),t.filter(function(t){return e.indexOf(t)===-1}).join(" ")}function d(e,a,u){function l(t){var e={},n=y(t).getBoundingClientRect();return G(["width","height","top","left"],function(t){var r=n[t];switch(t){case"top":r+=m.scrollTop;break;case"left":r+=m.scrollLeft}e[t]=Math.floor(r)+"px"}),e}function c(){var e=t(g,{addClass:i,delay:!0,from:l(a)});return e.$$willAnimate?e:null}function f(t){return t.attr("class")||""}function d(){var e=p(f(u)),n=h(e,b),r=h(b,e),a=t(g,{to:l(u),addClass:o+" "+n,removeClass:i+" "+r,delay:!0});return a.$$willAnimate?a:null}function v(){g.remove(),a.removeClass(n),u.removeClass(n)}var g=nt(y(a).cloneNode(!0)),b=p(f(g));a.addClass(n),u.addClass(n),g.addClass(r),$.append(g);var w,x=c();if(!x&&(w=d(),!w))return v();var _=x||w;return{start:function(){function t(){n&&n.end()}var e,n=_.start();return n.done(function(){return n=null,!w&&(w=d())?(n=w.start(),n.done(function(){n=null,v(),e.complete()}),n):(v(),void e.complete())}),e=new s({end:t,cancel:t})}}}function v(t,e,n,r){var i=g(t,rt),o=g(e,rt),a=[];if(G(r,function(t){var e=t.out,r=t.in,i=d(n,e,r);i&&a.push(i)}),i||o||0!==a.length)return{start:function(){function t(){G(e,function(t){t.end()})}var e=[];i&&e.push(i.start()),o&&e.push(o.start()),G(a,function(t){e.push(t.start())});var n=new s({end:t,cancel:t});return s.all(e,function(t){n.complete(t)}),n}}}function g(e){var n=e.element,r=e.options||{};e.structural&&(r.event=e.event,r.structural=!0,r.applyClassesEarly=!0,"leave"===e.event&&(r.onDone=r.domOperation)),r.preparationClasses&&(r.event=C(r.event,r.preparationClasses));var i=t(n,r);return i.$$willAnimate?i:null}if(!l.animations&&!l.transitions)return rt;var m=f[0].body,b=y(u),$=nt(e(b)||m.contains(b)?b:m);return function(t){return t.from&&t.to?v(t.from,t.to,t.classes,t.anchors):g(t)}}]}],St=["$animateProvider",function(t){this.$get=["$injector","$$AnimateRunner","$$jqLite",function(e,n,r){function i(n){n=X(n)?n:n.split(" ");for(var r=[],i={},o=0;o<n.length;o++){var a=n[o],s=t.$$registeredAnimations[a];s&&!i[a]&&(r.push(e.get(s)),i[a]=!0)}return r}var o=f(r);return function(t,e,r,a){function s(){a.domOperation(),o(t,a)}function u(){d=!0,s(),h(t,a)}function l(t,e,r,i,o){var a;switch(r){case"animate":a=[e,i.from,i.to,o];break;case"setClass":a=[e,m,y,o];break;case"addClass":a=[e,m,o];break;case"removeClass":a=[e,y,o];break;default:a=[e,o]}a.push(i);var s=t.apply(t,a);if(s)if(Z(s.start)&&(s=s.start()),s instanceof n)s.done(o);else if(Z(s))return s;return rt}function c(t,e,r,i,o){var a=[];return G(i,function(i){var s=i[o];s&&a.push(function(){var i,o,a=!1,u=function(t){a||(a=!0,(o||rt)(t),i.complete(!t))};return i=new n({end:function(){u()},cancel:function(){u(!0)}}),o=l(s,t,e,r,function(t){var e=t===!1;u(e)}),i})}),a}function f(t,e,r,i,o){var a=c(t,e,r,i,o);if(0===a.length){var s,u;"beforeSetClass"===o?(s=c(t,"removeClass",r,i,"beforeRemoveClass"),u=c(t,"addClass",r,i,"beforeAddClass")):"setClass"===o&&(s=c(t,"removeClass",r,i,"removeClass"),u=c(t,"addClass",r,i,"addClass")),s&&(a=a.concat(s)),u&&(a=a.concat(u))}if(0!==a.length)return function(t){var e=[];return a.length&&G(a,function(t){e.push(t())}),e.length?n.all(e,t):t(),function(t){G(e,function(e){t?e.cancel():e.end()})}}}var d=!1;3===arguments.length&&Q(r)&&(a=r,r=null),a=p(a),r||(r=t.attr("class")||"",a.addClass&&(r+=" "+a.addClass),a.removeClass&&(r+=" "+a.removeClass));var v,g,m=a.addClass,y=a.removeClass,b=i(r);if(b.length){var $,w;"leave"===e?(w="leave",$="afterLeave"):(w="before"+e.charAt(0).toUpperCase()+e.substr(1),$=e),"enter"!==e&&"move"!==e&&(v=f(t,e,a,b,w)),g=f(t,e,a,b,$)}if(v||g){var x;return{$$willAnimate:!0,end:function(){return x?x.end():(u(),x=new n,x.complete(!0)),x},start:function(){function t(t){u(t),x.complete(t)}function e(e){d||((r||rt)(e),t(e))}if(x)return x;x=new n;var r,i=[];return v&&i.push(function(t){r=v(t)}),i.length?i.push(function(t){s(),t(!0)}):s(),g&&i.push(function(t){r=g(t)}),x.setHost({end:function(){e()},cancel:function(){e(!0)}}),n.chain(i,t),x}}}}}]}],Tt=["$$animationProvider",function(t){t.drivers.push("$$animateJsDriver"),this.$get=["$$animateJs","$$AnimateRunner",function(t,e){function n(e){var n=e.element,r=e.event,i=e.options,o=e.classes;return t(n,r,o,i)}return function(t){if(t.from&&t.to){var r=n(t.from),i=n(t.to);if(!r&&!i)return;return{start:function(){function t(){return function(){G(o,function(t){t.end()})}}function n(t){a.complete(t)}var o=[];r&&o.push(r.start()),i&&o.push(i.start()),e.all(o,n);var a=new e({end:t(),cancel:t()});return a}}}return n(t)}}]}],Et="data-ng-animate",Mt="$ngAnimatePin",Dt=["$animateProvider",function(e){function r(t){if(!t)return null;var e=t.split(d),n=Object.create(null);return G(e,function(t){n[t]=!0}),n}function i(t,e){if(t&&e){var n=r(e);return t.split(d).some(function(t){return n[t]})}}function o(t,e,n,r){return v[t].some(function(t){return t(e,n,r)})}function a(t,e){var n=(t.addClass||"").length>0,r=(t.removeClass||"").length>0;return e?n&&r:n||r}var l=1,c=2,d=" ",v=this.rules={skip:[],cancel:[],join:[]};v.join.push(function(t,e,n){return!e.structural&&a(e)}),v.skip.push(function(t,e,n){return!e.structural&&!a(e)}),v.skip.push(function(t,e,n){return"leave"===n.event&&e.structural}),v.skip.push(function(t,e,n){return n.structural&&n.state===c&&!e.structural}),v.cancel.push(function(t,e,n){return n.structural&&e.structural}),v.cancel.push(function(t,e,n){return n.state===c&&e.structural}),v.cancel.push(function(t,e,n){if(n.structural)return!1;var r=e.addClass,o=e.removeClass,a=n.addClass,s=n.removeClass;return!(et(r)&&et(o)||et(a)&&et(s))&&(i(r,s)||i(o,a))}),this.$get=["$$rAF","$rootScope","$rootElement","$document","$$HashMap","$$animation","$$AnimateRunner","$templateRequest","$$jqLite","$$forceReflow","$$isDocumentHidden",function(r,i,d,v,m,w,x,_,C,k,S){function T(){var t=!1;return function(e){t?e():i.$$postDigest(function(){t=!0,e()})}}function E(t,e){return g(t,e,{})}function M(t,e,n){var r=y(e),i=y(t),o=[],a=H[n];return a&&G(a,function(t){it.call(t.node,r)?o.push(t.callback):"leave"===n&&it.call(t.node,i)&&o.push(t.callback)}),o}function D(t,e,n){var r=u(e);return t.filter(function(t){var e=t.node===r&&(!n||t.callback===n);return!e})}function A(t,e){"close"!==t||e[0].parentNode||ot.off(e)}function O(t,e,n){function u(e,n,i,o){C(function(){var e=M(v,t,n);e.length?r(function(){G(e,function(e){e(t,i,o)}),A(i,t)}):A(i,t)}),e.progress(n,i,o)}function f(e){$(t,m),rt(t,m),h(t,m),m.domOperation(),_.complete(!e)}var d,v,m=W(n);t=s(t),t&&(d=y(t),v=t.parent()),m=p(m);var _=new x,C=T();if(X(m.addClass)&&(m.addClass=m.addClass.join(" ")),m.addClass&&!tt(m.addClass)&&(m.addClass=null),X(m.removeClass)&&(m.removeClass=m.removeClass.join(" ")),m.removeClass&&!tt(m.removeClass)&&(m.removeClass=null),m.from&&!Q(m.from)&&(m.from=null),m.to&&!Q(m.to)&&(m.to=null),!d)return f(),_;var k=[d.getAttribute("class"),m.addClass,m.removeClass].join(" ");if(!Z(k))return f(),_;var D=["enter","move","leave"].indexOf(e)>=0,O=S(),R=!U||O||V.get(d),j=!R&&q.get(d)||{},L=!!j.state;if(R||L&&j.state===l||(R=!I(t,v,e)),R)return O&&u(_,e,"start"),f(),O&&u(_,e,"close"),_;D&&N(t);var H={structural:D,element:t,event:e,addClass:m.addClass,removeClass:m.removeClass,close:f,options:m,runner:_};if(L){var B=o("skip",t,H,j);if(B)return j.state===c?(f(),_):(g(t,j,H),j.runner);var z=o("cancel",t,H,j);if(z)if(j.state===c)j.runner.end();else{if(!j.structural)return g(t,j,H),j.runner;j.close()}else{var Y=o("join",t,H,j);if(Y){if(j.state!==c)return b(t,D?e:null,m),e=H.event=j.event,m=g(t,j,H),j.runner;E(t,H)}}}else E(t,H);var K=H.structural;if(K||(K="animate"===H.event&&Object.keys(H.options.to||{}).length>0||a(H)),!K)return f(),P(t),_;var J=(j.counter||0)+1;return H.counter=J,F(t,l,H),i.$$postDigest(function(){var n=q.get(d),r=!n;n=n||{};var i=t.parent()||[],o=i.length>0&&("animate"===n.event||n.structural||a(n));if(r||n.counter!==J||!o)return r&&(rt(t,m),h(t,m)),(r||D&&n.event!==e)&&(m.domOperation(),_.end()),void(o||P(t));e=!n.structural&&a(n,!0)?"setClass":n.event,F(t,c);var s=w(t,e,n.options);_.setHost(s),u(_,e,"start",{}),s.done(function(n){f(!n);var r=q.get(d);r&&r.counter===J&&P(y(t)),u(_,e,"close",{})})}),_}function N(t){var e=y(t),n=e.querySelectorAll("["+Et+"]");G(n,function(t){var e=parseInt(t.getAttribute(Et),10),n=q.get(t);if(n)switch(e){case c:n.runner.end();case l:q.remove(t)}})}function P(t){var e=y(t);e.removeAttribute(Et),q.remove(e)}function R(t,e){return y(t)===y(e)}function I(t,e,n){var r,i=nt(v[0].body),o=R(t,i)||"HTML"===t[0].nodeName,a=R(t,d),s=!1,u=V.get(y(t)),l=nt.data(t[0],Mt);for(l&&(e=l),e=y(e);e&&(a||(a=R(e,d)),e.nodeType===j);){var c=q.get(e)||{};if(!s){var f=V.get(e);if(f===!0&&u!==!1){u=!0;break}f===!1&&(u=!1),s=c.structural}if(et(r)||r===!0){var p=nt.data(e,B);K(p)&&(r=p)}if(s&&r===!1)break;if(o||(o=R(e,i)),o&&a)break;e=a||!(l=nt.data(e,Mt))?e.parentNode:y(l)}var h=(!s||r)&&u!==!0;return h&&a&&o}function F(t,e,n){n=n||{},n.state=e;var r=y(t);r.setAttribute(Et,e);var i=q.get(r),o=i?Y(i,n):n;q.put(r,o)}var q=new m,V=new m,U=null,L=i.$watch(function(){return 0===_.totalPendingRequests},function(t){t&&(L(),i.$$postDigest(function(){i.$$postDigest(function(){null===U&&(U=!0)})}))}),H=Object.create(null),z=e.classNameFilter(),Z=z?function(t){return z.test(t)}:function(){return!0},rt=f(C),it=t.Node.prototype.contains||function(t){return this===t||!!(16&this.compareDocumentPosition(t))},ot={on:function(t,e,n){var r=u(e);H[t]=H[t]||[],H[t].push({node:r,callback:n}),nt(e).on("$destroy",function(){var i=q.get(r);i||ot.off(t,e,n)})},off:function(t,e,n){if(1!==arguments.length||tt(arguments[0])){var r=H[t];r&&(H[t]=1===arguments.length?null:D(r,e,n))}else{e=arguments[0];for(var i in H)H[i]=D(H[i],e)}},pin:function(t,e){n(J(t),"element","not an element"),n(J(e),"parentElement","not an element"),t.data(Mt,e)},push:function(t,e,n,r){return n=n||{},n.domOperation=r,O(t,e,n)},enabled:function(t,e){var n=arguments.length;if(0===n)e=!!U;else{var r=J(t);if(r){var i=y(t);1===n?e=!V.get(i):V.put(i,!e)}else e=U=!!t}return e}};return ot}]}],At=["$animateProvider",function(t){function e(t,e){t.data(s,e)}function n(t){t.removeData(s)}function i(t){return t.data(s)}var o="ng-animate-ref",a=this.drivers=[],s="$$animationRunner";this.$get=["$$jqLite","$rootScope","$injector","$$AnimateRunner","$$HashMap","$$rAFScheduler",function(t,s,u,l,c,d){function v(t){function e(t){if(t.processed)return t;t.processed=!0;var n=t.domNode,r=n.parentNode;o.put(n,t);for(var a;r;){if(a=o.get(r)){a.processed||(a=e(a));break}r=r.parentNode}return(a||i).children.push(t),t}function n(t){var e,n=[],r=[];for(e=0;e<t.children.length;e++)r.push(t.children[e]);var i=r.length,o=0,a=[];for(e=0;e<r.length;e++){var s=r[e];i<=0&&(i=o,o=0,n.push(a),a=[]),a.push(s.fn),s.children.forEach(function(t){o++,r.push(t)}),i--}return a.length&&n.push(a),n}var r,i={children:[]},o=new c;for(r=0;r<t.length;r++){var a=t[r];o.put(a.domNode,t[r]={domNode:a.domNode,fn:a.fn,children:[]})}for(r=0;r<t.length;r++)e(t[r]);return n(i)}var g=[],m=f(t);return function(c,f,b){function $(t){var e="["+o+"]",n=t.hasAttribute(o)?[t]:t.querySelectorAll(e),r=[];return G(n,function(t){var e=t.getAttribute(o);e&&e.length&&r.push(t)}),r}function w(t){var e=[],n={};G(t,function(t,r){var i=t.element,a=y(i),s=t.event,u=["enter","move"].indexOf(s)>=0,l=t.structural?$(a):[];if(l.length){var c=u?"to":"from";G(l,function(t){var e=t.getAttribute(o);n[e]=n[e]||{},n[e][c]={animationID:r,element:nt(t)}})}else e.push(t)});var r={},i={};return G(n,function(n,o){var a=n.from,s=n.to;if(!a||!s){var u=a?a.animationID:s.animationID,l=u.toString();return void(r[l]||(r[l]=!0,e.push(t[u])))}var c=t[a.animationID],f=t[s.animationID],p=a.animationID.toString();if(!i[p]){var h=i[p]={structural:!0,beforeStart:function(){c.beforeStart(),f.beforeStart()},close:function(){c.close(),f.close()},classes:x(c.classes,f.classes),from:c,to:f,anchors:[]};h.classes.length?e.push(h):(e.push(c),e.push(f))}i[p].anchors.push({out:a.element,in:s.element})}),e}function x(t,e){t=t.split(" "),e=e.split(" ");for(var n=[],r=0;r<t.length;r++){var i=t[r];if("ng-"!==i.substring(0,3))for(var o=0;o<e.length;o++)if(i===e[o]){n.push(i);break}}return n.join(" ")}function _(t){for(var e=a.length-1;e>=0;e--){var n=a[e],r=u.get(n),i=r(t);if(i)return i}}function C(){c.addClass(H),A&&t.addClass(c,A),O&&(t.removeClass(c,O),O=null)}function k(t,e){function n(t){var n=i(t);n&&n.setHost(e)}t.from&&t.to?(n(t.from.element),n(t.to.element)):n(t.element)}function S(){var t=i(c);!t||"leave"===f&&b.$$domOperationFired||t.end()}function T(e){c.off("$destroy",S),n(c),m(c,b),h(c,b),b.domOperation(),A&&t.removeClass(c,A),c.removeClass(H),M.complete(!e)}b=p(b);var E=["enter","move","leave"].indexOf(f)>=0,M=new l({end:function(){T()},cancel:function(){T(!0)}});if(!a.length)return T(),M;e(c,M);var D=r(c.attr("class"),r(b.addClass,b.removeClass)),A=b.tempClasses;A&&(D+=" "+A,b.tempClasses=null);var O;return E&&(O="ng-"+f+L,t.addClass(c,O)),g.push({element:c,classes:D,event:f,structural:E,options:b,beforeStart:C,close:T}),c.on("$destroy",S),g.length>1?M:(s.$$postDigest(function(){var t=[];G(g,function(e){i(e.element)?t.push(e):e.close()}),g.length=0;var e=w(t),n=[];G(e,function(t){n.push({domNode:y(t.from?t.from.element:t.element),fn:function(){t.beforeStart();var e,n=t.close,r=t.anchors?t.from.element||t.to.element:t.element;if(i(r)){var o=_(t);o&&(e=o.start)}if(e){var a=e();a.done(function(t){n(!t)}),k(t,a)}else n()}})}),d(v(n))}),M)}}]}],Ot=["$animate","$rootScope",function(t,e){return{restrict:"A",transclude:"element",terminal:!0,priority:600,link:function(e,n,r,i,o){var a,s;e.$watchCollection(r.ngAnimateSwap||r.for,function(r){a&&t.leave(a),s&&(s.$destroy(),s=null),(r||0===r)&&(s=e.$new(),o(s,function(e){a=e,t.enter(e,null,n)}))})}}}];e.module("ngAnimate",[],function(){rt=e.noop,W=e.copy,Y=e.extend,nt=e.element,G=e.forEach,X=e.isArray,tt=e.isString,Q=e.isObject,et=e.isUndefined,K=e.isDefined,Z=e.isFunction,J=e.isElement}).directive("ngAnimateSwap",Ot).directive("ngAnimateChildren",mt).factory("$$rAFScheduler",gt).provider("$$animateQueue",Dt).provider("$$animation",At).provider("$animateCss",Ct).provider("$$animateCssDriver",kt).provider("$$animateJs",St).provider("$$animateJsDriver",Tt)}(window,window.angular)},23:function(t,e){/**
 * @license AngularJS v1.6.1
 * (c) 2010-2016 Google, Inc. http://angularjs.org
 * License: MIT
 */
!function(t,e){"use strict";function n(t){return null!=t&&""!==t&&"hasOwnProperty"!==t&&a.test("."+t)}function r(t,r){if(!n(r))throw o("badmember",'Dotted member path "@{0}" is invalid.',r);for(var i=r.split("."),a=0,s=i.length;a<s&&e.isDefined(t);a++){var u=i[a];t=null!==t?t[u]:void 0}return t}function i(t,n){n=n||{},e.forEach(n,function(t,e){delete n[e]});for(var r in t)!t.hasOwnProperty(r)||"$"===r.charAt(0)&&"$"===r.charAt(1)||(n[r]=t[r]);return n}var o=e.$$minErr("$resource"),a=/^(\.[a-zA-Z_$@][0-9a-zA-Z_$@]*)+$/;e.module("ngResource",["ng"]).provider("$resource",function(){var t=/^https?:\/\/\[[^\]]*][^\/]*/,n=this;this.defaults={stripTrailingSlashes:!0,cancellable:!1,actions:{get:{method:"GET"},save:{method:"POST"},query:{method:"GET",isArray:!0},remove:{method:"DELETE"},delete:{method:"DELETE"}}},this.$get=["$http","$log","$q","$timeout",function(a,s,u,l){function c(t,e){this.template=t,this.defaults=d({},n.defaults,e),this.urlParams={}}function f(t,e,$,w){function x(t,n){var i={};return n=d({},e,n),h(n,function(e,n){y(e)&&(e=e(t)),i[n]=e&&e.charAt&&"@"===e.charAt(0)?r(t,e.substr(1)):e}),i}function _(t){return t.resource}function C(t){i(t||{},this)}var k=new c(t,w);return $=d({},n.defaults.actions,$),C.prototype.toJSON=function(){var t=d({},this);return delete t.$promise,delete t.$resolved,t},h($,function(t,e){var n=/^(POST|PUT|PATCH)$/i.test(t.method),r=t.timeout,c=m(t.cancellable)?t.cancellable:k.defaults.cancellable;r&&!b(r)&&(s.debug("ngResource:\n  Only numeric values are allowed as `timeout`.\n  Promises are not supported in $resource, because the same value would be used for multiple requests. If you are looking for a way to cancel requests, you should use the `cancellable` option."),delete t.timeout,r=null),C[e]=function(s,f,m,b){function $(t){F.catch(p),M.resolve(t)}var w,S,T,E={};switch(arguments.length){case 4:T=b,S=m;case 3:case 2:if(!y(f)){E=s,w=f,S=m;break}if(y(s)){S=s,T=f;break}S=f,T=m;case 1:y(s)?S=s:n?w=s:E=s;break;case 0:break;default:throw o("badargs","Expected up to 4 arguments [params, data, success, error], got {0} arguments",arguments.length)}var M,D,A=this instanceof C,O=A?w:t.isArray?[]:new C(w),N={},P=t.interceptor&&t.interceptor.response||_,R=t.interceptor&&t.interceptor.responseError||void 0,I=!!T,j=!!R;h(t,function(t,e){switch(e){default:N[e]=v(t);break;case"params":case"isArray":case"interceptor":case"cancellable":}}),!A&&c&&(M=u.defer(),N.timeout=M.promise,r&&(D=l(M.resolve,r))),n&&(N.data=w),k.setUrlParams(N,d({},x(w,t.params||{}),E),t.url);var F=a(N).then(function(n){var r=n.data;if(r){if(g(r)!==!!t.isArray)throw o("badcfg","Error in resource configuration for action `{0}`. Expected response to contain an {1} but got an {2} (Request: {3} {4})",e,t.isArray?"array":"object",g(r)?"array":"object",N.method,N.url);if(t.isArray)O.length=0,h(r,function(t){"object"==typeof t?O.push(new C(t)):O.push(t)});else{var a=O.$promise;i(r,O),O.$promise=a}}return n.resource=O,n});return F=F.finally(function(){O.$resolved=!0,!A&&c&&(O.$cancelRequest=p,l.cancel(D),M=D=N.timeout=null)}),F=F.then(function(t){var e=P(t);return(S||p)(e,t.headers,t.status,t.statusText),e},I||j?function(t){return I&&T(t),j?R(t):u.reject(t)}:void 0),I&&!j&&F.catch(p),A?F:(O.$promise=F,O.$resolved=!1,c&&(O.$cancelRequest=$),O)},C.prototype["$"+e]=function(t,n,r){y(t)&&(r=n,n=t,t={});var i=C[e].call(this,t,this,n,r);return i.$promise||i}}),C.bind=function(n){var r=d({},e,n);return f(t,r,$,w)},C}var p=e.noop,h=e.forEach,d=e.extend,v=e.copy,g=e.isArray,m=e.isDefined,y=e.isFunction,b=e.isNumber,$=e.$$encodeUriQuery,w=e.$$encodeUriSegment;return c.prototype={setUrlParams:function(e,n,r){var i,a,s=this,u=r||s.template,l="",c=s.urlParams=Object.create(null);h(u.split(/\W/),function(t){if("hasOwnProperty"===t)throw o("badname","hasOwnProperty is not a valid parameter name.");!new RegExp("^\\d+$").test(t)&&t&&new RegExp("(^|[^\\\\]):"+t+"(\\W|$)").test(u)&&(c[t]={isQueryParamValue:new RegExp("\\?.*=:"+t+"(?:\\W|$)").test(u)})}),u=u.replace(/\\:/g,":"),u=u.replace(t,function(t){return l=t,""}),n=n||{},h(s.urlParams,function(t,e){i=n.hasOwnProperty(e)?n[e]:s.defaults[e],m(i)&&null!==i?(a=t.isQueryParamValue?$(i,!0):w(i),u=u.replace(new RegExp(":"+e+"(\\W|$)","g"),function(t,e){return a+e})):u=u.replace(new RegExp("(/?):"+e+"(\\W|$)","g"),function(t,e,n){return"/"===n.charAt(0)?n:e+n})}),s.defaults.stripTrailingSlashes&&(u=u.replace(/\/+$/,"")||"/"),u=u.replace(/\/\.(?=\w+($|\?))/,"."),e.url=l+u.replace(/\/\\\./,"/."),h(n,function(t,n){s.urlParams[n]||(e.params=e.params||{},e.params[n]=t)})}},f}]})}(window,window.angular)},24:function(t,e){/** 
* @version 2.1.8
* @license MIT
*/
!function(t,e){"use strict";t.module("smart-table",[]).run(["$templateCache",function(t){t.put("template/smart-table/pagination.html",'<nav ng-if="numPages && pages.length >= 2"><ul class="pagination"><li ng-repeat="page in pages" ng-class="{active: page==currentPage}"><a href="javascript: void(0);" ng-click="selectPage(page)">{{page}}</a></li></ul></nav>')}]),t.module("smart-table").constant("stConfig",{pagination:{template:"template/smart-table/pagination.html",itemsByPage:10,displayedPages:5},search:{delay:400,inputEvent:"input"},select:{mode:"single",selectedClass:"st-selected"},sort:{ascentClass:"st-sort-ascent",descentClass:"st-sort-descent",descendingFirst:!1,skipNatural:!1,delay:300},pipe:{delay:100}}),t.module("smart-table").controller("stTableController",["$scope","$parse","$filter","$attrs",function(n,r,i,o){function a(t){return t?[].concat(t):[]}function s(){m=a(l(n)),b===!0&&$.pipe()}function u(t,e){if(e.indexOf(".")!=-1){var n=e.split("."),i=n.pop(),o=n.join("."),a=r(o)(t);delete a[i],0==Object.keys(a).length&&u(t,o)}else delete t[e]}var l,c,f,p=o.stTable,h=r(p),d=h.assign,v=i("orderBy"),g=i("filter"),m=a(h(n)),y={sort:{},search:{},pagination:{start:0,totalItemCount:0}},b=!0,$=this;o.stSafeSrc&&(l=r(o.stSafeSrc),n.$watch(function(){var t=l(n);return t&&t.length?t[0]:e},function(t,e){t!==e&&s()}),n.$watch(function(){var t=l(n);return t?t.length:0},function(t,e){t!==m.length&&s()}),n.$watch(function(){return l(n)},function(t,e){t!==e&&(y.pagination.start=0,s())})),this.sortBy=function(e,n){return y.sort.predicate=e,y.sort.reverse=n===!0,t.isFunction(e)?y.sort.functionName=e.name:delete y.sort.functionName,y.pagination.start=0,this.pipe()},this.search=function(e,n){var i=y.search.predicateObject||{},o=n?n:"$";return e=t.isString(e)?e.trim():e,r(o).assign(i,e),e||u(i,o),y.search.predicateObject=i,y.pagination.start=0,this.pipe()},this.pipe=function(){var t,r=y.pagination;c=y.search.predicateObject?g(m,y.search.predicateObject):m,y.sort.predicate&&(c=v(c,y.sort.predicate,y.sort.reverse)),r.totalItemCount=c.length,r.number!==e&&(r.numberOfPages=c.length>0?Math.ceil(c.length/r.number):1,r.start=r.start>=c.length?(r.numberOfPages-1)*r.number:r.start,t=c.slice(r.start,r.start+parseInt(r.number))),d(n,t||c)},this.select=function(t,r){var i=a(h(n)),o=i.indexOf(t);o!==-1&&("single"===r?(t.isSelected=t.isSelected!==!0,f&&(f.isSelected=!1),f=t.isSelected===!0?t:e):i[o].isSelected=!i[o].isSelected)},this.slice=function(t,e){return y.pagination.start=t,y.pagination.number=e,this.pipe()},this.tableState=function(){return y},this.getFilteredCollection=function(){return c||m},this.setFilterFunction=function(t){g=i(t)},this.setSortFunction=function(t){v=i(t)},this.preventPipeOnWatch=function(){b=!1}}]).directive("stTable",function(){return{restrict:"A",controller:"stTableController",link:function(t,e,n,r){n.stSetFilter&&r.setFilterFunction(n.stSetFilter),n.stSetSort&&r.setSortFunction(n.stSetSort)}}}),t.module("smart-table").directive("stSearch",["stConfig","$timeout","$parse",function(t,e,n){return{require:"^stTable",link:function(r,i,o,a){var s=a,u=null,l=o.stDelay||t.search.delay,c=o.stInputEvent||t.search.inputEvent;o.$observe("stSearch",function(t,e){var n=i[0].value;t!==e&&n&&(a.tableState().search={},s.search(n,t))}),r.$watch(function(){return a.tableState().search},function(t,e){var r=o.stSearch||"$";t.predicateObject&&n(r)(t.predicateObject)!==i[0].value&&(i[0].value=n(r)(t.predicateObject)||"")},!0),i.bind(c,function(t){t=t.originalEvent||t,null!==u&&e.cancel(u),u=e(function(){s.search(t.target.value,o.stSearch||""),u=null},l)})}}}]),t.module("smart-table").directive("stSelectRow",["stConfig",function(t){return{restrict:"A",require:"^stTable",scope:{row:"=stSelectRow"},link:function(e,n,r,i){var o=r.stSelectMode||t.select.mode;n.bind("click",function(){e.$apply(function(){i.select(e.row,o)})}),e.$watch("row.isSelected",function(e){e===!0?n.addClass(t.select.selectedClass):n.removeClass(t.select.selectedClass)})}}}]),t.module("smart-table").directive("stSort",["stConfig","$parse","$timeout",function(n,r,i){return{restrict:"A",require:"^stTable",link:function(o,a,s,u){function l(){y?h=0===h?2:h-1:h++;var e;f=t.isFunction(p(o))||t.isArray(p(o))?p(o):s.stSort,h%3===0&&!!m!=!0?(h=0,u.tableState().sort={},u.tableState().pagination.start=0,e=u.pipe.bind(u)):e=u.sortBy.bind(u,f,h%2===0),null!==b&&i.cancel(b),$<0?e():b=i(e,$)}var c,f=s.stSort,p=r(f),h=0,d=s.stClassAscent||n.sort.ascentClass,v=s.stClassDescent||n.sort.descentClass,g=[d,v],m=s.stSkipNatural!==e?s.stSkipNatural:n.sort.skipNatural,y=s.stDescendingFirst!==e?s.stDescendingFirst:n.sort.descendingFirst,b=null,$=s.stDelay||n.sort.delay;s.stSortDefault&&(c=o.$eval(s.stSortDefault)!==e?o.$eval(s.stSortDefault):s.stSortDefault),a.bind("click",function(){f&&o.$apply(l)}),c&&(h="reverse"===c?1:0,l()),o.$watch(function(){return u.tableState().sort},function(t){t.predicate!==f?(h=0,a.removeClass(d).removeClass(v)):(h=t.reverse===!0?2:1,a.removeClass(g[h%2]).addClass(g[h-1]))},!0)}}}]),t.module("smart-table").directive("stPagination",["stConfig",function(t){return{restrict:"EA",require:"^stTable",scope:{stItemsByPage:"=?",stDisplayedPages:"=?",stPageChange:"&"},templateUrl:function(e,n){return n.stTemplate?n.stTemplate:t.pagination.template},link:function(e,n,r,i){function o(){var t,n,r=i.tableState().pagination,o=1,a=e.currentPage;for(e.totalItemCount=r.totalItemCount,e.currentPage=Math.floor(r.start/r.number)+1,o=Math.max(o,e.currentPage-Math.abs(Math.floor(e.stDisplayedPages/2))),t=o+e.stDisplayedPages,t>r.numberOfPages&&(t=r.numberOfPages+1,o=Math.max(1,t-e.stDisplayedPages)),e.pages=[],e.numPages=r.numberOfPages,n=o;n<t;n++)e.pages.push(n);a!==e.currentPage&&e.stPageChange({newPage:e.currentPage})}e.stItemsByPage=e.stItemsByPage?+e.stItemsByPage:t.pagination.itemsByPage,e.stDisplayedPages=e.stDisplayedPages?+e.stDisplayedPages:t.pagination.displayedPages,e.currentPage=1,e.pages=[],e.$watch(function(){return i.tableState().pagination},o,!0),e.$watch("stItemsByPage",function(t,n){t!==n&&e.selectPage(1)}),e.$watch("stDisplayedPages",o),e.selectPage=function(t){t>0&&t<=e.numPages&&i.slice((t-1)*e.stItemsByPage,e.stItemsByPage)},i.tableState().pagination.number||i.slice(0,e.stItemsByPage)}}}]),t.module("smart-table").directive("stPipe",["stConfig","$timeout",function(e,n){return{require:"stTable",scope:{stPipe:"="},link:{pre:function(r,i,o,a){var s=null;t.isFunction(r.stPipe)&&(a.preventPipeOnWatch(),a.pipe=function(){return null!==s&&n.cancel(s),s=n(function(){r.stPipe(a.tableState(),a)},e.pipe.delay)})},post:function(t,e,n,r){r.pipe()}}}}])}(angular)},25:function(t,e){angular.module("ui.bootstrap",["ui.bootstrap.tpls","ui.bootstrap.collapse","ui.bootstrap.tabindex","ui.bootstrap.accordion","ui.bootstrap.alert","ui.bootstrap.buttons","ui.bootstrap.carousel","ui.bootstrap.dateparser","ui.bootstrap.isClass","ui.bootstrap.datepicker","ui.bootstrap.position","ui.bootstrap.datepickerPopup","ui.bootstrap.debounce","ui.bootstrap.multiMap","ui.bootstrap.dropdown","ui.bootstrap.stackedMap","ui.bootstrap.modal","ui.bootstrap.paging","ui.bootstrap.pager","ui.bootstrap.pagination","ui.bootstrap.tooltip","ui.bootstrap.popover","ui.bootstrap.progressbar","ui.bootstrap.rating","ui.bootstrap.tabs","ui.bootstrap.timepicker","ui.bootstrap.typeahead"]),angular.module("ui.bootstrap.tpls",["uib/template/accordion/accordion-group.html","uib/template/accordion/accordion.html","uib/template/alert/alert.html","uib/template/carousel/carousel.html","uib/template/carousel/slide.html","uib/template/datepicker/datepicker.html","uib/template/datepicker/day.html","uib/template/datepicker/month.html","uib/template/datepicker/year.html","uib/template/datepickerPopup/popup.html","uib/template/modal/window.html","uib/template/pager/pager.html","uib/template/pagination/pagination.html","uib/template/tooltip/tooltip-html-popup.html","uib/template/tooltip/tooltip-popup.html","uib/template/tooltip/tooltip-template-popup.html","uib/template/popover/popover-html.html","uib/template/popover/popover-template.html","uib/template/popover/popover.html","uib/template/progressbar/bar.html","uib/template/progressbar/progress.html","uib/template/progressbar/progressbar.html","uib/template/rating/rating.html","uib/template/tabs/tab.html","uib/template/tabs/tabset.html","uib/template/timepicker/timepicker.html","uib/template/typeahead/typeahead-match.html","uib/template/typeahead/typeahead-popup.html"]),angular.module("ui.bootstrap.collapse",[]).directive("uibCollapse",["$animate","$q","$parse","$injector",function(t,e,n,r){var i=r.has("$animateCss")?r.get("$animateCss"):null;return{link:function(r,o,a){function s(){m=!!("horizontal"in a),m?(y={width:""},b={width:"0"}):(y={height:""},b={height:"0"}),r.$eval(a.uibCollapse)||o.addClass("in").addClass("collapse").attr("aria-expanded",!0).attr("aria-hidden",!1).css(y)}function u(t){return m?{width:t.scrollWidth+"px"}:{height:t.scrollHeight+"px"}}function l(){o.hasClass("collapse")&&o.hasClass("in")||e.resolve(h(r)).then(function(){o.removeClass("collapse").addClass("collapsing").attr("aria-expanded",!0).attr("aria-hidden",!1),i?i(o,{addClass:"in",easing:"ease",css:{overflow:"hidden"},to:u(o[0])}).start().finally(c):t.addClass(o,"in",{css:{overflow:"hidden"},to:u(o[0])}).then(c)})}function c(){o.removeClass("collapsing").addClass("collapse").css(y),d(r)}function f(){return o.hasClass("collapse")||o.hasClass("in")?void e.resolve(v(r)).then(function(){o.css(u(o[0])).removeClass("collapse").addClass("collapsing").attr("aria-expanded",!1).attr("aria-hidden",!0),i?i(o,{removeClass:"in",to:b}).start().finally(p):t.removeClass(o,"in",{to:b}).then(p)}):p()}function p(){o.css(b),o.removeClass("collapsing").addClass("collapse"),g(r)}var h=n(a.expanding),d=n(a.expanded),v=n(a.collapsing),g=n(a.collapsed),m=!1,y={},b={};s(),r.$watch(a.uibCollapse,function(t){t?f():l()})}}}]),angular.module("ui.bootstrap.tabindex",[]).directive("uibTabindexToggle",function(){return{restrict:"A",link:function(t,e,n){n.$observe("disabled",function(t){n.$set("tabindex",t?-1:null)})}}}),angular.module("ui.bootstrap.accordion",["ui.bootstrap.collapse","ui.bootstrap.tabindex"]).constant("uibAccordionConfig",{closeOthers:!0}).controller("UibAccordionController",["$scope","$attrs","uibAccordionConfig",function(t,e,n){this.groups=[],this.closeOthers=function(r){var i=angular.isDefined(e.closeOthers)?t.$eval(e.closeOthers):n.closeOthers;i&&angular.forEach(this.groups,function(t){t!==r&&(t.isOpen=!1)})},this.addGroup=function(t){var e=this;this.groups.push(t),t.$on("$destroy",function(n){e.removeGroup(t)})},this.removeGroup=function(t){var e=this.groups.indexOf(t);e!==-1&&this.groups.splice(e,1)}}]).directive("uibAccordion",function(){return{controller:"UibAccordionController",controllerAs:"accordion",transclude:!0,templateUrl:function(t,e){return e.templateUrl||"uib/template/accordion/accordion.html"}}}).directive("uibAccordionGroup",function(){return{require:"^uibAccordion",transclude:!0,restrict:"A",templateUrl:function(t,e){return e.templateUrl||"uib/template/accordion/accordion-group.html"},scope:{heading:"@",panelClass:"@?",isOpen:"=?",isDisabled:"=?"},controller:function(){this.setHeading=function(t){this.heading=t}},link:function(t,e,n,r){e.addClass("panel"),r.addGroup(t),t.openClass=n.openClass||"panel-open",t.panelClass=n.panelClass||"panel-default",t.$watch("isOpen",function(n){e.toggleClass(t.openClass,!!n),n&&r.closeOthers(t)}),t.toggleOpen=function(e){t.isDisabled||e&&32!==e.which||(t.isOpen=!t.isOpen)};var i="accordiongroup-"+t.$id+"-"+Math.floor(1e4*Math.random());t.headingId=i+"-tab",t.panelId=i+"-panel"}}}).directive("uibAccordionHeading",function(){return{transclude:!0,template:"",replace:!0,require:"^uibAccordionGroup",link:function(t,e,n,r,i){r.setHeading(i(t,angular.noop))}}}).directive("uibAccordionTransclude",function(){function t(){return"uib-accordion-header,data-uib-accordion-header,x-uib-accordion-header,uib\\:accordion-header,[uib-accordion-header],[data-uib-accordion-header],[x-uib-accordion-header]"}return{require:"^uibAccordionGroup",link:function(e,n,r,i){e.$watch(function(){return i[r.uibAccordionTransclude]},function(e){if(e){var r=angular.element(n[0].querySelector(t()));r.html(""),r.append(e)}})}}}),angular.module("ui.bootstrap.alert",[]).controller("UibAlertController",["$scope","$element","$attrs","$interpolate","$timeout",function(t,e,n,r,i){t.closeable=!!n.close,e.addClass("alert"),n.$set("role","alert"),t.closeable&&e.addClass("alert-dismissible");var o=angular.isDefined(n.dismissOnTimeout)?r(n.dismissOnTimeout)(t.$parent):null;o&&i(function(){t.close()},parseInt(o,10))}]).directive("uibAlert",function(){return{controller:"UibAlertController",controllerAs:"alert",restrict:"A",templateUrl:function(t,e){return e.templateUrl||"uib/template/alert/alert.html"},transclude:!0,scope:{close:"&"}}}),angular.module("ui.bootstrap.buttons",[]).constant("uibButtonConfig",{activeClass:"active",toggleEvent:"click"}).controller("UibButtonsController",["uibButtonConfig",function(t){this.activeClass=t.activeClass||"active",this.toggleEvent=t.toggleEvent||"click"}]).directive("uibBtnRadio",["$parse",function(t){return{require:["uibBtnRadio","ngModel"],controller:"UibButtonsController",controllerAs:"buttons",link:function(e,n,r,i){var o=i[0],a=i[1],s=t(r.uibUncheckable);n.find("input").css({display:"none"}),a.$render=function(){n.toggleClass(o.activeClass,angular.equals(a.$modelValue,e.$eval(r.uibBtnRadio)))},n.on(o.toggleEvent,function(){if(!r.disabled){var t=n.hasClass(o.activeClass);t&&!angular.isDefined(r.uncheckable)||e.$apply(function(){a.$setViewValue(t?null:e.$eval(r.uibBtnRadio)),a.$render()})}}),r.uibUncheckable&&e.$watch(s,function(t){r.$set("uncheckable",t?"":void 0)})}}}]).directive("uibBtnCheckbox",function(){return{require:["uibBtnCheckbox","ngModel"],controller:"UibButtonsController",controllerAs:"button",link:function(t,e,n,r){function i(){return a(n.btnCheckboxTrue,!0)}function o(){return a(n.btnCheckboxFalse,!1)}function a(e,n){return angular.isDefined(e)?t.$eval(e):n}var s=r[0],u=r[1];e.find("input").css({display:"none"}),u.$render=function(){e.toggleClass(s.activeClass,angular.equals(u.$modelValue,i()))},e.on(s.toggleEvent,function(){n.disabled||t.$apply(function(){u.$setViewValue(e.hasClass(s.activeClass)?o():i()),u.$render()})})}}}),angular.module("ui.bootstrap.carousel",[]).controller("UibCarouselController",["$scope","$element","$interval","$timeout","$animate",function(t,e,n,r,i){function o(){for(;b.length;)b.shift()}function a(t){for(var e=0;e<g.length;e++)g[e].slide.active=e===t}function s(n,r,u){if(!$){if(angular.extend(n,{direction:u}),angular.extend(g[y].slide||{},{direction:u}),i.enabled(e)&&!t.$currentTransition&&g[r].element&&v.slides.length>1){g[r].element.data(m,n.direction);var l=v.getCurrentIndex();angular.isNumber(l)&&g[l].element&&g[l].element.data(m,n.direction),t.$currentTransition=!0,i.on("addClass",g[r].element,function(e,n){if("close"===n&&(t.$currentTransition=null,i.off("addClass",e),b.length)){var r=b.pop().slide,a=r.index,u=a>v.getCurrentIndex()?"next":"prev";o(),s(r,a,u)}})}t.active=n.index,y=n.index,a(r),f()}}function u(t){for(var e=0;e<g.length;e++)if(g[e].slide===t)return e}function l(){h&&(n.cancel(h),h=null)}function c(e){e.length||(t.$currentTransition=null,o())}function f(){l();var e=+t.interval;!isNaN(e)&&e>0&&(h=n(p,e))}function p(){var e=+t.interval;d&&!isNaN(e)&&e>0&&g.length?t.next():t.pause()}var h,d,v=this,g=v.slides=t.slides=[],m="uib-slideDirection",y=t.active,b=[],$=!1;e.addClass("carousel"),v.addSlide=function(e,n){g.push({slide:e,element:n}),g.sort(function(t,e){return+t.slide.index-+e.slide.index}),(e.index===t.active||1===g.length&&!angular.isNumber(t.active))&&(t.$currentTransition&&(t.$currentTransition=null),y=e.index,t.active=e.index,a(y),v.select(g[u(e)]),1===g.length&&t.play())},v.getCurrentIndex=function(){for(var t=0;t<g.length;t++)if(g[t].slide.index===y)return t},v.next=t.next=function(){var e=(v.getCurrentIndex()+1)%g.length;return 0===e&&t.noWrap()?void t.pause():v.select(g[e],"next")},v.prev=t.prev=function(){var e=v.getCurrentIndex()-1<0?g.length-1:v.getCurrentIndex()-1;return t.noWrap()&&e===g.length-1?void t.pause():v.select(g[e],"prev")},v.removeSlide=function(e){var n=u(e),r=b.indexOf(g[n]);r!==-1&&b.splice(r,1),g.splice(n,1),g.length>0&&y===n?n>=g.length?(y=g.length-1,t.active=y,a(y),v.select(g[g.length-1])):(y=n,t.active=y,a(y),v.select(g[n])):y>n&&(y--,t.active=y),0===g.length&&(y=null,t.active=null,o())},v.select=t.select=function(e,n){var r=u(e.slide);void 0===n&&(n=r>v.getCurrentIndex()?"next":"prev"),e.slide.index===y||t.$currentTransition?e&&e.slide.index!==y&&t.$currentTransition&&b.push(g[r]):s(e.slide,r,n)},t.indexOfSlide=function(t){return+t.slide.index},t.isActive=function(e){return t.active===e.slide.index},t.isPrevDisabled=function(){return 0===t.active&&t.noWrap()},t.isNextDisabled=function(){return t.active===g.length-1&&t.noWrap()},t.pause=function(){t.noPause||(d=!1,l())},t.play=function(){d||(d=!0,f())},e.on("mouseenter",t.pause),e.on("mouseleave",t.play),t.$on("$destroy",function(){$=!0,l()}),t.$watch("noTransition",function(t){i.enabled(e,!t)}),t.$watch("interval",f),t.$watchCollection("slides",c),t.$watch("active",function(t){if(angular.isNumber(t)&&y!==t){for(var e=0;e<g.length;e++)if(g[e].slide.index===t){t=e;break}var n=g[t];n&&(a(t),v.select(g[t]),y=t)}})}]).directive("uibCarousel",function(){return{transclude:!0,controller:"UibCarouselController",controllerAs:"carousel",restrict:"A",templateUrl:function(t,e){return e.templateUrl||"uib/template/carousel/carousel.html"},scope:{active:"=",interval:"=",noTransition:"=",noPause:"=",noWrap:"&"}}}).directive("uibSlide",["$animate",function(t){return{require:"^uibCarousel",restrict:"A",transclude:!0,templateUrl:function(t,e){return e.templateUrl||"uib/template/carousel/slide.html"},scope:{actual:"=?",index:"=?"},link:function(e,n,r,i){n.addClass("item"),i.addSlide(e,n),e.$on("$destroy",function(){i.removeSlide(e)}),e.$watch("active",function(e){t[e?"addClass":"removeClass"](n,"active")})}}}]).animation(".item",["$animateCss",function(t){function e(t,e,n){t.removeClass(e),n&&n()}var n="uib-slideDirection";return{beforeAddClass:function(r,i,o){if("active"===i){var a=!1,s=r.data(n),u="next"===s?"left":"right",l=e.bind(this,r,u+" "+s,o);return r.addClass(s),t(r,{addClass:u}).start().done(l),function(){a=!0}}o()},beforeRemoveClass:function(r,i,o){if("active"===i){var a=!1,s=r.data(n),u="next"===s?"left":"right",l=e.bind(this,r,u,o);return t(r,{addClass:u}).start().done(l),function(){a=!0}}o()}}}]),angular.module("ui.bootstrap.dateparser",[]).service("uibDateParser",["$log","$locale","dateFilter","orderByFilter","filterFilter",function(t,e,n,r,i){function o(t){return i(y,{key:t},!0)[0]}function a(t){var e=[],n=t.split(""),i=t.indexOf("'");if(i>-1){var o=!1;t=t.split("");for(var a=i;a<t.length;a++)o?("'"===t[a]&&(a+1<t.length&&"'"===t[a+1]?(t[a+1]="$",n[a+1]=""):(n[a]="",o=!1)),t[a]="$"):"'"===t[a]&&(t[a]="$",n[a]="",o=!0);t=t.join("")}return angular.forEach(y,function(r){var i=t.indexOf(r.key);if(i>-1){t=t.split(""),n[i]="("+r.regex+")",t[i]="$";for(var o=i+1,a=i+r.key.length;o<a;o++)n[o]="",t[o]="$";t=t.join(""),e.push({index:i,key:r.key,apply:r.apply,matcher:r.regex})}}),{regex:new RegExp("^"+n.join("")+"$"),map:r(e,"index")}}function s(t){for(var e,n,r=[],i=0;i<t.length;)if(angular.isNumber(n)){if("'"===t.charAt(i))(i+1>=t.length||"'"!==t.charAt(i+1))&&(r.push(u(t,n,i)),n=null);else if(i===t.length)for(;n<t.length;)e=l(t,n),r.push(e),n=e.endIdx;i++}else"'"!==t.charAt(i)?(e=l(t,i),r.push(e.parser),i=e.endIdx):(n=i,i++);return r}function u(t,e,n){return function(){return t.substr(e+1,n-e-1)}}function l(t,e){for(var n=t.substr(e),r=0;r<y.length;r++)if(new RegExp("^"+y[r].key).test(n)){var i=y[r];return{endIdx:e+i.key.length,parser:i.formatter}}return{endIdx:e+1,parser:function(){return n.charAt(0)}}}function c(t,e,n){return!(n<1)&&(1===e&&n>28?29===n&&(t%4===0&&t%100!==0||t%400===0):3!==e&&5!==e&&8!==e&&10!==e||n<31)}function f(t){return parseInt(t,10)}function p(t,e){return t&&e?g(t,e):t}function h(t,e){return t&&e?g(t,e,!0):t}function d(t,e){t=t.replace(/:/g,"");var n=Date.parse("Jan 01, 1970 00:00:00 "+t)/6e4;return isNaN(n)?e:n}function v(t,e){return t=new Date(t.getTime()),t.setMinutes(t.getMinutes()+e),t}function g(t,e,n){n=n?-1:1;var r=t.getTimezoneOffset(),i=d(e,r);return v(t,n*(i-r))}var m,y,b=/[\\\^\$\*\+\?\|\[\]\(\)\.\{\}]/g;this.init=function(){m=e.id,this.parsers={},this.formatters={},y=[{key:"yyyy",regex:"\\d{4}",apply:function(t){this.year=+t},formatter:function(t){var e=new Date;return e.setFullYear(Math.abs(t.getFullYear())),n(e,"yyyy")}},{key:"yy",regex:"\\d{2}",apply:function(t){t=+t,this.year=t<69?t+2e3:t+1900},formatter:function(t){var e=new Date;return e.setFullYear(Math.abs(t.getFullYear())),n(e,"yy")}},{key:"y",regex:"\\d{1,4}",apply:function(t){this.year=+t},formatter:function(t){var e=new Date;return e.setFullYear(Math.abs(t.getFullYear())),n(e,"y")}},{key:"M!",regex:"0?[1-9]|1[0-2]",apply:function(t){this.month=t-1},formatter:function(t){var e=t.getMonth();return/^[0-9]$/.test(e)?n(t,"MM"):n(t,"M")}},{key:"MMMM",regex:e.DATETIME_FORMATS.MONTH.join("|"),apply:function(t){this.month=e.DATETIME_FORMATS.MONTH.indexOf(t)},formatter:function(t){return n(t,"MMMM")}},{key:"MMM",regex:e.DATETIME_FORMATS.SHORTMONTH.join("|"),apply:function(t){this.month=e.DATETIME_FORMATS.SHORTMONTH.indexOf(t)},formatter:function(t){return n(t,"MMM")}},{key:"MM",regex:"0[1-9]|1[0-2]",apply:function(t){this.month=t-1},formatter:function(t){return n(t,"MM")}},{key:"M",regex:"[1-9]|1[0-2]",apply:function(t){this.month=t-1},formatter:function(t){return n(t,"M")}},{key:"d!",regex:"[0-2]?[0-9]{1}|3[0-1]{1}",apply:function(t){this.date=+t},formatter:function(t){var e=t.getDate();return/^[1-9]$/.test(e)?n(t,"dd"):n(t,"d")}},{key:"dd",regex:"[0-2][0-9]{1}|3[0-1]{1}",apply:function(t){this.date=+t},formatter:function(t){return n(t,"dd")}},{key:"d",regex:"[1-2]?[0-9]{1}|3[0-1]{1}",apply:function(t){this.date=+t},formatter:function(t){return n(t,"d")}},{key:"EEEE",regex:e.DATETIME_FORMATS.DAY.join("|"),formatter:function(t){return n(t,"EEEE")}},{key:"EEE",regex:e.DATETIME_FORMATS.SHORTDAY.join("|"),formatter:function(t){return n(t,"EEE")}},{key:"HH",regex:"(?:0|1)[0-9]|2[0-3]",apply:function(t){this.hours=+t},formatter:function(t){return n(t,"HH")}},{key:"hh",regex:"0[0-9]|1[0-2]",apply:function(t){this.hours=+t},formatter:function(t){return n(t,"hh")}},{key:"H",regex:"1?[0-9]|2[0-3]",apply:function(t){this.hours=+t},formatter:function(t){return n(t,"H")}},{key:"h",regex:"[0-9]|1[0-2]",apply:function(t){this.hours=+t},formatter:function(t){return n(t,"h")}},{key:"mm",regex:"[0-5][0-9]",apply:function(t){this.minutes=+t},formatter:function(t){return n(t,"mm")}},{key:"m",regex:"[0-9]|[1-5][0-9]",apply:function(t){this.minutes=+t},formatter:function(t){return n(t,"m")}},{key:"sss",regex:"[0-9][0-9][0-9]",apply:function(t){this.milliseconds=+t},formatter:function(t){return n(t,"sss")}},{key:"ss",regex:"[0-5][0-9]",apply:function(t){this.seconds=+t},formatter:function(t){return n(t,"ss")}},{key:"s",regex:"[0-9]|[1-5][0-9]",apply:function(t){this.seconds=+t},formatter:function(t){return n(t,"s")}},{key:"a",regex:e.DATETIME_FORMATS.AMPMS.join("|"),apply:function(t){12===this.hours&&(this.hours=0),"PM"===t&&(this.hours+=12)},formatter:function(t){return n(t,"a")}},{key:"Z",regex:"[+-]\\d{4}",apply:function(t){var e=t.match(/([+-])(\d{2})(\d{2})/),n=e[1],r=e[2],i=e[3];this.hours+=f(n+r),this.minutes+=f(n+i)},formatter:function(t){return n(t,"Z")}},{key:"ww",regex:"[0-4][0-9]|5[0-3]",formatter:function(t){return n(t,"ww")}},{key:"w",regex:"[0-9]|[1-4][0-9]|5[0-3]",formatter:function(t){return n(t,"w")}},{key:"GGGG",regex:e.DATETIME_FORMATS.ERANAMES.join("|").replace(/\s/g,"\\s"),formatter:function(t){return n(t,"GGGG")}},{key:"GGG",regex:e.DATETIME_FORMATS.ERAS.join("|"),formatter:function(t){return n(t,"GGG")}},{key:"GG",regex:e.DATETIME_FORMATS.ERAS.join("|"),formatter:function(t){return n(t,"GG")}},{key:"G",regex:e.DATETIME_FORMATS.ERAS.join("|"),formatter:function(t){return n(t,"G")}}],angular.version.major>=1&&angular.version.minor>4&&y.push({key:"LLLL",regex:e.DATETIME_FORMATS.STANDALONEMONTH.join("|"),apply:function(t){this.month=e.DATETIME_FORMATS.STANDALONEMONTH.indexOf(t)},formatter:function(t){return n(t,"LLLL")}})},this.init(),this.getParser=function(t){var e=o(t);return e&&e.apply||null},this.overrideParser=function(t,e){var n=o(t);n&&angular.isFunction(e)&&(this.parsers={},n.apply=e)}.bind(this),this.filter=function(t,n){if(!angular.isDate(t)||isNaN(t)||!n)return"";n=e.DATETIME_FORMATS[n]||n,e.id!==m&&this.init(),this.formatters[n]||(this.formatters[n]=s(n));var r=this.formatters[n];return r.reduce(function(e,n){return e+n(t)},"")},this.parse=function(n,r,i){if(!angular.isString(n)||!r)return n;r=e.DATETIME_FORMATS[r]||r,r=r.replace(b,"\\$&"),e.id!==m&&this.init(),this.parsers[r]||(this.parsers[r]=a(r,"apply"));var o=this.parsers[r],s=o.regex,u=o.map,l=n.match(s),f=!1;if(l&&l.length){var p,h;angular.isDate(i)&&!isNaN(i.getTime())?p={year:i.getFullYear(),month:i.getMonth(),date:i.getDate(),hours:i.getHours(),minutes:i.getMinutes(),seconds:i.getSeconds(),milliseconds:i.getMilliseconds()}:(i&&t.warn("dateparser:","baseDate is not a valid date"),p={year:1900,month:0,date:1,hours:0,minutes:0,seconds:0,milliseconds:0});for(var d=1,v=l.length;d<v;d++){var g=u[d-1];"Z"===g.matcher&&(f=!0),g.apply&&g.apply.call(p,l[d])}var y=f?Date.prototype.setUTCFullYear:Date.prototype.setFullYear,$=f?Date.prototype.setUTCHours:Date.prototype.setHours;return c(p.year,p.month,p.date)&&(!angular.isDate(i)||isNaN(i.getTime())||f?(h=new Date(0),y.call(h,p.year,p.month,p.date),$.call(h,p.hours||0,p.minutes||0,p.seconds||0,p.milliseconds||0)):(h=new Date(i),y.call(h,p.year,p.month,p.date),$.call(h,p.hours,p.minutes,p.seconds,p.milliseconds))),h}},this.toTimezone=p,this.fromTimezone=h,this.timezoneToOffset=d,this.addDateMinutes=v,this.convertTimezoneToLocal=g}]),angular.module("ui.bootstrap.isClass",[]).directive("uibIsClass",["$animate",function(t){var e=/^\s*([\s\S]+?)\s+on\s+([\s\S]+?)\s*$/,n=/^\s*([\s\S]+?)\s+for\s+([\s\S]+?)\s*$/;return{restrict:"A",compile:function(r,i){function o(t,e,n){u.push(t),l.push({scope:t,element:e}),d.forEach(function(e,n){a(e,t)}),t.$on("$destroy",s)}function a(e,r){var i=e.match(n),o=r.$eval(i[1]),a=i[2],s=c[e];if(!s){var u=function(e){var n=null;l.some(function(t){var r=t.scope.$eval(p);if(r===e)return n=t,!0}),s.lastActivated!==n&&(s.lastActivated&&t.removeClass(s.lastActivated.element,o),n&&t.addClass(n.element,o),s.lastActivated=n)};c[e]=s={lastActivated:null,scope:r,watchFn:u,compareWithExp:a,watcher:r.$watch(a,u)}}s.watchFn(r.$eval(a))}function s(t){var e=t.targetScope,n=u.indexOf(e);if(u.splice(n,1),l.splice(n,1),u.length){var r=u[0];angular.forEach(c,function(t){t.scope===e&&(t.watcher=r.$watch(t.compareWithExp,t.watchFn),t.scope=r)})}else c={}}var u=[],l=[],c={},f=i.uibIsClass.match(e),p=f[2],h=f[1],d=h.split(",");return o}}}]),angular.module("ui.bootstrap.datepicker",["ui.bootstrap.dateparser","ui.bootstrap.isClass"]).value("$datepickerSuppressError",!1).value("$datepickerLiteralWarning",!0).constant("uibDatepickerConfig",{datepickerMode:"day",formatDay:"dd",formatMonth:"MMMM",formatYear:"yyyy",formatDayHeader:"EEE",formatDayTitle:"MMMM yyyy",formatMonthTitle:"yyyy",maxDate:null,maxMode:"year",minDate:null,minMode:"day",monthColumns:3,ngModelOptions:{},shortcutPropagation:!1,showWeeks:!0,yearColumns:5,yearRows:4}).controller("UibDatepickerController",["$scope","$element","$attrs","$parse","$interpolate","$locale","$log","dateFilter","uibDatepickerConfig","$datepickerLiteralWarning","$datepickerSuppressError","uibDateParser",function(t,e,n,r,i,o,a,s,u,l,c,f){function p(e){t.datepickerMode=e,t.datepickerOptions.datepickerMode=e}var h=this,d={$setViewValue:angular.noop},v={},g=[];e.addClass("uib-datepicker"),n.$set("role","application"),t.datepickerOptions||(t.datepickerOptions={}),this.modes=["day","month","year"],["customClass","dateDisabled","datepickerMode","formatDay","formatDayHeader","formatDayTitle","formatMonth","formatMonthTitle","formatYear","maxDate","maxMode","minDate","minMode","monthColumns","showWeeks","shortcutPropagation","startingDay","yearColumns","yearRows"].forEach(function(e){switch(e){case"customClass":case"dateDisabled":t[e]=t.datepickerOptions[e]||angular.noop;break;case"datepickerMode":t.datepickerMode=angular.isDefined(t.datepickerOptions.datepickerMode)?t.datepickerOptions.datepickerMode:u.datepickerMode;break;case"formatDay":case"formatDayHeader":case"formatDayTitle":case"formatMonth":case"formatMonthTitle":case"formatYear":h[e]=angular.isDefined(t.datepickerOptions[e])?i(t.datepickerOptions[e])(t.$parent):u[e];break;case"monthColumns":case"showWeeks":case"shortcutPropagation":case"yearColumns":case"yearRows":h[e]=angular.isDefined(t.datepickerOptions[e])?t.datepickerOptions[e]:u[e];break;case"startingDay":angular.isDefined(t.datepickerOptions.startingDay)?h.startingDay=t.datepickerOptions.startingDay:angular.isNumber(u.startingDay)?h.startingDay=u.startingDay:h.startingDay=(o.DATETIME_FORMATS.FIRSTDAYOFWEEK+8)%7;break;case"maxDate":case"minDate":t.$watch("datepickerOptions."+e,function(t){t?angular.isDate(t)?h[e]=f.fromTimezone(new Date(t),v.timezone):(l&&a.warn("Literal date support has been deprecated, please switch to date object usage"),h[e]=new Date(s(t,"medium"))):h[e]=u[e]?f.fromTimezone(new Date(u[e]),v.timezone):null,h.refreshView()});break;case"maxMode":case"minMode":t.datepickerOptions[e]?t.$watch(function(){return t.datepickerOptions[e]},function(n){h[e]=t[e]=angular.isDefined(n)?n:t.datepickerOptions[e],("minMode"===e&&h.modes.indexOf(t.datepickerOptions.datepickerMode)<h.modes.indexOf(h[e])||"maxMode"===e&&h.modes.indexOf(t.datepickerOptions.datepickerMode)>h.modes.indexOf(h[e]))&&(t.datepickerMode=h[e],t.datepickerOptions.datepickerMode=h[e])}):h[e]=t[e]=u[e]||null}}),t.uniqueId="datepicker-"+t.$id+"-"+Math.floor(1e4*Math.random()),t.disabled=angular.isDefined(n.disabled)||!1,angular.isDefined(n.ngDisabled)&&g.push(t.$parent.$watch(n.ngDisabled,function(e){t.disabled=e,h.refreshView()})),t.isActive=function(e){return 0===h.compare(e.date,h.activeDate)&&(t.activeDateId=e.uid,!0)},this.init=function(e){d=e,v=e.$options||t.datepickerOptions.ngModelOptions||u.ngModelOptions,t.datepickerOptions.initDate?(h.activeDate=f.fromTimezone(t.datepickerOptions.initDate,v.timezone)||new Date,t.$watch("datepickerOptions.initDate",function(t){t&&(d.$isEmpty(d.$modelValue)||d.$invalid)&&(h.activeDate=f.fromTimezone(t,v.timezone),h.refreshView())})):h.activeDate=new Date;var n=d.$modelValue?new Date(d.$modelValue):new Date;this.activeDate=isNaN(n)?f.fromTimezone(new Date,v.timezone):f.fromTimezone(n,v.timezone),d.$render=function(){h.render()}},this.render=function(){if(d.$viewValue){var t=new Date(d.$viewValue),e=!isNaN(t);e?this.activeDate=f.fromTimezone(t,v.timezone):c||a.error('Datepicker directive: "ng-model" value must be a Date object')}this.refreshView()},this.refreshView=function(){if(this.element){t.selectedDt=null,this._refreshView(),t.activeDt&&(t.activeDateId=t.activeDt.uid);var e=d.$viewValue?new Date(d.$viewValue):null;e=f.fromTimezone(e,v.timezone),d.$setValidity("dateDisabled",!e||this.element&&!this.isDisabled(e))}},this.createDateObject=function(e,n){var r=d.$viewValue?new Date(d.$viewValue):null;r=f.fromTimezone(r,v.timezone);var i=new Date;i=f.fromTimezone(i,v.timezone);var o=this.compare(e,i),a={date:e,label:f.filter(e,n),selected:r&&0===this.compare(e,r),disabled:this.isDisabled(e),past:o<0,current:0===o,future:o>0,customClass:this.customClass(e)||null};return r&&0===this.compare(e,r)&&(t.selectedDt=a),h.activeDate&&0===this.compare(a.date,h.activeDate)&&(t.activeDt=a),a},this.isDisabled=function(e){return t.disabled||this.minDate&&this.compare(e,this.minDate)<0||this.maxDate&&this.compare(e,this.maxDate)>0||t.dateDisabled&&t.dateDisabled({
date:e,mode:t.datepickerMode})},this.customClass=function(e){return t.customClass({date:e,mode:t.datepickerMode})},this.split=function(t,e){for(var n=[];t.length>0;)n.push(t.splice(0,e));return n},t.select=function(e){if(t.datepickerMode===h.minMode){var n=d.$viewValue?f.fromTimezone(new Date(d.$viewValue),v.timezone):new Date(0,0,0,0,0,0,0);n.setFullYear(e.getFullYear(),e.getMonth(),e.getDate()),n=f.toTimezone(n,v.timezone),d.$setViewValue(n),d.$render()}else h.activeDate=e,p(h.modes[h.modes.indexOf(t.datepickerMode)-1]),t.$emit("uib:datepicker.mode");t.$broadcast("uib:datepicker.focus")},t.move=function(t){var e=h.activeDate.getFullYear()+t*(h.step.years||0),n=h.activeDate.getMonth()+t*(h.step.months||0);h.activeDate.setFullYear(e,n,1),h.refreshView()},t.toggleMode=function(e){e=e||1,t.datepickerMode===h.maxMode&&1===e||t.datepickerMode===h.minMode&&e===-1||(p(h.modes[h.modes.indexOf(t.datepickerMode)+e]),t.$emit("uib:datepicker.mode"))},t.keys={13:"enter",32:"space",33:"pageup",34:"pagedown",35:"end",36:"home",37:"left",38:"up",39:"right",40:"down"};var m=function(){h.element[0].focus()};t.$on("uib:datepicker.focus",m),t.keydown=function(e){var n=t.keys[e.which];if(n&&!e.shiftKey&&!e.altKey&&!t.disabled)if(e.preventDefault(),h.shortcutPropagation||e.stopPropagation(),"enter"===n||"space"===n){if(h.isDisabled(h.activeDate))return;t.select(h.activeDate)}else!e.ctrlKey||"up"!==n&&"down"!==n?(h.handleKeyDown(n,e),h.refreshView()):t.toggleMode("up"===n?1:-1)},e.on("keydown",function(e){t.$apply(function(){t.keydown(e)})}),t.$on("$destroy",function(){for(;g.length;)g.shift()()})}]).controller("UibDaypickerController",["$scope","$element","dateFilter",function(t,e,n){function r(t,e){return 1!==e||t%4!==0||t%100===0&&t%400!==0?o[e]:29}function i(t){var e=new Date(t);e.setDate(e.getDate()+4-(e.getDay()||7));var n=e.getTime();return e.setMonth(0),e.setDate(1),Math.floor(Math.round((n-e)/864e5)/7)+1}var o=[31,28,31,30,31,30,31,31,30,31,30,31];this.step={months:1},this.element=e,this.init=function(e){angular.extend(e,this),t.showWeeks=e.showWeeks,e.refreshView()},this.getDates=function(t,e){for(var n,r=new Array(e),i=new Date(t),o=0;o<e;)n=new Date(i),r[o++]=n,i.setDate(i.getDate()+1);return r},this._refreshView=function(){var e=this.activeDate.getFullYear(),r=this.activeDate.getMonth(),o=new Date(this.activeDate);o.setFullYear(e,r,1);var a=this.startingDay-o.getDay(),s=a>0?7-a:-a,u=new Date(o);s>0&&u.setDate(-s+1);for(var l=this.getDates(u,42),c=0;c<42;c++)l[c]=angular.extend(this.createDateObject(l[c],this.formatDay),{secondary:l[c].getMonth()!==r,uid:t.uniqueId+"-"+c});t.labels=new Array(7);for(var f=0;f<7;f++)t.labels[f]={abbr:n(l[f].date,this.formatDayHeader),full:n(l[f].date,"EEEE")};if(t.title=n(this.activeDate,this.formatDayTitle),t.rows=this.split(l,7),t.showWeeks){t.weekNumbers=[];for(var p=(11-this.startingDay)%7,h=t.rows.length,d=0;d<h;d++)t.weekNumbers.push(i(t.rows[d][p].date))}},this.compare=function(t,e){var n=new Date(t.getFullYear(),t.getMonth(),t.getDate()),r=new Date(e.getFullYear(),e.getMonth(),e.getDate());return n.setFullYear(t.getFullYear()),r.setFullYear(e.getFullYear()),n-r},this.handleKeyDown=function(t,e){var n=this.activeDate.getDate();if("left"===t)n-=1;else if("up"===t)n-=7;else if("right"===t)n+=1;else if("down"===t)n+=7;else if("pageup"===t||"pagedown"===t){var i=this.activeDate.getMonth()+("pageup"===t?-1:1);this.activeDate.setMonth(i,1),n=Math.min(r(this.activeDate.getFullYear(),this.activeDate.getMonth()),n)}else"home"===t?n=1:"end"===t&&(n=r(this.activeDate.getFullYear(),this.activeDate.getMonth()));this.activeDate.setDate(n)}}]).controller("UibMonthpickerController",["$scope","$element","dateFilter",function(t,e,n){this.step={years:1},this.element=e,this.init=function(t){angular.extend(t,this),t.refreshView()},this._refreshView=function(){for(var e,r=new Array(12),i=this.activeDate.getFullYear(),o=0;o<12;o++)e=new Date(this.activeDate),e.setFullYear(i,o,1),r[o]=angular.extend(this.createDateObject(e,this.formatMonth),{uid:t.uniqueId+"-"+o});t.title=n(this.activeDate,this.formatMonthTitle),t.rows=this.split(r,this.monthColumns),t.yearHeaderColspan=this.monthColumns>3?this.monthColumns-2:1},this.compare=function(t,e){var n=new Date(t.getFullYear(),t.getMonth()),r=new Date(e.getFullYear(),e.getMonth());return n.setFullYear(t.getFullYear()),r.setFullYear(e.getFullYear()),n-r},this.handleKeyDown=function(t,e){var n=this.activeDate.getMonth();if("left"===t)n-=1;else if("up"===t)n-=this.monthColumns;else if("right"===t)n+=1;else if("down"===t)n+=this.monthColumns;else if("pageup"===t||"pagedown"===t){var r=this.activeDate.getFullYear()+("pageup"===t?-1:1);this.activeDate.setFullYear(r)}else"home"===t?n=0:"end"===t&&(n=11);this.activeDate.setMonth(n)}}]).controller("UibYearpickerController",["$scope","$element","dateFilter",function(t,e,n){function r(t){return parseInt((t-1)/o,10)*o+1}var i,o;this.element=e,this.yearpickerInit=function(){i=this.yearColumns,o=this.yearRows*i,this.step={years:o}},this._refreshView=function(){for(var e,n=new Array(o),a=0,s=r(this.activeDate.getFullYear());a<o;a++)e=new Date(this.activeDate),e.setFullYear(s+a,0,1),n[a]=angular.extend(this.createDateObject(e,this.formatYear),{uid:t.uniqueId+"-"+a});t.title=[n[0].label,n[o-1].label].join(" - "),t.rows=this.split(n,i),t.columns=i},this.compare=function(t,e){return t.getFullYear()-e.getFullYear()},this.handleKeyDown=function(t,e){var n=this.activeDate.getFullYear();"left"===t?n-=1:"up"===t?n-=i:"right"===t?n+=1:"down"===t?n+=i:"pageup"===t||"pagedown"===t?n+=("pageup"===t?-1:1)*o:"home"===t?n=r(this.activeDate.getFullYear()):"end"===t&&(n=r(this.activeDate.getFullYear())+o-1),this.activeDate.setFullYear(n)}}]).directive("uibDatepicker",function(){return{templateUrl:function(t,e){return e.templateUrl||"uib/template/datepicker/datepicker.html"},scope:{datepickerOptions:"=?"},require:["uibDatepicker","^ngModel"],restrict:"A",controller:"UibDatepickerController",controllerAs:"datepicker",link:function(t,e,n,r){var i=r[0],o=r[1];i.init(o)}}}).directive("uibDaypicker",function(){return{templateUrl:function(t,e){return e.templateUrl||"uib/template/datepicker/day.html"},require:["^uibDatepicker","uibDaypicker"],restrict:"A",controller:"UibDaypickerController",link:function(t,e,n,r){var i=r[0],o=r[1];o.init(i)}}}).directive("uibMonthpicker",function(){return{templateUrl:function(t,e){return e.templateUrl||"uib/template/datepicker/month.html"},require:["^uibDatepicker","uibMonthpicker"],restrict:"A",controller:"UibMonthpickerController",link:function(t,e,n,r){var i=r[0],o=r[1];o.init(i)}}}).directive("uibYearpicker",function(){return{templateUrl:function(t,e){return e.templateUrl||"uib/template/datepicker/year.html"},require:["^uibDatepicker","uibYearpicker"],restrict:"A",controller:"UibYearpickerController",link:function(t,e,n,r){var i=r[0];angular.extend(i,r[1]),i.yearpickerInit(),i.refreshView()}}}),angular.module("ui.bootstrap.position",[]).factory("$uibPosition",["$document","$window",function(t,e){var n,r,i={normal:/(auto|scroll)/,hidden:/(auto|scroll|hidden)/},o={auto:/\s?auto?\s?/i,primary:/^(top|bottom|left|right)$/,secondary:/^(top|bottom|left|right|center)$/,vertical:/^(top|bottom)$/},a=/(HTML|BODY)/;return{getRawNode:function(t){return t.nodeName?t:t[0]||t},parseStyle:function(t){return t=parseFloat(t),isFinite(t)?t:0},offsetParent:function(n){function r(t){return"static"===(e.getComputedStyle(t).position||"static")}n=this.getRawNode(n);for(var i=n.offsetParent||t[0].documentElement;i&&i!==t[0].documentElement&&r(i);)i=i.offsetParent;return i||t[0].documentElement},scrollbarWidth:function(i){if(i){if(angular.isUndefined(r)){var o=t.find("body");o.addClass("uib-position-body-scrollbar-measure"),r=e.innerWidth-o[0].clientWidth,r=isFinite(r)?r:0,o.removeClass("uib-position-body-scrollbar-measure")}return r}if(angular.isUndefined(n)){var a=angular.element('<div class="uib-position-scrollbar-measure"></div>');t.find("body").append(a),n=a[0].offsetWidth-a[0].clientWidth,n=isFinite(n)?n:0,a.remove()}return n},scrollbarPadding:function(t){t=this.getRawNode(t);var n=e.getComputedStyle(t),r=this.parseStyle(n.paddingRight),i=this.parseStyle(n.paddingBottom),o=this.scrollParent(t,!1,!0),s=this.scrollbarWidth(a.test(o.tagName));return{scrollbarWidth:s,widthOverflow:o.scrollWidth>o.clientWidth,right:r+s,originalRight:r,heightOverflow:o.scrollHeight>o.clientHeight,bottom:i+s,originalBottom:i}},isScrollable:function(t,n){t=this.getRawNode(t);var r=n?i.hidden:i.normal,o=e.getComputedStyle(t);return r.test(o.overflow+o.overflowY+o.overflowX)},scrollParent:function(n,r,o){n=this.getRawNode(n);var a=r?i.hidden:i.normal,s=t[0].documentElement,u=e.getComputedStyle(n);if(o&&a.test(u.overflow+u.overflowY+u.overflowX))return n;var l="absolute"===u.position,c=n.parentElement||s;if(c===s||"fixed"===u.position)return s;for(;c.parentElement&&c!==s;){var f=e.getComputedStyle(c);if(l&&"static"!==f.position&&(l=!1),!l&&a.test(f.overflow+f.overflowY+f.overflowX))break;c=c.parentElement}return c},position:function(n,r){n=this.getRawNode(n);var i=this.offset(n);if(r){var o=e.getComputedStyle(n);i.top-=this.parseStyle(o.marginTop),i.left-=this.parseStyle(o.marginLeft)}var a=this.offsetParent(n),s={top:0,left:0};return a!==t[0].documentElement&&(s=this.offset(a),s.top+=a.clientTop-a.scrollTop,s.left+=a.clientLeft-a.scrollLeft),{width:Math.round(angular.isNumber(i.width)?i.width:n.offsetWidth),height:Math.round(angular.isNumber(i.height)?i.height:n.offsetHeight),top:Math.round(i.top-s.top),left:Math.round(i.left-s.left)}},offset:function(n){n=this.getRawNode(n);var r=n.getBoundingClientRect();return{width:Math.round(angular.isNumber(r.width)?r.width:n.offsetWidth),height:Math.round(angular.isNumber(r.height)?r.height:n.offsetHeight),top:Math.round(r.top+(e.pageYOffset||t[0].documentElement.scrollTop)),left:Math.round(r.left+(e.pageXOffset||t[0].documentElement.scrollLeft))}},viewportOffset:function(n,r,i){n=this.getRawNode(n),i=i!==!1;var o=n.getBoundingClientRect(),a={top:0,left:0,bottom:0,right:0},s=r?t[0].documentElement:this.scrollParent(n),u=s.getBoundingClientRect();if(a.top=u.top+s.clientTop,a.left=u.left+s.clientLeft,s===t[0].documentElement&&(a.top+=e.pageYOffset,a.left+=e.pageXOffset),a.bottom=a.top+s.clientHeight,a.right=a.left+s.clientWidth,i){var l=e.getComputedStyle(s);a.top+=this.parseStyle(l.paddingTop),a.bottom-=this.parseStyle(l.paddingBottom),a.left+=this.parseStyle(l.paddingLeft),a.right-=this.parseStyle(l.paddingRight)}return{top:Math.round(o.top-a.top),bottom:Math.round(a.bottom-o.bottom),left:Math.round(o.left-a.left),right:Math.round(a.right-o.right)}},parsePlacement:function(t){var e=o.auto.test(t);return e&&(t=t.replace(o.auto,"")),t=t.split("-"),t[0]=t[0]||"top",o.primary.test(t[0])||(t[0]="top"),t[1]=t[1]||"center",o.secondary.test(t[1])||(t[1]="center"),e?t[2]=!0:t[2]=!1,t},positionElements:function(t,n,r,i){t=this.getRawNode(t),n=this.getRawNode(n);var a=angular.isDefined(n.offsetWidth)?n.offsetWidth:n.prop("offsetWidth"),s=angular.isDefined(n.offsetHeight)?n.offsetHeight:n.prop("offsetHeight");r=this.parsePlacement(r);var u=i?this.offset(t):this.position(t),l={top:0,left:0,placement:""};if(r[2]){var c=this.viewportOffset(t,i),f=e.getComputedStyle(n),p={width:a+Math.round(Math.abs(this.parseStyle(f.marginLeft)+this.parseStyle(f.marginRight))),height:s+Math.round(Math.abs(this.parseStyle(f.marginTop)+this.parseStyle(f.marginBottom)))};if(r[0]="top"===r[0]&&p.height>c.top&&p.height<=c.bottom?"bottom":"bottom"===r[0]&&p.height>c.bottom&&p.height<=c.top?"top":"left"===r[0]&&p.width>c.left&&p.width<=c.right?"right":"right"===r[0]&&p.width>c.right&&p.width<=c.left?"left":r[0],r[1]="top"===r[1]&&p.height-u.height>c.bottom&&p.height-u.height<=c.top?"bottom":"bottom"===r[1]&&p.height-u.height>c.top&&p.height-u.height<=c.bottom?"top":"left"===r[1]&&p.width-u.width>c.right&&p.width-u.width<=c.left?"right":"right"===r[1]&&p.width-u.width>c.left&&p.width-u.width<=c.right?"left":r[1],"center"===r[1])if(o.vertical.test(r[0])){var h=u.width/2-a/2;c.left+h<0&&p.width-u.width<=c.right?r[1]="left":c.right+h<0&&p.width-u.width<=c.left&&(r[1]="right")}else{var d=u.height/2-p.height/2;c.top+d<0&&p.height-u.height<=c.bottom?r[1]="top":c.bottom+d<0&&p.height-u.height<=c.top&&(r[1]="bottom")}}switch(r[0]){case"top":l.top=u.top-s;break;case"bottom":l.top=u.top+u.height;break;case"left":l.left=u.left-a;break;case"right":l.left=u.left+u.width}switch(r[1]){case"top":l.top=u.top;break;case"bottom":l.top=u.top+u.height-s;break;case"left":l.left=u.left;break;case"right":l.left=u.left+u.width-a;break;case"center":o.vertical.test(r[0])?l.left=u.left+u.width/2-a/2:l.top=u.top+u.height/2-s/2}return l.top=Math.round(l.top),l.left=Math.round(l.left),l.placement="center"===r[1]?r[0]:r[0]+"-"+r[1],l},adjustTop:function(t,e,n,r){if(t.indexOf("top")!==-1&&n!==r)return{top:e.top-r+"px"}},positionArrow:function(t,n){t=this.getRawNode(t);var r=t.querySelector(".tooltip-inner, .popover-inner");if(r){var i=angular.element(r).hasClass("tooltip-inner"),a=i?t.querySelector(".tooltip-arrow"):t.querySelector(".arrow");if(a){var s={top:"",bottom:"",left:"",right:""};if(n=this.parsePlacement(n),"center"===n[1])return void angular.element(a).css(s);var u="border-"+n[0]+"-width",l=e.getComputedStyle(a)[u],c="border-";c+=o.vertical.test(n[0])?n[0]+"-"+n[1]:n[1]+"-"+n[0],c+="-radius";var f=e.getComputedStyle(i?r:t)[c];switch(n[0]){case"top":s.bottom=i?"0":"-"+l;break;case"bottom":s.top=i?"0":"-"+l;break;case"left":s.right=i?"0":"-"+l;break;case"right":s.left=i?"0":"-"+l}s[n[1]]=f,angular.element(a).css(s)}}}}}]),angular.module("ui.bootstrap.datepickerPopup",["ui.bootstrap.datepicker","ui.bootstrap.position"]).value("$datepickerPopupLiteralWarning",!0).constant("uibDatepickerPopupConfig",{altInputFormats:[],appendToBody:!1,clearText:"Clear",closeOnDateSelection:!0,closeText:"Done",currentText:"Today",datepickerPopup:"yyyy-MM-dd",datepickerPopupTemplateUrl:"uib/template/datepickerPopup/popup.html",datepickerTemplateUrl:"uib/template/datepicker/datepicker.html",html5Types:{date:"yyyy-MM-dd","datetime-local":"yyyy-MM-ddTHH:mm:ss.sss",month:"yyyy-MM"},onOpenFocus:!0,showButtonBar:!0,placement:"auto bottom-left"}).controller("UibDatepickerPopupController",["$scope","$element","$attrs","$compile","$log","$parse","$window","$document","$rootScope","$uibPosition","dateFilter","uibDateParser","uibDatepickerPopupConfig","$timeout","uibDatepickerConfig","$datepickerPopupLiteralWarning",function(t,e,n,r,i,o,a,s,u,l,c,f,p,h,d,v){function g(e){var n=f.parse(e,x,t.date);if(isNaN(n))for(var r=0;r<P.length;r++)if(n=f.parse(e,P[r],t.date),!isNaN(n))return n;return n}function m(t){if(angular.isNumber(t)&&(t=new Date(t)),!t)return null;if(angular.isDate(t)&&!isNaN(t))return t;if(angular.isString(t)){var e=g(t);if(!isNaN(e))return f.toTimezone(e,O.timezone)}return A.$options&&A.$options.allowInvalid?t:void 0}function y(t,e){var r=t||e;return!n.ngRequired&&!r||(angular.isNumber(r)&&(r=new Date(r)),!r||(!(!angular.isDate(r)||isNaN(r))||!!angular.isString(r)&&!isNaN(g(r))))}function b(n){if(t.isOpen||!t.disabled){var r=N[0],i=e[0].contains(n.target),o=void 0!==r.contains&&r.contains(n.target);!t.isOpen||i||o||t.$apply(function(){t.isOpen=!1})}}function $(n){27===n.which&&t.isOpen?(n.preventDefault(),n.stopPropagation(),t.$apply(function(){t.isOpen=!1}),e[0].focus()):40!==n.which||t.isOpen||(n.preventDefault(),n.stopPropagation(),t.$apply(function(){t.isOpen=!0}))}function w(){if(t.isOpen){var r=angular.element(N[0].querySelector(".uib-datepicker-popup")),i=n.popupPlacement?n.popupPlacement:p.placement,o=l.positionElements(e,r,i,C);r.css({top:o.top+"px",left:o.left+"px"}),r.hasClass("uib-position-measure")&&r.removeClass("uib-position-measure")}}var x,_,C,k,S,T,E,M,D,A,O,N,P,R=!1,I=[];this.init=function(i){if(A=i,O=angular.isObject(i.$options)?i.$options:{timezone:null},_=angular.isDefined(n.closeOnDateSelection)?t.$parent.$eval(n.closeOnDateSelection):p.closeOnDateSelection,C=angular.isDefined(n.datepickerAppendToBody)?t.$parent.$eval(n.datepickerAppendToBody):p.appendToBody,k=angular.isDefined(n.onOpenFocus)?t.$parent.$eval(n.onOpenFocus):p.onOpenFocus,S=angular.isDefined(n.datepickerPopupTemplateUrl)?n.datepickerPopupTemplateUrl:p.datepickerPopupTemplateUrl,T=angular.isDefined(n.datepickerTemplateUrl)?n.datepickerTemplateUrl:p.datepickerTemplateUrl,P=angular.isDefined(n.altInputFormats)?t.$parent.$eval(n.altInputFormats):p.altInputFormats,t.showButtonBar=angular.isDefined(n.showButtonBar)?t.$parent.$eval(n.showButtonBar):p.showButtonBar,p.html5Types[n.type]?(x=p.html5Types[n.type],R=!0):(x=n.uibDatepickerPopup||p.datepickerPopup,n.$observe("uibDatepickerPopup",function(t,e){var n=t||p.datepickerPopup;if(n!==x&&(x=n,A.$modelValue=null,!x))throw new Error("uibDatepickerPopup must have a date format specified.")})),!x)throw new Error("uibDatepickerPopup must have a date format specified.");if(R&&n.uibDatepickerPopup)throw new Error("HTML5 date input types do not support custom formats.");E=angular.element("<div uib-datepicker-popup-wrap><div uib-datepicker></div></div>"),E.attr({"ng-model":"date","ng-change":"dateSelection(date)","template-url":S}),M=angular.element(E.children()[0]),M.attr("template-url",T),t.datepickerOptions||(t.datepickerOptions={}),R&&"month"===n.type&&(t.datepickerOptions.datepickerMode="month",t.datepickerOptions.minMode="month"),M.attr("datepicker-options","datepickerOptions"),R?A.$formatters.push(function(e){return t.date=f.fromTimezone(e,O.timezone),e}):(A.$$parserName="date",A.$validators.date=y,A.$parsers.unshift(m),A.$formatters.push(function(e){return A.$isEmpty(e)?(t.date=e,e):(angular.isNumber(e)&&(e=new Date(e)),t.date=f.fromTimezone(e,O.timezone),f.filter(t.date,x))})),A.$viewChangeListeners.push(function(){t.date=g(A.$viewValue)}),e.on("keydown",$),N=r(E)(t),E.remove(),C?s.find("body").append(N):e.after(N),t.$on("$destroy",function(){for(t.isOpen===!0&&(u.$$phase||t.$apply(function(){t.isOpen=!1})),N.remove(),e.off("keydown",$),s.off("click",b),D&&D.off("scroll",w),angular.element(a).off("resize",w);I.length;)I.shift()()})},t.getText=function(e){return t[e+"Text"]||p[e+"Text"]},t.isDisabled=function(e){"today"===e&&(e=f.fromTimezone(new Date,O.timezone));var n={};return angular.forEach(["minDate","maxDate"],function(e){t.datepickerOptions[e]?angular.isDate(t.datepickerOptions[e])?n[e]=new Date(t.datepickerOptions[e]):(v&&i.warn("Literal date support has been deprecated, please switch to date object usage"),n[e]=new Date(c(t.datepickerOptions[e],"medium"))):n[e]=null}),t.datepickerOptions&&n.minDate&&t.compare(e,n.minDate)<0||n.maxDate&&t.compare(e,n.maxDate)>0},t.compare=function(t,e){return new Date(t.getFullYear(),t.getMonth(),t.getDate())-new Date(e.getFullYear(),e.getMonth(),e.getDate())},t.dateSelection=function(n){t.date=n;var r=t.date?f.filter(t.date,x):null;e.val(r),A.$setViewValue(r),_&&(t.isOpen=!1,e[0].focus())},t.keydown=function(n){27===n.which&&(n.stopPropagation(),t.isOpen=!1,e[0].focus())},t.select=function(e,n){if(n.stopPropagation(),"today"===e){var r=new Date;angular.isDate(t.date)?(e=new Date(t.date),e.setFullYear(r.getFullYear(),r.getMonth(),r.getDate())):(e=f.fromTimezone(r,O.timezone),e.setHours(0,0,0,0))}t.dateSelection(e)},t.close=function(n){n.stopPropagation(),t.isOpen=!1,e[0].focus()},t.disabled=angular.isDefined(n.disabled)||!1,n.ngDisabled&&I.push(t.$parent.$watch(o(n.ngDisabled),function(e){t.disabled=e})),t.$watch("isOpen",function(r){r?t.disabled?t.isOpen=!1:h(function(){w(),k&&t.$broadcast("uib:datepicker.focus"),s.on("click",b);var r=n.popupPlacement?n.popupPlacement:p.placement;C||l.parsePlacement(r)[2]?(D=D||angular.element(l.scrollParent(e)),D&&D.on("scroll",w)):D=null,angular.element(a).on("resize",w)},0,!1):(s.off("click",b),D&&D.off("scroll",w),angular.element(a).off("resize",w))}),t.$on("uib:datepicker.mode",function(){h(w,0,!1)})}]).directive("uibDatepickerPopup",function(){return{require:["ngModel","uibDatepickerPopup"],controller:"UibDatepickerPopupController",scope:{datepickerOptions:"=?",isOpen:"=?",currentText:"@",clearText:"@",closeText:"@"},link:function(t,e,n,r){var i=r[0],o=r[1];o.init(i)}}}).directive("uibDatepickerPopupWrap",function(){return{restrict:"A",transclude:!0,templateUrl:function(t,e){return e.templateUrl||"uib/template/datepickerPopup/popup.html"}}}),angular.module("ui.bootstrap.debounce",[]).factory("$$debounce",["$timeout",function(t){return function(e,n){var r;return function(){var i=this,o=Array.prototype.slice.call(arguments);r&&t.cancel(r),r=t(function(){e.apply(i,o)},n)}}}]),angular.module("ui.bootstrap.multiMap",[]).factory("$$multiMap",function(){return{createNew:function(){var t={};return{entries:function(){return Object.keys(t).map(function(e){return{key:e,value:t[e]}})},get:function(e){return t[e]},hasKey:function(e){return!!t[e]},keys:function(){return Object.keys(t)},put:function(e,n){t[e]||(t[e]=[]),t[e].push(n)},remove:function(e,n){var r=t[e];if(r){var i=r.indexOf(n);i!==-1&&r.splice(i,1),r.length||delete t[e]}}}}}}),angular.module("ui.bootstrap.dropdown",["ui.bootstrap.multiMap","ui.bootstrap.position"]).constant("uibDropdownConfig",{appendToOpenClass:"uib-dropdown-open",openClass:"open"}).service("uibDropdownService",["$document","$rootScope","$$multiMap",function(t,e,n){var r=null,i=n.createNew();this.isOnlyOpen=function(t,e){var n=i.get(e);if(n){var r=n.reduce(function(e,n){return n.scope===t?n:e},{});if(r)return 1===n.length}return!1},this.open=function(e,n,a){if(r||t.on("click",o),r&&r!==e&&(r.isOpen=!1),r=e,a){var s=i.get(a);if(s){var u=s.map(function(t){return t.scope});u.indexOf(e)===-1&&i.put(a,{scope:e})}else i.put(a,{scope:e})}},this.close=function(e,n,a){if(r===e&&(t.off("click",o),t.off("keydown",this.keybindFilter),r=null),a){var s=i.get(a);if(s){var u=s.reduce(function(t,n){return n.scope===e?n:t},{});u&&i.remove(a,u)}}};var o=function(t){if(r&&!(t&&"disabled"===r.getAutoClose()||t&&3===t.which)){var n=r.getToggleElement();if(!(t&&n&&n[0].contains(t.target))){var i=r.getDropdownElement();t&&"outsideClick"===r.getAutoClose()&&i&&i[0].contains(t.target)||(r.focusToggleElement(),r.isOpen=!1,e.$$phase||r.$apply())}}};this.keybindFilter=function(t){if(r){var e=r.getDropdownElement(),n=r.getToggleElement(),i=e&&e[0].contains(t.target),a=n&&n[0].contains(t.target);27===t.which?(t.stopPropagation(),r.focusToggleElement(),o()):r.isKeynavEnabled()&&[38,40].indexOf(t.which)!==-1&&r.isOpen&&(i||a)&&(t.preventDefault(),t.stopPropagation(),r.focusDropdownEntry(t.which))}}}]).controller("UibDropdownController",["$scope","$element","$attrs","$parse","uibDropdownConfig","uibDropdownService","$animate","$uibPosition","$document","$compile","$templateRequest",function(t,e,n,r,i,o,a,s,u,l,c){var f,p,h=this,d=t.$new(),v=i.appendToOpenClass,g=i.openClass,m=angular.noop,y=n.onToggle?r(n.onToggle):angular.noop,b=!1,$=null,w=!1,x=u.find("body");e.addClass("dropdown"),this.init=function(){if(n.isOpen&&(p=r(n.isOpen),m=p.assign,t.$watch(p,function(t){d.isOpen=!!t})),angular.isDefined(n.dropdownAppendTo)){var i=r(n.dropdownAppendTo)(d);i&&($=angular.element(i))}b=angular.isDefined(n.dropdownAppendToBody),w=angular.isDefined(n.keyboardNav),b&&!$&&($=x),$&&h.dropdownMenu&&($.append(h.dropdownMenu),e.on("$destroy",function(){h.dropdownMenu.remove()}))},this.toggle=function(t){return d.isOpen=arguments.length?!!t:!d.isOpen,angular.isFunction(m)&&m(d,d.isOpen),d.isOpen},this.isOpen=function(){return d.isOpen},d.getToggleElement=function(){return h.toggleElement},d.getAutoClose=function(){return n.autoClose||"always"},d.getElement=function(){return e},d.isKeynavEnabled=function(){return w},d.focusDropdownEntry=function(t){var n=h.dropdownMenu?angular.element(h.dropdownMenu).find("a"):e.find("ul").eq(0).find("a");switch(t){case 40:angular.isNumber(h.selectedOption)?h.selectedOption=h.selectedOption===n.length-1?h.selectedOption:h.selectedOption+1:h.selectedOption=0;break;case 38:angular.isNumber(h.selectedOption)?h.selectedOption=0===h.selectedOption?0:h.selectedOption-1:h.selectedOption=n.length-1}n[h.selectedOption].focus()},d.getDropdownElement=function(){return h.dropdownMenu},d.focusToggleElement=function(){h.toggleElement&&h.toggleElement[0].focus()},d.$watch("isOpen",function(n,r){if($&&h.dropdownMenu){var i,p,w,x=s.positionElements(e,h.dropdownMenu,"bottom-left",!0),_=0;if(i={top:x.top+"px",display:n?"block":"none"},p=h.dropdownMenu.hasClass("dropdown-menu-right"),p?(i.left="auto",w=s.scrollbarPadding($),w.heightOverflow&&w.scrollbarWidth&&(_=w.scrollbarWidth),i.right=window.innerWidth-_-(x.left+e.prop("offsetWidth"))+"px"):(i.left=x.left+"px",i.right="auto"),!b){var C=s.offset($);i.top=x.top-C.top+"px",p?i.right=window.innerWidth-(x.left-C.left+e.prop("offsetWidth"))+"px":i.left=x.left-C.left+"px"}h.dropdownMenu.css(i)}var k=$?$:e,S=$?v:g,T=k.hasClass(S),E=o.isOnlyOpen(t,$);if(T===!n){var M;M=$?E?"removeClass":"addClass":n?"addClass":"removeClass",a[M](k,S).then(function(){angular.isDefined(n)&&n!==r&&y(t,{open:!!n})})}if(n)h.dropdownMenuTemplateUrl?c(h.dropdownMenuTemplateUrl).then(function(t){f=d.$new(),l(t.trim())(f,function(t){var e=t;h.dropdownMenu.replaceWith(e),h.dropdownMenu=e,u.on("keydown",o.keybindFilter)})}):u.on("keydown",o.keybindFilter),d.focusToggleElement(),o.open(d,e,$);else{if(o.close(d,e,$),h.dropdownMenuTemplateUrl){f&&f.$destroy();var D=angular.element('<ul class="dropdown-menu"></ul>');h.dropdownMenu.replaceWith(D),h.dropdownMenu=D}h.selectedOption=null}angular.isFunction(m)&&m(t,n)})}]).directive("uibDropdown",function(){return{controller:"UibDropdownController",link:function(t,e,n,r){r.init()}}}).directive("uibDropdownMenu",function(){return{restrict:"A",require:"?^uibDropdown",link:function(t,e,n,r){if(r&&!angular.isDefined(n.dropdownNested)){e.addClass("dropdown-menu");var i=n.templateUrl;i&&(r.dropdownMenuTemplateUrl=i),r.dropdownMenu||(r.dropdownMenu=e)}}}}).directive("uibDropdownToggle",function(){return{require:"?^uibDropdown",link:function(t,e,n,r){if(r){e.addClass("dropdown-toggle"),r.toggleElement=e;var i=function(i){i.preventDefault(),e.hasClass("disabled")||n.disabled||t.$apply(function(){r.toggle()})};e.on("click",i),e.attr({"aria-haspopup":!0,"aria-expanded":!1}),t.$watch(r.isOpen,function(t){e.attr("aria-expanded",!!t)}),t.$on("$destroy",function(){e.off("click",i)})}}}}),angular.module("ui.bootstrap.stackedMap",[]).factory("$$stackedMap",function(){return{createNew:function(){var t=[];return{add:function(e,n){t.push({key:e,value:n})},get:function(e){for(var n=0;n<t.length;n++)if(e===t[n].key)return t[n]},keys:function(){for(var e=[],n=0;n<t.length;n++)e.push(t[n].key);return e},top:function(){return t[t.length-1]},remove:function(e){for(var n=-1,r=0;r<t.length;r++)if(e===t[r].key){n=r;break}return t.splice(n,1)[0]},removeTop:function(){return t.pop()},length:function(){return t.length}}}}}),angular.module("ui.bootstrap.modal",["ui.bootstrap.multiMap","ui.bootstrap.stackedMap","ui.bootstrap.position"]).provider("$uibResolve",function(){var t=this;this.resolver=null,this.setResolver=function(t){this.resolver=t},this.$get=["$injector","$q",function(e,n){var r=t.resolver?e.get(t.resolver):null;return{resolve:function(t,i,o,a){if(r)return r.resolve(t,i,o,a);var s=[];return angular.forEach(t,function(t){angular.isFunction(t)||angular.isArray(t)?s.push(n.resolve(e.invoke(t))):angular.isString(t)?s.push(n.resolve(e.get(t))):s.push(n.resolve(t))}),n.all(s).then(function(e){var n={},r=0;return angular.forEach(t,function(t,i){n[i]=e[r++]}),n})}}}]}).directive("uibModalBackdrop",["$animate","$injector","$uibModalStack",function(t,e,n){function r(e,r,i){i.modalInClass&&(t.addClass(r,i.modalInClass),e.$on(n.NOW_CLOSING_EVENT,function(n,o){var a=o();e.modalOptions.animation?t.removeClass(r,i.modalInClass).then(a):a()}))}return{restrict:"A",compile:function(t,e){return t.addClass(e.backdropClass),r}}}]).directive("uibModalWindow",["$uibModalStack","$q","$animateCss","$document",function(t,e,n,r){return{scope:{index:"@"},restrict:"A",transclude:!0,templateUrl:function(t,e){return e.templateUrl||"uib/template/modal/window.html"},link:function(i,o,a){o.addClass(a.windowTopClass||""),i.size=a.size,i.close=function(e){var n=t.getTop();n&&n.value.backdrop&&"static"!==n.value.backdrop&&e.target===e.currentTarget&&(e.preventDefault(),e.stopPropagation(),t.dismiss(n.key,"backdrop click"))},o.on("click",i.close),i.$isRendered=!0;var s=e.defer();i.$$postDigest(function(){s.resolve()}),s.promise.then(function(){var s=null;a.modalInClass&&(s=n(o,{addClass:a.modalInClass}).start(),i.$on(t.NOW_CLOSING_EVENT,function(t,e){var r=e();n(o,{removeClass:a.modalInClass}).start().then(r)})),e.when(s).then(function(){var e=t.getTop();if(e&&t.modalRendered(e.key),!r[0].activeElement||!o[0].contains(r[0].activeElement)){var n=o[0].querySelector("[autofocus]");n?n.focus():o[0].focus()}})})}}}]).directive("uibModalAnimationClass",function(){return{compile:function(t,e){e.modalAnimation&&t.addClass(e.uibModalAnimationClass)}}}).directive("uibModalTransclude",["$animate",function(t){return{link:function(e,n,r,i,o){o(e.$parent,function(e){n.empty(),t.enter(e,n)})}}}]).factory("$uibModalStack",["$animate","$animateCss","$document","$compile","$rootScope","$q","$$multiMap","$$stackedMap","$uibPosition",function(t,e,n,r,i,o,a,s,u){function l(t){var e="-";return t.replace(D,function(t,n){return(n?e:"")+t.toLowerCase()})}function c(t){return!!(t.offsetWidth||t.offsetHeight||t.getClientRects().length)}function f(){for(var t=-1,e=_.keys(),n=0;n<e.length;n++)_.get(e[n]).value.backdrop&&(t=n);return t>-1&&t<S&&(t=S),t}function p(t,e){var n=_.get(t).value,r=n.appendTo;_.remove(t),T=_.top(),T&&(S=parseInt(T.value.modalDomEl.attr("index"),10)),v(n.modalDomEl,n.modalScope,function(){var e=n.openedClass||x;C.remove(e,t);var i=C.hasKey(e);r.toggleClass(e,i),!i&&w&&w.heightOverflow&&w.scrollbarWidth&&(w.originalRight?r.css({paddingRight:w.originalRight+"px"}):r.css({paddingRight:""}),w=null),h(!0)},n.closedDeferred),d(),e&&e.focus?e.focus():r.focus&&r.focus()}function h(t){var e;_.length()>0&&(e=_.top().value,e.modalDomEl.toggleClass(e.windowTopClass||"",t))}function d(){if(b&&f()===-1){var t=$;v(b,$,function(){t=null}),b=void 0,$=void 0}}function v(e,n,r,i){function a(){a.done||(a.done=!0,t.leave(e).then(function(){r&&r(),e.remove(),i&&i.resolve()}),n.$destroy())}var s,u=null,l=function(){return s||(s=o.defer(),u=s.promise),function(){s.resolve()}};return n.$broadcast(k.NOW_CLOSING_EVENT,l),o.when(u).then(a)}function g(t){if(t.isDefaultPrevented())return t;var e=_.top();if(e)switch(t.which){case 27:e.value.keyboard&&(t.preventDefault(),i.$apply(function(){k.dismiss(e.key,"escape key press")}));break;case 9:var n=k.loadFocusElementList(e),r=!1;t.shiftKey?(k.isFocusInFirstItem(t,n)||k.isModalFocused(t,e))&&(r=k.focusLastFocusableElement(n)):k.isFocusInLastItem(t,n)&&(r=k.focusFirstFocusableElement(n)),r&&(t.preventDefault(),t.stopPropagation())}}function m(t,e,n){return!t.value.modalScope.$broadcast("modal.closing",e,n).defaultPrevented}function y(){Array.prototype.forEach.call(document.querySelectorAll("["+E+"]"),function(t){var e=parseInt(t.getAttribute(E),10),n=e-1;t.setAttribute(E,n),n||(t.removeAttribute(E),t.removeAttribute("aria-hidden"))})}var b,$,w,x="modal-open",_=s.createNew(),C=a.createNew(),k={NOW_CLOSING_EVENT:"modal.stack.now-closing"},S=0,T=null,E="data-bootstrap-modal-aria-hidden-count",M="a[href], area[href], input:not([disabled]):not([tabindex='-1']), button:not([disabled]):not([tabindex='-1']),select:not([disabled]):not([tabindex='-1']), textarea:not([disabled]):not([tabindex='-1']), iframe, object, embed, *[tabindex]:not([tabindex='-1']), *[contenteditable=true]",D=/[A-Z]/g;return i.$watch(f,function(t){$&&($.index=t)}),n.on("keydown",g),i.$on("$destroy",function(){n.off("keydown",g)}),k.open=function(e,o){function a(t){function e(t){var e=t.parent()?t.parent().children():[];return Array.prototype.filter.call(e,function(e){return e!==t[0]})}if(t&&"BODY"!==t[0].tagName)return e(t).forEach(function(t){
var e="true"===t.getAttribute("aria-hidden"),n=parseInt(t.getAttribute(E),10);n||(n=e?1:0),t.setAttribute(E,n+1),t.setAttribute("aria-hidden","true")}),a(t.parent())}var s=n[0].activeElement,c=o.openedClass||x;h(!1),T=_.top(),_.add(e,{deferred:o.deferred,renderDeferred:o.renderDeferred,closedDeferred:o.closedDeferred,modalScope:o.scope,backdrop:o.backdrop,keyboard:o.keyboard,openedClass:o.openedClass,windowTopClass:o.windowTopClass,animation:o.animation,appendTo:o.appendTo}),C.put(c,e);var p=o.appendTo,d=f();if(!p.length)throw new Error("appendTo element not found. Make sure that the element passed is in DOM.");d>=0&&!b&&($=i.$new(!0),$.modalOptions=o,$.index=d,b=angular.element('<div uib-modal-backdrop="modal-backdrop"></div>'),b.attr({class:"modal-backdrop","ng-style":"{'z-index': 1040 + (index && 1 || 0) + index*10}","uib-modal-animation-class":"fade","modal-in-class":"in"}),o.backdropClass&&b.addClass(o.backdropClass),o.animation&&b.attr("modal-animation","true"),r(b)($),t.enter(b,p),u.isScrollable(p)&&(w=u.scrollbarPadding(p),w.heightOverflow&&w.scrollbarWidth&&p.css({paddingRight:w.right+"px"})));var v;o.component?(v=document.createElement(l(o.component.name)),v=angular.element(v),v.attr({resolve:"$resolve","modal-instance":"$uibModalInstance",close:"$close($value)",dismiss:"$dismiss($value)"})):v=o.content,S=T?parseInt(T.value.modalDomEl.attr("index"),10)+1:0;var g=angular.element('<div uib-modal-window="modal-window"></div>');g.attr({class:"modal","template-url":o.windowTemplateUrl,"window-top-class":o.windowTopClass,role:"dialog","aria-labelledby":o.ariaLabelledBy,"aria-describedby":o.ariaDescribedBy,size:o.size,index:S,animate:"animate","ng-style":"{'z-index': 1050 + $$topModalIndex*10, display: 'block'}",tabindex:-1,"uib-modal-animation-class":"fade","modal-in-class":"in"}).append(v),o.windowClass&&g.addClass(o.windowClass),o.animation&&g.attr("modal-animation","true"),p.addClass(c),o.scope&&(o.scope.$$topModalIndex=S),t.enter(r(g)(o.scope),p),_.top().value.modalDomEl=g,_.top().value.modalOpener=s,a(g)},k.close=function(t,e){var n=_.get(t);return y(),n&&m(n,e,!0)?(n.value.modalScope.$$uibDestructionScheduled=!0,n.value.deferred.resolve(e),p(t,n.value.modalOpener),!0):!n},k.dismiss=function(t,e){var n=_.get(t);return y(),n&&m(n,e,!1)?(n.value.modalScope.$$uibDestructionScheduled=!0,n.value.deferred.reject(e),p(t,n.value.modalOpener),!0):!n},k.dismissAll=function(t){for(var e=this.getTop();e&&this.dismiss(e.key,t);)e=this.getTop()},k.getTop=function(){return _.top()},k.modalRendered=function(t){var e=_.get(t);e&&e.value.renderDeferred.resolve()},k.focusFirstFocusableElement=function(t){return t.length>0&&(t[0].focus(),!0)},k.focusLastFocusableElement=function(t){return t.length>0&&(t[t.length-1].focus(),!0)},k.isModalFocused=function(t,e){if(t&&e){var n=e.value.modalDomEl;if(n&&n.length)return(t.target||t.srcElement)===n[0]}return!1},k.isFocusInFirstItem=function(t,e){return e.length>0&&(t.target||t.srcElement)===e[0]},k.isFocusInLastItem=function(t,e){return e.length>0&&(t.target||t.srcElement)===e[e.length-1]},k.loadFocusElementList=function(t){if(t){var e=t.value.modalDomEl;if(e&&e.length){var n=e[0].querySelectorAll(M);return n?Array.prototype.filter.call(n,function(t){return c(t)}):n}}},k}]).provider("$uibModal",function(){var t={options:{animation:!0,backdrop:!0,keyboard:!0},$get:["$rootScope","$q","$document","$templateRequest","$controller","$uibResolve","$uibModalStack",function(e,n,r,i,o,a,s){function u(t){return t.template?n.when(t.template):i(angular.isFunction(t.templateUrl)?t.templateUrl():t.templateUrl)}var l={},c=null;return l.getPromiseChain=function(){return c},l.open=function(i){function l(){return g}var f=n.defer(),p=n.defer(),h=n.defer(),d=n.defer(),v={result:f.promise,opened:p.promise,closed:h.promise,rendered:d.promise,close:function(t){return s.close(v,t)},dismiss:function(t){return s.dismiss(v,t)}};if(i=angular.extend({},t.options,i),i.resolve=i.resolve||{},i.appendTo=i.appendTo||r.find("body").eq(0),!i.component&&!i.template&&!i.templateUrl)throw new Error("One of component or template or templateUrl options is required.");var g;g=i.component?n.when(a.resolve(i.resolve,{},null,null)):n.all([u(i),a.resolve(i.resolve,{},null,null)]);var m;return m=c=n.all([c]).then(l,l).then(function(t){function n(e,n,r,i){e.$scope=a,e.$scope.$resolve={},r?e.$scope.$uibModalInstance=v:e.$uibModalInstance=v;var o=n?t[1]:t;angular.forEach(o,function(t,n){i&&(e[n]=t),e.$scope.$resolve[n]=t})}var r=i.scope||e,a=r.$new();a.$close=v.close,a.$dismiss=v.dismiss,a.$on("$destroy",function(){a.$$uibDestructionScheduled||a.$dismiss("$uibUnscheduledDestruction")});var u,l,c={scope:a,deferred:f,renderDeferred:d,closedDeferred:h,animation:i.animation,backdrop:i.backdrop,keyboard:i.keyboard,backdropClass:i.backdropClass,windowTopClass:i.windowTopClass,windowClass:i.windowClass,windowTemplateUrl:i.windowTemplateUrl,ariaLabelledBy:i.ariaLabelledBy,ariaDescribedBy:i.ariaDescribedBy,size:i.size,openedClass:i.openedClass,appendTo:i.appendTo},g={},m={};i.component?(n(g,!1,!0,!1),g.name=i.component,c.component=g):i.controller&&(n(m,!0,!1,!0),l=o(i.controller,m,!0,i.controllerAs),i.controllerAs&&i.bindToController&&(u=l.instance,u.$close=a.$close,u.$dismiss=a.$dismiss,angular.extend(u,{$resolve:m.$scope.$resolve},r)),u=l(),angular.isFunction(u.$onInit)&&u.$onInit()),i.component||(c.content=t[0]),s.open(v,c),p.resolve(!0)},function(t){p.reject(t),f.reject(t)}).finally(function(){c===m&&(c=null)}),v},l}]};return t}),angular.module("ui.bootstrap.paging",[]).factory("uibPaging",["$parse",function(t){return{create:function(e,n,r){e.setNumPages=r.numPages?t(r.numPages).assign:angular.noop,e.ngModelCtrl={$setViewValue:angular.noop},e._watchers=[],e.init=function(t,i){e.ngModelCtrl=t,e.config=i,t.$render=function(){e.render()},r.itemsPerPage?e._watchers.push(n.$parent.$watch(r.itemsPerPage,function(t){e.itemsPerPage=parseInt(t,10),n.totalPages=e.calculateTotalPages(),e.updatePage()})):e.itemsPerPage=i.itemsPerPage,n.$watch("totalItems",function(t,r){(angular.isDefined(t)||t!==r)&&(n.totalPages=e.calculateTotalPages(),e.updatePage())})},e.calculateTotalPages=function(){var t=e.itemsPerPage<1?1:Math.ceil(n.totalItems/e.itemsPerPage);return Math.max(t||0,1)},e.render=function(){n.page=parseInt(e.ngModelCtrl.$viewValue,10)||1},n.selectPage=function(t,r){r&&r.preventDefault();var i=!n.ngDisabled||!r;i&&n.page!==t&&t>0&&t<=n.totalPages&&(r&&r.target&&r.target.blur(),e.ngModelCtrl.$setViewValue(t),e.ngModelCtrl.$render())},n.getText=function(t){return n[t+"Text"]||e.config[t+"Text"]},n.noPrevious=function(){return 1===n.page},n.noNext=function(){return n.page===n.totalPages},e.updatePage=function(){e.setNumPages(n.$parent,n.totalPages),n.page>n.totalPages?n.selectPage(n.totalPages):e.ngModelCtrl.$render()},n.$on("$destroy",function(){for(;e._watchers.length;)e._watchers.shift()()})}}}]),angular.module("ui.bootstrap.pager",["ui.bootstrap.paging","ui.bootstrap.tabindex"]).controller("UibPagerController",["$scope","$attrs","uibPaging","uibPagerConfig",function(t,e,n,r){t.align=angular.isDefined(e.align)?t.$parent.$eval(e.align):r.align,n.create(this,t,e)}]).constant("uibPagerConfig",{itemsPerPage:10,previousText:"« Previous",nextText:"Next »",align:!0}).directive("uibPager",["uibPagerConfig",function(t){return{scope:{totalItems:"=",previousText:"@",nextText:"@",ngDisabled:"="},require:["uibPager","?ngModel"],restrict:"A",controller:"UibPagerController",controllerAs:"pager",templateUrl:function(t,e){return e.templateUrl||"uib/template/pager/pager.html"},link:function(e,n,r,i){n.addClass("pager");var o=i[0],a=i[1];a&&o.init(a,t)}}}]),angular.module("ui.bootstrap.pagination",["ui.bootstrap.paging","ui.bootstrap.tabindex"]).controller("UibPaginationController",["$scope","$attrs","$parse","uibPaging","uibPaginationConfig",function(t,e,n,r,i){function o(t,e,n){return{number:t,text:e,active:n}}function a(t,e){var n=[],r=1,i=e,a=angular.isDefined(u)&&u<e;a&&(l?(r=Math.max(t-Math.floor(u/2),1),i=r+u-1,i>e&&(i=e,r=i-u+1)):(r=(Math.ceil(t/u)-1)*u+1,i=Math.min(r+u-1,e)));for(var s=r;s<=i;s++){var h=o(s,p(s),s===t);n.push(h)}if(a&&u>0&&(!l||c||f)){if(r>1){if(!f||r>3){var d=o(r-1,"...",!1);n.unshift(d)}if(f){if(3===r){var v=o(2,"2",!1);n.unshift(v)}var g=o(1,"1",!1);n.unshift(g)}}if(i<e){if(!f||i<e-2){var m=o(i+1,"...",!1);n.push(m)}if(f){if(i===e-2){var y=o(e-1,e-1,!1);n.push(y)}var b=o(e,e,!1);n.push(b)}}}return n}var s=this,u=angular.isDefined(e.maxSize)?t.$parent.$eval(e.maxSize):i.maxSize,l=angular.isDefined(e.rotate)?t.$parent.$eval(e.rotate):i.rotate,c=angular.isDefined(e.forceEllipses)?t.$parent.$eval(e.forceEllipses):i.forceEllipses,f=angular.isDefined(e.boundaryLinkNumbers)?t.$parent.$eval(e.boundaryLinkNumbers):i.boundaryLinkNumbers,p=angular.isDefined(e.pageLabel)?function(n){return t.$parent.$eval(e.pageLabel,{$page:n})}:angular.identity;t.boundaryLinks=angular.isDefined(e.boundaryLinks)?t.$parent.$eval(e.boundaryLinks):i.boundaryLinks,t.directionLinks=angular.isDefined(e.directionLinks)?t.$parent.$eval(e.directionLinks):i.directionLinks,r.create(this,t,e),e.maxSize&&s._watchers.push(t.$parent.$watch(n(e.maxSize),function(t){u=parseInt(t,10),s.render()}));var h=this.render;this.render=function(){h(),t.page>0&&t.page<=t.totalPages&&(t.pages=a(t.page,t.totalPages))}}]).constant("uibPaginationConfig",{itemsPerPage:10,boundaryLinks:!1,boundaryLinkNumbers:!1,directionLinks:!0,firstText:"First",previousText:"Previous",nextText:"Next",lastText:"Last",rotate:!0,forceEllipses:!1}).directive("uibPagination",["$parse","uibPaginationConfig",function(t,e){return{scope:{totalItems:"=",firstText:"@",previousText:"@",nextText:"@",lastText:"@",ngDisabled:"="},require:["uibPagination","?ngModel"],restrict:"A",controller:"UibPaginationController",controllerAs:"pagination",templateUrl:function(t,e){return e.templateUrl||"uib/template/pagination/pagination.html"},link:function(t,n,r,i){n.addClass("pagination");var o=i[0],a=i[1];a&&o.init(a,e)}}}]),angular.module("ui.bootstrap.tooltip",["ui.bootstrap.position","ui.bootstrap.stackedMap"]).provider("$uibTooltip",function(){function t(t){var e=/[A-Z]/g,n="-";return t.replace(e,function(t,e){return(e?n:"")+t.toLowerCase()})}var e={placement:"top",placementClassPrefix:"",animation:!0,popupDelay:0,popupCloseDelay:0,useContentExp:!1},n={mouseenter:"mouseleave",click:"click",outsideClick:"outsideClick",focus:"blur",none:""},r={};this.options=function(t){angular.extend(r,t)},this.setTriggers=function(t){angular.extend(n,t)},this.$get=["$window","$compile","$timeout","$document","$uibPosition","$interpolate","$rootScope","$parse","$$stackedMap",function(i,o,a,s,u,l,c,f,p){function h(t){if(27===t.which){var e=d.top();e&&(e.value.close(),e=null)}}var d=p.createNew();return s.on("keyup",h),c.$on("$destroy",function(){s.off("keyup",h)}),function(i,c,p,h){function v(t){var e=(t||h.trigger||p).split(" "),r=e.map(function(t){return n[t]||t});return{show:e,hide:r}}h=angular.extend({},e,r,h);var g=t(i),m=l.startSymbol(),y=l.endSymbol(),b="<div "+g+'-popup uib-title="'+m+"title"+y+'" '+(h.useContentExp?'content-exp="contentExp()" ':'content="'+m+"content"+y+'" ')+'origin-scope="origScope" class="uib-position-measure '+c+'" tooltip-animation-class="fade"uib-tooltip-classes ng-class="{ in: isOpen }" ></div>';return{compile:function(t,e){var n=o(b);return function(t,e,r,o){function l(){V.isOpen?g():p()}function p(){q&&!t.$eval(r[c+"Enable"])||($(),_(),V.popupDelay?O||(O=a(m,V.popupDelay,!1)):m())}function g(){y(),V.popupCloseDelay?N||(N=a(b,V.popupCloseDelay,!1)):b()}function m(){return y(),$(),V.content?(w(),void V.$evalAsync(function(){V.isOpen=!0,C(!0),z()})):angular.noop}function y(){O&&(a.cancel(O),O=null),P&&(a.cancel(P),P=null)}function b(){V&&V.$evalAsync(function(){V&&(V.isOpen=!1,C(!1),V.animation?A||(A=a(x,150,!1)):x())})}function $(){N&&(a.cancel(N),N=null),A&&(a.cancel(A),A=null)}function w(){M||(D=V.$new(),M=n(D,function(t){j?s.find("body").append(t):e.after(t)}),d.add(V,{close:b}),k())}function x(){y(),$(),S(),M&&(M.remove(),M=null,R&&a.cancel(R)),d.remove(V),D&&(D.$destroy(),D=null)}function _(){V.title=r[c+"Title"],H?V.content=H(t):V.content=r[i],V.popupClass=r[c+"Class"],V.placement=angular.isDefined(r[c+"Placement"])?r[c+"Placement"]:h.placement;var e=u.parsePlacement(V.placement);I=e[1]?e[0]+"-"+e[1]:e[0];var n=parseInt(r[c+"PopupDelay"],10),o=parseInt(r[c+"PopupCloseDelay"],10);V.popupDelay=isNaN(n)?h.popupDelay:n,V.popupCloseDelay=isNaN(o)?h.popupCloseDelay:o}function C(e){L&&angular.isFunction(L.assign)&&L.assign(t,e)}function k(){B.length=0,H?(B.push(t.$watch(H,function(t){V.content=t,!t&&V.isOpen&&b()})),B.push(D.$watch(function(){U||(U=!0,D.$$postDigest(function(){U=!1,V&&V.isOpen&&z()}))}))):B.push(r.$observe(i,function(t){V.content=t,!t&&V.isOpen?b():z()})),B.push(r.$observe(c+"Title",function(t){V.title=t,V.isOpen&&z()})),B.push(r.$observe(c+"Placement",function(t){V.placement=t?t:h.placement,V.isOpen&&z()}))}function S(){B.length&&(angular.forEach(B,function(t){t()}),B.length=0)}function T(t){V&&V.isOpen&&M&&(e[0].contains(t.target)||M[0].contains(t.target)||g())}function E(){var n=[],i=[],o=t.$eval(r[c+"Trigger"]);W(),angular.isObject(o)?(Object.keys(o).forEach(function(t){n.push(t),i.push(o[t])}),F={show:n,hide:i}):F=v(o),"none"!==F.show&&F.show.forEach(function(t,n){"outsideClick"===t?(e.on("click",l),s.on("click",T)):t===F.hide[n]?e.on(t,l):t&&(e.on(t,p),e.on(F.hide[n],g)),e.on("keypress",function(t){27===t.which&&g()})})}var M,D,A,O,N,P,R,I,j=!!angular.isDefined(h.appendToBody)&&h.appendToBody,F=v(void 0),q=angular.isDefined(r[c+"Enable"]),V=t.$new(!0),U=!1,L=!!angular.isDefined(r[c+"IsOpen"])&&f(r[c+"IsOpen"]),H=!!h.useContentExp&&f(r[i]),B=[],z=function(){M&&M.html()&&(P||(P=a(function(){var t=u.positionElements(e,M,V.placement,j),n=angular.isDefined(M.offsetHeight)?M.offsetHeight:M.prop("offsetHeight"),r=j?u.offset(e):u.position(e);M.css({top:t.top+"px",left:t.left+"px"});var i=t.placement.split("-");M.hasClass(i[0])||(M.removeClass(I.split("-")[0]),M.addClass(i[0])),M.hasClass(h.placementClassPrefix+t.placement)||(M.removeClass(h.placementClassPrefix+I),M.addClass(h.placementClassPrefix+t.placement)),R=a(function(){var t=angular.isDefined(M.offsetHeight)?M.offsetHeight:M.prop("offsetHeight"),e=u.adjustTop(i,r,n,t);e&&M.css(e),R=null},0,!1),M.hasClass("uib-position-measure")?(u.positionArrow(M,t.placement),M.removeClass("uib-position-measure")):I!==t.placement&&u.positionArrow(M,t.placement),I=t.placement,P=null},0,!1)))};V.origScope=t,V.isOpen=!1,V.contentExp=function(){return V.content},r.$observe("disabled",function(t){t&&y(),t&&V.isOpen&&b()}),L&&t.$watch(L,function(t){V&&!t===V.isOpen&&l()});var W=function(){F.show.forEach(function(t){"outsideClick"===t?e.off("click",l):(e.off(t,p),e.off(t,l))}),F.hide.forEach(function(t){"outsideClick"===t?s.off("click",T):e.off(t,g)})};E();var Y=t.$eval(r[c+"Animation"]);V.animation=angular.isDefined(Y)?!!Y:h.animation;var G,X=c+"AppendToBody";G=X in r&&void 0===r[X]||t.$eval(r[X]),j=angular.isDefined(G)?G:j,t.$on("$destroy",function(){W(),x(),V=null})}}}}}]}).directive("uibTooltipTemplateTransclude",["$animate","$sce","$compile","$templateRequest",function(t,e,n,r){return{link:function(i,o,a){var s,u,l,c=i.$eval(a.tooltipTemplateTranscludeScope),f=0,p=function(){u&&(u.remove(),u=null),s&&(s.$destroy(),s=null),l&&(t.leave(l).then(function(){u=null}),u=l,l=null)};i.$watch(e.parseAsResourceUrl(a.uibTooltipTemplateTransclude),function(e){var a=++f;e?(r(e,!0).then(function(r){if(a===f){var i=c.$new(),u=r,h=n(u)(i,function(e){p(),t.enter(e,o)});s=i,l=h,s.$emit("$includeContentLoaded",e)}},function(){a===f&&(p(),i.$emit("$includeContentError",e))}),i.$emit("$includeContentRequested",e)):p()}),i.$on("$destroy",p)}}}]).directive("uibTooltipClasses",["$uibPosition",function(t){return{restrict:"A",link:function(e,n,r){if(e.placement){var i=t.parsePlacement(e.placement);n.addClass(i[0])}e.popupClass&&n.addClass(e.popupClass),e.animation&&n.addClass(r.tooltipAnimationClass)}}}]).directive("uibTooltipPopup",function(){return{restrict:"A",scope:{content:"@"},templateUrl:"uib/template/tooltip/tooltip-popup.html"}}).directive("uibTooltip",["$uibTooltip",function(t){return t("uibTooltip","tooltip","mouseenter")}]).directive("uibTooltipTemplatePopup",function(){return{restrict:"A",scope:{contentExp:"&",originScope:"&"},templateUrl:"uib/template/tooltip/tooltip-template-popup.html"}}).directive("uibTooltipTemplate",["$uibTooltip",function(t){return t("uibTooltipTemplate","tooltip","mouseenter",{useContentExp:!0})}]).directive("uibTooltipHtmlPopup",function(){return{restrict:"A",scope:{contentExp:"&"},templateUrl:"uib/template/tooltip/tooltip-html-popup.html"}}).directive("uibTooltipHtml",["$uibTooltip",function(t){return t("uibTooltipHtml","tooltip","mouseenter",{useContentExp:!0})}]),angular.module("ui.bootstrap.popover",["ui.bootstrap.tooltip"]).directive("uibPopoverTemplatePopup",function(){return{restrict:"A",scope:{uibTitle:"@",contentExp:"&",originScope:"&"},templateUrl:"uib/template/popover/popover-template.html"}}).directive("uibPopoverTemplate",["$uibTooltip",function(t){return t("uibPopoverTemplate","popover","click",{useContentExp:!0})}]).directive("uibPopoverHtmlPopup",function(){return{restrict:"A",scope:{contentExp:"&",uibTitle:"@"},templateUrl:"uib/template/popover/popover-html.html"}}).directive("uibPopoverHtml",["$uibTooltip",function(t){return t("uibPopoverHtml","popover","click",{useContentExp:!0})}]).directive("uibPopoverPopup",function(){return{restrict:"A",scope:{uibTitle:"@",content:"@"},templateUrl:"uib/template/popover/popover.html"}}).directive("uibPopover",["$uibTooltip",function(t){return t("uibPopover","popover","click")}]),angular.module("ui.bootstrap.progressbar",[]).constant("uibProgressConfig",{animate:!0,max:100}).controller("UibProgressController",["$scope","$attrs","uibProgressConfig",function(t,e,n){function r(){return angular.isDefined(t.maxParam)?t.maxParam:n.max}var i=this,o=angular.isDefined(e.animate)?t.$parent.$eval(e.animate):n.animate;this.bars=[],t.max=r(),this.addBar=function(t,e,n){o||e.css({transition:"none"}),this.bars.push(t),t.max=r(),t.title=n&&angular.isDefined(n.title)?n.title:"progressbar",t.$watch("value",function(e){t.recalculatePercentage()}),t.recalculatePercentage=function(){var e=i.bars.reduce(function(t,e){return e.percent=+(100*e.value/e.max).toFixed(2),t+e.percent},0);e>100&&(t.percent-=e-100)},t.$on("$destroy",function(){e=null,i.removeBar(t)})},this.removeBar=function(t){this.bars.splice(this.bars.indexOf(t),1),this.bars.forEach(function(t){t.recalculatePercentage()})},t.$watch("maxParam",function(t){i.bars.forEach(function(t){t.max=r(),t.recalculatePercentage()})})}]).directive("uibProgress",function(){return{replace:!0,transclude:!0,controller:"UibProgressController",require:"uibProgress",scope:{maxParam:"=?max"},templateUrl:"uib/template/progressbar/progress.html"}}).directive("uibBar",function(){return{replace:!0,transclude:!0,require:"^uibProgress",scope:{value:"=",type:"@"},templateUrl:"uib/template/progressbar/bar.html",link:function(t,e,n,r){r.addBar(t,e,n)}}}).directive("uibProgressbar",function(){return{replace:!0,transclude:!0,controller:"UibProgressController",scope:{value:"=",maxParam:"=?max",type:"@"},templateUrl:"uib/template/progressbar/progressbar.html",link:function(t,e,n,r){r.addBar(t,angular.element(e.children()[0]),{title:n.title})}}}),angular.module("ui.bootstrap.rating",[]).constant("uibRatingConfig",{max:5,stateOn:null,stateOff:null,enableReset:!0,titles:["one","two","three","four","five"]}).controller("UibRatingController",["$scope","$attrs","uibRatingConfig",function(t,e,n){var r={$setViewValue:angular.noop},i=this;this.init=function(i){r=i,r.$render=this.render,r.$formatters.push(function(t){return angular.isNumber(t)&&t<<0!==t&&(t=Math.round(t)),t}),this.stateOn=angular.isDefined(e.stateOn)?t.$parent.$eval(e.stateOn):n.stateOn,this.stateOff=angular.isDefined(e.stateOff)?t.$parent.$eval(e.stateOff):n.stateOff,this.enableReset=angular.isDefined(e.enableReset)?t.$parent.$eval(e.enableReset):n.enableReset;var o=angular.isDefined(e.titles)?t.$parent.$eval(e.titles):n.titles;this.titles=angular.isArray(o)&&o.length>0?o:n.titles;var a=angular.isDefined(e.ratingStates)?t.$parent.$eval(e.ratingStates):new Array(angular.isDefined(e.max)?t.$parent.$eval(e.max):n.max);t.range=this.buildTemplateObjects(a)},this.buildTemplateObjects=function(t){for(var e=0,n=t.length;e<n;e++)t[e]=angular.extend({index:e},{stateOn:this.stateOn,stateOff:this.stateOff,title:this.getTitle(e)},t[e]);return t},this.getTitle=function(t){return t>=this.titles.length?t+1:this.titles[t]},t.rate=function(e){if(!t.readonly&&e>=0&&e<=t.range.length){var n=i.enableReset&&r.$viewValue===e?0:e;r.$setViewValue(n),r.$render()}},t.enter=function(e){t.readonly||(t.value=e),t.onHover({value:e})},t.reset=function(){t.value=r.$viewValue,t.onLeave()},t.onKeydown=function(e){/(37|38|39|40)/.test(e.which)&&(e.preventDefault(),e.stopPropagation(),t.rate(t.value+(38===e.which||39===e.which?1:-1)))},this.render=function(){t.value=r.$viewValue,t.title=i.getTitle(t.value-1)}}]).directive("uibRating",function(){return{require:["uibRating","ngModel"],restrict:"A",scope:{readonly:"=?readOnly",onHover:"&",onLeave:"&"},controller:"UibRatingController",templateUrl:"uib/template/rating/rating.html",link:function(t,e,n,r){var i=r[0],o=r[1];i.init(o)}}}),angular.module("ui.bootstrap.tabs",[]).controller("UibTabsetController",["$scope",function(t){function e(t){for(var e=0;e<r.tabs.length;e++)if(r.tabs[e].index===t)return e}var n,r=this;r.tabs=[],r.select=function(t,o){if(!i){var a=e(n),s=r.tabs[a];if(s){if(s.tab.onDeselect({$event:o,$selectedIndex:t}),o&&o.isDefaultPrevented())return;s.tab.active=!1}var u=r.tabs[t];u?(u.tab.onSelect({$event:o}),u.tab.active=!0,r.active=u.index,n=u.index):!u&&angular.isDefined(n)&&(r.active=null,n=null)}},r.addTab=function(t){if(r.tabs.push({tab:t,index:t.index}),r.tabs.sort(function(t,e){return t.index>e.index?1:t.index<e.index?-1:0}),t.index===r.active||!angular.isDefined(r.active)&&1===r.tabs.length){var n=e(t.index);r.select(n)}},r.removeTab=function(t){for(var e,n=0;n<r.tabs.length;n++)if(r.tabs[n].tab===t){e=n;break}if(r.tabs[e].index===r.active){var i=e===r.tabs.length-1?e-1:e+1%r.tabs.length;r.select(i)}r.tabs.splice(e,1)},t.$watch("tabset.active",function(t){angular.isDefined(t)&&t!==n&&r.select(e(t))});var i;t.$on("$destroy",function(){i=!0})}]).directive("uibTabset",function(){return{transclude:!0,replace:!0,scope:{},bindToController:{active:"=?",type:"@"},controller:"UibTabsetController",controllerAs:"tabset",templateUrl:function(t,e){return e.templateUrl||"uib/template/tabs/tabset.html"},link:function(t,e,n){t.vertical=!!angular.isDefined(n.vertical)&&t.$parent.$eval(n.vertical),t.justified=!!angular.isDefined(n.justified)&&t.$parent.$eval(n.justified)}}}).directive("uibTab",["$parse",function(t){return{require:"^uibTabset",replace:!0,templateUrl:function(t,e){return e.templateUrl||"uib/template/tabs/tab.html"},transclude:!0,scope:{heading:"@",index:"=?",classes:"@?",onSelect:"&select",onDeselect:"&deselect"},controller:function(){},controllerAs:"tab",link:function(e,n,r,i,o){e.disabled=!1,r.disable&&e.$parent.$watch(t(r.disable),function(t){e.disabled=!!t}),angular.isUndefined(r.index)&&(i.tabs&&i.tabs.length?e.index=Math.max.apply(null,i.tabs.map(function(t){return t.index}))+1:e.index=0),angular.isUndefined(r.classes)&&(e.classes=""),e.select=function(t){if(!e.disabled){for(var n,r=0;r<i.tabs.length;r++)if(i.tabs[r].tab===e){n=r;break}i.select(n,t)}},i.addTab(e),e.$on("$destroy",function(){i.removeTab(e)}),e.$transcludeFn=o}}}]).directive("uibTabHeadingTransclude",function(){return{restrict:"A",require:"^uibTab",link:function(t,e){t.$watch("headingElement",function(t){t&&(e.html(""),e.append(t))})}}}).directive("uibTabContentTransclude",function(){function t(t){return t.tagName&&(t.hasAttribute("uib-tab-heading")||t.hasAttribute("data-uib-tab-heading")||t.hasAttribute("x-uib-tab-heading")||"uib-tab-heading"===t.tagName.toLowerCase()||"data-uib-tab-heading"===t.tagName.toLowerCase()||"x-uib-tab-heading"===t.tagName.toLowerCase()||"uib:tab-heading"===t.tagName.toLowerCase())}return{restrict:"A",require:"^uibTabset",link:function(e,n,r){var i=e.$eval(r.uibTabContentTransclude).tab;i.$transcludeFn(i.$parent,function(e){angular.forEach(e,function(e){t(e)?i.headingElement=e:n.append(e)})})}}}),angular.module("ui.bootstrap.timepicker",[]).constant("uibTimepickerConfig",{hourStep:1,minuteStep:1,secondStep:1,showMeridian:!0,showSeconds:!1,meridians:null,readonlyInput:!1,mousewheel:!0,arrowkeys:!0,showSpinners:!0,templateUrl:"uib/template/timepicker/timepicker.html"}).controller("UibTimepickerController",["$scope","$element","$attrs","$parse","$log","$locale","uibTimepickerConfig",function(t,e,n,r,i,o,a){function s(){var e=+t.hours,n=t.showMeridian?e>0&&e<13:e>=0&&e<24;if(n&&""!==t.hours)return t.showMeridian&&(12===e&&(e=0),t.meridian===C[1]&&(e+=12)),e}function u(){var e=+t.minutes,n=e>=0&&e<60;if(n&&""!==t.minutes)return e}function l(){var e=+t.seconds;return e>=0&&e<60?e:void 0}function c(t,e){return null===t?"":angular.isDefined(t)&&t.toString().length<2&&!e?"0"+t:t.toString()}function f(t){p(),_.$setViewValue(new Date(w)),h(t)}function p(){y&&y.$setValidity("hours",!0),b&&b.$setValidity("minutes",!0),$&&$.$setValidity("seconds",!0),_.$setValidity("time",!0),t.invalidHours=!1,t.invalidMinutes=!1,t.invalidSeconds=!1}function h(e){if(_.$modelValue){var n=w.getHours(),r=w.getMinutes(),i=w.getSeconds();t.showMeridian&&(n=0===n||12===n?12:n%12),t.hours="h"===e?n:c(n,!k),"m"!==e&&(t.minutes=c(r)),t.meridian=w.getHours()<12?C[0]:C[1],"s"!==e&&(t.seconds=c(i)),t.meridian=w.getHours()<12?C[0]:C[1]}else t.hours=null,t.minutes=null,t.seconds=null,t.meridian=C[0]}function d(t){w=g(w,t),f()}function v(t,e){return g(t,60*e)}function g(t,e){var n=new Date(t.getTime()+1e3*e),r=new Date(t);return r.setHours(n.getHours(),n.getMinutes(),n.getSeconds()),r}function m(){return(null===t.hours||""===t.hours)&&(null===t.minutes||""===t.minutes)&&(!t.showSeconds||t.showSeconds&&(null===t.seconds||""===t.seconds))}var y,b,$,w=new Date,x=[],_={$setViewValue:angular.noop},C=angular.isDefined(n.meridians)?t.$parent.$eval(n.meridians):a.meridians||o.DATETIME_FORMATS.AMPMS,k=!angular.isDefined(n.padHours)||t.$parent.$eval(n.padHours);t.tabindex=angular.isDefined(n.tabindex)?n.tabindex:0,e.removeAttr("tabindex"),this.init=function(e,r){_=e,_.$render=this.render,_.$formatters.unshift(function(t){return t?new Date(t):null});var i=r.eq(0),o=r.eq(1),s=r.eq(2);y=i.controller("ngModel"),b=o.controller("ngModel"),$=s.controller("ngModel");var u=angular.isDefined(n.mousewheel)?t.$parent.$eval(n.mousewheel):a.mousewheel;u&&this.setupMousewheelEvents(i,o,s);var l=angular.isDefined(n.arrowkeys)?t.$parent.$eval(n.arrowkeys):a.arrowkeys;l&&this.setupArrowkeyEvents(i,o,s),t.readonlyInput=angular.isDefined(n.readonlyInput)?t.$parent.$eval(n.readonlyInput):a.readonlyInput,this.setupInputEvents(i,o,s)};var S=a.hourStep;n.hourStep&&x.push(t.$parent.$watch(r(n.hourStep),function(t){S=+t}));var T=a.minuteStep;n.minuteStep&&x.push(t.$parent.$watch(r(n.minuteStep),function(t){T=+t}));var E;x.push(t.$parent.$watch(r(n.min),function(t){var e=new Date(t);E=isNaN(e)?void 0:e}));var M;x.push(t.$parent.$watch(r(n.max),function(t){var e=new Date(t);M=isNaN(e)?void 0:e}));var D=!1;n.ngDisabled&&x.push(t.$parent.$watch(r(n.ngDisabled),function(t){D=t})),t.noIncrementHours=function(){var t=v(w,60*S);return D||t>M||t<w&&t<E},t.noDecrementHours=function(){var t=v(w,60*-S);return D||t<E||t>w&&t>M},t.noIncrementMinutes=function(){var t=v(w,T);return D||t>M||t<w&&t<E},t.noDecrementMinutes=function(){var t=v(w,-T);return D||t<E||t>w&&t>M},t.noIncrementSeconds=function(){var t=g(w,A);return D||t>M||t<w&&t<E},t.noDecrementSeconds=function(){var t=g(w,-A);return D||t<E||t>w&&t>M},t.noToggleMeridian=function(){return w.getHours()<12?D||v(w,720)>M:D||v(w,-720)<E};var A=a.secondStep;n.secondStep&&x.push(t.$parent.$watch(r(n.secondStep),function(t){A=+t})),t.showSeconds=a.showSeconds,n.showSeconds&&x.push(t.$parent.$watch(r(n.showSeconds),function(e){t.showSeconds=!!e})),t.showMeridian=a.showMeridian,n.showMeridian&&x.push(t.$parent.$watch(r(n.showMeridian),function(e){if(t.showMeridian=!!e,_.$error.time){var n=s(),r=u();angular.isDefined(n)&&angular.isDefined(r)&&(w.setHours(n),f())}else h()})),this.setupMousewheelEvents=function(e,n,r){var i=function(t){t.originalEvent&&(t=t.originalEvent);var e=t.wheelDelta?t.wheelDelta:-t.deltaY;return t.detail||e>0};e.on("mousewheel wheel",function(e){D||t.$apply(i(e)?t.incrementHours():t.decrementHours()),e.preventDefault()}),n.on("mousewheel wheel",function(e){D||t.$apply(i(e)?t.incrementMinutes():t.decrementMinutes()),e.preventDefault()}),r.on("mousewheel wheel",function(e){D||t.$apply(i(e)?t.incrementSeconds():t.decrementSeconds()),e.preventDefault()})},this.setupArrowkeyEvents=function(e,n,r){e.on("keydown",function(e){D||(38===e.which?(e.preventDefault(),t.incrementHours(),t.$apply()):40===e.which&&(e.preventDefault(),t.decrementHours(),t.$apply()))}),n.on("keydown",function(e){D||(38===e.which?(e.preventDefault(),t.incrementMinutes(),t.$apply()):40===e.which&&(e.preventDefault(),t.decrementMinutes(),t.$apply()))}),r.on("keydown",function(e){D||(38===e.which?(e.preventDefault(),t.incrementSeconds(),t.$apply()):40===e.which&&(e.preventDefault(),t.decrementSeconds(),t.$apply()))})},this.setupInputEvents=function(e,n,r){if(t.readonlyInput)return t.updateHours=angular.noop,t.updateMinutes=angular.noop,void(t.updateSeconds=angular.noop);var i=function(e,n,r){_.$setViewValue(null),_.$setValidity("time",!1),angular.isDefined(e)&&(t.invalidHours=e,y&&y.$setValidity("hours",!1)),angular.isDefined(n)&&(t.invalidMinutes=n,b&&b.$setValidity("minutes",!1)),angular.isDefined(r)&&(t.invalidSeconds=r,$&&$.$setValidity("seconds",!1))};t.updateHours=function(){var t=s(),e=u();_.$setDirty(),angular.isDefined(t)&&angular.isDefined(e)?(w.setHours(t),w.setMinutes(e),w<E||w>M?i(!0):f("h")):i(!0)},e.on("blur",function(e){_.$setTouched(),m()?p():null===t.hours||""===t.hours?i(!0):!t.invalidHours&&t.hours<10&&t.$apply(function(){t.hours=c(t.hours,!k)})}),t.updateMinutes=function(){var t=u(),e=s();_.$setDirty(),angular.isDefined(t)&&angular.isDefined(e)?(w.setHours(e),w.setMinutes(t),w<E||w>M?i(void 0,!0):f("m")):i(void 0,!0)},n.on("blur",function(e){_.$setTouched(),m()?p():null===t.minutes?i(void 0,!0):!t.invalidMinutes&&t.minutes<10&&t.$apply(function(){t.minutes=c(t.minutes)})}),t.updateSeconds=function(){var t=l();_.$setDirty(),angular.isDefined(t)?(w.setSeconds(t),f("s")):i(void 0,void 0,!0)},r.on("blur",function(e){m()?p():!t.invalidSeconds&&t.seconds<10&&t.$apply(function(){t.seconds=c(t.seconds)})})},this.render=function(){var e=_.$viewValue;isNaN(e)?(_.$setValidity("time",!1),i.error('Timepicker directive: "ng-model" value must be a Date object, a number of milliseconds since 01.01.1970 or a string representing an RFC2822 or ISO 8601 date.')):(e&&(w=e),w<E||w>M?(_.$setValidity("time",!1),t.invalidHours=!0,t.invalidMinutes=!0):p(),h())},t.showSpinners=angular.isDefined(n.showSpinners)?t.$parent.$eval(n.showSpinners):a.showSpinners,t.incrementHours=function(){t.noIncrementHours()||d(60*S*60)},t.decrementHours=function(){t.noDecrementHours()||d(60*-S*60)},t.incrementMinutes=function(){t.noIncrementMinutes()||d(60*T)},t.decrementMinutes=function(){t.noDecrementMinutes()||d(60*-T)},t.incrementSeconds=function(){t.noIncrementSeconds()||d(A)},t.decrementSeconds=function(){t.noDecrementSeconds()||d(-A)},t.toggleMeridian=function(){var e=u(),n=s();t.noToggleMeridian()||(angular.isDefined(e)&&angular.isDefined(n)?d(720*(w.getHours()<12?60:-60)):t.meridian=t.meridian===C[0]?C[1]:C[0])},t.blur=function(){_.$setTouched()},t.$on("$destroy",function(){for(;x.length;)x.shift()()})}]).directive("uibTimepicker",["uibTimepickerConfig",function(t){
return{require:["uibTimepicker","?^ngModel"],restrict:"A",controller:"UibTimepickerController",controllerAs:"timepicker",scope:{},templateUrl:function(e,n){return n.templateUrl||t.templateUrl},link:function(t,e,n,r){var i=r[0],o=r[1];o&&i.init(o,e.find("input"))}}}]),angular.module("ui.bootstrap.typeahead",["ui.bootstrap.debounce","ui.bootstrap.position"]).factory("uibTypeaheadParser",["$parse",function(t){var e=/^\s*([\s\S]+?)(?:\s+as\s+([\s\S]+?))?\s+for\s+(?:([\$\w][\$\w\d]*))\s+in\s+([\s\S]+?)$/;return{parse:function(n){var r=n.match(e);if(!r)throw new Error('Expected typeahead specification in form of "_modelValue_ (as _label_)? for _item_ in _collection_" but got "'+n+'".');return{itemName:r[3],source:t(r[4]),viewMapper:t(r[2]||r[1]),modelMapper:t(r[1])}}}}]).controller("UibTypeaheadController",["$scope","$element","$attrs","$compile","$parse","$q","$timeout","$document","$window","$rootScope","$$debounce","$uibPosition","uibTypeaheadParser",function(t,e,n,r,i,o,a,s,u,l,c,f,p){function h(){V.moveInProgress||(V.moveInProgress=!0,V.$digest()),J()}function d(){V.position=D?f.offset(e):f.position(e),V.position.top+=e.prop("offsetHeight")}var v,g,m=[9,13,27,38,40],y=200,b=t.$eval(n.typeaheadMinLength);b||0===b||(b=1),t.$watch(n.typeaheadMinLength,function(t){b=t||0===t?t:1});var $=t.$eval(n.typeaheadWaitMs)||0,w=t.$eval(n.typeaheadEditable)!==!1;t.$watch(n.typeaheadEditable,function(t){w=t!==!1});var x,_,C=i(n.typeaheadLoading).assign||angular.noop,k=n.typeaheadShouldSelect?i(n.typeaheadShouldSelect):function(t,e){var n=e.$event;return 13===n.which||9===n.which},S=i(n.typeaheadOnSelect),T=!!angular.isDefined(n.typeaheadSelectOnBlur)&&t.$eval(n.typeaheadSelectOnBlur),E=i(n.typeaheadNoResults).assign||angular.noop,M=n.typeaheadInputFormatter?i(n.typeaheadInputFormatter):void 0,D=!!n.typeaheadAppendToBody&&t.$eval(n.typeaheadAppendToBody),A=n.typeaheadAppendTo?t.$eval(n.typeaheadAppendTo):null,O=t.$eval(n.typeaheadFocusFirst)!==!1,N=!!n.typeaheadSelectOnExact&&t.$eval(n.typeaheadSelectOnExact),P=i(n.typeaheadIsOpen).assign||angular.noop,R=t.$eval(n.typeaheadShowHint)||!1,I=i(n.ngModel),j=i(n.ngModel+"($$$p)"),F=function(e,n){return angular.isFunction(I(t))&&g&&g.$options&&g.$options.getterSetter?j(e,{$$$p:n}):I.assign(e,n)},q=p.parse(n.uibTypeahead),V=t.$new(),U=t.$on("$destroy",function(){V.$destroy()});V.$on("$destroy",U);var L="typeahead-"+V.$id+"-"+Math.floor(1e4*Math.random());e.attr({"aria-autocomplete":"list","aria-expanded":!1,"aria-owns":L});var H,B;R&&(H=angular.element("<div></div>"),H.css("position","relative"),e.after(H),B=e.clone(),B.attr("placeholder",""),B.attr("tabindex","-1"),B.val(""),B.css({position:"absolute",top:"0px",left:"0px","border-color":"transparent","box-shadow":"none",opacity:1,background:"none 0% 0% / auto repeat scroll padding-box border-box rgb(255, 255, 255)",color:"#999"}),e.css({position:"relative","vertical-align":"top","background-color":"transparent"}),B.attr("id")&&B.removeAttr("id"),H.append(B),B.after(e));var z=angular.element("<div uib-typeahead-popup></div>");z.attr({id:L,matches:"matches",active:"activeIdx",select:"select(activeIdx, evt)","move-in-progress":"moveInProgress",query:"query",position:"position","assign-is-open":"assignIsOpen(isOpen)",debounce:"debounceUpdate"}),angular.isDefined(n.typeaheadTemplateUrl)&&z.attr("template-url",n.typeaheadTemplateUrl),angular.isDefined(n.typeaheadPopupTemplateUrl)&&z.attr("popup-template-url",n.typeaheadPopupTemplateUrl);var W=function(){R&&B.val("")},Y=function(){V.matches=[],V.activeIdx=-1,e.attr("aria-expanded",!1),W()},G=function(t){return L+"-option-"+t};V.$watch("activeIdx",function(t){t<0?e.removeAttr("aria-activedescendant"):e.attr("aria-activedescendant",G(t))});var X=function(t,e){return!!(V.matches.length>e&&t)&&t.toUpperCase()===V.matches[e].label.toUpperCase()},K=function(n,r){var i={$viewValue:n};C(t,!0),E(t,!1),o.when(q.source(t,i)).then(function(o){var a=n===v.$viewValue;if(a&&x)if(o&&o.length>0){V.activeIdx=O?0:-1,E(t,!1),V.matches.length=0;for(var s=0;s<o.length;s++)i[q.itemName]=o[s],V.matches.push({id:G(s),label:q.viewMapper(V,i),model:o[s]});if(V.query=n,d(),e.attr("aria-expanded",!0),N&&1===V.matches.length&&X(n,0)&&(angular.isNumber(V.debounceUpdate)||angular.isObject(V.debounceUpdate)?c(function(){V.select(0,r)},angular.isNumber(V.debounceUpdate)?V.debounceUpdate:V.debounceUpdate.default):V.select(0,r)),R){var u=V.matches[0].label;angular.isString(n)&&n.length>0&&u.slice(0,n.length).toUpperCase()===n.toUpperCase()?B.val(n+u.slice(n.length)):B.val("")}}else Y(),E(t,!0);a&&C(t,!1)},function(){Y(),C(t,!1),E(t,!0)})};D&&(angular.element(u).on("resize",h),s.find("body").on("scroll",h));var J=c(function(){V.matches.length&&d(),V.moveInProgress=!1},y);V.moveInProgress=!1,V.query=void 0;var Z,Q=function(t){Z=a(function(){K(t)},$)},tt=function(){Z&&a.cancel(Z)};Y(),V.assignIsOpen=function(e){P(t,e)},V.select=function(r,i){var o,s,u={};_=!0,u[q.itemName]=s=V.matches[r].model,o=q.modelMapper(t,u),F(t,o),v.$setValidity("editable",!0),v.$setValidity("parse",!0),S(t,{$item:s,$model:o,$label:q.viewMapper(t,u),$event:i}),Y(),V.$eval(n.typeaheadFocusOnSelect)!==!1&&a(function(){e[0].focus()},0,!1)},e.on("keydown",function(e){if(0!==V.matches.length&&m.indexOf(e.which)!==-1){var n=k(t,{$event:e});if(V.activeIdx===-1&&n||9===e.which&&e.shiftKey)return Y(),void V.$digest();e.preventDefault();var r;switch(e.which){case 27:e.stopPropagation(),Y(),t.$digest();break;case 38:V.activeIdx=(V.activeIdx>0?V.activeIdx:V.matches.length)-1,V.$digest(),r=z[0].querySelectorAll(".uib-typeahead-match")[V.activeIdx],r.parentNode.scrollTop=r.offsetTop;break;case 40:V.activeIdx=(V.activeIdx+1)%V.matches.length,V.$digest(),r=z[0].querySelectorAll(".uib-typeahead-match")[V.activeIdx],r.parentNode.scrollTop=r.offsetTop;break;default:n&&V.$apply(function(){angular.isNumber(V.debounceUpdate)||angular.isObject(V.debounceUpdate)?c(function(){V.select(V.activeIdx,e)},angular.isNumber(V.debounceUpdate)?V.debounceUpdate:V.debounceUpdate.default):V.select(V.activeIdx,e)})}}}),e.on("focus",function(t){x=!0,0!==b||v.$viewValue||a(function(){K(v.$viewValue,t)},0)}),e.on("blur",function(t){T&&V.matches.length&&V.activeIdx!==-1&&!_&&(_=!0,V.$apply(function(){angular.isObject(V.debounceUpdate)&&angular.isNumber(V.debounceUpdate.blur)?c(function(){V.select(V.activeIdx,t)},V.debounceUpdate.blur):V.select(V.activeIdx,t)})),!w&&v.$error.editable&&(v.$setViewValue(),V.$apply(function(){v.$setValidity("editable",!0),v.$setValidity("parse",!0)}),e.val("")),x=!1,_=!1});var et=function(n){e[0]!==n.target&&3!==n.which&&0!==V.matches.length&&(Y(),l.$$phase||t.$digest())};s.on("click",et),t.$on("$destroy",function(){s.off("click",et),(D||A)&&nt.remove(),D&&(angular.element(u).off("resize",h),s.find("body").off("scroll",h)),z.remove(),R&&H.remove()});var nt=r(z)(V);D?s.find("body").append(nt):A?angular.element(A).eq(0).append(nt):e.after(nt),this.init=function(e,n){v=e,g=n,V.debounceUpdate=v.$options&&i(v.$options.debounce)(t),v.$parsers.unshift(function(e){return x=!0,0===b||e&&e.length>=b?$>0?(tt(),Q(e)):K(e):(C(t,!1),tt(),Y()),w?e:e?void v.$setValidity("editable",!1):(v.$setValidity("editable",!0),null)}),v.$formatters.push(function(e){var n,r,i={};return w||v.$setValidity("editable",!0),M?(i.$model=e,M(t,i)):(i[q.itemName]=e,n=q.viewMapper(t,i),i[q.itemName]=void 0,r=q.viewMapper(t,i),n!==r?n:e)})}}]).directive("uibTypeahead",function(){return{controller:"UibTypeaheadController",require:["ngModel","^?ngModelOptions","uibTypeahead"],link:function(t,e,n,r){r[2].init(r[0],r[1])}}}).directive("uibTypeaheadPopup",["$$debounce",function(t){return{scope:{matches:"=",query:"=",active:"=",position:"&",moveInProgress:"=",select:"&",assignIsOpen:"&",debounce:"&"},replace:!0,templateUrl:function(t,e){return e.popupTemplateUrl||"uib/template/typeahead/typeahead-popup.html"},link:function(e,n,r){e.templateUrl=r.templateUrl,e.isOpen=function(){var t=e.matches.length>0;return e.assignIsOpen({isOpen:t}),t},e.isActive=function(t){return e.active===t},e.selectActive=function(t){e.active=t},e.selectMatch=function(n,r){var i=e.debounce();angular.isNumber(i)||angular.isObject(i)?t(function(){e.select({activeIdx:n,evt:r})},angular.isNumber(i)?i:i.default):e.select({activeIdx:n,evt:r})}}}}]).directive("uibTypeaheadMatch",["$templateRequest","$compile","$parse",function(t,e,n){return{scope:{index:"=",match:"=",query:"="},link:function(r,i,o){var a=n(o.templateUrl)(r.$parent)||"uib/template/typeahead/typeahead-match.html";t(a).then(function(t){var n=angular.element(t.trim());i.replaceWith(n),e(n)(r)})}}}]).filter("uibTypeaheadHighlight",["$sce","$injector","$log",function(t,e,n){function r(t){return t.replace(/([.?*+^$[\]\\(){}|-])/g,"\\$1")}function i(t){return/<.*>/g.test(t)}var o;return o=e.has("$sanitize"),function(e,a){return!o&&i(e)&&n.warn("Unsafe use of typeahead please use ngSanitize"),e=a?(""+e).replace(new RegExp(r(a),"gi"),"<strong>$&</strong>"):e,o||(e=t.trustAsHtml(e)),e}}]),angular.module("uib/template/accordion/accordion-group.html",[]).run(["$templateCache",function(t){t.put("uib/template/accordion/accordion-group.html",'<div role="tab" id="{{::headingId}}" aria-selected="{{isOpen}}" class="panel-heading" ng-keypress="toggleOpen($event)">\n  <h4 class="panel-title">\n    <a role="button" data-toggle="collapse" href aria-expanded="{{isOpen}}" aria-controls="{{::panelId}}" tabindex="0" class="accordion-toggle" ng-click="toggleOpen()" uib-accordion-transclude="heading" ng-disabled="isDisabled" uib-tabindex-toggle><span uib-accordion-header ng-class="{\'text-muted\': isDisabled}">{{heading}}</span></a>\n  </h4>\n</div>\n<div id="{{::panelId}}" aria-labelledby="{{::headingId}}" aria-hidden="{{!isOpen}}" role="tabpanel" class="panel-collapse collapse" uib-collapse="!isOpen">\n  <div class="panel-body" ng-transclude></div>\n</div>\n')}]),angular.module("uib/template/accordion/accordion.html",[]).run(["$templateCache",function(t){t.put("uib/template/accordion/accordion.html",'<div role="tablist" class="panel-group" ng-transclude></div>')}]),angular.module("uib/template/alert/alert.html",[]).run(["$templateCache",function(t){t.put("uib/template/alert/alert.html",'<button ng-show="closeable" type="button" class="close" ng-click="close({$event: $event})">\n  <span aria-hidden="true">&times;</span>\n  <span class="sr-only">Close</span>\n</button>\n<div ng-transclude></div>\n')}]),angular.module("uib/template/carousel/carousel.html",[]).run(["$templateCache",function(t){t.put("uib/template/carousel/carousel.html",'<div class="carousel-inner" ng-transclude></div>\n<a role="button" href class="left carousel-control" ng-click="prev()" ng-class="{ disabled: isPrevDisabled() }" ng-show="slides.length > 1">\n  <span aria-hidden="true" class="glyphicon glyphicon-chevron-left"></span>\n  <span class="sr-only">previous</span>\n</a>\n<a role="button" href class="right carousel-control" ng-click="next()" ng-class="{ disabled: isNextDisabled() }" ng-show="slides.length > 1">\n  <span aria-hidden="true" class="glyphicon glyphicon-chevron-right"></span>\n  <span class="sr-only">next</span>\n</a>\n<ol class="carousel-indicators" ng-show="slides.length > 1">\n  <li ng-repeat="slide in slides | orderBy:indexOfSlide track by $index" ng-class="{ active: isActive(slide) }" ng-click="select(slide)">\n    <span class="sr-only">slide {{ $index + 1 }} of {{ slides.length }}<span ng-if="isActive(slide)">, currently active</span></span>\n  </li>\n</ol>\n')}]),angular.module("uib/template/carousel/slide.html",[]).run(["$templateCache",function(t){t.put("uib/template/carousel/slide.html",'<div class="text-center" ng-transclude></div>\n')}]),angular.module("uib/template/datepicker/datepicker.html",[]).run(["$templateCache",function(t){t.put("uib/template/datepicker/datepicker.html",'<div ng-switch="datepickerMode">\n  <div uib-daypicker ng-switch-when="day" tabindex="0" class="uib-daypicker"></div>\n  <div uib-monthpicker ng-switch-when="month" tabindex="0" class="uib-monthpicker"></div>\n  <div uib-yearpicker ng-switch-when="year" tabindex="0" class="uib-yearpicker"></div>\n</div>\n')}]),angular.module("uib/template/datepicker/day.html",[]).run(["$templateCache",function(t){t.put("uib/template/datepicker/day.html",'<table role="grid" aria-labelledby="{{::uniqueId}}-title" aria-activedescendant="{{activeDateId}}">\n  <thead>\n    <tr>\n      <th><button type="button" class="btn btn-default btn-sm pull-left uib-left" ng-click="move(-1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-left"></i><span class="sr-only">previous</span></button></th>\n      <th colspan="{{::5 + showWeeks}}"><button id="{{::uniqueId}}-title" role="heading" aria-live="assertive" aria-atomic="true" type="button" class="btn btn-default btn-sm uib-title" ng-click="toggleMode()" ng-disabled="datepickerMode === maxMode" tabindex="-1"><strong>{{title}}</strong></button></th>\n      <th><button type="button" class="btn btn-default btn-sm pull-right uib-right" ng-click="move(1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-right"></i><span class="sr-only">next</span></button></th>\n    </tr>\n    <tr>\n      <th ng-if="showWeeks" class="text-center"></th>\n      <th ng-repeat="label in ::labels track by $index" class="text-center"><small aria-label="{{::label.full}}">{{::label.abbr}}</small></th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr class="uib-weeks" ng-repeat="row in rows track by $index" role="row">\n      <td ng-if="showWeeks" class="text-center h6"><em>{{ weekNumbers[$index] }}</em></td>\n      <td ng-repeat="dt in row" class="uib-day text-center" role="gridcell"\n        id="{{::dt.uid}}"\n        ng-class="::dt.customClass">\n        <button type="button" class="btn btn-default btn-sm"\n          uib-is-class="\n            \'btn-info\' for selectedDt,\n            \'active\' for activeDt\n            on dt"\n          ng-click="select(dt.date)"\n          ng-disabled="::dt.disabled"\n          tabindex="-1"><span ng-class="::{\'text-muted\': dt.secondary, \'text-info\': dt.current}">{{::dt.label}}</span></button>\n      </td>\n    </tr>\n  </tbody>\n</table>\n')}]),angular.module("uib/template/datepicker/month.html",[]).run(["$templateCache",function(t){t.put("uib/template/datepicker/month.html",'<table role="grid" aria-labelledby="{{::uniqueId}}-title" aria-activedescendant="{{activeDateId}}">\n  <thead>\n    <tr>\n      <th><button type="button" class="btn btn-default btn-sm pull-left uib-left" ng-click="move(-1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-left"></i><span class="sr-only">previous</span></button></th>\n      <th colspan="{{::yearHeaderColspan}}"><button id="{{::uniqueId}}-title" role="heading" aria-live="assertive" aria-atomic="true" type="button" class="btn btn-default btn-sm uib-title" ng-click="toggleMode()" ng-disabled="datepickerMode === maxMode" tabindex="-1"><strong>{{title}}</strong></button></th>\n      <th><button type="button" class="btn btn-default btn-sm pull-right uib-right" ng-click="move(1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-right"></i><span class="sr-only">next</span></i></button></th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr class="uib-months" ng-repeat="row in rows track by $index" role="row">\n      <td ng-repeat="dt in row" class="uib-month text-center" role="gridcell"\n        id="{{::dt.uid}}"\n        ng-class="::dt.customClass">\n        <button type="button" class="btn btn-default"\n          uib-is-class="\n            \'btn-info\' for selectedDt,\n            \'active\' for activeDt\n            on dt"\n          ng-click="select(dt.date)"\n          ng-disabled="::dt.disabled"\n          tabindex="-1"><span ng-class="::{\'text-info\': dt.current}">{{::dt.label}}</span></button>\n      </td>\n    </tr>\n  </tbody>\n</table>\n')}]),angular.module("uib/template/datepicker/year.html",[]).run(["$templateCache",function(t){t.put("uib/template/datepicker/year.html",'<table role="grid" aria-labelledby="{{::uniqueId}}-title" aria-activedescendant="{{activeDateId}}">\n  <thead>\n    <tr>\n      <th><button type="button" class="btn btn-default btn-sm pull-left uib-left" ng-click="move(-1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-left"></i><span class="sr-only">previous</span></button></th>\n      <th colspan="{{::columns - 2}}"><button id="{{::uniqueId}}-title" role="heading" aria-live="assertive" aria-atomic="true" type="button" class="btn btn-default btn-sm uib-title" ng-click="toggleMode()" ng-disabled="datepickerMode === maxMode" tabindex="-1"><strong>{{title}}</strong></button></th>\n      <th><button type="button" class="btn btn-default btn-sm pull-right uib-right" ng-click="move(1)" tabindex="-1"><i aria-hidden="true" class="glyphicon glyphicon-chevron-right"></i><span class="sr-only">next</span></button></th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr class="uib-years" ng-repeat="row in rows track by $index" role="row">\n      <td ng-repeat="dt in row" class="uib-year text-center" role="gridcell"\n        id="{{::dt.uid}}"\n        ng-class="::dt.customClass">\n        <button type="button" class="btn btn-default"\n          uib-is-class="\n            \'btn-info\' for selectedDt,\n            \'active\' for activeDt\n            on dt"\n          ng-click="select(dt.date)"\n          ng-disabled="::dt.disabled"\n          tabindex="-1"><span ng-class="::{\'text-info\': dt.current}">{{::dt.label}}</span></button>\n      </td>\n    </tr>\n  </tbody>\n</table>\n')}]),angular.module("uib/template/datepickerPopup/popup.html",[]).run(["$templateCache",function(t){t.put("uib/template/datepickerPopup/popup.html",'<ul role="presentation" class="uib-datepicker-popup dropdown-menu uib-position-measure" dropdown-nested ng-if="isOpen" ng-keydown="keydown($event)" ng-click="$event.stopPropagation()">\n  <li ng-transclude></li>\n  <li ng-if="showButtonBar" class="uib-button-bar">\n    <span class="btn-group pull-left">\n      <button type="button" class="btn btn-sm btn-info uib-datepicker-current" ng-click="select(\'today\', $event)" ng-disabled="isDisabled(\'today\')">{{ getText(\'current\') }}</button>\n      <button type="button" class="btn btn-sm btn-danger uib-clear" ng-click="select(null, $event)">{{ getText(\'clear\') }}</button>\n    </span>\n    <button type="button" class="btn btn-sm btn-success pull-right uib-close" ng-click="close($event)">{{ getText(\'close\') }}</button>\n  </li>\n</ul>\n')}]),angular.module("uib/template/modal/window.html",[]).run(["$templateCache",function(t){t.put("uib/template/modal/window.html","<div class=\"modal-dialog {{size ? 'modal-' + size : ''}}\"><div class=\"modal-content\" uib-modal-transclude></div></div>\n")}]),angular.module("uib/template/pager/pager.html",[]).run(["$templateCache",function(t){t.put("uib/template/pager/pager.html",'<li ng-class="{disabled: noPrevious()||ngDisabled, previous: align}"><a href ng-click="selectPage(page - 1, $event)" ng-disabled="noPrevious()||ngDisabled" uib-tabindex-toggle>{{::getText(\'previous\')}}</a></li>\n<li ng-class="{disabled: noNext()||ngDisabled, next: align}"><a href ng-click="selectPage(page + 1, $event)" ng-disabled="noNext()||ngDisabled" uib-tabindex-toggle>{{::getText(\'next\')}}</a></li>\n')}]),angular.module("uib/template/pagination/pagination.html",[]).run(["$templateCache",function(t){t.put("uib/template/pagination/pagination.html",'<li ng-if="::boundaryLinks" ng-class="{disabled: noPrevious()||ngDisabled}" class="pagination-first"><a href ng-click="selectPage(1, $event)" ng-disabled="noPrevious()||ngDisabled" uib-tabindex-toggle>{{::getText(\'first\')}}</a></li>\n<li ng-if="::directionLinks" ng-class="{disabled: noPrevious()||ngDisabled}" class="pagination-prev"><a href ng-click="selectPage(page - 1, $event)" ng-disabled="noPrevious()||ngDisabled" uib-tabindex-toggle>{{::getText(\'previous\')}}</a></li>\n<li ng-repeat="page in pages track by $index" ng-class="{active: page.active,disabled: ngDisabled&&!page.active}" class="pagination-page"><a href ng-click="selectPage(page.number, $event)" ng-disabled="ngDisabled&&!page.active" uib-tabindex-toggle>{{page.text}}</a></li>\n<li ng-if="::directionLinks" ng-class="{disabled: noNext()||ngDisabled}" class="pagination-next"><a href ng-click="selectPage(page + 1, $event)" ng-disabled="noNext()||ngDisabled" uib-tabindex-toggle>{{::getText(\'next\')}}</a></li>\n<li ng-if="::boundaryLinks" ng-class="{disabled: noNext()||ngDisabled}" class="pagination-last"><a href ng-click="selectPage(totalPages, $event)" ng-disabled="noNext()||ngDisabled" uib-tabindex-toggle>{{::getText(\'last\')}}</a></li>\n')}]),angular.module("uib/template/tooltip/tooltip-html-popup.html",[]).run(["$templateCache",function(t){t.put("uib/template/tooltip/tooltip-html-popup.html",'<div class="tooltip-arrow"></div>\n<div class="tooltip-inner" ng-bind-html="contentExp()"></div>\n')}]),angular.module("uib/template/tooltip/tooltip-popup.html",[]).run(["$templateCache",function(t){t.put("uib/template/tooltip/tooltip-popup.html",'<div class="tooltip-arrow"></div>\n<div class="tooltip-inner" ng-bind="content"></div>\n')}]),angular.module("uib/template/tooltip/tooltip-template-popup.html",[]).run(["$templateCache",function(t){t.put("uib/template/tooltip/tooltip-template-popup.html",'<div class="tooltip-arrow"></div>\n<div class="tooltip-inner"\n  uib-tooltip-template-transclude="contentExp()"\n  tooltip-template-transclude-scope="originScope()"></div>\n')}]),angular.module("uib/template/popover/popover-html.html",[]).run(["$templateCache",function(t){t.put("uib/template/popover/popover-html.html",'<div class="arrow"></div>\n\n<div class="popover-inner">\n    <h3 class="popover-title" ng-bind="uibTitle" ng-if="uibTitle"></h3>\n    <div class="popover-content" ng-bind-html="contentExp()"></div>\n</div>\n')}]),angular.module("uib/template/popover/popover-template.html",[]).run(["$templateCache",function(t){t.put("uib/template/popover/popover-template.html",'<div class="arrow"></div>\n\n<div class="popover-inner">\n    <h3 class="popover-title" ng-bind="uibTitle" ng-if="uibTitle"></h3>\n    <div class="popover-content"\n      uib-tooltip-template-transclude="contentExp()"\n      tooltip-template-transclude-scope="originScope()"></div>\n</div>\n')}]),angular.module("uib/template/popover/popover.html",[]).run(["$templateCache",function(t){t.put("uib/template/popover/popover.html",'<div class="arrow"></div>\n\n<div class="popover-inner">\n    <h3 class="popover-title" ng-bind="uibTitle" ng-if="uibTitle"></h3>\n    <div class="popover-content" ng-bind="content"></div>\n</div>\n')}]),angular.module("uib/template/progressbar/bar.html",[]).run(["$templateCache",function(t){t.put("uib/template/progressbar/bar.html",'<div class="progress-bar" ng-class="type && \'progress-bar-\' + type" role="progressbar" aria-valuenow="{{value}}" aria-valuemin="0" aria-valuemax="{{max}}" ng-style="{width: (percent < 100 ? percent : 100) + \'%\'}" aria-valuetext="{{percent | number:0}}%" aria-labelledby="{{::title}}" ng-transclude></div>\n')}]),angular.module("uib/template/progressbar/progress.html",[]).run(["$templateCache",function(t){t.put("uib/template/progressbar/progress.html",'<div class="progress" ng-transclude aria-labelledby="{{::title}}"></div>')}]),angular.module("uib/template/progressbar/progressbar.html",[]).run(["$templateCache",function(t){t.put("uib/template/progressbar/progressbar.html",'<div class="progress">\n  <div class="progress-bar" ng-class="type && \'progress-bar-\' + type" role="progressbar" aria-valuenow="{{value}}" aria-valuemin="0" aria-valuemax="{{max}}" ng-style="{width: (percent < 100 ? percent : 100) + \'%\'}" aria-valuetext="{{percent | number:0}}%" aria-labelledby="{{::title}}" ng-transclude></div>\n</div>\n')}]),angular.module("uib/template/rating/rating.html",[]).run(["$templateCache",function(t){t.put("uib/template/rating/rating.html",'<span ng-mouseleave="reset()" ng-keydown="onKeydown($event)" tabindex="0" role="slider" aria-valuemin="0" aria-valuemax="{{range.length}}" aria-valuenow="{{value}}" aria-valuetext="{{title}}">\n    <span ng-repeat-start="r in range track by $index" class="sr-only">({{ $index < value ? \'*\' : \' \' }})</span>\n    <i ng-repeat-end ng-mouseenter="enter($index + 1)" ng-click="rate($index + 1)" class="glyphicon" ng-class="$index < value && (r.stateOn || \'glyphicon-star\') || (r.stateOff || \'glyphicon-star-empty\')" ng-attr-title="{{r.title}}"></i>\n</span>\n')}]),angular.module("uib/template/tabs/tab.html",[]).run(["$templateCache",function(t){t.put("uib/template/tabs/tab.html",'<li ng-class="[{active: active, disabled: disabled}, classes]" class="uib-tab nav-item">\n  <a href ng-click="select($event)" class="nav-link" uib-tab-heading-transclude>{{heading}}</a>\n</li>\n')}]),angular.module("uib/template/tabs/tabset.html",[]).run(["$templateCache",function(t){t.put("uib/template/tabs/tabset.html",'<div>\n  <ul class="nav nav-{{tabset.type || \'tabs\'}}" ng-class="{\'nav-stacked\': vertical, \'nav-justified\': justified}" ng-transclude></ul>\n  <div class="tab-content">\n    <div class="tab-pane"\n         ng-repeat="tab in tabset.tabs"\n         ng-class="{active: tabset.active === tab.index}"\n         uib-tab-content-transclude="tab">\n    </div>\n  </div>\n</div>\n')}]),angular.module("uib/template/timepicker/timepicker.html",[]).run(["$templateCache",function(t){t.put("uib/template/timepicker/timepicker.html",'<table class="uib-timepicker">\n  <tbody>\n    <tr class="text-center" ng-show="::showSpinners">\n      <td class="uib-increment hours"><a ng-click="incrementHours()" ng-class="{disabled: noIncrementHours()}" class="btn btn-link" ng-disabled="noIncrementHours()" tabindex="-1"><span class="glyphicon glyphicon-chevron-up"></span></a></td>\n      <td>&nbsp;</td>\n      <td class="uib-increment minutes"><a ng-click="incrementMinutes()" ng-class="{disabled: noIncrementMinutes()}" class="btn btn-link" ng-disabled="noIncrementMinutes()" tabindex="-1"><span class="glyphicon glyphicon-chevron-up"></span></a></td>\n      <td ng-show="showSeconds">&nbsp;</td>\n      <td ng-show="showSeconds" class="uib-increment seconds"><a ng-click="incrementSeconds()" ng-class="{disabled: noIncrementSeconds()}" class="btn btn-link" ng-disabled="noIncrementSeconds()" tabindex="-1"><span class="glyphicon glyphicon-chevron-up"></span></a></td>\n      <td ng-show="showMeridian"></td>\n    </tr>\n    <tr>\n      <td class="form-group uib-time hours" ng-class="{\'has-error\': invalidHours}">\n        <input type="text" placeholder="HH" ng-model="hours" ng-change="updateHours()" class="form-control text-center" ng-readonly="::readonlyInput" maxlength="2" tabindex="{{::tabindex}}" ng-disabled="noIncrementHours()" ng-blur="blur()">\n      </td>\n      <td class="uib-separator">:</td>\n      <td class="form-group uib-time minutes" ng-class="{\'has-error\': invalidMinutes}">\n        <input type="text" placeholder="MM" ng-model="minutes" ng-change="updateMinutes()" class="form-control text-center" ng-readonly="::readonlyInput" maxlength="2" tabindex="{{::tabindex}}" ng-disabled="noIncrementMinutes()" ng-blur="blur()">\n      </td>\n      <td ng-show="showSeconds" class="uib-separator">:</td>\n      <td class="form-group uib-time seconds" ng-class="{\'has-error\': invalidSeconds}" ng-show="showSeconds">\n        <input type="text" placeholder="SS" ng-model="seconds" ng-change="updateSeconds()" class="form-control text-center" ng-readonly="readonlyInput" maxlength="2" tabindex="{{::tabindex}}" ng-disabled="noIncrementSeconds()" ng-blur="blur()">\n      </td>\n      <td ng-show="showMeridian" class="uib-time am-pm"><button type="button" ng-class="{disabled: noToggleMeridian()}" class="btn btn-default text-center" ng-click="toggleMeridian()" ng-disabled="noToggleMeridian()" tabindex="{{::tabindex}}">{{meridian}}</button></td>\n    </tr>\n    <tr class="text-center" ng-show="::showSpinners">\n      <td class="uib-decrement hours"><a ng-click="decrementHours()" ng-class="{disabled: noDecrementHours()}" class="btn btn-link" ng-disabled="noDecrementHours()" tabindex="-1"><span class="glyphicon glyphicon-chevron-down"></span></a></td>\n      <td>&nbsp;</td>\n      <td class="uib-decrement minutes"><a ng-click="decrementMinutes()" ng-class="{disabled: noDecrementMinutes()}" class="btn btn-link" ng-disabled="noDecrementMinutes()" tabindex="-1"><span class="glyphicon glyphicon-chevron-down"></span></a></td>\n      <td ng-show="showSeconds">&nbsp;</td>\n      <td ng-show="showSeconds" class="uib-decrement seconds"><a ng-click="decrementSeconds()" ng-class="{disabled: noDecrementSeconds()}" class="btn btn-link" ng-disabled="noDecrementSeconds()" tabindex="-1"><span class="glyphicon glyphicon-chevron-down"></span></a></td>\n      <td ng-show="showMeridian"></td>\n    </tr>\n  </tbody>\n</table>\n')}]),angular.module("uib/template/typeahead/typeahead-match.html",[]).run(["$templateCache",function(t){t.put("uib/template/typeahead/typeahead-match.html",'<a href\n   tabindex="-1"\n   ng-bind-html="match.label | uibTypeaheadHighlight:query"\n   ng-attr-title="{{match.label}}"></a>\n')}]),angular.module("uib/template/typeahead/typeahead-popup.html",[]).run(["$templateCache",function(t){t.put("uib/template/typeahead/typeahead-popup.html",'<ul class="dropdown-menu" ng-show="isOpen() && !moveInProgress" ng-style="{top: position().top+\'px\', left: position().left+\'px\'}" role="listbox" aria-hidden="{{!isOpen()}}">\n    <li class="uib-typeahead-match" ng-repeat="match in matches track by $index" ng-class="{active: isActive($index) }" ng-mouseenter="selectActive($index)" ng-click="selectMatch($index, $event)" role="option" id="{{::match.id}}">\n        <div uib-typeahead-match index="$index" match="match" query="query" template-url="templateUrl"></div>\n    </li>\n</ul>\n')}]),angular.module("ui.bootstrap.carousel").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibCarouselCss&&angular.element(document).find("head").prepend('<style type="text/css">.ng-animate.item:not(.left):not(.right){-webkit-transition:0s ease-in-out left;transition:0s ease-in-out left}</style>'),angular.$$uibCarouselCss=!0}),angular.module("ui.bootstrap.datepicker").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibDatepickerCss&&angular.element(document).find("head").prepend('<style type="text/css">.uib-datepicker .uib-title{width:100%;}.uib-day button,.uib-month button,.uib-year button{min-width:100%;}.uib-left,.uib-right{width:100%}</style>'),angular.$$uibDatepickerCss=!0}),angular.module("ui.bootstrap.position").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibPositionCss&&angular.element(document).find("head").prepend('<style type="text/css">.uib-position-measure{display:block !important;visibility:hidden !important;position:absolute !important;top:-9999px !important;left:-9999px !important;}.uib-position-scrollbar-measure{position:absolute !important;top:-9999px !important;width:50px !important;height:50px !important;overflow:scroll !important;}.uib-position-body-scrollbar-measure{overflow:scroll !important;}</style>'),angular.$$uibPositionCss=!0}),angular.module("ui.bootstrap.datepickerPopup").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibDatepickerpopupCss&&angular.element(document).find("head").prepend('<style type="text/css">.uib-datepicker-popup.dropdown-menu{display:block;float:none;margin:0;}.uib-button-bar{padding:10px 9px 2px;}</style>'),angular.$$uibDatepickerpopupCss=!0}),angular.module("ui.bootstrap.tooltip").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibTooltipCss&&angular.element(document).find("head").prepend('<style type="text/css">[uib-tooltip-popup].tooltip.top-left > .tooltip-arrow,[uib-tooltip-popup].tooltip.top-right > .tooltip-arrow,[uib-tooltip-popup].tooltip.bottom-left > .tooltip-arrow,[uib-tooltip-popup].tooltip.bottom-right > .tooltip-arrow,[uib-tooltip-popup].tooltip.left-top > .tooltip-arrow,[uib-tooltip-popup].tooltip.left-bottom > .tooltip-arrow,[uib-tooltip-popup].tooltip.right-top > .tooltip-arrow,[uib-tooltip-popup].tooltip.right-bottom > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.top-left > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.top-right > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.bottom-left > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.bottom-right > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.left-top > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.left-bottom > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.right-top > .tooltip-arrow,[uib-tooltip-html-popup].tooltip.right-bottom > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.top-left > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.top-right > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.bottom-left > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.bottom-right > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.left-top > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.left-bottom > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.right-top > .tooltip-arrow,[uib-tooltip-template-popup].tooltip.right-bottom > .tooltip-arrow,[uib-popover-popup].popover.top-left > .arrow,[uib-popover-popup].popover.top-right > .arrow,[uib-popover-popup].popover.bottom-left > .arrow,[uib-popover-popup].popover.bottom-right > .arrow,[uib-popover-popup].popover.left-top > .arrow,[uib-popover-popup].popover.left-bottom > .arrow,[uib-popover-popup].popover.right-top > .arrow,[uib-popover-popup].popover.right-bottom > .arrow,[uib-popover-html-popup].popover.top-left > .arrow,[uib-popover-html-popup].popover.top-right > .arrow,[uib-popover-html-popup].popover.bottom-left > .arrow,[uib-popover-html-popup].popover.bottom-right > .arrow,[uib-popover-html-popup].popover.left-top > .arrow,[uib-popover-html-popup].popover.left-bottom > .arrow,[uib-popover-html-popup].popover.right-top > .arrow,[uib-popover-html-popup].popover.right-bottom > .arrow,[uib-popover-template-popup].popover.top-left > .arrow,[uib-popover-template-popup].popover.top-right > .arrow,[uib-popover-template-popup].popover.bottom-left > .arrow,[uib-popover-template-popup].popover.bottom-right > .arrow,[uib-popover-template-popup].popover.left-top > .arrow,[uib-popover-template-popup].popover.left-bottom > .arrow,[uib-popover-template-popup].popover.right-top > .arrow,[uib-popover-template-popup].popover.right-bottom > .arrow{top:auto;bottom:auto;left:auto;right:auto;margin:0;}[uib-popover-popup].popover,[uib-popover-html-popup].popover,[uib-popover-template-popup].popover{display:block !important;}</style>'),
angular.$$uibTooltipCss=!0}),angular.module("ui.bootstrap.timepicker").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibTimepickerCss&&angular.element(document).find("head").prepend('<style type="text/css">.uib-time input{width:50px;}</style>'),angular.$$uibTimepickerCss=!0}),angular.module("ui.bootstrap.typeahead").run(function(){!angular.$$csp().noInlineStyle&&!angular.$$uibTypeaheadCss&&angular.element(document).find("head").prepend('<style type="text/css">[uib-typeahead-popup].dropdown-menu{display:block;}</style>'),angular.$$uibTypeaheadCss=!0})},3:function(t,e,n){n(24),t.exports="smart-table"},36:function(t,e){/**
 * @license AngularJS v1.6.1
 * (c) 2010-2016 Google, Inc. http://angularjs.org
 * License: MIT
 */
!function(t){"use strict";function e(t,e){return e=e||Error,function(){var n,r,i=2,o=arguments,a=o[0],s="["+(t?t+":":"")+a+"] ",u=o[1];for(s+=u.replace(/\{\d+\}/g,function(t){var e=+t.slice(1,-1),n=e+i;return n<o.length?wt(o[n]):t}),s+="\nhttp://errors.angularjs.org/1.6.1/"+(t?t+"/":"")+a,r=i,n="?";r<o.length;r++,n="&")s+=n+"p"+(r-i)+"="+encodeURIComponent(wt(o[r]));return new e(s)}}function n(t){if(null==t||S(t))return!1;if(li(t)||w(t)||Jr&&t instanceof Jr)return!0;var e="length"in Object(t)&&t.length;return x(e)&&(e>=0&&(e-1 in t||t instanceof Array)||"function"==typeof t.item)}function r(t,e,i){var o,a;if(t)if(C(t))for(o in t)"prototype"!==o&&"length"!==o&&"name"!==o&&t.hasOwnProperty(o)&&e.call(i,t[o],o,t);else if(li(t)||n(t)){var s="object"!=typeof t;for(o=0,a=t.length;o<a;o++)(s||o in t)&&e.call(i,t[o],o,t)}else if(t.forEach&&t.forEach!==r)t.forEach(e,i,t);else if($(t))for(o in t)e.call(i,t[o],o,t);else if("function"==typeof t.hasOwnProperty)for(o in t)t.hasOwnProperty(o)&&e.call(i,t[o],o,t);else for(o in t)zr.call(t,o)&&e.call(i,t[o],o,t);return t}function i(t,e,n){for(var r=Object.keys(t).sort(),i=0;i<r.length;i++)e.call(n,t[r[i]],r[i]);return r}function o(t){return function(e,n){t(n,e)}}function a(){return++si}function s(t,e){e?t.$$hashKey=e:delete t.$$hashKey}function u(t,e,n){for(var r=t.$$hashKey,i=0,o=e.length;i<o;++i){var a=e[i];if(b(a)||C(a))for(var l=Object.keys(a),c=0,f=l.length;c<f;c++){var p=l[c],h=a[p];n&&b(h)?_(h)?t[p]=new Date(h.valueOf()):k(h)?t[p]=new RegExp(h):h.nodeName?t[p]=h.cloneNode(!0):R(h)?t[p]=h.clone():(b(t[p])||(t[p]=li(h)?[]:{}),u(t[p],[h],!0)):t[p]=h}}return s(t,r),t}function l(t){return u(t,ti.call(arguments,1),!1)}function c(t){return u(t,ti.call(arguments,1),!0)}function f(t){return parseInt(t,10)}function p(t,e){return l(Object.create(t),e)}function h(){}function d(t){return t}function v(t){return function(){return t}}function g(t){return C(t.toString)&&t.toString!==ri}function m(t){return"undefined"==typeof t}function y(t){return"undefined"!=typeof t}function b(t){return null!==t&&"object"==typeof t}function $(t){return null!==t&&"object"==typeof t&&!ii(t)}function w(t){return"string"==typeof t}function x(t){return"number"==typeof t}function _(t){return"[object Date]"===ri.call(t)}function C(t){return"function"==typeof t}function k(t){return"[object RegExp]"===ri.call(t)}function S(t){return t&&t.window===t}function T(t){return t&&t.$evalAsync&&t.$watch}function E(t){return"[object File]"===ri.call(t)}function M(t){return"[object FormData]"===ri.call(t)}function D(t){return"[object Blob]"===ri.call(t)}function A(t){return"boolean"==typeof t}function O(t){return t&&C(t.then)}function N(t){return t&&x(t.length)&&ci.test(ri.call(t))}function P(t){return"[object ArrayBuffer]"===ri.call(t)}function R(t){return!(!t||!(t.nodeName||t.prop&&t.attr&&t.find))}function I(t){var e,n={},r=t.split(",");for(e=0;e<r.length;e++)n[r[e]]=!0;return n}function j(t){return Wr(t.nodeName||t[0]&&t[0].nodeName)}function F(t,e){return Array.prototype.indexOf.call(t,e)!==-1}function q(t,e){var n=t.indexOf(e);return n>=0&&t.splice(n,1),n}function V(t,e){function n(t,e){var n,r=e.$$hashKey;if(li(t))for(var o=0,a=t.length;o<a;o++)e.push(i(t[o]));else if($(t))for(n in t)e[n]=i(t[n]);else if(t&&"function"==typeof t.hasOwnProperty)for(n in t)t.hasOwnProperty(n)&&(e[n]=i(t[n]));else for(n in t)zr.call(t,n)&&(e[n]=i(t[n]));return s(e,r),e}function i(t){if(!b(t))return t;var e=a.indexOf(t);if(e!==-1)return u[e];if(S(t)||T(t))throw oi("cpws","Can't copy! Making copies of Window or Scope instances is not supported.");var r=!1,i=o(t);return void 0===i&&(i=li(t)?[]:Object.create(ii(t)),r=!0),a.push(t),u.push(i),r?n(t,i):i}function o(t){switch(ri.call(t)){case"[object Int8Array]":case"[object Int16Array]":case"[object Int32Array]":case"[object Float32Array]":case"[object Float64Array]":case"[object Uint8Array]":case"[object Uint8ClampedArray]":case"[object Uint16Array]":case"[object Uint32Array]":return new t.constructor(i(t.buffer),t.byteOffset,t.length);case"[object ArrayBuffer]":if(!t.slice){var e=new ArrayBuffer(t.byteLength);return new Uint8Array(e).set(new Uint8Array(t)),e}return t.slice(0);case"[object Boolean]":case"[object Number]":case"[object String]":case"[object Date]":return new t.constructor(t.valueOf());case"[object RegExp]":var n=new RegExp(t.source,t.toString().match(/[^\/]*$/)[0]);return n.lastIndex=t.lastIndex,n;case"[object Blob]":return new t.constructor([t],{type:t.type})}if(C(t.cloneNode))return t.cloneNode(!0)}var a=[],u=[];if(e){if(N(e)||P(e))throw oi("cpta","Can't copy! TypedArray destination cannot be mutated.");if(t===e)throw oi("cpi","Can't copy! Source and destination are identical.");return li(e)?e.length=0:r(e,function(t,n){"$$hashKey"!==n&&delete e[n]}),a.push(t),u.push(e),n(t,e)}return i(t)}function U(t,e){if(t===e)return!0;if(null===t||null===e)return!1;if(t!==t&&e!==e)return!0;var n,r,i,o=typeof t,a=typeof e;if(o===a&&"object"===o){if(!li(t)){if(_(t))return!!_(e)&&U(t.getTime(),e.getTime());if(k(t))return!!k(e)&&t.toString()===e.toString();if(T(t)||T(e)||S(t)||S(e)||li(e)||_(e)||k(e))return!1;i=gt();for(r in t)if("$"!==r.charAt(0)&&!C(t[r])){if(!U(t[r],e[r]))return!1;i[r]=!0}for(r in e)if(!(r in i)&&"$"!==r.charAt(0)&&y(e[r])&&!C(e[r]))return!1;return!0}if(!li(e))return!1;if((n=t.length)===e.length){for(r=0;r<n;r++)if(!U(t[r],e[r]))return!1;return!0}}return!1}function L(t,e,n){return t.concat(ti.call(e,n))}function H(t,e){return ti.call(t,e||0)}function B(t,e){var n=arguments.length>2?H(arguments,2):[];return!C(e)||e instanceof RegExp?e:n.length?function(){return arguments.length?e.apply(t,L(n,arguments,0)):e.apply(t,n)}:function(){return arguments.length?e.apply(t,arguments):e.call(t)}}function z(e,n){var r=n;return"string"==typeof e&&"$"===e.charAt(0)&&"$"===e.charAt(1)?r=void 0:S(n)?r="$WINDOW":n&&t.document===n?r="$DOCUMENT":T(n)&&(r="$SCOPE"),r}function W(t,e){if(!m(t))return x(e)||(e=e?2:null),JSON.stringify(t,z,e)}function Y(t){return w(t)?JSON.parse(t):t}function G(t,e){t=t.replace(vi,"");var n=Date.parse("Jan 01, 1970 00:00:00 "+t)/6e4;return ui(n)?e:n}function X(t,e){return t=new Date(t.getTime()),t.setMinutes(t.getMinutes()+e),t}function K(t,e,n){n=n?-1:1;var r=t.getTimezoneOffset(),i=G(e,r);return X(t,n*(i-r))}function J(t){t=Jr(t).clone();try{t.empty()}catch(t){}var e=Jr("<div>").append(t).html();try{return t[0].nodeType===xi?Wr(e):e.match(/^(<[^>]+>)/)[1].replace(/^<([\w-]+)/,function(t,e){return"<"+Wr(e)})}catch(t){return Wr(e)}}function Z(t){try{return decodeURIComponent(t)}catch(t){}}function Q(t){var e={};return r((t||"").split("&"),function(t){var n,r,i;t&&(r=t=t.replace(/\+/g,"%20"),n=t.indexOf("="),n!==-1&&(r=t.substring(0,n),i=t.substring(n+1)),r=Z(r),y(r)&&(i=!y(i)||Z(i),zr.call(e,r)?li(e[r])?e[r].push(i):e[r]=[e[r],i]:e[r]=i))}),e}function tt(t){var e=[];return r(t,function(t,n){li(t)?r(t,function(t){e.push(nt(n,!0)+(t===!0?"":"="+nt(t,!0)))}):e.push(nt(n,!0)+(t===!0?"":"="+nt(t,!0)))}),e.length?e.join("&"):""}function et(t){return nt(t,!0).replace(/%26/gi,"&").replace(/%3D/gi,"=").replace(/%2B/gi,"+")}function nt(t,e){return encodeURIComponent(t).replace(/%40/gi,"@").replace(/%3A/gi,":").replace(/%24/g,"$").replace(/%2C/gi,",").replace(/%3B/gi,";").replace(/%20/g,e?"%20":"+")}function rt(t,e){var n,r,i=gi.length;for(r=0;r<i;++r)if(n=gi[r]+e,w(n=t.getAttribute(n)))return n;return null}function it(t){if(!t.currentScript)return!0;var e=t.currentScript.getAttribute("src"),n=t.createElement("a");if(n.href=e,t.location.origin===n.origin)return!0;switch(n.protocol){case"http:":case"https:":case"ftp:":case"blob:":case"file:":case"data:":return!0;default:return!1}}function ot(e,n){var i,o,a={};if(r(gi,function(t){var n=t+"app";!i&&e.hasAttribute&&e.hasAttribute(n)&&(i=e,o=e.getAttribute(n))}),r(gi,function(t){var n,r=t+"app";!i&&(n=e.querySelector("["+r.replace(":","\\:")+"]"))&&(i=n,o=n.getAttribute(r))}),i){if(!mi)return void t.console.error("Angular: disabling automatic bootstrap. <script> protocol indicates an extension, document.location.href does not match.");a.strictDi=null!==rt(i,"strict-di"),n(i,o?[o]:[],a)}}function at(e,n,i){b(i)||(i={});var o={strictDi:!1};i=l(o,i);var a=function(){if(e=Jr(e),e.injector()){var r=e[0]===t.document?"document":J(e);throw oi("btstrpd","App already bootstrapped with this element '{0}'",r.replace(/</,"&lt;").replace(/>/,"&gt;"))}n=n||[],n.unshift(["$provide",function(t){t.value("$rootElement",e)}]),i.debugInfoEnabled&&n.push(["$compileProvider",function(t){t.debugInfoEnabled(!0)}]),n.unshift("ng");var o=le(n,i.strictDi);return o.invoke(["$rootScope","$rootElement","$compile","$injector",function(t,e,n,r){t.$apply(function(){e.data("$injector",r),n(e)(t)})}]),o},s=/^NG_ENABLE_DEBUG_INFO!/,u=/^NG_DEFER_BOOTSTRAP!/;return t&&s.test(t.name)&&(i.debugInfoEnabled=!0,t.name=t.name.replace(s,"")),t&&!u.test(t.name)?a():(t.name=t.name.replace(u,""),ai.resumeBootstrap=function(t){return r(t,function(t){n.push(t)}),a()},void(C(ai.resumeDeferredBootstrap)&&ai.resumeDeferredBootstrap()))}function st(){t.name="NG_ENABLE_DEBUG_INFO!"+t.name,t.location.reload()}function ut(t){var e=ai.element(t).injector();if(!e)throw oi("test","no injector found for element argument to getTestability");return e.get("$$testability")}function lt(t,e){return e=e||"_",t.replace(yi,function(t,n){return(n?e:"")+t.toLowerCase()})}function ct(){var e;if(!bi){var n=di();Zr=m(n)?t.jQuery:n?t[n]:void 0,Zr&&Zr.fn.on?(Jr=Zr,l(Zr.fn,{scope:qi.scope,isolateScope:qi.isolateScope,controller:qi.controller,injector:qi.injector,inheritedData:qi.inheritedData}),e=Zr.cleanData,Zr.cleanData=function(t){for(var n,r,i=0;null!=(r=t[i]);i++)n=Zr._data(r,"events"),n&&n.$destroy&&Zr(r).triggerHandler("$destroy");e(t)}):Jr=Pt,ai.element=Jr,bi=!0}}function ft(t,e,n){if(!t)throw oi("areq","Argument '{0}' is {1}",e||"?",n||"required");return t}function pt(t,e,n){return n&&li(t)&&(t=t[t.length-1]),ft(C(t),e,"not a function, got "+(t&&"object"==typeof t?t.constructor.name||"Object":typeof t)),t}function ht(t,e){if("hasOwnProperty"===t)throw oi("badname","hasOwnProperty is not a valid {0} name",e)}function dt(t,e,n){if(!e)return t;for(var r,i=e.split("."),o=t,a=i.length,s=0;s<a;s++)r=i[s],t&&(t=(o=t)[r]);return!n&&C(t)?B(o,t):t}function vt(t){for(var e,n=t[0],r=t[t.length-1],i=1;n!==r&&(n=n.nextSibling);i++)(e||t[i]!==n)&&(e||(e=Jr(ti.call(t,0,i))),e.push(n));return e||t}function gt(){return Object.create(null)}function mt(t){if(null==t)return"";switch(typeof t){case"string":break;case"number":t=""+t;break;default:t=!g(t)||li(t)||_(t)?W(t):t.toString()}return t}function yt(t){function n(t,e,n){return t[e]||(t[e]=n())}var r=e("$injector"),i=e("ng"),o=n(t,"angular",Object);return o.$$minErr=o.$$minErr||e,n(o,"module",function(){var t={};return function(e,o,a){var s=function(t,e){if("hasOwnProperty"===t)throw i("badname","hasOwnProperty is not a valid {0} name",e)};return s(e,"module"),o&&t.hasOwnProperty(e)&&(t[e]=null),n(t,e,function(){function t(t,e,n,r){return r||(r=i),function(){return r[n||"push"]([t,e,arguments]),c}}function n(t,n,r){return r||(r=i),function(i,o){return o&&C(o)&&(o.$$moduleName=e),r.push([t,n,arguments]),c}}if(!o)throw r("nomod","Module '{0}' is not available! You either misspelled the module name or forgot to load it. If registering a module ensure that you specify the dependencies as the second argument.",e);var i=[],s=[],u=[],l=t("$injector","invoke","push",s),c={_invokeQueue:i,_configBlocks:s,_runBlocks:u,requires:o,name:e,provider:n("$provide","provider"),factory:n("$provide","factory"),service:n("$provide","service"),value:t("$provide","value"),constant:t("$provide","constant","unshift"),decorator:n("$provide","decorator",s),animation:n("$animateProvider","register"),filter:n("$filterProvider","register"),controller:n("$controllerProvider","register"),directive:n("$compileProvider","directive"),component:n("$compileProvider","component"),config:l,run:function(t){return u.push(t),this}};return a&&l(a),c})}})}function bt(t,e){if(li(t)){e=e||[];for(var n=0,r=t.length;n<r;n++)e[n]=t[n]}else if(b(t)){e=e||{};for(var i in t)"$"===i.charAt(0)&&"$"===i.charAt(1)||(e[i]=t[i])}return e||t}function $t(t){var e=[];return JSON.stringify(t,function(t,n){if(n=z(t,n),b(n)){if(e.indexOf(n)>=0)return"...";e.push(n)}return n})}function wt(t){return"function"==typeof t?t.toString().replace(/ \{[\s\S]*$/,""):m(t)?"undefined":"string"!=typeof t?$t(t):t}function xt(n){l(n,{bootstrap:at,copy:V,extend:l,merge:c,equals:U,element:Jr,forEach:r,injector:le,noop:h,bind:B,toJson:W,fromJson:Y,identity:d,isUndefined:m,isDefined:y,isString:w,isFunction:C,isObject:b,isNumber:x,isElement:R,isArray:li,version:Si,isDate:_,lowercase:Wr,uppercase:Yr,callbacks:{$$counter:0},getTestability:ut,reloadWithDebugInfo:st,$$minErr:e,$$csp:hi,$$encodeUriSegment:et,$$encodeUriQuery:nt,$$stringify:mt}),(Qr=yt(t))("ng",["ngLocale"],["$provide",function(t){t.provider({$$sanitizeUri:Tn}),t.provider("$compile",$e).directive({a:Yo,input:ha,textarea:ha,form:Qo,script:ss,select:cs,option:fs,ngBind:ga,ngBindHtml:ya,ngBindTemplate:ma,ngClass:$a,ngClassEven:xa,ngClassOdd:wa,ngCloak:_a,ngController:Ca,ngForm:ta,ngHide:ts,ngIf:Ta,ngInclude:Ea,ngInit:Da,ngNonBindable:za,ngPluralize:Xa,ngRepeat:Ka,ngShow:Qa,ngStyle:es,ngSwitch:ns,ngSwitchWhen:rs,ngSwitchDefault:is,ngOptions:Ga,ngTransclude:as,ngModel:La,ngList:Aa,ngChange:ba,pattern:hs,ngPattern:hs,required:ps,ngRequired:ps,minlength:vs,ngMinlength:vs,maxlength:ds,ngMaxlength:ds,ngValue:va,ngModelOptions:Ba}).directive({ngInclude:Ma}).directive(Go).directive(ka),t.provider({$anchorScroll:ce,$animate:eo,$animateCss:io,$$animateJs:Qi,$$animateQueue:to,$$AnimateRunner:ro,$$animateAsyncRun:no,$browser:ge,$cacheFactory:me,$controller:Se,$document:Te,$$isDocumentHidden:Ee,$exceptionHandler:Me,$filter:Ln,$$forceReflow:fo,$interpolate:He,$interval:Be,$http:qe,$httpParamSerializer:Ae,$httpParamSerializerJQLike:Oe,$httpBackend:Ue,$xhrFactory:Ve,$jsonpCallbacks:$o,$location:an,$log:sn,$parse:wn,$rootScope:Sn,$q:xn,$$q:_n,$sce:On,$sceDelegate:An,$sniffer:Nn,$templateCache:ye,$templateRequest:Pn,$$testability:Rn,$timeout:In,$window:qn,$$rAF:kn,$$jqLite:ne,$$HashMap:Hi,$$cookieReader:Un})}])}function _t(){return++Ei}function Ct(t){return St(t.replace(Di,"ms-"))}function kt(t,e){return e.toUpperCase()}function St(t){return t.replace(Mi,kt)}function Tt(t){return!Pi.test(t)}function Et(t){var e=t.nodeType;return e===$i||!e||e===Ci}function Mt(t){for(var e in Ti[t.ng339])return!0;return!1}function Dt(t){for(var e=0,n=t.length;e<n;e++)Ft(t[e])}function At(t,e){var n,i,o,a,s=e.createDocumentFragment(),u=[];if(Tt(t))u.push(e.createTextNode(t));else{for(n=s.appendChild(e.createElement("div")),i=(Ri.exec(t)||["",""])[1].toLowerCase(),o=ji[i]||ji._default,n.innerHTML=o[1]+t.replace(Ii,"<$1></$2>")+o[2],a=o[0];a--;)n=n.lastChild;u=L(u,n.childNodes),n=s.firstChild,n.textContent=""}return s.textContent="",s.innerHTML="",r(u,function(t){s.appendChild(t)}),s}function Ot(e,n){n=n||t.document;var r;return(r=Ni.exec(e))?[n.createElement(r[1])]:(r=At(e,n))?r.childNodes:[]}function Nt(t,e){var n=t.parentNode;n&&n.replaceChild(e,t),e.appendChild(t)}function Pt(t){if(t instanceof Pt)return t;var e;if(w(t)&&(t=fi(t),e=!0),!(this instanceof Pt)){if(e&&"<"!==t.charAt(0))throw Oi("nosel","Looking up elements via selectors is not supported by jqLite! See: http://docs.angularjs.org/api/angular.element");return new Pt(t)}e?Bt(this,Ot(t)):C(t)?Kt(t):Bt(this,t)}function Rt(t){return t.cloneNode(!0)}function It(t,e){if(e||Ft(t),t.querySelectorAll)for(var n=t.querySelectorAll("*"),r=0,i=n.length;r<i;r++)Ft(n[r])}function jt(t,e,n,i){if(y(i))throw Oi("offargs","jqLite#off() does not support the `selector` argument");var o=qt(t),a=o&&o.events,s=o&&o.handle;if(s)if(e){var u=function(e){var r=a[e];y(n)&&q(r||[],n),y(n)&&r&&r.length>0||(t.removeEventListener(e,s),delete a[e])};r(e.split(" "),function(t){u(t),Ai[t]&&u(Ai[t])})}else for(e in a)"$destroy"!==e&&t.removeEventListener(e,s),delete a[e]}function Ft(t,e){var n=t.ng339,r=n&&Ti[n];if(r){if(e)return void delete r.data[e];r.handle&&(r.events.$destroy&&r.handle({},"$destroy"),jt(t)),delete Ti[n],t.ng339=void 0}}function qt(t,e){var n=t.ng339,r=n&&Ti[n];return e&&!r&&(t.ng339=n=_t(),r=Ti[n]={events:{},data:{},handle:void 0}),r}function Vt(t,e,n){if(Et(t)){var r,i=y(n),o=!i&&e&&!b(e),a=!e,s=qt(t,!o),u=s&&s.data;if(i)u[St(e)]=n;else{if(a)return u;if(o)return u&&u[St(e)];for(r in e)u[St(r)]=e[r]}}}function Ut(t,e){return!!t.getAttribute&&(" "+(t.getAttribute("class")||"")+" ").replace(/[\n\t]/g," ").indexOf(" "+e+" ")>-1}function Lt(t,e){e&&t.setAttribute&&r(e.split(" "),function(e){t.setAttribute("class",fi((" "+(t.getAttribute("class")||"")+" ").replace(/[\n\t]/g," ").replace(" "+fi(e)+" "," ")))})}function Ht(t,e){if(e&&t.setAttribute){var n=(" "+(t.getAttribute("class")||"")+" ").replace(/[\n\t]/g," ");r(e.split(" "),function(t){t=fi(t),n.indexOf(" "+t+" ")===-1&&(n+=t+" ")}),t.setAttribute("class",fi(n))}}function Bt(t,e){if(e)if(e.nodeType)t[t.length++]=e;else{var n=e.length;if("number"==typeof n&&e.window!==e){if(n)for(var r=0;r<n;r++)t[t.length++]=e[r]}else t[t.length++]=e}}function zt(t,e){return Wt(t,"$"+(e||"ngController")+"Controller")}function Wt(t,e,n){t.nodeType===Ci&&(t=t.documentElement);for(var r=li(e)?e:[e];t;){for(var i=0,o=r.length;i<o;i++)if(y(n=Jr.data(t,r[i])))return n;t=t.parentNode||t.nodeType===ki&&t.host}}function Yt(t){for(It(t,!0);t.firstChild;)t.removeChild(t.firstChild)}function Gt(t,e){e||It(t);var n=t.parentNode;n&&n.removeChild(t)}function Xt(e,n){n=n||t,"complete"===n.document.readyState?n.setTimeout(e):Jr(n).on("load",e)}function Kt(e){function n(){t.document.removeEventListener("DOMContentLoaded",n),t.removeEventListener("load",n),e()}"complete"===t.document.readyState?t.setTimeout(e):(t.document.addEventListener("DOMContentLoaded",n),t.addEventListener("load",n))}function Jt(t,e){var n=Vi[e.toLowerCase()];return n&&Ui[j(t)]&&n}function Zt(t){return Li[t]}function Qt(t,e){var n=function(n,r){n.isDefaultPrevented=function(){return n.defaultPrevented};var i=e[r||n.type],o=i?i.length:0;if(o){if(m(n.immediatePropagationStopped)){var a=n.stopImmediatePropagation;n.stopImmediatePropagation=function(){n.immediatePropagationStopped=!0,n.stopPropagation&&n.stopPropagation(),a&&a.call(n)}}n.isImmediatePropagationStopped=function(){return n.immediatePropagationStopped===!0};var s=i.specialHandlerWrapper||te;o>1&&(i=bt(i));for(var u=0;u<o;u++)n.isImmediatePropagationStopped()||s(t,n,i[u])}};return n.elem=t,n}function te(t,e,n){n.call(t,e)}function ee(t,e,n){var r=e.relatedTarget;r&&(r===t||Fi.call(t,r))||n.call(t,e)}function ne(){this.$get=function(){return l(Pt,{hasClass:function(t,e){return t.attr&&(t=t[0]),Ut(t,e)},addClass:function(t,e){return t.attr&&(t=t[0]),Ht(t,e)},removeClass:function(t,e){return t.attr&&(t=t[0]),Lt(t,e)}})}}function re(t,e){var n=t&&t.$$hashKey;if(n)return"function"==typeof n&&(n=t.$$hashKey()),n;var r=typeof t;return n="function"===r||"object"===r&&null!==t?t.$$hashKey=r+":"+(e||a)():r+":"+t}function ie(t,e){if(e){var n=0;this.nextUid=function(){return++n}}r(t,this.put,this)}function oe(t){return Function.prototype.toString.call(t)+" "}function ae(t){var e=oe(t).replace(Gi,""),n=e.match(Bi)||e.match(zi);return n}function se(t){var e=ae(t);return e?"function("+(e[1]||"").replace(/[\s\r\n]+/," ")+")":"fn"}function ue(t,e,n){var i,o,a;if("function"==typeof t){if(!(i=t.$inject)){if(i=[],t.length){if(e)throw w(n)&&n||(n=t.name||se(t)),Xi("strictdi","{0} is not using explicit annotation and cannot be invoked in strict mode",n);o=ae(t),r(o[1].split(Wi),function(t){t.replace(Yi,function(t,e,n){i.push(n)})})}t.$inject=i}}else li(t)?(a=t.length-1,pt(t[a],"fn"),i=t.slice(0,a)):pt(t,"fn",!0);return i}function le(t,e){function n(t){return function(e,n){return b(e)?void r(e,o(t)):t(e,n)}}function i(t,e){if(ht(t,"service"),(C(e)||li(e))&&(e=_.instantiate(e)),!e.$get)throw Xi("pget","Provider '{0}' must define $get factory method.",t);return x[t+g]=e}function a(t,e){return function(){var n=T.invoke(e,this);if(m(n))throw Xi("undef","Provider '{0}' must return a value from $get factory method.",t);return n}}function s(t,e,n){return i(t,{$get:n!==!1?a(t,e):e})}function u(t,e){return s(t,["$injector",function(t){return t.instantiate(e)}])}function l(t,e){return s(t,v(e),!1)}function c(t,e){ht(t,"constant"),x[t]=e,k[t]=e}function f(t,e){var n=_.get(t+g),r=n.$get;n.$get=function(){var t=T.invoke(r,n);return T.invoke(e,null,{$delegate:t})}}function p(t){ft(m(t)||li(t),"modulesToLoad","not an array");var e,n=[];return r(t,function(t){function r(t){var e,n;for(e=0,n=t.length;e<n;e++){var r=t[e],i=_.get(r[0]);i[r[1]].apply(i,r[2])}}if(!$.get(t)){$.put(t,!0);try{w(t)?(e=Qr(t),n=n.concat(p(e.requires)).concat(e._runBlocks),r(e._invokeQueue),r(e._configBlocks)):C(t)?n.push(_.invoke(t)):li(t)?n.push(_.invoke(t)):pt(t,"module")}catch(e){throw li(t)&&(t=t[t.length-1]),e.message&&e.stack&&e.stack.indexOf(e.message)===-1&&(e=e.message+"\n"+e.stack),Xi("modulerr","Failed to instantiate module {0} due to:\n{1}",t,e.stack||e.message||e)}}}),n}function h(t,n){function r(e,r){if(t.hasOwnProperty(e)){if(t[e]===d)throw Xi("cdep","Circular dependency found: {0}",e+" <- "+y.join(" <- "));return t[e]}try{return y.unshift(e),t[e]=d,t[e]=n(e,r),t[e]}catch(n){throw t[e]===d&&delete t[e],n}finally{y.shift()}}function i(t,n,i){for(var o=[],a=le.$$annotate(t,e,i),s=0,u=a.length;s<u;s++){var l=a[s];if("string"!=typeof l)throw Xi("itkn","Incorrect injection token! Expected service name as string, got {0}",l);o.push(n&&n.hasOwnProperty(l)?n[l]:r(l,i))}return o}function o(t){if(Kr||"function"!=typeof t)return!1;var e=t.$$ngIsClass;return A(e)||(e=t.$$ngIsClass=/^(?:class\b|constructor\()/.test(oe(t))),e}function a(t,e,n,r){"string"==typeof n&&(r=n,n=null);var a=i(t,n,r);return li(t)&&(t=t[t.length-1]),o(t)?(a.unshift(null),new(Function.prototype.bind.apply(t,a))):t.apply(e,a)}function s(t,e,n){var r=li(t)?t[t.length-1]:t,o=i(t,e,n);return o.unshift(null),new(Function.prototype.bind.apply(r,o))}return{invoke:a,instantiate:s,get:r,annotate:le.$$annotate,has:function(e){return x.hasOwnProperty(e+g)||t.hasOwnProperty(e)}}}e=e===!0;var d={},g="Provider",y=[],$=new ie([],!0),x={$provide:{provider:n(i),factory:n(s),service:n(u),value:n(l),constant:n(c),decorator:f}},_=x.$injector=h(x,function(t,e){throw ai.isString(e)&&y.push(e),Xi("unpr","Unknown provider: {0}",y.join(" <- "))}),k={},S=h(k,function(t,e){var n=_.get(t+g,e);return T.invoke(n.$get,n,void 0,t)}),T=S;x["$injector"+g]={$get:v(S)};var E=p(t);return T=S.get("$injector"),T.strictDi=e,r(E,function(t){t&&T.invoke(t)}),T}function ce(){var t=!0;this.disableAutoScrolling=function(){t=!1},this.$get=["$window","$location","$rootScope",function(e,n,r){function i(t){var e=null;return Array.prototype.some.call(t,function(t){if("a"===j(t))return e=t,!0}),e}function o(){var t=s.yOffset;if(C(t))t=t();else if(R(t)){var n=t[0],r=e.getComputedStyle(n);t="fixed"!==r.position?0:n.getBoundingClientRect().bottom}else x(t)||(t=0);return t}function a(t){if(t){t.scrollIntoView();var n=o();if(n){var r=t.getBoundingClientRect().top;e.scrollBy(0,r-n)}}else e.scrollTo(0,0)}function s(t){t=w(t)?t:x(t)?t.toString():n.hash();var e;t?(e=u.getElementById(t))?a(e):(e=i(u.getElementsByName(t)))?a(e):"top"===t&&a(null):a(null)}var u=e.document;return t&&r.$watch(function(){return n.hash()},function(t,e){t===e&&""===t||Xt(function(){r.$evalAsync(s)})}),s}]}function fe(t,e){return t||e?t?e?(li(t)&&(t=t.join(" ")),li(e)&&(e=e.join(" ")),t+" "+e):t:e:""}function pe(t){for(var e=0;e<t.length;e++){var n=t[e];if(n.nodeType===Ji)return n}}function he(t){w(t)&&(t=t.split(" "));var e=gt();return r(t,function(t){t.length&&(e[t]=!0)}),e}function de(t){return b(t)?t:{}}function ve(t,e,n,i){function o(t){try{t.apply(null,H(arguments,1))}finally{if(y--,0===y)for(;b.length;)try{b.pop()()}catch(t){n.error(t)}}}function a(t){var e=t.indexOf("#");return e===-1?"":t.substr(e)}function s(){C=null,u(),l()}function u(){$=k(),$=m($)?null:$,U($,E)&&($=E),E=$}function l(){x===c.url()&&w===$||(x=c.url(),w=$,r(S,function(t){t(c.url(),$)}))}var c=this,f=t.location,p=t.history,d=t.setTimeout,v=t.clearTimeout,g={};c.isMock=!1;var y=0,b=[];c.$$completeOutstandingRequest=o,c.$$incOutstandingRequestCount=function(){y++},c.notifyWhenNoOutstandingRequests=function(t){0===y?t():b.push(t)};var $,w,x=f.href,_=e.find("base"),C=null,k=i.history?function(){try{return p.state}catch(t){}}:h;u(),w=$,c.url=function(e,n,r){if(m(r)&&(r=null),f!==t.location&&(f=t.location),p!==t.history&&(p=t.history),e){var o=w===r;if(x===e&&(!i.history||o))return c;var s=x&&Ke(x)===Ke(e);return x=e,w=r,!i.history||s&&o?(s||(C=e),n?f.replace(e):s?f.hash=a(e):f.href=e,f.href!==e&&(C=e)):(p[n?"replaceState":"pushState"](r,"",e),u(),w=$),C&&(C=e),c}return C||f.href.replace(/%27/g,"'")},c.state=function(){return $};var S=[],T=!1,E=null;c.onUrlChange=function(e){return T||(i.history&&Jr(t).on("popstate",s),Jr(t).on("hashchange",s),T=!0),S.push(e),e},c.$$applicationDestroyed=function(){Jr(t).off("hashchange popstate",s)},c.$$checkUrlChange=l,c.baseHref=function(){var t=_.attr("href");return t?t.replace(/^(https?:)?\/\/[^\/]*/,""):""},c.defer=function(t,e){var n;return y++,n=d(function(){delete g[n],o(t)},e||0),g[n]=!0,n},c.defer.cancel=function(t){return!!g[t]&&(delete g[t],v(t),o(h),!0)}}function ge(){this.$get=["$window","$log","$sniffer","$document",function(t,e,n,r){return new ve(t,r,e,n)}]}function me(){this.$get=function(){function t(t,r){function i(t){t!==p&&(h?h===t&&(h=t.n):h=t,o(t.n,t.p),o(t,p),p=t,p.n=null)}function o(t,e){t!==e&&(t&&(t.p=e),e&&(e.n=t))}if(t in n)throw e("$cacheFactory")("iid","CacheId '{0}' is already taken!",t);var a=0,s=l({},r,{id:t}),u=gt(),c=r&&r.capacity||Number.MAX_VALUE,f=gt(),p=null,h=null;return n[t]={put:function(t,e){if(!m(e)){if(c<Number.MAX_VALUE){var n=f[t]||(f[t]={key:t});i(n)}return t in u||a++,u[t]=e,a>c&&this.remove(h.key),e}},get:function(t){if(c<Number.MAX_VALUE){var e=f[t];if(!e)return;i(e)}return u[t]},remove:function(t){if(c<Number.MAX_VALUE){var e=f[t];if(!e)return;e===p&&(p=e.p),e===h&&(h=e.n),o(e.n,e.p),delete f[t]}t in u&&(delete u[t],a--)},removeAll:function(){u=gt(),a=0,f=gt(),p=h=null},destroy:function(){u=null,s=null,f=null,delete n[t]},info:function(){return l({},s,{size:a})}}}var n={};return t.info=function(){var t={};return r(n,function(e,n){t[n]=e.info()}),t},t.get=function(t){return n[t]},t}}function ye(){this.$get=["$cacheFactory",function(t){return t("templates")}]}function be(){}function $e(e,n){function i(t,e,n){var i=/^\s*([@&<]|=(\*?))(\??)\s*(\w*)\s*$/,o=gt();return r(t,function(t,r){if(t in E)return void(o[r]=E[t]);var a=t.match(i);if(!a)throw oo("iscp","Invalid {3} for directive '{0}'. Definition: {... {1}: '{2}' ...}",e,r,t,n?"controller bindings definition":"isolate scope definition");o[r]={mode:a[1][0],collection:"*"===a[2],optional:"?"===a[3],attrName:a[4]||r},a[4]&&(E[t]=o[r])}),o}function a(t,e){var n={isolateScope:null,bindToController:null};if(b(t.scope)&&(t.bindToController===!0?(n.bindToController=i(t.scope,e,!0),n.isolateScope={}):n.isolateScope=i(t.scope,e,!1)),b(t.bindToController)&&(n.bindToController=i(t.bindToController,e,!0)),n.bindToController&&!t.controller)throw oo("noctrl","Cannot bind to controller without directive '{0}'s controller.",e);return n}function s(t){var e=t.charAt(0);if(!e||e!==Wr(e))throw oo("baddir","Directive/Component name '{0}' is invalid. The first character must be a lowercase letter",t);if(t!==t.trim())throw oo("baddir","Directive/Component name '{0}' is invalid. The name should not contain leading or trailing whitespaces",t)}function u(t){var e=t.require||t.controller&&t.name;return!li(e)&&b(e)&&r(e,function(t,n){var r=t.match(k),i=t.substring(r[0].length);i||(e[n]=r[0]+n)}),e}function c(t,e){if(t&&(!w(t)||!/[EACM]/.test(t)))throw oo("badrestrict","Restrict property '{0}' of directive '{1}' is invalid",t,e);return t||"EA"}var f={},g="Directive",$=/^\s*directive:\s*([\w-]+)\s+(.*)$/,x=/(([\w-]+)(?::([^;]+))?;?)/,_=I("ngSrc,ngSrcset,src,srcset"),k=/^(?:(\^\^?)?(\?)?(\^\^?)?)?/,S=/^(on[a-z]+|formaction)$/,E=gt();this.directive=function t(n,i){return ft(n,"name"),ht(n,"directive"),w(n)?(s(n),ft(i,"directiveFactory"),f.hasOwnProperty(n)||(f[n]=[],e.factory(n+g,["$injector","$exceptionHandler",function(t,e){var i=[];return r(f[n],function(r,o){try{var a=t.invoke(r);C(a)?a={compile:v(a)}:!a.compile&&a.link&&(a.compile=v(a.link)),a.priority=a.priority||0,a.index=o,a.name=a.name||n,a.require=u(a),a.restrict=c(a.restrict,n),a.$$moduleName=r.$$moduleName,i.push(a)}catch(t){e(t)}}),i}])),f[n].push(i)):r(n,o(t)),this},this.component=function(t,e){function n(t){function n(e){return C(e)||li(e)?function(n,r){return t.invoke(e,this,{$element:n,$attrs:r})}:e}var o=e.template||e.templateUrl?e.template:"",a={controller:i,controllerAs:ke(e.controller)||e.controllerAs||"$ctrl",template:n(o),templateUrl:n(e.templateUrl),transclude:e.transclude,scope:{},bindToController:e.bindings||{},restrict:"E",require:e.require};return r(e,function(t,e){"$"===e.charAt(0)&&(a[e]=t)}),a}var i=e.controller||function(){};return r(e,function(t,e){"$"===e.charAt(0)&&(n[e]=t,C(i)&&(i[e]=t))}),n.$inject=["$injector"],this.directive(t,n)},this.aHrefSanitizationWhitelist=function(t){return y(t)?(n.aHrefSanitizationWhitelist(t),this):n.aHrefSanitizationWhitelist()},this.imgSrcSanitizationWhitelist=function(t){return y(t)?(n.imgSrcSanitizationWhitelist(t),this):n.imgSrcSanitizationWhitelist()};var M=!0;this.debugInfoEnabled=function(t){return y(t)?(M=t,this):M};var D=!1;this.preAssignBindingsEnabled=function(t){return y(t)?(D=t,this):D};var O=10;this.onChangesTtl=function(t){return arguments.length?(O=t,this):O};var N=!0;this.commentDirectivesEnabled=function(t){return arguments.length?(N=t,this):N};var P=!0;this.cssClassDirectivesEnabled=function(t){return arguments.length?(P=t,this):P},this.$get=["$injector","$interpolate","$exceptionHandler","$templateRequest","$parse","$controller","$rootScope","$sce","$animate","$$sanitizeUri",function(e,n,i,o,s,u,c,v,E,R){function I(){try{if(!--Mt)throw _t=void 0,oo("infchng","{0} $onChanges() iterations reached. Aborting!\n",O);c.$apply(function(){for(var t=[],e=0,n=_t.length;e<n;++e)try{_t[e]()}catch(e){t.push(e)}if(_t=void 0,t.length)throw t})}finally{Mt++}}function F(t,e){if(e){var n,r,i,o=Object.keys(e);for(n=0,r=o.length;n<r;n++)i=o[n],this[i]=e[i]}else this.$attr={};this.$$element=t}function V(t,e,n){kt.innerHTML="<span "+e+">";var r=kt.firstChild.attributes,i=r[0];r.removeNamedItem(i.name),i.value=n,t.attributes.setNamedItem(i)}function L(t,e){try{t.addClass(e)}catch(t){}}function z(t,e,n,r,i){t instanceof Jr||(t=Jr(t));var o=Y(t,e,t,n,r,i);z.$$addScopeClass(t);var a=null;return function(e,n,r){if(!t)throw oo("multilink","This element has already been linked.");ft(e,"scope"),i&&i.needsNewScope&&(e=e.$parent.$new()),r=r||{};var s=r.parentBoundTranscludeFn,u=r.transcludeControllers,l=r.futureParentElement;s&&s.$$boundTransclude&&(s=s.$$boundTransclude),a||(a=W(l));var c;if(c="html"!==a?Jr(vt(a,Jr("<div>").append(t).html())):n?qi.clone.call(t):t,u)for(var f in u)c.data("$"+f+"Controller",u[f].instance);return z.$$addScopeInfo(c,e),n&&n(c,e),o&&o(e,c,c,s),n||(t=o=null),c}}function W(t){var e=t&&t[0];return e&&"foreignobject"!==j(e)&&ri.call(e).match(/SVG/)?"svg":"html"}function Y(t,e,n,r,i,o){function a(t,n,r,i){var o,a,s,u,l,c,f,p,v;if(h){var g=n.length;for(v=new Array(g),l=0;l<d.length;l+=3)f=d[l],v[f]=n[f]}else v=n;for(l=0,c=d.length;l<c;)s=v[d[l++]],o=d[l++],a=d[l++],o?(o.scope?(u=t.$new(),z.$$addScopeInfo(Jr(s),u)):u=t,p=o.transcludeOnThisElement?X(t,o.transclude,i):!o.templateOnThisElement&&i?i:!i&&e?X(t,e):null,o(a,u,s,r,p)):a&&a(t,s.childNodes,void 0,i)}for(var s,u,l,c,f,p,h,d=[],v=li(t)||t instanceof Jr,g=0;g<t.length;g++)s=new F,11===Kr&&G(t,g,v),u=K(t[g],[],s,0===g?r:void 0,i),l=u.length?nt(u,t[g],s,e,n,null,[],[],o):null,l&&l.scope&&z.$$addScopeClass(s.$$element),f=l&&l.terminal||!(c=t[g].childNodes)||!c.length?null:Y(c,l?(l.transcludeOnThisElement||!l.templateOnThisElement)&&l.transclude:e),
(l||f)&&(d.push(g,l,f),p=!0,h=h||l),o=null;return p?a:null}function G(t,e,n){var r,i=t[e],o=i.parentNode;if(i.nodeType===xi)for(;;){if(r=o?i.nextSibling:t[e+1],!r||r.nodeType!==xi)break;i.nodeValue=i.nodeValue+r.nodeValue,r.parentNode&&r.parentNode.removeChild(r),n&&r===t[e+1]&&t.splice(e+1,1)}}function X(t,e,n){function r(r,i,o,a,s){return r||(r=t.$new(!1,s),r.$$transcluded=!0),e(r,i,{parentBoundTranscludeFn:n,transcludeControllers:o,futureParentElement:a})}var i=r.$$slots=gt();for(var o in e.$$slots)e.$$slots[o]?i[o]=X(t,e.$$slots[o],n):i[o]=null;return r}function K(t,e,n,r,i){var o,a,s,u=t.nodeType,l=n.$attr;switch(u){case $i:a=j(t),at(e,xe(a),"E",r,i);for(var c,f,p,h,d,v,g=t.attributes,m=0,y=g&&g.length;m<y;m++){var $=!1,_=!1;c=g[m],f=c.name,d=c.value,h=xe(f),v=Nt.test(h),v&&(f=f.replace(so,"").substr(8).replace(/_(.)/g,function(t,e){return e.toUpperCase()}));var C=h.match(Pt);C&&st(C[1])&&($=f,_=f.substr(0,f.length-5)+"end",f=f.substr(0,f.length-6)),p=xe(f.toLowerCase()),l[p]=f,!v&&n.hasOwnProperty(p)||(n[p]=d,Jt(t,p)&&(n[p]=!0)),yt(t,e,d,p,v),at(e,p,"A",r,i,$,_)}if("input"===a&&"hidden"===t.getAttribute("type")&&t.setAttribute("autocomplete","off"),!Et)break;if(s=t.className,b(s)&&(s=s.animVal),w(s)&&""!==s)for(;o=x.exec(s);)p=xe(o[2]),at(e,p,"C",r,i)&&(n[p]=fi(o[3])),s=s.substr(o.index+o[0].length);break;case xi:dt(e,t.nodeValue);break;case _i:if(!St)break;Z(t,e,n,r,i)}return e.sort(pt),e}function Z(t,e,n,r,i){try{var o=$.exec(t.nodeValue);if(o){var a=xe(o[1]);at(e,a,"M",r,i)&&(n[a]=fi(o[2]))}}catch(t){}}function Q(t,e,n){var r=[],i=0;if(e&&t.hasAttribute&&t.hasAttribute(e)){do{if(!t)throw oo("uterdir","Unterminated attribute, found '{0}' but no matching '{1}' found.",e,n);t.nodeType===$i&&(t.hasAttribute(e)&&i++,t.hasAttribute(n)&&i--),r.push(t),t=t.nextSibling}while(i>0)}else r.push(t);return Jr(r)}function tt(t,e,n){return function(r,i,o,a,s){return i=Q(i[0],e,n),t(r,i,o,a,s)}}function et(t,e,n,r,i,o){var a;return t?z(e,n,r,i,o):function(){return a||(a=z(e,n,r,i,o),e=n=o=null),a.apply(this,arguments)}}function nt(t,e,n,o,a,s,u,c,f){function p(t,e,n,r){t&&(n&&(t=tt(t,n,r)),t.require=d.require,t.directiveName=v,(k===d||d.$$isolateScope)&&(t=$t(t,{isolateScope:!0})),u.push(t)),e&&(n&&(e=tt(e,n,r)),e.require=d.require,e.directiveName=v,(k===d||d.$$isolateScope)&&(e=$t(e,{isolateScope:!0})),c.push(e))}function h(t,o,a,s,f){function p(t,e,n,r){var i;if(T(t)||(r=n,n=e,e=t,t=void 0),O&&(i=$),n||(n=O?E.parent():E),!r)return f(t,e,i,n,q);var o=f.$$slots[r];if(o)return o(t,e,i,n,q);if(m(o))throw oo("noslot",'No parent directive that requires a transclusion with slot name "{0}". Element: {1}',r,J(E))}var h,d,v,g,y,$,w,E,M,A;e===a?(M=n,E=n.$$element):(E=Jr(a),M=new F(E,n)),y=o,k?g=o.$new(!0):x&&(y=o.$parent),f&&(w=p,w.$$boundTransclude=f,w.isSlotFilled=function(t){return!!f.$$slots[t]}),_&&($=it(E,M,w,_,g,o,k)),k&&(z.$$addScopeInfo(E,g,!0,!(S&&(S===k||S===k.$$originalDirective))),z.$$addScopeClass(E,!0),g.$$isolateBindings=k.$$isolateBindings,A=xt(o,M,g,g.$$isolateBindings,k),A.removeWatches&&g.$on("$destroy",A.removeWatches));for(var N in $){var P=_[N],R=$[N],I=P.$$bindings.bindToController;if(D){I?R.bindingInfo=xt(y,M,R.instance,I,P):R.bindingInfo={};var j=R();j!==R.instance&&(R.instance=j,E.data("$"+P.name+"Controller",j),R.bindingInfo.removeWatches&&R.bindingInfo.removeWatches(),R.bindingInfo=xt(y,M,R.instance,I,P))}else R.instance=R(),E.data("$"+P.name+"Controller",R.instance),R.bindingInfo=xt(y,M,R.instance,I,P)}for(r(_,function(t,e){var n=t.require;t.bindToController&&!li(n)&&b(n)&&l($[e].instance,rt(e,n,E,$))}),r($,function(t){var e=t.instance;if(C(e.$onChanges))try{e.$onChanges(t.bindingInfo.initialChanges)}catch(t){i(t)}if(C(e.$onInit))try{e.$onInit()}catch(t){i(t)}C(e.$doCheck)&&(y.$watch(function(){e.$doCheck()}),e.$doCheck()),C(e.$onDestroy)&&y.$on("$destroy",function(){e.$onDestroy()})}),h=0,d=u.length;h<d;h++)v=u[h],wt(v,v.isolateScope?g:o,E,M,v.require&&rt(v.directiveName,v.require,E,$),w);var q=o;for(k&&(k.template||null===k.templateUrl)&&(q=g),t&&t(q,a.childNodes,void 0,f),h=c.length-1;h>=0;h--)v=c[h],wt(v,v.isolateScope?g:o,E,M,v.require&&rt(v.directiveName,v.require,E,$),w);r($,function(t){var e=t.instance;C(e.$postLink)&&e.$postLink()})}f=f||{};for(var d,v,g,y,$,w=-Number.MAX_VALUE,x=f.newScopeDirective,_=f.controllerDirectives,k=f.newIsolateScopeDirective,S=f.templateDirective,E=f.nonTlbTranscludeDirective,M=!1,A=!1,O=f.hasElementTranscludeDirective,N=n.$$element=Jr(e),P=s,R=o,I=!1,q=!1,V=0,U=t.length;V<U;V++){d=t[V];var L=d.$$start,W=d.$$end;if(L&&(N=Q(e,L,W)),g=void 0,w>d.priority)break;if($=d.scope,$&&(d.templateUrl||(b($)?(ht("new/isolated scope",k||x,d,N),k=d):ht("new/isolated scope",k,d,N)),x=x||d),v=d.name,!I&&(d.replace&&(d.templateUrl||d.template)||d.transclude&&!d.$$tlb)){for(var Y,G=V+1;Y=t[G++];)if(Y.transclude&&!Y.$$tlb||Y.replace&&(Y.templateUrl||Y.template)){q=!0;break}I=!0}if(!d.templateUrl&&d.controller&&(_=_||gt(),ht("'"+v+"' controller",_[v],d,N),_[v]=d),$=d.transclude)if(M=!0,d.$$tlb||(ht("transclusion",E,d,N),E=d),"element"===$)O=!0,w=d.priority,g=N,N=n.$$element=Jr(z.$$createComment(v,n[v])),e=N[0],bt(a,H(g),e),g[0].$$parentNode=g[0].parentNode,R=et(q,g,o,w,P&&P.name,{nonTlbTranscludeDirective:E});else{var X=gt();if(b($)){g=[];var Z=gt(),nt=gt();r($,function(t,e){var n="?"===t.charAt(0);t=n?t.substring(1):t,Z[t]=e,X[e]=null,nt[e]=n}),r(N.contents(),function(t){var e=Z[xe(j(t))];e?(nt[e]=!0,X[e]=X[e]||[],X[e].push(t)):g.push(t)}),r(nt,function(t,e){if(!t)throw oo("reqslot","Required transclusion slot `{0}` was not filled.",e)});for(var at in X)X[at]&&(X[at]=et(q,X[at],o))}else g=Jr(Rt(e)).contents();N.empty(),R=et(q,g,o,void 0,void 0,{needsNewScope:d.$$isolateScope||d.$$newScope}),R.$$slots=X}if(d.template)if(A=!0,ht("template",S,d,N),S=d,$=C(d.template)?d.template(N,n):d.template,$=Ot($),d.replace){if(P=d,g=Tt($)?[]:Ce(vt(d.templateNamespace,fi($))),e=g[0],1!==g.length||e.nodeType!==$i)throw oo("tplrt","Template for directive '{0}' must have exactly one root element. {1}",v,"");bt(a,N,e);var st={$attr:{}},lt=K(e,[],st),ft=t.splice(V+1,t.length-(V+1));(k||x)&&ot(lt,k,x),t=t.concat(lt).concat(ft),ut(n,st),U=t.length}else N.html($);if(d.templateUrl)A=!0,ht("template",S,d,N),S=d,d.replace&&(P=d),h=ct(t.splice(V,t.length-V),N,n,a,M&&R,u,c,{controllerDirectives:_,newScopeDirective:x!==d&&x,newIsolateScopeDirective:k,templateDirective:S,nonTlbTranscludeDirective:E}),U=t.length;else if(d.compile)try{y=d.compile(N,n,R);var pt=d.$$originalDirective||d;C(y)?p(null,B(pt,y),L,W):y&&p(B(pt,y.pre),B(pt,y.post),L,W)}catch(t){i(t,J(N))}d.terminal&&(h.terminal=!0,w=Math.max(w,d.priority))}return h.scope=x&&x.scope===!0,h.transcludeOnThisElement=M,h.templateOnThisElement=A,h.transclude=R,f.hasElementTranscludeDirective=O,h}function rt(t,e,n,i){var o;if(w(e)){var a=e.match(k),s=e.substring(a[0].length),u=a[1]||a[3],l="?"===a[2];if("^^"===u?n=n.parent():(o=i&&i[s],o=o&&o.instance),!o){var c="$"+s+"Controller";o=u?n.inheritedData(c):n.data(c)}if(!o&&!l)throw oo("ctreq","Controller '{0}', required by directive '{1}', can't be found!",s,t)}else if(li(e)){o=[];for(var f=0,p=e.length;f<p;f++)o[f]=rt(t,e[f],n,i)}else b(e)&&(o={},r(e,function(e,r){o[r]=rt(t,e,n,i)}));return o||null}function it(t,e,n,r,i,o,a){var s=gt();for(var l in r){var c=r[l],f={$scope:c===a||c.$$isolateScope?i:o,$element:t,$attrs:e,$transclude:n},p=c.controller;"@"===p&&(p=e[c.name]);var h=u(p,f,!0,c.controllerAs);s[c.name]=h,t.data("$"+c.name+"Controller",h.instance)}return s}function ot(t,e,n){for(var r=0,i=t.length;r<i;r++)t[r]=p(t[r],{$$isolateScope:e,$$newScope:n})}function at(t,n,r,i,o,s,u){if(n===o)return null;var l=null;if(f.hasOwnProperty(n))for(var c,h=e.get(n+g),d=0,v=h.length;d<v;d++)if(c=h[d],(m(i)||i>c.priority)&&c.restrict.indexOf(r)!==-1){if(s&&(c=p(c,{$$start:s,$$end:u})),!c.$$bindings){var y=c.$$bindings=a(c,c.name);b(y.isolateScope)&&(c.$$isolateBindings=y.isolateScope)}t.push(c),l=c}return l}function st(t){if(f.hasOwnProperty(t))for(var n,r=e.get(t+g),i=0,o=r.length;i<o;i++)if(n=r[i],n.multiElement)return!0;return!1}function ut(t,e){var n=e.$attr,i=t.$attr;r(t,function(r,i){"$"!==i.charAt(0)&&(e[i]&&e[i]!==r&&(r.length?r+=("style"===i?";":" ")+e[i]:r=e[i]),t.$set(i,r,!0,n[i]))}),r(e,function(e,r){t.hasOwnProperty(r)||"$"===r.charAt(0)||(t[r]=e,"class"!==r&&"style"!==r&&(i[r]=n[r]))})}function ct(t,e,n,a,s,u,l,c){var f,d,v=[],g=e[0],m=t.shift(),y=p(m,{templateUrl:null,transclude:null,replace:null,$$originalDirective:m}),$=C(m.templateUrl)?m.templateUrl(e,n):m.templateUrl,w=m.templateNamespace;return e.empty(),o($).then(function(i){var o,p,h,x;if(i=Ot(i),m.replace){if(h=Tt(i)?[]:Ce(vt(w,fi(i))),o=h[0],1!==h.length||o.nodeType!==$i)throw oo("tplrt","Template for directive '{0}' must have exactly one root element. {1}",m.name,$);p={$attr:{}},bt(a,e,o);var _=K(o,[],p);b(m.scope)&&ot(_,!0),t=_.concat(t),ut(n,p)}else o=g,e.html(i);for(t.unshift(y),f=nt(t,o,n,s,e,m,u,l,c),r(a,function(t,n){t===o&&(a[n]=e[0])}),d=Y(e[0].childNodes,s);v.length;){var C=v.shift(),k=v.shift(),S=v.shift(),T=v.shift(),E=e[0];if(!C.$$destroyed){if(k!==g){var M=k.className;c.hasElementTranscludeDirective&&m.replace||(E=Rt(o)),bt(S,Jr(k),E),L(Jr(E),M)}x=f.transcludeOnThisElement?X(C,f.transclude,T):T,f(d,C,E,a,x)}}v=null}).catch(function(t){t instanceof Error&&i(t)}).catch(h),function(t,e,n,r,i){var o=i;e.$$destroyed||(v?v.push(e,n,r,o):(f.transcludeOnThisElement&&(o=X(e,f.transclude,i)),f(d,e,n,r,o)))}}function pt(t,e){var n=e.priority-t.priority;return 0!==n?n:t.name!==e.name?t.name<e.name?-1:1:t.index-e.index}function ht(t,e,n,r){function i(t){return t?" (module: "+t+")":""}if(e)throw oo("multidir","Multiple directives [{0}{1}, {2}{3}] asking for {4} on: {5}",e.name,i(e.$$moduleName),n.name,i(n.$$moduleName),t,J(r))}function dt(t,e){var r=n(e,!0);r&&t.push({priority:0,compile:function(t){var e=t.parent(),n=!!e.length;return n&&z.$$addBindingClass(e),function(t,e){var i=e.parent();n||z.$$addBindingClass(i),z.$$addBindingInfo(i,r.expressions),t.$watch(r,function(t){e[0].nodeValue=t})}}})}function vt(e,n){switch(e=Wr(e||"html")){case"svg":case"math":var r=t.document.createElement("div");return r.innerHTML="<"+e+">"+n+"</"+e+">",r.childNodes[0].childNodes;default:return n}}function mt(t,e){if("srcdoc"===e)return v.HTML;var n=j(t);if("src"===e||"ngSrc"===e){if(["img","video","audio","source","track"].indexOf(n)===-1)return v.RESOURCE_URL}else if("xlinkHref"===e||"form"===n&&"action"===e||"link"===n&&"href"===e)return v.RESOURCE_URL}function yt(t,e,r,i,o){var a=mt(t,i),s=!o,u=_[i]||o,l=n(r,s,a,u);if(l){if("multiple"===i&&"select"===j(t))throw oo("selmulti","Binding to the 'multiple' attribute is not supported. Element: {0}",J(t));if(S.test(i))throw oo("nodomevents","Interpolations for HTML DOM event attributes are disallowed.  Please use the ng- versions (such as ng-click instead of onclick) instead.");e.push({priority:100,compile:function(){return{pre:function(t,e,o){var s=o.$$observers||(o.$$observers=gt()),c=o[i];c!==r&&(l=c&&n(c,!0,a,u),r=c),l&&(o[i]=l(t),(s[i]||(s[i]=[])).$$inter=!0,(o.$$observers&&o.$$observers[i].$$scope||t).$watch(l,function(t,e){"class"===i&&t!==e?o.$updateClass(t,e):o.$set(i,t)}))}}}})}}function bt(e,n,r){var i,o,a=n[0],s=n.length,u=a.parentNode;if(e)for(i=0,o=e.length;i<o;i++)if(e[i]===a){e[i++]=r;for(var l=i,c=l+s-1,f=e.length;l<f;l++,c++)c<f?e[l]=e[c]:delete e[l];e.length-=s-1,e.context===a&&(e.context=r);break}u&&u.replaceChild(r,a);var p=t.document.createDocumentFragment();for(i=0;i<s;i++)p.appendChild(n[i]);for(Jr.hasData(a)&&(Jr.data(r,Jr.data(a)),Jr(a).off("$destroy")),Jr.cleanData(p.querySelectorAll("*")),i=1;i<s;i++)delete n[i];n[0]=r,n.length=1}function $t(t,e){return l(function(){return t.apply(null,arguments)},t,e)}function wt(t,e,n,r,o,a){try{t(e,n,r,o,a)}catch(t){i(t,J(n))}}function xt(t,e,i,o,a){function u(e,n,r){!C(i.$onChanges)||n===r||n!==n&&r!==r||(_t||(t.$$postDigest(I),_t=[]),c||(c={},_t.push(l)),c[e]&&(r=c[e].previousValue),c[e]=new we(r,n))}function l(){i.$onChanges(c),c=void 0}var c,f=[],p={};return r(o,function(r,o){var l,c,d,v,g,m=r.attrName,y=r.optional,b=r.mode;switch(b){case"@":y||zr.call(e,m)||(i[o]=e[m]=void 0),g=e.$observe(m,function(t){if(w(t)||A(t)){var e=i[o];u(o,t,e),i[o]=t}}),e.$$observers[m].$$scope=t,l=e[m],w(l)?i[o]=n(l)(t):A(l)&&(i[o]=l),p[o]=new we(ao,i[o]),f.push(g);break;case"=":if(!zr.call(e,m)){if(y)break;e[m]=void 0}if(y&&!e[m])break;c=s(e[m]),v=c.literal?U:function(t,e){return t===e||t!==t&&e!==e},d=c.assign||function(){throw l=i[o]=c(t),oo("nonassign","Expression '{0}' in attribute '{1}' used with directive '{2}' is non-assignable!",e[m],m,a.name)},l=i[o]=c(t);var $=function(e){return v(e,i[o])||(v(e,l)?d(t,e=i[o]):i[o]=e),l=e};$.$stateful=!0,g=r.collection?t.$watchCollection(e[m],$):t.$watch(s(e[m],$),null,c.literal),f.push(g);break;case"<":if(!zr.call(e,m)){if(y)break;e[m]=void 0}if(y&&!e[m])break;c=s(e[m]);var x=c.literal,_=i[o]=c(t);p[o]=new we(ao,i[o]),g=t.$watch(c,function(t,e){if(e===t){if(e===_||x&&U(e,_))return;e=_}u(o,t,e),i[o]=t},x),f.push(g);break;case"&":if(c=e.hasOwnProperty(m)?s(e[m]):h,c===h&&y)break;i[o]=function(e){return c(t,e)}}}),{initialChanges:p,removeWatches:f.length&&function(){for(var t=0,e=f.length;t<e;++t)f[t]()}}}var _t,Ct=/^\w/,kt=t.document.createElement("div"),St=N,Et=P,Mt=O;F.prototype={$normalize:xe,$addClass:function(t){t&&t.length>0&&E.addClass(this.$$element,t)},$removeClass:function(t){t&&t.length>0&&E.removeClass(this.$$element,t)},$updateClass:function(t,e){var n=_e(t,e);n&&n.length&&E.addClass(this.$$element,n);var r=_e(e,t);r&&r.length&&E.removeClass(this.$$element,r)},$set:function(t,e,n,o){var a,s=this.$$element[0],u=Jt(s,t),l=Zt(t),c=t;if(u?(this.$$element.prop(t,e),o=u):l&&(this[l]=e,c=l),this[t]=e,o?this.$attr[t]=o:(o=this.$attr[t],o||(this.$attr[t]=o=lt(t,"-"))),a=j(this.$$element),"a"===a&&("href"===t||"xlinkHref"===t)||"img"===a&&"src"===t)this[t]=e=R(e,"src"===t);else if("img"===a&&"srcset"===t&&y(e)){for(var f="",p=fi(e),h=/(\s+\d+x\s*,|\s+\d+w\s*,|\s+,|,\s+)/,d=/\s/.test(p)?h:/(,)/,v=p.split(d),g=Math.floor(v.length/2),b=0;b<g;b++){var $=2*b;f+=R(fi(v[$]),!0),f+=" "+fi(v[$+1])}var w=fi(v[2*b]).split(/\s/);f+=R(fi(w[0]),!0),2===w.length&&(f+=" "+fi(w[1])),this[t]=e=f}n!==!1&&(null===e||m(e)?this.$$element.removeAttr(o):Ct.test(o)?this.$$element.attr(o,e):V(this.$$element[0],o,e));var x=this.$$observers;x&&r(x[c],function(t){try{t(e)}catch(t){i(t)}})},$observe:function(t,e){var n=this,r=n.$$observers||(n.$$observers=gt()),i=r[t]||(r[t]=[]);return i.push(e),c.$evalAsync(function(){i.$$inter||!n.hasOwnProperty(t)||m(n[t])||e(n[t])}),function(){q(i,e)}}};var Dt=n.startSymbol(),At=n.endSymbol(),Ot="{{"===Dt&&"}}"===At?d:function(t){return t.replace(/\{\{/g,Dt).replace(/}}/g,At)},Nt=/^ngAttr[A-Z]/,Pt=/^(.+)Start$/;return z.$$addBindingInfo=M?function(t,e){var n=t.data("$binding")||[];li(e)?n=n.concat(e):n.push(e),t.data("$binding",n)}:h,z.$$addBindingClass=M?function(t){L(t,"ng-binding")}:h,z.$$addScopeInfo=M?function(t,e,n,r){var i=n?r?"$isolateScopeNoTemplate":"$isolateScope":"$scope";t.data(i,e)}:h,z.$$addScopeClass=M?function(t,e){L(t,e?"ng-isolate-scope":"ng-scope")}:h,z.$$createComment=function(e,n){var r="";return M&&(r=" "+(e||"")+": ",n&&(r+=n+" ")),t.document.createComment(r)},z}]}function we(t,e){this.previousValue=t,this.currentValue=e}function xe(t){return t.replace(so,"").replace(uo,kt)}function _e(t,e){var n="",r=t.split(/\s+/),i=e.split(/\s+/);t:for(var o=0;o<r.length;o++){for(var a=r[o],s=0;s<i.length;s++)if(a===i[s])continue t;n+=(n.length>0?" ":"")+a}return n}function Ce(t){t=Jr(t);var e=t.length;if(e<=1)return t;for(;e--;){var n=t[e];(n.nodeType===_i||n.nodeType===xi&&""===n.nodeValue.trim())&&ei.call(t,e,1)}return t}function ke(t,e){if(e&&w(e))return e;if(w(t)){var n=co.exec(t);if(n)return n[3]}}function Se(){var t={},n=!1;this.has=function(e){return t.hasOwnProperty(e)},this.register=function(e,n){ht(e,"controller"),b(e)?l(t,e):t[e]=n},this.allowGlobals=function(){n=!0},this.$get=["$injector","$window",function(r,i){function o(t,n,r,i){if(!t||!b(t.$scope))throw e("$controller")("noscp","Cannot export controller '{0}' as '{1}'! No $scope object provided via `locals`.",i,n);t.$scope[n]=r}return function(e,a,s,u){var c,f,p,h;if(s=s===!0,u&&w(u)&&(h=u),w(e)){if(f=e.match(co),!f)throw lo("ctrlfmt","Badly formed controller string '{0}'. Must match `__name__ as __id__` or `__name__`.",e);if(p=f[1],h=h||f[3],e=t.hasOwnProperty(p)?t[p]:dt(a.$scope,p,!0)||(n?dt(i,p,!0):void 0),!e)throw lo("ctrlreg","The controller with the name '{0}' is not registered.",p);pt(e,p,!0)}if(s){var d=(li(e)?e[e.length-1]:e).prototype;return c=Object.create(d||null),h&&o(a,h,c,p||e.name),l(function(){var t=r.invoke(e,c,a,p);return t!==c&&(b(t)||C(t))&&(c=t,h&&o(a,h,c,p||e.name)),c},{instance:c,identifier:h})}return c=r.instantiate(e,a,p),h&&o(a,h,c,p||e.name),c}}]}function Te(){this.$get=["$window",function(t){return Jr(t.document)}]}function Ee(){this.$get=["$document","$rootScope",function(t,e){function n(){i=r.hidden}var r=t[0],i=r&&r.hidden;return t.on("visibilitychange",n),e.$on("$destroy",function(){t.off("visibilitychange",n)}),function(){return i}}]}function Me(){this.$get=["$log",function(t){return function(e,n){t.error.apply(t,arguments)}}]}function De(t){return b(t)?_(t)?t.toISOString():W(t):t}function Ae(){this.$get=function(){return function(t){if(!t)return"";var e=[];return i(t,function(t,n){null===t||m(t)||(li(t)?r(t,function(t){e.push(nt(n)+"="+nt(De(t)))}):e.push(nt(n)+"="+nt(De(t))))}),e.join("&")}}}function Oe(){this.$get=function(){return function(t){function e(t,o,a){null===t||m(t)||(li(t)?r(t,function(t,n){e(t,o+"["+(b(t)?n:"")+"]")}):b(t)&&!_(t)?i(t,function(t,n){e(t,o+(a?"":"[")+n+(a?"":"]"))}):n.push(nt(o)+"="+nt(De(t))))}if(!t)return"";var n=[];return e(t,"",!0),n.join("&")}}}function Ne(t,e){if(w(t)){var n=t.replace(mo,"").trim();if(n){var r=e("Content-Type");(r&&0===r.indexOf(po)||Pe(n))&&(t=Y(n))}}return t}function Pe(t){var e=t.match(vo);return e&&go[e[0]].test(t)}function Re(t){function e(t,e){t&&(i[t]=i[t]?i[t]+", "+e:e)}var n,i=gt();return w(t)?r(t.split("\n"),function(t){n=t.indexOf(":"),e(Wr(fi(t.substr(0,n))),fi(t.substr(n+1)))}):b(t)&&r(t,function(t,n){e(Wr(n),fi(t))}),i}function Ie(t){var e;return function(n){if(e||(e=Re(t)),n){var r=e[Wr(n)];return void 0===r&&(r=null),r}return e}}function je(t,e,n,i){return C(i)?i(t,e,n):(r(i,function(r){t=r(t,e,n)}),t)}function Fe(t){return 200<=t&&t<300}function qe(){var t=this.defaults={transformResponse:[Ne],transformRequest:[function(t){return!b(t)||E(t)||D(t)||M(t)?t:W(t)}],headers:{common:{Accept:"application/json, text/plain, */*"},post:bt(ho),put:bt(ho),patch:bt(ho)},xsrfCookieName:"XSRF-TOKEN",xsrfHeaderName:"X-XSRF-TOKEN",paramSerializer:"$httpParamSerializer",jsonpCallbackParam:"callback"},n=!1;this.useApplyAsync=function(t){return y(t)?(n=!!t,this):n};var i=this.interceptors=[];this.$get=["$browser","$httpBackend","$$cookieReader","$cacheFactory","$rootScope","$q","$injector","$sce",function(o,a,s,u,c,f,p,d){function v(n){function i(t,e){for(var n=0,r=e.length;n<r;){var i=e[n++],o=e[n++];t=t.then(i,o)}return e.length=0,t}function a(){o.$$completeOutstandingRequest(h)}function s(t,e){var n,i={};return r(t,function(t,r){C(t)?(n=t(e),null!=n&&(i[r]=n)):i[r]=t}),i}function u(e){var n,r,i,o=t.headers,a=l({},e.headers);o=l({},o.common,o[Wr(e.method)]);t:for(n in o){r=Wr(n);for(i in a)if(Wr(i)===r)continue t;a[n]=o[n]}return s(a,bt(e))}function c(e){var n=e.headers,i=je(e.data,Ie(n),void 0,e.transformRequest);return m(i)&&r(n,function(t,e){"content-type"===Wr(e)&&delete n[e]}),m(e.withCredentials)&&!m(t.withCredentials)&&(e.withCredentials=t.withCredentials),x(e,i).then(v,v)}function v(t){var e=l({},t);return e.data=je(t.data,t.headers,t.status,g.transformResponse),Fe(t.status)?e:f.reject(e)}if(!b(n))throw e("$http")("badreq","Http request configuration must be an object.  Received: {0}",n);if(!w(d.valueOf(n.url)))throw e("$http")("badreq","Http request configuration url must be a string or a $sce trusted object.  Received: {0}",n.url);var g=l({method:"get",transformRequest:t.transformRequest,transformResponse:t.transformResponse,paramSerializer:t.paramSerializer,jsonpCallbackParam:t.jsonpCallbackParam},n);g.headers=u(n),g.method=Yr(g.method),g.paramSerializer=w(g.paramSerializer)?p.get(g.paramSerializer):g.paramSerializer,o.$$incOutstandingRequestCount();var y=[],$=[],_=f.resolve(g);return r(T,function(t){(t.request||t.requestError)&&y.unshift(t.request,t.requestError),(t.response||t.responseError)&&$.push(t.response,t.responseError)}),_=i(_,y),_=_.then(c),_=i(_,$),_=_.finally(a)}function g(t){r(arguments,function(t){v[t]=function(e,n){return v(l({},n||{},{method:t,url:e}))}})}function $(t){r(arguments,function(t){v[t]=function(e,n,r){return v(l({},r||{},{method:t,url:e,data:n}))}})}function x(e,i){function o(t){if(t){var e={};return r(t,function(t,r){e[r]=function(e){function r(){t(e)}n?c.$applyAsync(r):c.$$phase?r():c.$apply(r)}}),e}}function u(t,e,r,i){function o(){l(e,t,r,i)}g&&(Fe(t)?g.put(M,[t,e,Re(r),i]):g.remove(M)),n?c.$applyAsync(o):(o(),c.$$phase||c.$apply())}function l(t,n,r,i){n=n>=-1?n:0,(Fe(n)?x.resolve:x.reject)({data:t,status:n,headers:Ie(r),config:e,statusText:i})}function p(t){l(t.data,t.status,bt(t.headers()),t.statusText)}function h(){var t=v.pendingRequests.indexOf(e);t!==-1&&v.pendingRequests.splice(t,1)}var g,$,x=f.defer(),C=x.promise,T=e.headers,E="jsonp"===Wr(e.method),M=e.url;if(E?M=d.getTrustedResourceUrl(M):w(M)||(M=d.valueOf(M)),M=_(M,e.paramSerializer(e.params)),E&&(M=k(M,e.jsonpCallbackParam)),v.pendingRequests.push(e),C.then(h,h),!e.cache&&!t.cache||e.cache===!1||"GET"!==e.method&&"JSONP"!==e.method||(g=b(e.cache)?e.cache:b(t.cache)?t.cache:S),g&&($=g.get(M),y($)?O($)?$.then(p,p):li($)?l($[1],$[0],bt($[2]),$[3]):l($,200,{},"OK"):g.put(M,C)),m($)){var D=Fn(e.url)?s()[e.xsrfCookieName||t.xsrfCookieName]:void 0;D&&(T[e.xsrfHeaderName||t.xsrfHeaderName]=D),a(e.method,M,i,u,T,e.timeout,e.withCredentials,e.responseType,o(e.eventHandlers),o(e.uploadEventHandlers))}return C}function _(t,e){return e.length>0&&(t+=(t.indexOf("?")===-1?"?":"&")+e),t}function k(t,e){if(/[&?][^=]+=JSON_CALLBACK/.test(t))throw yo("badjsonp",'Illegal use of JSON_CALLBACK in url, "{0}"',t);var n=new RegExp("[&?]"+e+"=");if(n.test(t))throw yo("badjsonp",'Illegal use of callback param, "{0}", in url, "{1}"',e,t);return t+=(t.indexOf("?")===-1?"?":"&")+e+"=JSON_CALLBACK"}var S=u("$http");t.paramSerializer=w(t.paramSerializer)?p.get(t.paramSerializer):t.paramSerializer;var T=[];return r(i,function(t){T.unshift(w(t)?p.get(t):p.invoke(t))}),v.pendingRequests=[],g("get","delete","head","jsonp"),$("post","put","patch"),v.defaults=t,v}]}function Ve(){this.$get=function(){return function(){return new t.XMLHttpRequest}}}function Ue(){this.$get=["$browser","$jsonpCallbacks","$document","$xhrFactory",function(t,e,n,r){return Le(t,r,t.defer,e,n[0])}]}function Le(t,e,n,i,o){function a(t,e,n){t=t.replace("JSON_CALLBACK",e);var r=o.createElement("script"),a=null;return r.type="text/javascript",r.src=t,r.async=!0,a=function(t){r.removeEventListener("load",a),r.removeEventListener("error",a),o.body.removeChild(r),r=null;var s=-1,u="unknown";t&&("load"!==t.type||i.wasCalled(e)||(t={type:"error"}),u=t.type,s="error"===t.type?404:200),n&&n(s,u)},r.addEventListener("load",a),r.addEventListener("error",a),o.body.appendChild(r),a}return function(o,s,u,l,c,f,p,h,d,v){function g(){w&&w(),x&&x.abort()}function b(t,e,r,i,o){y(C)&&n.cancel(C),w=x=null,t(e,r,i,o)}if(s=s||t.url(),"jsonp"===Wr(o))var $=i.createCallback(s),w=a(s,$,function(t,e){var n=200===t&&i.getResponse($);b(l,t,n,"",e),i.removeCallback($)});else{var x=e(o,s);x.open(o,s,!0),r(c,function(t,e){y(t)&&x.setRequestHeader(e,t)}),x.onload=function(){var t=x.statusText||"",e="response"in x?x.response:x.responseText,n=1223===x.status?204:x.status;0===n&&(n=e?200:"file"===jn(s).protocol?404:0),b(l,n,e,x.getAllResponseHeaders(),t)};var _=function(){b(l,-1,null,null,"")};if(x.onerror=_,x.onabort=_,x.ontimeout=_,r(d,function(t,e){x.addEventListener(e,t)}),r(v,function(t,e){x.upload.addEventListener(e,t)}),p&&(x.withCredentials=!0),h)try{x.responseType=h}catch(t){if("json"!==h)throw t}x.send(m(u)?null:u)}if(f>0)var C=n(g,f);else O(f)&&f.then(g)}}function He(){var t="{{",e="}}";this.startSymbol=function(e){return e?(t=e,this):t},this.endSymbol=function(t){return t?(e=t,this):e},this.$get=["$parse","$exceptionHandler","$sce",function(n,r,i){function o(t){return"\\\\\\"+t}function a(n){return n.replace(p,t).replace(h,e)}function s(t,e,n,r){var i=t.$watch(function(t){return i(),r(t)},e,n);return i}function u(o,u,p,h){function d(t){try{return t=A(t),h&&!y(t)?t:mt(t)}catch(t){r(bo.interr(o,t))}}if(!o.length||o.indexOf(t)===-1){var g;if(!u){var b=a(o);g=v(b),g.exp=o,g.expressions=[],g.$$watchDelegate=s}return g}h=!!h;for(var $,w,x,_=0,k=[],S=[],T=o.length,E=[],M=[];_<T;){if(($=o.indexOf(t,_))===-1||(w=o.indexOf(e,$+c))===-1){_!==T&&E.push(a(o.substring(_)));break}_!==$&&E.push(a(o.substring(_,$))),x=o.substring($+c,w),k.push(x),S.push(n(x,d)),_=w+f,M.push(E.length),E.push("")}if(p&&E.length>1&&bo.throwNoconcat(o),!u||k.length){var D=function(t){for(var e=0,n=k.length;e<n;e++){if(h&&m(t[e]))return;E[M[e]]=t[e]}return E.join("")},A=function(t){return p?i.getTrusted(p,t):i.valueOf(t)};return l(function(t){var e=0,n=k.length,i=new Array(n);try{for(;e<n;e++)i[e]=S[e](t);return D(i)}catch(t){r(bo.interr(o,t))}},{exp:o,expressions:k,$$watchDelegate:function(t,e){var n;return t.$watchGroup(S,function(r,i){var o=D(r);C(e)&&e.call(this,o,r!==i?n:o,t),n=o})}})}}var c=t.length,f=e.length,p=new RegExp(t.replace(/./g,o),"g"),h=new RegExp(e.replace(/./g,o),"g");return u.startSymbol=function(){return t},u.endSymbol=function(){return e},u}]}function Be(){this.$get=["$rootScope","$window","$q","$$q","$browser",function(t,e,n,r,i){function o(o,s,u,l){function c(){f?o.apply(null,p):o(v)}var f=arguments.length>4,p=f?H(arguments,4):[],h=e.setInterval,d=e.clearInterval,v=0,g=y(l)&&!l,m=(g?r:n).defer(),b=m.promise;return u=y(u)?u:0,b.$$intervalId=h(function(){g?i.defer(c):t.$evalAsync(c),m.notify(v++),u>0&&v>=u&&(m.resolve(v),d(b.$$intervalId),delete a[b.$$intervalId]),g||t.$apply()},s),a[b.$$intervalId]=m,b}var a={};return o.cancel=function(t){return!!(t&&t.$$intervalId in a)&&(a[t.$$intervalId].promise.catch(h),a[t.$$intervalId].reject("canceled"),e.clearInterval(t.$$intervalId),delete a[t.$$intervalId],!0)},o}]}function ze(t){for(var e=t.split("/"),n=e.length;n--;)e[n]=et(e[n]);return e.join("/")}function We(t,e){var n=jn(t);e.$$protocol=n.protocol,e.$$host=n.hostname,e.$$port=f(n.port)||xo[n.protocol]||null}function Ye(t,e){if(Co.test(t))throw _o("badpath",'Invalid url "{0}".',t);var n="/"!==t.charAt(0);n&&(t="/"+t);var r=jn(t);e.$$path=decodeURIComponent(n&&"/"===r.pathname.charAt(0)?r.pathname.substring(1):r.pathname),e.$$search=Q(r.search),e.$$hash=decodeURIComponent(r.hash),e.$$path&&"/"!==e.$$path.charAt(0)&&(e.$$path="/"+e.$$path)}function Ge(t,e){return t.slice(0,e.length)===e}function Xe(t,e){if(Ge(e,t))return e.substr(t.length)}function Ke(t){var e=t.indexOf("#");return e===-1?t:t.substr(0,e)}function Je(t){return t.replace(/(#.+)|#$/,"$1")}function Ze(t){return t.substr(0,Ke(t).lastIndexOf("/")+1)}function Qe(t){return t.substring(0,t.indexOf("/",t.indexOf("//")+2))}function tn(t,e,n){this.$$html5=!0,n=n||"",We(t,this),this.$$parse=function(t){var n=Xe(e,t);if(!w(n))throw _o("ipthprfx",'Invalid url "{0}", missing path prefix "{1}".',t,e);Ye(n,this),this.$$path||(this.$$path="/"),this.$$compose()},this.$$compose=function(){var t=tt(this.$$search),n=this.$$hash?"#"+et(this.$$hash):"";this.$$url=ze(this.$$path)+(t?"?"+t:"")+n,this.$$absUrl=e+this.$$url.substr(1)},this.$$parseLinkUrl=function(r,i){if(i&&"#"===i[0])return this.hash(i.slice(1)),!0;var o,a,s;return y(o=Xe(t,r))?(a=o,s=n&&y(o=Xe(n,o))?e+(Xe("/",o)||o):t+a):y(o=Xe(e,r))?s=e+o:e===r+"/"&&(s=e),s&&this.$$parse(s),!!s}}function en(t,e,n){We(t,this),this.$$parse=function(r){function i(t,e,n){var r,i=/^\/[A-Z]:(\/.*)/;return Ge(e,n)&&(e=e.replace(n,"")),i.exec(e)?t:(r=i.exec(t),r?r[1]:t)}var o,a=Xe(t,r)||Xe(e,r);m(a)||"#"!==a.charAt(0)?this.$$html5?o=a:(o="",m(a)&&(t=r,this.replace())):(o=Xe(n,a),m(o)&&(o=a)),Ye(o,this),this.$$path=i(this.$$path,o,t),this.$$compose()},this.$$compose=function(){var e=tt(this.$$search),r=this.$$hash?"#"+et(this.$$hash):"";this.$$url=ze(this.$$path)+(e?"?"+e:"")+r,this.$$absUrl=t+(this.$$url?n+this.$$url:"")},this.$$parseLinkUrl=function(e,n){return Ke(t)===Ke(e)&&(this.$$parse(e),!0)}}function nn(t,e,n){this.$$html5=!0,en.apply(this,arguments),this.$$parseLinkUrl=function(r,i){if(i&&"#"===i[0])return this.hash(i.slice(1)),!0;var o,a;return t===Ke(r)?o=r:(a=Xe(e,r))?o=t+n+a:e===r+"/"&&(o=e),o&&this.$$parse(o),!!o},this.$$compose=function(){var e=tt(this.$$search),r=this.$$hash?"#"+et(this.$$hash):"";this.$$url=ze(this.$$path)+(e?"?"+e:"")+r,this.$$absUrl=t+n+this.$$url}}function rn(t){return function(){return this[t]}}function on(t,e){return function(n){return m(n)?this[t]:(this[t]=e(n),this.$$compose(),this)}}function an(){var t="!",e={enabled:!1,requireBase:!0,rewriteLinks:!0};this.hashPrefix=function(e){return y(e)?(t=e,this):t},this.html5Mode=function(t){return A(t)?(e.enabled=t,this):b(t)?(A(t.enabled)&&(e.enabled=t.enabled),A(t.requireBase)&&(e.requireBase=t.requireBase),(A(t.rewriteLinks)||w(t.rewriteLinks))&&(e.rewriteLinks=t.rewriteLinks),this):e},this.$get=["$rootScope","$browser","$sniffer","$rootElement","$window",function(n,r,i,o,a){function s(t,e,n){var i=l.url(),o=l.$$state;try{r.url(t,e,n),l.$$state=r.state()}catch(t){throw l.url(i),l.$$state=o,t}}function u(t,e){n.$broadcast("$locationChangeSuccess",l.absUrl(),t,l.$$state,e)}var l,c,f,p=r.baseHref(),h=r.url();if(e.enabled){if(!p&&e.requireBase)throw _o("nobase","$location in HTML5 mode requires a <base> tag to be present!");f=Qe(h)+(p||"/"),c=i.history?tn:nn}else f=Ke(h),c=en;var d=Ze(f);l=new c(f,d,"#"+t),l.$$parseLinkUrl(h,h),l.$$state=r.state();var v=/^\s*(javascript|mailto):/i;o.on("click",function(t){var i=e.rewriteLinks;if(i&&!t.ctrlKey&&!t.metaKey&&!t.shiftKey&&2!==t.which&&2!==t.button){for(var s=Jr(t.target);"a"!==j(s[0]);)if(s[0]===o[0]||!(s=s.parent())[0])return;if(!w(i)||!m(s.attr(i))){var u=s.prop("href"),c=s.attr("href")||s.attr("xlink:href");b(u)&&"[object SVGAnimatedString]"===u.toString()&&(u=jn(u.animVal).href),v.test(u)||!u||s.attr("target")||t.isDefaultPrevented()||l.$$parseLinkUrl(u,c)&&(t.preventDefault(),l.absUrl()!==r.url()&&(n.$apply(),a.angular["ff-684208-preventDefault"]=!0))}}}),Je(l.absUrl())!==Je(h)&&r.url(l.absUrl(),!0);var g=!0;return r.onUrlChange(function(t,e){return Ge(t,d)?(n.$evalAsync(function(){var r,i=l.absUrl(),o=l.$$state;t=Je(t),l.$$parse(t),l.$$state=e,r=n.$broadcast("$locationChangeStart",t,i,e,o).defaultPrevented,l.absUrl()===t&&(r?(l.$$parse(i),l.$$state=o,s(i,!1,o)):(g=!1,u(i,o)))}),void(n.$$phase||n.$digest())):void(a.location.href=t)}),n.$watch(function(){var t=Je(r.url()),e=Je(l.absUrl()),o=r.state(),a=l.$$replace,c=t!==e||l.$$html5&&i.history&&o!==l.$$state;(g||c)&&(g=!1,n.$evalAsync(function(){var e=l.absUrl(),r=n.$broadcast("$locationChangeStart",e,t,l.$$state,o).defaultPrevented;l.absUrl()===e&&(r?(l.$$parse(t),l.$$state=o):(c&&s(e,a,o===l.$$state?null:l.$$state),u(t,o)))})),l.$$replace=!1}),l}]}function sn(){var t=!0,e=this;this.debugEnabled=function(e){return y(e)?(t=e,this):t},this.$get=["$window",function(n){function i(t){return t instanceof Error&&(t.stack?t=t.message&&t.stack.indexOf(t.message)===-1?"Error: "+t.message+"\n"+t.stack:t.stack:t.sourceURL&&(t=t.message+"\n"+t.sourceURL+":"+t.line)),t}function o(t){var e=n.console||{},o=e[t]||e.log||h,a=!1;try{a=!!o.apply}catch(t){}return a?function(){var t=[];return r(arguments,function(e){t.push(i(e))}),o.apply(e,t)}:function(t,e){o(t,null==e?"":e)}}return{log:o("log"),info:o("info"),warn:o("warn"),error:o("error"),debug:function(){var n=o("debug");return function(){
t&&n.apply(e,arguments)}}()}}]}function un(t){return t+""}function ln(t,e){return"undefined"!=typeof t?t:e}function cn(t,e){return"undefined"==typeof t?e:"undefined"==typeof e?t:t+e}function fn(t,e){var n=t(e);return!n.$stateful}function pn(t,e){var n,i,o;switch(t.type){case Ao.Program:n=!0,r(t.body,function(t){pn(t.expression,e),n=n&&t.expression.constant}),t.constant=n;break;case Ao.Literal:t.constant=!0,t.toWatch=[];break;case Ao.UnaryExpression:pn(t.argument,e),t.constant=t.argument.constant,t.toWatch=t.argument.toWatch;break;case Ao.BinaryExpression:pn(t.left,e),pn(t.right,e),t.constant=t.left.constant&&t.right.constant,t.toWatch=t.left.toWatch.concat(t.right.toWatch);break;case Ao.LogicalExpression:pn(t.left,e),pn(t.right,e),t.constant=t.left.constant&&t.right.constant,t.toWatch=t.constant?[]:[t];break;case Ao.ConditionalExpression:pn(t.test,e),pn(t.alternate,e),pn(t.consequent,e),t.constant=t.test.constant&&t.alternate.constant&&t.consequent.constant,t.toWatch=t.constant?[]:[t];break;case Ao.Identifier:t.constant=!1,t.toWatch=[t];break;case Ao.MemberExpression:pn(t.object,e),t.computed&&pn(t.property,e),t.constant=t.object.constant&&(!t.computed||t.property.constant),t.toWatch=[t];break;case Ao.CallExpression:o=!!t.filter&&fn(e,t.callee.name),n=o,i=[],r(t.arguments,function(t){pn(t,e),n=n&&t.constant,t.constant||i.push.apply(i,t.toWatch)}),t.constant=n,t.toWatch=o?i:[t];break;case Ao.AssignmentExpression:pn(t.left,e),pn(t.right,e),t.constant=t.left.constant&&t.right.constant,t.toWatch=[t];break;case Ao.ArrayExpression:n=!0,i=[],r(t.elements,function(t){pn(t,e),n=n&&t.constant,t.constant||i.push.apply(i,t.toWatch)}),t.constant=n,t.toWatch=i;break;case Ao.ObjectExpression:n=!0,i=[],r(t.properties,function(t){pn(t.value,e),n=n&&t.value.constant&&!t.computed,t.value.constant||i.push.apply(i,t.value.toWatch)}),t.constant=n,t.toWatch=i;break;case Ao.ThisExpression:t.constant=!1,t.toWatch=[];break;case Ao.LocalsExpression:t.constant=!1,t.toWatch=[]}}function hn(t){if(1===t.length){var e=t[0].expression,n=e.toWatch;return 1!==n.length?n:n[0]!==e?n:void 0}}function dn(t){return t.type===Ao.Identifier||t.type===Ao.MemberExpression}function vn(t){if(1===t.body.length&&dn(t.body[0].expression))return{type:Ao.AssignmentExpression,left:t.body[0].expression,right:{type:Ao.NGValueParameter},operator:"="}}function gn(t){return 0===t.body.length||1===t.body.length&&(t.body[0].expression.type===Ao.Literal||t.body[0].expression.type===Ao.ArrayExpression||t.body[0].expression.type===Ao.ObjectExpression)}function mn(t){return t.constant}function yn(t,e){this.astBuilder=t,this.$filter=e}function bn(t,e){this.astBuilder=t,this.$filter=e}function $n(t){return C(t.valueOf)?t.valueOf():To.call(t)}function wn(){var t,e,n=gt(),i={true:!0,false:!1,null:null,undefined:void 0};this.addLiteral=function(t,e){i[t]=e},this.setIdentifierFns=function(n,r){return t=n,e=r,this},this.$get=["$filter",function(o){function a(t,e){var r,i,a;switch(typeof t){case"string":if(t=t.trim(),a=t,r=n[a],!r){":"===t.charAt(0)&&":"===t.charAt(1)&&(i=!0,t=t.substring(2));var s=new Do(v),d=new Oo(s,o,v);r=d.parse(t),r.constant?r.$$watchDelegate=f:i?r.$$watchDelegate=r.literal?c:l:r.inputs&&(r.$$watchDelegate=u),n[a]=r}return p(r,e);case"function":return p(t,e);default:return p(h,e)}}function s(t,e){return null==t||null==e?t===e:("object"!=typeof t||(t=$n(t),"object"!=typeof t))&&(t===e||t!==t&&e!==e)}function u(t,e,n,r,i){var o,a=r.inputs;if(1===a.length){var u=s;return a=a[0],t.$watch(function(t){var e=a(t);return s(e,u)||(o=r(t,void 0,void 0,[e]),u=e&&$n(e)),o},e,n,i)}for(var l=[],c=[],f=0,p=a.length;f<p;f++)l[f]=s,c[f]=null;return t.$watch(function(t){for(var e=!1,n=0,i=a.length;n<i;n++){var u=a[n](t);(e||(e=!s(u,l[n])))&&(c[n]=u,l[n]=u&&$n(u))}return e&&(o=r(t,void 0,void 0,c)),o},e,n,i)}function l(t,e,n,r,i){function o(t){return r(t)}function a(t,n,r){l=t,C(e)&&e(t,n,r),y(t)&&r.$$postDigest(function(){y(l)&&s()})}var s,l;return s=r.inputs?u(t,a,n,r,i):t.$watch(o,a,n)}function c(t,e,n,i){function o(t){var e=!0;return r(t,function(t){y(t)||(e=!1)}),e}var a,s;return a=t.$watch(function(t){return i(t)},function(t,n,r){s=t,C(e)&&e(t,n,r),o(t)&&r.$$postDigest(function(){o(s)&&a()})},n)}function f(t,e,n,r){var i=t.$watch(function(t){return i(),r(t)},e,n);return i}function p(t,e){if(!e)return t;var n=t.$$watchDelegate,r=!1,i=n!==c&&n!==l,o=i?function(n,i,o,a){var s=r&&a?a[0]:t(n,i,o,a);return e(s,n,i)}:function(n,r,i,o){var a=t(n,r,i,o),s=e(a,n,r);return y(a)?s:a};return r=!t.inputs,t.$$watchDelegate&&t.$$watchDelegate!==u?(o.$$watchDelegate=t.$$watchDelegate,o.inputs=t.inputs):e.$stateful||(o.$$watchDelegate=u,o.inputs=t.inputs?t.inputs:[t]),o}var d=hi().noUnsafeEval,v={csp:d,literals:V(i),isIdentifierStart:C(t)&&t,isIdentifierContinue:C(e)&&e};return a}]}function xn(){var t=!0;this.$get=["$rootScope","$exceptionHandler",function(e,n){return Cn(function(t){e.$evalAsync(t)},n,t)}],this.errorOnUnhandledRejections=function(e){return y(e)?(t=e,this):t}}function _n(){var t=!0;this.$get=["$browser","$exceptionHandler",function(e,n){return Cn(function(t){e.defer(t)},n,t)}],this.errorOnUnhandledRejections=function(e){return y(e)?(t=e,this):t}}function Cn(t,n,i){function o(){return new a}function a(){var t=this.promise=new s;this.resolve=function(e){p(t,e)},this.reject=function(e){d(t,e)},this.notify=function(e){g(t,e)}}function s(){this.$$state={status:0}}function u(e){var n,r,o;o=e.pending,e.processScheduled=!1,e.pending=void 0;try{for(var a=0,s=o.length;a<s;++a){e.pur=!0,r=o[a][0],n=o[a][e.status];try{C(n)?p(r,n(e.value)):1===e.status?p(r,e.value):d(r,e.value)}catch(t){d(r,t)}}}finally{--T,i&&0===T&&t(c)}}function c(){for(;!T&&E.length;){var t=E.shift();if(!t.pur){t.pur=!0;var e="Possibly unhandled rejection: "+wt(t.value);t.value instanceof Error?n(t.value,e):n(e)}}}function f(e){!i||e.pending||2!==e.status||e.pur||(0===T&&0===E.length&&t(c),E.push(e)),!e.processScheduled&&e.pending&&(e.processScheduled=!0,++T,t(function(){u(e)}))}function p(t,e){t.$$state.status||(e===t?v(t,S("qcycle","Expected promise to be resolved with value other than itself '{0}'",e)):h(t,e))}function h(t,e){function n(e){a||(a=!0,h(t,e))}function r(e){a||(a=!0,v(t,e))}function i(e){g(t,e)}var o,a=!1;try{(b(e)||C(e))&&(o=e.then),C(o)?(t.$$state.status=-1,o.call(e,n,r,i)):(t.$$state.value=e,t.$$state.status=1,f(t.$$state))}catch(t){r(t)}}function d(t,e){t.$$state.status||v(t,e)}function v(t,e){t.$$state.value=e,t.$$state.status=2,f(t.$$state)}function g(e,r){var i=e.$$state.pending;e.$$state.status<=0&&i&&i.length&&t(function(){for(var t,e,o=0,a=i.length;o<a;o++){e=i[o][0],t=i[o][3];try{g(e,C(t)?t(r):r)}catch(t){n(t)}}})}function y(t){var e=new s;return d(e,t),e}function $(t,e,n){var r=null;try{C(n)&&(r=n())}catch(t){return y(t)}return O(r)?r.then(function(){return e(t)},y):e(t)}function w(t,e,n,r){var i=new s;return p(i,t),i.then(e,n,r)}function x(t){var e=new s,n=0,i=li(t)?[]:{};return r(t,function(t,r){n++,w(t).then(function(t){i[r]=t,--n||p(e,i)},function(t){d(e,t)})}),0===n&&p(e,i),e}function _(t){var e=o();return r(t,function(t){w(t).then(e.resolve,e.reject)}),e.promise}function k(t){function e(t){p(r,t)}function n(t){d(r,t)}if(!C(t))throw S("norslvr","Expected resolverFn, got '{0}'",t);var r=new s;return t(e,n),r}var S=e("$q",TypeError),T=0,E=[];l(s.prototype,{then:function(t,e,n){if(m(t)&&m(e)&&m(n))return this;var r=new s;return this.$$state.pending=this.$$state.pending||[],this.$$state.pending.push([r,t,e,n]),this.$$state.status>0&&f(this.$$state),r},catch:function(t){return this.then(null,t)},finally:function(t,e){return this.then(function(e){return $(e,M,t)},function(e){return $(e,y,t)},e)}});var M=w;return k.prototype=s.prototype,k.defer=o,k.reject=y,k.when=w,k.resolve=M,k.all=x,k.race=_,k}function kn(){this.$get=["$window","$timeout",function(t,e){var n=t.requestAnimationFrame||t.webkitRequestAnimationFrame,r=t.cancelAnimationFrame||t.webkitCancelAnimationFrame||t.webkitCancelRequestAnimationFrame,i=!!n,o=i?function(t){var e=n(t);return function(){r(e)}}:function(t){var n=e(t,16.66,!1);return function(){e.cancel(n)}};return o.supported=i,o}]}function Sn(){function t(t){function e(){this.$$watchers=this.$$nextSibling=this.$$childHead=this.$$childTail=null,this.$$listeners={},this.$$listenerCount={},this.$$watchersCount=0,this.$id=a(),this.$$ChildScope=null}return e.prototype=t,e}var i=10,o=e("$rootScope"),s=null,u=null;this.digestTtl=function(t){return arguments.length&&(i=t),i},this.$get=["$exceptionHandler","$parse","$browser",function(e,l,c){function f(t){t.currentScope.$$destroyed=!0}function p(t){9===Kr&&(t.$$childHead&&p(t.$$childHead),t.$$nextSibling&&p(t.$$nextSibling)),t.$parent=t.$$nextSibling=t.$$prevSibling=t.$$childHead=t.$$childTail=t.$root=t.$$watchers=null}function d(){this.$id=a(),this.$$phase=this.$parent=this.$$watchers=this.$$nextSibling=this.$$prevSibling=this.$$childHead=this.$$childTail=null,this.$root=this,this.$$destroyed=!1,this.$$listeners={},this.$$listenerCount={},this.$$watchersCount=0,this.$$isolateBindings=null}function v(t){if(k.$$phase)throw o("inprog","{0} already in progress",k.$$phase);k.$$phase=t}function g(){k.$$phase=null}function y(t,e){do t.$$watchersCount+=e;while(t=t.$parent)}function $(t,e,n){do t.$$listenerCount[n]-=e,0===t.$$listenerCount[n]&&delete t.$$listenerCount[n];while(t=t.$parent)}function w(){}function x(){for(;E.length;)try{E.shift()()}catch(t){e(t)}u=null}function _(){null===u&&(u=c.defer(function(){k.$apply(x)}))}d.prototype={constructor:d,$new:function(e,n){var r;return n=n||this,e?(r=new d,r.$root=this.$root):(this.$$ChildScope||(this.$$ChildScope=t(this)),r=new this.$$ChildScope),r.$parent=n,r.$$prevSibling=n.$$childTail,n.$$childHead?(n.$$childTail.$$nextSibling=r,n.$$childTail=r):n.$$childHead=n.$$childTail=r,(e||n!==this)&&r.$on("$destroy",f),r},$watch:function(t,e,n,r){var i=l(t);if(i.$$watchDelegate)return i.$$watchDelegate(this,e,n,i,t);var o=this,a=o.$$watchers,u={fn:e,last:w,get:i,exp:r||t,eq:!!n};return s=null,C(e)||(u.fn=h),a||(a=o.$$watchers=[],a.$$digestWatchIndex=-1),a.unshift(u),a.$$digestWatchIndex++,y(this,1),function(){var t=q(a,u);t>=0&&(y(o,-1),t<a.$$digestWatchIndex&&a.$$digestWatchIndex--),s=null}},$watchGroup:function(t,e){function n(){u=!1,l?(l=!1,e(o,o,s)):e(o,i,s)}var i=new Array(t.length),o=new Array(t.length),a=[],s=this,u=!1,l=!0;if(!t.length){var c=!0;return s.$evalAsync(function(){c&&e(o,o,s)}),function(){c=!1}}return 1===t.length?this.$watch(t[0],function(t,n,r){o[0]=t,i[0]=n,e(o,t===n?o:i,r)}):(r(t,function(t,e){var r=s.$watch(t,function(t,r){o[e]=t,i[e]=r,u||(u=!0,s.$evalAsync(n))});a.push(r)}),function(){for(;a.length;)a.shift()()})},$watchCollection:function(t,e){function r(t){o=t;var e,r,i,s,u;if(!m(o)){if(b(o))if(n(o)){a!==h&&(a=h,g=a.length=0,f++),e=o.length,g!==e&&(f++,a.length=g=e);for(var l=0;l<e;l++)u=a[l],s=o[l],i=u!==u&&s!==s,i||u===s||(f++,a[l]=s)}else{a!==d&&(a=d={},g=0,f++),e=0;for(r in o)zr.call(o,r)&&(e++,s=o[r],u=a[r],r in a?(i=u!==u&&s!==s,i||u===s||(f++,a[r]=s)):(g++,a[r]=s,f++));if(g>e){f++;for(r in a)zr.call(o,r)||(g--,delete a[r])}}else a!==o&&(a=o,f++);return f}}function i(){if(v?(v=!1,e(o,o,u)):e(o,s,u),c)if(b(o))if(n(o)){s=new Array(o.length);for(var t=0;t<o.length;t++)s[t]=o[t]}else{s={};for(var r in o)zr.call(o,r)&&(s[r]=o[r])}else s=o}r.$stateful=!0;var o,a,s,u=this,c=e.length>1,f=0,p=l(t,r),h=[],d={},v=!0,g=0;return this.$watch(p,i)},$digest:function(){var t,n,r,a,l,f,p,h,d,m,y,b=i,$=this,_=[];v("$digest"),c.$$checkUrlChange(),this===k&&null!==u&&(c.defer.cancel(u),x()),s=null;do{p=!1,d=$;for(var E=0;E<S.length;E++){try{y=S[E],y.scope.$eval(y.expression,y.locals)}catch(t){e(t)}s=null}S.length=0;t:do{if(f=d.$$watchers)for(f.$$digestWatchIndex=f.length;f.$$digestWatchIndex--;)try{if(t=f[f.$$digestWatchIndex])if(l=t.get,(n=l(d))===(r=t.last)||(t.eq?U(n,r):ui(n)&&ui(r))){if(t===s){p=!1;break t}}else p=!0,s=t,t.last=t.eq?V(n,null):n,a=t.fn,a(n,r===w?n:r,d),b<5&&(m=4-b,_[m]||(_[m]=[]),_[m].push({msg:C(t.exp)?"fn: "+(t.exp.name||t.exp.toString()):t.exp,newVal:n,oldVal:r}))}catch(t){e(t)}if(!(h=d.$$watchersCount&&d.$$childHead||d!==$&&d.$$nextSibling))for(;d!==$&&!(h=d.$$nextSibling);)d=d.$parent}while(d=h);if((p||S.length)&&!b--)throw g(),o("infdig","{0} $digest() iterations reached. Aborting!\nWatchers fired in the last 5 iterations: {1}",i,_)}while(p||S.length);for(g();M<T.length;)try{T[M++]()}catch(t){e(t)}T.length=M=0},$destroy:function(){if(!this.$$destroyed){var t=this.$parent;this.$broadcast("$destroy"),this.$$destroyed=!0,this===k&&c.$$applicationDestroyed(),y(this,-this.$$watchersCount);for(var e in this.$$listenerCount)$(this,this.$$listenerCount[e],e);t&&t.$$childHead===this&&(t.$$childHead=this.$$nextSibling),t&&t.$$childTail===this&&(t.$$childTail=this.$$prevSibling),this.$$prevSibling&&(this.$$prevSibling.$$nextSibling=this.$$nextSibling),this.$$nextSibling&&(this.$$nextSibling.$$prevSibling=this.$$prevSibling),this.$destroy=this.$digest=this.$apply=this.$evalAsync=this.$applyAsync=h,this.$on=this.$watch=this.$watchGroup=function(){return h},this.$$listeners={},this.$$nextSibling=null,p(this)}},$eval:function(t,e){return l(t)(this,e)},$evalAsync:function(t,e){k.$$phase||S.length||c.defer(function(){S.length&&k.$digest()}),S.push({scope:this,expression:l(t),locals:e})},$$postDigest:function(t){T.push(t)},$apply:function(t){try{v("$apply");try{return this.$eval(t)}finally{g()}}catch(t){e(t)}finally{try{k.$digest()}catch(t){throw e(t),t}}},$applyAsync:function(t){function e(){n.$eval(t)}var n=this;t&&E.push(e),t=l(t),_()},$on:function(t,e){var n=this.$$listeners[t];n||(this.$$listeners[t]=n=[]),n.push(e);var r=this;do r.$$listenerCount[t]||(r.$$listenerCount[t]=0),r.$$listenerCount[t]++;while(r=r.$parent);var i=this;return function(){var r=n.indexOf(e);r!==-1&&(n[r]=null,$(i,1,t))}},$emit:function(t,n){var r,i,o,a=[],s=this,u=!1,l={name:t,targetScope:s,stopPropagation:function(){u=!0},preventDefault:function(){l.defaultPrevented=!0},defaultPrevented:!1},c=L([l],arguments,1);do{for(r=s.$$listeners[t]||a,l.currentScope=s,i=0,o=r.length;i<o;i++)if(r[i])try{r[i].apply(null,c)}catch(t){e(t)}else r.splice(i,1),i--,o--;if(u)return l.currentScope=null,l;s=s.$parent}while(s);return l.currentScope=null,l},$broadcast:function(t,n){var r=this,i=r,o=r,a={name:t,targetScope:r,preventDefault:function(){a.defaultPrevented=!0},defaultPrevented:!1};if(!r.$$listenerCount[t])return a;for(var s,u,l,c=L([a],arguments,1);i=o;){for(a.currentScope=i,s=i.$$listeners[t]||[],u=0,l=s.length;u<l;u++)if(s[u])try{s[u].apply(null,c)}catch(t){e(t)}else s.splice(u,1),u--,l--;if(!(o=i.$$listenerCount[t]&&i.$$childHead||i!==r&&i.$$nextSibling))for(;i!==r&&!(o=i.$$nextSibling);)i=i.$parent}return a.currentScope=null,a}};var k=new d,S=k.$$asyncQueue=[],T=k.$$postDigestQueue=[],E=k.$$applyAsyncQueue=[],M=0;return k}]}function Tn(){var t=/^\s*(https?|ftp|mailto|tel|file):/,e=/^\s*((https?|ftp|file|blob):|data:image\/)/;this.aHrefSanitizationWhitelist=function(e){return y(e)?(t=e,this):t},this.imgSrcSanitizationWhitelist=function(t){return y(t)?(e=t,this):e},this.$get=function(){return function(n,r){var i,o=r?e:t;return i=jn(n).href,""===i||i.match(o)?n:"unsafe:"+i}}}function En(t){return t.replace(Ro,kt)}function Mn(t){if("self"===t)return t;if(w(t)){if(t.indexOf("***")>-1)throw No("iwcard","Illegal sequence *** in string matcher.  String: {0}",t);return t=pi(t).replace(/\\\*\\\*/g,".*").replace(/\\\*/g,"[^:/.?&;]*"),new RegExp("^"+t+"$")}if(k(t))return new RegExp("^"+t.source+"$");throw No("imatcher",'Matchers may only be "self", string patterns or RegExp objects')}function Dn(t){var e=[];return y(t)&&r(t,function(t){e.push(Mn(t))}),e}function An(){this.SCE_CONTEXTS=Po;var t=["self"],e=[];this.resourceUrlWhitelist=function(e){return arguments.length&&(t=Dn(e)),t},this.resourceUrlBlacklist=function(t){return arguments.length&&(e=Dn(t)),e},this.$get=["$injector",function(n){function r(t,e){return"self"===t?Fn(e):!!t.exec(e.href)}function i(n){var i,o,a=jn(n.toString()),s=!1;for(i=0,o=t.length;i<o;i++)if(r(t[i],a)){s=!0;break}if(s)for(i=0,o=e.length;i<o;i++)if(r(e[i],a)){s=!1;break}return s}function o(t){var e=function(t){this.$$unwrapTrustedValue=function(){return t}};return t&&(e.prototype=new t),e.prototype.valueOf=function(){return this.$$unwrapTrustedValue()},e.prototype.toString=function(){return this.$$unwrapTrustedValue().toString()},e}function a(t,e){var n=f.hasOwnProperty(t)?f[t]:null;if(!n)throw No("icontext","Attempted to trust a value in invalid context. Context: {0}; Value: {1}",t,e);if(null===e||m(e)||""===e)return e;if("string"!=typeof e)throw No("itype","Attempted to trust a non-string value in a content requiring a string: Context: {0}",t);return new n(e)}function s(t){return t instanceof c?t.$$unwrapTrustedValue():t}function u(t,e){if(null===e||m(e)||""===e)return e;var n=f.hasOwnProperty(t)?f[t]:null;if(n&&e instanceof n)return e.$$unwrapTrustedValue();if(t===Po.RESOURCE_URL){if(i(e))return e;throw No("insecurl","Blocked loading resource from url not allowed by $sceDelegate policy.  URL: {0}",e.toString())}if(t===Po.HTML)return l(e);throw No("unsafe","Attempting to use an unsafe value in a safe context.")}var l=function(t){throw No("unsafe","Attempting to use an unsafe value in a safe context.")};n.has("$sanitize")&&(l=n.get("$sanitize"));var c=o(),f={};return f[Po.HTML]=o(c),f[Po.CSS]=o(c),f[Po.URL]=o(c),f[Po.JS]=o(c),f[Po.RESOURCE_URL]=o(f[Po.URL]),{trustAs:a,getTrusted:u,valueOf:s}}]}function On(){var t=!0;this.enabled=function(e){return arguments.length&&(t=!!e),t},this.$get=["$parse","$sceDelegate",function(e,n){if(t&&Kr<8)throw No("iequirks","Strict Contextual Escaping does not support Internet Explorer version < 11 in quirks mode.  You can fix this by adding the text <!doctype html> to the top of your HTML document.  See http://docs.angularjs.org/api/ng.$sce for more information.");var i=bt(Po);i.isEnabled=function(){return t},i.trustAs=n.trustAs,i.getTrusted=n.getTrusted,i.valueOf=n.valueOf,t||(i.trustAs=i.getTrusted=function(t,e){return e},i.valueOf=d),i.parseAs=function(t,n){var r=e(n);return r.literal&&r.constant?r:e(n,function(e){return i.getTrusted(t,e)})};var o=i.parseAs,a=i.getTrusted,s=i.trustAs;return r(Po,function(t,e){var n=Wr(e);i[En("parse_as_"+n)]=function(e){return o(t,e)},i[En("get_trusted_"+n)]=function(e){return a(t,e)},i[En("trust_as_"+n)]=function(e){return s(t,e)}}),i}]}function Nn(){this.$get=["$window","$document",function(t,e){var n={},r=t.chrome&&(t.chrome.app&&t.chrome.app.runtime||!t.chrome.app&&t.chrome.runtime&&t.chrome.runtime.id),i=!r&&t.history&&t.history.pushState,o=f((/android (\d+)/.exec(Wr((t.navigator||{}).userAgent))||[])[1]),a=/Boxee/i.test((t.navigator||{}).userAgent),s=e[0]||{},u=s.body&&s.body.style,l=!1,c=!1;return u&&(l=!!("transition"in u||"webkitTransition"in u),c=!!("animation"in u||"webkitAnimation"in u)),{history:!(!i||o<4||a),hasEvent:function(t){if("input"===t&&Kr)return!1;if(m(n[t])){var e=s.createElement("div");n[t]="on"+t in e}return n[t]},csp:hi(),transitions:l,animations:c,android:o}}]}function Pn(){var t;this.httpOptions=function(e){return e?(t=e,this):t},this.$get=["$exceptionHandler","$templateCache","$http","$q","$sce",function(e,n,r,i,o){function a(s,u){function c(t){return u||(t=Io("tpload","Failed to load template: {0} (HTTP status: {1} {2})",s,t.status,t.statusText),e(t)),i.reject(t)}a.totalPendingRequests++,w(s)&&!m(n.get(s))||(s=o.getTrustedResourceUrl(s));var f=r.defaults&&r.defaults.transformResponse;return li(f)?f=f.filter(function(t){return t!==Ne}):f===Ne&&(f=null),r.get(s,l({cache:n,transformResponse:f},t)).finally(function(){a.totalPendingRequests--}).then(function(t){return n.put(s,t.data),t.data},c)}return a.totalPendingRequests=0,a}]}function Rn(){this.$get=["$rootScope","$browser","$location",function(t,e,n){var i={};return i.findBindings=function(t,e,n){var i=t.getElementsByClassName("ng-binding"),o=[];return r(i,function(t){var i=ai.element(t).data("$binding");i&&r(i,function(r){if(n){var i=new RegExp("(^|\\s)"+pi(e)+"(\\s|\\||$)");i.test(r)&&o.push(t)}else r.indexOf(e)!==-1&&o.push(t)})}),o},i.findModels=function(t,e,n){for(var r=["ng-","data-ng-","ng\\:"],i=0;i<r.length;++i){var o=n?"=":"*=",a="["+r[i]+"model"+o+'"'+e+'"]',s=t.querySelectorAll(a);if(s.length)return s}},i.getLocation=function(){return n.url()},i.setLocation=function(e){e!==n.url()&&(n.url(e),t.$digest())},i.whenStable=function(t){e.notifyWhenNoOutstandingRequests(t)},i}]}function In(){this.$get=["$rootScope","$browser","$q","$$q","$exceptionHandler",function(t,e,n,r,i){function o(o,s,u){C(o)||(u=s,s=o,o=h);var l,c=H(arguments,3),f=y(u)&&!u,p=(f?r:n).defer(),d=p.promise;return l=e.defer(function(){try{p.resolve(o.apply(null,c))}catch(t){p.reject(t),i(t)}finally{delete a[d.$$timeoutId]}f||t.$apply()},s),d.$$timeoutId=l,a[l]=p,d}var a={};return o.cancel=function(t){return!!(t&&t.$$timeoutId in a)&&(a[t.$$timeoutId].promise.catch(h),a[t.$$timeoutId].reject("canceled"),delete a[t.$$timeoutId],e.defer.cancel(t.$$timeoutId))},o}]}function jn(t){var e=t;return Kr&&(jo.setAttribute("href",e),e=jo.href),jo.setAttribute("href",e),{href:jo.href,protocol:jo.protocol?jo.protocol.replace(/:$/,""):"",host:jo.host,search:jo.search?jo.search.replace(/^\?/,""):"",hash:jo.hash?jo.hash.replace(/^#/,""):"",hostname:jo.hostname,port:jo.port,pathname:"/"===jo.pathname.charAt(0)?jo.pathname:"/"+jo.pathname}}function Fn(t){var e=w(t)?jn(t):t;return e.protocol===Fo.protocol&&e.host===Fo.host}function qn(){this.$get=v(t)}function Vn(t){function e(t){try{return t.cookie||""}catch(t){return""}}function n(t){try{return decodeURIComponent(t)}catch(e){return t}}var r=t[0]||{},i={},o="";return function(){var t,a,s,u,l,c=e(r);if(c!==o)for(o=c,t=o.split("; "),i={},s=0;s<t.length;s++)a=t[s],u=a.indexOf("="),u>0&&(l=n(a.substring(0,u)),m(i[l])&&(i[l]=n(a.substring(u+1))));return i}}function Un(){this.$get=Vn}function Ln(t){function e(i,o){if(b(i)){var a={};return r(i,function(t,n){a[n]=e(n,t)}),a}return t.factory(i+n,o)}var n="Filter";this.register=e,this.$get=["$injector",function(t){return function(e){return t.get(e+n)}}],e("currency",Yn),e("date",ur),e("filter",Hn),e("json",lr),e("limitTo",cr),e("lowercase",zo),e("number",Gn),e("orderBy",pr),e("uppercase",Wo)}function Hn(){return function(t,r,i,o){if(!n(t)){if(null==t)return t;throw e("filter")("notarray","Expected array but received: {0}",t)}o=o||"$";var a,s,u=Wn(r);switch(u){case"function":a=r;break;case"boolean":case"null":case"number":case"string":s=!0;case"object":a=Bn(r,i,o,s);break;default:return t}return Array.prototype.filter.call(t,a)}}function Bn(t,e,n,r){var i,o=b(t)&&n in t;return e===!0?e=U:C(e)||(e=function(t,e){return!m(t)&&(null===t||null===e?t===e:!(b(e)||b(t)&&!g(t))&&(t=Wr(""+t),e=Wr(""+e),t.indexOf(e)!==-1))}),i=function(i){return o&&!b(i)?zn(i,t[n],e,n,!1):zn(i,t,e,n,r)}}function zn(t,e,n,r,i,o){var a=Wn(t),s=Wn(e);if("string"===s&&"!"===e.charAt(0))return!zn(t,e.substring(1),n,r,i);if(li(t))return t.some(function(t){return zn(t,e,n,r,i)});switch(a){case"object":var u;if(i){for(u in t)if("$"!==u.charAt(0)&&zn(t[u],e,n,r,!0))return!0;return!o&&zn(t,e,n,r,!1)}if("object"===s){for(u in e){var l=e[u];if(!C(l)&&!m(l)){var c=u===r,f=c?t:t[u];if(!zn(f,l,n,r,c,c))return!1}}return!0}return n(t,e);case"function":return!1;default:return n(t,e)}}function Wn(t){return null===t?"null":typeof t}function Yn(t){var e=t.NUMBER_FORMATS;return function(t,n,r){return m(n)&&(n=e.CURRENCY_SYM),m(r)&&(r=e.PATTERNS[1].maxFrac),null==t?t:Jn(t,e.PATTERNS[1],e.GROUP_SEP,e.DECIMAL_SEP,r).replace(/\u00A4/g,n)}}function Gn(t){var e=t.NUMBER_FORMATS;return function(t,n){return null==t?t:Jn(t,e.PATTERNS[0],e.GROUP_SEP,e.DECIMAL_SEP,n)}}function Xn(t){var e,n,r,i,o,a=0;for((n=t.indexOf(Vo))>-1&&(t=t.replace(Vo,"")),(r=t.search(/e/i))>0?(n<0&&(n=r),n+=+t.slice(r+1),t=t.substring(0,r)):n<0&&(n=t.length),r=0;t.charAt(r)===Uo;r++);if(r===(o=t.length))e=[0],n=1;else{for(o--;t.charAt(o)===Uo;)o--;for(n-=r,e=[],i=0;r<=o;r++,i++)e[i]=+t.charAt(r)}return n>qo&&(e=e.splice(0,qo-1),a=n-1,n=1),{d:e,e:a,i:n}}function Kn(t,e,n,r){var i=t.d,o=i.length-t.i;e=m(e)?Math.min(Math.max(n,o),r):+e;var a=e+t.i,s=i[a];if(a>0){i.splice(Math.max(t.i,a));for(var u=a;u<i.length;u++)i[u]=0}else{o=Math.max(0,o),t.i=1,i.length=Math.max(1,a=e+1),i[0]=0;for(var l=1;l<a;l++)i[l]=0}if(s>=5)if(a-1<0){for(var c=0;c>a;c--)i.unshift(0),t.i++;i.unshift(1),t.i++}else i[a-1]++;for(;o<Math.max(0,e);o++)i.push(0);var f=i.reduceRight(function(t,e,n,r){return e+=t,r[n]=e%10,Math.floor(e/10)},0);f&&(i.unshift(f),t.i++)}function Jn(t,e,n,r,i){if(!w(t)&&!x(t)||isNaN(t))return"";var o,a=!isFinite(t),s=!1,u=Math.abs(t)+"",l="";if(a)l="∞";else{o=Xn(u),Kn(o,i,e.minFrac,e.maxFrac);var c=o.d,f=o.i,p=o.e,h=[];for(s=c.reduce(function(t,e){return t&&!e},!0);f<0;)c.unshift(0),f++;f>0?h=c.splice(f,c.length):(h=c,c=[0]);var d=[];for(c.length>=e.lgSize&&d.unshift(c.splice(-e.lgSize,c.length).join(""));c.length>e.gSize;)d.unshift(c.splice(-e.gSize,c.length).join(""));c.length&&d.unshift(c.join("")),l=d.join(n),h.length&&(l+=r+h.join("")),p&&(l+="e+"+p)}return t<0&&!s?e.negPre+l+e.negSuf:e.posPre+l+e.posSuf}function Zn(t,e,n,r){var i="";for((t<0||r&&t<=0)&&(r?t=-t+1:(t=-t,i="-")),t=""+t;t.length<e;)t=Uo+t;return n&&(t=t.substr(t.length-e)),i+t}function Qn(t,e,n,r,i){return n=n||0,function(o){var a=o["get"+t]();return(n>0||a>-n)&&(a+=n),0===a&&n===-12&&(a=12),Zn(a,e,r,i)}}function tr(t,e,n){return function(r,i){var o=r["get"+t](),a=(n?"STANDALONE":"")+(e?"SHORT":""),s=Yr(a+t);return i[s][o]}}function er(t,e,n){var r=-1*n,i=r>=0?"+":"";return i+=Zn(Math[r>0?"floor":"ceil"](r/60),2)+Zn(Math.abs(r%60),2)}function nr(t){var e=new Date(t,0,1).getDay();return new Date(t,0,(e<=4?5:12)-e)}function rr(t){return new Date(t.getFullYear(),t.getMonth(),t.getDate()+(4-t.getDay()))}function ir(t){return function(e){var n=nr(e.getFullYear()),r=rr(e),i=+r-+n,o=1+Math.round(i/6048e5);return Zn(o,t)}}function or(t,e){return t.getHours()<12?e.AMPMS[0]:e.AMPMS[1]}function ar(t,e){return t.getFullYear()<=0?e.ERAS[0]:e.ERAS[1]}function sr(t,e){return t.getFullYear()<=0?e.ERANAMES[0]:e.ERANAMES[1]}function ur(t){function e(t){var e;if(e=t.match(n)){var r=new Date(0),i=0,o=0,a=e[8]?r.setUTCFullYear:r.setFullYear,s=e[8]?r.setUTCHours:r.setHours;e[9]&&(i=f(e[9]+e[10]),o=f(e[9]+e[11])),a.call(r,f(e[1]),f(e[2])-1,f(e[3]));var u=f(e[4]||0)-i,l=f(e[5]||0)-o,c=f(e[6]||0),p=Math.round(1e3*parseFloat("0."+(e[7]||0)));return s.call(r,u,l,c,p),r}return t}var n=/^(\d{4})-?(\d\d)-?(\d\d)(?:T(\d\d)(?::?(\d\d)(?::?(\d\d)(?:\.(\d+))?)?)?(Z|([+-])(\d\d):?(\d\d))?)?$/;return function(n,i,o){var a,s,u="",l=[];if(i=i||"mediumDate",i=t.DATETIME_FORMATS[i]||i,w(n)&&(n=Bo.test(n)?f(n):e(n)),x(n)&&(n=new Date(n)),!_(n)||!isFinite(n.getTime()))return n;for(;i;)s=Ho.exec(i),s?(l=L(l,s,1),i=l.pop()):(l.push(i),i=null);var c=n.getTimezoneOffset();return o&&(c=G(o,c),n=K(n,o,!0)),r(l,function(e){a=Lo[e],u+=a?a(n,t.DATETIME_FORMATS,c):"''"===e?"'":e.replace(/(^'|'$)/g,"").replace(/''/g,"'")}),u}}function lr(){return function(t,e){return m(e)&&(e=2),W(t,e)}}function cr(){return function(t,e,r){return e=Math.abs(Number(e))===1/0?Number(e):f(e),ui(e)?t:(x(t)&&(t=t.toString()),n(t)?(r=!r||isNaN(r)?0:f(r),r=r<0?Math.max(0,t.length+r):r,e>=0?fr(t,r,r+e):0===r?fr(t,e,t.length):fr(t,Math.max(0,r+e),r)):t)}}function fr(t,e,n){return w(t)?t.slice(e,n):ti.call(t,e,n)}function pr(t){function r(e){return e.map(function(e){var n=1,r=d;if(C(e))r=e;else if(w(e)&&("+"!==e.charAt(0)&&"-"!==e.charAt(0)||(n="-"===e.charAt(0)?-1:1,e=e.substring(1)),""!==e&&(r=t(e),r.constant))){var i=r();r=function(t){return t[i]}}return{get:r,descending:n}})}function i(t){switch(typeof t){case"number":case"boolean":case"string":return!0;default:return!1}}function o(t){return C(t.valueOf)&&(t=t.valueOf(),i(t))?t:g(t)&&(t=t.toString(),i(t))?t:t}function a(t,e){var n=typeof t;return null===t?(n="string",t="null"):"object"===n&&(t=o(t)),{value:t,type:n,index:e}}function s(t,e){var n=0,r=t.type,i=e.type;if(r===i){var o=t.value,a=e.value;"string"===r?(o=o.toLowerCase(),a=a.toLowerCase()):"object"===r&&(b(o)&&(o=t.index),b(a)&&(a=e.index)),o!==a&&(n=o<a?-1:1)}else n=r<i?-1:1;return n}return function(t,i,o,u){function l(t,e){return{value:t,tieBreaker:{value:e,type:"number",index:e},predicateValues:f.map(function(n){return a(n.get(t),e)})}}function c(t,e){for(var n=0,r=f.length;n<r;n++){var i=h(t.predicateValues[n],e.predicateValues[n]);if(i)return i*f[n].descending*p}return h(t.tieBreaker,e.tieBreaker)*p}if(null==t)return t;if(!n(t))throw e("orderBy")("notarray","Expected array but received: {0}",t);li(i)||(i=[i]),0===i.length&&(i=["+"]);var f=r(i),p=o?-1:1,h=C(u)?u:s,d=Array.prototype.map.call(t,l);return d.sort(c),t=d.map(function(t){return t.value})}}function hr(t){return C(t)&&(t={link:t}),t.restrict=t.restrict||"AC",v(t)}function dr(t,e){t.$name=e}function vr(t,e,n,r,i){this.$$controls=[],this.$error={},this.$$success={},this.$pending=void 0,this.$name=i(e.name||e.ngForm||"")(n),this.$dirty=!1,this.$pristine=!0,this.$valid=!0,this.$invalid=!1,this.$submitted=!1,this.$$parentForm=Xo,this.$$element=t,this.$$animate=r,gr(this)}function gr(t){t.$$classCache={},t.$$classCache[Na]=!(t.$$classCache[Oa]=t.$$element.hasClass(Oa))}function mr(t){function e(t,e,n,r){t[e]||(t[e]={}),a(t[e],n,r)}function n(t,e,n,r){t[e]&&s(t[e],n,r),yr(t[e])&&(t[e]=void 0)}function r(t,e,n){n&&!t.$$classCache[e]?(t.$$animate.addClass(t.$$element,e),t.$$classCache[e]=!0):!n&&t.$$classCache[e]&&(t.$$animate.removeClass(t.$$element,e),t.$$classCache[e]=!1)}function i(t,e,n){e=e?"-"+lt(e,"-"):"",r(t,Oa+e,n===!0),r(t,Na+e,n===!1)}var o=t.clazz,a=t.set,s=t.unset;o.prototype.$setValidity=function(t,o,u){m(o)?e(this,"$pending",t,u):n(this,"$pending",t,u),A(o)?o?(s(this.$error,t,u),a(this.$$success,t,u)):(a(this.$error,t,u),s(this.$$success,t,u)):(s(this.$error,t,u),s(this.$$success,t,u)),this.$pending?(r(this,Ko,!0),this.$valid=this.$invalid=void 0,i(this,"",null)):(r(this,Ko,!1),this.$valid=yr(this.$error),this.$invalid=!this.$valid,i(this,"",this.$valid));var l;l=this.$pending&&this.$pending[t]?void 0:!this.$error[t]&&(!!this.$$success[t]||null),i(this,t,l),this.$$parentForm.$setValidity(t,l,this)}}function yr(t){if(t)for(var e in t)if(t.hasOwnProperty(e))return!1;return!0}function br(t){t.$formatters.push(function(e){return t.$isEmpty(e)?e:e.toString()})}function $r(t,e,n,r,i,o){wr(t,e,n,r,i,o),br(r)}function wr(t,e,n,r,i,o){var a=Wr(e[0].type);if(!i.android){var s=!1;e.on("compositionstart",function(){s=!0}),e.on("compositionend",function(){s=!1,l()})}var u,l=function(t){if(u&&(o.defer.cancel(u),u=null),!s){var i=e.val(),l=t&&t.type;"password"===a||n.ngTrim&&"false"===n.ngTrim||(i=fi(i)),(r.$viewValue!==i||""===i&&r.$$hasNativeValidators)&&r.$setViewValue(i,l)}};if(i.hasEvent("input"))e.on("input",l);else{var c=function(t,e,n){u||(u=o.defer(function(){u=null,e&&e.value===n||l(t)}))};e.on("keydown",function(t){var e=t.keyCode;91===e||15<e&&e<19||37<=e&&e<=40||c(t,this,this.value)}),i.hasEvent("paste")&&e.on("paste cut",c)}e.on("change",l),fa[a]&&r.$$hasNativeValidators&&a===n.type&&e.on(ca,function(t){if(!u){var e=this[Br],n=e.badInput,r=e.typeMismatch;u=o.defer(function(){u=null,e.badInput===n&&e.typeMismatch===r||l(t)})}}),r.$render=function(){var t=r.$isEmpty(r.$viewValue)?"":r.$viewValue;e.val()!==t&&e.val(t)}}function xr(t,e){if(_(t))return t;if(w(t)){sa.lastIndex=0;var n=sa.exec(t);if(n){var r=+n[1],i=+n[2],o=0,a=0,s=0,u=0,l=nr(r),c=7*(i-1);return e&&(o=e.getHours(),a=e.getMinutes(),s=e.getSeconds(),u=e.getMilliseconds()),new Date(r,0,l.getDate()+c,o,a,s,u)}}return NaN}function _r(t,e){return function(n,i){var o,a;if(_(n))return n;if(w(n)){if('"'===n.charAt(0)&&'"'===n.charAt(n.length-1)&&(n=n.substring(1,n.length-1)),ea.test(n))return new Date(n);if(t.lastIndex=0,o=t.exec(n))return o.shift(),a=i?{yyyy:i.getFullYear(),MM:i.getMonth()+1,dd:i.getDate(),HH:i.getHours(),mm:i.getMinutes(),ss:i.getSeconds(),sss:i.getMilliseconds()/1e3}:{yyyy:1970,MM:1,dd:1,HH:0,mm:0,ss:0,sss:0},r(o,function(t,n){
n<e.length&&(a[e[n]]=+t)}),new Date(a.yyyy,a.MM-1,a.dd,a.HH,a.mm,a.ss||0,1e3*a.sss||0)}return NaN}}function Cr(t,e,n,r){return function(i,o,a,s,u,l,c){function f(t){return t&&!(t.getTime&&t.getTime()!==t.getTime())}function p(t){return y(t)&&!_(t)?n(t)||void 0:t}kr(i,o,a,s),wr(i,o,a,s,u,l);var h,d=s&&s.$options.getOption("timezone");if(s.$$parserName=t,s.$parsers.push(function(t){if(s.$isEmpty(t))return null;if(e.test(t)){var r=n(t,h);return d&&(r=K(r,d)),r}}),s.$formatters.push(function(t){if(t&&!_(t))throw Va("datefmt","Expected `{0}` to be a date",t);return f(t)?(h=t,h&&d&&(h=K(h,d,!0)),c("date")(t,r,d)):(h=null,"")}),y(a.min)||a.ngMin){var v;s.$validators.min=function(t){return!f(t)||m(v)||n(t)>=v},a.$observe("min",function(t){v=p(t),s.$validate()})}if(y(a.max)||a.ngMax){var g;s.$validators.max=function(t){return!f(t)||m(g)||n(t)<=g},a.$observe("max",function(t){g=p(t),s.$validate()})}}}function kr(t,e,n,r){var i=e[0],o=r.$$hasNativeValidators=b(i.validity);o&&r.$parsers.push(function(t){var n=e.prop(Br)||{};return n.badInput||n.typeMismatch?void 0:t})}function Sr(t){t.$$parserName="number",t.$parsers.push(function(e){return t.$isEmpty(e)?null:ia.test(e)?parseFloat(e):void 0}),t.$formatters.push(function(e){if(!t.$isEmpty(e)){if(!x(e))throw Va("numfmt","Expected `{0}` to be a number",e);e=e.toString()}return e})}function Tr(t){return y(t)&&!x(t)&&(t=parseFloat(t)),ui(t)?void 0:t}function Er(t){return(0|t)===t}function Mr(t){var e=t.toString(),n=e.indexOf(".");if(n===-1){if(-1<t&&t<1){var r=/e-(\d+)$/.exec(e);if(r)return Number(r[1])}return 0}return e.length-n-1}function Dr(t,e,n){var r=Number(t);if(!Er(r)||!Er(e)||!Er(n)){var i=Math.max(Mr(r),Mr(e),Mr(n)),o=Math.pow(10,i);r*=o,e*=o,n*=o}return(r-e)%n===0}function Ar(t,e,n,r,i,o){kr(t,e,n,r),Sr(r),wr(t,e,n,r,i,o);var a,s;if((y(n.min)||n.ngMin)&&(r.$validators.min=function(t){return r.$isEmpty(t)||m(a)||t>=a},n.$observe("min",function(t){a=Tr(t),r.$validate()})),(y(n.max)||n.ngMax)&&(r.$validators.max=function(t){return r.$isEmpty(t)||m(s)||t<=s},n.$observe("max",function(t){s=Tr(t),r.$validate()})),y(n.step)||n.ngStep){var u;r.$validators.step=function(t,e){return r.$isEmpty(e)||m(u)||Dr(e,a||0,u)},n.$observe("step",function(t){u=Tr(t),r.$validate()})}}function Or(t,e,n,r,i,o){function a(t,r){e.attr(t,n[t]),n.$observe(t,r)}function s(t){if(f=Tr(t),!ui(r.$modelValue))if(c){var n=e.val();f>n&&(n=f,e.val(n)),r.$setViewValue(n)}else r.$validate()}function u(t){if(p=Tr(t),!ui(r.$modelValue))if(c){var n=e.val();p<n&&(e.val(p),n=p<f?f:p),r.$setViewValue(n)}else r.$validate()}function l(t){h=Tr(t),ui(r.$modelValue)||(c&&r.$viewValue!==e.val()?r.$setViewValue(e.val()):r.$validate())}kr(t,e,n,r),Sr(r),wr(t,e,n,r,i,o);var c=r.$$hasNativeValidators&&"range"===e[0].type,f=c?0:void 0,p=c?100:void 0,h=c?1:void 0,d=e[0].validity,v=y(n.min),g=y(n.max),b=y(n.step),$=r.$render;r.$render=c&&y(d.rangeUnderflow)&&y(d.rangeOverflow)?function(){$(),r.$setViewValue(e.val())}:$,v&&(r.$validators.min=c?function(){return!0}:function(t,e){return r.$isEmpty(e)||m(f)||e>=f},a("min",s)),g&&(r.$validators.max=c?function(){return!0}:function(t,e){return r.$isEmpty(e)||m(p)||e<=p},a("max",u)),b&&(r.$validators.step=c?function(){return!d.stepMismatch}:function(t,e){return r.$isEmpty(e)||m(h)||Dr(e,f||0,h)},a("step",l))}function Nr(t,e,n,r,i,o){wr(t,e,n,r,i,o),br(r),r.$$parserName="url",r.$validators.url=function(t,e){var n=t||e;return r.$isEmpty(n)||na.test(n)}}function Pr(t,e,n,r,i,o){wr(t,e,n,r,i,o),br(r),r.$$parserName="email",r.$validators.email=function(t,e){var n=t||e;return r.$isEmpty(n)||ra.test(n)}}function Rr(t,e,n,r){var i=!n.ngTrim||"false"!==fi(n.ngTrim);m(n.name)&&e.attr("name",a());var o=function(t){var o;e[0].checked&&(o=n.value,i&&(o=fi(o)),r.$setViewValue(o,t&&t.type))};e.on("click",o),r.$render=function(){var t=n.value;i&&(t=fi(t)),e[0].checked=t===r.$viewValue},n.$observe("value",r.$render)}function Ir(t,e,n,r,i){var o;if(y(r)){if(o=t(r),!o.constant)throw Va("constexpr","Expected constant expression for `{0}`, but saw `{1}`.",n,r);return o(e)}return i}function jr(t,e,n,r,i,o,a,s){var u=Ir(s,t,"ngTrueValue",n.ngTrueValue,!0),l=Ir(s,t,"ngFalseValue",n.ngFalseValue,!1),c=function(t){r.$setViewValue(e[0].checked,t&&t.type)};e.on("click",c),r.$render=function(){e[0].checked=r.$viewValue},r.$isEmpty=function(t){return t===!1},r.$formatters.push(function(t){return U(t,u)}),r.$parsers.push(function(t){return t?u:l})}function Fr(t,e){function n(t,e){if(!t||!t.length)return[];if(!e||!e.length)return t;var n=[];t:for(var r=0;r<t.length;r++){for(var i=t[r],o=0;o<e.length;o++)if(i===e[o])continue t;n.push(i)}return n}function i(t){return t&&t.split(" ")}function o(t){var e=t;return li(t)?e=t.map(o).join(" "):b(t)&&(e=Object.keys(t).filter(function(e){return t[e]}).join(" ")),e}function a(t){var e=t;if(li(t))e=t.map(a);else if(b(t)){var n=!1;e=Object.keys(t).filter(function(e){var r=t[e];return!n&&m(r)&&(n=!0),r}),n&&e.push(void 0)}return e}t="ngClass"+t;var s;return["$parse",function(u){return{restrict:"AC",link:function(l,c,f){function p(t){t=v(i(t),1),f.$addClass(t)}function h(t){t=v(i(t),-1),f.$removeClass(t)}function d(t,e){var r=i(t),o=i(e),a=n(r,o),s=n(o,r),u=v(a,-1),l=v(s,1);f.$addClass(l),f.$removeClass(u)}function v(t,e){var n=[];return r(t,function(t){(e>0||k[t])&&(k[t]=(k[t]||0)+e,k[t]===+(e>0)&&n.push(t))}),n.join(" ")}function g(t){t===e?p(b):h(b),S=t}function m(t){var e=o(t);e!==b&&y(e)}function y(t){S===e&&d(b,t),b=t}var b,$=f[t].trim(),w=":"===$.charAt(0)&&":"===$.charAt(1),x=w?a:o,_=u($,x),C=w?m:y,k=c.data("$classCounts"),S=!0;k||(k=gt(),c.data("$classCounts",k)),"ngClass"!==t&&(s||(s=u("$index",function(t){return 1&t})),l.$watch(s,g)),l.$watch(_,C,w)}}}]}function qr(t,e,n,r,i,o,a,s,u){this.$viewValue=Number.NaN,this.$modelValue=Number.NaN,this.$$rawModelValue=void 0,this.$validators={},this.$asyncValidators={},this.$parsers=[],this.$formatters=[],this.$viewChangeListeners=[],this.$untouched=!0,this.$touched=!1,this.$pristine=!0,this.$dirty=!1,this.$valid=!0,this.$invalid=!1,this.$error={},this.$$success={},this.$pending=void 0,this.$name=u(n.name||"",!1)(t),this.$$parentForm=Xo,this.$options=Ua,this.$$parsedNgModel=i(n.ngModel),this.$$parsedNgModelAssign=this.$$parsedNgModel.assign,this.$$ngModelGet=this.$$parsedNgModel,this.$$ngModelSet=this.$$parsedNgModelAssign,this.$$pendingDebounce=null,this.$$parserValid=void 0,this.$$currentValidationRunId=0,this.$$scope=t,this.$$attr=n,this.$$element=r,this.$$animate=o,this.$$timeout=a,this.$$parse=i,this.$$q=s,this.$$exceptionHandler=e,gr(this),Vr(this)}function Vr(t){t.$$scope.$watch(function(){var e=t.$$ngModelGet(t.$$scope);if(e!==t.$modelValue&&(t.$modelValue===t.$modelValue||e===e)){t.$modelValue=t.$$rawModelValue=e,t.$$parserValid=void 0;for(var n=t.$formatters,r=n.length,i=e;r--;)i=n[r](i);t.$viewValue!==i&&(t.$$updateEmptyClasses(i),t.$viewValue=t.$$lastCommittedViewValue=i,t.$render(),t.$$runValidators(t.$modelValue,t.$viewValue,h))}return e})}function Ur(t){this.$$options=t}function Lr(t,e){r(e,function(e,n){y(t[n])||(t[n]=e)})}var Hr=/^\/(.+)\/([a-z]*)$/,Br="validity",zr=Object.prototype.hasOwnProperty,Wr=function(t){return w(t)?t.toLowerCase():t},Yr=function(t){return w(t)?t.toUpperCase():t},Gr=function(t){return w(t)?t.replace(/[A-Z]/g,function(t){return String.fromCharCode(32|t.charCodeAt(0))}):t},Xr=function(t){return w(t)?t.replace(/[a-z]/g,function(t){return String.fromCharCode(t.charCodeAt(0)&-33)}):t};"i"!=="I".toLowerCase()&&(Wr=Gr,Yr=Xr);var Kr,Jr,Zr,Qr,ti=[].slice,ei=[].splice,ni=[].push,ri=Object.prototype.toString,ii=Object.getPrototypeOf,oi=e("ng"),ai=t.angular||(t.angular={}),si=0;Kr=t.document.documentMode;var ui=Number.isNaN||function(t){return t!==t};h.$inject=[],d.$inject=[];var li=Array.isArray,ci=/^\[object (?:Uint8|Uint8Clamped|Uint16|Uint32|Int8|Int16|Int32|Float32|Float64)Array]$/,fi=function(t){return w(t)?t.trim():t},pi=function(t){return t.replace(/([-()[\]{}+?*.$^|,:#<!\\])/g,"\\$1").replace(/\x08/g,"\\x08")},hi=function(){function e(){try{return new Function(""),!1}catch(t){return!0}}if(!y(hi.rules)){var n=t.document.querySelector("[ng-csp]")||t.document.querySelector("[data-ng-csp]");if(n){var r=n.getAttribute("ng-csp")||n.getAttribute("data-ng-csp");hi.rules={noUnsafeEval:!r||r.indexOf("no-unsafe-eval")!==-1,noInlineStyle:!r||r.indexOf("no-inline-style")!==-1}}else hi.rules={noUnsafeEval:e(),noInlineStyle:!1}}return hi.rules},di=function(){if(y(di.name_))return di.name_;var e,n,r,i,o=gi.length;for(n=0;n<o;++n)if(r=gi[n],e=t.document.querySelector("["+r.replace(":","\\:")+"jq]")){i=e.getAttribute(r+"jq");break}return di.name_=i},vi=/:/g,gi=["ng-","data-ng-","ng:","x-ng-"],mi=it(t.document),yi=/[A-Z]/g,bi=!1,$i=1,wi=2,xi=3,_i=8,Ci=9,ki=11,Si={full:"1.6.1",major:1,minor:6,dot:1,codeName:"promise-rectification"};Pt.expando="ng339";var Ti=Pt.cache={},Ei=1;Pt._data=function(t){return this.cache[t[this.expando]]||{}};var Mi=/-([a-z])/g,Di=/^-ms-/,Ai={mouseleave:"mouseout",mouseenter:"mouseover"},Oi=e("jqLite"),Ni=/^<([\w-]+)\s*\/?>(?:<\/\1>|)$/,Pi=/<|&#?\w+;/,Ri=/<([\w:-]+)/,Ii=/<(?!area|br|col|embed|hr|img|input|link|meta|param)(([\w:-]+)[^>]*)\/>/gi,ji={option:[1,'<select multiple="multiple">',"</select>"],thead:[1,"<table>","</table>"],col:[2,"<table><colgroup>","</colgroup></table>"],tr:[2,"<table><tbody>","</tbody></table>"],td:[3,"<table><tbody><tr>","</tr></tbody></table>"],_default:[0,"",""]};ji.optgroup=ji.option,ji.tbody=ji.tfoot=ji.colgroup=ji.caption=ji.thead,ji.th=ji.td;var Fi=t.Node.prototype.contains||function(t){return!!(16&this.compareDocumentPosition(t))},qi=Pt.prototype={ready:Kt,toString:function(){var t=[];return r(this,function(e){t.push(""+e)}),"["+t.join(", ")+"]"},eq:function(t){return Jr(t>=0?this[t]:this[this.length+t])},length:0,push:ni,sort:[].sort,splice:[].splice},Vi={};r("multiple,selected,checked,disabled,readOnly,required,open".split(","),function(t){Vi[Wr(t)]=t});var Ui={};r("input,select,option,textarea,button,form,details".split(","),function(t){Ui[t]=!0});var Li={ngMinlength:"minlength",ngMaxlength:"maxlength",ngMin:"min",ngMax:"max",ngPattern:"pattern",ngStep:"step"};r({data:Vt,removeData:Ft,hasData:Mt,cleanData:Dt},function(t,e){Pt[e]=t}),r({data:Vt,inheritedData:Wt,scope:function(t){return Jr.data(t,"$scope")||Wt(t.parentNode||t,["$isolateScope","$scope"])},isolateScope:function(t){return Jr.data(t,"$isolateScope")||Jr.data(t,"$isolateScopeNoTemplate")},controller:zt,injector:function(t){return Wt(t,"$injector")},removeAttr:function(t,e){t.removeAttribute(e)},hasClass:Ut,css:function(t,e,n){return e=Ct(e),y(n)?void(t.style[e]=n):t.style[e]},attr:function(t,e,n){var r,i=t.nodeType;if(i!==xi&&i!==wi&&i!==_i&&t.getAttribute){var o=Wr(e),a=Vi[o];return y(n)?void(null===n||n===!1&&a?t.removeAttribute(e):t.setAttribute(e,a?o:n)):(r=t.getAttribute(e),a&&null!==r&&(r=o),null===r?void 0:r)}},prop:function(t,e,n){return y(n)?void(t[e]=n):t[e]},text:function(){function t(t,e){if(m(e)){var n=t.nodeType;return n===$i||n===xi?t.textContent:""}t.textContent=e}return t.$dv="",t}(),val:function(t,e){if(m(e)){if(t.multiple&&"select"===j(t)){var n=[];return r(t.options,function(t){t.selected&&n.push(t.value||t.text)}),n}return t.value}t.value=e},html:function(t,e){return m(e)?t.innerHTML:(It(t,!0),void(t.innerHTML=e))},empty:Yt},function(t,e){Pt.prototype[e]=function(e,n){var r,i,o=this.length;if(t!==Yt&&m(2===t.length&&t!==Ut&&t!==zt?e:n)){if(b(e)){for(r=0;r<o;r++)if(t===Vt)t(this[r],e);else for(i in e)t(this[r],i,e[i]);return this}for(var a=t.$dv,s=m(a)?Math.min(o,1):o,u=0;u<s;u++){var l=t(this[u],e,n);a=a?a+l:l}return a}for(r=0;r<o;r++)t(this[r],e,n);return this}}),r({removeData:Ft,on:function(t,e,n,r){if(y(r))throw Oi("onargs","jqLite#on() does not support the `selector` or `eventData` parameters");if(Et(t)){var i=qt(t,!0),o=i.events,a=i.handle;a||(a=i.handle=Qt(t,o));for(var s=e.indexOf(" ")>=0?e.split(" "):[e],u=s.length,l=function(e,r,i){var s=o[e];s||(s=o[e]=[],s.specialHandlerWrapper=r,"$destroy"===e||i||t.addEventListener(e,a)),s.push(n)};u--;)e=s[u],Ai[e]?(l(Ai[e],ee),l(e,void 0,!0)):l(e)}},off:jt,one:function(t,e,n){t=Jr(t),t.on(e,function r(){t.off(e,n),t.off(e,r)}),t.on(e,n)},replaceWith:function(t,e){var n,i=t.parentNode;It(t),r(new Pt(e),function(e){n?i.insertBefore(e,n.nextSibling):i.replaceChild(e,t),n=e})},children:function(t){var e=[];return r(t.childNodes,function(t){t.nodeType===$i&&e.push(t)}),e},contents:function(t){return t.contentDocument||t.childNodes||[]},append:function(t,e){var n=t.nodeType;if(n===$i||n===ki){e=new Pt(e);for(var r=0,i=e.length;r<i;r++){var o=e[r];t.appendChild(o)}}},prepend:function(t,e){if(t.nodeType===$i){var n=t.firstChild;r(new Pt(e),function(e){t.insertBefore(e,n)})}},wrap:function(t,e){Nt(t,Jr(e).eq(0).clone()[0])},remove:Gt,detach:function(t){Gt(t,!0)},after:function(t,e){var n=t,r=t.parentNode;if(r){e=new Pt(e);for(var i=0,o=e.length;i<o;i++){var a=e[i];r.insertBefore(a,n.nextSibling),n=a}}},addClass:Ht,removeClass:Lt,toggleClass:function(t,e,n){e&&r(e.split(" "),function(e){var r=n;m(r)&&(r=!Ut(t,e)),(r?Ht:Lt)(t,e)})},parent:function(t){var e=t.parentNode;return e&&e.nodeType!==ki?e:null},next:function(t){return t.nextElementSibling},find:function(t,e){return t.getElementsByTagName?t.getElementsByTagName(e):[]},clone:Rt,triggerHandler:function(t,e,n){var i,o,a,s=e.type||e,u=qt(t),c=u&&u.events,f=c&&c[s];f&&(i={preventDefault:function(){this.defaultPrevented=!0},isDefaultPrevented:function(){return this.defaultPrevented===!0},stopImmediatePropagation:function(){this.immediatePropagationStopped=!0},isImmediatePropagationStopped:function(){return this.immediatePropagationStopped===!0},stopPropagation:h,type:s,target:t},e.type&&(i=l(i,e)),o=bt(f),a=n?[i].concat(n):[i],r(o,function(e){i.isImmediatePropagationStopped()||e.apply(t,a)}))}},function(t,e){Pt.prototype[e]=function(e,n,r){for(var i,o=0,a=this.length;o<a;o++)m(i)?(i=t(this[o],e,n,r),y(i)&&(i=Jr(i))):Bt(i,t(this[o],e,n,r));return y(i)?i:this}}),Pt.prototype.bind=Pt.prototype.on,Pt.prototype.unbind=Pt.prototype.off,ie.prototype={put:function(t,e){this[re(t,this.nextUid)]=e},get:function(t){return this[re(t,this.nextUid)]},remove:function(t){var e=this[t=re(t,this.nextUid)];return delete this[t],e}};var Hi=[function(){this.$get=[function(){return ie}]}],Bi=/^([^(]+?)=>/,zi=/^[^(]*\(\s*([^)]*)\)/m,Wi=/,/,Yi=/^\s*(_?)(\S+?)\1\s*$/,Gi=/((\/\/.*$)|(\/\*[\s\S]*?\*\/))/gm,Xi=e("$injector");le.$$annotate=ue;var Ki=e("$animate"),Ji=1,Zi="ng-animate",Qi=function(){this.$get=h},to=function(){var t=new ie,e=[];this.$get=["$$AnimateRunner","$rootScope",function(n,i){function o(t,e,n){var i=!1;return e&&(e=w(e)?e.split(" "):li(e)?e:[],r(e,function(e){e&&(i=!0,t[e]=n)})),i}function a(){r(e,function(e){var n=t.get(e);if(n){var i=he(e.attr("class")),o="",a="";r(n,function(t,e){var n=!!i[e];t!==n&&(t?o+=(o.length?" ":"")+e:a+=(a.length?" ":"")+e)}),r(e,function(t){o&&Ht(t,o),a&&Lt(t,a)}),t.remove(e)}}),e.length=0}function s(n,r,s){var u=t.get(n)||{},l=o(u,r,!0),c=o(u,s,!1);(l||c)&&(t.put(n,u),e.push(n),1===e.length&&i.$$postDigest(a))}return{enabled:h,on:h,off:h,pin:h,push:function(t,e,r,i){i&&i(),r=r||{},r.from&&t.css(r.from),r.to&&t.css(r.to),(r.addClass||r.removeClass)&&s(t,r.addClass,r.removeClass);var o=new n;return o.complete(),o}}}]},eo=["$provide",function(t){var e=this;this.$$registeredAnimations=Object.create(null),this.register=function(n,r){if(n&&"."!==n.charAt(0))throw Ki("notcsel","Expecting class selector starting with '.' got '{0}'.",n);var i=n+"-animation";e.$$registeredAnimations[n.substr(1)]=i,t.factory(i,r)},this.classNameFilter=function(t){if(1===arguments.length&&(this.$$classNameFilter=t instanceof RegExp?t:null,this.$$classNameFilter)){var e=new RegExp("(\\s+|\\/)"+Zi+"(\\s+|\\/)");if(e.test(this.$$classNameFilter.toString()))throw Ki("nongcls",'$animateProvider.classNameFilter(regex) prohibits accepting a regex value which matches/contains the "{0}" CSS class.',Zi)}return this.$$classNameFilter},this.$get=["$$animateQueue",function(t){function e(t,e,n){if(n){var r=pe(n);!r||r.parentNode||r.previousElementSibling||(n=null)}n?n.after(t):e.prepend(t)}return{on:t.on,off:t.off,pin:t.pin,enabled:t.enabled,cancel:function(t){t.end&&t.end()},enter:function(n,r,i,o){return r=r&&Jr(r),i=i&&Jr(i),r=r||i.parent(),e(n,r,i),t.push(n,"enter",de(o))},move:function(n,r,i,o){return r=r&&Jr(r),i=i&&Jr(i),r=r||i.parent(),e(n,r,i),t.push(n,"move",de(o))},leave:function(e,n){return t.push(e,"leave",de(n),function(){e.remove()})},addClass:function(e,n,r){return r=de(r),r.addClass=fe(r.addclass,n),t.push(e,"addClass",r)},removeClass:function(e,n,r){return r=de(r),r.removeClass=fe(r.removeClass,n),t.push(e,"removeClass",r)},setClass:function(e,n,r,i){return i=de(i),i.addClass=fe(i.addClass,n),i.removeClass=fe(i.removeClass,r),t.push(e,"setClass",i)},animate:function(e,n,r,i,o){return o=de(o),o.from=o.from?l(o.from,n):n,o.to=o.to?l(o.to,r):r,i=i||"ng-inline-animate",o.tempClasses=fe(o.tempClasses,i),t.push(e,"animate",o)}}}]}],no=function(){this.$get=["$$rAF",function(t){function e(e){n.push(e),n.length>1||t(function(){for(var t=0;t<n.length;t++)n[t]();n=[]})}var n=[];return function(){var t=!1;return e(function(){t=!0}),function(n){t?n():e(n)}}}]},ro=function(){this.$get=["$q","$sniffer","$$animateAsyncRun","$$isDocumentHidden","$timeout",function(t,e,n,i,o){function a(t){this.setHost(t);var e=n(),r=function(t){o(t,0,!1)};this._doneCallbacks=[],this._tick=function(t){i()?r(t):e(t)},this._state=0}var s=0,u=1,l=2;return a.chain=function(t,e){function n(){return r===t.length?void e(!0):void t[r](function(t){return t===!1?void e(!1):(r++,void n())})}var r=0;n()},a.all=function(t,e){function n(n){o=o&&n,++i===t.length&&e(o)}var i=0,o=!0;r(t,function(t){t.done(n)})},a.prototype={setHost:function(t){this.host=t||{}},done:function(t){this._state===l?t():this._doneCallbacks.push(t)},progress:h,getPromise:function(){if(!this.promise){var e=this;this.promise=t(function(t,n){e.done(function(e){e===!1?n():t()})})}return this.promise},then:function(t,e){return this.getPromise().then(t,e)},catch:function(t){return this.getPromise().catch(t)},finally:function(t){return this.getPromise().finally(t)},pause:function(){this.host.pause&&this.host.pause()},resume:function(){this.host.resume&&this.host.resume()},end:function(){this.host.end&&this.host.end(),this._resolve(!0)},cancel:function(){this.host.cancel&&this.host.cancel(),this._resolve(!1)},complete:function(t){var e=this;e._state===s&&(e._state=u,e._tick(function(){e._resolve(t)}))},_resolve:function(t){this._state!==l&&(r(this._doneCallbacks,function(e){e(t)}),this._doneCallbacks.length=0,this._state=l)}},a}]},io=function(){this.$get=["$$rAF","$q","$$AnimateRunner",function(t,e,n){return function(e,r){function i(){return t(function(){o(),s||u.complete(),s=!0}),u}function o(){a.addClass&&(e.addClass(a.addClass),a.addClass=null),a.removeClass&&(e.removeClass(a.removeClass),a.removeClass=null),a.to&&(e.css(a.to),a.to=null)}var a=r||{};a.$$prepared||(a=V(a)),a.cleanupStyles&&(a.from=a.to=null),a.from&&(e.css(a.from),a.from=null);var s,u=new n;return{start:i,end:i}}}]},oo=e("$compile"),ao=new be;$e.$inject=["$provide","$$sanitizeUriProvider"],we.prototype.isFirstChange=function(){return this.previousValue===ao};var so=/^((?:x|data)[:\-_])/i,uo=/[:\-_]+(.)/g,lo=e("$controller"),co=/^(\S+)(\s+as\s+([\w$]+))?$/,fo=function(){this.$get=["$document",function(t){return function(e){return e?!e.nodeType&&e instanceof Jr&&(e=e[0]):e=t[0].body,e.offsetWidth+1}}]},po="application/json",ho={"Content-Type":po+";charset=utf-8"},vo=/^\[|^\{(?!\{)/,go={"[":/]$/,"{":/}$/},mo=/^\)]\}',?\n/,yo=e("$http"),bo=ai.$interpolateMinErr=e("$interpolate");bo.throwNoconcat=function(t){throw bo("noconcat","Error while interpolating: {0}\nStrict Contextual Escaping disallows interpolations that concatenate multiple expressions when a trusted value is required.  See http://docs.angularjs.org/api/ng.$sce",t)},bo.interr=function(t,e){return bo("interr","Can't interpolate: {0}\n{1}",t,e.toString())};var $o=function(){this.$get=["$window",function(t){function e(t){var e=function(t){e.data=t,e.called=!0};return e.id=t,e}var n=t.angular.callbacks,r={};return{createCallback:function(t){var i="_"+(n.$$counter++).toString(36),o="angular.callbacks."+i,a=e(i);return r[o]=n[i]=a,o},wasCalled:function(t){return r[t].called},getResponse:function(t){return r[t].data},removeCallback:function(t){var e=r[t];delete n[e.id],delete r[t]}}}]},wo=/^([^?#]*)(\?([^#]*))?(#(.*))?$/,xo={http:80,https:443,ftp:21},_o=e("$location"),Co=/^\s*[\\/]{2,}/,ko={$$absUrl:"",$$html5:!1,$$replace:!1,absUrl:rn("$$absUrl"),url:function(t){if(m(t))return this.$$url;var e=wo.exec(t);return(e[1]||""===t)&&this.path(decodeURIComponent(e[1])),(e[2]||e[1]||""===t)&&this.search(e[3]||""),this.hash(e[5]||""),this},protocol:rn("$$protocol"),host:rn("$$host"),port:rn("$$port"),path:on("$$path",function(t){return t=null!==t?t.toString():"","/"===t.charAt(0)?t:"/"+t}),search:function(t,e){switch(arguments.length){case 0:return this.$$search;case 1:if(w(t)||x(t))t=t.toString(),this.$$search=Q(t);else{if(!b(t))throw _o("isrcharg","The first argument of the `$location#search()` call must be a string or an object.");t=V(t,{}),r(t,function(e,n){null==e&&delete t[n]}),this.$$search=t}break;default:m(e)||null===e?delete this.$$search[t]:this.$$search[t]=e}return this.$$compose(),this},hash:on("$$hash",function(t){return null!==t?t.toString():""}),replace:function(){return this.$$replace=!0,this}};r([nn,en,tn],function(t){t.prototype=Object.create(ko),t.prototype.state=function(e){if(!arguments.length)return this.$$state;if(t!==tn||!this.$$html5)throw _o("nostate","History API state support is available only in HTML5 mode and only in browsers supporting HTML5 History API");return this.$$state=m(e)?null:e,this}});var So=e("$parse"),To={}.constructor.prototype.valueOf,Eo=gt();r("+ - * / % === !== == != < > <= >= && || ! = |".split(" "),function(t){Eo[t]=!0});var Mo={n:"\n",f:"\f",r:"\r",t:"\t",v:"\v","'":"'",'"':'"'},Do=function(t){this.options=t};Do.prototype={constructor:Do,lex:function(t){for(this.text=t,this.index=0,this.tokens=[];this.index<this.text.length;){var e=this.text.charAt(this.index);if('"'===e||"'"===e)this.readString(e);else if(this.isNumber(e)||"."===e&&this.isNumber(this.peek()))this.readNumber();else if(this.isIdentifierStart(this.peekMultichar()))this.readIdent();else if(this.is(e,"(){}[].,;:?"))this.tokens.push({index:this.index,text:e}),this.index++;else if(this.isWhitespace(e))this.index++;else{var n=e+this.peek(),r=n+this.peek(2),i=Eo[e],o=Eo[n],a=Eo[r];if(i||o||a){var s=a?r:o?n:e;this.tokens.push({index:this.index,text:s,operator:!0}),this.index+=s.length}else this.throwError("Unexpected next character ",this.index,this.index+1)}}return this.tokens},is:function(t,e){return e.indexOf(t)!==-1},peek:function(t){var e=t||1;return this.index+e<this.text.length&&this.text.charAt(this.index+e)},isNumber:function(t){return"0"<=t&&t<="9"&&"string"==typeof t},isWhitespace:function(t){return" "===t||"\r"===t||"\t"===t||"\n"===t||"\v"===t||" "===t},isIdentifierStart:function(t){return this.options.isIdentifierStart?this.options.isIdentifierStart(t,this.codePointAt(t)):this.isValidIdentifierStart(t)},isValidIdentifierStart:function(t){return"a"<=t&&t<="z"||"A"<=t&&t<="Z"||"_"===t||"$"===t},isIdentifierContinue:function(t){return this.options.isIdentifierContinue?this.options.isIdentifierContinue(t,this.codePointAt(t)):this.isValidIdentifierContinue(t)},isValidIdentifierContinue:function(t,e){return this.isValidIdentifierStart(t,e)||this.isNumber(t)},codePointAt:function(t){return 1===t.length?t.charCodeAt(0):(t.charCodeAt(0)<<10)+t.charCodeAt(1)-56613888},peekMultichar:function(){var t=this.text.charAt(this.index),e=this.peek();if(!e)return t;var n=t.charCodeAt(0),r=e.charCodeAt(0);return n>=55296&&n<=56319&&r>=56320&&r<=57343?t+e:t},isExpOperator:function(t){return"-"===t||"+"===t||this.isNumber(t)},throwError:function(t,e,n){n=n||this.index;var r=y(e)?"s "+e+"-"+this.index+" ["+this.text.substring(e,n)+"]":" "+n;throw So("lexerr","Lexer Error: {0} at column{1} in expression [{2}].",t,r,this.text)},readNumber:function(){for(var t="",e=this.index;this.index<this.text.length;){var n=Wr(this.text.charAt(this.index));if("."===n||this.isNumber(n))t+=n;else{var r=this.peek();if("e"===n&&this.isExpOperator(r))t+=n;else if(this.isExpOperator(n)&&r&&this.isNumber(r)&&"e"===t.charAt(t.length-1))t+=n;else{if(!this.isExpOperator(n)||r&&this.isNumber(r)||"e"!==t.charAt(t.length-1))break;this.throwError("Invalid exponent")}}this.index++}this.tokens.push({index:e,text:t,constant:!0,value:Number(t)})},readIdent:function(){var t=this.index;for(this.index+=this.peekMultichar().length;this.index<this.text.length;){var e=this.peekMultichar();if(!this.isIdentifierContinue(e))break;this.index+=e.length}this.tokens.push({index:t,text:this.text.slice(t,this.index),identifier:!0})},readString:function(t){var e=this.index;this.index++;for(var n="",r=t,i=!1;this.index<this.text.length;){var o=this.text.charAt(this.index);if(r+=o,i){if("u"===o){var a=this.text.substring(this.index+1,this.index+5);a.match(/[\da-f]{4}/i)||this.throwError("Invalid unicode escape [\\u"+a+"]"),this.index+=4,n+=String.fromCharCode(parseInt(a,16))}else{var s=Mo[o];n+=s||o}i=!1}else if("\\"===o)i=!0;else{if(o===t)return this.index++,void this.tokens.push({index:e,text:r,constant:!0,value:n});n+=o}this.index++}this.throwError("Unterminated quote",e)}};var Ao=function(t,e){this.lexer=t,this.options=e};Ao.Program="Program",Ao.ExpressionStatement="ExpressionStatement",Ao.AssignmentExpression="AssignmentExpression",Ao.ConditionalExpression="ConditionalExpression",Ao.LogicalExpression="LogicalExpression",Ao.BinaryExpression="BinaryExpression",Ao.UnaryExpression="UnaryExpression",Ao.CallExpression="CallExpression",Ao.MemberExpression="MemberExpression",Ao.Identifier="Identifier",Ao.Literal="Literal",Ao.ArrayExpression="ArrayExpression",Ao.Property="Property",Ao.ObjectExpression="ObjectExpression",Ao.ThisExpression="ThisExpression",Ao.LocalsExpression="LocalsExpression",Ao.NGValueParameter="NGValueParameter",Ao.prototype={ast:function(t){this.text=t,this.tokens=this.lexer.lex(t);var e=this.program();return 0!==this.tokens.length&&this.throwError("is an unexpected token",this.tokens[0]),e},program:function(){for(var t=[];;)if(this.tokens.length>0&&!this.peek("}",")",";","]")&&t.push(this.expressionStatement()),!this.expect(";"))return{type:Ao.Program,body:t}},expressionStatement:function(){return{type:Ao.ExpressionStatement,expression:this.filterChain()}},filterChain:function(){for(var t=this.expression();this.expect("|");)t=this.filter(t);return t},expression:function(){return this.assignment()},assignment:function(){var t=this.ternary();if(this.expect("=")){if(!dn(t))throw So("lval","Trying to assign a value to a non l-value");t={type:Ao.AssignmentExpression,left:t,right:this.assignment(),operator:"="}}return t},ternary:function(){var t,e,n=this.logicalOR();return this.expect("?")&&(t=this.expression(),this.consume(":"))?(e=this.expression(),{type:Ao.ConditionalExpression,test:n,alternate:t,consequent:e}):n},logicalOR:function(){for(var t=this.logicalAND();this.expect("||");)t={type:Ao.LogicalExpression,operator:"||",left:t,right:this.logicalAND()};return t},logicalAND:function(){for(var t=this.equality();this.expect("&&");)t={type:Ao.LogicalExpression,operator:"&&",left:t,right:this.equality()};return t},equality:function(){for(var t,e=this.relational();t=this.expect("==","!=","===","!==");)e={type:Ao.BinaryExpression,operator:t.text,left:e,right:this.relational()};return e},relational:function(){for(var t,e=this.additive();t=this.expect("<",">","<=",">=");)e={type:Ao.BinaryExpression,operator:t.text,left:e,right:this.additive()};return e},additive:function(){for(var t,e=this.multiplicative();t=this.expect("+","-");)e={type:Ao.BinaryExpression,operator:t.text,left:e,right:this.multiplicative()};return e},multiplicative:function(){for(var t,e=this.unary();t=this.expect("*","/","%");)e={type:Ao.BinaryExpression,operator:t.text,left:e,right:this.unary()};return e},unary:function(){var t;return(t=this.expect("+","-","!"))?{type:Ao.UnaryExpression,operator:t.text,prefix:!0,argument:this.unary()}:this.primary()},primary:function(){var t;this.expect("(")?(t=this.filterChain(),this.consume(")")):this.expect("[")?t=this.arrayDeclaration():this.expect("{")?t=this.object():this.selfReferential.hasOwnProperty(this.peek().text)?t=V(this.selfReferential[this.consume().text]):this.options.literals.hasOwnProperty(this.peek().text)?t={type:Ao.Literal,value:this.options.literals[this.consume().text]}:this.peek().identifier?t=this.identifier():this.peek().constant?t=this.constant():this.throwError("not a primary expression",this.peek());for(var e;e=this.expect("(","[",".");)"("===e.text?(t={type:Ao.CallExpression,callee:t,arguments:this.parseArguments()},this.consume(")")):"["===e.text?(t={type:Ao.MemberExpression,object:t,property:this.expression(),computed:!0},this.consume("]")):"."===e.text?t={type:Ao.MemberExpression,object:t,property:this.identifier(),computed:!1}:this.throwError("IMPOSSIBLE");return t},filter:function(t){for(var e=[t],n={type:Ao.CallExpression,callee:this.identifier(),arguments:e,filter:!0};this.expect(":");)e.push(this.expression());return n},parseArguments:function(){var t=[];if(")"!==this.peekToken().text)do t.push(this.filterChain());while(this.expect(","));return t},identifier:function(){var t=this.consume();return t.identifier||this.throwError("is not a valid identifier",t),{type:Ao.Identifier,name:t.text}},constant:function(){return{type:Ao.Literal,value:this.consume().value}},arrayDeclaration:function(){var t=[];if("]"!==this.peekToken().text)do{if(this.peek("]"))break;t.push(this.expression())}while(this.expect(","));return this.consume("]"),{type:Ao.ArrayExpression,elements:t}},object:function(){var t,e=[];if("}"!==this.peekToken().text)do{if(this.peek("}"))break;t={type:Ao.Property,kind:"init"},this.peek().constant?(t.key=this.constant(),t.computed=!1,this.consume(":"),t.value=this.expression()):this.peek().identifier?(t.key=this.identifier(),t.computed=!1,this.peek(":")?(this.consume(":"),t.value=this.expression()):t.value=t.key):this.peek("[")?(this.consume("["),t.key=this.expression(),this.consume("]"),t.computed=!0,this.consume(":"),t.value=this.expression()):this.throwError("invalid key",this.peek()),e.push(t)}while(this.expect(","));return this.consume("}"),{type:Ao.ObjectExpression,properties:e}},throwError:function(t,e){throw So("syntax","Syntax Error: Token '{0}' {1} at column {2} of the expression [{3}] starting at [{4}].",e.text,t,e.index+1,this.text,this.text.substring(e.index))},consume:function(t){if(0===this.tokens.length)throw So("ueoe","Unexpected end of expression: {0}",this.text);var e=this.expect(t);return e||this.throwError("is unexpected, expecting ["+t+"]",this.peek()),e},peekToken:function(){if(0===this.tokens.length)throw So("ueoe","Unexpected end of expression: {0}",this.text);return this.tokens[0]},peek:function(t,e,n,r){return this.peekAhead(0,t,e,n,r)},peekAhead:function(t,e,n,r,i){if(this.tokens.length>t){var o=this.tokens[t],a=o.text;if(a===e||a===n||a===r||a===i||!e&&!n&&!r&&!i)return o}return!1},expect:function(t,e,n,r){var i=this.peek(t,e,n,r);return!!i&&(this.tokens.shift(),i)},selfReferential:{this:{type:Ao.ThisExpression},$locals:{type:Ao.LocalsExpression}}},yn.prototype={compile:function(t){var e=this,n=this.astBuilder.ast(t);this.state={nextId:0,filters:{},fn:{vars:[],body:[],own:{}},assign:{vars:[],body:[],own:{}},inputs:[]},pn(n,e.$filter);var i,o="";if(this.stage="assign",i=vn(n)){this.state.computing="assign";var a=this.nextId();this.recurse(i,a),this.return_(a),o="fn.assign="+this.generateFunction("assign","s,v,l")}var s=hn(n.body);e.stage="inputs",r(s,function(t,n){var r="fn"+n;e.state[r]={vars:[],body:[],own:{}},e.state.computing=r;var i=e.nextId();e.recurse(t,i),e.return_(i),e.state.inputs.push(r),
t.watchId=n}),this.state.computing="fn",this.stage="main",this.recurse(n);var u='"'+this.USE+" "+this.STRICT+'";\n'+this.filterPrefix()+"var fn="+this.generateFunction("fn","s,l,a,i")+o+this.watchFns()+"return fn;",l=new Function("$filter","getStringValue","ifDefined","plus",u)(this.$filter,un,ln,cn);return this.state=this.stage=void 0,l.literal=gn(n),l.constant=mn(n),l},USE:"use",STRICT:"strict",watchFns:function(){var t=[],e=this.state.inputs,n=this;return r(e,function(e){t.push("var "+e+"="+n.generateFunction(e,"s"))}),e.length&&t.push("fn.inputs=["+e.join(",")+"];"),t.join("")},generateFunction:function(t,e){return"function("+e+"){"+this.varsPrefix(t)+this.body(t)+"};"},filterPrefix:function(){var t=[],e=this;return r(this.state.filters,function(n,r){t.push(n+"=$filter("+e.escape(r)+")")}),t.length?"var "+t.join(",")+";":""},varsPrefix:function(t){return this.state[t].vars.length?"var "+this.state[t].vars.join(",")+";":""},body:function(t){return this.state[t].body.join("")},recurse:function(t,e,n,i,o,a){var s,u,l,c,f,p=this;if(i=i||h,!a&&y(t.watchId))return e=e||this.nextId(),void this.if_("i",this.lazyAssign(e,this.computedMember("i",t.watchId)),this.lazyRecurse(t,e,n,i,o,!0));switch(t.type){case Ao.Program:r(t.body,function(e,n){p.recurse(e.expression,void 0,void 0,function(t){u=t}),n!==t.body.length-1?p.current().body.push(u,";"):p.return_(u)});break;case Ao.Literal:c=this.escape(t.value),this.assign(e,c),i(e||c);break;case Ao.UnaryExpression:this.recurse(t.argument,void 0,void 0,function(t){u=t}),c=t.operator+"("+this.ifDefined(u,0)+")",this.assign(e,c),i(c);break;case Ao.BinaryExpression:this.recurse(t.left,void 0,void 0,function(t){s=t}),this.recurse(t.right,void 0,void 0,function(t){u=t}),c="+"===t.operator?this.plus(s,u):"-"===t.operator?this.ifDefined(s,0)+t.operator+this.ifDefined(u,0):"("+s+")"+t.operator+"("+u+")",this.assign(e,c),i(c);break;case Ao.LogicalExpression:e=e||this.nextId(),p.recurse(t.left,e),p.if_("&&"===t.operator?e:p.not(e),p.lazyRecurse(t.right,e)),i(e);break;case Ao.ConditionalExpression:e=e||this.nextId(),p.recurse(t.test,e),p.if_(e,p.lazyRecurse(t.alternate,e),p.lazyRecurse(t.consequent,e)),i(e);break;case Ao.Identifier:e=e||this.nextId(),n&&(n.context="inputs"===p.stage?"s":this.assign(this.nextId(),this.getHasOwnProperty("l",t.name)+"?l:s"),n.computed=!1,n.name=t.name),p.if_("inputs"===p.stage||p.not(p.getHasOwnProperty("l",t.name)),function(){p.if_("inputs"===p.stage||"s",function(){o&&1!==o&&p.if_(p.isNull(p.nonComputedMember("s",t.name)),p.lazyAssign(p.nonComputedMember("s",t.name),"{}")),p.assign(e,p.nonComputedMember("s",t.name))})},e&&p.lazyAssign(e,p.nonComputedMember("l",t.name))),i(e);break;case Ao.MemberExpression:s=n&&(n.context=this.nextId())||this.nextId(),e=e||this.nextId(),p.recurse(t.object,s,void 0,function(){p.if_(p.notNull(s),function(){t.computed?(u=p.nextId(),p.recurse(t.property,u),p.getStringValue(u),o&&1!==o&&p.if_(p.not(p.computedMember(s,u)),p.lazyAssign(p.computedMember(s,u),"{}")),c=p.computedMember(s,u),p.assign(e,c),n&&(n.computed=!0,n.name=u)):(o&&1!==o&&p.if_(p.isNull(p.nonComputedMember(s,t.property.name)),p.lazyAssign(p.nonComputedMember(s,t.property.name),"{}")),c=p.nonComputedMember(s,t.property.name),p.assign(e,c),n&&(n.computed=!1,n.name=t.property.name))},function(){p.assign(e,"undefined")}),i(e)},!!o);break;case Ao.CallExpression:e=e||this.nextId(),t.filter?(u=p.filter(t.callee.name),l=[],r(t.arguments,function(t){var e=p.nextId();p.recurse(t,e),l.push(e)}),c=u+"("+l.join(",")+")",p.assign(e,c),i(e)):(u=p.nextId(),s={},l=[],p.recurse(t.callee,u,s,function(){p.if_(p.notNull(u),function(){r(t.arguments,function(e){p.recurse(e,t.constant?void 0:p.nextId(),void 0,function(t){l.push(t)})}),c=s.name?p.member(s.context,s.name,s.computed)+"("+l.join(",")+")":u+"("+l.join(",")+")",p.assign(e,c)},function(){p.assign(e,"undefined")}),i(e)}));break;case Ao.AssignmentExpression:u=this.nextId(),s={},this.recurse(t.left,void 0,s,function(){p.if_(p.notNull(s.context),function(){p.recurse(t.right,u),c=p.member(s.context,s.name,s.computed)+t.operator+u,p.assign(e,c),i(e||c)})},1);break;case Ao.ArrayExpression:l=[],r(t.elements,function(e){p.recurse(e,t.constant?void 0:p.nextId(),void 0,function(t){l.push(t)})}),c="["+l.join(",")+"]",this.assign(e,c),i(e||c);break;case Ao.ObjectExpression:l=[],f=!1,r(t.properties,function(t){t.computed&&(f=!0)}),f?(e=e||this.nextId(),this.assign(e,"{}"),r(t.properties,function(t){t.computed?(s=p.nextId(),p.recurse(t.key,s)):s=t.key.type===Ao.Identifier?t.key.name:""+t.key.value,u=p.nextId(),p.recurse(t.value,u),p.assign(p.member(e,s,t.computed),u)})):(r(t.properties,function(e){p.recurse(e.value,t.constant?void 0:p.nextId(),void 0,function(t){l.push(p.escape(e.key.type===Ao.Identifier?e.key.name:""+e.key.value)+":"+t)})}),c="{"+l.join(",")+"}",this.assign(e,c)),i(e||c);break;case Ao.ThisExpression:this.assign(e,"s"),i(e||"s");break;case Ao.LocalsExpression:this.assign(e,"l"),i(e||"l");break;case Ao.NGValueParameter:this.assign(e,"v"),i(e||"v")}},getHasOwnProperty:function(t,e){var n=t+"."+e,r=this.current().own;return r.hasOwnProperty(n)||(r[n]=this.nextId(!1,t+"&&("+this.escape(e)+" in "+t+")")),r[n]},assign:function(t,e){if(t)return this.current().body.push(t,"=",e,";"),t},filter:function(t){return this.state.filters.hasOwnProperty(t)||(this.state.filters[t]=this.nextId(!0)),this.state.filters[t]},ifDefined:function(t,e){return"ifDefined("+t+","+this.escape(e)+")"},plus:function(t,e){return"plus("+t+","+e+")"},return_:function(t){this.current().body.push("return ",t,";")},if_:function(t,e,n){if(t===!0)e();else{var r=this.current().body;r.push("if(",t,"){"),e(),r.push("}"),n&&(r.push("else{"),n(),r.push("}"))}},not:function(t){return"!("+t+")"},isNull:function(t){return t+"==null"},notNull:function(t){return t+"!=null"},nonComputedMember:function(t,e){var n=/^[$_a-zA-Z][$_a-zA-Z0-9]*$/,r=/[^$_a-zA-Z0-9]/g;return n.test(e)?t+"."+e:t+'["'+e.replace(r,this.stringEscapeFn)+'"]'},computedMember:function(t,e){return t+"["+e+"]"},member:function(t,e,n){return n?this.computedMember(t,e):this.nonComputedMember(t,e)},getStringValue:function(t){this.assign(t,"getStringValue("+t+")")},lazyRecurse:function(t,e,n,r,i,o){var a=this;return function(){a.recurse(t,e,n,r,i,o)}},lazyAssign:function(t,e){var n=this;return function(){n.assign(t,e)}},stringEscapeRegex:/[^ a-zA-Z0-9]/g,stringEscapeFn:function(t){return"\\u"+("0000"+t.charCodeAt(0).toString(16)).slice(-4)},escape:function(t){if(w(t))return"'"+t.replace(this.stringEscapeRegex,this.stringEscapeFn)+"'";if(x(t))return t.toString();if(t===!0)return"true";if(t===!1)return"false";if(null===t)return"null";if("undefined"==typeof t)return"undefined";throw So("esc","IMPOSSIBLE")},nextId:function(t,e){var n="v"+this.state.nextId++;return t||this.current().vars.push(n+(e?"="+e:"")),n},current:function(){return this.state[this.state.computing]}},bn.prototype={compile:function(t){var e=this,n=this.astBuilder.ast(t);pn(n,e.$filter);var i,o;(i=vn(n))&&(o=this.recurse(i));var a,s=hn(n.body);s&&(a=[],r(s,function(t,n){var r=e.recurse(t);t.input=r,a.push(r),t.watchId=n}));var u=[];r(n.body,function(t){u.push(e.recurse(t.expression))});var l=0===n.body.length?h:1===n.body.length?u[0]:function(t,e){var n;return r(u,function(r){n=r(t,e)}),n};return o&&(l.assign=function(t,e,n){return o(t,n,e)}),a&&(l.inputs=a),l.literal=gn(n),l.constant=mn(n),l},recurse:function(t,e,n){var i,o,a,s=this;if(t.input)return this.inputs(t.input,t.watchId);switch(t.type){case Ao.Literal:return this.value(t.value,e);case Ao.UnaryExpression:return o=this.recurse(t.argument),this["unary"+t.operator](o,e);case Ao.BinaryExpression:return i=this.recurse(t.left),o=this.recurse(t.right),this["binary"+t.operator](i,o,e);case Ao.LogicalExpression:return i=this.recurse(t.left),o=this.recurse(t.right),this["binary"+t.operator](i,o,e);case Ao.ConditionalExpression:return this["ternary?:"](this.recurse(t.test),this.recurse(t.alternate),this.recurse(t.consequent),e);case Ao.Identifier:return s.identifier(t.name,e,n);case Ao.MemberExpression:return i=this.recurse(t.object,!1,!!n),t.computed||(o=t.property.name),t.computed&&(o=this.recurse(t.property)),t.computed?this.computedMember(i,o,e,n):this.nonComputedMember(i,o,e,n);case Ao.CallExpression:return a=[],r(t.arguments,function(t){a.push(s.recurse(t))}),t.filter&&(o=this.$filter(t.callee.name)),t.filter||(o=this.recurse(t.callee,!0)),t.filter?function(t,n,r,i){for(var s=[],u=0;u<a.length;++u)s.push(a[u](t,n,r,i));var l=o.apply(void 0,s,i);return e?{context:void 0,name:void 0,value:l}:l}:function(t,n,r,i){var s,u=o(t,n,r,i);if(null!=u.value){for(var l=[],c=0;c<a.length;++c)l.push(a[c](t,n,r,i));s=u.value.apply(u.context,l)}return e?{value:s}:s};case Ao.AssignmentExpression:return i=this.recurse(t.left,!0,1),o=this.recurse(t.right),function(t,n,r,a){var s=i(t,n,r,a),u=o(t,n,r,a);return s.context[s.name]=u,e?{value:u}:u};case Ao.ArrayExpression:return a=[],r(t.elements,function(t){a.push(s.recurse(t))}),function(t,n,r,i){for(var o=[],s=0;s<a.length;++s)o.push(a[s](t,n,r,i));return e?{value:o}:o};case Ao.ObjectExpression:return a=[],r(t.properties,function(t){t.computed?a.push({key:s.recurse(t.key),computed:!0,value:s.recurse(t.value)}):a.push({key:t.key.type===Ao.Identifier?t.key.name:""+t.key.value,computed:!1,value:s.recurse(t.value)})}),function(t,n,r,i){for(var o={},s=0;s<a.length;++s)a[s].computed?o[a[s].key(t,n,r,i)]=a[s].value(t,n,r,i):o[a[s].key]=a[s].value(t,n,r,i);return e?{value:o}:o};case Ao.ThisExpression:return function(t){return e?{value:t}:t};case Ao.LocalsExpression:return function(t,n){return e?{value:n}:n};case Ao.NGValueParameter:return function(t,n,r){return e?{value:r}:r}}},"unary+":function(t,e){return function(n,r,i,o){var a=t(n,r,i,o);return a=y(a)?+a:0,e?{value:a}:a}},"unary-":function(t,e){return function(n,r,i,o){var a=t(n,r,i,o);return a=y(a)?-a:-0,e?{value:a}:a}},"unary!":function(t,e){return function(n,r,i,o){var a=!t(n,r,i,o);return e?{value:a}:a}},"binary+":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a),u=e(r,i,o,a),l=cn(s,u);return n?{value:l}:l}},"binary-":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a),u=e(r,i,o,a),l=(y(s)?s:0)-(y(u)?u:0);return n?{value:l}:l}},"binary*":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)*e(r,i,o,a);return n?{value:s}:s}},"binary/":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)/e(r,i,o,a);return n?{value:s}:s}},"binary%":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)%e(r,i,o,a);return n?{value:s}:s}},"binary===":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)===e(r,i,o,a);return n?{value:s}:s}},"binary!==":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)!==e(r,i,o,a);return n?{value:s}:s}},"binary==":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)==e(r,i,o,a);return n?{value:s}:s}},"binary!=":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)!=e(r,i,o,a);return n?{value:s}:s}},"binary<":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)<e(r,i,o,a);return n?{value:s}:s}},"binary>":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)>e(r,i,o,a);return n?{value:s}:s}},"binary<=":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)<=e(r,i,o,a);return n?{value:s}:s}},"binary>=":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)>=e(r,i,o,a);return n?{value:s}:s}},"binary&&":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)&&e(r,i,o,a);return n?{value:s}:s}},"binary||":function(t,e,n){return function(r,i,o,a){var s=t(r,i,o,a)||e(r,i,o,a);return n?{value:s}:s}},"ternary?:":function(t,e,n,r){return function(i,o,a,s){var u=t(i,o,a,s)?e(i,o,a,s):n(i,o,a,s);return r?{value:u}:u}},value:function(t,e){return function(){return e?{context:void 0,name:void 0,value:t}:t}},identifier:function(t,e,n){return function(r,i,o,a){var s=i&&t in i?i:r;n&&1!==n&&s&&null==s[t]&&(s[t]={});var u=s?s[t]:void 0;return e?{context:s,name:t,value:u}:u}},computedMember:function(t,e,n,r){return function(i,o,a,s){var u,l,c=t(i,o,a,s);return null!=c&&(u=e(i,o,a,s),u=un(u),r&&1!==r&&c&&!c[u]&&(c[u]={}),l=c[u]),n?{context:c,name:u,value:l}:l}},nonComputedMember:function(t,e,n,r){return function(i,o,a,s){var u=t(i,o,a,s);r&&1!==r&&u&&null==u[e]&&(u[e]={});var l=null!=u?u[e]:void 0;return n?{context:u,name:e,value:l}:l}},inputs:function(t,e){return function(n,r,i,o){return o?o[e]:t(n,r,i)}}};var Oo=function(t,e,n){this.lexer=t,this.$filter=e,this.options=n,this.ast=new Ao(t,n),this.astCompiler=n.csp?new bn(this.ast,e):new yn(this.ast,e)};Oo.prototype={constructor:Oo,parse:function(t){return this.astCompiler.compile(t)}};var No=e("$sce"),Po={HTML:"html",CSS:"css",URL:"url",RESOURCE_URL:"resourceUrl",JS:"js"},Ro=/_([a-z])/g,Io=e("$compile"),jo=t.document.createElement("a"),Fo=jn(t.location.href);Vn.$inject=["$document"],Ln.$inject=["$provide"];var qo=22,Vo=".",Uo="0";Yn.$inject=["$locale"],Gn.$inject=["$locale"];var Lo={yyyy:Qn("FullYear",4,0,!1,!0),yy:Qn("FullYear",2,0,!0,!0),y:Qn("FullYear",1,0,!1,!0),MMMM:tr("Month"),MMM:tr("Month",!0),MM:Qn("Month",2,1),M:Qn("Month",1,1),LLLL:tr("Month",!1,!0),dd:Qn("Date",2),d:Qn("Date",1),HH:Qn("Hours",2),H:Qn("Hours",1),hh:Qn("Hours",2,-12),h:Qn("Hours",1,-12),mm:Qn("Minutes",2),m:Qn("Minutes",1),ss:Qn("Seconds",2),s:Qn("Seconds",1),sss:Qn("Milliseconds",3),EEEE:tr("Day"),EEE:tr("Day",!0),a:or,Z:er,ww:ir(2),w:ir(1),G:ar,GG:ar,GGG:ar,GGGG:sr},Ho=/((?:[^yMLdHhmsaZEwG']+)|(?:'(?:[^']|'')*')|(?:E+|y+|M+|L+|d+|H+|h+|m+|s+|a|Z|G+|w+))(.*)/,Bo=/^-?\d+$/;ur.$inject=["$locale"];var zo=v(Wr),Wo=v(Yr);pr.$inject=["$parse"];var Yo=v({restrict:"E",compile:function(t,e){if(!e.href&&!e.xlinkHref)return function(t,e){if("a"===e[0].nodeName.toLowerCase()){var n="[object SVGAnimatedString]"===ri.call(e.prop("href"))?"xlink:href":"href";e.on("click",function(t){e.attr(n)||t.preventDefault()})}}}}),Go={};r(Vi,function(t,e){function n(t,n,i){t.$watch(i[r],function(t){i.$set(e,!!t)})}if("multiple"!==t){var r=xe("ng-"+e),i=n;"checked"===t&&(i=function(t,e,i){i.ngModel!==i[r]&&n(t,e,i)}),Go[r]=function(){return{restrict:"A",priority:100,link:i}}}}),r(Li,function(t,e){Go[e]=function(){return{priority:100,link:function(t,n,r){if("ngPattern"===e&&"/"===r.ngPattern.charAt(0)){var i=r.ngPattern.match(Hr);if(i)return void r.$set("ngPattern",new RegExp(i[1],i[2]))}t.$watch(r[e],function(t){r.$set(e,t)})}}}}),r(["src","srcset","href"],function(t){var e=xe("ng-"+t);Go[e]=function(){return{priority:99,link:function(n,r,i){var o=t,a=t;"href"===t&&"[object SVGAnimatedString]"===ri.call(r.prop("href"))&&(a="xlinkHref",i.$attr[a]="xlink:href",o=null),i.$observe(e,function(e){return e?(i.$set(a,e),void(Kr&&o&&r.prop(o,i[a]))):void("href"===t&&i.$set(a,null))})}}}});var Xo={$addControl:h,$$renameControl:dr,$removeControl:h,$setValidity:h,$setDirty:h,$setPristine:h,$setSubmitted:h},Ko="ng-pending",Jo="ng-submitted";vr.$inject=["$element","$attrs","$scope","$animate","$interpolate"],vr.prototype={$rollbackViewValue:function(){r(this.$$controls,function(t){t.$rollbackViewValue()})},$commitViewValue:function(){r(this.$$controls,function(t){t.$commitViewValue()})},$addControl:function(t){ht(t.$name,"input"),this.$$controls.push(t),t.$name&&(this[t.$name]=t),t.$$parentForm=this},$$renameControl:function(t,e){var n=t.$name;this[n]===t&&delete this[n],this[e]=t,t.$name=e},$removeControl:function(t){t.$name&&this[t.$name]===t&&delete this[t.$name],r(this.$pending,function(e,n){this.$setValidity(n,null,t)},this),r(this.$error,function(e,n){this.$setValidity(n,null,t)},this),r(this.$$success,function(e,n){this.$setValidity(n,null,t)},this),q(this.$$controls,t),t.$$parentForm=Xo},$setDirty:function(){this.$$animate.removeClass(this.$$element,Pa),this.$$animate.addClass(this.$$element,Ra),this.$dirty=!0,this.$pristine=!1,this.$$parentForm.$setDirty()},$setPristine:function(){this.$$animate.setClass(this.$$element,Pa,Ra+" "+Jo),this.$dirty=!1,this.$pristine=!0,this.$submitted=!1,r(this.$$controls,function(t){t.$setPristine()})},$setUntouched:function(){r(this.$$controls,function(t){t.$setUntouched()})},$setSubmitted:function(){this.$$animate.addClass(this.$$element,Jo),this.$submitted=!0,this.$$parentForm.$setSubmitted()}},mr({clazz:vr,set:function(t,e,n){var r=t[e];if(r){var i=r.indexOf(n);i===-1&&r.push(n)}else t[e]=[n]},unset:function(t,e,n){var r=t[e];r&&(q(r,n),0===r.length&&delete t[e])}});var Zo=function(t){return["$timeout","$parse",function(e,n){function r(t){return""===t?n('this[""]').assign:n(t).assign||h}var i={name:"form",restrict:t?"EAC":"E",require:["form","^^?form"],controller:vr,compile:function(n,i){n.addClass(Pa).addClass(Oa);var o=i.name?"name":!(!t||!i.ngForm)&&"ngForm";return{pre:function(t,n,i,a){var s=a[0];if(!("action"in i)){var u=function(e){t.$apply(function(){s.$commitViewValue(),s.$setSubmitted()}),e.preventDefault()};n[0].addEventListener("submit",u),n.on("$destroy",function(){e(function(){n[0].removeEventListener("submit",u)},0,!1)})}var c=a[1]||s.$$parentForm;c.$addControl(s);var f=o?r(s.$name):h;o&&(f(t,s),i.$observe(o,function(e){s.$name!==e&&(f(t,void 0),s.$$parentForm.$$renameControl(s,e),(f=r(s.$name))(t,s))})),n.on("$destroy",function(){s.$$parentForm.$removeControl(s),f(t,void 0),l(s,Xo)})}}}};return i}]},Qo=Zo(),ta=Zo(!0),ea=/^\d{4,}-[01]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d\.\d+(?:[+-][0-2]\d:[0-5]\d|Z)$/,na=/^[a-z][a-z\d.+-]*:\/*(?:[^:@]+(?::[^@]+)?@)?(?:[^\s:\/?#]+|\[[a-f\d:]+])(?::\d+)?(?:\/[^?#]*)?(?:\?[^#]*)?(?:#.*)?$/i,ra=/^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+\/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+\/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/,ia=/^\s*(-|\+)?(\d+|(\d*(\.\d*)))([eE][+-]?\d+)?\s*$/,oa=/^(\d{4,})-(\d{2})-(\d{2})$/,aa=/^(\d{4,})-(\d\d)-(\d\d)T(\d\d):(\d\d)(?::(\d\d)(\.\d{1,3})?)?$/,sa=/^(\d{4,})-W(\d\d)$/,ua=/^(\d{4,})-(\d\d)$/,la=/^(\d\d):(\d\d)(?::(\d\d)(\.\d{1,3})?)?$/,ca="keydown wheel mousedown",fa=gt();r("date,datetime-local,month,time,week".split(","),function(t){fa[t]=!0});var pa={text:$r,date:Cr("date",oa,_r(oa,["yyyy","MM","dd"]),"yyyy-MM-dd"),"datetime-local":Cr("datetimelocal",aa,_r(aa,["yyyy","MM","dd","HH","mm","ss","sss"]),"yyyy-MM-ddTHH:mm:ss.sss"),time:Cr("time",la,_r(la,["HH","mm","ss","sss"]),"HH:mm:ss.sss"),week:Cr("week",sa,xr,"yyyy-Www"),month:Cr("month",ua,_r(ua,["yyyy","MM"]),"yyyy-MM"),number:Ar,url:Nr,email:Pr,radio:Rr,range:Or,checkbox:jr,hidden:h,button:h,submit:h,reset:h,file:h},ha=["$browser","$sniffer","$filter","$parse",function(t,e,n,r){return{restrict:"E",require:["?ngModel"],link:{pre:function(i,o,a,s){s[0]&&(pa[Wr(a.type)]||pa.text)(i,o,a,s[0],e,t,n,r)}}}}],da=/^(true|false|\d+)$/,va=function(){function t(t,e,n){t.prop("value",n),e.$set("value",n)}return{restrict:"A",priority:100,compile:function(e,n){return da.test(n.ngValue)?function(e,n,r){var i=e.$eval(r.ngValue);t(n,r,i)}:function(e,n,r){e.$watch(r.ngValue,function(e){t(n,r,e)})}}}},ga=["$compile",function(t){return{restrict:"AC",compile:function(e){return t.$$addBindingClass(e),function(e,n,r){t.$$addBindingInfo(n,r.ngBind),n=n[0],e.$watch(r.ngBind,function(t){n.textContent=mt(t)})}}}}],ma=["$interpolate","$compile",function(t,e){return{compile:function(n){return e.$$addBindingClass(n),function(n,r,i){var o=t(r.attr(i.$attr.ngBindTemplate));e.$$addBindingInfo(r,o.expressions),r=r[0],i.$observe("ngBindTemplate",function(t){r.textContent=m(t)?"":t})}}}}],ya=["$sce","$parse","$compile",function(t,e,n){return{restrict:"A",compile:function(r,i){var o=e(i.ngBindHtml),a=e(i.ngBindHtml,function(e){return t.valueOf(e)});return n.$$addBindingClass(r),function(e,r,i){n.$$addBindingInfo(r,i.ngBindHtml),e.$watch(a,function(){var n=o(e);r.html(t.getTrustedHtml(n)||"")})}}}}],ba=v({restrict:"A",require:"ngModel",link:function(t,e,n,r){r.$viewChangeListeners.push(function(){t.$eval(n.ngChange)})}}),$a=Fr("",!0),wa=Fr("Odd",0),xa=Fr("Even",1),_a=hr({compile:function(t,e){e.$set("ngCloak",void 0),t.removeClass("ng-cloak")}}),Ca=[function(){return{restrict:"A",scope:!0,controller:"@",priority:500}}],ka={},Sa={blur:!0,focus:!0};r("click dblclick mousedown mouseup mouseover mouseout mousemove mouseenter mouseleave keydown keyup keypress submit focus blur copy cut paste".split(" "),function(t){var e=xe("ng-"+t);ka[e]=["$parse","$rootScope",function(n,r){return{restrict:"A",compile:function(i,o){var a=n(o[e],null,!0);return function(e,n){n.on(t,function(n){var i=function(){a(e,{$event:n})};Sa[t]&&r.$$phase?e.$evalAsync(i):e.$apply(i)})}}}}]});var Ta=["$animate","$compile",function(t,e){return{multiElement:!0,transclude:"element",priority:600,terminal:!0,restrict:"A",$$tlb:!0,link:function(n,r,i,o,a){var s,u,l;n.$watch(i.ngIf,function(n){n?u||a(function(n,o){u=o,n[n.length++]=e.$$createComment("end ngIf",i.ngIf),s={clone:n},t.enter(n,r.parent(),r)}):(l&&(l.remove(),l=null),u&&(u.$destroy(),u=null),s&&(l=vt(s.clone),t.leave(l).done(function(t){t!==!1&&(l=null)}),s=null))})}}}],Ea=["$templateRequest","$anchorScroll","$animate",function(t,e,n){return{restrict:"ECA",priority:400,terminal:!0,transclude:"element",controller:ai.noop,compile:function(r,i){var o=i.ngInclude||i.src,a=i.onload||"",s=i.autoscroll;return function(r,i,u,l,c){var f,p,h,d=0,v=function(){p&&(p.remove(),p=null),f&&(f.$destroy(),f=null),h&&(n.leave(h).done(function(t){t!==!1&&(p=null)}),p=h,h=null)};r.$watch(o,function(o){var u=function(t){t===!1||!y(s)||s&&!r.$eval(s)||e()},p=++d;o?(t(o,!0).then(function(t){if(!r.$$destroyed&&p===d){var e=r.$new();l.template=t;var s=c(e,function(t){v(),n.enter(t,null,i).done(u)});f=e,h=s,f.$emit("$includeContentLoaded",o),r.$eval(a)}},function(){r.$$destroyed||p===d&&(v(),r.$emit("$includeContentError",o))}),r.$emit("$includeContentRequested",o)):(v(),l.template=null)})}}}}],Ma=["$compile",function(e){return{restrict:"ECA",priority:-400,require:"ngInclude",link:function(n,r,i,o){return ri.call(r[0]).match(/SVG/)?(r.empty(),void e(At(o.template,t.document).childNodes)(n,function(t){r.append(t)},{futureParentElement:r})):(r.html(o.template),void e(r.contents())(n))}}}],Da=hr({priority:450,compile:function(){return{pre:function(t,e,n){t.$eval(n.ngInit)}}}}),Aa=function(){return{restrict:"A",priority:100,require:"ngModel",link:function(t,e,n,i){var o=n.ngList||", ",a="false"!==n.ngTrim,s=a?fi(o):o,u=function(t){if(!m(t)){var e=[];return t&&r(t.split(s),function(t){t&&e.push(a?fi(t):t)}),e}};i.$parsers.push(u),i.$formatters.push(function(t){if(li(t))return t.join(o)}),i.$isEmpty=function(t){return!t||!t.length}}}},Oa="ng-valid",Na="ng-invalid",Pa="ng-pristine",Ra="ng-dirty",Ia="ng-untouched",ja="ng-touched",Fa="ng-empty",qa="ng-not-empty",Va=e("ngModel");qr.$inject=["$scope","$exceptionHandler","$attrs","$element","$parse","$animate","$timeout","$q","$interpolate"],qr.prototype={$$initGetterSetters:function(){if(this.$options.getOption("getterSetter")){var t=this.$$parse(this.$$attr.ngModel+"()"),e=this.$$parse(this.$$attr.ngModel+"($$$p)");this.$$ngModelGet=function(e){var n=this.$$parsedNgModel(e);return C(n)&&(n=t(e)),n},this.$$ngModelSet=function(t,n){C(this.$$parsedNgModel(t))?e(t,{$$$p:n}):this.$$parsedNgModelAssign(t,n)}}else if(!this.$$parsedNgModel.assign)throw Va("nonassign","Expression '{0}' is non-assignable. Element: {1}",this.$$attr.ngModel,J(this.$$element))},$render:h,$isEmpty:function(t){return m(t)||""===t||null===t||t!==t},$$updateEmptyClasses:function(t){this.$isEmpty(t)?(this.$$animate.removeClass(this.$$element,qa),this.$$animate.addClass(this.$$element,Fa)):(this.$$animate.removeClass(this.$$element,Fa),this.$$animate.addClass(this.$$element,qa))},$setPristine:function(){this.$dirty=!1,this.$pristine=!0,this.$$animate.removeClass(this.$$element,Ra),this.$$animate.addClass(this.$$element,Pa)},$setDirty:function(){this.$dirty=!0,this.$pristine=!1,this.$$animate.removeClass(this.$$element,Pa),this.$$animate.addClass(this.$$element,Ra),this.$$parentForm.$setDirty()},$setUntouched:function(){this.$touched=!1,this.$untouched=!0,this.$$animate.setClass(this.$$element,Ia,ja)},$setTouched:function(){this.$touched=!0,this.$untouched=!1,this.$$animate.setClass(this.$$element,ja,Ia)},$rollbackViewValue:function(){this.$$timeout.cancel(this.$$pendingDebounce),this.$viewValue=this.$$lastCommittedViewValue,this.$render()},$validate:function(){if(!ui(this.$modelValue)){var t=this.$$lastCommittedViewValue,e=this.$$rawModelValue,n=this.$valid,r=this.$modelValue,i=this.$options.getOption("allowInvalid"),o=this;this.$$runValidators(e,t,function(t){i||n===t||(o.$modelValue=t?e:void 0,o.$modelValue!==r&&o.$$writeModelToScope())})}},$$runValidators:function(t,e,n){function i(){var t=c.$$parserName||"parse";return m(c.$$parserValid)?(s(t,null),!0):(c.$$parserValid||(r(c.$validators,function(t,e){s(e,null)}),r(c.$asyncValidators,function(t,e){s(e,null)})),s(t,c.$$parserValid),c.$$parserValid)}function o(){var n=!0;return r(c.$validators,function(r,i){var o=Boolean(r(t,e));n=n&&o,s(i,o)}),!!n||(r(c.$asyncValidators,function(t,e){s(e,null)}),!1)}function a(){var n=[],i=!0;r(c.$asyncValidators,function(r,o){var a=r(t,e);if(!O(a))throw Va("nopromise","Expected asynchronous validator to return a promise but got '{0}' instead.",a);s(o,void 0),n.push(a.then(function(){s(o,!0)},function(){i=!1,s(o,!1)}))}),n.length?c.$$q.all(n).then(function(){u(i)},h):u(!0)}function s(t,e){l===c.$$currentValidationRunId&&c.$setValidity(t,e)}function u(t){l===c.$$currentValidationRunId&&n(t)}this.$$currentValidationRunId++;var l=this.$$currentValidationRunId,c=this;return i()&&o()?void a():void u(!1)},$commitViewValue:function(){var t=this.$viewValue;this.$$timeout.cancel(this.$$pendingDebounce),(this.$$lastCommittedViewValue!==t||""===t&&this.$$hasNativeValidators)&&(this.$$updateEmptyClasses(t),this.$$lastCommittedViewValue=t,this.$pristine&&this.$setDirty(),this.$$parseAndValidate())},$$parseAndValidate:function(){function t(){r.$modelValue!==o&&r.$$writeModelToScope()}var e=this.$$lastCommittedViewValue,n=e,r=this;if(this.$$parserValid=!m(n)||void 0,this.$$parserValid)for(var i=0;i<this.$parsers.length;i++)if(n=this.$parsers[i](n),m(n)){this.$$parserValid=!1;break}ui(this.$modelValue)&&(this.$modelValue=this.$$ngModelGet(this.$$scope));var o=this.$modelValue,a=this.$options.getOption("allowInvalid");this.$$rawModelValue=n,a&&(this.$modelValue=n,t()),this.$$runValidators(n,this.$$lastCommittedViewValue,function(e){a||(r.$modelValue=e?n:void 0,t())})},$$writeModelToScope:function(){this.$$ngModelSet(this.$$scope,this.$modelValue),r(this.$viewChangeListeners,function(t){try{t()}catch(t){this.$$exceptionHandler(t)}},this)},$setViewValue:function(t,e){this.$viewValue=t,this.$options.getOption("updateOnDefault")&&this.$$debounceViewValueCommit(e)},$$debounceViewValueCommit:function(t){var e=this.$options.getOption("debounce");x(e[t])?e=e[t]:x(e.default)&&(e=e.default),this.$$timeout.cancel(this.$$pendingDebounce);var n=this;e>0?this.$$pendingDebounce=this.$$timeout(function(){n.$commitViewValue()},e):this.$$scope.$root.$$phase?this.$commitViewValue():this.$$scope.$apply(function(){n.$commitViewValue()})}},mr({clazz:qr,set:function(t,e){t[e]=!0},unset:function(t,e){delete t[e]}});var Ua,La=["$rootScope",function(t){return{restrict:"A",require:["ngModel","^?form","^?ngModelOptions"],controller:qr,priority:1,compile:function(e){return e.addClass(Pa).addClass(Ia).addClass(Oa),{pre:function(t,e,n,r){var i=r[0],o=r[1]||i.$$parentForm,a=r[2];a&&(i.$options=a.$options),i.$$initGetterSetters(),o.$addControl(i),n.$observe("name",function(t){i.$name!==t&&i.$$parentForm.$$renameControl(i,t)}),t.$on("$destroy",function(){i.$$parentForm.$removeControl(i)})},post:function(e,n,r,i){function o(){a.$setTouched()}var a=i[0];a.$options.getOption("updateOn")&&n.on(a.$options.getOption("updateOn"),function(t){a.$$debounceViewValueCommit(t&&t.type)}),n.on("blur",function(){a.$touched||(t.$$phase?e.$evalAsync(o):e.$apply(o))})}}}}}],Ha=/(\s+|^)default(\s+|$)/;Ur.prototype={getOption:function(t){return this.$$options[t]},createChild:function(t){var e=!1;return t=l({},t),r(t,function(n,r){"$inherit"===n?"*"===r?e=!0:(t[r]=this.$$options[r],"updateOn"===r&&(t.updateOnDefault=this.$$options.updateOnDefault)):"updateOn"===r&&(t.updateOnDefault=!1,t[r]=fi(n.replace(Ha,function(){return t.updateOnDefault=!0," "})))},this),e&&(delete t["*"],Lr(t,this.$$options)),Lr(t,Ua.$$options),new Ur(t)}},Ua=new Ur({updateOn:"",updateOnDefault:!0,debounce:0,getterSetter:!1,allowInvalid:!1,timezone:null});var Ba=function(){function t(t,e){this.$$attrs=t,this.$$scope=e}return t.$inject=["$attrs","$scope"],t.prototype={$onInit:function(){var t=this.parentCtrl?this.parentCtrl.$options:Ua,e=this.$$scope.$eval(this.$$attrs.ngModelOptions);this.$options=t.createChild(e)}},{restrict:"A",priority:10,require:{parentCtrl:"?^^ngModelOptions"},bindToController:!0,controller:t}},za=hr({terminal:!0,priority:1e3}),Wa=e("ngOptions"),Ya=/^\s*([\s\S]+?)(?:\s+as\s+([\s\S]+?))?(?:\s+group\s+by\s+([\s\S]+?))?(?:\s+disable\s+when\s+([\s\S]+?))?\s+for\s+(?:([$\w][$\w]*)|(?:\(\s*([$\w][$\w]*)\s*,\s*([$\w][$\w]*)\s*\)))\s+in\s+([\s\S]+?)(?:\s+track\s+by\s+([\s\S]+?))?$/,Ga=["$compile","$document","$parse",function(e,i,o){function a(t,e,r){function i(t,e,n,r,i){this.selectValue=t,this.viewValue=e,this.label=n,this.group=r,this.disabled=i}function a(t){var e;if(!l&&n(t))e=t;else{e=[];for(var r in t)t.hasOwnProperty(r)&&"$"!==r.charAt(0)&&e.push(r)}return e}var s=t.match(Ya);if(!s)throw Wa("iexp","Expected expression in form of '_select_ (as _label_)? for (_key_,)?_value_ in _collection_' but got '{0}'. Element: {1}",t,J(e));var u=s[5]||s[7],l=s[6],c=/ as /.test(s[0])&&s[1],f=s[9],p=o(s[2]?s[1]:u),h=c&&o(c),d=h||p,v=f&&o(f),g=f?function(t,e){return v(r,e)}:function(t){return re(t)},m=function(t,e){return g(t,_(t,e))},y=o(s[2]||s[1]),b=o(s[3]||""),$=o(s[4]||""),w=o(s[8]),x={},_=l?function(t,e){return x[l]=e,x[u]=t,x}:function(t){return x[u]=t,x};return{trackBy:f,getTrackByValue:m,getWatchables:o(w,function(t){var e=[];t=t||[];for(var n=a(t),i=n.length,o=0;o<i;o++){var u=t===n?o:n[o],l=t[u],c=_(l,u),f=g(l,c);if(e.push(f),s[2]||s[1]){var p=y(r,c);e.push(p)}if(s[4]){var h=$(r,c);e.push(h)}}return e}),getOptions:function(){for(var t=[],e={},n=w(r)||[],o=a(n),s=o.length,u=0;u<s;u++){var l=n===o?u:o[u],c=n[l],p=_(c,l),h=d(r,p),v=g(h,p),x=y(r,p),C=b(r,p),k=$(r,p),S=new i(v,h,x,C,k);t.push(S),e[v]=S}return{items:t,selectValueMap:e,getOptionFromViewValue:function(t){return e[m(t)]},getViewValueFromOption:function(t){return f?V(t.viewValue):t.viewValue}}}}}function s(t,n,o,s){function c(t,e){var n=u.cloneNode(!1);e.appendChild(n),p(t,n)}function f(t){var e=_.getOptionFromViewValue(t),n=e&&e.element;return n&&!n.selected&&(n.selected=!0),e}function p(t,e){t.element=e,e.disabled=t.disabled,t.label!==e.label&&(e.label=t.label,e.textContent=t.label),e.value=t.selectValue}function h(){var t=_&&d.readValue();if(_)for(var e=_.items.length-1;e>=0;e--){var r=_.items[e];Gt(y(r.group)?r.element.parentNode:r.element)}_=C.getOptions();var i={};if(w&&n.prepend(d.emptyOption),_.items.forEach(function(t){var e;y(t.group)?(e=i[t.group],e||(e=l.cloneNode(!1),k.appendChild(e),e.label=null===t.group?"null":t.group,i[t.group]=e),c(t,e)):c(t,k)}),n[0].appendChild(k),v.$render(),!v.$isEmpty(t)){var o=d.readValue(),a=C.trackBy||g;(a?U(t,o):t===o)||(v.$setViewValue(o),v.$render())}}for(var d=s[0],v=s[1],g=o.multiple,m=0,b=n.children(),$=b.length;m<$;m++)if(""===b[m].value){d.hasEmptyOption=!0,d.emptyOption=b.eq(m);break}var w=!!d.emptyOption,x=Jr(u.cloneNode(!1));x.val("?");var _,C=a(o.ngOptions,n,t),k=i[0].createDocumentFragment();d.generateUnknownOptionValue=function(t){return"?"},g?(d.writeValue=function(t){var e=t&&t.map(f)||[];_.items.forEach(function(t){t.element.selected&&!F(e,t)&&(t.element.selected=!1)})},d.readValue=function(){var t=n.val()||[],e=[];return r(t,function(t){var n=_.selectValueMap[t];n&&!n.disabled&&e.push(_.getViewValueFromOption(n))}),e},C.trackBy&&t.$watchCollection(function(){if(li(v.$viewValue))return v.$viewValue.map(function(t){return C.getTrackByValue(t);
})},function(){v.$render()})):(d.writeValue=function(t){var e=_.selectValueMap[n.val()],r=_.getOptionFromViewValue(t);e&&e.element.removeAttribute("selected"),r?(n[0].value!==r.selectValue&&(d.removeUnknownOption(),d.unselectEmptyOption(),n[0].value=r.selectValue,r.element.selected=!0),r.element.setAttribute("selected","selected")):w?d.selectEmptyOption():d.unknownOption.parent().length?d.updateUnknownOption(t):d.renderUnknownOption(t)},d.readValue=function(){var t=_.selectValueMap[n.val()];return t&&!t.disabled?(d.unselectEmptyOption(),d.removeUnknownOption(),_.getViewValueFromOption(t)):null},C.trackBy&&t.$watch(function(){return C.getTrackByValue(v.$viewValue)},function(){v.$render()})),w&&(d.emptyOption.remove(),e(d.emptyOption)(t),d.emptyOption[0].nodeType===_i?(d.hasEmptyOption=!1,d.registerOption=function(t,e){""===e.val()&&(d.hasEmptyOption=!0,d.emptyOption=e,d.emptyOption.removeClass("ng-scope"),v.$render(),e.on("$destroy",function(){d.hasEmptyOption=!1,d.emptyOption=void 0}))}):d.emptyOption.removeClass("ng-scope")),n.empty(),h(),t.$watchCollection(C.getWatchables,h)}var u=t.document.createElement("option"),l=t.document.createElement("optgroup");return{restrict:"A",terminal:!0,require:["select","ngModel"],link:{pre:function(t,e,n,r){r[0].registerOption=h},post:s}}}],Xa=["$locale","$interpolate","$log",function(t,e,n){var i=/{}/g,o=/^when(Minus)?(.+)$/;return{link:function(a,s,u){function l(t){s.text(t||"")}var c,f=u.count,p=u.$attr.when&&s.attr(u.$attr.when),d=u.offset||0,v=a.$eval(p)||{},g={},y=e.startSymbol(),b=e.endSymbol(),$=y+f+"-"+d+b,w=ai.noop;r(u,function(t,e){var n=o.exec(e);if(n){var r=(n[1]?"-":"")+Wr(n[2]);v[r]=s.attr(u.$attr[e])}}),r(v,function(t,n){g[n]=e(t.replace(i,$))}),a.$watch(f,function(e){var r=parseFloat(e),i=ui(r);if(i||r in v||(r=t.pluralCat(r-d)),!(r===c||i&&ui(c))){w();var o=g[r];m(o)?(null!=e&&n.debug("ngPluralize: no rule defined for '"+r+"' in "+p),w=h,l()):w=a.$watch(o,l),c=r}})}}}],Ka=["$parse","$animate","$compile",function(t,i,o){var a="$$NG_REMOVED",s=e("ngRepeat"),u=function(t,e,n,r,i,o,a){t[n]=r,i&&(t[i]=o),t.$index=e,t.$first=0===e,t.$last=e===a-1,t.$middle=!(t.$first||t.$last),t.$odd=!(t.$even=0===(1&e))},l=function(t){return t.clone[0]},c=function(t){return t.clone[t.clone.length-1]};return{restrict:"A",multiElement:!0,transclude:"element",priority:1e3,terminal:!0,$$tlb:!0,compile:function(e,f){var p=f.ngRepeat,h=o.$$createComment("end ngRepeat",p),d=p.match(/^\s*([\s\S]+?)\s+in\s+([\s\S]+?)(?:\s+as\s+([\s\S]+?))?(?:\s+track\s+by\s+([\s\S]+?))?\s*$/);if(!d)throw s("iexp","Expected expression in form of '_item_ in _collection_[ track by _id_]' but got '{0}'.",p);var v=d[1],g=d[2],m=d[3],y=d[4];if(d=v.match(/^(?:(\s*[$\w]+)|\(\s*([$\w]+)\s*,\s*([$\w]+)\s*\))$/),!d)throw s("iidexp","'_item_' in '_item_ in _collection_' should be an identifier or '(_key_, _value_)' expression, but got '{0}'.",v);var b=d[3]||d[1],$=d[2];if(m&&(!/^[$a-zA-Z_][$a-zA-Z0-9_]*$/.test(m)||/^(null|undefined|this|\$index|\$first|\$middle|\$last|\$even|\$odd|\$parent|\$root|\$id)$/.test(m)))throw s("badident","alias '{0}' is invalid --- must be a valid JS identifier which is not a reserved name.",m);var w,x,_,C,k={$id:re};return y?w=t(y):(_=function(t,e){return re(e)},C=function(t){return t}),function(t,e,o,f,d){w&&(x=function(e,n,r){return $&&(k[$]=e),k[b]=n,k.$index=r,w(t,k)});var v=gt();t.$watchCollection(g,function(o){var f,g,y,w,k,S,T,E,M,D,A,O,N=e[0],P=gt();if(m&&(t[m]=o),n(o))M=o,E=x||_;else{E=x||C,M=[];for(var R in o)zr.call(o,R)&&"$"!==R.charAt(0)&&M.push(R)}for(w=M.length,A=new Array(w),f=0;f<w;f++)if(k=o===M?f:M[f],S=o[k],T=E(k,S,f),v[T])D=v[T],delete v[T],P[T]=D,A[f]=D;else{if(P[T])throw r(A,function(t){t&&t.scope&&(v[t.id]=t)}),s("dupes","Duplicates in a repeater are not allowed. Use 'track by' expression to specify unique keys. Repeater: {0}, Duplicate key: {1}, Duplicate value: {2}",p,T,S);A[f]={id:T,scope:void 0,clone:void 0},P[T]=!0}for(var I in v){if(D=v[I],O=vt(D.clone),i.leave(O),O[0].parentNode)for(f=0,g=O.length;f<g;f++)O[f][a]=!0;D.scope.$destroy()}for(f=0;f<w;f++)if(k=o===M?f:M[f],S=o[k],D=A[f],D.scope){y=N;do y=y.nextSibling;while(y&&y[a]);l(D)!==y&&i.move(vt(D.clone),null,N),N=c(D),u(D.scope,f,b,S,$,k,w)}else d(function(t,e){D.scope=e;var n=h.cloneNode(!1);t[t.length++]=n,i.enter(t,null,N),N=n,D.clone=t,P[D.id]=D,u(D.scope,f,b,S,$,k,w)});v=P})}}}}],Ja="ng-hide",Za="ng-hide-animate",Qa=["$animate",function(t){return{restrict:"A",multiElement:!0,link:function(e,n,r){e.$watch(r.ngShow,function(e){t[e?"removeClass":"addClass"](n,Ja,{tempClasses:Za})})}}}],ts=["$animate",function(t){return{restrict:"A",multiElement:!0,link:function(e,n,r){e.$watch(r.ngHide,function(e){t[e?"addClass":"removeClass"](n,Ja,{tempClasses:Za})})}}}],es=hr(function(t,e,n){t.$watch(n.ngStyle,function(t,n){n&&t!==n&&r(n,function(t,n){e.css(n,"")}),t&&e.css(t)},!0)}),ns=["$animate","$compile",function(t,e){return{require:"ngSwitch",controller:["$scope",function(){this.cases={}}],link:function(n,i,o,a){var s=o.ngSwitch||o.on,u=[],l=[],c=[],f=[],p=function(t,e){return function(n){n!==!1&&t.splice(e,1)}};n.$watch(s,function(n){for(var i,o;c.length;)t.cancel(c.pop());for(i=0,o=f.length;i<o;++i){var s=vt(l[i].clone);f[i].$destroy();var h=c[i]=t.leave(s);h.done(p(c,i))}l.length=0,f.length=0,(u=a.cases["!"+n]||a.cases["?"])&&r(u,function(n){n.transclude(function(r,i){f.push(i);var o=n.element;r[r.length++]=e.$$createComment("end ngSwitchWhen");var a={clone:r};l.push(a),t.enter(r,o.parent(),o)})})})}}}],rs=hr({transclude:"element",priority:1200,require:"^ngSwitch",multiElement:!0,link:function(t,e,n,i,o){var a=n.ngSwitchWhen.split(n.ngSwitchWhenSeparator).sort().filter(function(t,e,n){return n[e-1]!==t});r(a,function(t){i.cases["!"+t]=i.cases["!"+t]||[],i.cases["!"+t].push({transclude:o,element:e})})}}),is=hr({transclude:"element",priority:1200,require:"^ngSwitch",multiElement:!0,link:function(t,e,n,r,i){r.cases["?"]=r.cases["?"]||[],r.cases["?"].push({transclude:i,element:e})}}),os=e("ngTransclude"),as=["$compile",function(t){return{restrict:"EAC",terminal:!0,compile:function(e){var n=t(e.contents());return e.empty(),function(t,e,r,i,o){function a(t,n){t.length&&u(t)?e.append(t):(s(),n.$destroy())}function s(){n(t,function(t){e.append(t)})}function u(t){for(var e=0,n=t.length;e<n;e++){var r=t[e];if(r.nodeType!==xi||r.nodeValue.trim())return!0}}if(!o)throw os("orphan","Illegal use of ngTransclude directive in the template! No parent directive that requires a transclusion found. Element: {0}",J(e));r.ngTransclude===r.$attr.ngTransclude&&(r.ngTransclude="");var l=r.ngTransclude||r.ngTranscludeSlot;o(a,null,l),l&&!o.isSlotFilled(l)&&s()}}}}],ss=["$templateCache",function(t){return{restrict:"E",terminal:!0,compile:function(e,n){if("text/ng-template"===n.type){var r=n.id,i=e[0].text;t.put(r,i)}}}}],us={$setViewValue:h,$render:h},ls=["$element","$scope",function(e,n){function r(){u||(u=!0,n.$$postDigest(function(){u=!1,a.ngModelCtrl.$render()}))}function i(t){l||(l=!0,n.$$postDigest(function(){n.$$destroyed||(l=!1,a.ngModelCtrl.$setViewValue(a.readValue()),t&&a.ngModelCtrl.$render())}))}function o(t){t.prop("selected",!0),t.attr("selected",!0)}var a=this,s=new ie;a.selectValueMap={},a.ngModelCtrl=us,a.multiple=!1,a.unknownOption=Jr(t.document.createElement("option")),a.hasEmptyOption=!1,a.emptyOption=void 0,a.renderUnknownOption=function(t){var n=a.generateUnknownOptionValue(t);a.unknownOption.val(n),e.prepend(a.unknownOption),o(a.unknownOption),e.val(n)},a.updateUnknownOption=function(t){var n=a.generateUnknownOptionValue(t);a.unknownOption.val(n),o(a.unknownOption),e.val(n)},a.generateUnknownOptionValue=function(t){return"? "+re(t)+" ?"},a.removeUnknownOption=function(){a.unknownOption.parent()&&a.unknownOption.remove()},a.selectEmptyOption=function(){a.emptyOption&&(e.val(""),o(a.emptyOption))},a.unselectEmptyOption=function(){a.hasEmptyOption&&a.emptyOption.removeAttr("selected")},n.$on("$destroy",function(){a.renderUnknownOption=h}),a.readValue=function(){var t=e.val(),n=t in a.selectValueMap?a.selectValueMap[t]:t;return a.hasOption(n)?n:null},a.writeValue=function(t){var n=e[0].options[e[0].selectedIndex];if(n&&n.removeAttribute("selected"),a.hasOption(t)){a.removeUnknownOption();var r=re(t);e.val(r in a.selectValueMap?r:t);var i=e[0].options[e[0].selectedIndex];o(Jr(i))}else null==t&&a.emptyOption?(a.removeUnknownOption(),a.selectEmptyOption()):a.unknownOption.parent().length?a.updateUnknownOption(t):a.renderUnknownOption(t)},a.addOption=function(t,e){if(e[0].nodeType!==_i){ht(t,'"option value"'),""===t&&(a.hasEmptyOption=!0,a.emptyOption=e);var n=s.get(t)||0;s.put(t,n+1),r()}},a.removeOption=function(t){var e=s.get(t);e&&(1===e?(s.remove(t),""===t&&(a.hasEmptyOption=!1,a.emptyOption=void 0)):s.put(t,e-1))},a.hasOption=function(t){return!!s.get(t)};var u=!1,l=!1;a.registerOption=function(t,e,n,r,o){if(n.$attr.ngValue){var s,u=NaN;n.$observe("value",function(t){var n,r=e.prop("selected");y(u)&&(a.removeOption(s),delete a.selectValueMap[u],n=!0),u=re(t),s=t,a.selectValueMap[u]=t,a.addOption(t,e),e.attr("value",u),n&&r&&i()})}else r?n.$observe("value",function(t){a.readValue();var n,r=e.prop("selected");y(s)&&(a.removeOption(s),n=!0),s=t,a.addOption(t,e),n&&r&&i()}):o?t.$watch(o,function(t,r){n.$set("value",t);var o=e.prop("selected");r!==t&&a.removeOption(r),a.addOption(t,e),r&&o&&i()}):a.addOption(n.value,e);n.$observe("disabled",function(t){("true"===t||t&&e.prop("selected"))&&(a.multiple?i(!0):(a.ngModelCtrl.$setViewValue(null),a.ngModelCtrl.$render()))}),e.on("$destroy",function(){var t=a.readValue(),e=n.value;a.removeOption(e),a.ngModelCtrl.$render(),(a.multiple&&t&&t.indexOf(e)!==-1||t===e)&&i(!0)})}}],cs=function(){function t(t,e,n,i){var o=i[0],a=i[1];if(!a)return void(o.registerOption=h);if(o.ngModelCtrl=a,e.on("change",function(){o.removeUnknownOption(),t.$apply(function(){a.$setViewValue(o.readValue())})}),n.multiple){o.multiple=!0,o.readValue=function(){var t=[];return r(e.find("option"),function(e){if(e.selected&&!e.disabled){var n=e.value;t.push(n in o.selectValueMap?o.selectValueMap[n]:n)}}),t},o.writeValue=function(t){var n=new ie(t);r(e.find("option"),function(t){t.selected=y(n.get(t.value))||y(n.get(o.selectValueMap[t.value]))})};var s,u=NaN;t.$watch(function(){u!==a.$viewValue||U(s,a.$viewValue)||(s=bt(a.$viewValue),a.$render()),u=a.$viewValue}),a.$isEmpty=function(t){return!t||0===t.length}}}function e(t,e,n,r){var i=r[1];if(i){var o=r[0];i.$render=function(){o.writeValue(i.$viewValue)}}}return{restrict:"E",require:["select","?ngModel"],controller:ls,priority:1,link:{pre:t,post:e}}},fs=["$interpolate",function(t){return{restrict:"E",priority:100,compile:function(e,n){var r,i;return y(n.ngValue)||(y(n.value)?r=t(n.value,!0):(i=t(e.text(),!0),i||n.$set("value",e.text()))),function(t,e,n){var o="$selectController",a=e.parent(),s=a.data(o)||a.parent().data(o);s&&s.registerOption(t,e,n,r,i)}}}}],ps=function(){return{restrict:"A",require:"?ngModel",link:function(t,e,n,r){r&&(n.required=!0,r.$validators.required=function(t,e){return!n.required||!r.$isEmpty(e)},n.$observe("required",function(){r.$validate()}))}}},hs=function(){return{restrict:"A",require:"?ngModel",link:function(t,n,r,i){if(i){var o,a=r.ngPattern||r.pattern;r.$observe("pattern",function(t){if(w(t)&&t.length>0&&(t=new RegExp("^"+t+"$")),t&&!t.test)throw e("ngPattern")("noregexp","Expected {0} to be a RegExp but was {1}. Element: {2}",a,t,J(n));o=t||void 0,i.$validate()}),i.$validators.pattern=function(t,e){return i.$isEmpty(e)||m(o)||o.test(e)}}}}},ds=function(){return{restrict:"A",require:"?ngModel",link:function(t,e,n,r){if(r){var i=-1;n.$observe("maxlength",function(t){var e=f(t);i=ui(e)?-1:e,r.$validate()}),r.$validators.maxlength=function(t,e){return i<0||r.$isEmpty(e)||e.length<=i}}}}},vs=function(){return{restrict:"A",require:"?ngModel",link:function(t,e,n,r){if(r){var i=0;n.$observe("minlength",function(t){i=f(t)||0,r.$validate()}),r.$validators.minlength=function(t,e){return r.$isEmpty(e)||e.length>=i}}}}};return t.angular.bootstrap?void(t.console&&console.log("WARNING: Tried to load angular more than once.")):(ct(),xt(ai),ai.module("ngLocale",[],["$provide",function(t){function e(t){t+="";var e=t.indexOf(".");return e==-1?0:t.length-e-1}function n(t,n){var r=n;void 0===r&&(r=Math.min(e(t),3));var i=Math.pow(10,r),o=(t*i|0)%i;return{v:r,f:o}}var r={ZERO:"zero",ONE:"one",TWO:"two",FEW:"few",MANY:"many",OTHER:"other"};t.value("$locale",{DATETIME_FORMATS:{AMPMS:["AM","PM"],DAY:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],ERANAMES:["Before Christ","Anno Domini"],ERAS:["BC","AD"],FIRSTDAYOFWEEK:6,MONTH:["January","February","March","April","May","June","July","August","September","October","November","December"],SHORTDAY:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],SHORTMONTH:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],STANDALONEMONTH:["January","February","March","April","May","June","July","August","September","October","November","December"],WEEKENDRANGE:[5,6],fullDate:"EEEE, MMMM d, y",longDate:"MMMM d, y",medium:"MMM d, y h:mm:ss a",mediumDate:"MMM d, y",mediumTime:"h:mm:ss a",short:"M/d/yy h:mm a",shortDate:"M/d/yy",shortTime:"h:mm a"},NUMBER_FORMATS:{CURRENCY_SYM:"$",DECIMAL_SEP:".",GROUP_SEP:",",PATTERNS:[{gSize:3,lgSize:3,maxFrac:3,minFrac:0,minInt:1,negPre:"-",negSuf:"",posPre:"",posSuf:""},{gSize:3,lgSize:3,maxFrac:2,minFrac:2,minInt:1,negPre:"-¤",negSuf:"",posPre:"¤",posSuf:""}]},id:"en-us",localeID:"en_US",pluralCat:function(t,e){var i=0|t,o=n(t,e);return 1==i&&0==o.v?r.ONE:r.OTHER}})}]),void Jr(function(){ot(t.document,at)}))}(window),!window.angular.$$csp().noInlineStyle&&window.angular.element(document.head).prepend('<style type="text/css">@charset "UTF-8";[ng\\:cloak],[ng-cloak],[data-ng-cloak],[x-ng-cloak],.ng-cloak,.x-ng-cloak,.ng-hide:not(.ng-hide-animate){display:none !important;}ng\\:form{display:block;}.ng-animate-shim{visibility:hidden;}.ng-anchor{position:absolute;}</style>')},37:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.affix"),o="object"==typeof e&&e;i||r.data("bs.affix",i=new n(this,o)),"string"==typeof e&&i[e]()})}var n=function(e,r){this.options=t.extend({},n.DEFAULTS,r),this.$target=t(this.options.target).on("scroll.bs.affix.data-api",t.proxy(this.checkPosition,this)).on("click.bs.affix.data-api",t.proxy(this.checkPositionWithEventLoop,this)),this.$element=t(e),this.affixed=null,this.unpin=null,this.pinnedOffset=null,this.checkPosition()};n.VERSION="3.3.7",n.RESET="affix affix-top affix-bottom",n.DEFAULTS={offset:0,target:window},n.prototype.getState=function(t,e,n,r){var i=this.$target.scrollTop(),o=this.$element.offset(),a=this.$target.height();if(null!=n&&"top"==this.affixed)return i<n&&"top";if("bottom"==this.affixed)return null!=n?!(i+this.unpin<=o.top)&&"bottom":!(i+a<=t-r)&&"bottom";var s=null==this.affixed,u=s?i:o.top,l=s?a:e;return null!=n&&i<=n?"top":null!=r&&u+l>=t-r&&"bottom"},n.prototype.getPinnedOffset=function(){if(this.pinnedOffset)return this.pinnedOffset;this.$element.removeClass(n.RESET).addClass("affix");var t=this.$target.scrollTop(),e=this.$element.offset();return this.pinnedOffset=e.top-t},n.prototype.checkPositionWithEventLoop=function(){setTimeout(t.proxy(this.checkPosition,this),1)},n.prototype.checkPosition=function(){if(this.$element.is(":visible")){var e=this.$element.height(),r=this.options.offset,i=r.top,o=r.bottom,a=Math.max(t(document).height(),t(document.body).height());"object"!=typeof r&&(o=i=r),"function"==typeof i&&(i=r.top(this.$element)),"function"==typeof o&&(o=r.bottom(this.$element));var s=this.getState(a,e,i,o);if(this.affixed!=s){null!=this.unpin&&this.$element.css("top","");var u="affix"+(s?"-"+s:""),l=t.Event(u+".bs.affix");if(this.$element.trigger(l),l.isDefaultPrevented())return;this.affixed=s,this.unpin="bottom"==s?this.getPinnedOffset():null,this.$element.removeClass(n.RESET).addClass(u).trigger(u.replace("affix","affixed")+".bs.affix")}"bottom"==s&&this.$element.offset({top:a-e-o})}};var r=t.fn.affix;t.fn.affix=e,t.fn.affix.Constructor=n,t.fn.affix.noConflict=function(){return t.fn.affix=r,this},t(window).on("load",function(){t('[data-spy="affix"]').each(function(){var n=t(this),r=n.data();r.offset=r.offset||{},null!=r.offsetBottom&&(r.offset.bottom=r.offsetBottom),null!=r.offsetTop&&(r.offset.top=r.offsetTop),e.call(n,r)})})}(t)}).call(e,n(1))},38:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var n=t(this),i=n.data("bs.alert");i||n.data("bs.alert",i=new r(this)),"string"==typeof e&&i[e].call(n)})}var n='[data-dismiss="alert"]',r=function(e){t(e).on("click",n,this.close)};r.VERSION="3.3.7",r.TRANSITION_DURATION=150,r.prototype.close=function(e){function n(){a.detach().trigger("closed.bs.alert").remove()}var i=t(this),o=i.attr("data-target");o||(o=i.attr("href"),o=o&&o.replace(/.*(?=#[^\s]*$)/,""));var a=t("#"===o?[]:o);e&&e.preventDefault(),a.length||(a=i.closest(".alert")),a.trigger(e=t.Event("close.bs.alert")),e.isDefaultPrevented()||(a.removeClass("in"),t.support.transition&&a.hasClass("fade")?a.one("bsTransitionEnd",n).emulateTransitionEnd(r.TRANSITION_DURATION):n())};var i=t.fn.alert;t.fn.alert=e,t.fn.alert.Constructor=r,t.fn.alert.noConflict=function(){return t.fn.alert=i,this},t(document).on("click.bs.alert.data-api",n,r.prototype.close)}(t)}).call(e,n(1))},39:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.button"),o="object"==typeof e&&e;i||r.data("bs.button",i=new n(this,o)),"toggle"==e?i.toggle():e&&i.setState(e)})}var n=function(e,r){this.$element=t(e),this.options=t.extend({},n.DEFAULTS,r),this.isLoading=!1};n.VERSION="3.3.7",n.DEFAULTS={loadingText:"loading..."},n.prototype.setState=function(e){var n="disabled",r=this.$element,i=r.is("input")?"val":"html",o=r.data();e+="Text",null==o.resetText&&r.data("resetText",r[i]()),setTimeout(t.proxy(function(){r[i](null==o[e]?this.options[e]:o[e]),"loadingText"==e?(this.isLoading=!0,r.addClass(n).attr(n,n).prop(n,!0)):this.isLoading&&(this.isLoading=!1,r.removeClass(n).removeAttr(n).prop(n,!1))},this),0)},n.prototype.toggle=function(){var t=!0,e=this.$element.closest('[data-toggle="buttons"]');if(e.length){var n=this.$element.find("input");"radio"==n.prop("type")?(n.prop("checked")&&(t=!1),e.find(".active").removeClass("active"),this.$element.addClass("active")):"checkbox"==n.prop("type")&&(n.prop("checked")!==this.$element.hasClass("active")&&(t=!1),this.$element.toggleClass("active")),n.prop("checked",this.$element.hasClass("active")),t&&n.trigger("change")}else this.$element.attr("aria-pressed",!this.$element.hasClass("active")),this.$element.toggleClass("active")};var r=t.fn.button;t.fn.button=e,t.fn.button.Constructor=n,t.fn.button.noConflict=function(){return t.fn.button=r,this},t(document).on("click.bs.button.data-api",'[data-toggle^="button"]',function(n){var r=t(n.target).closest(".btn");e.call(r,"toggle"),t(n.target).is('input[type="radio"], input[type="checkbox"]')||(n.preventDefault(),r.is("input,button")?r.trigger("focus"):r.find("input:visible,button:visible").first().trigger("focus"))}).on("focus.bs.button.data-api blur.bs.button.data-api",'[data-toggle^="button"]',function(e){t(e.target).closest(".btn").toggleClass("focus",/^focus(in)?$/.test(e.type))})}(t)}).call(e,n(1))},4:function(t,e,n){n(25),t.exports="ui.bootstrap"},40:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.carousel"),o=t.extend({},n.DEFAULTS,r.data(),"object"==typeof e&&e),a="string"==typeof e?e:o.slide;i||r.data("bs.carousel",i=new n(this,o)),"number"==typeof e?i.to(e):a?i[a]():o.interval&&i.pause().cycle()})}var n=function(e,n){this.$element=t(e),this.$indicators=this.$element.find(".carousel-indicators"),this.options=n,this.paused=null,this.sliding=null,this.interval=null,this.$active=null,this.$items=null,this.options.keyboard&&this.$element.on("keydown.bs.carousel",t.proxy(this.keydown,this)),"hover"==this.options.pause&&!("ontouchstart"in document.documentElement)&&this.$element.on("mouseenter.bs.carousel",t.proxy(this.pause,this)).on("mouseleave.bs.carousel",t.proxy(this.cycle,this))};n.VERSION="3.3.7",n.TRANSITION_DURATION=600,n.DEFAULTS={interval:5e3,pause:"hover",wrap:!0,keyboard:!0},n.prototype.keydown=function(t){if(!/input|textarea/i.test(t.target.tagName)){switch(t.which){case 37:this.prev();break;case 39:this.next();break;default:return}t.preventDefault()}},n.prototype.cycle=function(e){return e||(this.paused=!1),this.interval&&clearInterval(this.interval),this.options.interval&&!this.paused&&(this.interval=setInterval(t.proxy(this.next,this),this.options.interval)),this},n.prototype.getItemIndex=function(t){return this.$items=t.parent().children(".item"),this.$items.index(t||this.$active)},n.prototype.getItemForDirection=function(t,e){var n=this.getItemIndex(e),r="prev"==t&&0===n||"next"==t&&n==this.$items.length-1;if(r&&!this.options.wrap)return e;var i="prev"==t?-1:1,o=(n+i)%this.$items.length;return this.$items.eq(o)},n.prototype.to=function(t){var e=this,n=this.getItemIndex(this.$active=this.$element.find(".item.active"));if(!(t>this.$items.length-1||t<0))return this.sliding?this.$element.one("slid.bs.carousel",function(){e.to(t)}):n==t?this.pause().cycle():this.slide(t>n?"next":"prev",this.$items.eq(t))},n.prototype.pause=function(e){return e||(this.paused=!0),this.$element.find(".next, .prev").length&&t.support.transition&&(this.$element.trigger(t.support.transition.end),this.cycle(!0)),this.interval=clearInterval(this.interval),this},n.prototype.next=function(){if(!this.sliding)return this.slide("next")},n.prototype.prev=function(){if(!this.sliding)return this.slide("prev")},n.prototype.slide=function(e,r){var i=this.$element.find(".item.active"),o=r||this.getItemForDirection(e,i),a=this.interval,s="next"==e?"left":"right",u=this;if(o.hasClass("active"))return this.sliding=!1;var l=o[0],c=t.Event("slide.bs.carousel",{relatedTarget:l,direction:s});if(this.$element.trigger(c),!c.isDefaultPrevented()){if(this.sliding=!0,a&&this.pause(),this.$indicators.length){this.$indicators.find(".active").removeClass("active");var f=t(this.$indicators.children()[this.getItemIndex(o)]);f&&f.addClass("active")}var p=t.Event("slid.bs.carousel",{relatedTarget:l,direction:s});return t.support.transition&&this.$element.hasClass("slide")?(o.addClass(e),o[0].offsetWidth,i.addClass(s),o.addClass(s),i.one("bsTransitionEnd",function(){o.removeClass([e,s].join(" ")).addClass("active"),i.removeClass(["active",s].join(" ")),u.sliding=!1,setTimeout(function(){u.$element.trigger(p)},0)}).emulateTransitionEnd(n.TRANSITION_DURATION)):(i.removeClass("active"),o.addClass("active"),this.sliding=!1,this.$element.trigger(p)),a&&this.cycle(),this}};var r=t.fn.carousel;t.fn.carousel=e,t.fn.carousel.Constructor=n,t.fn.carousel.noConflict=function(){return t.fn.carousel=r,this};var i=function(n){var r,i=t(this),o=t(i.attr("data-target")||(r=i.attr("href"))&&r.replace(/.*(?=#[^\s]+$)/,""));if(o.hasClass("carousel")){var a=t.extend({},o.data(),i.data()),s=i.attr("data-slide-to");s&&(a.interval=!1),e.call(o,a),s&&o.data("bs.carousel").to(s),n.preventDefault()}};t(document).on("click.bs.carousel.data-api","[data-slide]",i).on("click.bs.carousel.data-api","[data-slide-to]",i),t(window).on("load",function(){t('[data-ride="carousel"]').each(function(){var n=t(this);e.call(n,n.data())})})}(t)}).call(e,n(1))},41:function(t,e,n){(function(t){+function(t){"use strict";function e(e){var n,r=e.attr("data-target")||(n=e.attr("href"))&&n.replace(/.*(?=#[^\s]+$)/,"");return t(r)}function n(e){return this.each(function(){var n=t(this),i=n.data("bs.collapse"),o=t.extend({},r.DEFAULTS,n.data(),"object"==typeof e&&e);!i&&o.toggle&&/show|hide/.test(e)&&(o.toggle=!1),i||n.data("bs.collapse",i=new r(this,o)),"string"==typeof e&&i[e]()})}var r=function(e,n){this.$element=t(e),this.options=t.extend({},r.DEFAULTS,n),this.$trigger=t('[data-toggle="collapse"][href="#'+e.id+'"],[data-toggle="collapse"][data-target="#'+e.id+'"]'),this.transitioning=null,this.options.parent?this.$parent=this.getParent():this.addAriaAndCollapsedClass(this.$element,this.$trigger),this.options.toggle&&this.toggle()};r.VERSION="3.3.7",r.TRANSITION_DURATION=350,r.DEFAULTS={toggle:!0},r.prototype.dimension=function(){var t=this.$element.hasClass("width");return t?"width":"height"},r.prototype.show=function(){if(!this.transitioning&&!this.$element.hasClass("in")){var e,i=this.$parent&&this.$parent.children(".panel").children(".in, .collapsing");if(!(i&&i.length&&(e=i.data("bs.collapse"),e&&e.transitioning))){var o=t.Event("show.bs.collapse");if(this.$element.trigger(o),!o.isDefaultPrevented()){i&&i.length&&(n.call(i,"hide"),e||i.data("bs.collapse",null));var a=this.dimension();this.$element.removeClass("collapse").addClass("collapsing")[a](0).attr("aria-expanded",!0),this.$trigger.removeClass("collapsed").attr("aria-expanded",!0),this.transitioning=1;var s=function(){this.$element.removeClass("collapsing").addClass("collapse in")[a](""),this.transitioning=0,this.$element.trigger("shown.bs.collapse")};if(!t.support.transition)return s.call(this);var u=t.camelCase(["scroll",a].join("-"));this.$element.one("bsTransitionEnd",t.proxy(s,this)).emulateTransitionEnd(r.TRANSITION_DURATION)[a](this.$element[0][u])}}}},r.prototype.hide=function(){if(!this.transitioning&&this.$element.hasClass("in")){var e=t.Event("hide.bs.collapse");if(this.$element.trigger(e),!e.isDefaultPrevented()){var n=this.dimension();this.$element[n](this.$element[n]())[0].offsetHeight,this.$element.addClass("collapsing").removeClass("collapse in").attr("aria-expanded",!1),this.$trigger.addClass("collapsed").attr("aria-expanded",!1),this.transitioning=1;var i=function(){this.transitioning=0,this.$element.removeClass("collapsing").addClass("collapse").trigger("hidden.bs.collapse")};return t.support.transition?void this.$element[n](0).one("bsTransitionEnd",t.proxy(i,this)).emulateTransitionEnd(r.TRANSITION_DURATION):i.call(this)}}},r.prototype.toggle=function(){this[this.$element.hasClass("in")?"hide":"show"]()},r.prototype.getParent=function(){return t(this.options.parent).find('[data-toggle="collapse"][data-parent="'+this.options.parent+'"]').each(t.proxy(function(n,r){var i=t(r);this.addAriaAndCollapsedClass(e(i),i)},this)).end()},r.prototype.addAriaAndCollapsedClass=function(t,e){var n=t.hasClass("in");t.attr("aria-expanded",n),e.toggleClass("collapsed",!n).attr("aria-expanded",n)};var i=t.fn.collapse;t.fn.collapse=n,t.fn.collapse.Constructor=r,t.fn.collapse.noConflict=function(){return t.fn.collapse=i,this},t(document).on("click.bs.collapse.data-api",'[data-toggle="collapse"]',function(r){var i=t(this);i.attr("data-target")||r.preventDefault();var o=e(i),a=o.data("bs.collapse"),s=a?"toggle":i.data();n.call(o,s)})}(t)}).call(e,n(1))},42:function(t,e,n){(function(t){+function(t){"use strict";function e(e){var n=e.attr("data-target");n||(n=e.attr("href"),n=n&&/#[A-Za-z]/.test(n)&&n.replace(/.*(?=#[^\s]*$)/,""));var r=n&&t(n);return r&&r.length?r:e.parent()}function n(n){n&&3===n.which||(t(i).remove(),t(o).each(function(){var r=t(this),i=e(r),o={relatedTarget:this};i.hasClass("open")&&(n&&"click"==n.type&&/input|textarea/i.test(n.target.tagName)&&t.contains(i[0],n.target)||(i.trigger(n=t.Event("hide.bs.dropdown",o)),n.isDefaultPrevented()||(r.attr("aria-expanded","false"),i.removeClass("open").trigger(t.Event("hidden.bs.dropdown",o)))))}))}function r(e){return this.each(function(){var n=t(this),r=n.data("bs.dropdown");r||n.data("bs.dropdown",r=new a(this)),"string"==typeof e&&r[e].call(n)})}var i=".dropdown-backdrop",o='[data-toggle="dropdown"]',a=function(e){t(e).on("click.bs.dropdown",this.toggle)};a.VERSION="3.3.7",a.prototype.toggle=function(r){var i=t(this);if(!i.is(".disabled, :disabled")){var o=e(i),a=o.hasClass("open");if(n(),!a){"ontouchstart"in document.documentElement&&!o.closest(".navbar-nav").length&&t(document.createElement("div")).addClass("dropdown-backdrop").insertAfter(t(this)).on("click",n);var s={relatedTarget:this};if(o.trigger(r=t.Event("show.bs.dropdown",s)),r.isDefaultPrevented())return;i.trigger("focus").attr("aria-expanded","true"),o.toggleClass("open").trigger(t.Event("shown.bs.dropdown",s))}return!1}},a.prototype.keydown=function(n){if(/(38|40|27|32)/.test(n.which)&&!/input|textarea/i.test(n.target.tagName)){var r=t(this);if(n.preventDefault(),n.stopPropagation(),!r.is(".disabled, :disabled")){var i=e(r),a=i.hasClass("open");if(!a&&27!=n.which||a&&27==n.which)return 27==n.which&&i.find(o).trigger("focus"),r.trigger("click");var s=" li:not(.disabled):visible a",u=i.find(".dropdown-menu"+s);if(u.length){var l=u.index(n.target);38==n.which&&l>0&&l--,40==n.which&&l<u.length-1&&l++,~l||(l=0),u.eq(l).trigger("focus")}}}};var s=t.fn.dropdown;t.fn.dropdown=r,t.fn.dropdown.Constructor=a,t.fn.dropdown.noConflict=function(){return t.fn.dropdown=s,this},t(document).on("click.bs.dropdown.data-api",n).on("click.bs.dropdown.data-api",".dropdown form",function(t){t.stopPropagation()}).on("click.bs.dropdown.data-api",o,a.prototype.toggle).on("keydown.bs.dropdown.data-api",o,a.prototype.keydown).on("keydown.bs.dropdown.data-api",".dropdown-menu",a.prototype.keydown)}(t)}).call(e,n(1))},43:function(t,e,n){(function(t){+function(t){"use strict";function e(e,r){return this.each(function(){var i=t(this),o=i.data("bs.modal"),a=t.extend({},n.DEFAULTS,i.data(),"object"==typeof e&&e);o||i.data("bs.modal",o=new n(this,a)),"string"==typeof e?o[e](r):a.show&&o.show(r)})}var n=function(e,n){this.options=n,this.$body=t(document.body),this.$element=t(e),this.$dialog=this.$element.find(".modal-dialog"),this.$backdrop=null,this.isShown=null,this.originalBodyPad=null,this.scrollbarWidth=0,this.ignoreBackdropClick=!1,this.options.remote&&this.$element.find(".modal-content").load(this.options.remote,t.proxy(function(){this.$element.trigger("loaded.bs.modal")},this))};n.VERSION="3.3.7",n.TRANSITION_DURATION=300,n.BACKDROP_TRANSITION_DURATION=150,n.DEFAULTS={backdrop:!0,keyboard:!0,show:!0},n.prototype.toggle=function(t){return this.isShown?this.hide():this.show(t)},n.prototype.show=function(e){var r=this,i=t.Event("show.bs.modal",{relatedTarget:e});this.$element.trigger(i),this.isShown||i.isDefaultPrevented()||(this.isShown=!0,this.checkScrollbar(),this.setScrollbar(),this.$body.addClass("modal-open"),this.escape(),this.resize(),this.$element.on("click.dismiss.bs.modal",'[data-dismiss="modal"]',t.proxy(this.hide,this)),this.$dialog.on("mousedown.dismiss.bs.modal",function(){r.$element.one("mouseup.dismiss.bs.modal",function(e){t(e.target).is(r.$element)&&(r.ignoreBackdropClick=!0)})}),this.backdrop(function(){var i=t.support.transition&&r.$element.hasClass("fade");r.$element.parent().length||r.$element.appendTo(r.$body),r.$element.show().scrollTop(0),r.adjustDialog(),i&&r.$element[0].offsetWidth,r.$element.addClass("in"),r.enforceFocus();var o=t.Event("shown.bs.modal",{relatedTarget:e});i?r.$dialog.one("bsTransitionEnd",function(){r.$element.trigger("focus").trigger(o)}).emulateTransitionEnd(n.TRANSITION_DURATION):r.$element.trigger("focus").trigger(o)}))},n.prototype.hide=function(e){e&&e.preventDefault(),e=t.Event("hide.bs.modal"),this.$element.trigger(e),this.isShown&&!e.isDefaultPrevented()&&(this.isShown=!1,this.escape(),this.resize(),t(document).off("focusin.bs.modal"),this.$element.removeClass("in").off("click.dismiss.bs.modal").off("mouseup.dismiss.bs.modal"),this.$dialog.off("mousedown.dismiss.bs.modal"),t.support.transition&&this.$element.hasClass("fade")?this.$element.one("bsTransitionEnd",t.proxy(this.hideModal,this)).emulateTransitionEnd(n.TRANSITION_DURATION):this.hideModal())},n.prototype.enforceFocus=function(){t(document).off("focusin.bs.modal").on("focusin.bs.modal",t.proxy(function(t){document===t.target||this.$element[0]===t.target||this.$element.has(t.target).length||this.$element.trigger("focus");
},this))},n.prototype.escape=function(){this.isShown&&this.options.keyboard?this.$element.on("keydown.dismiss.bs.modal",t.proxy(function(t){27==t.which&&this.hide()},this)):this.isShown||this.$element.off("keydown.dismiss.bs.modal")},n.prototype.resize=function(){this.isShown?t(window).on("resize.bs.modal",t.proxy(this.handleUpdate,this)):t(window).off("resize.bs.modal")},n.prototype.hideModal=function(){var t=this;this.$element.hide(),this.backdrop(function(){t.$body.removeClass("modal-open"),t.resetAdjustments(),t.resetScrollbar(),t.$element.trigger("hidden.bs.modal")})},n.prototype.removeBackdrop=function(){this.$backdrop&&this.$backdrop.remove(),this.$backdrop=null},n.prototype.backdrop=function(e){var r=this,i=this.$element.hasClass("fade")?"fade":"";if(this.isShown&&this.options.backdrop){var o=t.support.transition&&i;if(this.$backdrop=t(document.createElement("div")).addClass("modal-backdrop "+i).appendTo(this.$body),this.$element.on("click.dismiss.bs.modal",t.proxy(function(t){return this.ignoreBackdropClick?void(this.ignoreBackdropClick=!1):void(t.target===t.currentTarget&&("static"==this.options.backdrop?this.$element[0].focus():this.hide()))},this)),o&&this.$backdrop[0].offsetWidth,this.$backdrop.addClass("in"),!e)return;o?this.$backdrop.one("bsTransitionEnd",e).emulateTransitionEnd(n.BACKDROP_TRANSITION_DURATION):e()}else if(!this.isShown&&this.$backdrop){this.$backdrop.removeClass("in");var a=function(){r.removeBackdrop(),e&&e()};t.support.transition&&this.$element.hasClass("fade")?this.$backdrop.one("bsTransitionEnd",a).emulateTransitionEnd(n.BACKDROP_TRANSITION_DURATION):a()}else e&&e()},n.prototype.handleUpdate=function(){this.adjustDialog()},n.prototype.adjustDialog=function(){var t=this.$element[0].scrollHeight>document.documentElement.clientHeight;this.$element.css({paddingLeft:!this.bodyIsOverflowing&&t?this.scrollbarWidth:"",paddingRight:this.bodyIsOverflowing&&!t?this.scrollbarWidth:""})},n.prototype.resetAdjustments=function(){this.$element.css({paddingLeft:"",paddingRight:""})},n.prototype.checkScrollbar=function(){var t=window.innerWidth;if(!t){var e=document.documentElement.getBoundingClientRect();t=e.right-Math.abs(e.left)}this.bodyIsOverflowing=document.body.clientWidth<t,this.scrollbarWidth=this.measureScrollbar()},n.prototype.setScrollbar=function(){var t=parseInt(this.$body.css("padding-right")||0,10);this.originalBodyPad=document.body.style.paddingRight||"",this.bodyIsOverflowing&&this.$body.css("padding-right",t+this.scrollbarWidth)},n.prototype.resetScrollbar=function(){this.$body.css("padding-right",this.originalBodyPad)},n.prototype.measureScrollbar=function(){var t=document.createElement("div");t.className="modal-scrollbar-measure",this.$body.append(t);var e=t.offsetWidth-t.clientWidth;return this.$body[0].removeChild(t),e};var r=t.fn.modal;t.fn.modal=e,t.fn.modal.Constructor=n,t.fn.modal.noConflict=function(){return t.fn.modal=r,this},t(document).on("click.bs.modal.data-api",'[data-toggle="modal"]',function(n){var r=t(this),i=r.attr("href"),o=t(r.attr("data-target")||i&&i.replace(/.*(?=#[^\s]+$)/,"")),a=o.data("bs.modal")?"toggle":t.extend({remote:!/#/.test(i)&&i},o.data(),r.data());r.is("a")&&n.preventDefault(),o.one("show.bs.modal",function(t){t.isDefaultPrevented()||o.one("hidden.bs.modal",function(){r.is(":visible")&&r.trigger("focus")})}),e.call(o,a,this)})}(t)}).call(e,n(1))},44:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.popover"),o="object"==typeof e&&e;!i&&/destroy|hide/.test(e)||(i||r.data("bs.popover",i=new n(this,o)),"string"==typeof e&&i[e]())})}var n=function(t,e){this.init("popover",t,e)};if(!t.fn.tooltip)throw new Error("Popover requires tooltip.js");n.VERSION="3.3.7",n.DEFAULTS=t.extend({},t.fn.tooltip.Constructor.DEFAULTS,{placement:"right",trigger:"click",content:"",template:'<div class="popover" role="tooltip"><div class="arrow"></div><h3 class="popover-title"></h3><div class="popover-content"></div></div>'}),n.prototype=t.extend({},t.fn.tooltip.Constructor.prototype),n.prototype.constructor=n,n.prototype.getDefaults=function(){return n.DEFAULTS},n.prototype.setContent=function(){var t=this.tip(),e=this.getTitle(),n=this.getContent();t.find(".popover-title")[this.options.html?"html":"text"](e),t.find(".popover-content").children().detach().end()[this.options.html?"string"==typeof n?"html":"append":"text"](n),t.removeClass("fade top bottom left right in"),t.find(".popover-title").html()||t.find(".popover-title").hide()},n.prototype.hasContent=function(){return this.getTitle()||this.getContent()},n.prototype.getContent=function(){var t=this.$element,e=this.options;return t.attr("data-content")||("function"==typeof e.content?e.content.call(t[0]):e.content)},n.prototype.arrow=function(){return this.$arrow=this.$arrow||this.tip().find(".arrow")};var r=t.fn.popover;t.fn.popover=e,t.fn.popover.Constructor=n,t.fn.popover.noConflict=function(){return t.fn.popover=r,this}}(t)}).call(e,n(1))},45:function(t,e,n){(function(t){+function(t){"use strict";function e(n,r){this.$body=t(document.body),this.$scrollElement=t(t(n).is(document.body)?window:n),this.options=t.extend({},e.DEFAULTS,r),this.selector=(this.options.target||"")+" .nav li > a",this.offsets=[],this.targets=[],this.activeTarget=null,this.scrollHeight=0,this.$scrollElement.on("scroll.bs.scrollspy",t.proxy(this.process,this)),this.refresh(),this.process()}function n(n){return this.each(function(){var r=t(this),i=r.data("bs.scrollspy"),o="object"==typeof n&&n;i||r.data("bs.scrollspy",i=new e(this,o)),"string"==typeof n&&i[n]()})}e.VERSION="3.3.7",e.DEFAULTS={offset:10},e.prototype.getScrollHeight=function(){return this.$scrollElement[0].scrollHeight||Math.max(this.$body[0].scrollHeight,document.documentElement.scrollHeight)},e.prototype.refresh=function(){var e=this,n="offset",r=0;this.offsets=[],this.targets=[],this.scrollHeight=this.getScrollHeight(),t.isWindow(this.$scrollElement[0])||(n="position",r=this.$scrollElement.scrollTop()),this.$body.find(this.selector).map(function(){var e=t(this),i=e.data("target")||e.attr("href"),o=/^#./.test(i)&&t(i);return o&&o.length&&o.is(":visible")&&[[o[n]().top+r,i]]||null}).sort(function(t,e){return t[0]-e[0]}).each(function(){e.offsets.push(this[0]),e.targets.push(this[1])})},e.prototype.process=function(){var t,e=this.$scrollElement.scrollTop()+this.options.offset,n=this.getScrollHeight(),r=this.options.offset+n-this.$scrollElement.height(),i=this.offsets,o=this.targets,a=this.activeTarget;if(this.scrollHeight!=n&&this.refresh(),e>=r)return a!=(t=o[o.length-1])&&this.activate(t);if(a&&e<i[0])return this.activeTarget=null,this.clear();for(t=i.length;t--;)a!=o[t]&&e>=i[t]&&(void 0===i[t+1]||e<i[t+1])&&this.activate(o[t])},e.prototype.activate=function(e){this.activeTarget=e,this.clear();var n=this.selector+'[data-target="'+e+'"],'+this.selector+'[href="'+e+'"]',r=t(n).parents("li").addClass("active");r.parent(".dropdown-menu").length&&(r=r.closest("li.dropdown").addClass("active")),r.trigger("activate.bs.scrollspy")},e.prototype.clear=function(){t(this.selector).parentsUntil(this.options.target,".active").removeClass("active")};var r=t.fn.scrollspy;t.fn.scrollspy=n,t.fn.scrollspy.Constructor=e,t.fn.scrollspy.noConflict=function(){return t.fn.scrollspy=r,this},t(window).on("load.bs.scrollspy.data-api",function(){t('[data-spy="scroll"]').each(function(){var e=t(this);n.call(e,e.data())})})}(t)}).call(e,n(1))},46:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.tab");i||r.data("bs.tab",i=new n(this)),"string"==typeof e&&i[e]()})}var n=function(e){this.element=t(e)};n.VERSION="3.3.7",n.TRANSITION_DURATION=150,n.prototype.show=function(){var e=this.element,n=e.closest("ul:not(.dropdown-menu)"),r=e.data("target");if(r||(r=e.attr("href"),r=r&&r.replace(/.*(?=#[^\s]*$)/,"")),!e.parent("li").hasClass("active")){var i=n.find(".active:last a"),o=t.Event("hide.bs.tab",{relatedTarget:e[0]}),a=t.Event("show.bs.tab",{relatedTarget:i[0]});if(i.trigger(o),e.trigger(a),!a.isDefaultPrevented()&&!o.isDefaultPrevented()){var s=t(r);this.activate(e.closest("li"),n),this.activate(s,s.parent(),function(){i.trigger({type:"hidden.bs.tab",relatedTarget:e[0]}),e.trigger({type:"shown.bs.tab",relatedTarget:i[0]})})}}},n.prototype.activate=function(e,r,i){function o(){a.removeClass("active").find("> .dropdown-menu > .active").removeClass("active").end().find('[data-toggle="tab"]').attr("aria-expanded",!1),e.addClass("active").find('[data-toggle="tab"]').attr("aria-expanded",!0),s?(e[0].offsetWidth,e.addClass("in")):e.removeClass("fade"),e.parent(".dropdown-menu").length&&e.closest("li.dropdown").addClass("active").end().find('[data-toggle="tab"]').attr("aria-expanded",!0),i&&i()}var a=r.find("> .active"),s=i&&t.support.transition&&(a.length&&a.hasClass("fade")||!!r.find("> .fade").length);a.length&&s?a.one("bsTransitionEnd",o).emulateTransitionEnd(n.TRANSITION_DURATION):o(),a.removeClass("in")};var r=t.fn.tab;t.fn.tab=e,t.fn.tab.Constructor=n,t.fn.tab.noConflict=function(){return t.fn.tab=r,this};var i=function(n){n.preventDefault(),e.call(t(this),"show")};t(document).on("click.bs.tab.data-api",'[data-toggle="tab"]',i).on("click.bs.tab.data-api",'[data-toggle="pill"]',i)}(t)}).call(e,n(1))},47:function(t,e,n){(function(t){+function(t){"use strict";function e(e){return this.each(function(){var r=t(this),i=r.data("bs.tooltip"),o="object"==typeof e&&e;!i&&/destroy|hide/.test(e)||(i||r.data("bs.tooltip",i=new n(this,o)),"string"==typeof e&&i[e]())})}var n=function(t,e){this.type=null,this.options=null,this.enabled=null,this.timeout=null,this.hoverState=null,this.$element=null,this.inState=null,this.init("tooltip",t,e)};n.VERSION="3.3.7",n.TRANSITION_DURATION=150,n.DEFAULTS={animation:!0,placement:"top",selector:!1,template:'<div class="tooltip" role="tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>',trigger:"hover focus",title:"",delay:0,html:!1,container:!1,viewport:{selector:"body",padding:0}},n.prototype.init=function(e,n,r){if(this.enabled=!0,this.type=e,this.$element=t(n),this.options=this.getOptions(r),this.$viewport=this.options.viewport&&t(t.isFunction(this.options.viewport)?this.options.viewport.call(this,this.$element):this.options.viewport.selector||this.options.viewport),this.inState={click:!1,hover:!1,focus:!1},this.$element[0]instanceof document.constructor&&!this.options.selector)throw new Error("`selector` option must be specified when initializing "+this.type+" on the window.document object!");for(var i=this.options.trigger.split(" "),o=i.length;o--;){var a=i[o];if("click"==a)this.$element.on("click."+this.type,this.options.selector,t.proxy(this.toggle,this));else if("manual"!=a){var s="hover"==a?"mouseenter":"focusin",u="hover"==a?"mouseleave":"focusout";this.$element.on(s+"."+this.type,this.options.selector,t.proxy(this.enter,this)),this.$element.on(u+"."+this.type,this.options.selector,t.proxy(this.leave,this))}}this.options.selector?this._options=t.extend({},this.options,{trigger:"manual",selector:""}):this.fixTitle()},n.prototype.getDefaults=function(){return n.DEFAULTS},n.prototype.getOptions=function(e){return e=t.extend({},this.getDefaults(),this.$element.data(),e),e.delay&&"number"==typeof e.delay&&(e.delay={show:e.delay,hide:e.delay}),e},n.prototype.getDelegateOptions=function(){var e={},n=this.getDefaults();return this._options&&t.each(this._options,function(t,r){n[t]!=r&&(e[t]=r)}),e},n.prototype.enter=function(e){var n=e instanceof this.constructor?e:t(e.currentTarget).data("bs."+this.type);return n||(n=new this.constructor(e.currentTarget,this.getDelegateOptions()),t(e.currentTarget).data("bs."+this.type,n)),e instanceof t.Event&&(n.inState["focusin"==e.type?"focus":"hover"]=!0),n.tip().hasClass("in")||"in"==n.hoverState?void(n.hoverState="in"):(clearTimeout(n.timeout),n.hoverState="in",n.options.delay&&n.options.delay.show?void(n.timeout=setTimeout(function(){"in"==n.hoverState&&n.show()},n.options.delay.show)):n.show())},n.prototype.isInStateTrue=function(){for(var t in this.inState)if(this.inState[t])return!0;return!1},n.prototype.leave=function(e){var n=e instanceof this.constructor?e:t(e.currentTarget).data("bs."+this.type);if(n||(n=new this.constructor(e.currentTarget,this.getDelegateOptions()),t(e.currentTarget).data("bs."+this.type,n)),e instanceof t.Event&&(n.inState["focusout"==e.type?"focus":"hover"]=!1),!n.isInStateTrue())return clearTimeout(n.timeout),n.hoverState="out",n.options.delay&&n.options.delay.hide?void(n.timeout=setTimeout(function(){"out"==n.hoverState&&n.hide()},n.options.delay.hide)):n.hide()},n.prototype.show=function(){var e=t.Event("show.bs."+this.type);if(this.hasContent()&&this.enabled){this.$element.trigger(e);var r=t.contains(this.$element[0].ownerDocument.documentElement,this.$element[0]);if(e.isDefaultPrevented()||!r)return;var i=this,o=this.tip(),a=this.getUID(this.type);this.setContent(),o.attr("id",a),this.$element.attr("aria-describedby",a),this.options.animation&&o.addClass("fade");var s="function"==typeof this.options.placement?this.options.placement.call(this,o[0],this.$element[0]):this.options.placement,u=/\s?auto?\s?/i,l=u.test(s);l&&(s=s.replace(u,"")||"top"),o.detach().css({top:0,left:0,display:"block"}).addClass(s).data("bs."+this.type,this),this.options.container?o.appendTo(this.options.container):o.insertAfter(this.$element),this.$element.trigger("inserted.bs."+this.type);var c=this.getPosition(),f=o[0].offsetWidth,p=o[0].offsetHeight;if(l){var h=s,d=this.getPosition(this.$viewport);s="bottom"==s&&c.bottom+p>d.bottom?"top":"top"==s&&c.top-p<d.top?"bottom":"right"==s&&c.right+f>d.width?"left":"left"==s&&c.left-f<d.left?"right":s,o.removeClass(h).addClass(s)}var v=this.getCalculatedOffset(s,c,f,p);this.applyPlacement(v,s);var g=function(){var t=i.hoverState;i.$element.trigger("shown.bs."+i.type),i.hoverState=null,"out"==t&&i.leave(i)};t.support.transition&&this.$tip.hasClass("fade")?o.one("bsTransitionEnd",g).emulateTransitionEnd(n.TRANSITION_DURATION):g()}},n.prototype.applyPlacement=function(e,n){var r=this.tip(),i=r[0].offsetWidth,o=r[0].offsetHeight,a=parseInt(r.css("margin-top"),10),s=parseInt(r.css("margin-left"),10);isNaN(a)&&(a=0),isNaN(s)&&(s=0),e.top+=a,e.left+=s,t.offset.setOffset(r[0],t.extend({using:function(t){r.css({top:Math.round(t.top),left:Math.round(t.left)})}},e),0),r.addClass("in");var u=r[0].offsetWidth,l=r[0].offsetHeight;"top"==n&&l!=o&&(e.top=e.top+o-l);var c=this.getViewportAdjustedDelta(n,e,u,l);c.left?e.left+=c.left:e.top+=c.top;var f=/top|bottom/.test(n),p=f?2*c.left-i+u:2*c.top-o+l,h=f?"offsetWidth":"offsetHeight";r.offset(e),this.replaceArrow(p,r[0][h],f)},n.prototype.replaceArrow=function(t,e,n){this.arrow().css(n?"left":"top",50*(1-t/e)+"%").css(n?"top":"left","")},n.prototype.setContent=function(){var t=this.tip(),e=this.getTitle();t.find(".tooltip-inner")[this.options.html?"html":"text"](e),t.removeClass("fade in top bottom left right")},n.prototype.hide=function(e){function r(){"in"!=i.hoverState&&o.detach(),i.$element&&i.$element.removeAttr("aria-describedby").trigger("hidden.bs."+i.type),e&&e()}var i=this,o=t(this.$tip),a=t.Event("hide.bs."+this.type);if(this.$element.trigger(a),!a.isDefaultPrevented())return o.removeClass("in"),t.support.transition&&o.hasClass("fade")?o.one("bsTransitionEnd",r).emulateTransitionEnd(n.TRANSITION_DURATION):r(),this.hoverState=null,this},n.prototype.fixTitle=function(){var t=this.$element;(t.attr("title")||"string"!=typeof t.attr("data-original-title"))&&t.attr("data-original-title",t.attr("title")||"").attr("title","")},n.prototype.hasContent=function(){return this.getTitle()},n.prototype.getPosition=function(e){e=e||this.$element;var n=e[0],r="BODY"==n.tagName,i=n.getBoundingClientRect();null==i.width&&(i=t.extend({},i,{width:i.right-i.left,height:i.bottom-i.top}));var o=window.SVGElement&&n instanceof window.SVGElement,a=r?{top:0,left:0}:o?null:e.offset(),s={scroll:r?document.documentElement.scrollTop||document.body.scrollTop:e.scrollTop()},u=r?{width:t(window).width(),height:t(window).height()}:null;return t.extend({},i,s,u,a)},n.prototype.getCalculatedOffset=function(t,e,n,r){return"bottom"==t?{top:e.top+e.height,left:e.left+e.width/2-n/2}:"top"==t?{top:e.top-r,left:e.left+e.width/2-n/2}:"left"==t?{top:e.top+e.height/2-r/2,left:e.left-n}:{top:e.top+e.height/2-r/2,left:e.left+e.width}},n.prototype.getViewportAdjustedDelta=function(t,e,n,r){var i={top:0,left:0};if(!this.$viewport)return i;var o=this.options.viewport&&this.options.viewport.padding||0,a=this.getPosition(this.$viewport);if(/right|left/.test(t)){var s=e.top-o-a.scroll,u=e.top+o-a.scroll+r;s<a.top?i.top=a.top-s:u>a.top+a.height&&(i.top=a.top+a.height-u)}else{var l=e.left-o,c=e.left+o+n;l<a.left?i.left=a.left-l:c>a.right&&(i.left=a.left+a.width-c)}return i},n.prototype.getTitle=function(){var t,e=this.$element,n=this.options;return t=e.attr("data-original-title")||("function"==typeof n.title?n.title.call(e[0]):n.title)},n.prototype.getUID=function(t){do t+=~~(1e6*Math.random());while(document.getElementById(t));return t},n.prototype.tip=function(){if(!this.$tip&&(this.$tip=t(this.options.template),1!=this.$tip.length))throw new Error(this.type+" `template` option must consist of exactly 1 top-level element!");return this.$tip},n.prototype.arrow=function(){return this.$arrow=this.$arrow||this.tip().find(".tooltip-arrow")},n.prototype.enable=function(){this.enabled=!0},n.prototype.disable=function(){this.enabled=!1},n.prototype.toggleEnabled=function(){this.enabled=!this.enabled},n.prototype.toggle=function(e){var n=this;e&&(n=t(e.currentTarget).data("bs."+this.type),n||(n=new this.constructor(e.currentTarget,this.getDelegateOptions()),t(e.currentTarget).data("bs."+this.type,n))),e?(n.inState.click=!n.inState.click,n.isInStateTrue()?n.enter(n):n.leave(n)):n.tip().hasClass("in")?n.leave(n):n.enter(n)},n.prototype.destroy=function(){var t=this;clearTimeout(this.timeout),this.hide(function(){t.$element.off("."+t.type).removeData("bs."+t.type),t.$tip&&t.$tip.detach(),t.$tip=null,t.$arrow=null,t.$viewport=null,t.$element=null})};var r=t.fn.tooltip;t.fn.tooltip=e,t.fn.tooltip.Constructor=n,t.fn.tooltip.noConflict=function(){return t.fn.tooltip=r,this}}(t)}).call(e,n(1))},48:function(t,e,n){(function(t){+function(t){"use strict";function e(){var t=document.createElement("bootstrap"),e={WebkitTransition:"webkitTransitionEnd",MozTransition:"transitionend",OTransition:"oTransitionEnd otransitionend",transition:"transitionend"};for(var n in e)if(void 0!==t.style[n])return{end:e[n]};return!1}t.fn.emulateTransitionEnd=function(e){var n=!1,r=this;t(this).one("bsTransitionEnd",function(){n=!0});var i=function(){n||t(r).trigger(t.support.transition.end)};return setTimeout(i,e),this},t(function(){t.support.transition=e(),t.support.transition&&(t.event.special.bsTransitionEnd={bindType:t.support.transition.end,delegateType:t.support.transition.end,handle:function(e){if(t(e.target).is(this))return e.handleObj.handler.apply(this,arguments)}})})}(t)}).call(e,n(1))},5:function(t,e,n){n(137),t.exports="ng-ip-address"},6:function(t,e,n){n(23),t.exports="ngResource"},7:function(t,e){/**
 * AngularJS module that adds support for specifying default child views for abstract states when using ui-router.
 *
 * @link https://github.com/nonplus/angular-ui-router-default
 *
 * @license angular-ui-router-default v0.0.6
 * (c) Copyright Stepan Riha <github@nonplus.net>
 * License MIT
 */
!function(n){"use strict";var r="ui.router.default";"undefined"!=typeof t&&"undefined"!=typeof e&&t.exports===e&&(t.exports=r);var i=10;n.module(r,["ui.router"]).config(["$provide",function(t){t.decorator("$state",["$delegate","$injector","$q",function(t,e,n){var r,o=t.transitionTo;return t.transitionTo=function(a,s,u){function l(){function e(e){return u!==r?n.reject(new Error("transition superseded")):e?(a(),d=e,l()):o.call(t,d,v,g)}function a(){if(p===i)throw new Error("Too many abstract state default redirects");p+=1}var s=h.get(d,h.$current);if(!s)return o.call(t,d,v,g);d=s.name;var u=c(s);return r=u,n.when(u).then(e)}function c(t){function e(t){return"."===t[0]?d+t:t}return t&&t.abstract&&(t.abstract!==!0||t.default)?f(t).then(e):null}function f(t){var r;return r=t.default?t.default:t.abstract,r instanceof Function||r instanceof Array?n.when(e.invoke(r)):n.when(r)}var p=0,h=this,d=a.name||a,v=s,g=u;return l()},t}])}])}(window.angular)},8:function(t,e,n){!function(t,n){n(e)}(this,function(t){"use strict";function e(t){return function(e,n){return zl(t(e),n)}}function n(t,e){return[t,e]}function r(t,e,n){var r=(e-t)/Math.max(0,n),i=Math.floor(Math.log(r)/Math.LN10),o=r/Math.pow(10,i);return i>=0?(o>=lc?10:o>=cc?5:o>=fc?2:1)*Math.pow(10,i):-Math.pow(10,-i)/(o>=lc?10:o>=cc?5:o>=fc?2:1)}function i(t,e,n){var r=Math.abs(e-t)/Math.max(0,n),i=Math.pow(10,Math.floor(Math.log(r)/Math.LN10)),o=r/i;return o>=lc?i*=10:o>=cc?i*=5:o>=fc&&(i*=2),e<t?-i:i}function o(t){return t.length}function a(t){return"translate("+(t+.5)+",0)"}function s(t){return"translate(0,"+(t+.5)+")"}function u(t){var e=Math.max(0,t.bandwidth()-1)/2;return t.round()&&(e=Math.round(e)),function(n){return t(n)+e}}function l(){return!this.__axis}function c(t,e){function n(n){var a=null==i?e.ticks?e.ticks.apply(e,r):e.domain():i,s=null==o?e.tickFormat?e.tickFormat.apply(e,r):Dc:o,g=Math.max(c,0)+p,m=e.range(),y=m[0]+.5,b=m[m.length-1]+.5,$=(e.bandwidth?u:Dc)(e.copy()),w=n.selection?n.selection():n,x=w.selectAll(".domain").data([null]),_=w.selectAll(".tick").data(a,e).order(),C=_.exit(),k=_.enter().append("g").attr("class","tick"),S=_.select("line"),T=_.select("text");x=x.merge(x.enter().insert("path",".tick").attr("class","domain").attr("stroke","#000")),_=_.merge(k),S=S.merge(k.append("line").attr("stroke","#000").attr(d+"2",h*c)),T=T.merge(k.append("text").attr("fill","#000").attr(d,h*g).attr("dy",t===Ac?"0em":t===Nc?"0.71em":"0.32em")),n!==w&&(x=x.transition(n),_=_.transition(n),S=S.transition(n),T=T.transition(n),C=C.transition(n).attr("opacity",Rc).attr("transform",function(t){return isFinite(t=$(t))?v(t):this.getAttribute("transform")}),k.attr("opacity",Rc).attr("transform",function(t){var e=this.parentNode.__axis;return v(e&&isFinite(e=e(t))?e:$(t))})),C.remove(),x.attr("d",t===Pc||t==Oc?"M"+h*f+","+y+"H0.5V"+b+"H"+h*f:"M"+y+","+h*f+"V0.5H"+b+"V"+h*f),_.attr("opacity",1).attr("transform",function(t){return v($(t))}),S.attr(d+"2",h*c),T.attr(d,h*g).text(s),w.filter(l).attr("fill","none").attr("font-size",10).attr("font-family","sans-serif").attr("text-anchor",t===Oc?"start":t===Pc?"end":"middle"),w.each(function(){this.__axis=$})}var r=[],i=null,o=null,c=6,f=6,p=3,h=t===Ac||t===Pc?-1:1,d=t===Pc||t===Oc?"x":"y",v=t===Ac||t===Nc?a:s;return n.scale=function(t){return arguments.length?(e=t,n):e},n.ticks=function(){return r=Mc.call(arguments),n},n.tickArguments=function(t){return arguments.length?(r=null==t?[]:Mc.call(t),n):r.slice()},n.tickValues=function(t){return arguments.length?(i=null==t?null:Mc.call(t),n):i&&i.slice()},n.tickFormat=function(t){return arguments.length?(o=t,n):o},n.tickSize=function(t){return arguments.length?(c=f=+t,n):c},n.tickSizeInner=function(t){return arguments.length?(c=+t,n):c},n.tickSizeOuter=function(t){return arguments.length?(f=+t,n):f},n.tickPadding=function(t){return arguments.length?(p=+t,n):p},n}function f(t){return c(Ac,t)}function p(t){return c(Oc,t)}function h(t){return c(Nc,t)}function d(t){return c(Pc,t)}function v(){for(var t,e=0,n=arguments.length,r={};e<n;++e){if(!(t=arguments[e]+"")||t in r)throw new Error("illegal type: "+t);r[t]=[]}return new g(r)}function g(t){this._=t}function m(t,e){return t.trim().split(/^|\s+/).map(function(t){var n="",r=t.indexOf(".");if(r>=0&&(n=t.slice(r+1),t=t.slice(0,r)),t&&!e.hasOwnProperty(t))throw new Error("unknown type: "+t);return{type:t,name:n}})}function y(t,e){for(var n,r=0,i=t.length;r<i;++r)if((n=t[r]).name===e)return n.value}function b(t,e,n){for(var r=0,i=t.length;r<i;++r)if(t[r].name===e){t[r]=Ic,t=t.slice(0,r).concat(t.slice(r+1));break}return null!=n&&t.push({name:e,value:n}),t}function $(t){return function(){var e=this.ownerDocument,n=this.namespaceURI;return n===jc&&e.documentElement.namespaceURI===jc?e.createElement(t):e.createElementNS(n,t)}}function w(t){return function(){return this.ownerDocument.createElementNS(t.space,t.local)}}function x(){return new _}function _(){this._="@"+(++Uc).toString(36)}function C(t,e,n){return t=k(t,e,n),function(e){var n=e.relatedTarget;n&&(n===this||8&n.compareDocumentPosition(this))||t.call(this,e)}}function k(e,n,r){return function(i){var o=t.event;t.event=i;try{e.call(this,this.__data__,n,r)}finally{t.event=o}}}function S(t){return t.trim().split(/^|\s+/).map(function(t){var e="",n=t.indexOf(".");return n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),{type:t,name:e}})}function T(t){return function(){var e=this.__on;if(e){for(var n,r=0,i=-1,o=e.length;r<o;++r)n=e[r],t.type&&n.type!==t.type||n.name!==t.name?e[++i]=n:this.removeEventListener(n.type,n.listener,n.capture);++i?e.length=i:delete this.__on}}}function E(t,e,n){var r=Wc.hasOwnProperty(t.type)?C:k;return function(i,o,a){var s,u=this.__on,l=r(e,o,a);if(u)for(var c=0,f=u.length;c<f;++c)if((s=u[c]).type===t.type&&s.name===t.name)return this.removeEventListener(s.type,s.listener,s.capture),this.addEventListener(s.type,s.listener=l,s.capture=n),void(s.value=e);this.addEventListener(t.type,l,n),s={type:t.type,name:t.name,value:e,listener:l,capture:n},u?u.push(s):this.__on=[s]}}function M(e,n,r,i){var o=t.event;e.sourceEvent=t.event,t.event=e;try{return n.apply(r,i)}finally{t.event=o}}function D(){}function A(){return[]}function O(t,e){this.ownerDocument=t.ownerDocument,this.namespaceURI=t.namespaceURI,this._next=null,this._parent=t,this.__data__=e}function N(t,e,n,r,i,o){for(var a,s=0,u=e.length,l=o.length;s<l;++s)(a=e[s])?(a.__data__=o[s],r[s]=a):n[s]=new O(t,o[s]);for(;s<u;++s)(a=e[s])&&(i[s]=a)}function P(t,e,n,r,i,o,a){var s,u,l,c={},f=e.length,p=o.length,h=new Array(f);for(s=0;s<f;++s)(u=e[s])&&(h[s]=l=sf+a.call(u,u.__data__,s,e),l in c?i[s]=u:c[l]=u);for(s=0;s<p;++s)l=sf+a.call(t,o[s],s,o),(u=c[l])?(r[s]=u,u.__data__=o[s],c[l]=null):n[s]=new O(t,o[s]);for(s=0;s<f;++s)(u=e[s])&&c[h[s]]===u&&(i[s]=u)}function R(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}function I(t){return function(){this.removeAttribute(t)}}function j(t){return function(){this.removeAttributeNS(t.space,t.local)}}function F(t,e){return function(){this.setAttribute(t,e)}}function q(t,e){return function(){this.setAttributeNS(t.space,t.local,e)}}function V(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttribute(t):this.setAttribute(t,n)}}function U(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttributeNS(t.space,t.local):this.setAttributeNS(t.space,t.local,n)}}function L(t){return function(){this.style.removeProperty(t)}}function H(t,e,n){return function(){this.style.setProperty(t,e,n)}}function B(t,e,n){return function(){var r=e.apply(this,arguments);null==r?this.style.removeProperty(t):this.style.setProperty(t,r,n)}}function z(t,e){return t.style.getPropertyValue(e)||$f(t).getComputedStyle(t,null).getPropertyValue(e)}function W(t){return function(){delete this[t]}}function Y(t,e){return function(){this[t]=e}}function G(t,e){return function(){var n=e.apply(this,arguments);null==n?delete this[t]:this[t]=n}}function X(t){return t.trim().split(/^|\s+/)}function K(t){return t.classList||new J(t)}function J(t){this._node=t,this._names=X(t.getAttribute("class")||"")}function Z(t,e){for(var n=K(t),r=-1,i=e.length;++r<i;)n.add(e[r])}function Q(t,e){for(var n=K(t),r=-1,i=e.length;++r<i;)n.remove(e[r])}function tt(t){return function(){Z(this,t)}}function et(t){return function(){Q(this,t)}}function nt(t,e){return function(){(e.apply(this,arguments)?Z:Q)(this,t)}}function rt(){this.textContent=""}function it(t){return function(){this.textContent=t}}function ot(t){return function(){var e=t.apply(this,arguments);this.textContent=null==e?"":e}}function at(){this.innerHTML=""}function st(t){return function(){this.innerHTML=t}}function ut(t){return function(){var e=t.apply(this,arguments);this.innerHTML=null==e?"":e}}function lt(){this.nextSibling&&this.parentNode.appendChild(this)}function ct(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}function ft(){return null}function pt(){var t=this.parentNode;t&&t.removeChild(this)}function ht(t,e,n){var r=$f(t),i=r.CustomEvent;"function"==typeof i?i=new i(e,n):(i=r.document.createEvent("Event"),n?(i.initEvent(e,n.bubbles,n.cancelable),i.detail=n.detail):i.initEvent(e,!1,!1)),t.dispatchEvent(i)}function dt(t,e){return function(){return ht(this,t,e)}}function vt(t,e){return function(){return ht(this,t,e.apply(this,arguments))}}function gt(t,e){this._groups=t,this._parents=e}function mt(){return new gt([[document.documentElement]],Nf)}function yt(){t.event.stopImmediatePropagation()}function bt(t,e){var n=t.document.documentElement,r=Pf(t).on("dragstart.drag",null);e&&(r.on("click.drag",Ff,!0),setTimeout(function(){r.on("click.drag",null)},0)),"onselectstart"in n?r.on("selectstart.drag",null):(n.style.MozUserSelect=n.__noselect,delete n.__noselect)}function $t(t,e,n,r,i,o,a,s,u,l){this.target=t,this.type=e,this.subject=n,this.identifier=r,this.active=i,this.x=o,this.y=a,this.dx=s,this.dy=u,this._=l}function wt(){return!t.event.button}function xt(){return this.parentNode}function _t(e){return null==e?{x:t.event.x,y:t.event.y}:e}function Ct(t,e){var n=Object.create(t.prototype);for(var r in e)n[r]=e[r];return n}function kt(){}function St(t){var e;return t=(t+"").trim().toLowerCase(),(e=Gf.exec(t))?(e=parseInt(e[1],16),new At(e>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1)):(e=Xf.exec(t))?Tt(parseInt(e[1],16)):(e=Kf.exec(t))?new At(e[1],e[2],e[3],1):(e=Jf.exec(t))?new At(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=Zf.exec(t))?Et(e[1],e[2],e[3],e[4]):(e=Qf.exec(t))?Et(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=tp.exec(t))?Ot(e[1],e[2]/100,e[3]/100,1):(e=ep.exec(t))?Ot(e[1],e[2]/100,e[3]/100,e[4]):np.hasOwnProperty(t)?Tt(np[t]):"transparent"===t?new At(NaN,NaN,NaN,0):null}function Tt(t){return new At(t>>16&255,t>>8&255,255&t,1)}function Et(t,e,n,r){return r<=0&&(t=e=n=NaN),new At(t,e,n,r)}function Mt(t){return t instanceof kt||(t=St(t)),t?(t=t.rgb(),new At(t.r,t.g,t.b,t.opacity)):new At}function Dt(t,e,n,r){return 1===arguments.length?Mt(t):new At(t,e,n,null==r?1:r)}function At(t,e,n,r){this.r=+t,this.g=+e,this.b=+n,this.opacity=+r}function Ot(t,e,n,r){return r<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new Rt(t,e,n,r)}function Nt(t){if(t instanceof Rt)return new Rt(t.h,t.s,t.l,t.opacity);if(t instanceof kt||(t=St(t)),!t)return new Rt;if(t instanceof Rt)return t;t=t.rgb();var e=t.r/255,n=t.g/255,r=t.b/255,i=Math.min(e,n,r),o=Math.max(e,n,r),a=NaN,s=o-i,u=(o+i)/2;return s?(a=e===o?(n-r)/s+6*(n<r):n===o?(r-e)/s+2:(e-n)/s+4,s/=u<.5?o+i:2-o-i,a*=60):s=u>0&&u<1?0:a,new Rt(a,s,u,t.opacity)}function Pt(t,e,n,r){return 1===arguments.length?Nt(t):new Rt(t,e,n,null==r?1:r)}function Rt(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function It(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}function jt(t){if(t instanceof qt)return new qt(t.l,t.a,t.b,t.opacity);if(t instanceof Wt){var e=t.h*rp;return new qt(t.l,Math.cos(e)*t.c,Math.sin(e)*t.c,t.opacity)}t instanceof At||(t=Mt(t));var n=Ht(t.r),r=Ht(t.g),i=Ht(t.b),o=Vt((.4124564*n+.3575761*r+.1804375*i)/ap),a=Vt((.2126729*n+.7151522*r+.072175*i)/sp),s=Vt((.0193339*n+.119192*r+.9503041*i)/up);return new qt(116*a-16,500*(o-a),200*(a-s),t.opacity)}function Ft(t,e,n,r){return 1===arguments.length?jt(t):new qt(t,e,n,null==r?1:r)}function qt(t,e,n,r){this.l=+t,this.a=+e,this.b=+n,this.opacity=+r}function Vt(t){return t>pp?Math.pow(t,1/3):t/fp+lp}function Ut(t){return t>cp?t*t*t:fp*(t-lp)}function Lt(t){return 255*(t<=.0031308?12.92*t:1.055*Math.pow(t,1/2.4)-.055)}function Ht(t){return(t/=255)<=.04045?t/12.92:Math.pow((t+.055)/1.055,2.4)}function Bt(t){if(t instanceof Wt)return new Wt(t.h,t.c,t.l,t.opacity);t instanceof qt||(t=jt(t));var e=Math.atan2(t.b,t.a)*ip;return new Wt(e<0?e+360:e,Math.sqrt(t.a*t.a+t.b*t.b),t.l,t.opacity)}function zt(t,e,n,r){return 1===arguments.length?Bt(t):new Wt(t,e,n,null==r?1:r)}function Wt(t,e,n,r){this.h=+t,this.c=+e,this.l=+n,this.opacity=+r}function Yt(t){if(t instanceof Xt)return new Xt(t.h,t.s,t.l,t.opacity);t instanceof At||(t=Mt(t));var e=t.r/255,n=t.g/255,r=t.b/255,i=($p*r+yp*e-bp*n)/($p+yp-bp),o=r-i,a=(mp*(n-i)-vp*o)/gp,s=Math.sqrt(a*a+o*o)/(mp*i*(1-i)),u=s?Math.atan2(a,o)*ip-120:NaN;return new Xt(u<0?u+360:u,s,i,t.opacity)}function Gt(t,e,n,r){return 1===arguments.length?Yt(t):new Xt(t,e,n,null==r?1:r)}function Xt(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function Kt(t,e,n,r,i){var o=t*t,a=o*t;return((1-3*t+3*o-a)*e+(4-6*o+3*a)*n+(1+3*t+3*o-3*a)*r+a*i)/6}function Jt(t,e){return function(n){return t+n*e}}function Zt(t,e,n){return t=Math.pow(t,n),e=Math.pow(e,n)-t,n=1/n,function(r){return Math.pow(t+r*e,n)}}function Qt(t,e){var n=e-t;return n?Jt(t,n>180||n<-180?n-360*Math.round(n/360):n):Mp(isNaN(t)?e:t)}function te(t){return 1===(t=+t)?ee:function(e,n){return n-e?Zt(e,n,t):Mp(isNaN(e)?n:e)}}function ee(t,e){var n=e-t;return n?Jt(t,n):Mp(isNaN(t)?e:t)}function ne(t){return function(e){var n,r,i=e.length,o=new Array(i),a=new Array(i),s=new Array(i);for(n=0;n<i;++n)r=Dt(e[n]),o[n]=r.r||0,a[n]=r.g||0,s[n]=r.b||0;return o=t(o),a=t(a),s=t(s),r.opacity=1,function(t){return r.r=o(t),r.g=a(t),r.b=s(t),r+""}}}function re(t){return function(){return t}}function ie(t){return function(e){return t(e)+""}}function oe(t){return"none"===t?Hp:(wp||(wp=document.createElement("DIV"),xp=document.documentElement,_p=document.defaultView),wp.style.transform=t,t=_p.getComputedStyle(xp.appendChild(wp),null).getPropertyValue("transform"),xp.removeChild(wp),t=t.slice(7,-1).split(","),Bp(+t[0],+t[1],+t[2],+t[3],+t[4],+t[5]))}function ae(t){return null==t?Hp:(Cp||(Cp=document.createElementNS("http://www.w3.org/2000/svg","g")),Cp.setAttribute("transform",t),(t=Cp.transform.baseVal.consolidate())?(t=t.matrix,Bp(t.a,t.b,t.c,t.d,t.e,t.f)):Hp)}function se(t,e,n,r){function i(t){return t.length?t.pop()+" ":""}function o(t,r,i,o,a,s){if(t!==i||r!==o){var u=a.push("translate(",null,e,null,n);s.push({i:u-4,x:Rp(t,i)},{i:u-2,x:Rp(r,o)})}else(i||o)&&a.push("translate("+i+e+o+n)}function a(t,e,n,o){t!==e?(t-e>180?e+=360:e-t>180&&(t+=360),o.push({i:n.push(i(n)+"rotate(",null,r)-2,x:Rp(t,e)})):e&&n.push(i(n)+"rotate("+e+r)}function s(t,e,n,o){t!==e?o.push({i:n.push(i(n)+"skewX(",null,r)-2,x:Rp(t,e)}):e&&n.push(i(n)+"skewX("+e+r)}function u(t,e,n,r,o,a){if(t!==n||e!==r){var s=o.push(i(o)+"scale(",null,",",null,")");a.push({i:s-4,x:Rp(t,n)},{i:s-2,x:Rp(e,r)})}else 1===n&&1===r||o.push(i(o)+"scale("+n+","+r+")")}return function(e,n){var r=[],i=[];return e=t(e),n=t(n),o(e.translateX,e.translateY,n.translateX,n.translateY,r,i),a(e.rotate,n.rotate,r,i),s(e.skewX,n.skewX,r,i),u(e.scaleX,e.scaleY,n.scaleX,n.scaleY,r,i),e=n=null,function(t){for(var e,n=-1,o=i.length;++n<o;)r[(e=i[n]).i]=e.x(t);return r.join("")}}}function ue(t){return((t=Math.exp(t))+1/t)/2}function le(t){return((t=Math.exp(t))-1/t)/2}function ce(t){return((t=Math.exp(2*t))-1)/(t+1)}function fe(t){return function(e,n){var r=t((e=Pt(e)).h,(n=Pt(n)).h),i=ee(e.s,n.s),o=ee(e.l,n.l),a=ee(e.opacity,n.opacity);return function(t){return e.h=r(t),e.s=i(t),e.l=o(t),e.opacity=a(t),e+""}}}function pe(t,e){var n=ee((t=Ft(t)).l,(e=Ft(e)).l),r=ee(t.a,e.a),i=ee(t.b,e.b),o=ee(t.opacity,e.opacity);return function(e){return t.l=n(e),t.a=r(e),t.b=i(e),t.opacity=o(e),t+""}}function he(t){return function(e,n){var r=t((e=zt(e)).h,(n=zt(n)).h),i=ee(e.c,n.c),o=ee(e.l,n.l),a=ee(e.opacity,n.opacity);return function(t){return e.h=r(t),e.c=i(t),e.l=o(t),e.opacity=a(t),e+""}}}function de(t){return function e(n){function r(e,r){var i=t((e=Gt(e)).h,(r=Gt(r)).h),o=ee(e.s,r.s),a=ee(e.l,r.l),s=ee(e.opacity,r.opacity);return function(t){return e.h=i(t),e.s=o(t),e.l=a(Math.pow(t,n)),e.opacity=s(t),e+""}}return n=+n,r.gamma=e,r}(1)}function ve(){return ch||(hh(ge),ch=ph.now()+fh)}function ge(){ch=0}function me(){this._call=this._time=this._next=null}function ye(t,e,n){var r=new me;return r.restart(t,e,n),r}function be(){ve(),++oh;for(var t,e=kp;e;)(t=ch-e._time)>=0&&e._call.call(null,t),e=e._next;--oh}function $e(){ch=(lh=ph.now())+fh,oh=ah=0;try{be()}finally{oh=0,xe(),ch=0}}function we(){var t=ph.now(),e=t-lh;e>uh&&(fh-=e,lh=t)}function xe(){for(var t,e,n=kp,r=1/0;n;)n._call?(r>n._time&&(r=n._time),t=n,n=n._next):(e=n._next,n._next=null,n=t?t._next=e:kp=e);Sp=t,_e(r)}function _e(t){if(!oh){ah&&(ah=clearTimeout(ah));var e=t-ch;e>24?(t<1/0&&(ah=setTimeout($e,e)),sh&&(sh=clearInterval(sh))):(sh||(lh=ch,sh=setInterval(we,uh)),oh=1,hh($e))}}function Ce(t,e){var n=t.__transition;if(!n||!(n=n[e])||n.state>yh)throw new Error("too late");return n}function ke(t,e){var n=t.__transition;if(!n||!(n=n[e])||n.state>$h)throw new Error("too late");return n}function Se(t,e){var n=t.__transition;if(!n||!(n=n[e]))throw new Error("too late");return n}function Te(t,e,n){function r(t){n.state=bh,n.timer.restart(i,n.delay,n.time),n.delay<=t&&i(t-n.delay)}function i(r){var l,c,f,p;if(n.state!==bh)return a();for(l in u)if(p=u[l],p.name===n.name){if(p.state===wh)return dh(i);p.state===xh?(p.state=Ch,p.timer.stop(),p.on.call("interrupt",t,t.__data__,p.index,p.group),delete u[l]):+l<e&&(p.state=Ch,p.timer.stop(),delete u[l])}if(dh(function(){n.state===wh&&(n.state=xh,n.timer.restart(o,n.delay,n.time),o(r))}),n.state=$h,n.on.call("start",t,t.__data__,n.index,n.group),n.state===$h){for(n.state=wh,s=new Array(f=n.tween.length),l=0,c=-1;l<f;++l)(p=n.tween[l].value.call(t,t.__data__,n.index,n.group))&&(s[++c]=p);s.length=c+1}}function o(e){for(var r=e<n.duration?n.ease.call(null,e/n.duration):(n.timer.restart(a),n.state=_h,1),i=-1,o=s.length;++i<o;)s[i].call(null,r);n.state===_h&&(n.on.call("end",t,t.__data__,n.index,n.group),a())}function a(){n.state=Ch,n.timer.stop(),delete u[e];for(var r in u)return;delete t.__transition}var s,u=t.__transition;u[e]=n,n.timer=ye(r,0,n.time)}function Ee(t,e){var n,r;return function(){var i=ke(this,t),o=i.tween;if(o!==n){r=n=o;for(var a=0,s=r.length;a<s;++a)if(r[a].name===e){r=r.slice(),r.splice(a,1);break}}i.tween=r}}function Me(t,e,n){var r,i;if("function"!=typeof n)throw new Error;return function(){var o=ke(this,t),a=o.tween;if(a!==r){i=(r=a).slice();for(var s={name:e,value:n},u=0,l=i.length;u<l;++u)if(i[u].name===e){i[u]=s;break}u===l&&i.push(s)}o.tween=i}}function De(t,e,n){var r=t._id;return t.each(function(){var t=ke(this,r);(t.value||(t.value={}))[e]=n.apply(this,arguments)}),function(t){return Se(t,r).value[e]}}function Ae(t){return function(){this.removeAttribute(t)}}function Oe(t){return function(){this.removeAttributeNS(t.space,t.local)}}function Ne(t,e,n){var r,i;return function(){var o=this.getAttribute(t);return o===n?null:o===r?i:i=e(r=o,n)}}function Pe(t,e,n){var r,i;return function(){var o=this.getAttributeNS(t.space,t.local);return o===n?null:o===r?i:i=e(r=o,n)}}function Re(t,e,n){var r,i,o;return function(){var a,s=n(this);return null==s?void this.removeAttribute(t):(a=this.getAttribute(t),a===s?null:a===r&&s===i?o:o=e(r=a,i=s))}}function Ie(t,e,n){var r,i,o;return function(){var a,s=n(this);return null==s?void this.removeAttributeNS(t.space,t.local):(a=this.getAttributeNS(t.space,t.local),a===s?null:a===r&&s===i?o:o=e(r=a,i=s))}}function je(t,e){function n(){var n=this,r=e.apply(n,arguments);return r&&function(e){n.setAttributeNS(t.space,t.local,r(e))}}return n._value=e,n}function Fe(t,e){function n(){var n=this,r=e.apply(n,arguments);return r&&function(e){n.setAttribute(t,r(e))}}return n._value=e,n}function qe(t,e){return function(){Ce(this,t).delay=+e.apply(this,arguments)}}function Ve(t,e){return e=+e,function(){Ce(this,t).delay=e}}function Ue(t,e){return function(){ke(this,t).duration=+e.apply(this,arguments)}}function Le(t,e){return e=+e,function(){ke(this,t).duration=e}}function He(t,e){if("function"!=typeof e)throw new Error;return function(){ke(this,t).ease=e}}function Be(t){return(t+"").trim().split(/^|\s+/).every(function(t){var e=t.indexOf(".");return e>=0&&(t=t.slice(0,e)),!t||"start"===t})}function ze(t,e,n){var r,i,o=Be(e)?Ce:ke;return function(){var a=o(this,t),s=a.on;s!==r&&(i=(r=s).copy()).on(e,n),a.on=i}}function We(t){return function(){var e=this.parentNode;for(var n in this.__transition)if(+n!==t)return;e&&e.removeChild(this)}}function Ye(t,e){var n,r,i;return function(){var o=z(this,t),a=(this.style.removeProperty(t),z(this,t));return o===a?null:o===n&&a===r?i:i=e(n=o,r=a)}}function Ge(t){return function(){this.style.removeProperty(t)}}function Xe(t,e,n){var r,i;return function(){var o=z(this,t);return o===n?null:o===r?i:i=e(r=o,n)}}function Ke(t,e,n){var r,i,o;return function(){var a=z(this,t),s=n(this);return null==s&&(this.style.removeProperty(t),s=z(this,t)),a===s?null:a===r&&s===i?o:o=e(r=a,i=s)}}function Je(t,e,n){function r(){var r=this,i=e.apply(r,arguments);return i&&function(e){r.style.setProperty(t,i(e),n)}}return r._value=e,r}function Ze(t){return function(){this.textContent=t}}function Qe(t){return function(){var e=t(this);this.textContent=null==e?"":e}}function tn(t,e,n,r){this._groups=t,this._parents=e,this._name=n,this._id=r}function en(t){return mt().transition(t)}function nn(){return++Yh}function rn(t){return+t}function on(t){return t*t}function an(t){return t*(2-t)}function sn(t){return((t*=2)<=1?t*t:--t*(2-t)+1)/2}function un(t){return t*t*t}function ln(t){return--t*t*t+1}function cn(t){return((t*=2)<=1?t*t*t:(t-=2)*t*t+2)/2}function fn(t){return 1-Math.cos(t*td)}function pn(t){return Math.sin(t*td)}function hn(t){return(1-Math.cos(Qh*t))/2}function dn(t){return Math.pow(2,10*t-10)}function vn(t){return 1-Math.pow(2,-10*t)}function gn(t){return((t*=2)<=1?Math.pow(2,10*t-10):2-Math.pow(2,10-10*t))/2}function mn(t){return 1-Math.sqrt(1-t*t)}function yn(t){return Math.sqrt(1- --t*t)}function bn(t){return((t*=2)<=1?1-Math.sqrt(1-t*t):Math.sqrt(1-(t-=2)*t)+1)/2}function $n(t){return 1-wn(1-t)}function wn(t){return(t=+t)<ed?cd*t*t:t<rd?cd*(t-=nd)*t+id:t<ad?cd*(t-=od)*t+sd:cd*(t-=ud)*t+ld}function xn(t){return((t*=2)<=1?1-wn(1-t):wn(t-1)+1)/2}function _n(t,e){for(var n;!(n=t.__transition)||!(n=n[e]);)if(!(t=t.parentNode))return wd.time=ve(),wd;return n}function Cn(){t.event.stopImmediatePropagation()}function kn(t){return{type:t}}function Sn(){return!t.event.button}function Tn(){var t=this.ownerSVGElement||this;return[[0,0],[t.width.baseVal.value,t.height.baseVal.value]]}function En(t){for(;!t.__brush;)if(!(t=t.parentNode))return;return t.__brush}function Mn(t){return t[0][0]===t[1][0]||t[0][1]===t[1][1]}function Dn(t){var e=t.__brush;return e?e.dim.output(e.selection):null}function An(){return Nn(Od)}function On(){return Nn(Nd)}function Nn(e){function n(t){var n=t.property("__brush",s).selectAll(".overlay").data([kn("overlay")]);n.enter().append("rect").attr("class","overlay").attr("pointer-events","all").attr("cursor",Rd.overlay).merge(n).each(function(){var t=En(this).extent;Pf(this).attr("x",t[0][0]).attr("y",t[0][1]).attr("width",t[1][0]-t[0][0]).attr("height",t[1][1]-t[0][1])}),t.selectAll(".selection").data([kn("selection")]).enter().append("rect").attr("class","selection").attr("cursor",Rd.selection).attr("fill","#777").attr("fill-opacity",.3).attr("stroke","#fff").attr("shape-rendering","crispEdges");var i=t.selectAll(".handle").data(e.handles,function(t){return t.type});i.exit().remove(),i.enter().append("rect").attr("class",function(t){return"handle handle--"+t.type}).attr("cursor",function(t){return Rd[t.type]}),t.each(r).attr("fill","none").attr("pointer-events","all").style("-webkit-tap-highlight-color","rgba(0,0,0,0)").on("mousedown.brush touchstart.brush",a)}function r(){var t=Pf(this),e=En(this).selection;e?(t.selectAll(".selection").style("display",null).attr("x",e[0][0]).attr("y",e[0][1]).attr("width",e[1][0]-e[0][0]).attr("height",e[1][1]-e[0][1]),t.selectAll(".handle").style("display",null).attr("x",function(t){return"e"===t.type[t.type.length-1]?e[1][0]-p/2:e[0][0]-p/2}).attr("y",function(t){return"s"===t.type[0]?e[1][1]-p/2:e[0][1]-p/2}).attr("width",function(t){return"n"===t.type||"s"===t.type?e[1][0]-e[0][0]+p:p}).attr("height",function(t){return"e"===t.type||"w"===t.type?e[1][1]-e[0][1]+p:p})):t.selectAll(".selection,.handle").style("display","none").attr("x",null).attr("y",null).attr("width",null).attr("height",null)}function i(t,e){return t.__brush.emitter||new o(t,e)}function o(t,e){this.that=t,this.args=e,this.state=t.__brush,this.active=0}function a(){function n(){var t=Jc(C);!I||x||_||(Math.abs(t[0]-F[0])>Math.abs(t[1]-F[1])?_=!0:x=!0),F=t,w=!0,Td(),o()}function o(){var t;switch(b=F[0]-j[0],$=F[1]-j[1],S){case Md:case Ed:T&&(b=Math.max(O-f,Math.min(P-v,b)),p=f+b,g=v+b),E&&($=Math.max(N-h,Math.min(R-m,$)),d=h+$,y=m+$);break;case Dd:T<0?(b=Math.max(O-f,Math.min(P-f,b)),p=f+b,g=v):T>0&&(b=Math.max(O-v,Math.min(P-v,b)),p=f,g=v+b),E<0?($=Math.max(N-h,Math.min(R-h,$)),d=h+$,y=m):E>0&&($=Math.max(N-m,Math.min(R-m,$)),d=h,y=m+$);break;case Ad:T&&(p=Math.max(O,Math.min(P,f-b*T)),g=Math.max(O,Math.min(P,v+b*T))),E&&(d=Math.max(N,Math.min(R,h-$*E)),y=Math.max(N,Math.min(R,m+$*E)))}g<p&&(T*=-1,t=f,f=v,v=t,t=p,p=g,g=t,k in Id&&U.attr("cursor",Rd[k=Id[k]])),y<d&&(E*=-1,t=h,h=m,m=t,t=d,d=y,y=t,k in jd&&U.attr("cursor",Rd[k=jd[k]])),M.selection&&(A=M.selection),x&&(p=A[0][0],g=A[1][0]),_&&(d=A[0][1],y=A[1][1]),A[0][0]===p&&A[0][1]===d&&A[1][0]===g&&A[1][1]===y||(M.selection=[[p,d],[g,y]],r.call(C),q.brush())}function a(){if(Cn(),t.event.touches){if(t.event.touches.length)return;u&&clearTimeout(u),u=setTimeout(function(){u=null},500),V.on("touchmove.brush touchend.brush touchcancel.brush",null)}else bt(t.event.view,w),L.on("keydown.brush keyup.brush mousemove.brush mouseup.brush",null);V.attr("pointer-events","all"),U.attr("cursor",Rd.overlay),M.selection&&(A=M.selection),Mn(A)&&(M.selection=null,r.call(C)),q.end()}function s(){switch(t.event.keyCode){case 16:I=T&&E;break;case 18:S===Dd&&(T&&(v=g-b*T,f=p+b*T),E&&(m=y-$*E,h=d+$*E),S=Ad,o());break;case 32:S!==Dd&&S!==Ad||(T<0?v=g-b:T>0&&(f=p-b),E<0?m=y-$:E>0&&(h=d-$),S=Md,U.attr("cursor",Rd.selection),o());break;default:return}Td()}function l(){switch(t.event.keyCode){case 16:I&&(x=_=I=!1,o());break;case 18:S===Ad&&(T<0?v=g:T>0&&(f=p),E<0?m=y:E>0&&(h=d),S=Dd,o());break;case 32:S===Md&&(t.event.altKey?(T&&(v=g-b*T,f=p+b*T),E&&(m=y-$*E,h=d+$*E),S=Ad):(T<0?v=g:T>0&&(f=p),E<0?m=y:E>0&&(h=d),S=Dd),U.attr("cursor",Rd[k]),o());break;default:return}Td()}if(t.event.touches){if(t.event.changedTouches.length<t.event.touches.length)return Td()}else if(u)return;if(c.apply(this,arguments)){var f,p,h,d,v,g,m,y,b,$,w,x,_,C=this,k=t.event.target.__data__.type,S="selection"===(t.event.metaKey?k="overlay":k)?Ed:t.event.altKey?Ad:Dd,T=e===Nd?null:Fd[k],E=e===Od?null:qd[k],M=En(C),D=M.extent,A=M.selection,O=D[0][0],N=D[0][1],P=D[1][0],R=D[1][1],I=T&&E&&t.event.shiftKey,j=Jc(C),F=j,q=i(C,arguments).beforestart();"overlay"===k?M.selection=A=[[f=e===Nd?O:j[0],h=e===Od?N:j[1]],[v=e===Nd?P:f,m=e===Od?R:h]]:(f=A[0][0],h=A[0][1],v=A[1][0],m=A[1][1]),p=f,d=h,g=v,y=m;var V=Pf(C).attr("pointer-events","none"),U=V.selectAll(".overlay").attr("cursor",Rd[k]);if(t.event.touches)V.on("touchmove.brush",n,!0).on("touchend.brush touchcancel.brush",a,!0);else{var L=Pf(t.event.view).on("keydown.brush",s,!0).on("keyup.brush",l,!0).on("mousemove.brush",n,!0).on("mouseup.brush",a,!0);qf(t.event.view)}Cn(),Sh(C),r.call(C),q.start()}}function s(){var t=this.__brush||{selection:null};return t.extent=l.apply(this,arguments),t.dim=e,t}var u,l=Tn,c=Sn,f=v(n,"start","brush","end"),p=6;return n.move=function(t,n){t.selection?t.on("start.brush",function(){i(this,arguments).beforestart().start()}).on("interrupt.brush end.brush",function(){i(this,arguments).end()}).tween("brush",function(){function t(t){a.selection=1===t&&Mn(l)?null:c(t),r.call(o),s.brush()}var o=this,a=o.__brush,s=i(o,arguments),u=a.selection,l=e.input("function"==typeof n?n.apply(this,arguments):n,a.extent),c=Vp(u,l);return u&&l?t:t(1)}):t.each(function(){var t=this,o=arguments,a=t.__brush,s=e.input("function"==typeof n?n.apply(t,o):n,a.extent),u=i(t,o).beforestart();Sh(t),a.selection=null==s||Mn(s)?null:s,r.call(t),u.start().brush().end()})},o.prototype={beforestart:function(){return 1===++this.active&&(this.state.emitter=this,this.starting=!0),this},start:function(){return this.starting&&(this.starting=!1,this.emit("start")),this},brush:function(){return this.emit("brush"),this},end:function(){return 0===--this.active&&(delete this.state.emitter,this.emit("end")),this},emit:function(t){M(new Sd(n,t,e.output(this.state.selection)),f.apply,f,[t,this.that,this.args])}},n.extent=function(t){return arguments.length?(l="function"==typeof t?t:kd([[+t[0][0],+t[0][1]],[+t[1][0],+t[1][1]]]),n):l},n.filter=function(t){return arguments.length?(c="function"==typeof t?t:kd(!!t),n):c},n.handleSize=function(t){return arguments.length?(p=+t,n):p},n.on=function(){var t=f.on.apply(f,arguments);return t===f?n:t},n}function Pn(t){return function(e,n){return t(e.source.value+e.target.value,n.source.value+n.target.value)}}function Rn(){this._x0=this._y0=this._x1=this._y1=null,this._=""}function In(){return new Rn}function jn(t){return t.source}function Fn(t){return t.target}function qn(t){return t.radius}function Vn(t){return t.startAngle}function Un(t){return t.endAngle}function Ln(){}function Hn(t,e){var n=new Ln;if(t instanceof Ln)t.each(function(t,e){n.set(e,t)});else if(Array.isArray(t)){var r,i=-1,o=t.length;if(null==e)for(;++i<o;)n.set(i,t[i]);else for(;++i<o;)n.set(e(r=t[i],i,t),r)}else if(t)for(var a in t)n.set(a,t[a]);return n}function Bn(){return{}}function zn(t,e,n){t[e]=n}function Wn(){return Hn()}function Yn(t,e,n){t.set(e,n)}function Gn(){}function Xn(t,e){var n=new Gn;if(t instanceof Gn)t.each(function(t){n.add(t)});else if(t){var r=-1,i=t.length;if(null==e)for(;++r<i;)n.add(t[r]);else for(;++r<i;)n.add(e(t[r],r,t))}return n}function Kn(t){return new Function("d","return {"+t.map(function(t,e){return JSON.stringify(t)+": d["+e+"]"}).join(",")+"}")}function Jn(t,e){var n=Kn(t);return function(r,i){return e(n(r),i,t)}}function Zn(t){var e=Object.create(null),n=[];return t.forEach(function(t){for(var r in t)r in e||n.push(e[r]=r)}),n}function Qn(t,e,n,r){if(isNaN(e)||isNaN(n))return t;var i,o,a,s,u,l,c,f,p,h=t._root,d={data:r},v=t._x0,g=t._y0,m=t._x1,y=t._y1;if(!h)return t._root=d,t;for(;h.length;)if((l=e>=(o=(v+m)/2))?v=o:m=o,(c=n>=(a=(g+y)/2))?g=a:y=a,i=h,!(h=h[f=c<<1|l]))return i[f]=d,t;if(s=+t._x.call(null,h.data),u=+t._y.call(null,h.data),e===s&&n===u)return d.next=h,i?i[f]=d:t._root=d,t;do i=i?i[f]=new Array(4):t._root=new Array(4),(l=e>=(o=(v+m)/2))?v=o:m=o,(c=n>=(a=(g+y)/2))?g=a:y=a;while((f=c<<1|l)===(p=(u>=a)<<1|s>=o));return i[p]=h,i[f]=d,t}function tr(t){var e,n,r,i,o=t.length,a=new Array(o),s=new Array(o),u=1/0,l=1/0,c=-(1/0),f=-(1/0);for(n=0;n<o;++n)isNaN(r=+this._x.call(null,e=t[n]))||isNaN(i=+this._y.call(null,e))||(a[n]=r,s[n]=i,r<u&&(u=r),r>c&&(c=r),i<l&&(l=i),i>f&&(f=i));for(c<u&&(u=this._x0,c=this._x1),f<l&&(l=this._y0,f=this._y1),this.cover(u,l).cover(c,f),n=0;n<o;++n)Qn(this,a[n],s[n],t[n]);return this}function er(t){for(var e=0,n=t.length;e<n;++e)this.remove(t[e]);return this}function nr(t){return t[0]}function rr(t){return t[1]}
function ir(t,e,n){var r=new or(null==e?nr:e,null==n?rr:n,NaN,NaN,NaN,NaN);return null==t?r:r.addAll(t)}function or(t,e,n,r,i,o){this._x=t,this._y=e,this._x0=n,this._y0=r,this._x1=i,this._y1=o,this._root=void 0}function ar(t){for(var e={data:t.data},n=e;t=t.next;)n=n.next={data:t.data};return e}function sr(t){return t.x+t.vx}function ur(t){return t.y+t.vy}function lr(t){return t.index}function cr(t,e){var n=t.get(e);if(!n)throw new Error("missing: "+e);return n}function fr(t){return t.x}function pr(t){return t.y}function hr(t){return new dr(t)}function dr(t){if(!(e=Zv.exec(t)))throw new Error("invalid format: "+t);var e,n=e[1]||" ",r=e[2]||">",i=e[3]||"-",o=e[4]||"",a=!!e[5],s=e[6]&&+e[6],u=!!e[7],l=e[8]&&+e[8].slice(1),c=e[9]||"";"n"===c?(u=!0,c="g"):Jv[c]||(c=""),(a||"0"===n&&"="===r)&&(a=!0,n="0",r="="),this.fill=n,this.align=r,this.sign=i,this.symbol=o,this.zero=a,this.width=s,this.comma=u,this.precision=l,this.type=c}function vr(e){return Qv=ng(e),t.format=Qv.format,t.formatPrefix=Qv.formatPrefix,Qv}function gr(){this.reset()}function mr(t,e,n){var r=t.s=e+n,i=r-e,o=r-i;t.t=e-o+(n-i)}function yr(t){return t>1?0:t<-1?Ug:Math.acos(t)}function br(t){return t>1?Lg:t<-1?-Lg:Math.asin(t)}function $r(t){return(t=em(t/2))*t}function wr(){}function xr(t,e){t&&am.hasOwnProperty(t.type)&&am[t.type](t,e)}function _r(t,e,n){var r,i=-1,o=t.length-n;for(e.lineStart();++i<o;)r=t[i],e.point(r[0],r[1],r[2]);e.lineEnd()}function Cr(t,e){var n=-1,r=t.length;for(e.polygonStart();++n<r;)_r(t[n],e,1);e.polygonEnd()}function kr(){cm.point=Tr}function Sr(){Er(sg,ug)}function Tr(t,e){cm.point=Er,sg=t,ug=e,t*=Wg,e*=Wg,lg=t,cg=Kg(e=e/2+Hg),fg=em(e)}function Er(t,e){t*=Wg,e*=Wg,e=e/2+Hg;var n=t-lg,r=n>=0?1:-1,i=r*n,o=Kg(e),a=em(e),s=fg*a,u=cg*o+s*Kg(i),l=s*r*em(i);um.add(Xg(l,u)),lg=t,cg=o,fg=a}function Mr(t){return[Xg(t[1],t[0]),br(t[2])]}function Dr(t){var e=t[0],n=t[1],r=Kg(n);return[r*Kg(e),r*em(e),em(n)]}function Ar(t,e){return t[0]*e[0]+t[1]*e[1]+t[2]*e[2]}function Or(t,e){return[t[1]*e[2]-t[2]*e[1],t[2]*e[0]-t[0]*e[2],t[0]*e[1]-t[1]*e[0]]}function Nr(t,e){t[0]+=e[0],t[1]+=e[1],t[2]+=e[2]}function Pr(t,e){return[t[0]*e,t[1]*e,t[2]*e]}function Rr(t){var e=rm(t[0]*t[0]+t[1]*t[1]+t[2]*t[2]);t[0]/=e,t[1]/=e,t[2]/=e}function Ir(t,e){$g.push(wg=[pg=t,dg=t]),e<hg&&(hg=e),e>vg&&(vg=e)}function jr(t,e){var n=Dr([t*Wg,e*Wg]);if(bg){var r=Or(bg,n),i=[r[1],-r[0],0],o=Or(i,r);Rr(o),o=Mr(o);var a,s=t-gg,u=s>0?1:-1,l=o[0]*zg*u,c=Yg(s)>180;c^(u*gg<l&&l<u*t)?(a=o[1]*zg,a>vg&&(vg=a)):(l=(l+360)%360-180,c^(u*gg<l&&l<u*t)?(a=-o[1]*zg,a<hg&&(hg=a)):(e<hg&&(hg=e),e>vg&&(vg=e))),c?t<gg?Hr(pg,t)>Hr(pg,dg)&&(dg=t):Hr(t,dg)>Hr(pg,dg)&&(pg=t):dg>=pg?(t<pg&&(pg=t),t>dg&&(dg=t)):t>gg?Hr(pg,t)>Hr(pg,dg)&&(dg=t):Hr(t,dg)>Hr(pg,dg)&&(pg=t)}else $g.push(wg=[pg=t,dg=t]);e<hg&&(hg=e),e>vg&&(vg=e),bg=n,gg=t}function Fr(){hm.point=jr}function qr(){wg[0]=pg,wg[1]=dg,hm.point=Ir,bg=null}function Vr(t,e){if(bg){var n=t-gg;pm.add(Yg(n)>180?n+(n>0?360:-360):n)}else mg=t,yg=e;cm.point(t,e),jr(t,e)}function Ur(){cm.lineStart()}function Lr(){Vr(mg,yg),cm.lineEnd(),Yg(pm)>qg&&(pg=-(dg=180)),wg[0]=pg,wg[1]=dg,bg=null}function Hr(t,e){return(e-=t)<0?e+360:e}function Br(t,e){return t[0]-e[0]}function zr(t,e){return t[0]<=t[1]?t[0]<=e&&e<=t[1]:e<t[0]||t[1]<e}function Wr(t,e){t*=Wg,e*=Wg;var n=Kg(e);Yr(n*Kg(t),n*em(t),em(e))}function Yr(t,e,n){++xg,Cg+=(t-Cg)/xg,kg+=(e-kg)/xg,Sg+=(n-Sg)/xg}function Gr(){vm.point=Xr}function Xr(t,e){t*=Wg,e*=Wg;var n=Kg(e);Rg=n*Kg(t),Ig=n*em(t),jg=em(e),vm.point=Kr,Yr(Rg,Ig,jg)}function Kr(t,e){t*=Wg,e*=Wg;var n=Kg(e),r=n*Kg(t),i=n*em(t),o=em(e),a=Xg(rm((a=Ig*o-jg*i)*a+(a=jg*r-Rg*o)*a+(a=Rg*i-Ig*r)*a),Rg*r+Ig*i+jg*o);_g+=a,Tg+=a*(Rg+(Rg=r)),Eg+=a*(Ig+(Ig=i)),Mg+=a*(jg+(jg=o)),Yr(Rg,Ig,jg)}function Jr(){vm.point=Wr}function Zr(){vm.point=ti}function Qr(){ei(Ng,Pg),vm.point=Wr}function ti(t,e){Ng=t,Pg=e,t*=Wg,e*=Wg,vm.point=ei;var n=Kg(e);Rg=n*Kg(t),Ig=n*em(t),jg=em(e),Yr(Rg,Ig,jg)}function ei(t,e){t*=Wg,e*=Wg;var n=Kg(e),r=n*Kg(t),i=n*em(t),o=em(e),a=Ig*o-jg*i,s=jg*r-Rg*o,u=Rg*i-Ig*r,l=rm(a*a+s*s+u*u),c=br(l),f=l&&-c/l;Dg+=f*a,Ag+=f*s,Og+=f*u,_g+=c,Tg+=c*(Rg+(Rg=r)),Eg+=c*(Ig+(Ig=i)),Mg+=c*(jg+(jg=o)),Yr(Rg,Ig,jg)}function ni(t,e){return[t>Ug?t-Bg:t<-Ug?t+Bg:t,e]}function ri(t,e,n){return(t%=Bg)?e||n?ym(oi(t),ai(e,n)):oi(t):e||n?ai(e,n):ni}function ii(t){return function(e,n){return e+=t,[e>Ug?e-Bg:e<-Ug?e+Bg:e,n]}}function oi(t){var e=ii(t);return e.invert=ii(-t),e}function ai(t,e){function n(t,e){var n=Kg(e),s=Kg(t)*n,u=em(t)*n,l=em(e),c=l*r+s*i;return[Xg(u*o-c*a,s*r-l*i),br(c*o+u*a)]}var r=Kg(t),i=em(t),o=Kg(e),a=em(e);return n.invert=function(t,e){var n=Kg(e),s=Kg(t)*n,u=em(t)*n,l=em(e),c=l*o-u*a;return[Xg(u*o+l*a,s*r+c*i),br(c*r-s*i)]},n}function si(t,e,n,r,i,o){if(n){var a=Kg(e),s=em(e),u=r*n;null==i?(i=e+r*Bg,o=e-u/2):(i=ui(a,i),o=ui(a,o),(r>0?i<o:i>o)&&(i+=r*Bg));for(var l,c=i;r>0?c>o:c<o;c-=u)l=Mr([a,-s*Kg(c),-s*em(c)]),t.point(l[0],l[1])}}function ui(t,e){e=Dr(e),e[0]-=t,Rr(e);var n=yr(-e[1]);return((-e[2]<0?-n:n)+Bg-qg)%Bg}function li(t,e,n,r){this.x=t,this.z=e,this.o=n,this.e=r,this.v=!1,this.n=this.p=null}function ci(t){if(e=t.length){for(var e,n,r=0,i=t[0];++r<e;)i.n=n=t[r],n.p=i,i=n;i.n=n=t[0],n.p=i}}function fi(t,e,n,r){function i(i,o){return t<=i&&i<=n&&e<=o&&o<=r}function o(i,o,s,l){var c=0,f=0;if(null==i||(c=a(i,s))!==(f=a(o,s))||u(i,o)<0^s>0){do l.point(0===c||3===c?t:n,c>1?r:e);while((c=(c+s+4)%4)!==f)}else l.point(o[0],o[1])}function a(r,i){return Yg(r[0]-t)<qg?i>0?0:3:Yg(r[0]-n)<qg?i>0?2:1:Yg(r[1]-e)<qg?i>0?1:0:i>0?3:2}function s(t,e){return u(t.x,e.x)}function u(t,e){var n=a(t,1),r=a(e,1);return n!==r?n-r:0===n?e[1]-t[1]:1===n?t[0]-e[0]:2===n?t[1]-e[1]:e[0]-t[0]}return function(a){function u(t,e){i(t,e)&&S.point(t,e)}function l(){for(var e=0,n=0,i=g.length;n<i;++n)for(var o,a,s=g[n],u=1,l=s.length,c=s[0],f=c[0],p=c[1];u<l;++u)o=f,a=p,c=s[u],f=c[0],p=c[1],a<=r?p>r&&(f-o)*(r-a)>(p-a)*(t-o)&&++e:p<=r&&(f-o)*(r-a)<(p-a)*(t-o)&&--e;return e}function c(){S=T,v=[],g=[],k=!0}function f(){var t=l(),e=k&&t,n=(v=wc(v)).length;(e||n)&&(a.polygonStart(),e&&(a.lineStart(),o(null,null,1,a),a.lineEnd()),n&&Rm(v,s,t,o,a),a.polygonEnd()),S=a,v=g=m=null}function p(){E.point=d,g&&g.push(m=[]),C=!0,_=!1,w=x=NaN}function h(){v&&(d(y,b),$&&_&&T.rejoin(),v.push(T.result())),E.point=u,_&&S.lineEnd()}function d(o,a){var s=i(o,a);if(g&&m.push([o,a]),C)y=o,b=a,$=s,C=!1,s&&(S.lineStart(),S.point(o,a));else if(s&&_)S.point(o,a);else{var u=[w=Math.max(jm,Math.min(Im,w)),x=Math.max(jm,Math.min(Im,x))],l=[o=Math.max(jm,Math.min(Im,o)),a=Math.max(jm,Math.min(Im,a))];Nm(u,l,t,e,n,r)?(_||(S.lineStart(),S.point(u[0],u[1])),S.point(l[0],l[1]),s||S.lineEnd(),k=!1):s&&(S.lineStart(),S.point(o,a),k=!1)}w=o,x=a,_=s}var v,g,m,y,b,$,w,x,_,C,k,S=a,T=Om(),E={point:u,lineStart:p,lineEnd:h,polygonStart:c,polygonEnd:f};return E}}function pi(){Lm.point=di,Lm.lineEnd=hi}function hi(){Lm.point=Lm.lineEnd=wr}function di(t,e){t*=Wg,e*=Wg,bm=t,$m=em(e),wm=Kg(e),Lm.point=vi}function vi(t,e){t*=Wg,e*=Wg;var n=em(e),r=Kg(e),i=Yg(t-bm),o=Kg(i),a=em(i),s=r*a,u=wm*n-$m*r*o,l=$m*n+wm*r*o;Um.add(Xg(rm(s*s+u*u),l)),bm=t,$m=n,wm=r}function gi(t,e){return!(!t||!Gm.hasOwnProperty(t.type))&&Gm[t.type](t,e)}function mi(t,e){return 0===Wm(t,e)}function yi(t,e){var n=Wm(t[0],t[1]),r=Wm(t[0],e),i=Wm(e,t[1]);return r+i<=n+qg}function bi(t,e){return!!Vm(t.map($i),wi(e))}function $i(t){return t=t.map(wi),t.pop(),t}function wi(t){return[t[0]*Wg,t[1]*Wg]}function xi(t,e,n){var r=uc(t,e-qg,n).concat(e);return function(t){return r.map(function(e){return[t,e]})}}function _i(t,e,n){var r=uc(t,e-qg,n).concat(e);return function(t){return r.map(function(e){return[e,t]})}}function Ci(){function t(){return{type:"MultiLineString",coordinates:e()}}function e(){return uc(Jg(o/g)*g,i,g).map(p).concat(uc(Jg(l/m)*m,u,m).map(h)).concat(uc(Jg(r/d)*d,n,d).filter(function(t){return Yg(t%g)>qg}).map(c)).concat(uc(Jg(s/v)*v,a,v).filter(function(t){return Yg(t%m)>qg}).map(f))}var n,r,i,o,a,s,u,l,c,f,p,h,d=10,v=d,g=90,m=360,y=2.5;return t.lines=function(){return e().map(function(t){return{type:"LineString",coordinates:t}})},t.outline=function(){return{type:"Polygon",coordinates:[p(o).concat(h(u).slice(1),p(i).reverse().slice(1),h(l).reverse().slice(1))]}},t.extent=function(e){return arguments.length?t.extentMajor(e).extentMinor(e):t.extentMinor()},t.extentMajor=function(e){return arguments.length?(o=+e[0][0],i=+e[1][0],l=+e[0][1],u=+e[1][1],o>i&&(e=o,o=i,i=e),l>u&&(e=l,l=u,u=e),t.precision(y)):[[o,l],[i,u]]},t.extentMinor=function(e){return arguments.length?(r=+e[0][0],n=+e[1][0],s=+e[0][1],a=+e[1][1],r>n&&(e=r,r=n,n=e),s>a&&(e=s,s=a,a=e),t.precision(y)):[[r,s],[n,a]]},t.step=function(e){return arguments.length?t.stepMajor(e).stepMinor(e):t.stepMinor()},t.stepMajor=function(e){return arguments.length?(g=+e[0],m=+e[1],t):[g,m]},t.stepMinor=function(e){return arguments.length?(d=+e[0],v=+e[1],t):[d,v]},t.precision=function(e){return arguments.length?(y=+e,c=xi(s,a,90),f=_i(r,n,y),p=xi(l,u,90),h=_i(o,i,y),t):y},t.extentMajor([[-180,-90+qg],[180,90-qg]]).extentMinor([[-180,-80-qg],[180,80+qg]])}function ki(){return Ci()()}function Si(){ty.point=Ti}function Ti(t,e){ty.point=Ei,xm=Cm=t,_m=km=e}function Ei(t,e){Qm.add(km*t-Cm*e),Cm=t,km=e}function Mi(){Ei(xm,_m)}function Di(t,e){t<ey&&(ey=t),t>ry&&(ry=t),e<ny&&(ny=e),e>iy&&(iy=e)}function Ai(t,e){ay+=t,sy+=e,++uy}function Oi(){vy.point=Ni}function Ni(t,e){vy.point=Pi,Ai(Em=t,Mm=e)}function Pi(t,e){var n=t-Em,r=e-Mm,i=rm(n*n+r*r);ly+=i*(Em+t)/2,cy+=i*(Mm+e)/2,fy+=i,Ai(Em=t,Mm=e)}function Ri(){vy.point=Ai}function Ii(){vy.point=Fi}function ji(){qi(Sm,Tm)}function Fi(t,e){vy.point=qi,Ai(Sm=Em=t,Tm=Mm=e)}function qi(t,e){var n=t-Em,r=e-Mm,i=rm(n*n+r*r);ly+=i*(Em+t)/2,cy+=i*(Mm+e)/2,fy+=i,i=Mm*t-Em*e,py+=i*(Em+t),hy+=i*(Mm+e),dy+=3*i,Ai(Em=t,Mm=e)}function Vi(t){this._context=t}function Ui(t,e){xy.point=Li,my=by=t,yy=$y=e}function Li(t,e){by-=t,$y-=e,wy.add(rm(by*by+$y*$y)),by=t,$y=e}function Hi(){this._string=[]}function Bi(t){return"m0,"+t+"a"+t+","+t+" 0 1,1 0,"+-2*t+"a"+t+","+t+" 0 1,1 0,"+2*t+"z"}function zi(t){return t.length>1}function Wi(t,e){return((t=t.x)[0]<0?t[1]-Lg-qg:Lg-t[1])-((e=e.x)[0]<0?e[1]-Lg-qg:Lg-e[1])}function Yi(t){var e,n=NaN,r=NaN,i=NaN;return{lineStart:function(){t.lineStart(),e=1},point:function(o,a){var s=o>0?Ug:-Ug,u=Yg(o-n);Yg(u-Ug)<qg?(t.point(n,r=(r+a)/2>0?Lg:-Lg),t.point(i,r),t.lineEnd(),t.lineStart(),t.point(s,r),t.point(o,r),e=0):i!==s&&u>=Ug&&(Yg(n-i)<qg&&(n-=i*qg),Yg(o-s)<qg&&(o-=s*qg),r=Gi(n,r,o,a),t.point(i,r),t.lineEnd(),t.lineStart(),t.point(s,r),e=0),t.point(n=o,r=a),i=s},lineEnd:function(){t.lineEnd(),n=r=NaN},clean:function(){return 2-e}}}function Gi(t,e,n,r){var i,o,a=em(t-n);return Yg(a)>qg?Gg((em(e)*(o=Kg(r))*em(n)-em(r)*(i=Kg(e))*em(t))/(i*o*a)):(e+r)/2}function Xi(t,e,n,r){var i;if(null==t)i=n*Lg,r.point(-Ug,i),r.point(0,i),r.point(Ug,i),r.point(Ug,0),r.point(Ug,-i),r.point(0,-i),r.point(-Ug,-i),r.point(-Ug,0),r.point(-Ug,i);else if(Yg(t[0]-e[0])>qg){var o=t[0]<e[0]?Ug:-Ug;i=n*o/2,r.point(-o,i),r.point(0,i),r.point(o,i)}else r.point(e[0],e[1])}function Ki(t){return function(e){var n=new Ji;for(var r in t)n[r]=t[r];return n.stream=e,n}}function Ji(){}function Zi(t,e,n){var r=e[1][0]-e[0][0],i=e[1][1]-e[0][1],o=t.clipExtent&&t.clipExtent();t.scale(150).translate([0,0]),null!=o&&t.clipExtent(null),sm(n,t.stream(oy));var a=oy.result(),s=Math.min(r/(a[1][0]-a[0][0]),i/(a[1][1]-a[0][1])),u=+e[0][0]+(r-s*(a[1][0]+a[0][0]))/2,l=+e[0][1]+(i-s*(a[1][1]+a[0][1]))/2;return null!=o&&t.clipExtent(o),t.scale(150*s).translate([u,l])}function Qi(t,e,n){return Zi(t,[[0,0],e],n)}function to(t){return Ki({point:function(e,n){e=t(e,n),this.stream.point(e[0],e[1])}})}function eo(t,e){function n(r,i,o,a,s,u,l,c,f,p,h,d,v,g){var m=l-r,y=c-i,b=m*m+y*y;if(b>4*e&&v--){var $=a+p,w=s+h,x=u+d,_=rm($*$+w*w+x*x),C=br(x/=_),k=Yg(Yg(x)-1)<qg||Yg(o-f)<qg?(o+f)/2:Xg(w,$),S=t(k,C),T=S[0],E=S[1],M=T-r,D=E-i,A=y*M-m*D;(A*A/b>e||Yg((m*M+y*D)/b-.5)>.3||a*p+s*h+u*d<My)&&(n(r,i,o,a,s,u,T,E,k,$/=_,w/=_,x,v,g),g.point(T,E),n(T,E,k,$,w,x,l,c,f,p,h,d,v,g))}}return function(e){function r(n,r){n=t(n,r),e.point(n[0],n[1])}function i(){m=NaN,x.point=o,e.lineStart()}function o(r,i){var o=Dr([r,i]),a=t(r,i);n(m,y,g,b,$,w,m=a[0],y=a[1],g=r,b=o[0],$=o[1],w=o[2],Ey,e),e.point(m,y)}function a(){x.point=r,e.lineEnd()}function s(){i(),x.point=u,x.lineEnd=l}function u(t,e){o(c=t,e),f=m,p=y,h=b,d=$,v=w,x.point=o}function l(){n(m,y,g,b,$,w,f,p,c,h,d,v,Ey,e),x.lineEnd=a,a()}var c,f,p,h,d,v,g,m,y,b,$,w,x={point:r,lineStart:i,lineEnd:a,polygonStart:function(){e.polygonStart(),x.lineStart=s},polygonEnd:function(){e.polygonEnd(),x.lineStart=i}};return x}}function no(t){return ro(function(){return t})()}function ro(t){function e(t){return t=c(t[0]*Wg,t[1]*Wg),[t[0]*g+s,u-t[1]*g]}function n(t){return t=c.invert((t[0]-s)/g,(u-t[1])/g),t&&[t[0]*zg,t[1]*zg]}function r(t,e){return t=a(t,e),[t[0]*g+s,u-t[1]*g]}function i(){c=ym(l=ri(w,x,_),a);var t=a(b,$);return s=m-t[0]*g,u=y+t[1]*g,o()}function o(){return d=v=null,e}var a,s,u,l,c,f,p,h,d,v,g=150,m=480,y=250,b=0,$=0,w=0,x=0,_=0,C=null,k=ky,S=null,T=Jm,E=.5,M=Dy(r,E);return e.stream=function(t){return d&&v===t?d:d=Ay(k(l,M(T(v=t))))},e.clipAngle=function(t){return arguments.length?(k=+t?Sy(C=t*Wg,6*Wg):(C=null,ky),o()):C*zg},e.clipExtent=function(t){return arguments.length?(T=null==t?(S=f=p=h=null,Jm):fi(S=+t[0][0],f=+t[0][1],p=+t[1][0],h=+t[1][1]),o()):null==S?null:[[S,f],[p,h]]},e.scale=function(t){return arguments.length?(g=+t,i()):g},e.translate=function(t){return arguments.length?(m=+t[0],y=+t[1],i()):[m,y]},e.center=function(t){return arguments.length?(b=t[0]%360*Wg,$=t[1]%360*Wg,i()):[b*zg,$*zg]},e.rotate=function(t){return arguments.length?(w=t[0]%360*Wg,x=t[1]%360*Wg,_=t.length>2?t[2]%360*Wg:0,i()):[w*zg,x*zg,_*zg]},e.precision=function(t){return arguments.length?(M=Dy(r,E=t*t),o()):rm(E)},e.fitExtent=function(t,n){return Zi(e,t,n)},e.fitSize=function(t,n){return Qi(e,t,n)},function(){return a=t.apply(this,arguments),e.invert=a.invert&&n,i()}}function io(t){var e=0,n=Ug/3,r=ro(t),i=r(e,n);return i.parallels=function(t){return arguments.length?r(e=t[0]*Wg,n=t[1]*Wg):[e*zg,n*zg]},i}function oo(t){function e(t,e){return[t*n,em(e)/n]}var n=Kg(t);return e.invert=function(t,e){return[t/n,br(e*n)]},e}function ao(t,e){function n(t,e){var n=rm(o-2*i*em(e))/i;return[n*em(t*=i),a-n*Kg(t)]}var r=em(t),i=(r+em(e))/2;if(Yg(i)<qg)return oo(t);var o=1+r*(2*i-r),a=rm(o)/i;return n.invert=function(t,e){var n=a-e;return[Xg(t,Yg(n))/i*nm(n),br((o-(t*t+n*n)*i*i)/(2*i))]},n}function so(t){var e=t.length;return{point:function(n,r){for(var i=-1;++i<e;)t[i].point(n,r)},sphere:function(){for(var n=-1;++n<e;)t[n].sphere()},lineStart:function(){for(var n=-1;++n<e;)t[n].lineStart()},lineEnd:function(){for(var n=-1;++n<e;)t[n].lineEnd()},polygonStart:function(){for(var n=-1;++n<e;)t[n].polygonStart()},polygonEnd:function(){for(var n=-1;++n<e;)t[n].polygonEnd()}}}function uo(t){return function(e,n){var r=Kg(e),i=Kg(n),o=t(r*i);return[o*i*em(e),o*em(n)]}}function lo(t){return function(e,n){var r=rm(e*e+n*n),i=t(r),o=em(i),a=Kg(i);return[Xg(e*o,r*a),br(r&&n*o/r)]}}function co(t,e){return[t,Qg(im((Lg+e)/2))]}function fo(t){function e(){var e=Ug*s(),a=o(Dm(o.rotate()).invert([0,0]));return l(null==c?[[a[0]-e,a[1]-e],[a[0]+e,a[1]+e]]:t===co?[[Math.max(a[0]-e,c),n],[Math.min(a[0]+e,r),i]]:[[c,Math.max(a[1]-e,n)],[r,Math.min(a[1]+e,i)]])}var n,r,i,o=no(t),a=o.center,s=o.scale,u=o.translate,l=o.clipExtent,c=null;return o.scale=function(t){return arguments.length?(s(t),e()):s()},o.translate=function(t){return arguments.length?(u(t),e()):u()},o.center=function(t){return arguments.length?(a(t),e()):a()},o.clipExtent=function(t){return arguments.length?(null==t?c=n=r=i=null:(c=+t[0][0],n=+t[0][1],r=+t[1][0],i=+t[1][1]),e()):null==c?null:[[c,n],[r,i]]},e()}function po(t){return im((Lg+t)/2)}function ho(t,e){function n(t,e){o>0?e<-Lg+qg&&(e=-Lg+qg):e>Lg-qg&&(e=Lg-qg);var n=o/tm(po(e),i);return[n*em(i*t),o-n*Kg(i*t)]}var r=Kg(t),i=t===e?em(t):Qg(r/Kg(e))/Qg(po(e)/po(t)),o=r*tm(po(t),i)/i;return i?(n.invert=function(t,e){var n=o-e,r=nm(i)*rm(t*t+n*n);return[Xg(t,Yg(n))/i*nm(n),2*Gg(tm(o/r,1/i))-Lg]},n):co}function vo(t,e){return[t,e]}function go(t,e){function n(t,e){var n=o-e,r=i*t;return[n*em(r),o-n*Kg(r)]}var r=Kg(t),i=t===e?em(t):(r-Kg(e))/(e-t),o=r/i+t;return Yg(i)<qg?vo:(n.invert=function(t,e){var n=o-e;return[Xg(t,Yg(n))/i*nm(n),o-nm(i)*rm(t*t+n*n)]},n)}function mo(t,e){var n=Kg(e),r=Kg(t)*n;return[n*em(t)/r,em(e)/r]}function yo(t,e,n,r){return 1===t&&1===e&&0===n&&0===r?Jm:Ki({point:function(i,o){this.stream.point(i*t+n,o*e+r)}})}function bo(t,e){return[Kg(e)*em(t),em(e)]}function $o(t,e){var n=Kg(e),r=1+Kg(t)*n;return[n*em(t)/r,em(e)/r]}function wo(t,e){return[Qg(im((Lg+e)/2)),-t]}function xo(t,e){return t.parent===e.parent?1:2}function _o(t){return t.reduce(Co,0)/t.length}function Co(t,e){return t+e.x}function ko(t){return 1+t.reduce(So,0)}function So(t,e){return Math.max(t,e.y)}function To(t){for(var e;e=t.children;)t=e[0];return t}function Eo(t){for(var e;e=t.children;)t=e[e.length-1];return t}function Mo(t){var e=0,n=t.children,r=n&&n.length;if(r)for(;--r>=0;)e+=n[r].value;else e=1;t.value=e}function Do(t,e){if(t===e)return t;var n=t.ancestors(),r=e.ancestors(),i=null;for(t=n.pop(),e=r.pop();t===e;)i=t,t=n.pop(),e=r.pop();return i}function Ao(t,e){var n,r,i,o,a,s=new Io(t),u=+t.value&&(s.value=t.value),l=[s];for(null==e&&(e=No);n=l.pop();)if(u&&(n.value=+n.data.value),(i=e(n.data))&&(a=i.length))for(n.children=new Array(a),o=a-1;o>=0;--o)l.push(r=n.children[o]=new Io(i[o])),r.parent=n,r.depth=n.depth+1;return s.eachBefore(Ro)}function Oo(){return Ao(this).eachBefore(Po)}function No(t){return t.children}function Po(t){t.data=t.data.data}function Ro(t){var e=0;do t.height=e;while((t=t.parent)&&t.height<++e)}function Io(t){this.data=t,this.depth=this.height=0,this.parent=null}function jo(t){this._=t,this.next=null}function Fo(t,e){var n=e.x-t.x,r=e.y-t.y,i=t.r-e.r;return i*i+1e-6>n*n+r*r}function qo(t,e){var n,r,i,o=null,a=t.head;switch(e.length){case 1:n=Vo(e[0]);break;case 2:n=Uo(e[0],e[1]);break;case 3:n=Lo(e[0],e[1],e[2])}for(;a;)i=a._,r=a.next,n&&Fo(n,i)?o=a:(o?(t.tail=o,o.next=null):t.head=t.tail=null,e.push(i),n=qo(t,e),e.pop(),t.head?(a.next=t.head,t.head=a):(a.next=null,t.head=t.tail=a),o=t.tail,o.next=r),a=r;return t.tail=o,n}function Vo(t){return{x:t.x,y:t.y,r:t.r}}function Uo(t,e){var n=t.x,r=t.y,i=t.r,o=e.x,a=e.y,s=e.r,u=o-n,l=a-r,c=s-i,f=Math.sqrt(u*u+l*l);return{x:(n+o+u/f*c)/2,y:(r+a+l/f*c)/2,r:(f+i+s)/2}}function Lo(t,e,n){var r=t.x,i=t.y,o=t.r,a=e.x,s=e.y,u=e.r,l=n.x,c=n.y,f=n.r,p=2*(r-a),h=2*(i-s),d=2*(u-o),v=r*r+i*i-o*o-a*a-s*s+u*u,g=2*(r-l),m=2*(i-c),y=2*(f-o),b=r*r+i*i-o*o-l*l-c*c+f*f,$=g*h-p*m,w=(h*b-m*v)/$-r,x=(m*d-h*y)/$,_=(g*v-p*b)/$-i,C=(p*y-g*d)/$,k=x*x+C*C-1,S=2*(w*x+_*C+o),T=w*w+_*_-o*o,E=(-S-Math.sqrt(S*S-4*k*T))/(2*k);return{x:w+x*E+r,y:_+C*E+i,r:E}}function Ho(t,e,n){var r=t.x,i=t.y,o=e.r+n.r,a=t.r+n.r,s=e.x-r,u=e.y-i,l=s*s+u*u;if(l){var c=.5+((a*=a)-(o*=o))/(2*l),f=Math.sqrt(Math.max(0,2*o*(a+l)-(a-=l)*a-o*o))/(2*l);n.x=r+c*s+f*u,n.y=i+c*u-f*s}else n.x=r+a,n.y=i}function Bo(t,e){var n=e.x-t.x,r=e.y-t.y,i=t.r+e.r;return i*i-1e-6>n*n+r*r}function zo(t,e,n){var r=t._,i=t.next._,o=r.r+i.r,a=(r.x*i.r+i.x*r.r)/o-e,s=(r.y*i.r+i.y*r.r)/o-n;return a*a+s*s}function Wo(t){this._=t,this.next=null,this.previous=null}function Yo(t){if(!(i=t.length))return 0;var e,n,r,i;if(e=t[0],e.x=0,e.y=0,!(i>1))return e.r;if(n=t[1],e.x=-n.r,n.x=e.r,n.y=0,!(i>2))return e.r+n.r;Ho(n,e,r=t[2]);var o,a,s,u,l,c,f,p=e.r*e.r,h=n.r*n.r,d=r.r*r.r,v=p+h+d,g=p*e.x+h*n.x+d*r.x,m=p*e.y+h*n.y+d*r.y;e=new Wo(e),n=new Wo(n),r=new Wo(r),e.next=r.previous=n,n.next=e.previous=r,r.next=n.previous=e;t:for(s=3;s<i;++s){Ho(e._,n._,r=t[s]),r=new Wo(r),u=n.next,l=e.previous,c=n._.r,f=e._.r;do if(c<=f){if(Bo(u._,r._)){n=u,e.next=n,n.previous=e,--s;continue t}c+=u._.r,u=u.next}else{if(Bo(l._,r._)){e=l,e.next=n,n.previous=e,--s;continue t}f+=l._.r,l=l.previous}while(u!==l.next);for(r.previous=e,r.next=n,e.next=n.previous=n=r,v+=d=r._.r*r._.r,g+=d*r._.x,m+=d*r._.y,p=zo(e,o=g/v,a=m/v);(r=r.next)!==n;)(d=zo(r,o,a))<p&&(e=r,p=d);n=e.next}for(e=[n._],r=n;(r=r.next)!==n;)e.push(r._);for(r=sb(e),s=0;s<i;++s)e=t[s],e.x-=r.x,e.y-=r.y;return r.r}function Go(t){return null==t?null:Xo(t)}function Xo(t){if("function"!=typeof t)throw new Error;return t}function Ko(){return 0}function Jo(t){return Math.sqrt(t.value)}function Zo(t){return function(e){e.children||(e.r=Math.max(0,+t(e)||0))}}function Qo(t,e){return function(n){if(r=n.children){var r,i,o,a=r.length,s=t(n)*e||0;if(s)for(i=0;i<a;++i)r[i].r+=s;if(o=Yo(r),s)for(i=0;i<a;++i)r[i].r-=s;n.r=o+s}}}function ta(t){return function(e){var n=e.parent;e.r*=t,n&&(e.x=n.x+t*e.x,e.y=n.y+t*e.y)}}function ea(t){return t.id}function na(t){return t.parentId}function ra(t,e){return t.parent===e.parent?1:2}function ia(t){var e=t.children;return e?e[0]:t.t}function oa(t){var e=t.children;return e?e[e.length-1]:t.t}function aa(t,e,n){var r=n/(e.i-t.i);e.c-=r,e.s+=n,t.c+=r,e.z+=n,e.m+=n}function sa(t){for(var e,n=0,r=0,i=t.children,o=i.length;--o>=0;)e=i[o],e.z+=n,e.m+=n,n+=e.s+(r+=e.c)}function ua(t,e,n){return t.a.parent===e.parent?t.a:n}function la(t,e){this._=t,this.parent=null,this.children=null,this.A=null,this.a=this,this.z=0,this.m=0,this.c=0,this.s=0,this.t=null,this.i=e}function ca(t){for(var e,n,r,i,o,a=new la(t,0),s=[a];e=s.pop();)if(r=e._.children)for(e.children=new Array(o=r.length),i=o-1;i>=0;--i)s.push(n=e.children[i]=new la(r[i],i)),n.parent=e;return(a.parent=new la(null,0)).children=[a],a}function fa(t,e,n,r,i,o){for(var a,s,u,l,c,f,p,h,d,v,g,m=[],y=e.children,b=0,$=0,w=y.length,x=e.value;b<w;){u=i-n,l=o-r;do c=y[$++].value;while(!c&&$<w);for(f=p=c,v=Math.max(l/u,u/l)/(x*t),g=c*c*v,d=Math.max(p/g,g/f);$<w;++$){if(c+=s=y[$].value,s<f&&(f=s),s>p&&(p=s),g=c*c*v,h=Math.max(p/g,g/f),h>d){c-=s;break}d=h}m.push(a={value:c,dice:u<l,children:y.slice(b,$)}),a.dice?pb(a,n,r,i,x?r+=l*c/x:o):bb(a,n,r,x?n+=u*c/x:i,o),x-=c,b=$}return m}function pa(t,e){return t[0]-e[0]||t[1]-e[1]}function ha(t){for(var e=t.length,n=[0,1],r=2,i=2;i<e;++i){for(;r>1&&Eb(t[n[r-2]],t[n[r-1]],t[i])<=0;)--r;n[r++]=i}return n.slice(0,r)}function da(t){this._size=t,this._call=this._error=null,this._tasks=[],this._data=[],this._waiting=this._active=this._ended=this._start=0}function va(t){if(!t._start)try{ga(t)}catch(e){if(t._tasks[t._ended+t._active-1])ya(t,e);else if(!t._data)throw e}}function ga(t){for(;t._start=t._waiting&&t._active<t._size;){var e=t._ended+t._active,n=t._tasks[e],r=n.length-1,i=n[r];n[r]=ma(t,e),--t._waiting,++t._active,n=i.apply(null,n),t._tasks[e]&&(t._tasks[e]=n||Nb)}}function ma(t,e){return function(n,r){t._tasks[e]&&(--t._active,++t._ended,t._tasks[e]=null,null==t._error&&(null!=n?ya(t,n):(t._data[e]=r,t._waiting?va(t):ba(t))))}}function ya(t,e){var n,r=t._tasks.length;for(t._error=e,t._data=void 0,t._waiting=NaN;--r>=0;)if((n=t._tasks[r])&&(t._tasks[r]=null,n.abort))try{n.abort()}catch(t){}t._active=NaN,ba(t)}function ba(t){if(!t._active&&t._call){var e=t._data;t._data=void 0,t._call(t._error,e)}}function $a(t){if(null==t)t=1/0;else if(!((t=+t)>=1))throw new Error("invalid concurrency");return new da(t)}function wa(t){return function(e,n){t(null==e?n:null)}}function xa(t){var e=t.responseType;return e&&"text"!==e?t.response:t.responseText}function _a(t,e){return function(n){return t(n.responseText,e)}}function Ca(t){function e(e){var o=e+"",a=n.get(o);if(!a){if(i!==Qb)return i;n.set(o,a=r.push(e))}return t[(a-1)%t.length]}var n=Hn(),r=[],i=Qb;return t=null==t?[]:Zb.call(t),e.domain=function(t){if(!arguments.length)return r.slice();r=[],n=Hn();for(var i,o,a=-1,s=t.length;++a<s;)n.has(o=(i=t[a])+"")||n.set(o,r.push(i));return e},e.range=function(n){return arguments.length?(t=Zb.call(n),e):t.slice()},e.unknown=function(t){return arguments.length?(i=t,e):i},e.copy=function(){return Ca().domain(r).range(t).unknown(i)},e}function ka(){function t(){var t=i().length,r=a[1]<a[0],f=a[r-0],p=a[1-r];e=(p-f)/Math.max(1,t-u+2*l),s&&(e=Math.floor(e)),f+=(p-f-e*(t-u))*c,n=e*(1-u),s&&(f=Math.round(f),n=Math.round(n));var h=uc(t).map(function(t){return f+e*t});return o(r?h.reverse():h)}var e,n,r=Ca().unknown(void 0),i=r.domain,o=r.range,a=[0,1],s=!1,u=0,l=0,c=.5;return delete r.unknown,r.domain=function(e){return arguments.length?(i(e),t()):i()},r.range=function(e){return arguments.length?(a=[+e[0],+e[1]],t()):a.slice()},r.rangeRound=function(e){return a=[+e[0],+e[1]],s=!0,t()},r.bandwidth=function(){return n},r.step=function(){return e},r.round=function(e){return arguments.length?(s=!!e,t()):s},r.padding=function(e){return arguments.length?(u=l=Math.max(0,Math.min(1,e)),t()):u},r.paddingInner=function(e){return arguments.length?(u=Math.max(0,Math.min(1,e)),t()):u},r.paddingOuter=function(e){return arguments.length?(l=Math.max(0,Math.min(1,e)),t()):l},r.align=function(e){return arguments.length?(c=Math.max(0,Math.min(1,e)),t()):c},r.copy=function(){return ka().domain(i()).range(a).round(s).paddingInner(u).paddingOuter(l).align(c)},t()}function Sa(t){var e=t.copy;return t.padding=t.paddingOuter,delete t.paddingInner,delete t.paddingOuter,t.copy=function(){return Sa(e())},t}function Ta(){return Sa(ka().paddingInner(1))}function Ea(t,e){return(e-=t=+t)?function(n){return(n-t)/e}:t$(e)}function Ma(t){return function(e,n){var r=t(e=+e,n=+n);return function(t){return t<=e?0:t>=n?1:r(t)}}}function Da(t){return function(e,n){var r=t(e=+e,n=+n);return function(t){return t<=0?e:t>=1?n:r(t)}}}function Aa(t,e,n,r){var i=t[0],o=t[1],a=e[0],s=e[1];return o<i?(i=n(o,i),a=r(s,a)):(i=n(i,o),a=r(a,s)),function(t){return a(i(t))}}function Oa(t,e,n,r){var i=Math.min(t.length,e.length)-1,o=new Array(i),a=new Array(i),s=-1;for(t[i]<t[0]&&(t=t.slice().reverse(),e=e.slice().reverse());++s<i;)o[s]=n(t[s],t[s+1]),a[s]=r(e[s],e[s+1]);return function(e){var n=Gl(t,e,1,i)-1;return a[n](o[n](e))}}function Na(t,e){return e.domain(t.domain()).range(t.range()).interpolate(t.interpolate()).clamp(t.clamp())}function Pa(t,e){function n(){return i=Math.min(s.length,u.length)>2?Oa:Aa,o=a=null,r}function r(e){return(o||(o=i(s,u,c?Ma(t):t,l)))(+e)}var i,o,a,s=n$,u=n$,l=Vp,c=!1;return r.invert=function(t){return(a||(a=i(u,s,Ea,c?Da(e):e)))(+t)},r.domain=function(t){return arguments.length?(s=Jb.call(t,e$),n()):s.slice()},r.range=function(t){return arguments.length?(u=Zb.call(t),n()):u.slice()},r.rangeRound=function(t){return u=Zb.call(t),l=Up,n()},r.clamp=function(t){return arguments.length?(c=!!t,n()):c},r.interpolate=function(t){return arguments.length?(l=t,n()):l},n()}function Ra(t){var e=t.domain;return t.ticks=function(t){var n=e();return pc(n[0],n[n.length-1],null==t?10:t)},t.tickFormat=function(t,n){return r$(e(),t,n)},t.nice=function(n){null==n&&(n=10);var i,o=e(),a=0,s=o.length-1,u=o[a],l=o[s];return l<u&&(i=u,u=l,l=i,i=a,a=s,s=i),i=r(u,l,n),i>0?(u=Math.floor(u/i)*i,l=Math.ceil(l/i)*i,i=r(u,l,n)):i<0&&(u=Math.ceil(u*i)/i,l=Math.floor(l*i)/i,i=r(u,l,n)),i>0?(o[a]=Math.floor(u/i)*i,o[s]=Math.ceil(l/i)*i,e(o)):i<0&&(o[a]=Math.ceil(u*i)/i,o[s]=Math.floor(l*i)/i,e(o)),t},t}function Ia(){var t=Pa(Ea,Rp);return t.copy=function(){return Na(t,Ia())},Ra(t)}function ja(){function t(t){return+t}var e=[0,1];return t.invert=t,t.domain=t.range=function(n){return arguments.length?(e=Jb.call(n,e$),t):e.slice()},t.copy=function(){return ja().domain(e)},Ra(t)}function Fa(t,e){return(e=Math.log(e/t))?function(n){return Math.log(n/t)/e}:t$(e)}function qa(t,e){return t<0?function(n){return-Math.pow(-e,n)*Math.pow(-t,1-n)}:function(n){return Math.pow(e,n)*Math.pow(t,1-n)}}function Va(t){return isFinite(t)?+("1e"+t):t<0?0:t}function Ua(t){return 10===t?Va:t===Math.E?Math.exp:function(e){return Math.pow(t,e)}}function La(t){return t===Math.E?Math.log:10===t&&Math.log10||2===t&&Math.log2||(t=Math.log(t),function(e){return Math.log(e)/t})}function Ha(t){return function(e){return-t(-e)}}function Ba(){function e(){return o=La(i),a=Ua(i),r()[0]<0&&(o=Ha(o),a=Ha(a)),n}var n=Pa(Fa,qa).domain([1,10]),r=n.domain,i=10,o=La(10),a=Ua(10);return n.base=function(t){return arguments.length?(i=+t,e()):i},n.domain=function(t){return arguments.length?(r(t),e()):r()},n.ticks=function(t){var e,n=r(),s=n[0],u=n[n.length-1];(e=u<s)&&(p=s,s=u,u=p);var l,c,f,p=o(s),h=o(u),d=null==t?10:+t,v=[];if(!(i%1)&&h-p<d){if(p=Math.round(p)-1,h=Math.round(h)+1,s>0){for(;p<h;++p)for(c=1,l=a(p);c<i;++c)if(f=l*c,!(f<s)){if(f>u)break;v.push(f)}}else for(;p<h;++p)for(c=i-1,l=a(p);c>=1;--c)if(f=l*c,!(f<s)){if(f>u)break;v.push(f)}}else v=pc(p,h,Math.min(h-p,d)).map(a);return e?v.reverse():v},n.tickFormat=function(e,r){if(null==r&&(r=10===i?".0e":","),"function"!=typeof r&&(r=t.format(r)),e===1/0)return r;null==e&&(e=10);var s=Math.max(1,i*e/n.ticks().length);return function(t){var e=t/a(Math.round(o(t)));return e*i<i-.5&&(e*=i),e<=s?r(t):""}},n.nice=function(){return r(i$(r(),{floor:function(t){return a(Math.floor(o(t)))},ceil:function(t){return a(Math.ceil(o(t)))}}))},n.copy=function(){return Na(n,Ba().base(i))},n}function za(t,e){return t<0?-Math.pow(-t,e):Math.pow(t,e)}function Wa(){function t(t,e){return(e=za(e,n)-(t=za(t,n)))?function(r){return(za(r,n)-t)/e}:t$(e)}function e(t,e){return e=za(e,n)-(t=za(t,n)),function(r){return za(t+e*r,1/n)}}var n=1,r=Pa(t,e),i=r.domain;return r.exponent=function(t){return arguments.length?(n=+t,i(i())):n},r.copy=function(){return Na(r,Wa().exponent(n))},Ra(r)}function Ya(){return Wa().exponent(.5)}function Ga(){function t(){var t=0,o=Math.max(1,r.length);for(i=new Array(o-1);++t<o;)i[t-1]=vc(n,t/o);return e}function e(t){if(!isNaN(t=+t))return r[Gl(i,t)]}var n=[],r=[],i=[];return e.invertExtent=function(t){var e=r.indexOf(t);return e<0?[NaN,NaN]:[e>0?i[e-1]:n[0],e<i.length?i[e]:n[n.length-1]]},e.domain=function(e){if(!arguments.length)return n.slice();n=[];for(var r,i=0,o=e.length;i<o;++i)r=e[i],null==r||isNaN(r=+r)||n.push(r);return n.sort(zl),t()},e.range=function(e){return arguments.length?(r=Zb.call(e),t()):r.slice()},e.quantiles=function(){return i.slice()},e.copy=function(){return Ga().domain(n).range(r)},e}function Xa(){function t(t){if(t<=t)return a[Gl(o,t,0,i)]}function e(){var e=-1;for(o=new Array(i);++e<i;)o[e]=((e+1)*r-(e-i)*n)/(i+1);return t}var n=0,r=1,i=1,o=[.5],a=[0,1];return t.domain=function(t){return arguments.length?(n=+t[0],r=+t[1],e()):[n,r]},t.range=function(t){return arguments.length?(i=(a=Zb.call(t)).length-1,e()):a.slice()},t.invertExtent=function(t){var e=a.indexOf(t);return e<0?[NaN,NaN]:e<1?[n,o[0]]:e>=i?[o[i-1],r]:[o[e-1],o[e]]},t.copy=function(){return Xa().domain([n,r]).range(a)},Ra(t)}function Ka(){function t(t){if(t<=t)return n[Gl(e,t,0,r)]}var e=[.5],n=[0,1],r=1;return t.domain=function(i){return arguments.length?(e=Zb.call(i),r=Math.min(e.length,n.length-1),t):e.slice()},t.range=function(i){return arguments.length?(n=Zb.call(i),r=Math.min(e.length,n.length-1),t):n.slice()},t.invertExtent=function(t){var r=n.indexOf(t);return[e[r-1],e[r]]},t.copy=function(){return Ka().domain(e).range(n)},t}function Ja(t,e,n,r){function i(e){return t(e=new Date(+e)),e}return i.floor=i,i.ceil=function(n){return t(n=new Date(n-1)),e(n,1),t(n),n},i.round=function(t){var e=i(t),n=i.ceil(t);return t-e<n-t?e:n},i.offset=function(t,n){return e(t=new Date(+t),null==n?1:Math.floor(n)),t},i.range=function(n,r,o){var a=[];if(n=i.ceil(n),o=null==o?1:Math.floor(o),!(n<r&&o>0))return a;do a.push(new Date(+n));while(e(n,o),t(n),n<r);return a},i.filter=function(n){return Ja(function(e){if(e>=e)for(;t(e),!n(e);)e.setTime(e-1)},function(t,r){if(t>=t)for(;--r>=0;)for(;e(t,1),!n(t););})},n&&(i.count=function(e,r){return o$.setTime(+e),a$.setTime(+r),t(o$),t(a$),Math.floor(n(o$,a$))},i.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?i.filter(r?function(e){return r(e)%t===0}:function(e){return i.count(0,e)%t===0}):i:null}),i}function Za(t){return Ja(function(e){e.setDate(e.getDate()-(e.getDay()+7-t)%7),e.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+7*e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*c$)/h$;
})}function Qa(t){return Ja(function(e){e.setUTCDate(e.getUTCDate()-(e.getUTCDay()+7-t)%7),e.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+7*e)},function(t,e){return(e-t)/h$})}function ts(t){if(0<=t.y&&t.y<100){var e=new Date(-1,t.m,t.d,t.H,t.M,t.S,t.L);return e.setFullYear(t.y),e}return new Date(t.y,t.m,t.d,t.H,t.M,t.S,t.L)}function es(t){if(0<=t.y&&t.y<100){var e=new Date(Date.UTC(-1,t.m,t.d,t.H,t.M,t.S,t.L));return e.setUTCFullYear(t.y),e}return new Date(Date.UTC(t.y,t.m,t.d,t.H,t.M,t.S,t.L))}function ns(t){return{y:t,m:0,d:1,H:0,M:0,S:0,L:0}}function rs(t){function e(t,e){return function(n){var r,i,o,a=[],s=-1,u=0,l=t.length;for(n instanceof Date||(n=new Date(+n));++s<l;)37===t.charCodeAt(s)&&(a.push(t.slice(u,s)),null!=(i=fw[r=t.charAt(++s)])?r=t.charAt(++s):i="e"===r?" ":"0",(o=e[r])&&(r=o(n,i)),a.push(r),u=s+1);return a.push(t.slice(u,s)),a.join("")}}function n(t,e){return function(n){var i=ns(1900),o=r(i,t,n+="",0);if(o!=n.length)return null;if("p"in i&&(i.H=i.H%12+12*i.p),"W"in i||"U"in i){"w"in i||(i.w="W"in i?1:0);var a="Z"in i?es(ns(i.y)).getUTCDay():e(ns(i.y)).getDay();i.m=0,i.d="W"in i?(i.w+6)%7+7*i.W-(a+5)%7:i.w+7*i.U-(a+6)%7}return"Z"in i?(i.H+=i.Z/100|0,i.M+=i.Z%100,es(i)):e(i)}}function r(t,e,n,r){for(var i,o,a=0,s=e.length,u=n.length;a<s;){if(r>=u)return-1;if(i=e.charCodeAt(a++),37===i){if(i=e.charAt(a++),o=L[i in fw?e.charAt(a++):i],!o||(r=o(t,n,r))<0)return-1}else if(i!=n.charCodeAt(r++))return-1}return r}function i(t,e,n){var r=D.exec(e.slice(n));return r?(t.p=A[r[0].toLowerCase()],n+r[0].length):-1}function o(t,e,n){var r=P.exec(e.slice(n));return r?(t.w=R[r[0].toLowerCase()],n+r[0].length):-1}function a(t,e,n){var r=O.exec(e.slice(n));return r?(t.w=N[r[0].toLowerCase()],n+r[0].length):-1}function s(t,e,n){var r=F.exec(e.slice(n));return r?(t.m=q[r[0].toLowerCase()],n+r[0].length):-1}function u(t,e,n){var r=I.exec(e.slice(n));return r?(t.m=j[r[0].toLowerCase()],n+r[0].length):-1}function l(t,e,n){return r(t,x,e,n)}function c(t,e,n){return r(t,_,e,n)}function f(t,e,n){return r(t,C,e,n)}function p(t){return T[t.getDay()]}function h(t){return S[t.getDay()]}function d(t){return M[t.getMonth()]}function v(t){return E[t.getMonth()]}function g(t){return k[+(t.getHours()>=12)]}function m(t){return T[t.getUTCDay()]}function y(t){return S[t.getUTCDay()]}function b(t){return M[t.getUTCMonth()]}function $(t){return E[t.getUTCMonth()]}function w(t){return k[+(t.getUTCHours()>=12)]}var x=t.dateTime,_=t.date,C=t.time,k=t.periods,S=t.days,T=t.shortDays,E=t.months,M=t.shortMonths,D=as(k),A=ss(k),O=as(S),N=ss(S),P=as(T),R=ss(T),I=as(E),j=ss(E),F=as(M),q=ss(M),V={a:p,A:h,b:d,B:v,c:null,d:xs,e:xs,H:_s,I:Cs,j:ks,L:Ss,m:Ts,M:Es,p:g,S:Ms,U:Ds,w:As,W:Os,x:null,X:null,y:Ns,Y:Ps,Z:Rs,"%":Ks},U={a:m,A:y,b:b,B:$,c:null,d:Is,e:Is,H:js,I:Fs,j:qs,L:Vs,m:Us,M:Ls,p:w,S:Hs,U:Bs,w:zs,W:Ws,x:null,X:null,y:Ys,Y:Gs,Z:Xs,"%":Ks},L={a:o,A:a,b:s,B:u,c:l,d:vs,e:vs,H:ms,I:ms,j:gs,L:$s,m:ds,M:ys,p:i,S:bs,U:ls,w:us,W:cs,x:c,X:f,y:ps,Y:fs,Z:hs,"%":ws};return V.x=e(_,V),V.X=e(C,V),V.c=e(x,V),U.x=e(_,U),U.X=e(C,U),U.c=e(x,U),{format:function(t){var n=e(t+="",V);return n.toString=function(){return t},n},parse:function(t){var e=n(t+="",ts);return e.toString=function(){return t},e},utcFormat:function(t){var n=e(t+="",U);return n.toString=function(){return t},n},utcParse:function(t){var e=n(t,es);return e.toString=function(){return t},e}}}function is(t,e,n){var r=t<0?"-":"",i=(r?-t:t)+"",o=i.length;return r+(o<n?new Array(n-o+1).join(e)+i:i)}function os(t){return t.replace(dw,"\\$&")}function as(t){return new RegExp("^(?:"+t.map(os).join("|")+")","i")}function ss(t){for(var e={},n=-1,r=t.length;++n<r;)e[t[n].toLowerCase()]=n;return e}function us(t,e,n){var r=pw.exec(e.slice(n,n+1));return r?(t.w=+r[0],n+r[0].length):-1}function ls(t,e,n){var r=pw.exec(e.slice(n));return r?(t.U=+r[0],n+r[0].length):-1}function cs(t,e,n){var r=pw.exec(e.slice(n));return r?(t.W=+r[0],n+r[0].length):-1}function fs(t,e,n){var r=pw.exec(e.slice(n,n+4));return r?(t.y=+r[0],n+r[0].length):-1}function ps(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function hs(t,e,n){var r=/^(Z)|([+-]\d\d)(?:\:?(\d\d))?/.exec(e.slice(n,n+6));return r?(t.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function ds(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.m=r[0]-1,n+r[0].length):-1}function vs(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.d=+r[0],n+r[0].length):-1}function gs(t,e,n){var r=pw.exec(e.slice(n,n+3));return r?(t.m=0,t.d=+r[0],n+r[0].length):-1}function ms(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.H=+r[0],n+r[0].length):-1}function ys(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.M=+r[0],n+r[0].length):-1}function bs(t,e,n){var r=pw.exec(e.slice(n,n+2));return r?(t.S=+r[0],n+r[0].length):-1}function $s(t,e,n){var r=pw.exec(e.slice(n,n+3));return r?(t.L=+r[0],n+r[0].length):-1}function ws(t,e,n){var r=hw.exec(e.slice(n,n+1));return r?n+r[0].length:-1}function xs(t,e){return is(t.getDate(),e,2)}function _s(t,e){return is(t.getHours(),e,2)}function Cs(t,e){return is(t.getHours()%12||12,e,2)}function ks(t,e){return is(1+$$.count(F$(t),t),e,3)}function Ss(t,e){return is(t.getMilliseconds(),e,3)}function Ts(t,e){return is(t.getMonth()+1,e,2)}function Es(t,e){return is(t.getMinutes(),e,2)}function Ms(t,e){return is(t.getSeconds(),e,2)}function Ds(t,e){return is(x$.count(F$(t),t),e,2)}function As(t){return t.getDay()}function Os(t,e){return is(_$.count(F$(t),t),e,2)}function Ns(t,e){return is(t.getFullYear()%100,e,2)}function Ps(t,e){return is(t.getFullYear()%1e4,e,4)}function Rs(t){var e=t.getTimezoneOffset();return(e>0?"-":(e*=-1,"+"))+is(e/60|0,"0",2)+is(e%60,"0",2)}function Is(t,e){return is(t.getUTCDate(),e,2)}function js(t,e){return is(t.getUTCHours(),e,2)}function Fs(t,e){return is(t.getUTCHours()%12||12,e,2)}function qs(t,e){return is(1+B$.count(uw(t),t),e,3)}function Vs(t,e){return is(t.getUTCMilliseconds(),e,3)}function Us(t,e){return is(t.getUTCMonth()+1,e,2)}function Ls(t,e){return is(t.getUTCMinutes(),e,2)}function Hs(t,e){return is(t.getUTCSeconds(),e,2)}function Bs(t,e){return is(W$.count(uw(t),t),e,2)}function zs(t){return t.getUTCDay()}function Ws(t,e){return is(Y$.count(uw(t),t),e,2)}function Ys(t,e){return is(t.getUTCFullYear()%100,e,2)}function Gs(t,e){return is(t.getUTCFullYear()%1e4,e,4)}function Xs(){return"+0000"}function Ks(){return"%"}function Js(e){return lw=rs(e),t.timeFormat=lw.format,t.timeParse=lw.parse,t.utcFormat=lw.utcFormat,t.utcParse=lw.utcParse,lw}function Zs(t){return t.toISOString()}function Qs(t){var e=new Date(t);return isNaN(e)?null:e}function tu(t){return new Date(t)}function eu(t){return t instanceof Date?+t:+new Date(+t)}function nu(t,e,n,r,o,a,s,u,l){function c(i){return(s(i)<i?v:a(i)<i?g:o(i)<i?m:r(i)<i?y:e(i)<i?n(i)<i?b:$:t(i)<i?w:x)(i)}function f(e,n,r,o){if(null==e&&(e=10),"number"==typeof e){var a=Math.abs(r-n)/e,s=Wl(function(t){return t[2]}).right(_,a);s===_.length?(o=i(n/Cw,r/Cw,e),e=t):s?(s=_[a/_[s-1][2]<_[s][2]/a?s-1:s],o=s[1],e=s[0]):(o=i(n,r,e),e=u)}return null==o?e:e.every(o)}var p=Pa(Ea,Rp),h=p.invert,d=p.domain,v=l(".%L"),g=l(":%S"),m=l("%I:%M"),y=l("%I %p"),b=l("%a %d"),$=l("%b %d"),w=l("%B"),x=l("%Y"),_=[[s,1,yw],[s,5,5*yw],[s,15,15*yw],[s,30,30*yw],[a,1,bw],[a,5,5*bw],[a,15,15*bw],[a,30,30*bw],[o,1,$w],[o,3,3*$w],[o,6,6*$w],[o,12,12*$w],[r,1,ww],[r,2,2*ww],[n,1,xw],[e,1,_w],[e,3,3*_w],[t,1,Cw]];return p.invert=function(t){return new Date(h(t))},p.domain=function(t){return arguments.length?d(Jb.call(t,eu)):d().map(tu)},p.ticks=function(t,e){var n,r=d(),i=r[0],o=r[r.length-1],a=o<i;return a&&(n=i,i=o,o=n),n=f(t,i,o,e),n=n?n.range(i,o+1):[],a?n.reverse():n},p.tickFormat=function(t,e){return null==e?c:l(e)},p.nice=function(t,e){var n=d();return(t=f(t,n[0],n[n.length-1],e))?d(i$(n,t)):p},p.copy=function(){return Na(p,nu(t,e,n,r,o,a,s,u,l))},p}function ru(t){var e=t.length;return function(n){return t[Math.max(0,Math.min(e-1,Math.floor(n*e)))]}}function iu(t){function e(e){var o=(e-n)/(r-n);return t(i?Math.max(0,Math.min(1,o)):o)}var n=0,r=1,i=!1;return e.domain=function(t){return arguments.length?(n=+t[0],r=+t[1],e):[n,r]},e.clamp=function(t){return arguments.length?(i=!!t,e):i},e.interpolator=function(n){return arguments.length?(t=n,e):t},e.copy=function(){return iu(t).domain([n,r]).clamp(i)},Ra(e)}function ou(t){return t>1?0:t<-1?Kw:Math.acos(t)}function au(t){return t>=1?Jw:t<=-1?-Jw:Math.asin(t)}function su(t){return t.innerRadius}function uu(t){return t.outerRadius}function lu(t){return t.startAngle}function cu(t){return t.endAngle}function fu(t){return t&&t.padAngle}function pu(t,e,n,r,i,o,a,s){var u=n-t,l=r-e,c=a-i,f=s-o,p=(c*(e-o)-f*(t-i))/(f*u-c*l);return[t+p*u,e+p*l]}function hu(t,e,n,r,i,o,a){var s=t-n,u=e-r,l=(a?o:-o)/Gw(s*s+u*u),c=l*u,f=-l*s,p=t+c,h=e+f,d=n+c,v=r+f,g=(p+d)/2,m=(h+v)/2,y=d-p,b=v-h,$=y*y+b*b,w=i-o,x=p*v-d*h,_=(b<0?-1:1)*Gw(zw(0,w*w*$-x*x)),C=(x*b-y*_)/$,k=(-x*y-b*_)/$,S=(x*b+y*_)/$,T=(-x*y+b*_)/$,E=C-g,M=k-m,D=S-g,A=T-m;return E*E+M*M>D*D+A*A&&(C=S,k=T),{cx:C,cy:k,x01:-c,y01:-f,x11:C*(i/w-1),y11:k*(i/w-1)}}function du(t){this._context=t}function vu(t){return t[0]}function gu(t){return t[1]}function mu(t){this._curve=t}function yu(t){function e(e){return new mu(t(e))}return e._curve=t,e}function bu(t){var e=t.curve;return t.angle=t.x,delete t.x,t.radius=t.y,delete t.y,t.curve=function(t){return arguments.length?e(yu(t)):e()._curve},t}function $u(t){return t.source}function wu(t){return t.target}function xu(t){function e(){var e,s=lx.call(arguments),u=n.apply(this,s),l=r.apply(this,s);if(a||(a=e=In()),t(a,+i.apply(this,(s[0]=u,s)),+o.apply(this,s),+i.apply(this,(s[0]=l,s)),+o.apply(this,s)),e)return a=null,e+""||null}var n=$u,r=wu,i=vu,o=gu,a=null;return e.source=function(t){return arguments.length?(n=t,e):n},e.target=function(t){return arguments.length?(r=t,e):r},e.x=function(t){return arguments.length?(i="function"==typeof t?t:Uw(+t),e):i},e.y=function(t){return arguments.length?(o="function"==typeof t?t:Uw(+t),e):o},e.context=function(t){return arguments.length?(a=null==t?null:t,e):a},e}function _u(t,e,n,r,i){t.moveTo(e,n),t.bezierCurveTo(e=(e+r)/2,n,e,i,r,i)}function Cu(t,e,n,r,i){t.moveTo(e,n),t.bezierCurveTo(e,n=(n+i)/2,r,n,r,i)}function ku(t,e,n,r,i){var o=cx(e,n),a=cx(e,n=(n+i)/2),s=cx(r,n),u=cx(r,i);t.moveTo(o[0],o[1]),t.bezierCurveTo(a[0],a[1],s[0],s[1],u[0],u[1])}function Su(){return xu(_u)}function Tu(){return xu(Cu)}function Eu(){var t=xu(ku);return t.angle=t.x,delete t.x,t.radius=t.y,delete t.y,t}function Mu(t,e,n){t._context.bezierCurveTo((2*t._x0+t._x1)/3,(2*t._y0+t._y1)/3,(t._x0+2*t._x1)/3,(t._y0+2*t._y1)/3,(t._x0+4*t._x1+e)/6,(t._y0+4*t._y1+n)/6)}function Du(t){this._context=t}function Au(t){this._context=t}function Ou(t){this._context=t}function Nu(t,e){this._basis=new Du(t),this._beta=e}function Pu(t,e,n){t._context.bezierCurveTo(t._x1+t._k*(t._x2-t._x0),t._y1+t._k*(t._y2-t._y0),t._x2+t._k*(t._x1-e),t._y2+t._k*(t._y1-n),t._x2,t._y2)}function Ru(t,e){this._context=t,this._k=(1-e)/6}function Iu(t,e){this._context=t,this._k=(1-e)/6}function ju(t,e){this._context=t,this._k=(1-e)/6}function Fu(t,e,n){var r=t._x1,i=t._y1,o=t._x2,a=t._y2;if(t._l01_a>Xw){var s=2*t._l01_2a+3*t._l01_a*t._l12_a+t._l12_2a,u=3*t._l01_a*(t._l01_a+t._l12_a);r=(r*s-t._x0*t._l12_2a+t._x2*t._l01_2a)/u,i=(i*s-t._y0*t._l12_2a+t._y2*t._l01_2a)/u}if(t._l23_a>Xw){var l=2*t._l23_2a+3*t._l23_a*t._l12_a+t._l12_2a,c=3*t._l23_a*(t._l23_a+t._l12_a);o=(o*l+t._x1*t._l23_2a-e*t._l12_2a)/c,a=(a*l+t._y1*t._l23_2a-n*t._l12_2a)/c}t._context.bezierCurveTo(r,i,o,a,t._x2,t._y2)}function qu(t,e){this._context=t,this._alpha=e}function Vu(t,e){this._context=t,this._alpha=e}function Uu(t,e){this._context=t,this._alpha=e}function Lu(t){this._context=t}function Hu(t){return t<0?-1:1}function Bu(t,e,n){var r=t._x1-t._x0,i=e-t._x1,o=(t._y1-t._y0)/(r||i<0&&-0),a=(n-t._y1)/(i||r<0&&-0),s=(o*i+a*r)/(r+i);return(Hu(o)+Hu(a))*Math.min(Math.abs(o),Math.abs(a),.5*Math.abs(s))||0}function zu(t,e){var n=t._x1-t._x0;return n?(3*(t._y1-t._y0)/n-e)/2:e}function Wu(t,e,n){var r=t._x0,i=t._y0,o=t._x1,a=t._y1,s=(o-r)/3;t._context.bezierCurveTo(r+s,i+s*e,o-s,a-s*n,o,a)}function Yu(t){this._context=t}function Gu(t){this._context=new Xu(t)}function Xu(t){this._context=t}function Ku(t){return new Yu(t)}function Ju(t){return new Gu(t)}function Zu(t){this._context=t}function Qu(t){var e,n,r=t.length-1,i=new Array(r),o=new Array(r),a=new Array(r);for(i[0]=0,o[0]=2,a[0]=t[0]+2*t[1],e=1;e<r-1;++e)i[e]=1,o[e]=4,a[e]=4*t[e]+2*t[e+1];for(i[r-1]=2,o[r-1]=7,a[r-1]=8*t[r-1]+t[r],e=1;e<r;++e)n=i[e]/o[e-1],o[e]-=n,a[e]-=n*a[e-1];for(i[r-1]=a[r-1]/o[r-1],e=r-2;e>=0;--e)i[e]=(a[e]-i[e+1])/o[e];for(o[r-1]=(t[r]+i[r-1])/2,e=0;e<r-1;++e)o[e]=2*t[e+1]-i[e+1];return[i,o]}function tl(t,e){this._context=t,this._t=e}function el(t){return new tl(t,0)}function nl(t){return new tl(t,1)}function rl(t,e){return t[e]}function il(t){for(var e,n=0,r=-1,i=t.length;++r<i;)(e=+t[r][1])&&(n+=e);return n}function ol(t){return t[0]}function al(t){return t[1]}function sl(){this._=null}function ul(t){t.U=t.C=t.L=t.R=t.P=t.N=null}function ll(t,e){var n=e,r=e.R,i=n.U;i?i.L===n?i.L=r:i.R=r:t._=r,r.U=i,n.U=r,n.R=r.L,n.R&&(n.R.U=n),r.L=n}function cl(t,e){var n=e,r=e.L,i=n.U;i?i.L===n?i.L=r:i.R=r:t._=r,r.U=i,n.U=r,n.L=r.R,n.L&&(n.L.U=n),r.R=n}function fl(t){for(;t.L;)t=t.L;return t}function pl(t,e,n,r){var i=[null,null],o=s_.push(i)-1;return i.left=t,i.right=e,n&&dl(i,t,e,n),r&&dl(i,e,t,r),o_[t.index].halfedges.push(o),o_[e.index].halfedges.push(o),i}function hl(t,e,n){var r=[e,n];return r.left=t,r}function dl(t,e,n,r){t[0]||t[1]?t.left===n?t[1]=r:t[0]=r:(t[0]=r,t.left=e,t.right=n)}function vl(t,e,n,r,i){var o,a=t[0],s=t[1],u=a[0],l=a[1],c=s[0],f=s[1],p=0,h=1,d=c-u,v=f-l;if(o=e-u,d||!(o>0)){if(o/=d,d<0){if(o<p)return;o<h&&(h=o)}else if(d>0){if(o>h)return;o>p&&(p=o)}if(o=r-u,d||!(o<0)){if(o/=d,d<0){if(o>h)return;o>p&&(p=o)}else if(d>0){if(o<p)return;o<h&&(h=o)}if(o=n-l,v||!(o>0)){if(o/=v,v<0){if(o<p)return;o<h&&(h=o)}else if(v>0){if(o>h)return;o>p&&(p=o)}if(o=i-l,v||!(o<0)){if(o/=v,v<0){if(o>h)return;o>p&&(p=o)}else if(v>0){if(o<p)return;o<h&&(h=o)}return!(p>0||h<1)||(p>0&&(t[0]=[u+p*d,l+p*v]),h<1&&(t[1]=[u+h*d,l+h*v]),!0)}}}}}function gl(t,e,n,r,i){var o=t[1];if(o)return!0;var a,s,u=t[0],l=t.left,c=t.right,f=l[0],p=l[1],h=c[0],d=c[1],v=(f+h)/2,g=(p+d)/2;if(d===p){if(v<e||v>=r)return;if(f>h){if(u){if(u[1]>=i)return}else u=[v,n];o=[v,i]}else{if(u){if(u[1]<n)return}else u=[v,i];o=[v,n]}}else if(a=(f-h)/(d-p),s=g-a*v,a<-1||a>1)if(f>h){if(u){if(u[1]>=i)return}else u=[(n-s)/a,n];o=[(i-s)/a,i]}else{if(u){if(u[1]<n)return}else u=[(i-s)/a,i];o=[(n-s)/a,n]}else if(p<d){if(u){if(u[0]>=r)return}else u=[e,a*e+s];o=[r,a*r+s]}else{if(u){if(u[0]<e)return}else u=[r,a*r+s];o=[e,a*e+s]}return t[0]=u,t[1]=o,!0}function ml(t,e,n,r){for(var i,o=s_.length;o--;)gl(i=s_[o],t,e,n,r)&&vl(i,t,e,n,r)&&(Math.abs(i[0][0]-i[1][0])>c_||Math.abs(i[0][1]-i[1][1])>c_)||delete s_[o]}function yl(t){return o_[t.index]={site:t,halfedges:[]}}function bl(t,e){var n=t.site,r=e.left,i=e.right;return n===i&&(i=r,r=n),i?Math.atan2(i[1]-r[1],i[0]-r[0]):(n===r?(r=e[1],i=e[0]):(r=e[0],i=e[1]),Math.atan2(r[0]-i[0],i[1]-r[1]))}function $l(t,e){return e[+(e.left!==t.site)]}function wl(t,e){return e[+(e.left===t.site)]}function xl(){for(var t,e,n,r,i=0,o=o_.length;i<o;++i)if((t=o_[i])&&(r=(e=t.halfedges).length)){var a=new Array(r),s=new Array(r);for(n=0;n<r;++n)a[n]=n,s[n]=bl(t,s_[e[n]]);for(a.sort(function(t,e){return s[e]-s[t]}),n=0;n<r;++n)s[n]=e[a[n]];for(n=0;n<r;++n)e[n]=s[n]}}function _l(t,e,n,r){var i,o,a,s,u,l,c,f,p,h,d,v,g=o_.length,m=!0;for(i=0;i<g;++i)if(o=o_[i]){for(a=o.site,u=o.halfedges,s=u.length;s--;)s_[u[s]]||u.splice(s,1);for(s=0,l=u.length;s<l;)h=wl(o,s_[u[s]]),d=h[0],v=h[1],c=$l(o,s_[u[++s%l]]),f=c[0],p=c[1],(Math.abs(d-f)>c_||Math.abs(v-p)>c_)&&(u.splice(s,0,s_.push(hl(a,h,Math.abs(d-t)<c_&&r-v>c_?[t,Math.abs(f-t)<c_?p:r]:Math.abs(v-r)<c_&&n-d>c_?[Math.abs(p-r)<c_?f:n,r]:Math.abs(d-n)<c_&&v-e>c_?[n,Math.abs(f-n)<c_?p:e]:Math.abs(v-e)<c_&&d-t>c_?[Math.abs(p-e)<c_?f:t,e]:null))-1),++l);l&&(m=!1)}if(m){var y,b,$,w=1/0;for(i=0,m=null;i<g;++i)(o=o_[i])&&(a=o.site,y=a[0]-t,b=a[1]-e,$=y*y+b*b,$<w&&(w=$,m=o));if(m){var x=[t,e],_=[t,r],C=[n,r],k=[n,e];m.halfedges.push(s_.push(hl(a=m.site,x,_))-1,s_.push(hl(a,_,C))-1,s_.push(hl(a,C,k))-1,s_.push(hl(a,k,x))-1)}}for(i=0;i<g;++i)(o=o_[i])&&(o.halfedges.length||delete o_[i])}function Cl(){ul(this),this.x=this.y=this.arc=this.site=this.cy=null}function kl(t){var e=t.P,n=t.N;if(e&&n){var r=e.site,i=t.site,o=n.site;if(r!==o){var a=i[0],s=i[1],u=r[0]-a,l=r[1]-s,c=o[0]-a,f=o[1]-s,p=2*(u*f-l*c);if(!(p>=-f_)){var h=u*u+l*l,d=c*c+f*f,v=(f*h-l*d)/p,g=(u*d-c*h)/p,m=u_.pop()||new Cl;m.arc=t,m.site=i,m.x=v+a,m.y=(m.cy=g+s)+Math.sqrt(v*v+g*g),t.circle=m;for(var y=null,b=a_._;b;)if(m.y<b.y||m.y===b.y&&m.x<=b.x){if(!b.L){y=b.P;break}b=b.L}else{if(!b.R){y=b;break}b=b.R}a_.insert(y,m),y||(r_=m)}}}}function Sl(t){var e=t.circle;e&&(e.P||(r_=e.N),a_.remove(e),u_.push(e),ul(e),t.circle=null)}function Tl(){ul(this),this.edge=this.site=this.circle=null}function El(t){var e=l_.pop()||new Tl;return e.site=t,e}function Ml(t){Sl(t),i_.remove(t),l_.push(t),ul(t)}function Dl(t){var e=t.circle,n=e.x,r=e.cy,i=[n,r],o=t.P,a=t.N,s=[t];Ml(t);for(var u=o;u.circle&&Math.abs(n-u.circle.x)<c_&&Math.abs(r-u.circle.cy)<c_;)o=u.P,s.unshift(u),Ml(u),u=o;s.unshift(u),Sl(u);for(var l=a;l.circle&&Math.abs(n-l.circle.x)<c_&&Math.abs(r-l.circle.cy)<c_;)a=l.N,s.push(l),Ml(l),l=a;s.push(l),Sl(l);var c,f=s.length;for(c=1;c<f;++c)l=s[c],u=s[c-1],dl(l.edge,u.site,l.site,i);u=s[0],l=s[f-1],l.edge=pl(u.site,l.site,null,i),kl(u),kl(l)}function Al(t){for(var e,n,r,i,o=t[0],a=t[1],s=i_._;s;)if(r=Ol(s,a)-o,r>c_)s=s.L;else{if(i=o-Nl(s,a),!(i>c_)){r>-c_?(e=s.P,n=s):i>-c_?(e=s,n=s.N):e=n=s;break}if(!s.R){e=s;break}s=s.R}yl(t);var u=El(t);if(i_.insert(e,u),e||n){if(e===n)return Sl(e),n=El(e.site),i_.insert(u,n),u.edge=n.edge=pl(e.site,u.site),kl(e),void kl(n);if(!n)return void(u.edge=pl(e.site,u.site));Sl(e),Sl(n);var l=e.site,c=l[0],f=l[1],p=t[0]-c,h=t[1]-f,d=n.site,v=d[0]-c,g=d[1]-f,m=2*(p*g-h*v),y=p*p+h*h,b=v*v+g*g,$=[(g*y-h*b)/m+c,(p*b-v*y)/m+f];dl(n.edge,l,d,$),u.edge=pl(l,t,null,$),n.edge=pl(t,d,null,$),kl(e),kl(n)}}function Ol(t,e){var n=t.site,r=n[0],i=n[1],o=i-e;if(!o)return r;var a=t.P;if(!a)return-(1/0);n=a.site;var s=n[0],u=n[1],l=u-e;if(!l)return s;var c=s-r,f=1/o-1/l,p=c/l;return f?(-p+Math.sqrt(p*p-2*f*(c*c/(-2*l)-u+l/2+i-o/2)))/f+r:(r+s)/2}function Nl(t,e){var n=t.N;if(n)return Ol(n,e);var r=t.site;return r[1]===e?r[0]:1/0}function Pl(t,e,n){return(t[0]-n[0])*(e[1]-t[1])-(t[0]-e[0])*(n[1]-t[1])}function Rl(t,e){return e[1]-t[1]||e[0]-t[0]}function Il(t,e){var n,r,i,o=t.sort(Rl).pop();for(s_=[],o_=new Array(t.length),i_=new sl,a_=new sl;;)if(i=r_,o&&(!i||o[1]<i.y||o[1]===i.y&&o[0]<i.x))o[0]===n&&o[1]===r||(Al(o),n=o[0],r=o[1]),o=t.pop();else{if(!i)break;Dl(i.arc)}if(xl(),e){var a=+e[0][0],s=+e[0][1],u=+e[1][0],l=+e[1][1];ml(a,s,u,l),_l(a,s,u,l)}this.edges=s_,this.cells=o_,i_=a_=s_=o_=null}function jl(t,e,n){this.target=t,this.type=e,this.transform=n}function Fl(t,e,n){this.k=t,this.x=e,this.y=n}function ql(t){return t.__zoom||d_}function Vl(){t.event.stopImmediatePropagation()}function Ul(){return!t.event.button}function Ll(){var t,e,n=this;return n instanceof SVGElement?(n=n.ownerSVGElement||n,t=n.width.baseVal.value,e=n.height.baseVal.value):(t=n.clientWidth,e=n.clientHeight),[[0,0],[t,e]]}function Hl(){return this.__zoom||d_}var Bl="4.9.1",zl=function(t,e){return t<e?-1:t>e?1:t>=e?0:NaN},Wl=function(t){return 1===t.length&&(t=e(t)),{left:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)<0?r=o+1:i=o}return r},right:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)>0?i=o:r=o+1}return r}}},Yl=Wl(zl),Gl=Yl.right,Xl=Yl.left,Kl=function(t,e){null==e&&(e=n);for(var r=0,i=t.length-1,o=t[0],a=new Array(i<0?0:i);r<i;)a[r]=e(o,o=t[++r]);return a},Jl=function(t,e,r){var i,o,a,s,u=t.length,l=e.length,c=new Array(u*l);for(null==r&&(r=n),i=a=0;i<u;++i)for(s=t[i],o=0;o<l;++o,++a)c[a]=r(s,e[o]);return c},Zl=function(t,e){return e<t?-1:e>t?1:e>=t?0:NaN},Ql=function(t){return null===t?NaN:+t},tc=function(t,e){var n,r,i=t.length,o=0,a=-1,s=0,u=0;if(null==e)for(;++a<i;)isNaN(n=Ql(t[a]))||(r=n-s,s+=r/++o,u+=r*(n-s));else for(;++a<i;)isNaN(n=Ql(e(t[a],a,t)))||(r=n-s,s+=r/++o,u+=r*(n-s));if(o>1)return u/(o-1)},ec=function(t,e){var n=tc(t,e);return n?Math.sqrt(n):n},nc=function(t,e){var n,r,i,o=t.length,a=-1;if(null==e){for(;++a<o;)if(null!=(n=t[a])&&n>=n)for(r=i=n;++a<o;)null!=(n=t[a])&&(r>n&&(r=n),i<n&&(i=n))}else for(;++a<o;)if(null!=(n=e(t[a],a,t))&&n>=n)for(r=i=n;++a<o;)null!=(n=e(t[a],a,t))&&(r>n&&(r=n),i<n&&(i=n));return[r,i]},rc=Array.prototype,ic=rc.slice,oc=rc.map,ac=function(t){return function(){return t}},sc=function(t){return t},uc=function(t,e,n){t=+t,e=+e,n=(i=arguments.length)<2?(e=t,t=0,1):i<3?1:+n;for(var r=-1,i=0|Math.max(0,Math.ceil((e-t)/n)),o=new Array(i);++r<i;)o[r]=t+r*n;return o},lc=Math.sqrt(50),cc=Math.sqrt(10),fc=Math.sqrt(2),pc=function(t,e,n){var i,o,a,s=e<t,u=-1;if(s&&(i=t,t=e,e=i),0===(a=r(t,e,n))||!isFinite(a))return[];if(a>0)for(t=Math.ceil(t/a),e=Math.floor(e/a),o=new Array(i=Math.ceil(e-t+1));++u<i;)o[u]=(t+u)*a;else for(t=Math.floor(t*a),e=Math.ceil(e*a),o=new Array(i=Math.ceil(t-e+1));++u<i;)o[u]=(t-u)/a;return s&&o.reverse(),o},hc=function(t){return Math.ceil(Math.log(t.length)/Math.LN2)+1},dc=function(){function t(t){var o,a,s=t.length,u=new Array(s);for(o=0;o<s;++o)u[o]=e(t[o],o,t);var l=n(u),c=l[0],f=l[1],p=r(u,c,f);Array.isArray(p)||(p=i(c,f,p),p=uc(Math.ceil(c/p)*p,Math.floor(f/p)*p,p));for(var h=p.length;p[0]<=c;)p.shift(),--h;for(;p[h-1]>f;)p.pop(),--h;var d,v=new Array(h+1);for(o=0;o<=h;++o)d=v[o]=[],d.x0=o>0?p[o-1]:c,d.x1=o<h?p[o]:f;for(o=0;o<s;++o)a=u[o],c<=a&&a<=f&&v[Gl(p,a,0,h)].push(t[o]);return v}var e=sc,n=nc,r=hc;return t.value=function(n){return arguments.length?(e="function"==typeof n?n:ac(n),t):e},t.domain=function(e){return arguments.length?(n="function"==typeof e?e:ac([e[0],e[1]]),t):n},t.thresholds=function(e){return arguments.length?(r="function"==typeof e?e:ac(Array.isArray(e)?ic.call(e):e),t):r},t},vc=function(t,e,n){if(null==n&&(n=Ql),r=t.length){if((e=+e)<=0||r<2)return+n(t[0],0,t);if(e>=1)return+n(t[r-1],r-1,t);var r,i=(r-1)*e,o=Math.floor(i),a=+n(t[o],o,t),s=+n(t[o+1],o+1,t);return a+(s-a)*(i-o)}},gc=function(t,e,n){return t=oc.call(t,Ql).sort(zl),Math.ceil((n-e)/(2*(vc(t,.75)-vc(t,.25))*Math.pow(t.length,-1/3)))},mc=function(t,e,n){return Math.ceil((n-e)/(3.5*ec(t)*Math.pow(t.length,-1/3)))},yc=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&n>r&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&n>r&&(r=n);return r},bc=function(t,e){var n,r=t.length,i=r,o=-1,a=0;if(null==e)for(;++o<r;)isNaN(n=Ql(t[o]))?--i:a+=n;else for(;++o<r;)isNaN(n=Ql(e(t[o],o,t)))?--i:a+=n;if(i)return a/i},$c=function(t,e){var n,r=t.length,i=-1,o=[];if(null==e)for(;++i<r;)isNaN(n=Ql(t[i]))||o.push(n);else for(;++i<r;)isNaN(n=Ql(e(t[i],i,t)))||o.push(n);return vc(o.sort(zl),.5)},wc=function(t){for(var e,n,r,i=t.length,o=-1,a=0;++o<i;)a+=t[o].length;for(n=new Array(a);--i>=0;)for(r=t[i],e=r.length;--e>=0;)n[--a]=r[e];return n},xc=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&r>n&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&r>n&&(r=n);return r},_c=function(t,e){for(var n=e.length,r=new Array(n);n--;)r[n]=t[e[n]];return r},Cc=function(t,e){if(n=t.length){var n,r,i=0,o=0,a=t[o];for(null==e&&(e=zl);++i<n;)(e(r=t[i],a)<0||0!==e(a,a))&&(a=r,o=i);return 0===e(a,a)?o:void 0}},kc=function(t,e,n){for(var r,i,o=(null==n?t.length:n)-(e=null==e?0:+e);o;)i=Math.random()*o--|0,r=t[o+e],t[o+e]=t[i+e],t[i+e]=r;return t},Sc=function(t,e){var n,r=t.length,i=-1,o=0;if(null==e)for(;++i<r;)(n=+t[i])&&(o+=n);else for(;++i<r;)(n=+e(t[i],i,t))&&(o+=n);return o},Tc=function(t){if(!(i=t.length))return[];for(var e=-1,n=xc(t,o),r=new Array(n);++e<n;)for(var i,a=-1,s=r[e]=new Array(i);++a<i;)s[a]=t[a][e];return r},Ec=function(){return Tc(arguments)},Mc=Array.prototype.slice,Dc=function(t){return t},Ac=1,Oc=2,Nc=3,Pc=4,Rc=1e-6,Ic={value:function(){}};g.prototype=v.prototype={constructor:g,on:function(t,e){var n,r=this._,i=m(t+"",r),o=-1,a=i.length;{if(!(arguments.length<2)){if(null!=e&&"function"!=typeof e)throw new Error("invalid callback: "+e);for(;++o<a;)if(n=(t=i[o]).type)r[n]=b(r[n],t.name,e);else if(null==e)for(n in r)r[n]=b(r[n],t.name,null);return this}for(;++o<a;)if((n=(t=i[o]).type)&&(n=y(r[n],t.name)))return n}},copy:function(){var t={},e=this._;for(var n in e)t[n]=e[n].slice();return new g(t)},call:function(t,e){if((n=arguments.length-2)>0)for(var n,r,i=new Array(n),o=0;o<n;++o)i[o]=arguments[o+2];if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(r=this._[t],o=0,n=r.length;o<n;++o)r[o].value.apply(e,i)},apply:function(t,e,n){if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(var r=this._[t],i=0,o=r.length;i<o;++i)r[i].value.apply(e,n)}};var jc="http://www.w3.org/1999/xhtml",Fc={svg:"http://www.w3.org/2000/svg",xhtml:jc,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},qc=function(t){var e=t+="",n=e.indexOf(":");return n>=0&&"xmlns"!==(e=t.slice(0,n))&&(t=t.slice(n+1)),Fc.hasOwnProperty(e)?{space:Fc[e],local:t}:t},Vc=function(t){var e=qc(t);return(e.local?w:$)(e)},Uc=0;_.prototype=x.prototype={constructor:_,get:function(t){for(var e=this._;!(e in t);)if(!(t=t.parentNode))return;return t[e]},set:function(t,e){return t[this._]=e},remove:function(t){return this._ in t&&delete t[this._]},toString:function(){return this._}};var Lc=function(t){return function(){return this.matches(t)}};if("undefined"!=typeof document){var Hc=document.documentElement;if(!Hc.matches){var Bc=Hc.webkitMatchesSelector||Hc.msMatchesSelector||Hc.mozMatchesSelector||Hc.oMatchesSelector;Lc=function(t){return function(){return Bc.call(this,t)}}}}var zc=Lc,Wc={};if(t.event=null,"undefined"!=typeof document){var Yc=document.documentElement;"onmouseenter"in Yc||(Wc={mouseenter:"mouseover",mouseleave:"mouseout"})}var Gc=function(t,e,n){var r,i,o=S(t+""),a=o.length;{if(!(arguments.length<2)){for(s=e?E:T,null==n&&(n=!1),r=0;r<a;++r)this.each(s(o[r],e,n));return this}var s=this.node().__on;if(s)for(var u,l=0,c=s.length;l<c;++l)for(r=0,u=s[l];r<a;++r)if((i=o[r]).type===u.type&&i.name===u.name)return u.value}},Xc=function(){for(var e,n=t.event;e=n.sourceEvent;)n=e;return n},Kc=function(t,e){var n=t.ownerSVGElement||t;if(n.createSVGPoint){var r=n.createSVGPoint();return r.x=e.clientX,r.y=e.clientY,r=r.matrixTransform(t.getScreenCTM().inverse()),[r.x,r.y]}var i=t.getBoundingClientRect();return[e.clientX-i.left-t.clientLeft,e.clientY-i.top-t.clientTop]},Jc=function(t){var e=Xc();return e.changedTouches&&(e=e.changedTouches[0]),Kc(t,e)},Zc=function(t){return null==t?D:function(){return this.querySelector(t)}},Qc=function(t){"function"!=typeof t&&(t=Zc(t));for(var e=this._groups,n=e.length,r=new Array(n),i=0;i<n;++i)for(var o,a,s=e[i],u=s.length,l=r[i]=new Array(u),c=0;c<u;++c)(o=s[c])&&(a=t.call(o,o.__data__,c,s))&&("__data__"in o&&(a.__data__=o.__data__),l[c]=a);return new gt(r,this._parents)},tf=function(t){return null==t?A:function(){return this.querySelectorAll(t)}},ef=function(t){"function"!=typeof t&&(t=tf(t));for(var e=this._groups,n=e.length,r=[],i=[],o=0;o<n;++o)for(var a,s=e[o],u=s.length,l=0;l<u;++l)(a=s[l])&&(r.push(t.call(a,a.__data__,l,s)),i.push(a));return new gt(r,i)},nf=function(t){"function"!=typeof t&&(t=zc(t));for(var e=this._groups,n=e.length,r=new Array(n),i=0;i<n;++i)for(var o,a=e[i],s=a.length,u=r[i]=[],l=0;l<s;++l)(o=a[l])&&t.call(o,o.__data__,l,a)&&u.push(o);return new gt(r,this._parents)},rf=function(t){return new Array(t.length)},of=function(){return new gt(this._enter||this._groups.map(rf),this._parents)};O.prototype={constructor:O,appendChild:function(t){return this._parent.insertBefore(t,this._next)},insertBefore:function(t,e){return this._parent.insertBefore(t,e)},querySelector:function(t){return this._parent.querySelector(t)},querySelectorAll:function(t){return this._parent.querySelectorAll(t)}};var af=function(t){return function(){return t}},sf="$",uf=function(t,e){if(!t)return h=new Array(this.size()),l=-1,this.each(function(t){h[++l]=t}),h;var n=e?P:N,r=this._parents,i=this._groups;"function"!=typeof t&&(t=af(t));for(var o=i.length,a=new Array(o),s=new Array(o),u=new Array(o),l=0;l<o;++l){var c=r[l],f=i[l],p=f.length,h=t.call(c,c&&c.__data__,l,r),d=h.length,v=s[l]=new Array(d),g=a[l]=new Array(d),m=u[l]=new Array(p);n(c,f,v,g,m,h,e);for(var y,b,$=0,w=0;$<d;++$)if(y=v[$]){for($>=w&&(w=$+1);!(b=g[w])&&++w<d;);y._next=b||null}}return a=new gt(a,r),a._enter=s,a._exit=u,a},lf=function(){return new gt(this._exit||this._groups.map(rf),this._parents)},cf=function(t){for(var e=this._groups,n=t._groups,r=e.length,i=n.length,o=Math.min(r,i),a=new Array(r),s=0;s<o;++s)for(var u,l=e[s],c=n[s],f=l.length,p=a[s]=new Array(f),h=0;h<f;++h)(u=l[h]||c[h])&&(p[h]=u);for(;s<r;++s)a[s]=e[s];return new gt(a,this._parents)},ff=function(){for(var t=this._groups,e=-1,n=t.length;++e<n;)for(var r,i=t[e],o=i.length-1,a=i[o];--o>=0;)(r=i[o])&&(a&&a!==r.nextSibling&&a.parentNode.insertBefore(r,a),a=r);return this},pf=function(t){function e(e,n){return e&&n?t(e.__data__,n.__data__):!e-!n}t||(t=R);for(var n=this._groups,r=n.length,i=new Array(r),o=0;o<r;++o){for(var a,s=n[o],u=s.length,l=i[o]=new Array(u),c=0;c<u;++c)(a=s[c])&&(l[c]=a);l.sort(e)}return new gt(i,this._parents).order()},hf=function(){var t=arguments[0];return arguments[0]=this,t.apply(null,arguments),this},df=function(){var t=new Array(this.size()),e=-1;return this.each(function(){t[++e]=this}),t},vf=function(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var r=t[e],i=0,o=r.length;i<o;++i){var a=r[i];if(a)return a}return null},gf=function(){var t=0;return this.each(function(){++t}),t},mf=function(){return!this.node()},yf=function(t){for(var e=this._groups,n=0,r=e.length;n<r;++n)for(var i,o=e[n],a=0,s=o.length;a<s;++a)(i=o[a])&&t.call(i,i.__data__,a,o);return this},bf=function(t,e){var n=qc(t);if(arguments.length<2){var r=this.node();return n.local?r.getAttributeNS(n.space,n.local):r.getAttribute(n)}return this.each((null==e?n.local?j:I:"function"==typeof e?n.local?U:V:n.local?q:F)(n,e))},$f=function(t){return t.ownerDocument&&t.ownerDocument.defaultView||t.document&&t||t.defaultView},wf=function(t,e,n){return arguments.length>1?this.each((null==e?L:"function"==typeof e?B:H)(t,e,null==n?"":n)):z(this.node(),t)},xf=function(t,e){return arguments.length>1?this.each((null==e?W:"function"==typeof e?G:Y)(t,e)):this.node()[t]};J.prototype={add:function(t){var e=this._names.indexOf(t);e<0&&(this._names.push(t),this._node.setAttribute("class",this._names.join(" ")))},remove:function(t){var e=this._names.indexOf(t);e>=0&&(this._names.splice(e,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(t){return this._names.indexOf(t)>=0}};var _f=function(t,e){var n=X(t+"");if(arguments.length<2){for(var r=K(this.node()),i=-1,o=n.length;++i<o;)if(!r.contains(n[i]))return!1;return!0}return this.each(("function"==typeof e?nt:e?tt:et)(n,e))},Cf=function(t){return arguments.length?this.each(null==t?rt:("function"==typeof t?ot:it)(t)):this.node().textContent},kf=function(t){return arguments.length?this.each(null==t?at:("function"==typeof t?ut:st)(t)):this.node().innerHTML},Sf=function(){return this.each(lt)},Tf=function(){return this.each(ct)},Ef=function(t){var e="function"==typeof t?t:Vc(t);return this.select(function(){return this.appendChild(e.apply(this,arguments))})},Mf=function(t,e){var n="function"==typeof t?t:Vc(t),r=null==e?ft:"function"==typeof e?e:Zc(e);return this.select(function(){
return this.insertBefore(n.apply(this,arguments),r.apply(this,arguments)||null)})},Df=function(){return this.each(pt)},Af=function(t){return arguments.length?this.property("__data__",t):this.node().__data__},Of=function(t,e){return this.each(("function"==typeof e?vt:dt)(t,e))},Nf=[null];gt.prototype=mt.prototype={constructor:gt,select:Qc,selectAll:ef,filter:nf,data:uf,enter:of,exit:lf,merge:cf,order:ff,sort:pf,call:hf,nodes:df,node:vf,size:gf,empty:mf,each:yf,attr:bf,style:wf,property:xf,classed:_f,text:Cf,html:kf,raise:Sf,lower:Tf,append:Ef,insert:Mf,remove:Df,datum:Af,on:Gc,dispatch:Of};var Pf=function(t){return"string"==typeof t?new gt([[document.querySelector(t)]],[document.documentElement]):new gt([[t]],Nf)},Rf=function(t){return"string"==typeof t?new gt([document.querySelectorAll(t)],[document.documentElement]):new gt([null==t?[]:t],Nf)},If=function(t,e,n){arguments.length<3&&(n=e,e=Xc().changedTouches);for(var r,i=0,o=e?e.length:0;i<o;++i)if((r=e[i]).identifier===n)return Kc(t,r);return null},jf=function(t,e){null==e&&(e=Xc().touches);for(var n=0,r=e?e.length:0,i=new Array(r);n<r;++n)i[n]=Kc(t,e[n]);return i},Ff=function(){t.event.preventDefault(),t.event.stopImmediatePropagation()},qf=function(t){var e=t.document.documentElement,n=Pf(t).on("dragstart.drag",Ff,!0);"onselectstart"in e?n.on("selectstart.drag",Ff,!0):(e.__noselect=e.style.MozUserSelect,e.style.MozUserSelect="none")},Vf=function(t){return function(){return t}};$t.prototype.on=function(){var t=this._.on.apply(this._,arguments);return t===this._?this:t};var Uf=function(){function e(t){t.on("mousedown.drag",n).on("touchstart.drag",o).on("touchmove.drag",a).on("touchend.drag touchcancel.drag",s).style("-webkit-tap-highlight-color","rgba(0,0,0,0)")}function n(){if(!p&&h.apply(this,arguments)){var e=u("mouse",d.apply(this,arguments),Jc,this,arguments);e&&(Pf(t.event.view).on("mousemove.drag",r,!0).on("mouseup.drag",i,!0),qf(t.event.view),yt(),f=!1,l=t.event.clientX,c=t.event.clientY,e("start"))}}function r(){if(Ff(),!f){var e=t.event.clientX-l,n=t.event.clientY-c;f=e*e+n*n>$}m.mouse("drag")}function i(){Pf(t.event.view).on("mousemove.drag mouseup.drag",null),bt(t.event.view,f),Ff(),m.mouse("end")}function o(){if(h.apply(this,arguments)){var e,n,r=t.event.changedTouches,i=d.apply(this,arguments),o=r.length;for(e=0;e<o;++e)(n=u(r[e].identifier,i,If,this,arguments))&&(yt(),n("start"))}}function a(){var e,n,r=t.event.changedTouches,i=r.length;for(e=0;e<i;++e)(n=m[r[e].identifier])&&(Ff(),n("drag"))}function s(){var e,n,r=t.event.changedTouches,i=r.length;for(p&&clearTimeout(p),p=setTimeout(function(){p=null},500),e=0;e<i;++e)(n=m[r[e].identifier])&&(yt(),n("end"))}function u(n,r,i,o,a){var s,u,l,c=i(r,n),f=y.copy();if(M(new $t(e,"beforestart",s,n,b,c[0],c[1],0,0,f),function(){return null!=(t.event.subject=s=g.apply(o,a))&&(u=s.x-c[0]||0,l=s.y-c[1]||0,!0)}))return function t(p){var h,d=c;switch(p){case"start":m[n]=t,h=b++;break;case"end":delete m[n],--b;case"drag":c=i(r,n),h=b}M(new $t(e,p,s,n,h,c[0]+u,c[1]+l,c[0]-d[0],c[1]-d[1],f),f.apply,f,[p,o,a])}}var l,c,f,p,h=wt,d=xt,g=_t,m={},y=v("start","drag","end"),b=0,$=0;return e.filter=function(t){return arguments.length?(h="function"==typeof t?t:Vf(!!t),e):h},e.container=function(t){return arguments.length?(d="function"==typeof t?t:Vf(t),e):d},e.subject=function(t){return arguments.length?(g="function"==typeof t?t:Vf(t),e):g},e.on=function(){var t=y.on.apply(y,arguments);return t===y?e:t},e.clickDistance=function(t){return arguments.length?($=(t=+t)*t,e):Math.sqrt($)},e},Lf=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t},Hf=.7,Bf=1/Hf,zf="\\s*([+-]?\\d+)\\s*",Wf="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",Yf="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",Gf=/^#([0-9a-f]{3})$/,Xf=/^#([0-9a-f]{6})$/,Kf=new RegExp("^rgb\\("+[zf,zf,zf]+"\\)$"),Jf=new RegExp("^rgb\\("+[Yf,Yf,Yf]+"\\)$"),Zf=new RegExp("^rgba\\("+[zf,zf,zf,Wf]+"\\)$"),Qf=new RegExp("^rgba\\("+[Yf,Yf,Yf,Wf]+"\\)$"),tp=new RegExp("^hsl\\("+[Wf,Yf,Yf]+"\\)$"),ep=new RegExp("^hsla\\("+[Wf,Yf,Yf,Wf]+"\\)$"),np={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};Lf(kt,St,{displayable:function(){return this.rgb().displayable()},toString:function(){return this.rgb()+""}}),Lf(At,Dt,Ct(kt,{brighter:function(t){return t=null==t?Bf:Math.pow(Bf,t),new At(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?Hf:Math.pow(Hf,t),new At(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return 0<=this.r&&this.r<=255&&0<=this.g&&this.g<=255&&0<=this.b&&this.b<=255&&0<=this.opacity&&this.opacity<=1},toString:function(){var t=this.opacity;return t=isNaN(t)?1:Math.max(0,Math.min(1,t)),(1===t?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}})),Lf(Rt,Pt,Ct(kt,{brighter:function(t){return t=null==t?Bf:Math.pow(Bf,t),new Rt(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?Hf:Math.pow(Hf,t),new Rt(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*e,i=2*n-r;return new At(It(t>=240?t-240:t+120,i,r),It(t,i,r),It(t<120?t+240:t-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1}}));var rp=Math.PI/180,ip=180/Math.PI,op=18,ap=.95047,sp=1,up=1.08883,lp=4/29,cp=6/29,fp=3*cp*cp,pp=cp*cp*cp;Lf(qt,Ft,Ct(kt,{brighter:function(t){return new qt(this.l+op*(null==t?1:t),this.a,this.b,this.opacity)},darker:function(t){return new qt(this.l-op*(null==t?1:t),this.a,this.b,this.opacity)},rgb:function(){var t=(this.l+16)/116,e=isNaN(this.a)?t:t+this.a/500,n=isNaN(this.b)?t:t-this.b/200;return t=sp*Ut(t),e=ap*Ut(e),n=up*Ut(n),new At(Lt(3.2404542*e-1.5371385*t-.4985314*n),Lt(-.969266*e+1.8760108*t+.041556*n),Lt(.0556434*e-.2040259*t+1.0572252*n),this.opacity)}})),Lf(Wt,zt,Ct(kt,{brighter:function(t){return new Wt(this.h,this.c,this.l+op*(null==t?1:t),this.opacity)},darker:function(t){return new Wt(this.h,this.c,this.l-op*(null==t?1:t),this.opacity)},rgb:function(){return jt(this).rgb()}}));var hp=-.14861,dp=1.78277,vp=-.29227,gp=-.90649,mp=1.97294,yp=mp*gp,bp=mp*dp,$p=dp*vp-gp*hp;Lf(Xt,Gt,Ct(kt,{brighter:function(t){return t=null==t?Bf:Math.pow(Bf,t),new Xt(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?Hf:Math.pow(Hf,t),new Xt(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=isNaN(this.h)?0:(this.h+120)*rp,e=+this.l,n=isNaN(this.s)?0:this.s*e*(1-e),r=Math.cos(t),i=Math.sin(t);return new At(255*(e+n*(hp*r+dp*i)),255*(e+n*(vp*r+gp*i)),255*(e+n*(mp*r)),this.opacity)}}));var wp,xp,_p,Cp,kp,Sp,Tp=function(t){var e=t.length-1;return function(n){var r=n<=0?n=0:n>=1?(n=1,e-1):Math.floor(n*e),i=t[r],o=t[r+1],a=r>0?t[r-1]:2*i-o,s=r<e-1?t[r+2]:2*o-i;return Kt((n-r/e)*e,a,i,o,s)}},Ep=function(t){var e=t.length;return function(n){var r=Math.floor(((n%=1)<0?++n:n)*e),i=t[(r+e-1)%e],o=t[r%e],a=t[(r+1)%e],s=t[(r+2)%e];return Kt((n-r/e)*e,i,o,a,s)}},Mp=function(t){return function(){return t}},Dp=function t(e){function n(t,e){var n=r((t=Dt(t)).r,(e=Dt(e)).r),i=r(t.g,e.g),o=r(t.b,e.b),a=ee(t.opacity,e.opacity);return function(e){return t.r=n(e),t.g=i(e),t.b=o(e),t.opacity=a(e),t+""}}var r=te(e);return n.gamma=t,n}(1),Ap=ne(Tp),Op=ne(Ep),Np=function(t,e){var n,r=e?e.length:0,i=t?Math.min(r,t.length):0,o=new Array(r),a=new Array(r);for(n=0;n<i;++n)o[n]=Vp(t[n],e[n]);for(;n<r;++n)a[n]=e[n];return function(t){for(n=0;n<i;++n)a[n]=o[n](t);return a}},Pp=function(t,e){var n=new Date;return t=+t,e-=t,function(r){return n.setTime(t+e*r),n}},Rp=function(t,e){return t=+t,e-=t,function(n){return t+e*n}},Ip=function(t,e){var n,r={},i={};null!==t&&"object"==typeof t||(t={}),null!==e&&"object"==typeof e||(e={});for(n in e)n in t?r[n]=Vp(t[n],e[n]):i[n]=e[n];return function(t){for(n in r)i[n]=r[n](t);return i}},jp=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,Fp=new RegExp(jp.source,"g"),qp=function(t,e){var n,r,i,o=jp.lastIndex=Fp.lastIndex=0,a=-1,s=[],u=[];for(t+="",e+="";(n=jp.exec(t))&&(r=Fp.exec(e));)(i=r.index)>o&&(i=e.slice(o,i),s[a]?s[a]+=i:s[++a]=i),(n=n[0])===(r=r[0])?s[a]?s[a]+=r:s[++a]=r:(s[++a]=null,u.push({i:a,x:Rp(n,r)})),o=Fp.lastIndex;return o<e.length&&(i=e.slice(o),s[a]?s[a]+=i:s[++a]=i),s.length<2?u[0]?ie(u[0].x):re(e):(e=u.length,function(t){for(var n,r=0;r<e;++r)s[(n=u[r]).i]=n.x(t);return s.join("")})},Vp=function(t,e){var n,r=typeof e;return null==e||"boolean"===r?Mp(e):("number"===r?Rp:"string"===r?(n=St(e))?(e=n,Dp):qp:e instanceof St?Dp:e instanceof Date?Pp:Array.isArray(e)?Np:"function"!=typeof e.valueOf&&"function"!=typeof e.toString||isNaN(e)?Ip:Rp)(t,e)},Up=function(t,e){return t=+t,e-=t,function(n){return Math.round(t+e*n)}},Lp=180/Math.PI,Hp={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1},Bp=function(t,e,n,r,i,o){var a,s,u;return(a=Math.sqrt(t*t+e*e))&&(t/=a,e/=a),(u=t*n+e*r)&&(n-=t*u,r-=e*u),(s=Math.sqrt(n*n+r*r))&&(n/=s,r/=s,u/=s),t*r<e*n&&(t=-t,e=-e,u=-u,a=-a),{translateX:i,translateY:o,rotate:Math.atan2(e,t)*Lp,skewX:Math.atan(u)*Lp,scaleX:a,scaleY:s}},zp=se(oe,"px, ","px)","deg)"),Wp=se(ae,", ",")",")"),Yp=Math.SQRT2,Gp=2,Xp=4,Kp=1e-12,Jp=function(t,e){var n,r,i=t[0],o=t[1],a=t[2],s=e[0],u=e[1],l=e[2],c=s-i,f=u-o,p=c*c+f*f;if(p<Kp)r=Math.log(l/a)/Yp,n=function(t){return[i+t*c,o+t*f,a*Math.exp(Yp*t*r)]};else{var h=Math.sqrt(p),d=(l*l-a*a+Xp*p)/(2*a*Gp*h),v=(l*l-a*a-Xp*p)/(2*l*Gp*h),g=Math.log(Math.sqrt(d*d+1)-d),m=Math.log(Math.sqrt(v*v+1)-v);r=(m-g)/Yp,n=function(t){var e=t*r,n=ue(g),s=a/(Gp*h)*(n*ce(Yp*e+g)-le(g));return[i+s*c,o+s*f,a*n/ue(Yp*e+g)]}}return n.duration=1e3*r,n},Zp=fe(Qt),Qp=fe(ee),th=he(Qt),eh=he(ee),nh=de(Qt),rh=de(ee),ih=function(t,e){for(var n=new Array(e),r=0;r<e;++r)n[r]=t(r/(e-1));return n},oh=0,ah=0,sh=0,uh=1e3,lh=0,ch=0,fh=0,ph="object"==typeof performance&&performance.now?performance:Date,hh="function"==typeof requestAnimationFrame?requestAnimationFrame:function(t){setTimeout(t,17)};me.prototype=ye.prototype={constructor:me,restart:function(t,e,n){if("function"!=typeof t)throw new TypeError("callback is not a function");n=(null==n?ve():+n)+(null==e?0:+e),this._next||Sp===this||(Sp?Sp._next=this:kp=this,Sp=this),this._call=t,this._time=n,_e()},stop:function(){this._call&&(this._call=null,this._time=1/0,_e())}};var dh=function(t,e,n){var r=new me;return e=null==e?0:+e,r.restart(function(n){r.stop(),t(n+e)},e,n),r},vh=function(t,e,n){var r=new me,i=e;return null==e?(r.restart(t,e,n),r):(e=+e,n=null==n?ve():+n,r.restart(function o(a){a+=i,r.restart(o,i+=e,n),t(a)},e,n),r)},gh=v("start","end","interrupt"),mh=[],yh=0,bh=1,$h=2,wh=3,xh=4,_h=5,Ch=6,kh=function(t,e,n,r,i,o){var a=t.__transition;if(a){if(n in a)return}else t.__transition={};Te(t,n,{name:e,index:r,group:i,on:gh,tween:mh,time:o.time,delay:o.delay,duration:o.duration,ease:o.ease,timer:null,state:yh})},Sh=function(t,e){var n,r,i,o=t.__transition,a=!0;if(o){e=null==e?null:e+"";for(i in o)(n=o[i]).name===e?(r=n.state>$h&&n.state<_h,n.state=Ch,n.timer.stop(),r&&n.on.call("interrupt",t,t.__data__,n.index,n.group),delete o[i]):a=!1;a&&delete t.__transition}},Th=function(t){return this.each(function(){Sh(this,t)})},Eh=function(t,e){var n=this._id;if(t+="",arguments.length<2){for(var r,i=Se(this.node(),n).tween,o=0,a=i.length;o<a;++o)if((r=i[o]).name===t)return r.value;return null}return this.each((null==e?Ee:Me)(n,t,e))},Mh=function(t,e){var n;return("number"==typeof e?Rp:e instanceof St?Dp:(n=St(e))?(e=n,Dp):qp)(t,e)},Dh=function(t,e){var n=qc(t),r="transform"===n?Wp:Mh;return this.attrTween(t,"function"==typeof e?(n.local?Ie:Re)(n,r,De(this,"attr."+t,e)):null==e?(n.local?Oe:Ae)(n):(n.local?Pe:Ne)(n,r,e+""))},Ah=function(t,e){var n="attr."+t;if(arguments.length<2)return(n=this.tween(n))&&n._value;if(null==e)return this.tween(n,null);if("function"!=typeof e)throw new Error;var r=qc(t);return this.tween(n,(r.local?je:Fe)(r,e))},Oh=function(t){var e=this._id;return arguments.length?this.each(("function"==typeof t?qe:Ve)(e,t)):Se(this.node(),e).delay},Nh=function(t){var e=this._id;return arguments.length?this.each(("function"==typeof t?Ue:Le)(e,t)):Se(this.node(),e).duration},Ph=function(t){var e=this._id;return arguments.length?this.each(He(e,t)):Se(this.node(),e).ease},Rh=function(t){"function"!=typeof t&&(t=zc(t));for(var e=this._groups,n=e.length,r=new Array(n),i=0;i<n;++i)for(var o,a=e[i],s=a.length,u=r[i]=[],l=0;l<s;++l)(o=a[l])&&t.call(o,o.__data__,l,a)&&u.push(o);return new tn(r,this._parents,this._name,this._id)},Ih=function(t){if(t._id!==this._id)throw new Error;for(var e=this._groups,n=t._groups,r=e.length,i=n.length,o=Math.min(r,i),a=new Array(r),s=0;s<o;++s)for(var u,l=e[s],c=n[s],f=l.length,p=a[s]=new Array(f),h=0;h<f;++h)(u=l[h]||c[h])&&(p[h]=u);for(;s<r;++s)a[s]=e[s];return new tn(a,this._parents,this._name,this._id)},jh=function(t,e){var n=this._id;return arguments.length<2?Se(this.node(),n).on.on(t):this.each(ze(n,t,e))},Fh=function(){return this.on("end.remove",We(this._id))},qh=function(t){var e=this._name,n=this._id;"function"!=typeof t&&(t=Zc(t));for(var r=this._groups,i=r.length,o=new Array(i),a=0;a<i;++a)for(var s,u,l=r[a],c=l.length,f=o[a]=new Array(c),p=0;p<c;++p)(s=l[p])&&(u=t.call(s,s.__data__,p,l))&&("__data__"in s&&(u.__data__=s.__data__),f[p]=u,kh(f[p],e,n,p,f,Se(s,n)));return new tn(o,this._parents,e,n)},Vh=function(t){var e=this._name,n=this._id;"function"!=typeof t&&(t=tf(t));for(var r=this._groups,i=r.length,o=[],a=[],s=0;s<i;++s)for(var u,l=r[s],c=l.length,f=0;f<c;++f)if(u=l[f]){for(var p,h=t.call(u,u.__data__,f,l),d=Se(u,n),v=0,g=h.length;v<g;++v)(p=h[v])&&kh(p,e,n,v,h,d);o.push(h),a.push(u)}return new tn(o,a,e,n)},Uh=mt.prototype.constructor,Lh=function(){return new Uh(this._groups,this._parents)},Hh=function(t,e,n){var r="transform"==(t+="")?zp:Mh;return null==e?this.styleTween(t,Ye(t,r)).on("end.style."+t,Ge(t)):this.styleTween(t,"function"==typeof e?Ke(t,r,De(this,"style."+t,e)):Xe(t,r,e+""),n)},Bh=function(t,e,n){var r="style."+(t+="");if(arguments.length<2)return(r=this.tween(r))&&r._value;if(null==e)return this.tween(r,null);if("function"!=typeof e)throw new Error;return this.tween(r,Je(t,e,null==n?"":n))},zh=function(t){return this.tween("text","function"==typeof t?Qe(De(this,"text",t)):Ze(null==t?"":t+""))},Wh=function(){for(var t=this._name,e=this._id,n=nn(),r=this._groups,i=r.length,o=0;o<i;++o)for(var a,s=r[o],u=s.length,l=0;l<u;++l)if(a=s[l]){var c=Se(a,e);kh(a,t,n,l,s,{time:c.time+c.delay+c.duration,delay:0,duration:c.duration,ease:c.ease})}return new tn(r,this._parents,t,n)},Yh=0,Gh=mt.prototype;tn.prototype=en.prototype={constructor:tn,select:qh,selectAll:Vh,filter:Rh,merge:Ih,selection:Lh,transition:Wh,call:Gh.call,nodes:Gh.nodes,node:Gh.node,size:Gh.size,empty:Gh.empty,each:Gh.each,on:jh,attr:Dh,attrTween:Ah,style:Hh,styleTween:Bh,text:zh,remove:Fh,tween:Eh,delay:Oh,duration:Nh,ease:Ph};var Xh=3,Kh=function t(e){function n(t){return Math.pow(t,e)}return e=+e,n.exponent=t,n}(Xh),Jh=function t(e){function n(t){return 1-Math.pow(1-t,e)}return e=+e,n.exponent=t,n}(Xh),Zh=function t(e){function n(t){return((t*=2)<=1?Math.pow(t,e):2-Math.pow(2-t,e))/2}return e=+e,n.exponent=t,n}(Xh),Qh=Math.PI,td=Qh/2,ed=4/11,nd=6/11,rd=8/11,id=.75,od=9/11,ad=10/11,sd=.9375,ud=21/22,ld=63/64,cd=1/ed/ed,fd=1.70158,pd=function t(e){function n(t){return t*t*((e+1)*t-e)}return e=+e,n.overshoot=t,n}(fd),hd=function t(e){function n(t){return--t*t*((e+1)*t+e)+1}return e=+e,n.overshoot=t,n}(fd),dd=function t(e){function n(t){return((t*=2)<1?t*t*((e+1)*t-e):(t-=2)*t*((e+1)*t+e)+2)/2}return e=+e,n.overshoot=t,n}(fd),vd=2*Math.PI,gd=1,md=.3,yd=function t(e,n){function r(t){return e*Math.pow(2,10*--t)*Math.sin((i-t)/n)}var i=Math.asin(1/(e=Math.max(1,e)))*(n/=vd);return r.amplitude=function(e){return t(e,n*vd)},r.period=function(n){return t(e,n)},r}(gd,md),bd=function t(e,n){function r(t){return 1-e*Math.pow(2,-10*(t=+t))*Math.sin((t+i)/n)}var i=Math.asin(1/(e=Math.max(1,e)))*(n/=vd);return r.amplitude=function(e){return t(e,n*vd)},r.period=function(n){return t(e,n)},r}(gd,md),$d=function t(e,n){function r(t){return((t=2*t-1)<0?e*Math.pow(2,10*t)*Math.sin((i-t)/n):2-e*Math.pow(2,-10*t)*Math.sin((i+t)/n))/2}var i=Math.asin(1/(e=Math.max(1,e)))*(n/=vd);return r.amplitude=function(e){return t(e,n*vd)},r.period=function(n){return t(e,n)},r}(gd,md),wd={time:null,delay:0,duration:250,ease:cn},xd=function(t){var e,n;t instanceof tn?(e=t._id,t=t._name):(e=nn(),(n=wd).time=ve(),t=null==t?null:t+"");for(var r=this._groups,i=r.length,o=0;o<i;++o)for(var a,s=r[o],u=s.length,l=0;l<u;++l)(a=s[l])&&kh(a,t,e,l,s,n||_n(a,e));return new tn(r,this._parents,t,e)};mt.prototype.interrupt=Th,mt.prototype.transition=xd;var _d=[null],Cd=function(t,e){var n,r,i=t.__transition;if(i){e=null==e?null:e+"";for(r in i)if((n=i[r]).state>bh&&n.name===e)return new tn([[t]],_d,e,+r)}return null},kd=function(t){return function(){return t}},Sd=function(t,e,n){this.target=t,this.type=e,this.selection=n},Td=function(){t.event.preventDefault(),t.event.stopImmediatePropagation()},Ed={name:"drag"},Md={name:"space"},Dd={name:"handle"},Ad={name:"center"},Od={name:"x",handles:["e","w"].map(kn),input:function(t,e){return t&&[[t[0],e[0][1]],[t[1],e[1][1]]]},output:function(t){return t&&[t[0][0],t[1][0]]}},Nd={name:"y",handles:["n","s"].map(kn),input:function(t,e){return t&&[[e[0][0],t[0]],[e[1][0],t[1]]]},output:function(t){return t&&[t[0][1],t[1][1]]}},Pd={name:"xy",handles:["n","e","s","w","nw","ne","se","sw"].map(kn),input:function(t){return t},output:function(t){return t}},Rd={overlay:"crosshair",selection:"move",n:"ns-resize",e:"ew-resize",s:"ns-resize",w:"ew-resize",nw:"nwse-resize",ne:"nesw-resize",se:"nwse-resize",sw:"nesw-resize"},Id={e:"w",w:"e",nw:"ne",ne:"nw",se:"sw",sw:"se"},jd={n:"s",s:"n",nw:"sw",ne:"se",se:"ne",sw:"nw"},Fd={overlay:1,selection:1,n:null,e:1,s:null,w:-1,nw:-1,ne:1,se:1,sw:-1},qd={overlay:1,selection:1,n:-1,e:null,s:1,w:null,nw:-1,ne:-1,se:1,sw:1},Vd=function(){return Nn(Pd)},Ud=Math.cos,Ld=Math.sin,Hd=Math.PI,Bd=Hd/2,zd=2*Hd,Wd=Math.max,Yd=function(){function t(t){var o,a,s,u,l,c,f=t.length,p=[],h=uc(f),d=[],v=[],g=v.groups=new Array(f),m=new Array(f*f);for(o=0,l=-1;++l<f;){for(a=0,c=-1;++c<f;)a+=t[l][c];p.push(a),d.push(uc(f)),o+=a}for(n&&h.sort(function(t,e){return n(p[t],p[e])}),r&&d.forEach(function(e,n){e.sort(function(e,i){return r(t[n][e],t[n][i])})}),o=Wd(0,zd-e*f)/o,u=o?e:zd/f,a=0,l=-1;++l<f;){for(s=a,c=-1;++c<f;){var y=h[l],b=d[y][c],$=t[y][b],w=a,x=a+=$*o;m[b*f+y]={index:y,subindex:b,startAngle:w,endAngle:x,value:$}}g[y]={index:y,startAngle:s,endAngle:a,value:p[y]},a+=u}for(l=-1;++l<f;)for(c=l-1;++c<f;){var _=m[c*f+l],C=m[l*f+c];(_.value||C.value)&&v.push(_.value<C.value?{source:C,target:_}:{source:_,target:C})}return i?v.sort(i):v}var e=0,n=null,r=null,i=null;return t.padAngle=function(n){return arguments.length?(e=Wd(0,n),t):e},t.sortGroups=function(e){return arguments.length?(n=e,t):n},t.sortSubgroups=function(e){return arguments.length?(r=e,t):r},t.sortChords=function(e){return arguments.length?(null==e?i=null:(i=Pn(e))._=e,t):i&&i._},t},Gd=Array.prototype.slice,Xd=function(t){return function(){return t}},Kd=Math.PI,Jd=2*Kd,Zd=1e-6,Qd=Jd-Zd;Rn.prototype=In.prototype={constructor:Rn,moveTo:function(t,e){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)},closePath:function(){null!==this._x1&&(this._x1=this._x0,this._y1=this._y0,this._+="Z")},lineTo:function(t,e){this._+="L"+(this._x1=+t)+","+(this._y1=+e)},quadraticCurveTo:function(t,e,n,r){this._+="Q"+ +t+","+ +e+","+(this._x1=+n)+","+(this._y1=+r)},bezierCurveTo:function(t,e,n,r,i,o){this._+="C"+ +t+","+ +e+","+ +n+","+ +r+","+(this._x1=+i)+","+(this._y1=+o)},arcTo:function(t,e,n,r,i){t=+t,e=+e,n=+n,r=+r,i=+i;var o=this._x1,a=this._y1,s=n-t,u=r-e,l=o-t,c=a-e,f=l*l+c*c;if(i<0)throw new Error("negative radius: "+i);if(null===this._x1)this._+="M"+(this._x1=t)+","+(this._y1=e);else if(f>Zd)if(Math.abs(c*s-u*l)>Zd&&i){var p=n-o,h=r-a,d=s*s+u*u,v=p*p+h*h,g=Math.sqrt(d),m=Math.sqrt(f),y=i*Math.tan((Kd-Math.acos((d+f-v)/(2*g*m)))/2),b=y/m,$=y/g;Math.abs(b-1)>Zd&&(this._+="L"+(t+b*l)+","+(e+b*c)),this._+="A"+i+","+i+",0,0,"+ +(c*p>l*h)+","+(this._x1=t+$*s)+","+(this._y1=e+$*u)}else this._+="L"+(this._x1=t)+","+(this._y1=e);else;},arc:function(t,e,n,r,i,o){t=+t,e=+e,n=+n;var a=n*Math.cos(r),s=n*Math.sin(r),u=t+a,l=e+s,c=1^o,f=o?r-i:i-r;if(n<0)throw new Error("negative radius: "+n);null===this._x1?this._+="M"+u+","+l:(Math.abs(this._x1-u)>Zd||Math.abs(this._y1-l)>Zd)&&(this._+="L"+u+","+l),n&&(f<0&&(f=f%Jd+Jd),f>Qd?this._+="A"+n+","+n+",0,1,"+c+","+(t-a)+","+(e-s)+"A"+n+","+n+",0,1,"+c+","+(this._x1=u)+","+(this._y1=l):f>Zd&&(this._+="A"+n+","+n+",0,"+ +(f>=Kd)+","+c+","+(this._x1=t+n*Math.cos(i))+","+(this._y1=e+n*Math.sin(i))))},rect:function(t,e,n,r){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)+"h"+ +n+"v"+ +r+"h"+-n+"Z"},toString:function(){return this._}};var tv=function(){function t(){var t,s=Gd.call(arguments),u=e.apply(this,s),l=n.apply(this,s),c=+r.apply(this,(s[0]=u,s)),f=i.apply(this,s)-Bd,p=o.apply(this,s)-Bd,h=c*Ud(f),d=c*Ld(f),v=+r.apply(this,(s[0]=l,s)),g=i.apply(this,s)-Bd,m=o.apply(this,s)-Bd;if(a||(a=t=In()),a.moveTo(h,d),a.arc(0,0,c,f,p),f===g&&p===m||(a.quadraticCurveTo(0,0,v*Ud(g),v*Ld(g)),a.arc(0,0,v,g,m)),a.quadraticCurveTo(0,0,h,d),a.closePath(),t)return a=null,t+""||null}var e=jn,n=Fn,r=qn,i=Vn,o=Un,a=null;return t.radius=function(e){return arguments.length?(r="function"==typeof e?e:Xd(+e),t):r},t.startAngle=function(e){return arguments.length?(i="function"==typeof e?e:Xd(+e),t):i},t.endAngle=function(e){return arguments.length?(o="function"==typeof e?e:Xd(+e),t):o},t.source=function(n){return arguments.length?(e=n,t):e},t.target=function(e){return arguments.length?(n=e,t):n},t.context=function(e){return arguments.length?(a=null==e?null:e,t):a},t},ev="$";Ln.prototype=Hn.prototype={constructor:Ln,has:function(t){return ev+t in this},get:function(t){return this[ev+t]},set:function(t,e){return this[ev+t]=e,this},remove:function(t){var e=ev+t;return e in this&&delete this[e]},clear:function(){for(var t in this)t[0]===ev&&delete this[t]},keys:function(){var t=[];for(var e in this)e[0]===ev&&t.push(e.slice(1));return t},values:function(){var t=[];for(var e in this)e[0]===ev&&t.push(this[e]);return t},entries:function(){var t=[];for(var e in this)e[0]===ev&&t.push({key:e.slice(1),value:this[e]});return t},size:function(){var t=0;for(var e in this)e[0]===ev&&++t;return t},empty:function(){for(var t in this)if(t[0]===ev)return!1;return!0},each:function(t){for(var e in this)e[0]===ev&&t(this[e],e.slice(1),this)}};var nv=function(){function t(e,i,a,s){if(i>=o.length)return null!=r?r(e):null!=n?e.sort(n):e;for(var u,l,c,f=-1,p=e.length,h=o[i++],d=Hn(),v=a();++f<p;)(c=d.get(u=h(l=e[f])+""))?c.push(l):d.set(u,[l]);return d.each(function(e,n){s(v,n,t(e,i,a,s))}),v}function e(t,n){if(++n>o.length)return t;var i,s=a[n-1];return null!=r&&n>=o.length?i=t.entries():(i=[],t.each(function(t,r){i.push({key:r,values:e(t,n)})})),null!=s?i.sort(function(t,e){return s(t.key,e.key)}):i}var n,r,i,o=[],a=[];return i={object:function(e){return t(e,0,Bn,zn)},map:function(e){return t(e,0,Wn,Yn)},entries:function(n){return e(t(n,0,Wn,Yn),0)},key:function(t){return o.push(t),i},sortKeys:function(t){return a[o.length-1]=t,i},sortValues:function(t){return n=t,i},rollup:function(t){return r=t,i}}},rv=Hn.prototype;Gn.prototype=Xn.prototype={constructor:Gn,has:rv.has,add:function(t){return t+="",this[ev+t]=t,this},remove:rv.remove,clear:rv.clear,values:rv.keys,size:rv.size,empty:rv.empty,each:rv.each};var iv=function(t){var e=[];for(var n in t)e.push(n);return e},ov=function(t){var e=[];for(var n in t)e.push(t[n]);return e},av=function(t){var e=[];for(var n in t)e.push({key:n,value:t[n]});return e},sv=function(t){function e(t,e){var r,i,o=n(t,function(t,n){return r?r(t,n-1):(i=t,void(r=e?Jn(t,e):Kn(t)))});return o.columns=i,o}function n(t,e){function n(){if(c>=l)return a;if(i)return i=!1,o;var e,n=c;if(34===t.charCodeAt(n)){for(var r=n;r++<l;)if(34===t.charCodeAt(r)){if(34!==t.charCodeAt(r+1))break;++r}return c=r+2,e=t.charCodeAt(r+1),13===e?(i=!0,10===t.charCodeAt(r+2)&&++c):10===e&&(i=!0),t.slice(n+1,r).replace(/""/g,'"')}for(;c<l;){var s=1;if(e=t.charCodeAt(c++),10===e)i=!0;else if(13===e)i=!0,10===t.charCodeAt(c)&&(++c,++s);else if(e!==u)continue;return t.slice(n,c-s)}return t.slice(n)}for(var r,i,o={},a={},s=[],l=t.length,c=0,f=0;(r=n())!==a;){for(var p=[];r!==o&&r!==a;)p.push(r),r=n();e&&null==(p=e(p,f++))||s.push(p)}return s}function r(e,n){return null==n&&(n=Zn(e)),[n.map(a).join(t)].concat(e.map(function(e){return n.map(function(t){return a(e[t])}).join(t)})).join("\n")}function i(t){return t.map(o).join("\n")}function o(e){return e.map(a).join(t)}function a(t){return null==t?"":s.test(t+="")?'"'+t.replace(/\"/g,'""')+'"':t}var s=new RegExp('["'+t+"\n\r]"),u=t.charCodeAt(0);return{parse:e,parseRows:n,format:r,formatRows:i}},uv=sv(","),lv=uv.parse,cv=uv.parseRows,fv=uv.format,pv=uv.formatRows,hv=sv("\t"),dv=hv.parse,vv=hv.parseRows,gv=hv.format,mv=hv.formatRows,yv=function(t,e){function n(){var n,i,o=r.length,a=0,s=0;for(n=0;n<o;++n)i=r[n],a+=i.x,s+=i.y;for(a=a/o-t,s=s/o-e,n=0;n<o;++n)i=r[n],i.x-=a,i.y-=s}var r;return null==t&&(t=0),null==e&&(e=0),n.initialize=function(t){r=t},n.x=function(e){return arguments.length?(t=+e,n):t},n.y=function(t){return arguments.length?(e=+t,n):e},n},bv=function(t){return function(){return t}},$v=function(){return 1e-6*(Math.random()-.5)},wv=function(t){var e=+this._x.call(null,t),n=+this._y.call(null,t);return Qn(this.cover(e,n),e,n,t)},xv=function(t,e){if(isNaN(t=+t)||isNaN(e=+e))return this;var n=this._x0,r=this._y0,i=this._x1,o=this._y1;if(isNaN(n))i=(n=Math.floor(t))+1,o=(r=Math.floor(e))+1;else{if(!(n>t||t>i||r>e||e>o))return this;var a,s,u=i-n,l=this._root;switch(s=(e<(r+o)/2)<<1|t<(n+i)/2){case 0:do a=new Array(4),a[s]=l,l=a;while(u*=2,i=n+u,o=r+u,t>i||e>o);break;case 1:do a=new Array(4),a[s]=l,l=a;while(u*=2,n=i-u,o=r+u,n>t||e>o);break;case 2:do a=new Array(4),a[s]=l,l=a;while(u*=2,i=n+u,r=o-u,t>i||r>e);break;case 3:do a=new Array(4),a[s]=l,l=a;while(u*=2,n=i-u,r=o-u,n>t||r>e)}this._root&&this._root.length&&(this._root=l)}return this._x0=n,this._y0=r,this._x1=i,this._y1=o,this},_v=function(){var t=[];return this.visit(function(e){if(!e.length)do t.push(e.data);while(e=e.next)}),t},Cv=function(t){return arguments.length?this.cover(+t[0][0],+t[0][1]).cover(+t[1][0],+t[1][1]):isNaN(this._x0)?void 0:[[this._x0,this._y0],[this._x1,this._y1]]},kv=function(t,e,n,r,i){this.node=t,this.x0=e,this.y0=n,this.x1=r,this.y1=i},Sv=function(t,e,n){var r,i,o,a,s,u,l,c=this._x0,f=this._y0,p=this._x1,h=this._y1,d=[],v=this._root;for(v&&d.push(new kv(v,c,f,p,h)),null==n?n=1/0:(c=t-n,f=e-n,p=t+n,h=e+n,n*=n);u=d.pop();)if(!(!(v=u.node)||(i=u.x0)>p||(o=u.y0)>h||(a=u.x1)<c||(s=u.y1)<f))if(v.length){var g=(i+a)/2,m=(o+s)/2;d.push(new kv(v[3],g,m,a,s),new kv(v[2],i,m,g,s),new kv(v[1],g,o,a,m),new kv(v[0],i,o,g,m)),(l=(e>=m)<<1|t>=g)&&(u=d[d.length-1],d[d.length-1]=d[d.length-1-l],d[d.length-1-l]=u)}else{var y=t-+this._x.call(null,v.data),b=e-+this._y.call(null,v.data),$=y*y+b*b;if($<n){var w=Math.sqrt(n=$);c=t-w,f=e-w,p=t+w,h=e+w,r=v.data}}return r},Tv=function(t){if(isNaN(o=+this._x.call(null,t))||isNaN(a=+this._y.call(null,t)))return this;var e,n,r,i,o,a,s,u,l,c,f,p,h=this._root,d=this._x0,v=this._y0,g=this._x1,m=this._y1;if(!h)return this;if(h.length)for(;;){if((l=o>=(s=(d+g)/2))?d=s:g=s,(c=a>=(u=(v+m)/2))?v=u:m=u,e=h,!(h=h[f=c<<1|l]))return this;if(!h.length)break;(e[f+1&3]||e[f+2&3]||e[f+3&3])&&(n=e,p=f)}for(;h.data!==t;)if(r=h,!(h=h.next))return this;return(i=h.next)&&delete h.next,r?(i?r.next=i:delete r.next,this):e?(i?e[f]=i:delete e[f],(h=e[0]||e[1]||e[2]||e[3])&&h===(e[3]||e[2]||e[1]||e[0])&&!h.length&&(n?n[p]=h:this._root=h),this):(this._root=i,this)},Ev=function(){return this._root},Mv=function(){var t=0;return this.visit(function(e){if(!e.length)do++t;while(e=e.next)}),t},Dv=function(t){var e,n,r,i,o,a,s=[],u=this._root;for(u&&s.push(new kv(u,this._x0,this._y0,this._x1,this._y1));e=s.pop();)if(!t(u=e.node,r=e.x0,i=e.y0,o=e.x1,a=e.y1)&&u.length){var l=(r+o)/2,c=(i+a)/2;(n=u[3])&&s.push(new kv(n,l,c,o,a)),(n=u[2])&&s.push(new kv(n,r,c,l,a)),(n=u[1])&&s.push(new kv(n,l,i,o,c)),(n=u[0])&&s.push(new kv(n,r,i,l,c))}return this},Av=function(t){var e,n=[],r=[];for(this._root&&n.push(new kv(this._root,this._x0,this._y0,this._x1,this._y1));e=n.pop();){var i=e.node;if(i.length){var o,a=e.x0,s=e.y0,u=e.x1,l=e.y1,c=(a+u)/2,f=(s+l)/2;(o=i[0])&&n.push(new kv(o,a,s,c,f)),(o=i[1])&&n.push(new kv(o,c,s,u,f)),(o=i[2])&&n.push(new kv(o,a,f,c,l)),(o=i[3])&&n.push(new kv(o,c,f,u,l))}r.push(e)}for(;e=r.pop();)t(e.node,e.x0,e.y0,e.x1,e.y1);return this},Ov=function(t){return arguments.length?(this._x=t,this):this._x},Nv=function(t){return arguments.length?(this._y=t,this):this._y},Pv=ir.prototype=or.prototype;Pv.copy=function(){var t,e,n=new or(this._x,this._y,this._x0,this._y0,this._x1,this._y1),r=this._root;if(!r)return n;if(!r.length)return n._root=ar(r),n;for(t=[{source:r,target:n._root=new Array(4)}];r=t.pop();)for(var i=0;i<4;++i)(e=r.source[i])&&(e.length?t.push({source:e,target:r.target[i]=new Array(4)}):r.target[i]=ar(e));
return n},Pv.add=wv,Pv.addAll=tr,Pv.cover=xv,Pv.data=_v,Pv.extent=Cv,Pv.find=Sv,Pv.remove=Tv,Pv.removeAll=er,Pv.root=Ev,Pv.size=Mv,Pv.visit=Dv,Pv.visitAfter=Av,Pv.x=Ov,Pv.y=Nv;var Rv,Iv=function(t){function e(){function t(t,e,n,r,i){var o=t.data,s=t.r,h=f+s;{if(!o)return e>l+h||r<l-h||n>c+h||i<c-h;if(o.index>u.index){var d=l-o.x-o.vx,v=c-o.y-o.vy,g=d*d+v*v;g<h*h&&(0===d&&(d=$v(),g+=d*d),0===v&&(v=$v(),g+=v*v),g=(h-(g=Math.sqrt(g)))/g*a,u.vx+=(d*=g)*(h=(s*=s)/(p+s)),u.vy+=(v*=g)*h,o.vx-=d*(h=1-h),o.vy-=v*h)}}}for(var e,r,u,l,c,f,p,h=i.length,d=0;d<s;++d)for(r=ir(i,sr,ur).visitAfter(n),e=0;e<h;++e)u=i[e],f=o[u.index],p=f*f,l=u.x+u.vx,c=u.y+u.vy,r.visit(t)}function n(t){if(t.data)return t.r=o[t.data.index];for(var e=t.r=0;e<4;++e)t[e]&&t[e].r>t.r&&(t.r=t[e].r)}function r(){if(i){var e,n,r=i.length;for(o=new Array(r),e=0;e<r;++e)n=i[e],o[n.index]=+t(n,e,i)}}var i,o,a=1,s=1;return"function"!=typeof t&&(t=bv(null==t?1:+t)),e.initialize=function(t){i=t,r()},e.iterations=function(t){return arguments.length?(s=+t,e):s},e.strength=function(t){return arguments.length?(a=+t,e):a},e.radius=function(n){return arguments.length?(t="function"==typeof n?n:bv(+n),r(),e):t},e},jv=function(t){function e(t){return 1/Math.min(l[t.source.index],l[t.target.index])}function n(e){for(var n=0,r=t.length;n<d;++n)for(var i,o,u,l,f,p,h,v=0;v<r;++v)i=t[v],o=i.source,u=i.target,l=u.x+u.vx-o.x-o.vx||$v(),f=u.y+u.vy-o.y-o.vy||$v(),p=Math.sqrt(l*l+f*f),p=(p-s[v])/p*e*a[v],l*=p,f*=p,u.vx-=l*(h=c[v]),u.vy-=f*h,o.vx+=l*(h=1-h),o.vy+=f*h}function r(){if(u){var e,n,r=u.length,p=t.length,h=Hn(u,f);for(e=0,l=new Array(r);e<p;++e)n=t[e],n.index=e,"object"!=typeof n.source&&(n.source=cr(h,n.source)),"object"!=typeof n.target&&(n.target=cr(h,n.target)),l[n.source.index]=(l[n.source.index]||0)+1,l[n.target.index]=(l[n.target.index]||0)+1;for(e=0,c=new Array(p);e<p;++e)n=t[e],c[e]=l[n.source.index]/(l[n.source.index]+l[n.target.index]);a=new Array(p),i(),s=new Array(p),o()}}function i(){if(u)for(var e=0,n=t.length;e<n;++e)a[e]=+p(t[e],e,t)}function o(){if(u)for(var e=0,n=t.length;e<n;++e)s[e]=+h(t[e],e,t)}var a,s,u,l,c,f=lr,p=e,h=bv(30),d=1;return null==t&&(t=[]),n.initialize=function(t){u=t,r()},n.links=function(e){return arguments.length?(t=e,r(),n):t},n.id=function(t){return arguments.length?(f=t,n):f},n.iterations=function(t){return arguments.length?(d=+t,n):d},n.strength=function(t){return arguments.length?(p="function"==typeof t?t:bv(+t),i(),n):p},n.distance=function(t){return arguments.length?(h="function"==typeof t?t:bv(+t),o(),n):h},n},Fv=10,qv=Math.PI*(3-Math.sqrt(5)),Vv=function(t){function e(){n(),h.call("tick",o),a<s&&(p.stop(),h.call("end",o))}function n(){var e,n,r=t.length;for(a+=(l-a)*u,f.each(function(t){t(a)}),e=0;e<r;++e)n=t[e],null==n.fx?n.x+=n.vx*=c:(n.x=n.fx,n.vx=0),null==n.fy?n.y+=n.vy*=c:(n.y=n.fy,n.vy=0)}function r(){for(var e,n=0,r=t.length;n<r;++n){if(e=t[n],e.index=n,isNaN(e.x)||isNaN(e.y)){var i=Fv*Math.sqrt(n),o=n*qv;e.x=i*Math.cos(o),e.y=i*Math.sin(o)}(isNaN(e.vx)||isNaN(e.vy))&&(e.vx=e.vy=0)}}function i(e){return e.initialize&&e.initialize(t),e}var o,a=1,s=.001,u=1-Math.pow(s,1/300),l=0,c=.6,f=Hn(),p=ye(e),h=v("tick","end");return null==t&&(t=[]),r(),o={tick:n,restart:function(){return p.restart(e),o},stop:function(){return p.stop(),o},nodes:function(e){return arguments.length?(t=e,r(),f.each(i),o):t},alpha:function(t){return arguments.length?(a=+t,o):a},alphaMin:function(t){return arguments.length?(s=+t,o):s},alphaDecay:function(t){return arguments.length?(u=+t,o):+u},alphaTarget:function(t){return arguments.length?(l=+t,o):l},velocityDecay:function(t){return arguments.length?(c=1-t,o):1-c},force:function(t,e){return arguments.length>1?(null==e?f.remove(t):f.set(t,i(e)),o):f.get(t)},find:function(e,n,r){var i,o,a,s,u,l=0,c=t.length;for(null==r?r=1/0:r*=r,l=0;l<c;++l)s=t[l],i=e-s.x,o=n-s.y,a=i*i+o*o,a<r&&(u=s,r=a);return u},on:function(t,e){return arguments.length>1?(h.on(t,e),o):h.on(t)}}},Uv=function(){function t(t){var e,s=i.length,u=ir(i,fr,pr).visitAfter(n);for(a=t,e=0;e<s;++e)o=i[e],u.visit(r)}function e(){if(i){var t,e,n=i.length;for(s=new Array(n),t=0;t<n;++t)e=i[t],s[e.index]=+u(e,t,i)}}function n(t){var e,n,r,i,o,a=0;if(t.length){for(r=i=o=0;o<4;++o)(e=t[o])&&(n=e.value)&&(a+=n,r+=n*e.x,i+=n*e.y);t.x=r/a,t.y=i/a}else{e=t,e.x=e.data.x,e.y=e.data.y;do a+=s[e.data.index];while(e=e.next)}t.value=a}function r(t,e,n,r){if(!t.value)return!0;var i=t.x-o.x,u=t.y-o.y,p=r-e,h=i*i+u*u;if(p*p/f<h)return h<c&&(0===i&&(i=$v(),h+=i*i),0===u&&(u=$v(),h+=u*u),h<l&&(h=Math.sqrt(l*h)),o.vx+=i*t.value*a/h,o.vy+=u*t.value*a/h),!0;if(!(t.length||h>=c)){(t.data!==o||t.next)&&(0===i&&(i=$v(),h+=i*i),0===u&&(u=$v(),h+=u*u),h<l&&(h=Math.sqrt(l*h)));do t.data!==o&&(p=s[t.data.index]*a/h,o.vx+=i*p,o.vy+=u*p);while(t=t.next)}}var i,o,a,s,u=bv(-30),l=1,c=1/0,f=.81;return t.initialize=function(t){i=t,e()},t.strength=function(n){return arguments.length?(u="function"==typeof n?n:bv(+n),e(),t):u},t.distanceMin=function(e){return arguments.length?(l=e*e,t):Math.sqrt(l)},t.distanceMax=function(e){return arguments.length?(c=e*e,t):Math.sqrt(c)},t.theta=function(e){return arguments.length?(f=e*e,t):Math.sqrt(f)},t},Lv=function(t){function e(t){for(var e,n=0,a=r.length;n<a;++n)e=r[n],e.vx+=(o[n]-e.x)*i[n]*t}function n(){if(r){var e,n=r.length;for(i=new Array(n),o=new Array(n),e=0;e<n;++e)i[e]=isNaN(o[e]=+t(r[e],e,r))?0:+a(r[e],e,r)}}var r,i,o,a=bv(.1);return"function"!=typeof t&&(t=bv(null==t?0:+t)),e.initialize=function(t){r=t,n()},e.strength=function(t){return arguments.length?(a="function"==typeof t?t:bv(+t),n(),e):a},e.x=function(r){return arguments.length?(t="function"==typeof r?r:bv(+r),n(),e):t},e},Hv=function(t){function e(t){for(var e,n=0,a=r.length;n<a;++n)e=r[n],e.vy+=(o[n]-e.y)*i[n]*t}function n(){if(r){var e,n=r.length;for(i=new Array(n),o=new Array(n),e=0;e<n;++e)i[e]=isNaN(o[e]=+t(r[e],e,r))?0:+a(r[e],e,r)}}var r,i,o,a=bv(.1);return"function"!=typeof t&&(t=bv(null==t?0:+t)),e.initialize=function(t){r=t,n()},e.strength=function(t){return arguments.length?(a="function"==typeof t?t:bv(+t),n(),e):a},e.y=function(r){return arguments.length?(t="function"==typeof r?r:bv(+r),n(),e):t},e},Bv=function(t,e){if((n=(t=e?t.toExponential(e-1):t.toExponential()).indexOf("e"))<0)return null;var n,r=t.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+t.slice(n+1)]},zv=function(t){return t=Bv(Math.abs(t)),t?t[1]:NaN},Wv=function(t,e){return function(n,r){for(var i=n.length,o=[],a=0,s=t[0],u=0;i>0&&s>0&&(u+s+1>r&&(s=Math.max(1,r-u)),o.push(n.substring(i-=s,i+s)),!((u+=s+1)>r));)s=t[a=(a+1)%t.length];return o.reverse().join(e)}},Yv=function(t){return function(e){return e.replace(/[0-9]/g,function(e){return t[+e]})}},Gv=function(t,e){t=t.toPrecision(e);t:for(var n,r=t.length,i=1,o=-1;i<r;++i)switch(t[i]){case".":o=n=i;break;case"0":0===o&&(o=i),n=i;break;case"e":break t;default:o>0&&(o=0)}return o>0?t.slice(0,o)+t.slice(n+1):t},Xv=function(t,e){var n=Bv(t,e);if(!n)return t+"";var r=n[0],i=n[1],o=i-(Rv=3*Math.max(-8,Math.min(8,Math.floor(i/3))))+1,a=r.length;return o===a?r:o>a?r+new Array(o-a+1).join("0"):o>0?r.slice(0,o)+"."+r.slice(o):"0."+new Array(1-o).join("0")+Bv(t,Math.max(0,e+o-1))[0]},Kv=function(t,e){var n=Bv(t,e);if(!n)return t+"";var r=n[0],i=n[1];return i<0?"0."+new Array(-i).join("0")+r:r.length>i+1?r.slice(0,i+1)+"."+r.slice(i+1):r+new Array(i-r.length+2).join("0")},Jv={"":Gv,"%":function(t,e){return(100*t).toFixed(e)},b:function(t){return Math.round(t).toString(2)},c:function(t){return t+""},d:function(t){return Math.round(t).toString(10)},e:function(t,e){return t.toExponential(e)},f:function(t,e){return t.toFixed(e)},g:function(t,e){return t.toPrecision(e)},o:function(t){return Math.round(t).toString(8)},p:function(t,e){return Kv(100*t,e)},r:Kv,s:Xv,X:function(t){return Math.round(t).toString(16).toUpperCase()},x:function(t){return Math.round(t).toString(16)}},Zv=/^(?:(.)?([<>=^]))?([+\-\( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?([a-z%])?$/i;hr.prototype=dr.prototype,dr.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(null==this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(null==this.precision?"":"."+Math.max(0,0|this.precision))+this.type};var Qv,tg=function(t){return t},eg=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"],ng=function(t){function e(t){function e(t){var e,i,s,c=g,$=m;if("c"===v)$=y(t)+$,t="";else{t=+t;var w=t<0;if(t=y(Math.abs(t),d),w&&0===+t&&(w=!1),c=(w?"("===l?l:"-":"-"===l||"("===l?"":l)+c,$=$+("s"===v?eg[8+Rv/3]:"")+(w&&"("===l?")":""),b)for(e=-1,i=t.length;++e<i;)if(s=t.charCodeAt(e),48>s||s>57){$=(46===s?o+t.slice(e+1):t.slice(e))+$,t=t.slice(0,e);break}}h&&!f&&(t=r(t,1/0));var x=c.length+t.length+$.length,_=x<p?new Array(p-x+1).join(n):"";switch(h&&f&&(t=r(_+t,_.length?p-$.length:1/0),_=""),u){case"<":t=c+t+$+_;break;case"=":t=c+_+t+$;break;case"^":t=_.slice(0,x=_.length>>1)+c+t+$+_.slice(x);break;default:t=_+c+t+$}return a(t)}t=hr(t);var n=t.fill,u=t.align,l=t.sign,c=t.symbol,f=t.zero,p=t.width,h=t.comma,d=t.precision,v=t.type,g="$"===c?i[0]:"#"===c&&/[boxX]/.test(v)?"0"+v.toLowerCase():"",m="$"===c?i[1]:/[%p]/.test(v)?s:"",y=Jv[v],b=!v||/[defgprs%]/.test(v);return d=null==d?v?6:12:/[gprs]/.test(v)?Math.max(1,Math.min(21,d)):Math.max(0,Math.min(20,d)),e.toString=function(){return t+""},e}function n(t,n){var r=e((t=hr(t),t.type="f",t)),i=3*Math.max(-8,Math.min(8,Math.floor(zv(n)/3))),o=Math.pow(10,-i),a=eg[8+i/3];return function(t){return r(o*t)+a}}var r=t.grouping&&t.thousands?Wv(t.grouping,t.thousands):tg,i=t.currency,o=t.decimal,a=t.numerals?Yv(t.numerals):tg,s=t.percent||"%";return{format:e,formatPrefix:n}};vr({decimal:".",thousands:",",grouping:[3],currency:["$",""]});var rg=function(t){return Math.max(0,-zv(Math.abs(t)))},ig=function(t,e){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(zv(e)/3)))-zv(Math.abs(t)))},og=function(t,e){return t=Math.abs(t),e=Math.abs(e)-t,Math.max(0,zv(e)-zv(t))+1},ag=function(){return new gr};gr.prototype={constructor:gr,reset:function(){this.s=this.t=0},add:function(t){mr(Fg,t,this.t),mr(this,Fg.s,this.s),this.s?this.t+=Fg.t:this.s=Fg.t},valueOf:function(){return this.s}};var sg,ug,lg,cg,fg,pg,hg,dg,vg,gg,mg,yg,bg,$g,wg,xg,_g,Cg,kg,Sg,Tg,Eg,Mg,Dg,Ag,Og,Ng,Pg,Rg,Ig,jg,Fg=new gr,qg=1e-6,Vg=1e-12,Ug=Math.PI,Lg=Ug/2,Hg=Ug/4,Bg=2*Ug,zg=180/Ug,Wg=Ug/180,Yg=Math.abs,Gg=Math.atan,Xg=Math.atan2,Kg=Math.cos,Jg=Math.ceil,Zg=Math.exp,Qg=Math.log,tm=Math.pow,em=Math.sin,nm=Math.sign||function(t){return t>0?1:t<0?-1:0},rm=Math.sqrt,im=Math.tan,om={Feature:function(t,e){xr(t.geometry,e)},FeatureCollection:function(t,e){for(var n=t.features,r=-1,i=n.length;++r<i;)xr(n[r].geometry,e)}},am={Sphere:function(t,e){e.sphere()},Point:function(t,e){t=t.coordinates,e.point(t[0],t[1],t[2])},MultiPoint:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)t=n[r],e.point(t[0],t[1],t[2])},LineString:function(t,e){_r(t.coordinates,e,0)},MultiLineString:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)_r(n[r],e,0)},Polygon:function(t,e){Cr(t.coordinates,e)},MultiPolygon:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)Cr(n[r],e)},GeometryCollection:function(t,e){for(var n=t.geometries,r=-1,i=n.length;++r<i;)xr(n[r],e)}},sm=function(t,e){t&&om.hasOwnProperty(t.type)?om[t.type](t,e):xr(t,e)},um=ag(),lm=ag(),cm={point:wr,lineStart:wr,lineEnd:wr,polygonStart:function(){um.reset(),cm.lineStart=kr,cm.lineEnd=Sr},polygonEnd:function(){var t=+um;lm.add(t<0?Bg+t:t),this.lineStart=this.lineEnd=this.point=wr},sphere:function(){lm.add(Bg)}},fm=function(t){return lm.reset(),sm(t,cm),2*lm},pm=ag(),hm={point:Ir,lineStart:Fr,lineEnd:qr,polygonStart:function(){hm.point=Vr,hm.lineStart=Ur,hm.lineEnd=Lr,pm.reset(),cm.polygonStart()},polygonEnd:function(){cm.polygonEnd(),hm.point=Ir,hm.lineStart=Fr,hm.lineEnd=qr,um<0?(pg=-(dg=180),hg=-(vg=90)):pm>qg?vg=90:pm<-qg&&(hg=-90),wg[0]=pg,wg[1]=dg}},dm=function(t){var e,n,r,i,o,a,s;if(vg=dg=-(pg=hg=1/0),$g=[],sm(t,hm),n=$g.length){for($g.sort(Br),e=1,r=$g[0],o=[r];e<n;++e)i=$g[e],zr(r,i[0])||zr(r,i[1])?(Hr(r[0],i[1])>Hr(r[0],r[1])&&(r[1]=i[1]),Hr(i[0],r[1])>Hr(r[0],r[1])&&(r[0]=i[0])):o.push(r=i);for(a=-(1/0),n=o.length-1,e=0,r=o[n];e<=n;r=i,++e)i=o[e],(s=Hr(r[1],i[0]))>a&&(a=s,pg=i[0],dg=r[1])}return $g=wg=null,pg===1/0||hg===1/0?[[NaN,NaN],[NaN,NaN]]:[[pg,hg],[dg,vg]]},vm={sphere:wr,point:Wr,lineStart:Gr,lineEnd:Jr,polygonStart:function(){vm.lineStart=Zr,vm.lineEnd=Qr},polygonEnd:function(){vm.lineStart=Gr,vm.lineEnd=Jr}},gm=function(t){xg=_g=Cg=kg=Sg=Tg=Eg=Mg=Dg=Ag=Og=0,sm(t,vm);var e=Dg,n=Ag,r=Og,i=e*e+n*n+r*r;return i<Vg&&(e=Tg,n=Eg,r=Mg,_g<qg&&(e=Cg,n=kg,r=Sg),i=e*e+n*n+r*r,i<Vg)?[NaN,NaN]:[Xg(n,e)*zg,br(r/rm(i))*zg]},mm=function(t){return function(){return t}},ym=function(t,e){function n(n,r){return n=t(n,r),e(n[0],n[1])}return t.invert&&e.invert&&(n.invert=function(n,r){return n=e.invert(n,r),n&&t.invert(n[0],n[1])}),n};ni.invert=ni;var bm,$m,wm,xm,_m,Cm,km,Sm,Tm,Em,Mm,Dm=function(t){function e(e){return e=t(e[0]*Wg,e[1]*Wg),e[0]*=zg,e[1]*=zg,e}return t=ri(t[0]*Wg,t[1]*Wg,t.length>2?t[2]*Wg:0),e.invert=function(e){return e=t.invert(e[0]*Wg,e[1]*Wg),e[0]*=zg,e[1]*=zg,e},e},Am=function(){function t(t,e){n.push(t=r(t,e)),t[0]*=zg,t[1]*=zg}function e(){var t=i.apply(this,arguments),e=o.apply(this,arguments)*Wg,u=a.apply(this,arguments)*Wg;return n=[],r=ri(-t[0]*Wg,-t[1]*Wg,0).invert,si(s,e,u,1),t={type:"Polygon",coordinates:[n]},n=r=null,t}var n,r,i=mm([0,0]),o=mm(90),a=mm(6),s={point:t};return e.center=function(t){return arguments.length?(i="function"==typeof t?t:mm([+t[0],+t[1]]),e):i},e.radius=function(t){return arguments.length?(o="function"==typeof t?t:mm(+t),e):o},e.precision=function(t){return arguments.length?(a="function"==typeof t?t:mm(+t),e):a},e},Om=function(){var t,e=[];return{point:function(e,n){t.push([e,n])},lineStart:function(){e.push(t=[])},lineEnd:wr,rejoin:function(){e.length>1&&e.push(e.pop().concat(e.shift()))},result:function(){var n=e;return e=[],t=null,n}}},Nm=function(t,e,n,r,i,o){var a,s=t[0],u=t[1],l=e[0],c=e[1],f=0,p=1,h=l-s,d=c-u;if(a=n-s,h||!(a>0)){if(a/=h,h<0){if(a<f)return;a<p&&(p=a)}else if(h>0){if(a>p)return;a>f&&(f=a)}if(a=i-s,h||!(a<0)){if(a/=h,h<0){if(a>p)return;a>f&&(f=a)}else if(h>0){if(a<f)return;a<p&&(p=a)}if(a=r-u,d||!(a>0)){if(a/=d,d<0){if(a<f)return;a<p&&(p=a)}else if(d>0){if(a>p)return;a>f&&(f=a)}if(a=o-u,d||!(a<0)){if(a/=d,d<0){if(a>p)return;a>f&&(f=a)}else if(d>0){if(a<f)return;a<p&&(p=a)}return f>0&&(t[0]=s+f*h,t[1]=u+f*d),p<1&&(e[0]=s+p*h,e[1]=u+p*d),!0}}}}},Pm=function(t,e){return Yg(t[0]-e[0])<qg&&Yg(t[1]-e[1])<qg},Rm=function(t,e,n,r,i){var o,a,s=[],u=[];if(t.forEach(function(t){if(!((e=t.length-1)<=0)){var e,n,r=t[0],a=t[e];if(Pm(r,a)){for(i.lineStart(),o=0;o<e;++o)i.point((r=t[o])[0],r[1]);return void i.lineEnd()}s.push(n=new li(r,t,null,!0)),u.push(n.o=new li(r,null,n,!1)),s.push(n=new li(a,t,null,!1)),u.push(n.o=new li(a,null,n,!0))}}),s.length){for(u.sort(e),ci(s),ci(u),o=0,a=u.length;o<a;++o)u[o].e=n=!n;for(var l,c,f=s[0];;){for(var p=f,h=!0;p.v;)if((p=p.n)===f)return;l=p.z,i.lineStart();do{if(p.v=p.o.v=!0,p.e){if(h)for(o=0,a=l.length;o<a;++o)i.point((c=l[o])[0],c[1]);else r(p.x,p.n.x,1,i);p=p.n}else{if(h)for(l=p.p.z,o=l.length-1;o>=0;--o)i.point((c=l[o])[0],c[1]);else r(p.x,p.p.x,-1,i);p=p.p}p=p.o,l=p.z,h=!h}while(!p.v);i.lineEnd()}}},Im=1e9,jm=-Im,Fm=function(){var t,e,n,r=0,i=0,o=960,a=500;return n={stream:function(n){return t&&e===n?t:t=fi(r,i,o,a)(e=n)},extent:function(s){return arguments.length?(r=+s[0][0],i=+s[0][1],o=+s[1][0],a=+s[1][1],t=e=null,n):[[r,i],[o,a]]}}},qm=ag(),Vm=function(t,e){var n=e[0],r=e[1],i=[em(n),-Kg(n),0],o=0,a=0;qm.reset();for(var s=0,u=t.length;s<u;++s)if(c=(l=t[s]).length)for(var l,c,f=l[c-1],p=f[0],h=f[1]/2+Hg,d=em(h),v=Kg(h),g=0;g<c;++g,p=y,d=$,v=w,f=m){var m=l[g],y=m[0],b=m[1]/2+Hg,$=em(b),w=Kg(b),x=y-p,_=x>=0?1:-1,C=_*x,k=C>Ug,S=d*$;if(qm.add(Xg(S*_*em(C),v*w+S*Kg(C))),o+=k?x+_*Bg:x,k^p>=n^y>=n){var T=Or(Dr(f),Dr(m));Rr(T);var E=Or(i,T);Rr(E);var M=(k^x>=0?-1:1)*br(E[2]);(r>M||r===M&&(T[0]||T[1]))&&(a+=k^x>=0?1:-1)}}return(o<-qg||o<qg&&qm<-qg)^1&a},Um=ag(),Lm={sphere:wr,point:wr,lineStart:pi,lineEnd:wr,polygonStart:wr,polygonEnd:wr},Hm=function(t){return Um.reset(),sm(t,Lm),+Um},Bm=[null,null],zm={type:"LineString",coordinates:Bm},Wm=function(t,e){return Bm[0]=t,Bm[1]=e,Hm(zm)},Ym={Feature:function(t,e){return gi(t.geometry,e)},FeatureCollection:function(t,e){for(var n=t.features,r=-1,i=n.length;++r<i;)if(gi(n[r].geometry,e))return!0;return!1}},Gm={Sphere:function(){return!0},Point:function(t,e){return mi(t.coordinates,e)},MultiPoint:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(mi(n[r],e))return!0;return!1},LineString:function(t,e){return yi(t.coordinates,e)},MultiLineString:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(yi(n[r],e))return!0;return!1},Polygon:function(t,e){return bi(t.coordinates,e)},MultiPolygon:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)if(bi(n[r],e))return!0;return!1},GeometryCollection:function(t,e){for(var n=t.geometries,r=-1,i=n.length;++r<i;)if(gi(n[r],e))return!0;return!1}},Xm=function(t,e){return(t&&Ym.hasOwnProperty(t.type)?Ym[t.type]:gi)(t,e)},Km=function(t,e){var n=t[0]*Wg,r=t[1]*Wg,i=e[0]*Wg,o=e[1]*Wg,a=Kg(r),s=em(r),u=Kg(o),l=em(o),c=a*Kg(n),f=a*em(n),p=u*Kg(i),h=u*em(i),d=2*br(rm($r(o-r)+a*u*$r(i-n))),v=em(d),g=d?function(t){var e=em(t*=d)/v,n=em(d-t)/v,r=n*c+e*p,i=n*f+e*h,o=n*s+e*l;return[Xg(i,r)*zg,Xg(o,rm(r*r+i*i))*zg]}:function(){return[n*zg,r*zg]};return g.distance=d,g},Jm=function(t){return t},Zm=ag(),Qm=ag(),ty={point:wr,lineStart:wr,lineEnd:wr,polygonStart:function(){ty.lineStart=Si,ty.lineEnd=Mi},polygonEnd:function(){ty.lineStart=ty.lineEnd=ty.point=wr,Zm.add(Yg(Qm)),Qm.reset()},result:function(){var t=Zm/2;return Zm.reset(),t}},ey=1/0,ny=ey,ry=-ey,iy=ry,oy={point:Di,lineStart:wr,lineEnd:wr,polygonStart:wr,polygonEnd:wr,result:function(){var t=[[ey,ny],[ry,iy]];return ry=iy=-(ny=ey=1/0),t}},ay=0,sy=0,uy=0,ly=0,cy=0,fy=0,py=0,hy=0,dy=0,vy={point:Ai,lineStart:Oi,lineEnd:Ri,polygonStart:function(){vy.lineStart=Ii,vy.lineEnd=ji},polygonEnd:function(){vy.point=Ai,vy.lineStart=Oi,vy.lineEnd=Ri},result:function(){var t=dy?[py/dy,hy/dy]:fy?[ly/fy,cy/fy]:uy?[ay/uy,sy/uy]:[NaN,NaN];return ay=sy=uy=ly=cy=fy=py=hy=dy=0,t}};Vi.prototype={_radius:4.5,pointRadius:function(t){return this._radius=t,this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._context.closePath(),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._context.moveTo(t,e),this._point=1;break;case 1:this._context.lineTo(t,e);break;default:this._context.moveTo(t+this._radius,e),this._context.arc(t,e,this._radius,0,Bg)}},result:wr};var gy,my,yy,by,$y,wy=ag(),xy={point:wr,lineStart:function(){xy.point=Ui},lineEnd:function(){gy&&Li(my,yy),xy.point=wr},polygonStart:function(){gy=!0},polygonEnd:function(){gy=null},result:function(){var t=+wy;return wy.reset(),t}};Hi.prototype={_radius:4.5,_circle:Bi(4.5),pointRadius:function(t){return(t=+t)!==this._radius&&(this._radius=t,this._circle=null),this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._string.push("Z"),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._string.push("M",t,",",e),this._point=1;break;case 1:this._string.push("L",t,",",e);break;default:null==this._circle&&(this._circle=Bi(this._radius)),this._string.push("M",t,",",e,this._circle)}},result:function(){if(this._string.length){var t=this._string.join("");return this._string=[],t}return null}};var _y=function(t,e){function n(t){return t&&("function"==typeof o&&i.pointRadius(+o.apply(this,arguments)),sm(t,r(i))),i.result()}var r,i,o=4.5;return n.area=function(t){return sm(t,r(ty)),ty.result()},n.measure=function(t){return sm(t,r(xy)),xy.result()},n.bounds=function(t){return sm(t,r(oy)),oy.result()},n.centroid=function(t){return sm(t,r(vy)),vy.result()},n.projection=function(e){return arguments.length?(r=null==e?(t=null,Jm):(t=e).stream,n):t},n.context=function(t){return arguments.length?(i=null==t?(e=null,new Hi):new Vi(e=t),"function"!=typeof o&&i.pointRadius(o),n):e},n.pointRadius=function(t){return arguments.length?(o="function"==typeof t?t:(i.pointRadius(+t),+t),n):o},n.projection(t).context(e)},Cy=function(t,e,n,r){return function(i,o){function a(e,n){var r=i(e,n);t(e=r[0],n=r[1])&&o.point(e,n)}function s(t,e){var n=i(t,e);g.point(n[0],n[1])}function u(){w.point=s,g.lineStart()}function l(){w.point=a,g.lineEnd()}function c(t,e){v.push([t,e]);var n=i(t,e);b.point(n[0],n[1])}function f(){b.lineStart(),v=[]}function p(){c(v[0][0],v[0][1]),b.lineEnd();var t,e,n,r,i=b.clean(),a=y.result(),s=a.length;if(v.pop(),h.push(v),v=null,s)if(1&i){if(n=a[0],(e=n.length-1)>0){for($||(o.polygonStart(),$=!0),o.lineStart(),t=0;t<e;++t)o.point((r=n[t])[0],r[1]);o.lineEnd()}}else s>1&&2&i&&a.push(a.pop().concat(a.shift())),d.push(a.filter(zi))}var h,d,v,g=e(o),m=i.invert(r[0],r[1]),y=Om(),b=e(y),$=!1,w={point:a,lineStart:u,lineEnd:l,polygonStart:function(){w.point=c,w.lineStart=f,w.lineEnd=p,d=[],h=[]},polygonEnd:function(){w.point=a,w.lineStart=u,w.lineEnd=l,d=wc(d);var t=Vm(h,m);d.length?($||(o.polygonStart(),$=!0),Rm(d,Wi,t,n,o)):t&&($||(o.polygonStart(),$=!0),o.lineStart(),n(null,null,1,o),o.lineEnd()),$&&(o.polygonEnd(),$=!1),d=h=null},sphere:function(){o.polygonStart(),o.lineStart(),n(null,null,1,o),o.lineEnd(),o.polygonEnd()}};return w}},ky=Cy(function(){return!0},Yi,Xi,[-Ug,-Lg]),Sy=function(t,e){function n(n,r,i,o){si(o,t,e,i,n,r)}function r(t,e){return Kg(t)*Kg(e)>s}function i(t){var e,n,i,s,c;return{lineStart:function(){s=i=!1,c=1},point:function(f,p){var h,d=[f,p],v=r(f,p),g=u?v?0:a(f,p):v?a(f+(f<0?Ug:-Ug),p):0;if(!e&&(s=i=v)&&t.lineStart(),v!==i&&(h=o(e,d),(!h||Pm(e,h)||Pm(d,h))&&(d[0]+=qg,d[1]+=qg,v=r(d[0],d[1]))),v!==i)c=0,v?(t.lineStart(),h=o(d,e),t.point(h[0],h[1])):(h=o(e,d),t.point(h[0],h[1]),t.lineEnd()),e=h;else if(l&&e&&u^v){var m;g&n||!(m=o(d,e,!0))||(c=0,u?(t.lineStart(),t.point(m[0][0],m[0][1]),t.point(m[1][0],m[1][1]),t.lineEnd()):(t.point(m[1][0],m[1][1]),t.lineEnd(),t.lineStart(),t.point(m[0][0],m[0][1])))}!v||e&&Pm(e,d)||t.point(d[0],d[1]),e=d,i=v,n=g},lineEnd:function(){i&&t.lineEnd(),e=null},clean:function(){return c|(s&&i)<<1}}}function o(t,e,n){var r=Dr(t),i=Dr(e),o=[1,0,0],a=Or(r,i),u=Ar(a,a),l=a[0],c=u-l*l;if(!c)return!n&&t;var f=s*u/c,p=-s*l/c,h=Or(o,a),d=Pr(o,f),v=Pr(a,p);Nr(d,v);var g=h,m=Ar(d,g),y=Ar(g,g),b=m*m-y*(Ar(d,d)-1);if(!(b<0)){var $=rm(b),w=Pr(g,(-m-$)/y);if(Nr(w,d),w=Mr(w),!n)return w;var x,_=t[0],C=e[0],k=t[1],S=e[1];C<_&&(x=_,_=C,C=x);var T=C-_,E=Yg(T-Ug)<qg,M=E||T<qg;if(!E&&S<k&&(x=k,k=S,S=x),M?E?k+S>0^w[1]<(Yg(w[0]-_)<qg?k:S):k<=w[1]&&w[1]<=S:T>Ug^(_<=w[0]&&w[0]<=C)){var D=Pr(g,(-m+$)/y);return Nr(D,d),[w,Mr(D)]}}}function a(e,n){var r=u?t:Ug-t,i=0;return e<-r?i|=1:e>r&&(i|=2),n<-r?i|=4:n>r&&(i|=8),i}var s=Kg(t),u=s>0,l=Yg(s)>qg;return Cy(r,i,n,u?[0,-t]:[-Ug,t-Ug])},Ty=function(t){return{stream:Ki(t)}};Ji.prototype={constructor:Ji,point:function(t,e){this.stream.point(t,e)},sphere:function(){this.stream.sphere()},lineStart:function(){this.stream.lineStart()},lineEnd:function(){this.stream.lineEnd()},polygonStart:function(){this.stream.polygonStart()},polygonEnd:function(){this.stream.polygonEnd()}};var Ey=16,My=Kg(30*Wg),Dy=function(t,e){return+e?eo(t,e):to(t)},Ay=Ki({point:function(t,e){this.stream.point(t*Wg,e*Wg)}}),Oy=function(){return io(ao).scale(155.424).center([0,33.6442])},Ny=function(){return Oy().parallels([29.5,45.5]).scale(1070).translate([480,250]).rotate([96,0]).center([-.6,38.7])},Py=function(){function t(t){var e=t[0],n=t[1];return s=null,i.point(e,n),s||(o.point(e,n),s)||(a.point(e,n),s)}function e(){return n=r=null,t}var n,r,i,o,a,s,u=Ny(),l=Oy().rotate([154,0]).center([-2,58.5]).parallels([55,65]),c=Oy().rotate([157,0]).center([-3,19.9]).parallels([8,18]),f={point:function(t,e){s=[t,e]}};return t.invert=function(t){var e=u.scale(),n=u.translate(),r=(t[0]-n[0])/e,i=(t[1]-n[1])/e;return(i>=.12&&i<.234&&r>=-.425&&r<-.214?l:i>=.166&&i<.234&&r>=-.214&&r<-.115?c:u).invert(t)},t.stream=function(t){return n&&r===t?n:n=so([u.stream(r=t),l.stream(t),c.stream(t)])},t.precision=function(t){return arguments.length?(u.precision(t),l.precision(t),c.precision(t),e()):u.precision()},t.scale=function(e){return arguments.length?(u.scale(e),l.scale(.35*e),c.scale(e),t.translate(u.translate())):u.scale()},t.translate=function(t){if(!arguments.length)return u.translate();var n=u.scale(),r=+t[0],s=+t[1];return i=u.translate(t).clipExtent([[r-.455*n,s-.238*n],[r+.455*n,s+.238*n]]).stream(f),o=l.translate([r-.307*n,s+.201*n]).clipExtent([[r-.425*n+qg,s+.12*n+qg],[r-.214*n-qg,s+.234*n-qg]]).stream(f),a=c.translate([r-.205*n,s+.212*n]).clipExtent([[r-.214*n+qg,s+.166*n+qg],[r-.115*n-qg,s+.234*n-qg]]).stream(f),e()},t.fitExtent=function(e,n){return Zi(t,e,n)},t.fitSize=function(e,n){return Qi(t,e,n)},t.scale(1070)},Ry=uo(function(t){return rm(2/(1+t))});Ry.invert=lo(function(t){return 2*br(t/2)});var Iy=function(){return no(Ry).scale(124.75).clipAngle(179.999)},jy=uo(function(t){return(t=yr(t))&&t/em(t)});jy.invert=lo(function(t){return t});var Fy=function(){return no(jy).scale(79.4188).clipAngle(179.999)};co.invert=function(t,e){return[t,2*Gg(Zg(e))-Lg]};var qy=function(){return fo(co).scale(961/Bg)},Vy=function(){return io(ho).scale(109.5).parallels([30,30])};vo.invert=vo;var Uy=function(){return no(vo).scale(152.63)},Ly=function(){return io(go).scale(131.154).center([0,13.9389])};mo.invert=lo(Gg);var Hy=function(){return no(mo).scale(144.049).clipAngle(60)},By=function(){function t(){return i=o=null,a}var e,n,r,i,o,a,s=1,u=0,l=0,c=1,f=1,p=Jm,h=null,d=Jm;return a={stream:function(t){return i&&o===t?i:i=p(d(o=t))},clipExtent:function(i){return arguments.length?(d=null==i?(h=e=n=r=null,Jm):fi(h=+i[0][0],e=+i[0][1],n=+i[1][0],r=+i[1][1]),t()):null==h?null:[[h,e],[n,r]]},scale:function(e){return arguments.length?(p=yo((s=+e)*c,s*f,u,l),t()):s},translate:function(e){return arguments.length?(p=yo(s*c,s*f,u=+e[0],l=+e[1]),t()):[u,l]},reflectX:function(e){return arguments.length?(p=yo(s*(c=e?-1:1),s*f,u,l),t()):c<0},reflectY:function(e){return arguments.length?(p=yo(s*c,s*(f=e?-1:1),u,l),t()):f<0},fitExtent:function(t,e){return Zi(a,t,e)},fitSize:function(t,e){return Qi(a,t,e)}}};bo.invert=lo(br);var zy=function(){return no(bo).scale(249.5).clipAngle(90+qg)};$o.invert=lo(function(t){return 2*Gg(t)});var Wy=function(){return no($o).scale(250).clipAngle(142)};wo.invert=function(t,e){return[-e,2*Gg(Zg(t))-Lg]};var Yy=function(){var t=fo(wo),e=t.center,n=t.rotate;return t.center=function(t){return arguments.length?e([-t[1],t[0]]):(t=e(),[t[1],-t[0]])},t.rotate=function(t){return arguments.length?n([t[0],t[1],t.length>2?t[2]+90:90]):(t=n(),[t[0],t[1],t[2]-90])},n([0,0,90]).scale(159.155)},Gy=function(){function t(t){var o,a=0;t.eachAfter(function(t){var n=t.children;n?(t.x=_o(n),t.y=ko(n)):(t.x=o?a+=e(t,o):0,t.y=0,o=t)});var s=To(t),u=Eo(t),l=s.x-e(s,u)/2,c=u.x+e(u,s)/2;return t.eachAfter(i?function(e){e.x=(e.x-t.x)*n,e.y=(t.y-e.y)*r}:function(e){e.x=(e.x-l)/(c-l)*n,e.y=(1-(t.y?e.y/t.y:1))*r})}var e=xo,n=1,r=1,i=!1;return t.separation=function(n){return arguments.length?(e=n,t):e},t.size=function(e){return arguments.length?(i=!1,n=+e[0],r=+e[1],t):i?null:[n,r]},t.nodeSize=function(e){return arguments.length?(i=!0,n=+e[0],r=+e[1],t):i?[n,r]:null},t},Xy=function(){return this.eachAfter(Mo)},Ky=function(t){var e,n,r,i,o=this,a=[o];do for(e=a.reverse(),a=[];o=e.pop();)if(t(o),n=o.children)for(r=0,i=n.length;r<i;++r)a.push(n[r]);while(a.length);return this},Jy=function(t){for(var e,n,r=this,i=[r];r=i.pop();)if(t(r),e=r.children)for(n=e.length-1;n>=0;--n)i.push(e[n]);return this},Zy=function(t){for(var e,n,r,i=this,o=[i],a=[];i=o.pop();)if(a.push(i),e=i.children)for(n=0,r=e.length;n<r;++n)o.push(e[n]);for(;i=a.pop();)t(i);return this},Qy=function(t){return this.eachAfter(function(e){for(var n=+t(e.data)||0,r=e.children,i=r&&r.length;--i>=0;)n+=r[i].value;e.value=n})},tb=function(t){return this.eachBefore(function(e){e.children&&e.children.sort(t)})},eb=function(t){for(var e=this,n=Do(e,t),r=[e];e!==n;)e=e.parent,r.push(e);for(var i=r.length;t!==n;)r.splice(i,0,t),t=t.parent;return r},nb=function(){for(var t=this,e=[t];t=t.parent;)e.push(t);return e},rb=function(){var t=[];return this.each(function(e){t.push(e)}),t},ib=function(){var t=[];return this.eachBefore(function(e){e.children||t.push(e)}),t},ob=function(){var t=this,e=[];return t.each(function(n){n!==t&&e.push({source:n.parent,target:n})}),e};Io.prototype=Ao.prototype={constructor:Io,count:Xy,each:Ky,eachAfter:Zy,eachBefore:Jy,sum:Qy,sort:tb,path:eb,ancestors:nb,descendants:rb,leaves:ib,links:ob,copy:Oo};var ab=function(t){for(var e,n=(t=t.slice()).length,r=null,i=r;n;){var o=new jo(t[n-1]);i=i?i.next=o:r=o,t[e]=t[--n]}return{head:r,tail:i}},sb=function(t){return qo(ab(t),[])},ub=function(t){return Yo(t),t},lb=function(t){return function(){return t}},cb=function(){function t(t){return t.x=n/2,t.y=r/2,e?t.eachBefore(Zo(e)).eachAfter(Qo(i,.5)).eachBefore(ta(1)):t.eachBefore(Zo(Jo)).eachAfter(Qo(Ko,1)).eachAfter(Qo(i,t.r/Math.min(n,r))).eachBefore(ta(Math.min(n,r)/(2*t.r))),t}var e=null,n=1,r=1,i=Ko;return t.radius=function(n){return arguments.length?(e=Go(n),t):e},t.size=function(e){return arguments.length?(n=+e[0],r=+e[1],t):[n,r]},t.padding=function(e){return arguments.length?(i="function"==typeof e?e:lb(+e),t):i},t},fb=function(t){t.x0=Math.round(t.x0),t.y0=Math.round(t.y0),t.x1=Math.round(t.x1),t.y1=Math.round(t.y1)},pb=function(t,e,n,r,i){for(var o,a=t.children,s=-1,u=a.length,l=t.value&&(r-e)/t.value;++s<u;)o=a[s],o.y0=n,o.y1=i,o.x0=e,o.x1=e+=o.value*l},hb=function(){function t(t){var a=t.height+1;return t.x0=t.y0=i,t.x1=n,t.y1=r/a,t.eachBefore(e(r,a)),o&&t.eachBefore(fb),t}function e(t,e){return function(n){n.children&&pb(n,n.x0,t*(n.depth+1)/e,n.x1,t*(n.depth+2)/e);var r=n.x0,o=n.y0,a=n.x1-i,s=n.y1-i;a<r&&(r=a=(r+a)/2),s<o&&(o=s=(o+s)/2),n.x0=r,n.y0=o,n.x1=a,n.y1=s}}var n=1,r=1,i=0,o=!1;return t.round=function(e){return arguments.length?(o=!!e,t):o},t.size=function(e){return arguments.length?(n=+e[0],r=+e[1],t):[n,r]},t.padding=function(e){return arguments.length?(i=+e,t):i},t},db="$",vb={depth:-1},gb={},mb=function(){function t(t){var r,i,o,a,s,u,l,c=t.length,f=new Array(c),p={};for(i=0;i<c;++i)r=t[i],s=f[i]=new Io(r),null!=(u=e(r,i,t))&&(u+="")&&(l=db+(s.id=u),p[l]=l in p?gb:s);for(i=0;i<c;++i)if(s=f[i],u=n(t[i],i,t),null!=u&&(u+="")){if(a=p[db+u],!a)throw new Error("missing: "+u);if(a===gb)throw new Error("ambiguous: "+u);a.children?a.children.push(s):a.children=[s],s.parent=a}else{if(o)throw new Error("multiple roots");o=s}if(!o)throw new Error("no root");if(o.parent=vb,o.eachBefore(function(t){t.depth=t.parent.depth+1,--c}).eachBefore(Ro),o.parent=null,c>0)throw new Error("cycle");return o}var e=ea,n=na;return t.id=function(n){return arguments.length?(e=Xo(n),t):e},t.parentId=function(e){return arguments.length?(n=Xo(e),t):n},t};la.prototype=Object.create(Io.prototype);var yb=function(){function t(t){var r=ca(t);if(r.eachAfter(e),r.parent.m=-r.z,r.eachBefore(n),u)t.eachBefore(i);else{var l=t,c=t,f=t;t.eachBefore(function(t){t.x<l.x&&(l=t),t.x>c.x&&(c=t),t.depth>f.depth&&(f=t)});var p=l===c?1:o(l,c)/2,h=p-l.x,d=a/(c.x+p+h),v=s/(f.depth||1);t.eachBefore(function(t){t.x=(t.x+h)*d,t.y=t.depth*v})}return t}function e(t){var e=t.children,n=t.parent.children,i=t.i?n[t.i-1]:null;if(e){sa(t);var a=(e[0].z+e[e.length-1].z)/2;i?(t.z=i.z+o(t._,i._),t.m=t.z-a):t.z=a}else i&&(t.z=i.z+o(t._,i._));t.parent.A=r(t,i,t.parent.A||n[0])}function n(t){t._.x=t.z+t.parent.m,t.m+=t.parent.m}function r(t,e,n){if(e){for(var r,i=t,a=t,s=e,u=i.parent.children[0],l=i.m,c=a.m,f=s.m,p=u.m;s=oa(s),i=ia(i),s&&i;)u=ia(u),a=oa(a),a.a=t,r=s.z+f-i.z-l+o(s._,i._),r>0&&(aa(ua(s,t,n),t,r),l+=r,c+=r),f+=s.m,l+=i.m,p+=u.m,c+=a.m;
s&&!oa(a)&&(a.t=s,a.m+=f-c),i&&!ia(u)&&(u.t=i,u.m+=l-p,n=t)}return n}function i(t){t.x*=a,t.y=t.depth*s}var o=ra,a=1,s=1,u=null;return t.separation=function(e){return arguments.length?(o=e,t):o},t.size=function(e){return arguments.length?(u=!1,a=+e[0],s=+e[1],t):u?null:[a,s]},t.nodeSize=function(e){return arguments.length?(u=!0,a=+e[0],s=+e[1],t):u?[a,s]:null},t},bb=function(t,e,n,r,i){for(var o,a=t.children,s=-1,u=a.length,l=t.value&&(i-n)/t.value;++s<u;)o=a[s],o.x0=e,o.x1=r,o.y0=n,o.y1=n+=o.value*l},$b=(1+Math.sqrt(5))/2,wb=function t(e){function n(t,n,r,i,o){fa(e,t,n,r,i,o)}return n.ratio=function(e){return t((e=+e)>1?e:1)},n}($b),xb=function(){function t(t){return t.x0=t.y0=0,t.x1=i,t.y1=o,t.eachBefore(e),a=[0],r&&t.eachBefore(fb),t}function e(t){var e=a[t.depth],r=t.x0+e,i=t.y0+e,o=t.x1-e,p=t.y1-e;o<r&&(r=o=(r+o)/2),p<i&&(i=p=(i+p)/2),t.x0=r,t.y0=i,t.x1=o,t.y1=p,t.children&&(e=a[t.depth+1]=s(t)/2,r+=f(t)-e,i+=u(t)-e,o-=l(t)-e,p-=c(t)-e,o<r&&(r=o=(r+o)/2),p<i&&(i=p=(i+p)/2),n(t,r,i,o,p))}var n=wb,r=!1,i=1,o=1,a=[0],s=Ko,u=Ko,l=Ko,c=Ko,f=Ko;return t.round=function(e){return arguments.length?(r=!!e,t):r},t.size=function(e){return arguments.length?(i=+e[0],o=+e[1],t):[i,o]},t.tile=function(e){return arguments.length?(n=Xo(e),t):n},t.padding=function(e){return arguments.length?t.paddingInner(e).paddingOuter(e):t.paddingInner()},t.paddingInner=function(e){return arguments.length?(s="function"==typeof e?e:lb(+e),t):s},t.paddingOuter=function(e){return arguments.length?t.paddingTop(e).paddingRight(e).paddingBottom(e).paddingLeft(e):t.paddingTop()},t.paddingTop=function(e){return arguments.length?(u="function"==typeof e?e:lb(+e),t):u},t.paddingRight=function(e){return arguments.length?(l="function"==typeof e?e:lb(+e),t):l},t.paddingBottom=function(e){return arguments.length?(c="function"==typeof e?e:lb(+e),t):c},t.paddingLeft=function(e){return arguments.length?(f="function"==typeof e?e:lb(+e),t):f},t},_b=function(t,e,n,r,i){function o(t,e,n,r,i,a,s){if(t>=e-1){var l=u[t];return l.x0=r,l.y0=i,l.x1=a,l.y1=s,void 0}for(var f=c[t],p=n/2+f,h=t+1,d=e-1;h<d;){var v=h+d>>>1;c[v]<p?h=v+1:d=v}p-c[h-1]<c[h]-p&&t+1<h&&--h;var g=c[h]-f,m=n-g;if(a-r>s-i){var y=(r*m+a*g)/n;o(t,h,g,r,i,y,s),o(h,e,m,y,i,a,s)}else{var b=(i*m+s*g)/n;o(t,h,g,r,i,a,b),o(h,e,m,r,b,a,s)}}var a,s,u=t.children,l=u.length,c=new Array(l+1);for(c[0]=s=a=0;a<l;++a)c[a+1]=s+=u[a].value;o(0,l,t.value,e,n,r,i)},Cb=function(t,e,n,r,i){(1&t.depth?bb:pb)(t,e,n,r,i)},kb=function t(e){function n(t,n,r,i,o){if((a=t._squarify)&&a.ratio===e)for(var a,s,u,l,c,f=-1,p=a.length,h=t.value;++f<p;){for(s=a[f],u=s.children,l=s.value=0,c=u.length;l<c;++l)s.value+=u[l].value;s.dice?pb(s,n,r,i,r+=(o-r)*s.value/h):bb(s,n,r,n+=(i-n)*s.value/h,o),h-=s.value}else t._squarify=a=fa(e,t,n,r,i,o),a.ratio=e}return n.ratio=function(e){return t((e=+e)>1?e:1)},n}($b),Sb=function(t){for(var e,n=-1,r=t.length,i=t[r-1],o=0;++n<r;)e=i,i=t[n],o+=e[1]*i[0]-e[0]*i[1];return o/2},Tb=function(t){for(var e,n,r=-1,i=t.length,o=0,a=0,s=t[i-1],u=0;++r<i;)e=s,s=t[r],u+=n=e[0]*s[1]-s[0]*e[1],o+=(e[0]+s[0])*n,a+=(e[1]+s[1])*n;return u*=3,[o/u,a/u]},Eb=function(t,e,n){return(e[0]-t[0])*(n[1]-t[1])-(e[1]-t[1])*(n[0]-t[0])},Mb=function(t){if((n=t.length)<3)return null;var e,n,r=new Array(n),i=new Array(n);for(e=0;e<n;++e)r[e]=[+t[e][0],+t[e][1],e];for(r.sort(pa),e=0;e<n;++e)i[e]=[r[e][0],-r[e][1]];var o=ha(r),a=ha(i),s=a[0]===o[0],u=a[a.length-1]===o[o.length-1],l=[];for(e=o.length-1;e>=0;--e)l.push(t[r[o[e]][2]]);for(e=+s;e<a.length-u;++e)l.push(t[r[a[e]][2]]);return l},Db=function(t,e){for(var n,r,i=t.length,o=t[i-1],a=e[0],s=e[1],u=o[0],l=o[1],c=!1,f=0;f<i;++f)o=t[f],n=o[0],r=o[1],r>s!=l>s&&a<(u-n)*(s-r)/(l-r)+n&&(c=!c),u=n,l=r;return c},Ab=function(t){for(var e,n,r=-1,i=t.length,o=t[i-1],a=o[0],s=o[1],u=0;++r<i;)e=a,n=s,o=t[r],a=o[0],s=o[1],e-=a,n-=s,u+=Math.sqrt(e*e+n*n);return u},Ob=[].slice,Nb={};da.prototype=$a.prototype={constructor:da,defer:function(t){if("function"!=typeof t)throw new Error("invalid callback");if(this._call)throw new Error("defer after await");if(null!=this._error)return this;var e=Ob.call(arguments,1);return e.push(t),++this._waiting,this._tasks.push(e),va(this),this},abort:function(){return null==this._error&&ya(this,new Error("abort")),this},await:function(t){if("function"!=typeof t)throw new Error("invalid callback");if(this._call)throw new Error("multiple await");return this._call=function(e,n){t.apply(null,[e].concat(n))},ba(this),this},awaitAll:function(t){if("function"!=typeof t)throw new Error("invalid callback");if(this._call)throw new Error("multiple await");return this._call=t,ba(this),this}};var Pb=function(){return Math.random()},Rb=function t(e){function n(t,n){return t=null==t?0:+t,n=null==n?1:+n,1===arguments.length?(n=t,t=0):n-=t,function(){return e()*n+t}}return n.source=t,n}(Pb),Ib=function t(e){function n(t,n){var r,i;return t=null==t?0:+t,n=null==n?1:+n,function(){var o;if(null!=r)o=r,r=null;else do r=2*e()-1,o=2*e()-1,i=r*r+o*o;while(!i||i>1);return t+n*o*Math.sqrt(-2*Math.log(i)/i)}}return n.source=t,n}(Pb),jb=function t(e){function n(){var t=Ib.source(e).apply(this,arguments);return function(){return Math.exp(t())}}return n.source=t,n}(Pb),Fb=function t(e){function n(t){return function(){for(var n=0,r=0;r<t;++r)n+=e();return n}}return n.source=t,n}(Pb),qb=function t(e){function n(t){var n=Fb.source(e)(t);return function(){return n()/t}}return n.source=t,n}(Pb),Vb=function t(e){function n(t){return function(){return-Math.log(1-e())/t}}return n.source=t,n}(Pb),Ub=function(t,e){function n(t){var e,n=l.status;if(!n&&xa(l)||n>=200&&n<300||304===n){if(o)try{e=o.call(r,l)}catch(t){return void s.call("error",r,t)}else e=l;s.call("load",r,e)}else s.call("error",r,t)}var r,i,o,a,s=v("beforesend","progress","load","error"),u=Hn(),l=new XMLHttpRequest,c=null,f=null,p=0;if("undefined"==typeof XDomainRequest||"withCredentials"in l||!/^(http(s)?:)?\/\//.test(t)||(l=new XDomainRequest),"onload"in l?l.onload=l.onerror=l.ontimeout=n:l.onreadystatechange=function(t){l.readyState>3&&n(t)},l.onprogress=function(t){s.call("progress",r,t)},r={header:function(t,e){return t=(t+"").toLowerCase(),arguments.length<2?u.get(t):(null==e?u.remove(t):u.set(t,e+""),r)},mimeType:function(t){return arguments.length?(i=null==t?null:t+"",r):i},responseType:function(t){return arguments.length?(a=t,r):a},timeout:function(t){return arguments.length?(p=+t,r):p},user:function(t){return arguments.length<1?c:(c=null==t?null:t+"",r)},password:function(t){return arguments.length<1?f:(f=null==t?null:t+"",r)},response:function(t){return o=t,r},get:function(t,e){return r.send("GET",t,e)},post:function(t,e){return r.send("POST",t,e)},send:function(e,n,o){return l.open(e,t,!0,c,f),null==i||u.has("accept")||u.set("accept",i+",*/*"),l.setRequestHeader&&u.each(function(t,e){l.setRequestHeader(e,t)}),null!=i&&l.overrideMimeType&&l.overrideMimeType(i),null!=a&&(l.responseType=a),p>0&&(l.timeout=p),null==o&&"function"==typeof n&&(o=n,n=null),null!=o&&1===o.length&&(o=wa(o)),null!=o&&r.on("error",o).on("load",function(t){o(null,t)}),s.call("beforesend",r,l),l.send(null==n?null:n),r},abort:function(){return l.abort(),r},on:function(){var t=s.on.apply(s,arguments);return t===s?r:t}},null!=e){if("function"!=typeof e)throw new Error("invalid callback: "+e);return r.get(e)}return r},Lb=function(t,e){return function(n,r){var i=Ub(n).mimeType(t).response(e);if(null!=r){if("function"!=typeof r)throw new Error("invalid callback: "+r);return i.get(r)}return i}},Hb=Lb("text/html",function(t){return document.createRange().createContextualFragment(t.responseText)}),Bb=Lb("application/json",function(t){return JSON.parse(t.responseText)}),zb=Lb("text/plain",function(t){return t.responseText}),Wb=Lb("application/xml",function(t){var e=t.responseXML;if(!e)throw new Error("parse error");return e}),Yb=function(t,e){return function(n,r,i){arguments.length<3&&(i=r,r=null);var o=Ub(n).mimeType(t);return o.row=function(t){return arguments.length?o.response(_a(e,r=t)):r},o.row(r),i?o.get(i):o}},Gb=Yb("text/csv",lv),Xb=Yb("text/tab-separated-values",dv),Kb=Array.prototype,Jb=Kb.map,Zb=Kb.slice,Qb={name:"implicit"},t$=function(t){return function(){return t}},e$=function(t){return+t},n$=[0,1],r$=function(e,n,r){var o,a=e[0],s=e[e.length-1],u=i(a,s,null==n?10:n);switch(r=hr(null==r?",f":r),r.type){case"s":var l=Math.max(Math.abs(a),Math.abs(s));return null!=r.precision||isNaN(o=ig(u,l))||(r.precision=o),t.formatPrefix(r,l);case"":case"e":case"g":case"p":case"r":null!=r.precision||isNaN(o=og(u,Math.max(Math.abs(a),Math.abs(s))))||(r.precision=o-("e"===r.type));break;case"f":case"%":null!=r.precision||isNaN(o=rg(u))||(r.precision=o-2*("%"===r.type))}return t.format(r)},i$=function(t,e){t=t.slice();var n,r=0,i=t.length-1,o=t[r],a=t[i];return a<o&&(n=r,r=i,i=n,n=o,o=a,a=n),t[r]=e.floor(o),t[i]=e.ceil(a),t},o$=new Date,a$=new Date,s$=Ja(function(){},function(t,e){t.setTime(+t+e)},function(t,e){return e-t});s$.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?Ja(function(e){e.setTime(Math.floor(e/t)*t)},function(e,n){e.setTime(+e+n*t)},function(e,n){return(n-e)/t}):s$:null};var u$=s$.range,l$=1e3,c$=6e4,f$=36e5,p$=864e5,h$=6048e5,d$=Ja(function(t){t.setTime(Math.floor(t/l$)*l$)},function(t,e){t.setTime(+t+e*l$)},function(t,e){return(e-t)/l$},function(t){return t.getUTCSeconds()}),v$=d$.range,g$=Ja(function(t){t.setTime(Math.floor(t/c$)*c$)},function(t,e){t.setTime(+t+e*c$)},function(t,e){return(e-t)/c$},function(t){return t.getMinutes()}),m$=g$.range,y$=Ja(function(t){var e=t.getTimezoneOffset()*c$%f$;e<0&&(e+=f$),t.setTime(Math.floor((+t-e)/f$)*f$+e)},function(t,e){t.setTime(+t+e*f$)},function(t,e){return(e-t)/f$},function(t){return t.getHours()}),b$=y$.range,$$=Ja(function(t){t.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*c$)/p$},function(t){return t.getDate()-1}),w$=$$.range,x$=Za(0),_$=Za(1),C$=Za(2),k$=Za(3),S$=Za(4),T$=Za(5),E$=Za(6),M$=x$.range,D$=_$.range,A$=C$.range,O$=k$.range,N$=S$.range,P$=T$.range,R$=E$.range,I$=Ja(function(t){t.setDate(1),t.setHours(0,0,0,0)},function(t,e){t.setMonth(t.getMonth()+e)},function(t,e){return e.getMonth()-t.getMonth()+12*(e.getFullYear()-t.getFullYear())},function(t){return t.getMonth()}),j$=I$.range,F$=Ja(function(t){t.setMonth(0,1),t.setHours(0,0,0,0)},function(t,e){t.setFullYear(t.getFullYear()+e)},function(t,e){return e.getFullYear()-t.getFullYear()},function(t){return t.getFullYear()});F$.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Ja(function(e){e.setFullYear(Math.floor(e.getFullYear()/t)*t),e.setMonth(0,1),e.setHours(0,0,0,0)},function(e,n){e.setFullYear(e.getFullYear()+n*t)}):null};var q$=F$.range,V$=Ja(function(t){t.setUTCSeconds(0,0)},function(t,e){t.setTime(+t+e*c$)},function(t,e){return(e-t)/c$},function(t){return t.getUTCMinutes()}),U$=V$.range,L$=Ja(function(t){t.setUTCMinutes(0,0,0)},function(t,e){t.setTime(+t+e*f$)},function(t,e){return(e-t)/f$},function(t){return t.getUTCHours()}),H$=L$.range,B$=Ja(function(t){t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+e)},function(t,e){return(e-t)/p$},function(t){return t.getUTCDate()-1}),z$=B$.range,W$=Qa(0),Y$=Qa(1),G$=Qa(2),X$=Qa(3),K$=Qa(4),J$=Qa(5),Z$=Qa(6),Q$=W$.range,tw=Y$.range,ew=G$.range,nw=X$.range,rw=K$.range,iw=J$.range,ow=Z$.range,aw=Ja(function(t){t.setUTCDate(1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCMonth(t.getUTCMonth()+e)},function(t,e){return e.getUTCMonth()-t.getUTCMonth()+12*(e.getUTCFullYear()-t.getUTCFullYear())},function(t){return t.getUTCMonth()}),sw=aw.range,uw=Ja(function(t){t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCFullYear(t.getUTCFullYear()+e)},function(t,e){return e.getUTCFullYear()-t.getUTCFullYear()},function(t){return t.getUTCFullYear()});uw.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Ja(function(e){e.setUTCFullYear(Math.floor(e.getUTCFullYear()/t)*t),e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},function(e,n){e.setUTCFullYear(e.getUTCFullYear()+n*t)}):null};var lw,cw=uw.range,fw={"-":"",_:" ",0:"0"},pw=/^\s*\d+/,hw=/^%/,dw=/[\\\^\$\*\+\?\|\[\]\(\)\.\{\}]/g;Js({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]});var vw="%Y-%m-%dT%H:%M:%S.%LZ",gw=Date.prototype.toISOString?Zs:t.utcFormat(vw),mw=+new Date("2000-01-01T00:00:00.000Z")?Qs:t.utcParse(vw),yw=1e3,bw=60*yw,$w=60*bw,ww=24*$w,xw=7*ww,_w=30*ww,Cw=365*ww,kw=function(){return nu(F$,I$,x$,$$,y$,g$,d$,s$,t.timeFormat).domain([new Date(2e3,0,1),new Date(2e3,0,2)])},Sw=function(){return nu(uw,aw,W$,B$,L$,V$,d$,s$,t.utcFormat).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)])},Tw=function(t){return t.match(/.{6}/g).map(function(t){return"#"+t})},Ew=Tw("1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf"),Mw=Tw("393b795254a36b6ecf9c9ede6379398ca252b5cf6bcedb9c8c6d31bd9e39e7ba52e7cb94843c39ad494ad6616be7969c7b4173a55194ce6dbdde9ed6"),Dw=Tw("3182bd6baed69ecae1c6dbefe6550dfd8d3cfdae6bfdd0a231a35474c476a1d99bc7e9c0756bb19e9ac8bcbddcdadaeb636363969696bdbdbdd9d9d9"),Aw=Tw("1f77b4aec7e8ff7f0effbb782ca02c98df8ad62728ff98969467bdc5b0d58c564bc49c94e377c2f7b6d27f7f7fc7c7c7bcbd22dbdb8d17becf9edae5"),Ow=rh(Gt(300,.5,0),Gt(-240,.5,1)),Nw=rh(Gt(-100,.75,.35),Gt(80,1.5,.8)),Pw=rh(Gt(260,.75,.35),Gt(80,1.5,.8)),Rw=Gt(),Iw=function(t){(t<0||t>1)&&(t-=Math.floor(t));var e=Math.abs(t-.5);return Rw.h=360*t-100,Rw.s=1.5-1.5*e,Rw.l=.8-.9*e,Rw+""},jw=ru(Tw("44015444025645045745055946075a46085c460a5d460b5e470d60470e6147106347116447136548146748166848176948186a481a6c481b6d481c6e481d6f481f70482071482173482374482475482576482677482878482979472a7a472c7a472d7b472e7c472f7d46307e46327e46337f463480453581453781453882443983443a83443b84433d84433e85423f854240864241864142874144874045884046883f47883f48893e49893e4a893e4c8a3d4d8a3d4e8a3c4f8a3c508b3b518b3b528b3a538b3a548c39558c39568c38588c38598c375a8c375b8d365c8d365d8d355e8d355f8d34608d34618d33628d33638d32648e32658e31668e31678e31688e30698e306a8e2f6b8e2f6c8e2e6d8e2e6e8e2e6f8e2d708e2d718e2c718e2c728e2c738e2b748e2b758e2a768e2a778e2a788e29798e297a8e297b8e287c8e287d8e277e8e277f8e27808e26818e26828e26828e25838e25848e25858e24868e24878e23888e23898e238a8d228b8d228c8d228d8d218e8d218f8d21908d21918c20928c20928c20938c1f948c1f958b1f968b1f978b1f988b1f998a1f9a8a1e9b8a1e9c891e9d891f9e891f9f881fa0881fa1881fa1871fa28720a38620a48621a58521a68522a78522a88423a98324aa8325ab8225ac8226ad8127ad8128ae8029af7f2ab07f2cb17e2db27d2eb37c2fb47c31b57b32b67a34b67935b77937b87838b9773aba763bbb753dbc743fbc7340bd7242be7144bf7046c06f48c16e4ac16d4cc26c4ec36b50c46a52c56954c56856c66758c7655ac8645cc8635ec96260ca6063cb5f65cb5e67cc5c69cd5b6ccd5a6ece5870cf5773d05675d05477d1537ad1517cd2507fd34e81d34d84d44b86d54989d5488bd6468ed64590d74393d74195d84098d83e9bd93c9dd93ba0da39a2da37a5db36a8db34aadc32addc30b0dd2fb2dd2db5de2bb8de29bade28bddf26c0df25c2df23c5e021c8e020cae11fcde11dd0e11cd2e21bd5e21ad8e219dae319dde318dfe318e2e418e5e419e7e419eae51aece51befe51cf1e51df4e61ef6e620f8e621fbe723fde725")),Fw=ru(Tw("00000401000501010601010802010902020b02020d03030f03031204041405041606051806051a07061c08071e0907200a08220b09240c09260d0a290e0b2b100b2d110c2f120d31130d34140e36150e38160f3b180f3d19103f1a10421c10441d11471e114920114b21114e22115024125325125527125829115a2a115c2c115f2d11612f116331116533106734106936106b38106c390f6e3b0f703d0f713f0f72400f74420f75440f764510774710784910784a10794c117a4e117b4f127b51127c52137c54137d56147d57157e59157e5a167e5c167f5d177f5f187f601880621980641a80651a80671b80681c816a1c816b1d816d1d816e1e81701f81721f817320817521817621817822817922827b23827c23827e24828025828125818326818426818627818827818928818b29818c29818e2a81902a81912b81932b80942c80962c80982d80992d809b2e7f9c2e7f9e2f7fa02f7fa1307ea3307ea5317ea6317da8327daa337dab337cad347cae347bb0357bb2357bb3367ab5367ab73779b83779ba3878bc3978bd3977bf3a77c03a76c23b75c43c75c53c74c73d73c83e73ca3e72cc3f71cd4071cf4070d0416fd2426fd3436ed5446dd6456cd8456cd9466bdb476adc4869de4968df4a68e04c67e24d66e34e65e44f64e55064e75263e85362e95462ea5661eb5760ec5860ed5a5fee5b5eef5d5ef05f5ef1605df2625df2645cf3655cf4675cf4695cf56b5cf66c5cf66e5cf7705cf7725cf8745cf8765cf9785df9795df97b5dfa7d5efa7f5efa815ffb835ffb8560fb8761fc8961fc8a62fc8c63fc8e64fc9065fd9266fd9467fd9668fd9869fd9a6afd9b6bfe9d6cfe9f6dfea16efea36ffea571fea772fea973feaa74feac76feae77feb078feb27afeb47bfeb67cfeb77efeb97ffebb81febd82febf84fec185fec287fec488fec68afec88cfeca8dfecc8ffecd90fecf92fed194fed395fed597fed799fed89afdda9cfddc9efddea0fde0a1fde2a3fde3a5fde5a7fde7a9fde9aafdebacfcecaefceeb0fcf0b2fcf2b4fcf4b6fcf6b8fcf7b9fcf9bbfcfbbdfcfdbf")),qw=ru(Tw("00000401000501010601010802010a02020c02020e03021004031204031405041706041907051b08051d09061f0a07220b07240c08260d08290e092b10092d110a30120a32140b34150b37160b39180c3c190c3e1b0c411c0c431e0c451f0c48210c4a230c4c240c4f260c51280b53290b552b0b572d0b592f0a5b310a5c320a5e340a5f3609613809623909633b09643d09653e0966400a67420a68440a68450a69470b6a490b6a4a0c6b4c0c6b4d0d6c4f0d6c510e6c520e6d540f6d550f6d57106e59106e5a116e5c126e5d126e5f136e61136e62146e64156e65156e67166e69166e6a176e6c186e6d186e6f196e71196e721a6e741a6e751b6e771c6d781c6d7a1d6d7c1d6d7d1e6d7f1e6c801f6c82206c84206b85216b87216b88226a8a226a8c23698d23698f24699025689225689326679526679727669827669a28659b29649d29649f2a63a02a63a22b62a32c61a52c60a62d60a82e5fa92e5eab2f5ead305dae305cb0315bb1325ab3325ab43359b63458b73557b93556ba3655bc3754bd3853bf3952c03a51c13a50c33b4fc43c4ec63d4dc73e4cc83f4bca404acb4149cc4248ce4347cf4446d04545d24644d34743d44842d54a41d74b3fd84c3ed94d3dda4e3cdb503bdd513ade5238df5337e05536e15635e25734e35933e45a31e55c30e65d2fe75e2ee8602de9612bea632aeb6429eb6628ec6726ed6925ee6a24ef6c23ef6e21f06f20f1711ff1731df2741cf3761bf37819f47918f57b17f57d15f67e14f68013f78212f78410f8850ff8870ef8890cf98b0bf98c0af98e09fa9008fa9207fa9407fb9606fb9706fb9906fb9b06fb9d07fc9f07fca108fca309fca50afca60cfca80dfcaa0ffcac11fcae12fcb014fcb216fcb418fbb61afbb81dfbba1ffbbc21fbbe23fac026fac228fac42afac62df9c72ff9c932f9cb35f8cd37f8cf3af7d13df7d340f6d543f6d746f5d949f5db4cf4dd4ff4df53f4e156f3e35af3e55df2e661f2e865f2ea69f1ec6df1ed71f1ef75f1f179f2f27df2f482f3f586f3f68af4f88ef5f992f6fa96f8fb9af9fc9dfafda1fcffa4")),Vw=ru(Tw("0d088710078813078916078a19068c1b068d1d068e20068f2206902406912605912805922a05932c05942e05952f059631059733059735049837049938049a3a049a3c049b3e049c3f049c41049d43039e44039e46039f48039f4903a04b03a14c02a14e02a25002a25102a35302a35502a45601a45801a45901a55b01a55c01a65e01a66001a66100a76300a76400a76600a76700a86900a86a00a86c00a86e00a86f00a87100a87201a87401a87501a87701a87801a87a02a87b02a87d03a87e03a88004a88104a78305a78405a78606a68707a68808a68a09a58b0aa58d0ba58e0ca48f0da4910ea3920fa39410a29511a19613a19814a099159f9a169f9c179e9d189d9e199da01a9ca11b9ba21d9aa31e9aa51f99a62098a72197a82296aa2395ab2494ac2694ad2793ae2892b02991b12a90b22b8fb32c8eb42e8db52f8cb6308bb7318ab83289ba3388bb3488bc3587bd3786be3885bf3984c03a83c13b82c23c81c33d80c43e7fc5407ec6417dc7427cc8437bc9447aca457acb4679cc4778cc4977cd4a76ce4b75cf4c74d04d73d14e72d24f71d35171d45270d5536fd5546ed6556dd7566cd8576bd9586ada5a6ada5b69db5c68dc5d67dd5e66de5f65de6164df6263e06363e16462e26561e26660e3685fe4695ee56a5de56b5de66c5ce76e5be76f5ae87059e97158e97257ea7457eb7556eb7655ec7754ed7953ed7a52ee7b51ef7c51ef7e50f07f4ff0804ef1814df1834cf2844bf3854bf3874af48849f48948f58b47f58c46f68d45f68f44f79044f79143f79342f89441f89540f9973ff9983ef99a3efa9b3dfa9c3cfa9e3bfb9f3afba139fba238fca338fca537fca636fca835fca934fdab33fdac33fdae32fdaf31fdb130fdb22ffdb42ffdb52efeb72dfeb82cfeba2cfebb2bfebd2afebe2afec029fdc229fdc328fdc527fdc627fdc827fdca26fdcb26fccd25fcce25fcd025fcd225fbd324fbd524fbd724fad824fada24f9dc24f9dd25f8df25f8e125f7e225f7e425f6e626f6e826f5e926f5eb27f4ed27f3ee27f3f027f2f227f1f426f1f525f0f724f0f921")),Uw=function(t){return function(){return t}},Lw=Math.abs,Hw=Math.atan2,Bw=Math.cos,zw=Math.max,Ww=Math.min,Yw=Math.sin,Gw=Math.sqrt,Xw=1e-12,Kw=Math.PI,Jw=Kw/2,Zw=2*Kw,Qw=function(){function t(){var t,l,c=+e.apply(this,arguments),f=+n.apply(this,arguments),p=o.apply(this,arguments)-Jw,h=a.apply(this,arguments)-Jw,d=Lw(h-p),v=h>p;if(u||(u=t=In()),f<c&&(l=f,f=c,c=l),f>Xw)if(d>Zw-Xw)u.moveTo(f*Bw(p),f*Yw(p)),u.arc(0,0,f,p,h,!v),c>Xw&&(u.moveTo(c*Bw(h),c*Yw(h)),u.arc(0,0,c,h,p,v));else{var g,m,y=p,b=h,$=p,w=h,x=d,_=d,C=s.apply(this,arguments)/2,k=C>Xw&&(i?+i.apply(this,arguments):Gw(c*c+f*f)),S=Ww(Lw(f-c)/2,+r.apply(this,arguments)),T=S,E=S;if(k>Xw){var M=au(k/c*Yw(C)),D=au(k/f*Yw(C));(x-=2*M)>Xw?(M*=v?1:-1,$+=M,w-=M):(x=0,$=w=(p+h)/2),(_-=2*D)>Xw?(D*=v?1:-1,y+=D,b-=D):(_=0,y=b=(p+h)/2)}var A=f*Bw(y),O=f*Yw(y),N=c*Bw(w),P=c*Yw(w);if(S>Xw){var R=f*Bw(b),I=f*Yw(b),j=c*Bw($),F=c*Yw($);if(d<Kw){var q=x>Xw?pu(A,O,j,F,R,I,N,P):[N,P],V=A-q[0],U=O-q[1],L=R-q[0],H=I-q[1],B=1/Yw(ou((V*L+U*H)/(Gw(V*V+U*U)*Gw(L*L+H*H)))/2),z=Gw(q[0]*q[0]+q[1]*q[1]);T=Ww(S,(c-z)/(B-1)),E=Ww(S,(f-z)/(B+1))}}_>Xw?E>Xw?(g=hu(j,F,A,O,f,E,v),m=hu(R,I,N,P,f,E,v),u.moveTo(g.cx+g.x01,g.cy+g.y01),E<S?u.arc(g.cx,g.cy,E,Hw(g.y01,g.x01),Hw(m.y01,m.x01),!v):(u.arc(g.cx,g.cy,E,Hw(g.y01,g.x01),Hw(g.y11,g.x11),!v),u.arc(0,0,f,Hw(g.cy+g.y11,g.cx+g.x11),Hw(m.cy+m.y11,m.cx+m.x11),!v),u.arc(m.cx,m.cy,E,Hw(m.y11,m.x11),Hw(m.y01,m.x01),!v))):(u.moveTo(A,O),u.arc(0,0,f,y,b,!v)):u.moveTo(A,O),c>Xw&&x>Xw?T>Xw?(g=hu(N,P,R,I,c,-T,v),m=hu(A,O,j,F,c,-T,v),u.lineTo(g.cx+g.x01,g.cy+g.y01),T<S?u.arc(g.cx,g.cy,T,Hw(g.y01,g.x01),Hw(m.y01,m.x01),!v):(u.arc(g.cx,g.cy,T,Hw(g.y01,g.x01),Hw(g.y11,g.x11),!v),u.arc(0,0,c,Hw(g.cy+g.y11,g.cx+g.x11),Hw(m.cy+m.y11,m.cx+m.x11),v),u.arc(m.cx,m.cy,T,Hw(m.y11,m.x11),Hw(m.y01,m.x01),!v))):u.arc(0,0,c,w,$,v):u.lineTo(N,P)}else u.moveTo(0,0);if(u.closePath(),t)return u=null,t+""||null}var e=su,n=uu,r=Uw(0),i=null,o=lu,a=cu,s=fu,u=null;return t.centroid=function(){var t=(+e.apply(this,arguments)+ +n.apply(this,arguments))/2,r=(+o.apply(this,arguments)+ +a.apply(this,arguments))/2-Kw/2;return[Bw(r)*t,Yw(r)*t]},t.innerRadius=function(n){return arguments.length?(e="function"==typeof n?n:Uw(+n),t):e},t.outerRadius=function(e){return arguments.length?(n="function"==typeof e?e:Uw(+e),t):n},t.cornerRadius=function(e){return arguments.length?(r="function"==typeof e?e:Uw(+e),t):r},t.padRadius=function(e){return arguments.length?(i=null==e?null:"function"==typeof e?e:Uw(+e),t):i},t.startAngle=function(e){return arguments.length?(o="function"==typeof e?e:Uw(+e),t):o},t.endAngle=function(e){return arguments.length?(a="function"==typeof e?e:Uw(+e),t):a},t.padAngle=function(e){return arguments.length?(s="function"==typeof e?e:Uw(+e),t):s},t.context=function(e){return arguments.length?(u=null==e?null:e,t):u},t};du.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:this._context.lineTo(t,e)}}};var tx=function(t){return new du(t)},ex=function(){function t(t){var s,u,l,c=t.length,f=!1;for(null==i&&(a=o(l=In())),s=0;s<=c;++s)!(s<c&&r(u=t[s],s,t))===f&&((f=!f)?a.lineStart():a.lineEnd()),f&&a.point(+e(u,s,t),+n(u,s,t));if(l)return a=null,l+""||null}var e=vu,n=gu,r=Uw(!0),i=null,o=tx,a=null;return t.x=function(n){return arguments.length?(e="function"==typeof n?n:Uw(+n),t):e},t.y=function(e){return arguments.length?(n="function"==typeof e?e:Uw(+e),t):n},t.defined=function(e){return arguments.length?(r="function"==typeof e?e:Uw(!!e),t):r},t.curve=function(e){return arguments.length?(o=e,null!=i&&(a=o(i)),t):o},t.context=function(e){return arguments.length?(null==e?i=a=null:a=o(i=e),t):i},t},nx=function(){function t(t){var e,c,f,p,h,d=t.length,v=!1,g=new Array(d),m=new Array(d);for(null==s&&(l=u(h=In())),e=0;e<=d;++e){if(!(e<d&&a(p=t[e],e,t))===v)if(v=!v)c=e,l.areaStart(),l.lineStart();else{for(l.lineEnd(),l.lineStart(),f=e-1;f>=c;--f)l.point(g[f],m[f]);l.lineEnd(),l.areaEnd()}v&&(g[e]=+n(p,e,t),m[e]=+i(p,e,t),l.point(r?+r(p,e,t):g[e],o?+o(p,e,t):m[e]))}if(h)return l=null,h+""||null}function e(){return ex().defined(a).curve(u).context(s)}var n=vu,r=null,i=Uw(0),o=gu,a=Uw(!0),s=null,u=tx,l=null;return t.x=function(e){return arguments.length?(n="function"==typeof e?e:Uw(+e),r=null,t):n},t.x0=function(e){return arguments.length?(n="function"==typeof e?e:Uw(+e),t):n},t.x1=function(e){return arguments.length?(r=null==e?null:"function"==typeof e?e:Uw(+e),t):r},t.y=function(e){return arguments.length?(i="function"==typeof e?e:Uw(+e),o=null,t):i},t.y0=function(e){return arguments.length?(i="function"==typeof e?e:Uw(+e),t):i},t.y1=function(e){return arguments.length?(o=null==e?null:"function"==typeof e?e:Uw(+e),t):o},t.lineX0=t.lineY0=function(){return e().x(n).y(i)},t.lineY1=function(){return e().x(n).y(o)},t.lineX1=function(){return e().x(r).y(i)},t.defined=function(e){return arguments.length?(a="function"==typeof e?e:Uw(!!e),t):a},t.curve=function(e){return arguments.length?(u=e,null!=s&&(l=u(s)),t):u},t.context=function(e){return arguments.length?(null==e?s=l=null:l=u(s=e),t):s},t},rx=function(t,e){return e<t?-1:e>t?1:e>=t?0:NaN},ix=function(t){return t},ox=function(){function t(t){var s,u,l,c,f,p=t.length,h=0,d=new Array(p),v=new Array(p),g=+i.apply(this,arguments),m=Math.min(Zw,Math.max(-Zw,o.apply(this,arguments)-g)),y=Math.min(Math.abs(m)/p,a.apply(this,arguments)),b=y*(m<0?-1:1);for(s=0;s<p;++s)(f=v[d[s]=s]=+e(t[s],s,t))>0&&(h+=f);for(null!=n?d.sort(function(t,e){return n(v[t],v[e])}):null!=r&&d.sort(function(e,n){return r(t[e],t[n])}),s=0,l=h?(m-p*b)/h:0;s<p;++s,g=c)u=d[s],f=v[u],c=g+(f>0?f*l:0)+b,v[u]={data:t[u],index:s,value:f,startAngle:g,endAngle:c,padAngle:y};return v}var e=ix,n=rx,r=null,i=Uw(0),o=Uw(Zw),a=Uw(0);return t.value=function(n){return arguments.length?(e="function"==typeof n?n:Uw(+n),t):e},t.sortValues=function(e){return arguments.length?(n=e,r=null,t):n},t.sort=function(e){return arguments.length?(r=e,n=null,t):r},t.startAngle=function(e){return arguments.length?(i="function"==typeof e?e:Uw(+e),t):i},t.endAngle=function(e){return arguments.length?(o="function"==typeof e?e:Uw(+e),t):o},t.padAngle=function(e){return arguments.length?(a="function"==typeof e?e:Uw(+e),t):a},t},ax=yu(tx);mu.prototype={areaStart:function(){this._curve.areaStart()},areaEnd:function(){this._curve.areaEnd()},lineStart:function(){this._curve.lineStart()},lineEnd:function(){this._curve.lineEnd()},point:function(t,e){this._curve.point(e*Math.sin(t),e*-Math.cos(t))}};var sx=function(){return bu(ex().curve(ax))},ux=function(){var t=nx().curve(ax),e=t.curve,n=t.lineX0,r=t.lineX1,i=t.lineY0,o=t.lineY1;return t.angle=t.x,delete t.x,t.startAngle=t.x0,delete t.x0,t.endAngle=t.x1,delete t.x1,t.radius=t.y,delete t.y,t.innerRadius=t.y0,delete t.y0,t.outerRadius=t.y1,delete t.y1,t.lineStartAngle=function(){return bu(n())},delete t.lineX0,t.lineEndAngle=function(){return bu(r())},delete t.lineX1,t.lineInnerRadius=function(){return bu(i())},delete t.lineY0,t.lineOuterRadius=function(){return bu(o())},delete t.lineY1,t.curve=function(t){return arguments.length?e(yu(t)):e()._curve},t},lx=Array.prototype.slice,cx=function(t,e){return[(e=+e)*Math.cos(t-=Math.PI/2),e*Math.sin(t)]},fx={draw:function(t,e){var n=Math.sqrt(e/Kw);t.moveTo(n,0),t.arc(0,0,n,0,Zw)}},px={draw:function(t,e){var n=Math.sqrt(e/5)/2;t.moveTo(-3*n,-n),t.lineTo(-n,-n),t.lineTo(-n,-3*n),t.lineTo(n,-3*n),t.lineTo(n,-n),t.lineTo(3*n,-n),t.lineTo(3*n,n),t.lineTo(n,n),t.lineTo(n,3*n),t.lineTo(-n,3*n),t.lineTo(-n,n),t.lineTo(-3*n,n),t.closePath()}},hx=Math.sqrt(1/3),dx=2*hx,vx={draw:function(t,e){var n=Math.sqrt(e/dx),r=n*hx;t.moveTo(0,-n),t.lineTo(r,0),t.lineTo(0,n),t.lineTo(-r,0),t.closePath()}},gx=.8908130915292852,mx=Math.sin(Kw/10)/Math.sin(7*Kw/10),yx=Math.sin(Zw/10)*mx,bx=-Math.cos(Zw/10)*mx,$x={draw:function(t,e){var n=Math.sqrt(e*gx),r=yx*n,i=bx*n;t.moveTo(0,-n),t.lineTo(r,i);for(var o=1;o<5;++o){var a=Zw*o/5,s=Math.cos(a),u=Math.sin(a);t.lineTo(u*n,-s*n),t.lineTo(s*r-u*i,u*r+s*i)}t.closePath()}},wx={draw:function(t,e){var n=Math.sqrt(e),r=-n/2;t.rect(r,r,n,n)}},xx=Math.sqrt(3),_x={draw:function(t,e){var n=-Math.sqrt(e/(3*xx));t.moveTo(0,2*n),t.lineTo(-xx*n,-n),t.lineTo(xx*n,-n),t.closePath()}},Cx=-.5,kx=Math.sqrt(3)/2,Sx=1/Math.sqrt(12),Tx=3*(Sx/2+1),Ex={draw:function(t,e){var n=Math.sqrt(e/Tx),r=n/2,i=n*Sx,o=r,a=n*Sx+n,s=-o,u=a;t.moveTo(r,i),t.lineTo(o,a),t.lineTo(s,u),t.lineTo(Cx*r-kx*i,kx*r+Cx*i),t.lineTo(Cx*o-kx*a,kx*o+Cx*a),t.lineTo(Cx*s-kx*u,kx*s+Cx*u),t.lineTo(Cx*r+kx*i,Cx*i-kx*r),t.lineTo(Cx*o+kx*a,Cx*a-kx*o),t.lineTo(Cx*s+kx*u,Cx*u-kx*s),t.closePath()}},Mx=[fx,px,vx,wx,$x,_x,Ex],Dx=function(){function t(){var t;if(r||(r=t=In()),e.apply(this,arguments).draw(r,+n.apply(this,arguments)),t)return r=null,t+""||null}var e=Uw(fx),n=Uw(64),r=null;return t.type=function(n){return arguments.length?(e="function"==typeof n?n:Uw(n),t):e},t.size=function(e){return arguments.length?(n="function"==typeof e?e:Uw(+e),t):n},t.context=function(e){return arguments.length?(r=null==e?null:e,t):r},t},Ax=function(){};Du.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:Mu(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:Mu(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}};var Ox=function(t){return new Du(t)};Au.prototype={areaStart:Ax,areaEnd:Ax,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x2,this._y2),this._context.closePath();break;case 2:this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break;case 3:this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x2=t,this._y2=e;break;case 1:this._point=2,this._x3=t,this._y3=e;break;case 2:this._point=3,this._x4=t,this._y4=e,this._context.moveTo((this._x0+4*this._x1+t)/6,(this._y0+4*this._y1+e)/6);break;default:Mu(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}};var Nx=function(t){return new Au(t)};Ou.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+t)/6,r=(this._y0+4*this._y1+e)/6;this._line?this._context.lineTo(n,r):this._context.moveTo(n,r);break;case 3:this._point=4;default:Mu(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}};var Px=function(t){return new Ou(t)};Nu.prototype={lineStart:function(){this._x=[],this._y=[],this._basis.lineStart()},lineEnd:function(){var t=this._x,e=this._y,n=t.length-1;if(n>0)for(var r,i=t[0],o=e[0],a=t[n]-i,s=e[n]-o,u=-1;++u<=n;)r=u/n,this._basis.point(this._beta*t[u]+(1-this._beta)*(i+r*a),this._beta*e[u]+(1-this._beta)*(o+r*s));this._x=this._y=null,this._basis.lineEnd()},point:function(t,e){
this._x.push(+t),this._y.push(+e)}};var Rx=function t(e){function n(t){return 1===e?new Du(t):new Nu(t,e)}return n.beta=function(e){return t(+e)},n}(.85);Ru.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:Pu(this,this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2,this._x1=t,this._y1=e;break;case 2:this._point=3;default:Pu(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var Ix=function t(e){function n(t){return new Ru(t,e)}return n.tension=function(e){return t(+e)},n}(0);Iu.prototype={areaStart:Ax,areaEnd:Ax,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Pu(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var jx=function t(e){function n(t){return new Iu(t,e)}return n.tension=function(e){return t(+e)},n}(0);ju.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Pu(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var Fx=function t(e){function n(t){return new ju(t,e)}return n.tension=function(e){return t(+e)},n}(0);qu.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:this.point(this._x2,this._y2)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3;default:Fu(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var qx=function t(e){function n(t){return e?new qu(t,e):new Ru(t,0)}return n.alpha=function(e){return t(+e)},n}(.5);Vu.prototype={areaStart:Ax,areaEnd:Ax,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Fu(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var Vx=function t(e){function n(t){return e?new Vu(t,e):new Iu(t,0)}return n.alpha=function(e){return t(+e)},n}(.5);Uu.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Fu(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}};var Ux=function t(e){function n(t){return e?new Uu(t,e):new ju(t,0)}return n.alpha=function(e){return t(+e)},n}(.5);Lu.prototype={areaStart:Ax,areaEnd:Ax,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(t,e){t=+t,e=+e,this._point?this._context.lineTo(t,e):(this._point=1,this._context.moveTo(t,e))}};var Lx=function(t){return new Lu(t)};Yu.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:Wu(this,this._t0,zu(this,this._t0))}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){var n=NaN;if(t=+t,e=+e,t!==this._x1||e!==this._y1){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,Wu(this,zu(this,n=Bu(this,t,e)),n);break;default:Wu(this,this._t0,n=Bu(this,t,e))}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e,this._t0=n}}},(Gu.prototype=Object.create(Yu.prototype)).point=function(t,e){Yu.prototype.point.call(this,e,t)},Xu.prototype={moveTo:function(t,e){this._context.moveTo(e,t)},closePath:function(){this._context.closePath()},lineTo:function(t,e){this._context.lineTo(e,t)},bezierCurveTo:function(t,e,n,r,i,o){this._context.bezierCurveTo(e,t,r,n,o,i)}},Zu.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var t=this._x,e=this._y,n=t.length;if(n)if(this._line?this._context.lineTo(t[0],e[0]):this._context.moveTo(t[0],e[0]),2===n)this._context.lineTo(t[1],e[1]);else for(var r=Qu(t),i=Qu(e),o=0,a=1;a<n;++o,++a)this._context.bezierCurveTo(r[0][o],i[0][o],r[1][o],i[1][o],t[a],e[a]);(this._line||0!==this._line&&1===n)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(t,e){this._x.push(+t),this._y.push(+e)}};var Hx=function(t){return new Zu(t)};tl.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&2===this._point&&this._context.lineTo(this._x,this._y),(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:if(this._t<=0)this._context.lineTo(this._x,e),this._context.lineTo(t,e);else{var n=this._x*(1-this._t)+t*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,e)}}this._x=t,this._y=e}};var Bx=function(t){return new tl(t,.5)},zx=function(t,e){if((i=t.length)>1)for(var n,r,i,o=1,a=t[e[0]],s=a.length;o<i;++o)for(r=a,a=t[e[o]],n=0;n<s;++n)a[n][1]+=a[n][0]=isNaN(r[n][1])?r[n][0]:r[n][1]},Wx=function(t){for(var e=t.length,n=new Array(e);--e>=0;)n[e]=e;return n},Yx=function(){function t(t){var o,a,s=e.apply(this,arguments),u=t.length,l=s.length,c=new Array(l);for(o=0;o<l;++o){for(var f,p=s[o],h=c[o]=new Array(u),d=0;d<u;++d)h[d]=f=[0,+i(t[d],p,d,t)],f.data=t[d];h.key=p}for(o=0,a=n(c);o<l;++o)c[a[o]].index=o;return r(c,a),c}var e=Uw([]),n=Wx,r=zx,i=rl;return t.keys=function(n){return arguments.length?(e="function"==typeof n?n:Uw(lx.call(n)),t):e},t.value=function(e){return arguments.length?(i="function"==typeof e?e:Uw(+e),t):i},t.order=function(e){return arguments.length?(n=null==e?Wx:"function"==typeof e?e:Uw(lx.call(e)),t):n},t.offset=function(e){return arguments.length?(r=null==e?zx:e,t):r},t},Gx=function(t,e){if((r=t.length)>0){for(var n,r,i,o=0,a=t[0].length;o<a;++o){for(i=n=0;n<r;++n)i+=t[n][o][1]||0;if(i)for(n=0;n<r;++n)t[n][o][1]/=i}zx(t,e)}},Xx=function(t,e){if((s=t.length)>1)for(var n,r,i,o,a,s,u=0,l=t[e[0]].length;u<l;++u)for(o=a=0,n=0;n<s;++n)(i=(r=t[e[n]][u])[1]-r[0])>=0?(r[0]=o,r[1]=o+=i):i<0?(r[1]=a,r[0]=a+=i):r[0]=o},Kx=function(t,e){if((n=t.length)>0){for(var n,r=0,i=t[e[0]],o=i.length;r<o;++r){for(var a=0,s=0;a<n;++a)s+=t[a][r][1]||0;i[r][1]+=i[r][0]=-s/2}zx(t,e)}},Jx=function(t,e){if((i=t.length)>0&&(r=(n=t[e[0]]).length)>0){for(var n,r,i,o=0,a=1;a<r;++a){for(var s=0,u=0,l=0;s<i;++s){for(var c=t[e[s]],f=c[a][1]||0,p=c[a-1][1]||0,h=(f-p)/2,d=0;d<s;++d){var v=t[e[d]],g=v[a][1]||0,m=v[a-1][1]||0;h+=g-m}u+=f,l+=h*f}n[a-1][1]+=n[a-1][0]=o,u&&(o-=l/u)}n[a-1][1]+=n[a-1][0]=o,zx(t,e)}},Zx=function(t){var e=t.map(il);return Wx(t).sort(function(t,n){return e[t]-e[n]})},Qx=function(t){return Zx(t).reverse()},t_=function(t){var e,n,r=t.length,i=t.map(il),o=Wx(t).sort(function(t,e){return i[e]-i[t]}),a=0,s=0,u=[],l=[];for(e=0;e<r;++e)n=o[e],a<s?(a+=i[n],u.push(n)):(s+=i[n],l.push(n));return l.reverse().concat(u)},e_=function(t){return Wx(t).reverse()},n_=function(t){return function(){return t}};sl.prototype={constructor:sl,insert:function(t,e){var n,r,i;if(t){if(e.P=t,e.N=t.N,t.N&&(t.N.P=e),t.N=e,t.R){for(t=t.R;t.L;)t=t.L;t.L=e}else t.R=e;n=t}else this._?(t=fl(this._),e.P=null,e.N=t,t.P=t.L=e,n=t):(e.P=e.N=null,this._=e,n=null);for(e.L=e.R=null,e.U=n,e.C=!0,t=e;n&&n.C;)r=n.U,n===r.L?(i=r.R,i&&i.C?(n.C=i.C=!1,r.C=!0,t=r):(t===n.R&&(ll(this,n),t=n,n=t.U),n.C=!1,r.C=!0,cl(this,r))):(i=r.L,i&&i.C?(n.C=i.C=!1,r.C=!0,t=r):(t===n.L&&(cl(this,n),t=n,n=t.U),n.C=!1,r.C=!0,ll(this,r))),n=t.U;this._.C=!1},remove:function(t){t.N&&(t.N.P=t.P),t.P&&(t.P.N=t.N),t.N=t.P=null;var e,n,r,i=t.U,o=t.L,a=t.R;if(n=o?a?fl(a):o:a,i?i.L===t?i.L=n:i.R=n:this._=n,o&&a?(r=n.C,n.C=t.C,n.L=o,o.U=n,n!==a?(i=n.U,n.U=t.U,t=n.R,i.L=t,n.R=a,a.U=n):(n.U=i,i=n,t=n.R)):(r=t.C,t=n),t&&(t.U=i),!r){if(t&&t.C)return void(t.C=!1);do{if(t===this._)break;if(t===i.L){if(e=i.R,e.C&&(e.C=!1,i.C=!0,ll(this,i),e=i.R),e.L&&e.L.C||e.R&&e.R.C){e.R&&e.R.C||(e.L.C=!1,e.C=!0,cl(this,e),e=i.R),e.C=i.C,i.C=e.R.C=!1,ll(this,i),t=this._;break}}else if(e=i.L,e.C&&(e.C=!1,i.C=!0,cl(this,i),e=i.L),e.L&&e.L.C||e.R&&e.R.C){e.L&&e.L.C||(e.R.C=!1,e.C=!0,ll(this,e),e=i.L),e.C=i.C,i.C=e.L.C=!1,cl(this,i),t=this._;break}e.C=!0,t=i,i=i.U}while(!t.C);t&&(t.C=!1)}}};var r_,i_,o_,a_,s_,u_=[],l_=[],c_=1e-6,f_=1e-12;Il.prototype={constructor:Il,polygons:function(){var t=this.edges;return this.cells.map(function(e){var n=e.halfedges.map(function(n){return $l(e,t[n])});return n.data=e.site.data,n})},triangles:function(){var t=[],e=this.edges;return this.cells.forEach(function(n,r){if(o=(i=n.halfedges).length)for(var i,o,a,s=n.site,u=-1,l=e[i[o-1]],c=l.left===s?l.right:l.left;++u<o;)a=c,l=e[i[u]],c=l.left===s?l.right:l.left,a&&c&&r<a.index&&r<c.index&&Pl(s,a,c)<0&&t.push([s.data,a.data,c.data])}),t},links:function(){return this.edges.filter(function(t){return t.right}).map(function(t){return{source:t.left.data,target:t.right.data}})},find:function(t,e,n){for(var r,i,o=this,a=o._found||0,s=o.cells.length;!(i=o.cells[a]);)if(++a>=s)return null;var u=t-i.site[0],l=e-i.site[1],c=u*u+l*l;do i=o.cells[r=a],a=null,i.halfedges.forEach(function(n){var r=o.edges[n],s=r.left;if(s!==i.site&&s||(s=r.right)){var u=t-s[0],l=e-s[1],f=u*u+l*l;f<c&&(c=f,a=s.index)}});while(null!==a);return o._found=r,null==n||c<=n*n?i.site:null}};var p_=function(){function t(t){return new Il(t.map(function(r,i){var o=[Math.round(e(r,i,t)/c_)*c_,Math.round(n(r,i,t)/c_)*c_];return o.index=i,o.data=r,o}),r)}var e=ol,n=al,r=null;return t.polygons=function(e){return t(e).polygons()},t.links=function(e){return t(e).links()},t.triangles=function(e){return t(e).triangles()},t.x=function(n){return arguments.length?(e="function"==typeof n?n:n_(+n),t):e},t.y=function(e){return arguments.length?(n="function"==typeof e?e:n_(+e),t):n},t.extent=function(e){return arguments.length?(r=null==e?null:[[+e[0][0],+e[0][1]],[+e[1][0],+e[1][1]]],t):r&&[[r[0][0],r[0][1]],[r[1][0],r[1][1]]]},t.size=function(e){return arguments.length?(r=null==e?null:[[0,0],[+e[0],+e[1]]],t):r&&[r[1][0]-r[0][0],r[1][1]-r[0][1]]},t},h_=function(t){return function(){return t}};Fl.prototype={constructor:Fl,scale:function(t){return 1===t?this:new Fl(this.k*t,this.x,this.y)},translate:function(t,e){return 0===t&0===e?this:new Fl(this.k,this.x+this.k*t,this.y+this.k*e)},apply:function(t){return[t[0]*this.k+this.x,t[1]*this.k+this.y]},applyX:function(t){return t*this.k+this.x},applyY:function(t){return t*this.k+this.y},invert:function(t){return[(t[0]-this.x)/this.k,(t[1]-this.y)/this.k]},invertX:function(t){return(t-this.x)/this.k},invertY:function(t){return(t-this.y)/this.k},rescaleX:function(t){return t.copy().domain(t.range().map(this.invertX,this).map(t.invert,t))},rescaleY:function(t){return t.copy().domain(t.range().map(this.invertY,this).map(t.invert,t))},toString:function(){return"translate("+this.x+","+this.y+") scale("+this.k+")"}};var d_=new Fl(1,0,0);ql.prototype=Fl.prototype;var v_=function(){t.event.preventDefault(),t.event.stopImmediatePropagation()},g_=function(){function e(t){t.on("wheel.zoom",l).on("mousedown.zoom",c).on("dblclick.zoom",f).on("touchstart.zoom",p).on("touchmove.zoom",h).on("touchend.zoom touchcancel.zoom",d).style("-webkit-tap-highlight-color","rgba(0,0,0,0)").property("__zoom",Hl)}function n(t,e){return e=Math.max($,Math.min(w,e)),e===t.k?t:new Fl(e,t.x,t.y)}function r(t,e,n){var r=e[0]-n[0]*t.k,i=e[1]-n[1]*t.k;return r===t.x&&i===t.y?t:new Fl(t.k,r,i)}function i(t,e){var n=t.invertX(e[0][0])-x,r=t.invertX(e[1][0])-_,i=t.invertY(e[0][1])-C,o=t.invertY(e[1][1])-k;return t.translate(r>n?(n+r)/2:Math.min(0,n)||Math.max(0,r),o>i?(i+o)/2:Math.min(0,i)||Math.max(0,o))}function o(t){return[(+t[0][0]+ +t[1][0])/2,(+t[0][1]+ +t[1][1])/2]}function a(t,e,n){t.on("start.zoom",function(){s(this,arguments).start()}).on("interrupt.zoom end.zoom",function(){s(this,arguments).end()}).tween("zoom",function(){var t=this,r=arguments,i=s(t,r),a=b.apply(t,r),u=n||o(a),l=Math.max(a[1][0]-a[0][0],a[1][1]-a[0][1]),c=t.__zoom,f="function"==typeof e?e.apply(t,r):e,p=T(c.invert(u).concat(l/c.k),f.invert(u).concat(l/f.k));return function(t){if(1===t)t=f;else{var e=p(t),n=l/e[2];t=new Fl(n,u[0]-e[0]*n,u[1]-e[1]*n)}i.zoom(null,t)}})}function s(t,e){for(var n,r=0,i=E.length;r<i;++r)if((n=E[r]).that===t)return n;return new u(t,e)}function u(t,e){this.that=t,this.args=e,this.index=-1,this.active=0,this.extent=b.apply(t,e)}function l(){function e(){o.wheel=null,o.end()}if(y.apply(this,arguments)){var o=s(this,arguments),a=this.__zoom,u=Math.max($,Math.min(w,a.k*Math.pow(2,-t.event.deltaY*(t.event.deltaMode?120:1)/500))),l=Jc(this);if(o.wheel)o.mouse[0][0]===l[0]&&o.mouse[0][1]===l[1]||(o.mouse[1]=a.invert(o.mouse[0]=l)),clearTimeout(o.wheel);else{if(a.k===u)return;o.mouse=[l,a.invert(l)],Sh(this),o.start()}v_(),o.wheel=setTimeout(e,O),o.zoom("mouse",i(r(n(a,u),o.mouse[0],o.mouse[1]),o.extent))}}function c(){function e(){if(v_(),!o.moved){var e=t.event.clientX-l,n=t.event.clientY-c;o.moved=e*e+n*n>N}o.zoom("mouse",i(r(o.that.__zoom,o.mouse[0]=Jc(o.that),o.mouse[1]),o.extent))}function n(){a.on("mousemove.zoom mouseup.zoom",null),bt(t.event.view,o.moved),v_(),o.end()}if(!m&&y.apply(this,arguments)){var o=s(this,arguments),a=Pf(t.event.view).on("mousemove.zoom",e,!0).on("mouseup.zoom",n,!0),u=Jc(this),l=t.event.clientX,c=t.event.clientY;qf(t.event.view),Vl(),o.mouse=[u,this.__zoom.invert(u)],Sh(this),o.start()}}function f(){if(y.apply(this,arguments)){var o=this.__zoom,s=Jc(this),u=o.invert(s),l=o.k*(t.event.shiftKey?.5:2),c=i(r(n(o,l),s,u),b.apply(this,arguments));v_(),S>0?Pf(this).transition().duration(S).call(a,c,s):Pf(this).call(e.transform,c)}}function p(){if(y.apply(this,arguments)){var e,n,r,i,o=s(this,arguments),a=t.event.changedTouches,u=a.length;for(Vl(),n=0;n<u;++n)r=a[n],i=If(this,a,r.identifier),i=[i,this.__zoom.invert(i),r.identifier],o.touch0?o.touch1||(o.touch1=i):(o.touch0=i,e=!0);return g&&(g=clearTimeout(g),!o.touch1)?(o.end(),i=Pf(this).on("dblclick.zoom"),void(i&&i.apply(this,arguments))):void(e&&(g=setTimeout(function(){g=null},A),Sh(this),o.start()))}}function h(){var e,o,a,u,l=s(this,arguments),c=t.event.changedTouches,f=c.length;for(v_(),g&&(g=clearTimeout(g)),e=0;e<f;++e)o=c[e],a=If(this,c,o.identifier),l.touch0&&l.touch0[2]===o.identifier?l.touch0[0]=a:l.touch1&&l.touch1[2]===o.identifier&&(l.touch1[0]=a);if(o=l.that.__zoom,l.touch1){var p=l.touch0[0],h=l.touch0[1],d=l.touch1[0],v=l.touch1[1],m=(m=d[0]-p[0])*m+(m=d[1]-p[1])*m,y=(y=v[0]-h[0])*y+(y=v[1]-h[1])*y;o=n(o,Math.sqrt(m/y)),a=[(p[0]+d[0])/2,(p[1]+d[1])/2],u=[(h[0]+v[0])/2,(h[1]+v[1])/2]}else{if(!l.touch0)return;a=l.touch0[0],u=l.touch0[1]}l.zoom("touch",i(r(o,a,u),l.extent))}function d(){var e,n,r=s(this,arguments),i=t.event.changedTouches,o=i.length;for(Vl(),m&&clearTimeout(m),m=setTimeout(function(){m=null},A),e=0;e<o;++e)n=i[e],r.touch0&&r.touch0[2]===n.identifier?delete r.touch0:r.touch1&&r.touch1[2]===n.identifier&&delete r.touch1;r.touch1&&!r.touch0&&(r.touch0=r.touch1,delete r.touch1),r.touch0?r.touch0[1]=this.__zoom.invert(r.touch0[0]):r.end()}var g,m,y=Ul,b=Ll,$=0,w=1/0,x=-w,_=w,C=x,k=_,S=250,T=Jp,E=[],D=v("start","zoom","end"),A=500,O=150,N=0;return e.transform=function(t,e){var n=t.selection?t.selection():t;n.property("__zoom",Hl),t!==n?a(t,e):n.interrupt().each(function(){s(this,arguments).start().zoom(null,"function"==typeof e?e.apply(this,arguments):e).end()})},e.scaleBy=function(t,n){e.scaleTo(t,function(){var t=this.__zoom.k,e="function"==typeof n?n.apply(this,arguments):n;return t*e})},e.scaleTo=function(t,a){e.transform(t,function(){var t=b.apply(this,arguments),e=this.__zoom,s=o(t),u=e.invert(s),l="function"==typeof a?a.apply(this,arguments):a;return i(r(n(e,l),s,u),t)})},e.translateBy=function(t,n,r){e.transform(t,function(){return i(this.__zoom.translate("function"==typeof n?n.apply(this,arguments):n,"function"==typeof r?r.apply(this,arguments):r),b.apply(this,arguments))})},u.prototype={start:function(){return 1===++this.active&&(this.index=E.push(this)-1,this.emit("start")),this},zoom:function(t,e){return this.mouse&&"mouse"!==t&&(this.mouse[1]=e.invert(this.mouse[0])),this.touch0&&"touch"!==t&&(this.touch0[1]=e.invert(this.touch0[0])),this.touch1&&"touch"!==t&&(this.touch1[1]=e.invert(this.touch1[0])),this.that.__zoom=e,this.emit("zoom"),this},end:function(){return 0===--this.active&&(E.splice(this.index,1),this.index=-1,this.emit("end")),this},emit:function(t){M(new jl(e,t,this.that.__zoom),D.apply,D,[t,this.that,this.args])}},e.filter=function(t){return arguments.length?(y="function"==typeof t?t:h_(!!t),e):y},e.extent=function(t){return arguments.length?(b="function"==typeof t?t:h_([[+t[0][0],+t[0][1]],[+t[1][0],+t[1][1]]]),e):b},e.scaleExtent=function(t){return arguments.length?($=+t[0],w=+t[1],e):[$,w]},e.translateExtent=function(t){return arguments.length?(x=+t[0][0],_=+t[1][0],C=+t[0][1],k=+t[1][1],e):[[x,C],[_,k]]},e.duration=function(t){return arguments.length?(S=+t,e):S},e.interpolate=function(t){return arguments.length?(T=t,e):T},e.on=function(){var t=D.on.apply(D,arguments);return t===D?e:t},e.clickDistance=function(t){return arguments.length?(N=(t=+t)*t,e):Math.sqrt(N)},e};t.version=Bl,t.bisect=Gl,t.bisectRight=Gl,t.bisectLeft=Xl,t.ascending=zl,t.bisector=Wl,t.cross=Jl,t.descending=Zl,t.deviation=ec,t.extent=nc,t.histogram=dc,t.thresholdFreedmanDiaconis=gc,t.thresholdScott=mc,t.thresholdSturges=hc,t.max=yc,t.mean=bc,t.median=$c,t.merge=wc,t.min=xc,t.pairs=Kl,t.permute=_c,t.quantile=vc,t.range=uc,t.scan=Cc,t.shuffle=kc,t.sum=Sc,t.ticks=pc,t.tickIncrement=r,t.tickStep=i,t.transpose=Tc,t.variance=tc,t.zip=Ec,t.axisTop=f,t.axisRight=p,t.axisBottom=h,t.axisLeft=d,t.brush=Vd,t.brushX=An,t.brushY=On,t.brushSelection=Dn,t.chord=Yd,t.ribbon=tv,t.nest=nv,t.set=Xn,t.map=Hn,t.keys=iv,t.values=ov,t.entries=av,t.color=St,t.rgb=Dt,t.hsl=Pt,t.lab=Ft,t.hcl=zt,t.cubehelix=Gt,t.dispatch=v,t.drag=Uf,t.dragDisable=qf,t.dragEnable=bt,t.dsvFormat=sv,t.csvParse=lv,t.csvParseRows=cv,t.csvFormat=fv,t.csvFormatRows=pv,t.tsvParse=dv,t.tsvParseRows=vv,t.tsvFormat=gv,t.tsvFormatRows=mv,t.easeLinear=rn,t.easeQuad=sn,t.easeQuadIn=on,t.easeQuadOut=an,t.easeQuadInOut=sn,t.easeCubic=cn,t.easeCubicIn=un,t.easeCubicOut=ln,t.easeCubicInOut=cn,t.easePoly=Zh,t.easePolyIn=Kh,t.easePolyOut=Jh,t.easePolyInOut=Zh,t.easeSin=hn,t.easeSinIn=fn,t.easeSinOut=pn,t.easeSinInOut=hn,t.easeExp=gn,t.easeExpIn=dn,t.easeExpOut=vn,t.easeExpInOut=gn,t.easeCircle=bn,t.easeCircleIn=mn,t.easeCircleOut=yn,t.easeCircleInOut=bn,t.easeBounce=wn,t.easeBounceIn=$n,t.easeBounceOut=wn,t.easeBounceInOut=xn,t.easeBack=dd,t.easeBackIn=pd,t.easeBackOut=hd,t.easeBackInOut=dd,t.easeElastic=bd,t.easeElasticIn=yd,t.easeElasticOut=bd,t.easeElasticInOut=$d,t.forceCenter=yv,t.forceCollide=Iv,t.forceLink=jv,t.forceManyBody=Uv,t.forceSimulation=Vv,t.forceX=Lv,t.forceY=Hv,t.formatDefaultLocale=vr,t.formatLocale=ng,t.formatSpecifier=hr,t.precisionFixed=rg,t.precisionPrefix=ig,t.precisionRound=og,t.geoArea=fm,t.geoBounds=dm,t.geoCentroid=gm,t.geoCircle=Am,t.geoClipExtent=Fm,t.geoContains=Xm,t.geoDistance=Wm,t.geoGraticule=Ci,t.geoGraticule10=ki,t.geoInterpolate=Km,t.geoLength=Hm,t.geoPath=_y,t.geoAlbers=Ny,t.geoAlbersUsa=Py,t.geoAzimuthalEqualArea=Iy,t.geoAzimuthalEqualAreaRaw=Ry,t.geoAzimuthalEquidistant=Fy,t.geoAzimuthalEquidistantRaw=jy,t.geoConicConformal=Vy,t.geoConicConformalRaw=ho,t.geoConicEqualArea=Oy,t.geoConicEqualAreaRaw=ao,t.geoConicEquidistant=Ly,t.geoConicEquidistantRaw=go,t.geoEquirectangular=Uy,t.geoEquirectangularRaw=vo,t.geoGnomonic=Hy,t.geoGnomonicRaw=mo,t.geoIdentity=By,t.geoProjection=no,t.geoProjectionMutator=ro,t.geoMercator=qy,t.geoMercatorRaw=co,t.geoOrthographic=zy,t.geoOrthographicRaw=bo,t.geoStereographic=Wy,t.geoStereographicRaw=$o,t.geoTransverseMercator=Yy,t.geoTransverseMercatorRaw=wo,t.geoRotation=Dm,t.geoStream=sm,t.geoTransform=Ty,t.cluster=Gy,t.hierarchy=Ao,t.pack=cb,t.packSiblings=ub,t.packEnclose=sb,t.partition=hb,t.stratify=mb,t.tree=yb,t.treemap=xb,t.treemapBinary=_b,t.treemapDice=pb,t.treemapSlice=bb,t.treemapSliceDice=Cb,t.treemapSquarify=wb,t.treemapResquarify=kb,t.interpolate=Vp,t.interpolateArray=Np,t.interpolateBasis=Tp,t.interpolateBasisClosed=Ep,t.interpolateDate=Pp,t.interpolateNumber=Rp,t.interpolateObject=Ip,t.interpolateRound=Up,t.interpolateString=qp,t.interpolateTransformCss=zp,t.interpolateTransformSvg=Wp,t.interpolateZoom=Jp,t.interpolateRgb=Dp,t.interpolateRgbBasis=Ap,t.interpolateRgbBasisClosed=Op,t.interpolateHsl=Zp,t.interpolateHslLong=Qp,t.interpolateLab=pe,t.interpolateHcl=th,t.interpolateHclLong=eh,t.interpolateCubehelix=nh,t.interpolateCubehelixLong=rh,t.quantize=ih,t.path=In,t.polygonArea=Sb,t.polygonCentroid=Tb;t.polygonHull=Mb;t.polygonContains=Db,t.polygonLength=Ab,t.quadtree=ir,t.queue=$a,t.randomUniform=Rb,t.randomNormal=Ib,t.randomLogNormal=jb,t.randomBates=qb,t.randomIrwinHall=Fb,t.randomExponential=Vb,t.request=Ub,t.html=Hb,t.json=Bb,t.text=zb,t.xml=Wb,t.csv=Gb,t.tsv=Xb,t.scaleBand=ka,t.scalePoint=Ta,t.scaleIdentity=ja,t.scaleLinear=Ia,t.scaleLog=Ba,t.scaleOrdinal=Ca,t.scaleImplicit=Qb,t.scalePow=Wa,t.scaleSqrt=Ya,t.scaleQuantile=Ga,t.scaleQuantize=Xa,t.scaleThreshold=Ka,t.scaleTime=kw,t.scaleUtc=Sw,t.schemeCategory10=Ew,t.schemeCategory20b=Mw,t.schemeCategory20c=Dw,t.schemeCategory20=Aw,t.interpolateCubehelixDefault=Ow,t.interpolateRainbow=Iw,t.interpolateWarm=Nw,t.interpolateCool=Pw,t.interpolateViridis=jw,t.interpolateMagma=Fw,t.interpolateInferno=qw,t.interpolatePlasma=Vw,t.scaleSequential=iu,t.creator=Vc,t.local=x,t.matcher=zc,t.mouse=Jc,t.namespace=qc,t.namespaces=Fc,t.select=Pf,t.selectAll=Rf,t.selection=mt,t.selector=Zc,t.selectorAll=tf,t.style=z,t.touch=If,t.touches=jf,t.window=$f,t.customEvent=M,t.arc=Qw,t.area=nx,t.line=ex,t.pie=ox,t.radialArea=ux,t.radialLine=sx,t.linkHorizontal=Su,t.linkVertical=Tu,t.linkRadial=Eu,t.symbol=Dx,t.symbols=Mx,t.symbolCircle=fx,t.symbolCross=px,t.symbolDiamond=vx,t.symbolSquare=wx,t.symbolStar=$x,t.symbolTriangle=_x,t.symbolWye=Ex,t.curveBasisClosed=Nx,t.curveBasisOpen=Px,t.curveBasis=Ox,t.curveBundle=Rx,t.curveCardinalClosed=jx,t.curveCardinalOpen=Fx,t.curveCardinal=Ix,t.curveCatmullRomClosed=Vx,t.curveCatmullRomOpen=Ux,t.curveCatmullRom=qx,t.curveLinearClosed=Lx,t.curveLinear=tx,t.curveMonotoneX=Ku,t.curveMonotoneY=Ju,t.curveNatural=Hx,t.curveStep=Bx,t.curveStepAfter=nl,t.curveStepBefore=el,t.stack=Yx,t.stackOffsetExpand=Gx,t.stackOffsetDiverging=Xx,t.stackOffsetNone=zx,t.stackOffsetSilhouette=Kx,t.stackOffsetWiggle=Jx,t.stackOrderAscending=Zx,t.stackOrderDescending=Qx,t.stackOrderInsideOut=t_,t.stackOrderNone=Wx,t.stackOrderReverse=e_,t.timeInterval=Ja,t.timeMillisecond=s$,t.timeMilliseconds=u$,t.utcMillisecond=s$,t.utcMilliseconds=u$,t.timeSecond=d$,t.timeSeconds=v$,t.utcSecond=d$,t.utcSeconds=v$,t.timeMinute=g$,t.timeMinutes=m$,t.timeHour=y$,t.timeHours=b$,t.timeDay=$$,t.timeDays=w$,t.timeWeek=x$,t.timeWeeks=M$,t.timeSunday=x$,t.timeSundays=M$,t.timeMonday=_$,t.timeMondays=D$,t.timeTuesday=C$,t.timeTuesdays=A$,t.timeWednesday=k$,t.timeWednesdays=O$,t.timeThursday=S$,t.timeThursdays=N$,t.timeFriday=T$,t.timeFridays=P$,t.timeSaturday=E$,t.timeSaturdays=R$,t.timeMonth=I$,t.timeMonths=j$,t.timeYear=F$,t.timeYears=q$,t.utcMinute=V$,t.utcMinutes=U$,t.utcHour=L$,t.utcHours=H$,t.utcDay=B$,t.utcDays=z$,t.utcWeek=W$,t.utcWeeks=Q$,t.utcSunday=W$,t.utcSundays=Q$,t.utcMonday=Y$,t.utcMondays=tw,t.utcTuesday=G$,t.utcTuesdays=ew,t.utcWednesday=X$,t.utcWednesdays=nw,t.utcThursday=K$,t.utcThursdays=rw,t.utcFriday=J$,t.utcFridays=iw,t.utcSaturday=Z$,t.utcSaturdays=ow,t.utcMonth=aw,t.utcMonths=sw,t.utcYear=uw,t.utcYears=cw,t.timeFormatDefaultLocale=Js,t.timeFormatLocale=rs,t.isoFormat=gw,t.isoParse=mw,t.now=ve,t.timer=ye,t.timerFlush=be,t.timeout=dh,t.interval=vh,t.transition=en,t.active=Cd,t.interrupt=Sh,t.voronoi=p_,t.zoom=g_,t.zoomTransform=ql,t.zoomIdentity=d_,Object.defineProperty(t,"__esModule",{value:!0})})}});